Package: randomForestSRC
Version: 3.7.0
Date: 2026-08-31
Title: Fast Unified Random Forests for Survival, Regression, and
        Classification (RF-SRC)
Authors@R: c(person("Hemant", "Ishwaran", email = "hemant.ishwaran@gmail.com", role = "aut"),
    person("Udaya B.", "Kogalur", email = "ubk@kogalur.com", role = c("aut", "cre")))
Author: Hemant Ishwaran [aut],
  Udaya B. Kogalur [aut, cre]
Maintainer: Udaya B. Kogalur <ubk@kogalur.com>
BugReports: https://github.com/kogalur/randomForestSRC/issues/
Depends: R (>= 4.3.0),
Imports: parallel, data.tree, DiagrammeR
Suggests: survival, pec, prodlim, mlbench, interp, caret, cluster, fst,
        data.table
SystemRequirements: OpenMP
Description: Fast OpenMP parallel computing of Breiman's random forests for univariate, multivariate, unsupervised, survival, competing risks, class imbalanced classification and quantile regression. New Mahalanobis splitting for correlated outcomes.  Extreme random forests and randomized splitting.  Suite of imputation methods for missing data.  Fast random forests using subsampling. Confidence regions and standard errors for variable importance. New improved holdout importance. Case-specific importance.  Minimal depth variable importance. Visualize trees on your Safari or Google Chrome browser. Anonymous random forests for data privacy.
License: GPL (>= 3)
URL: https://www.randomforestsrc.org/ https://ishwaran.org/
NeedsCompilation: yes
Packaged: 2026-08-31 13:25:34 UTC; kogalur
Repository: CRAN
Date/Publication: 2026-08-31 19:40:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-08-31 21:20:39 UTC; unix
Archs: randomForestSRC.so.dSYM

Package: randomForestSRC
Version: 3.6.2
Date: 2026-04-18
Title: Fast Unified Random Forests for Survival, Regression, and
        Classification (RF-SRC)
Authors@R: c(person("Hemant", "Ishwaran", email = "hemant.ishwaran@gmail.com", role = "aut"),
    person("Udaya B.", "Kogalur", email = "ubk@kogalur.com", role = c("aut", "cre")))
Author: Hemant Ishwaran [aut],
  Udaya B. Kogalur [aut, cre]
Maintainer: Udaya B. Kogalur <ubk@kogalur.com>
BugReports: https://github.com/kogalur/randomForestSRC/issues/
Depends: R (>= 4.3.0),
Imports: parallel, data.tree, DiagrammeR
Suggests: survival, pec, prodlim, mlbench, interp, caret, cluster, fst,
        data.table
SystemRequirements: OpenMP
Description: Fast OpenMP parallel computing of Breiman's random forests for univariate, multivariate, unsupervised, survival, competing risks, class imbalanced classification and quantile regression. New Mahalanobis splitting for correlated outcomes.  Extreme random forests and randomized splitting.  Suite of imputation methods for missing data.  Fast random forests using subsampling. Confidence regions and standard errors for variable importance. New improved holdout importance. Case-specific importance.  Minimal depth variable importance. Visualize trees on your Safari or Google Chrome browser. Anonymous random forests for data privacy.
License: GPL (>= 3)
URL: https://www.randomforestsrc.org/ https://ishwaran.org/
NeedsCompilation: yes
Packaged: 2026-04-18 12:58:13 UTC; kogalur
Repository: CRAN
Date/Publication: 2026-04-19 05:10:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 12:41:23 UTC; unix
Archs: randomForestSRC.so.dSYM

useDynLib(randomForestSRC, .registration = TRUE)

importFrom("data.tree", "FromDataFrameNetwork", "SetGraphStyle",
           "SetEdgeStyle", "SetNodeStyle", "Do", "GetDefaultTooltip")
importFrom("DiagrammeR", "render_graph", "create_graph",
           "create_node_df", "create_edge_df", "add_global_graph_attrs")
importFrom("parallel", mclapply)
importFrom("graphics", "abline", "axis", "box", "boxplot", "bxp",
           "legend", "lines", "matlines", "matplot", "mtext", "par", "plot",
           "plot.new", "plot.window", "points", "rug", "segments",
           "strwidth", "text", "title", "grconvertX", "grconvertY", "plot.default")
importFrom("stats", "as.dist", "as.formula", "cutree", "dlnorm", "formula",
           "hclust", "lowess", "median", "model.matrix", "na.omit",
           "optim", "pgamma", "plnorm", "pnorm", "predict", 
           "quantile", "qnorm", "runif", "sd", "supsmu", "var", "wilcox.test", "rnorm",
           "lm", "coef", "binom.test", "rmultinom", "setNames")
importFrom("utils", "installed.packages", "txtProgressBar", "setTxtProgressBar",
           "write.table", "tail", "object.size", "combn", "flush.console")
importFrom("grDevices", "gray")

export(extract.bootsample,
       extract.quantile,
       extract.subsample,
       get.auc,
       get.auct.survival,
       get.bayes.rule,
       get.brier.error,
       get.brier.survival,
       get.cindex,
       get.confusion,
       get.imbalanced.performance,
       get.imbalanced.optimize,
       get.logloss,
       get.misclass.error,
       get.mv.cserror,
       get.mv.csvimp,
       get.mv.error,
       get.mv.error.block,
       get.mv.formula,
       get.mv.predicted,
       get.mv.vimp,
       get.partial.plot.data,
       get.pr.auc,
       get.pr.curve,
       get.quantile,
       get.quantile.crps,
       get.quantile.stat,
       get.rfq.threshold,
       get.tree,
       get.tree.rfsrc,
       holdout.vimp,
       holdout.vimp.rfsrc,
       imbalanced,
       imbalanced.rfsrc,
       impute,       
       impute.rfsrc,
       impute.learn,
       impute.learn.rfsrc,
       impute.ood.rfsrc,
       impute.ood,
       load.impute.learn,
       load.impute.learn.rfsrc,
       max.subtree,
       max.subtree.rfsrc,
       partial,
       partial.rfsrc,
       plotBrierAUC,
       plot.competing.risk,
       plot.competing.risk.rfsrc,
       plot.quantreg,
       plot.quantreg.rfsrc,
       plot.rfsrc,
       plot.subsample,
       plot.subsample.rfsrc,       
       plot.survival,
       plot.survival.rfsrc,
       plot.variable,
       plot.variable.rfsrc,       
       predict.rfsrc,
       print.rfsrc,
       print.bootsample,
       print.bootsample.rfsrc,
       print.imbalanced.performance,
       print.subsample,
       print.subsample.rfsrc,
       quantreg,
       quantreg.rfsrc,
       rfsrc,
       rfsrc.anonymous,
       rfsrc.cart,
       rfsrc.fast,
       rfsrc.news,
       save.impute.learn,
       save.impute.learn.rfsrc,
       sid.perf.metric,
       sidClustering,
       sidClustering.rfsrc,
       subsample,
       subsample.rfsrc,
       tune,
       tune.rfsrc,
       tune.nodesize,
       tune.nodesize.rfsrc,
       vimp,
       vimp.rfsrc)




export(fast.save)
export(fast.load)
export(fast.save.list)
export(fast.load.list)
export(lsos)



S3method(max, subtree)
S3method(max, subtree.rfsrc)

S3method(plot, competing.risk)
S3method(plot, competing.risk.rfsrc)
S3method(plot, quantreg)
S3method(plot, quantreg.rfsrc)
S3method(plot, rfsrc)
S3method(plot, subsample)
S3method(plot, subsample.rfsrc)
S3method(plot, survival)
S3method(plot, survival.rfsrc)
S3method(plot, variable)
S3method(plot, variable.rfsrc)

S3method(predict, rfsrc)
S3method(predict, impute.learn)
S3method(predict, impute.learn.rfsrc)


S3method(print, rfsrc)
S3method(print, bootsample)
S3method(print, bootsample.rfsrc)
S3method(print, imbalanced.performance)
S3method(print, impute.learn)
S3method(print, impute.learn.rfsrc)
S3method(print, subsample)
S3method(print, subsample.rfsrc)

coerce.multivariate <- function(x, outcome.target) {
  ## Warning:  This functon assumes that get.univariate.target has been called first, to
  ## verify the coherency of the target.  This means that the target exists in the forest object, and that
  ## it contains outcome statistics.
  ## If this is a multivarate family, we coerce the object, based on outcome.target
  ## into a univaritate regression or classification object.
  x$univariate <- TRUE
  if (x$family == "regr+" | x$family == "class+" | x$family == "mix+") {
    ## coerce the mulitvariate object into a univariate object
    x.coerced <- unlist(list(x$classOutput, x$regrOutput), recursive = FALSE)[[outcome.target]]
    x$univariate <- FALSE
    x$yvar <- x$yvar[, outcome.target]
    ## test for factors - ordered factors are treated as factors!
    if (is.factor(x$yvar) || is.ordered(x$yvar)) {
      x$family <- "class"
    }
    else {
      x$family <- "regr"
    }
    ## make various assignments to the coerced object.
    x$predicted <- x.coerced$predicted
    x$predicted.oob <- x.coerced$predicted.oob
    x$class <- x.coerced$class
    x$class.oob <- x.coerced$class.oob
    x$err.rate <- x.coerced$err.rate
    x$err.block.rate <- x.coerced$err.block.rate
    x$importance <- x.coerced$importance
    x$yvar.names <- outcome.target
    x$cse.num <- x.coerced$cse.num
    x$cse.den <- x.coerced$cse.den
    x$csv.num <- x.coerced$csv.num
    x$csv.den <- x.coerced$csv.den
  }
  x$outcome.target <- outcome.target
  x
}
get.mv.cserror <- function(obj, standardize = FALSE) {
  ## acquire y variable names
  ynms <- obj$yvar.names
  ## does not apply to survival or competing risk
  if (obj$family == "surv" ||  obj$family == "surv-CR") {
    return(NULL)
  }
  ## loop over the yvar names acquring the csv if it is present
  err <- lapply(ynms, function(nn) {
    ## coerce to a univariate object for uniformity
    o.coerce <- coerce.multivariate(obj, nn)
    e.num <- o.coerce$cse.num
    e.den <- o.coerce$cse.den
    err <- NULL
    if (!is.null(e.num)) {
      ## classification
      if (!(o.coerce$family == "regr" || o.coerce$family == "regr+")) {
        err <- e.num / e.den
      }
      ## regression: standardization can be applied
      else {
        yvar <- 1
        if (standardize) {
          yvar <- var(o.coerce$yvar, na.rm = TRUE)
          if (is.na(yvar) || yvar == 0) {
            yvar <- 1
          }
        }
        err <- e.num / (e.den * yvar)
      }
    }
    ## return
    err
  })
  ## if the first entry is NULL make the entire list NULL
  if (is.null(err[[1]])) {
    err <- NULL
  }  
  ## add ynames - for multivariate objects only
  if (!is.null(err) && length(err)>1) {
    names(err) <- ynms
  }
  if (!is.null(err) && length(err)==1) {
    err <- err[[1]]
  }
  ### return the goodies
  err 
}
## get case-specific error - this is not the same as
## Loss(predicted.oob,y)
get.mv.cserror <- function(obj, standardize = FALSE) {
  ## acquire y variable names
  ynms <- obj$yvar.names
  ## does not apply to survival or competing risk
  if (obj$family == "surv" ||  obj$family == "surv-CR") {
    return(NULL)
  }
  ## loop over the yvar names acquring the csv if it is present
  err <- lapply(ynms, function(nn) {
    ## coerce to a univariate object for uniformity
    o.coerce <- coerce.multivariate(obj, nn)
    e.num <- o.coerce$cse.num
    e.den <- o.coerce$cse.den
    err <- NULL
    if (!is.null(e.num)) {
      ## classification
      if (!(o.coerce$family == "regr" || o.coerce$family == "regr+")) {
        err <- e.num / e.den
      }
      ## regression: standardization can be applied
      else {
        yvar <- 1
        if (standardize) {
          yvar <- var(o.coerce$yvar, na.rm = TRUE)
          if (is.na(yvar) || yvar == 0) {
            yvar <- 1
          }
        }
        err <- e.num / (e.den * yvar)
      }
    }
    ## return
    err
  })
  ## if the first entry is NULL make the entire list NULL
  if (is.null(err[[1]])) {
    err <- NULL
  }  
  ## add ynames - for multivariate objects only
  if (!is.null(err) && length(err)>1) {
    names(err) <- ynms
  }
  if (!is.null(err) && length(err)==1) {
    err <- err[[1]]
  }
  ### return the goodies
  err 
}
## get case-specific VIMP
get.mv.csvimp <- function(obj, standardize = FALSE) {
  ## acquire y variable names
  ynms <- obj$yvar.names
  ## does not apply to survival or competing risk
  if (obj$family == "surv" ||  obj$family == "surv-CR") {
    return(NULL)
  }
  ## loop over the yvar names acquring the csv if it is present
  vmp <- lapply(ynms, function(nn) {
    ## coerce to a univariate object for uniformity
    o.coerce <- coerce.multivariate(obj, nn)
    v.num <- o.coerce$csv.num
    v.den <- o.coerce$csv.den
    v <- NULL
    if (!is.null(v.num)) {
      ## classification
      if (o.coerce$family != "regr") {
        v <- v.num / v.den
      }
      ## regression: standardization can be applied
      else {
        yvar <- 1
        if (standardize) {
          yvar <- var(o.coerce$yvar, na.rm = TRUE)
          if (is.na(yvar) || yvar == 0) {
            yvar <- 1
          }
        }
        v <- v.num / (v.den * yvar)
      }
      ## pretty the vimp up
      if (is.null(dim(v))) {
        v <- cbind(v)
      }
      ## make allowance for joint vimp
      if (length(o.coerce$importance) == 1) {
        v <- v[, 1, drop = FALSE]
      }
      colnames(v) <- names(o.coerce$importance)
    }
    ## return
    v
  })
  ## if the first entry is NULL make the entire list NULL
  if (is.null(vmp[[1]])) {
    vmp <- NULL
  }  
  ## add ynames - for multivariate objects only
  if (!is.null(vmp) && length(vmp)>1) {
    names(vmp) <- ynms
  }
  if (!is.null(vmp) && length(vmp)==1) {
    vmp <- vmp[[1]]
  }
  ### return the goodies
  vmp 
}
get.mv.error <- function(obj, standardize = FALSE, pretty = TRUE, block = FALSE) {
  ## acquire yvar names - don't want "censoring" for surv and surv-CR
  nms <- NULL
  ynms <- obj$yvar.names
  if (obj$family == "surv" ||  obj$family == "surv-CR") {
    ynms <- ynms[1]
  }
  ## pretty option is not allowed for blocked error
  if (block) {
    pretty <- FALSE
  }
  ## loop over the yvar names acquring error if it is present
  err <- lapply(ynms, function(nn) {
    o.coerce <- coerce.multivariate(obj, nn)
    if (!block) {
      er <- o.coerce$err.rate
    }
    else {
      er <- o.coerce$err.block.rate
    }
    if (!is.null(er)) {
      if (o.coerce$family != "regr" || !standardize) {
        if (pretty && o.coerce$family == "class") {##pretty can only pull first vimp column "all"
            er <- utils::tail(cbind(er)[, 1], 1)
        }
        ## now returns survival error rates more coherently
        else {
          if (!block) {
            er <- utils::tail(er, 1)
            rownames(er) <- NULL
          }
        }
      }
      else {##standardized VIMP only applies to regression
        if (!block) {
          er <- utils::tail(er, 1) / var(o.coerce$yvar, na.rm = TRUE)
        }
        else {
          er <- er / var(o.coerce$yvar, na.rm = TRUE)
        }
      }
    }
    if (is.null(dim(er))) {
      nms <<- c(nms, paste(nn))
    }
    else {
      nms <<- c(nms, colnames(er))
    }
    er
  })
  ## if the first entry is NULL make the entire list NULL
  if (is.null(err[[1]])) {
    err <- NULL
  }  
  ## return as a vector for convenient interpretation?
  ## vector reduces information for classification to "all"
  if (!is.null(err)) {
    if (pretty) {
      err <- unlist(err)
      names(err) <- nms
    }
    else {
      names(err) <- ynms
    }
  }
  ### return the goodies
  err
}
get.mv.error.block <- function(obj, standardize = FALSE) {
  ## use get.mv.error but request blocked error rate
  get.mv.error(obj, standardize = standardize, block = TRUE)
}
get.mv.formula <- function(ynames) {
  ynames <- sapply(ynames, as.name) ##adds backticks to handle non-standard ynames
  as.formula(paste("Multivar(", paste(ynames, collapse = ","),paste(") ~ ."), sep = ""))
}
get.mv.predicted <- function(obj, oob = TRUE) {
  ## loop over the yvar names acquring the vimp if it is present
  ## acquire yvar names - don't want "censoring" for surv and surv-CR
  nms <- NULL
  ynms <- obj$yvar.names
  if (obj$family == "surv" ||  obj$family == "surv-CR") {
    ynms <- ynms[1]
  }
  pred <- do.call(cbind, lapply(ynms, function(nn) {
    o.coerce <- coerce.multivariate(obj, nn)
    if (o.coerce$family == "class" || o.coerce$family == "surv-CR") {
      ## ensembles can now be either oob, inbag, or all 05/06/2018
      if (!is.null(o.coerce$predicted.oob)) {
        nms <<- c(nms, paste(nn, ".", colnames(o.coerce$predicted.oob), sep = ""))
      }
      else {
        nms <<- c(nms, paste(nn, ".", colnames(o.coerce$predicted), sep = ""))
      }
    }
    else {
      nms <<- c(nms, paste(nn))
    }
    ## user may request OOB when it doesn't exist: noted 04/04/2018
    if (oob && !is.null(o.coerce$predicted.oob)) {
      o.coerce$predicted.oob
    }
    else {
      o.coerce$predicted
    }
  }))
  ## pretty names
  colnames(pred) <- nms
  pred
}
get.mv.vimp <- function(obj, standardize = FALSE, pretty = TRUE) {
  ## acquire yvar names - don't want "censoring" for surv and surv-CR
  ynms <- obj$yvar.names
  if (obj$family == "surv" ||  obj$family == "surv-CR") {
    ynms <- ynms[1]
  }
  ## loop over the yvar names acquring the vimp if it is present
  vmp <- lapply(ynms, function(nn) {
    o.coerce <- coerce.multivariate(obj, nn)
    v <- o.coerce$importance
    if (!is.null(v)) {
      if (o.coerce$family != "regr" || !standardize) {
        if (pretty && o.coerce$family == "class") {##pretty can only pull first vimp column "all"
          v <- v[, 1]
        }
      }
      else {##standardized VIMP only applies to regression
        v <- v / var(o.coerce$yvar, na.rm = TRUE)
      }
      if (is.null(dim(v))) {
        v <- cbind(v)
        colnames(v) <- nn
      }
    }
    v
  })
  ## if the first entry is NULL make the entire list NULL
  if (is.null(vmp[[1]])) {
    vmp <- NULL
  }  
  ## return as a matrix for convenient interpretation?
  ## matrix reduces information for classification to "all"
  if (!is.null(vmp)) {
    if (pretty) {
      vmp <- do.call(cbind, vmp)
    }
    else {
      names(vmp) <- ynms
    }
  }
  ### return the goodies
  vmp 
}
get.univariate.target <- function(x, outcome.target = NULL) {
  ## This function takes a grow, grow-equivalent, or predict object and returns a single coherent target.
  ## That is, if no target has been specified, the first regression outcome with statistics is chosen.
  ## If no regression outcome exists, the first classification outcome with statistics is chosen.
  ## If the target is specified, the object is verified to contain the target outcome statistics
  ## for that y-var.  If none exist, the function will error.
  if (x$family == "regr+" | x$family == "class+" | x$family == "mix+") {
    if (is.null(outcome.target)) {
      ## Check the y-vars against regression and then classification.
      ## We choose the "first" variable, favoring regression, then
      ## classification.
      target <- match(c("regrOutput", "classOutput"), names(x))
      target <- target[!is.na(target)]
      if(length(target) > 0) {
        do.break <- FALSE
        for (i in target) {
          for (j in 1:length(x[[i]])) {
            if (length(x[[i]][[j]]) > 0) {
              ## This is a non-null output.
              outcome.target <- names(x[[i]][j])
              ## Exit the loop.
              do.break <- TRUE
              break
            }
          }
          if (do.break == TRUE) {
            break
          }
        }
      }
      else {
        ## Something would have to be seriously wrong for this to happen.
        stop("No outcomes found in object.  Please contact technical support.")
      }
    }
    else {
      ## Check that one and only one target has been specified.
      if (sum(is.element(outcome.target, x$yvar.names)) != 1) {
        stop("Specified target was not found or too many target outcomes were supplied (only one is allowed).")
      }
      ## A target outcome has been specified.  Verify that it contains outcome statistics.
      target <- match(c("regrOutput", "classOutput"), names(x))
      target <- target[!is.na(target)]
      found = FALSE
      if(length(target) > 0) {
        do.break <- FALSE
        for (i in target) {
          for (j in 1:length(x[[i]])) {
            if (length(x[[i]][[j]]) > 0) {
              ## This is a non-null output.
              if (outcome.target == names(x[[i]][j])) {
                found = TRUE
                ## Exit the loop.
                do.break <- TRUE
                break
              }
            }
          }
          if (do.break == TRUE) {
            break
          }
        }     
      }
      if (!found) {
        stop("Target outcome has been correctly specified but it did not contain outcome statistics.")
      }
    }
  }
  ## This function will return NULL if the function is not
  ## multivariate.  Otherwise, the outcome and its associated statistics is
  ## guaranteed to exist in the object.
  outcome.target
}
####################################################################
##
##  SVD-generalized inverse UTILITY FUNCTIONS
##
####################################################################
## generalized inverse of matrix X (typically X=covariance matrix)
ginverse <- function(X, tol = sqrt(.Machine$double.eps), test = FALSE) {
  X <- as.matrix(X)
  if (nrow(X) >= ncol(X)) {
    Xinv <- ginverse.workhorse(X, tol)
  }
  else {
    Xinv <- t(ginverse.workhorse(t(X), tol))
  }
  if (test) {
    p1 <- summary(c(X %*% Xinv %*% X - X))
    p2 <- summary(c(Xinv %*% X %*% Xinv - Xinv))
    p3 <- summary(c(X %*% Xinv - t(X %*% Xinv)))
    cat("test 3 properties of generalized inverse:\n")
    print(p1);print(p2);print(p3)
  }
  Xinv
}
## generalized inverse workhorse
ginverse.workhorse <- function (X, tol) {
  Xsvd <- svd(X)
  Positive <- Xsvd$d > max(tol * Xsvd$d[1L], 0)
  if (all(Positive)) {
    Xsvd$v %*% (1/Xsvd$d * t(Xsvd$u))
  }
  else if (!any(Positive)) { 
    array(0, dim(X)[2L:1L])
  }
  else {
    Xsvd$v[, Positive, drop = FALSE] %*% ((1/Xsvd$d[Positive]) * 
                       t(Xsvd$u[, Positive, drop = FALSE]))
  }
}
## minimum least squares via generalized inverse 
ginverse.mlse <- function(f, data) {
  ## acquire y and x
  ## convert x to a data matrix
  ynms <- all.vars(as.formula(f))[1]
  x <- data.matrix(data[, colnames(data) != ynms, drop = FALSE])
  y <- data[, ynms]
  ## mlse
  o.mlse <- ginverse.mlse.workhorse(y, x)
  ## ols
  o.lm <- lm(y~., data.frame(y = y, x))
  ## add mlse to ols
  tbl <- summary(o.lm)$coef
  ## for p>n the OLS may fail ... break this up into two cases
  if (nrow(tbl) == length(o.mlse$coef)) {
    cnms <- colnames(tbl); cnms[1] <- "ols"
    tbl <- data.frame(tbl)
    colnames(tbl) <- cnms
    tbl$mls <- o.mlse$coef
  }
  else {## ... looks like OLS failed
    tbl <- data.frame(ols = coef(o.lm),
                      mls = o.mlse$coef)
  }
  ## return the goodies
  print(tbl)
  invisible(list(summary = tbl, ols = o.lm, mls = o.mlse, y = y, x = x))
}
ginverse.mlse.workhorse <- function(y, x) {
  A <- as.matrix(data.matrix(cbind(1, x)))
  n <- nrow(A)
  p <- ncol(x)
  bhat <- c(ginverse(A) %*% y)
  yhat <- c(A %*% c(bhat))
  ssq <- sum((y - yhat)^2, na.rm = TRUE)
  ss0 <- sum((y - mean(y, na.rm = TRUE))^2, na.rm = TRUE)
  rsq <- (1 - ssq / ss0)
  rsq.adj <- NA
  if ((p + 1) < n) {
    rsq.adj <- (1 - (n - 1) / (n - p - 1) * (ssq / ss0))
  }
  names(bhat) <- c("Int", colnames(x))
  list(coef = bhat, fitted.values = yhat, rsq = rsq, rsq.adj = rsq.adj)
}
## generalized inverse of A=(X^T X) using svd
ginverseA <- function (X, ridge = 0, tol = sqrt(.Machine$double.eps)) {
  if (!is.matrix(X)) {
    X <- as.matrix(X)
  }
  ## we can sometimes expect NA or infinite values in the data
  ## for example repeated recursive splitting can yield features with no data
  ## clean this up by setting colums where this occurs to a constant - zero will do
  na.pt <- apply(is.na(X), 2, any)
  if (sum(na.pt) > 0) {
    X[, na.pt] <- 0
  }
  inf.pt <- apply(is.infinite(X), 2, any)
  if (sum(inf.pt) > 0) {
    X[, inf.pt] <- 0
  }
  ## calculate A
  A <- t(X) %*% X
  ## ridge stabilization
  if (ridge > 0) {
    A <- A + diag(ridge, ncol(A))
  }
  ginverse(A, tol)
}

sidClustering.rfsrc <- function(data,
                                method = "sid",
                                k = NULL,
                                reduce = TRUE,
                                ntree = 500,
                                ntree.reduce = function(p, vtry){100 * p / vtry},
                                fast = FALSE,
                                x.no.sid = NULL,
                                use.sid.for.x = TRUE,
                                x.only = NULL, y.only = NULL,
                                dist.sharpen = TRUE,
                                ...)
{
  ##------------------------------------------------------------------
  ##
  ## preliminary processing
  ##
  ##------------------------------------------------------------------
  ## make sure we have a data frame
  if (!is.data.frame(data)) {
    data <- data.frame(data)
  }
  ## decide which method to use
  method <- match.arg(method, c("sid",
                                "unsupv", 
                                "shi-horvath",  "sh",  "SH",  "sh1", "SH1", "sh-1", "SH-1",
                                "shi-horvath2", "sh2", "SH2", "sh-2", "SH-2"))
  if (grepl("sid", method)) {
    methodClass <- 1
  }
  else if (grepl("sh", method)) {
    methodClass <- 2
  }
  else if (grepl("un", method)) {
    methodClass <- 3
  }
  else {
    stop("method not set correctly:", method)
  }
  ## set k if user has not provided a value
  if (is.null(k)) {
    k <- 1:max(1, min(10, nrow(data) - 1))
  }
  else {
    k <- k[k<=nrow(data)]
    k <- k[k != 0]
    if (length(k) == 0) {
      k <- 1:max(1, min(10, nrow(data) - 1))
    }
  }
  ## determine the grow interface - rfsrc or rfsrc.fast?
  if (!fast) {
    rfsrc.grow <- "rfsrc"
  }
  else {
    rfsrc.grow <- "rfsrc.fast"
  }
  ##------------------------------------------------------------------
  ##
  ## run SH-1 to acquire cheap but effective VIMP for dimension reduction
  ##
  ## here's the rationale for Zcut and eps ...
  ## Let I=importance and s=standard error
  ## The left edge of the CI for a strong variable satisfies:
  ##
  ## I - s Zcut > eps' > 0  (eps' quantifies how far left CI is away from 0)
  ##
  ## With some rearrangment:
  ##
  ## I / (eps'/Zcut + s) > Zcut
  ##
  ## Let eps=eps'/Zcut, we have
  ##
  ## I / (eps + s) > Zcut
  ##
  ## As illustration, eps = .25, Zcut = 2 ---> eps'=.5
  ##
  ##------------------------------------------------------------------
  if (methodClass == 1 && reduce && nrow(data) > 2) {
    ## pull unnamed parameters to match against rfsrc formal arguments
    sv.f <- as.formula("classes ~ .")
    dots <- list(...)
    dots$importance <- TRUE
    dots$splitrule <- NULL ##needed to allow Mahalanobis splitting for sidClustering
    rfnames <- names(formals(rfsrc))
    rfnames <- rfnames[rfnames != "distance" & rfnames != "proximity"]
    ## remove any column with less than two unique values
    void.var <- sapply(data, function(x){length(unique(x, na.rm = TRUE)) < 2})
    if (sum(void.var) > 0) {
      data[, which(void.var)] <- NULL
    }
    ## ... the remaining part of the code block works if there is only 1 variable left
    ## ... there needs to be enough unique values always 
    ## run the supervised classifier and acquire VIMP
    dots.sv <- dots
    dots.sv$mtry <- NULL
    o <- do.call(holdout.vimp, c(list(formula = sv.f, data = make.sh(data, 1), ntree = ntree.reduce),
                     dots.sv[names(dots.sv) %in% rfnames]))
    ## identify significant variables
    Z <- o$importance[, 1]
    ## we need at least TWO (2) variables to pass filtering (otherwise there will be no SID)
    if (sum(Z > 0) > 1) {
      xvar.filter <- names(Z)[Z > 0]
    }
    else {
      ## there were less than two variables
      ## careful - it's possible there's only one variable
      xvar.filter <- names(Z)[order(Z, decreasing = T)[1:min(length(Z), 2)]]
    }
    ## subset the data
    data <- data[, xvar.filter, drop = FALSE]
  }
  ##-----------------------------------------------------------
  ##
  ## sid clustering method
  ##
  ##-----------------------------------------------------------
  if (methodClass == 1) {
    ## terminate if there is one case or less than two variables
    terminate <- FALSE
    ## did the user provide protected x-values - convert to data frame if so
    if (!is.null(x.no.sid) && !is.character(x.no.sid)) {
      x.no.sid <- data.frame(x.no.sid)
    }
    ## did the user provide protected x-values as a character vector?
    if (!is.null(x.no.sid) && is.character(x.no.sid)) {
      drops <- (names(data) %in% x.no.sid)
      if (sum(drops) > 0) {##check that x.no.sid is still viable - reduce (filtering) could remove it
        x.no.sid <- data[, drops, drop = FALSE]
        if (sum(!drops) > 0) {
          data <- data[ , !drops, drop = FALSE]
        }
        else {
          terminate <- TRUE
        }
      }
      else {
        x.no.sid <- NULL
      }
    }
    ## last chance to fail 
    if (terminate || ncol(data) < 2 || nrow(data) < 2) {
      rO <- list(clustering = rep(1, nrow(data)), rf = NULL, dist = NULL, sid = NULL, outcome.names = NULL)
      class(rO) <- c("rfsrc-failed", "sidClustering", method)
      return(invisible(rO))
    }
    ## sidIFY the data
    sid.o <- make.sid(data)
    rf.dat <- data.frame(sid.o$y, sid.o$x)
    if (use.sid.for.x) {
      outcome.names <- colnames(sid.o$y)
    }
    else {
      outcome.names <- colnames(sid.o$x)
    }
    ## if there are protected x-values, add them now to the data
    if (!is.null(x.no.sid) && is.data.frame(x.no.sid) && sum(names(data) %in% names(x.no.sid)) == 0) {
      rf.dat <- data.frame(rf.dat, x.no.sid)
      outcome.names <- c(outcome.names, colnames(x.no.sid))
    }
    ##remove the original data
    rm(data)
    ## make the mv formula
    mv.f <- as.formula(paste("Multivar(", paste(outcome.names, collapse = ","), paste(") ~ ."), sep = ""))
    ## pull unnamed parameters to match against rfsrc formal arguments
    ### extract the distance (oob by default)
    dots <- list(...)
    if (is.null(dots$distance)) {
      dots$distance <- "oob"
    }
    rfnames <- names(formals(rfsrc))
    ## mvRF call - in extreme cases (n=1,2 sample size) this can fail
    rf <- tryCatch({do.call(rfsrc.grow,
            c(list(formula = mv.f, data = rf.dat, ntree = ntree),
             dots[names(dots) %in% rfnames]))}, error=function(ex){NULL})
    if (is.null(rf)) {## call failed
      rO <- list(clustering = rep(1, nrow(rf.dat)), rf = NULL, dist = NULL, sid = NULL, outcome.names = NULL)
      class(rO) <- c("rfsrc-failed", "sidClustering", method)
      return(invisible(rO))
    }
    d <- rf$distance
    ## euclidean distance sharpening (depends on distance.sharpening option - this is slow!!!!)
    if (dist.sharpen) {
      d <- distance(d)
    }
    ## OOB distance may contain NA's - set these to 1
    d[is.na(d)] <- 1
  }##end RF-SID
  ##-----------------------------------------------------------
  ##
  ## Shi-Horvath (mode 1 and 2) clustering
  ## converts problem to classification using artifical class data
  ##
  ##-----------------------------------------------------------
  if (methodClass == 2) {
    ## set NULL values
    sid.o <- outcome.names <- NULL
    ## make SH synthetic class data - mode 1 or 2
    if (grepl("2", method)) {
      mode <- 2
      method <- "sh2"
    }
    else {
      mode <- 1
      method <- "sh1"
    }
    hv.dat <- make.sh(data, mode)
    ## make the supervised formula
    sv.f <- as.formula("classes ~ .")
    ## pull unnamed parameters to match against rfsrc formal arguments
    ## extract the distance (oob by default)
    dots <- list(...)
    if (is.null(dots$proximity)) {
      dots$proximity <- "oob"
    }
    rfnames <- names(formals(rfsrc))
    ## run the supervised classifier and extract the proximity distance
    rf <- do.call(rfsrc.grow, c(list(formula = sv.f, data = hv.dat, ntree = ntree),
            dots[names(dots) %in% rfnames]))
    d <- 1 - rf$proximity[rf$yvar == 1, rf$yvar == 1]
    ## OOB proximity may contain NA's - set these to 1
    d[is.na(d)] <- 1
  }
  ##-----------------------------------------------------------
  ##
  ## plain vanilla unsupervised splitting
  ## multivariate splitting x against x
  ##
  ##-----------------------------------------------------------
  if (methodClass == 3) {
    ## set NULL values
    sid.o <- outcome.names <- NULL
    ## scale the y-values
    y.dat <- scale(data.matrix(data))
    ## make mv formula 
    ## duplicate the data
    ## watch out for x.only and y.only requests
    outcome.values <- colnames(y.dat[, setdiff(colnames(y.dat), x.only), drop = FALSE])
    mv.f <- get.mv.formula(outcome.values)
    rf.data <- data.frame(y.dat, data[, setdiff(colnames(data), y.only), drop = FALSE])
    ## remove data
    rm(data)
    ## pull unnamed parameters to match against rfsrc formal arguments
    ### extract the distance (oob by default)
    dots <- list(...)
    if (is.null(dots$distance)) {
      dots$distance <- "oob"
    }
    rfnames <- names(formals(rfsrc))
    ## mvRF call - in extreme cases (n=1,2 sample size) this can fail
    rf <- tryCatch({do.call(rfsrc.grow,
                            c(list(formula = mv.f, data = rf.data, ntree = ntree), 
                              dots[names(dots) %in% rfnames]))}, error=function(ex){NULL})
    if (is.null(rf)) {## call failed
      rO <- list(clustering = rep(1, nrow(rf.data)), rf = NULL, dist = NULL)
      #class(rO) <- c("rfsrc-failed", "sgreedy.unsupv")
      return(invisible(rO))
    }
    d <- rf$distance
    ## euclidean distance sharpening (depends on distance.sharpening option - this is slow!!!!)
    if (dist.sharpen) {
      d <- distance(d)
    }
    ## OOB distance may contain NA's - set these to 1
    d[is.na(d)] <- 1
  }    
  ##-----------------------------------------------------------
  ##
  ## hierarchical clustering
  ##
  ##-----------------------------------------------------------
  clus <- cutree(hclust(as.dist(d)), k = k)
  ##-----------------------------------------------------------
  ##
  ## return the goodies
  ##
  ##-----------------------------------------------------------
  rO <- list(clustering = clus, rf = rf, dist = d, sid = sid.o, outcome.names = outcome.names)
  class(rO) <- c("rfsrc", "sidClustering", method)
  invisible(rO)
}
sidClustering <- sidClustering.rfsrc

impute.rfsrc <- function(formula, data,
                         ntree = 100, nodesize = 1, nsplit = 10,
                         nimpute = 2, fast = FALSE, blocks,                         
                         mf.q, max.iter = 10, eps = 0.01,
                         ytry = NULL, always.use = NULL, verbose = TRUE,
                         full.sweep = FALSE,      ## optional final sweep
                         restore.integer = TRUE,
                         ...)
{
  ##--------------------------------------------------------------
  ##
  ## preliminary checks to make sure impute makes sense
  ##
  ##--------------------------------------------------------------
  ## terminate if there is no data
  if (missing(data)) {
    stop("data is missing")
  }
  ## must be a data frame
  if (!is.data.frame(data)) {
    stop("'data' must be a data frame.", call. = FALSE)
  }
  ## identify the missing data
  ## if none/all: return the data
  which.na <- is.na(data)
  if (!any(which.na) || all(which.na)) {
    return(invisible(data))
  }
  ##--------------------------------------------------------------
  ##
  ## extract additional options specified by user
  ## we lock this down - here are the options we allow
  ##
  ##--------------------------------------------------------------
  ## list of allowed parameters
  rfnames <- c("mtry",
               "splitrule",
               "bootstrap",
               "sampsize",
               "samptype")
  ## get all user-specified dots
  dots.all <- list(...)
  ## get the permissible hidden options for main impute
  dots <- dots.all[names(dots.all) %in% rfnames]
  ## full sweep options (own list under 'full.sweep.options')
  fs.opts <- dots.all[["full.sweep.options"]]
  if (is.null(fs.opts)) fs.opts <- list()
  fs.dots <- fs.opts[names(fs.opts) %in% rfnames]
  fs.ntree    <- if (!is.null(fs.opts$ntree))    fs.opts$ntree    else 500
  fs.nodesize <- if (!is.null(fs.opts$nodesize)) fs.opts$nodesize else NULL
  fs.nsplit   <- if (!is.null(fs.opts$nsplit))   fs.opts$nsplit   else 10
  ##--------------------------------------------------------------
  ##
  ## process the data
  ##
  ##--------------------------------------------------------------
  p <- ncol(data)
  n <- nrow(data)
  all.r.na <- rowSums(which.na) == p
  all.c.na <- colSums(which.na) == n
  data <- data[!all.r.na, !all.c.na, drop = FALSE]
  which.na <- which.na[!all.r.na, !all.c.na, drop = FALSE]
  if (!any(which.na)) {
    return(data)
  }
  p <- ncol(data)
  n <- nrow(data)
  all.var.names <- colnames(data)
  ## integer-valued numeric columns are detected before imputation
  ## contaminates their support.  Actual integer columns are restored
  ## as integer vectors; numeric columns with integer support remain
  ## numeric but have imputed values rounded to the integer grid.
  integer.support <- .integer.support.map(data)
  ## mforest details
  if (missing(mf.q)) {
    mforest <- FALSE
  }
  else {
    mforest <- TRUE
  }
  ## set the number of blocks used to subdivide the data
  if (!missing(blocks)) {
    blocks <- cv.folds(nrow(data), max(1, blocks))
  }
  else {
    blocks <- list(1:nrow(data))
  }
  ##--------------------------------------------------------------
  ## METHOD 1: default impute call
  ##--------------------------------------------------------------
  if (!mforest) {
    if (missing(formula)) {
      if (is.null(ytry)) {
        ytry <- min(p - 1, max(25, ceiling(sqrt(p))))
      }
      dots$formula <- as.formula(paste("Unsupervised(", ytry, ") ~ ."))
      dots$splitrule <- NULL
    }
    else {
      dots$formula <- formula
    }
    nullBlocks <- lapply(blocks, function(blk) {
      dta <- data[blk,, drop = FALSE]
      retO <- tryCatch({do.call("generic.impute.rfsrc",
                      c(list(data = dta,
                             ntree = ntree,
                             nodesize = nodesize,
                             nsplit = nsplit,
                             nimpute = nimpute,
                             fast = fast), dots))}, error = function(e) {NULL})
      if (!is.null(retO)) {
        if (!is.null(retO$missing$row)) {
          blk <- blk[-retO$missing$row]
        }
        if (!is.null(retO$missing$col)) {
          ynames <- all.var.names[-retO$missing$col]
        }
          else {
            ynames <- all.var.names
          }
        data[blk, ynames] <<- retO$data[, ynames, drop = FALSE]
      }
      NULL
    })
    rm(nullBlocks)
  }
  ##--------------------------------------------------------------
  ## METHOD 2: mforest
  ##--------------------------------------------------------------
  if (mforest) {
    x.na <- lapply(1:p, function(k) {
      if (sum(which.na[, k]) > 0) {
        as.numeric(which(which.na[, k]))
      }
        else {
          NULL
        }
    })
    which.x.na <- which(sapply(x.na, length) > 0)
    names(x.na) <- all.var.names <- colnames(data)
    var.names <- all.var.names[which.x.na] 
    if (!is.null(always.use)) {
      always.use <- is.element(all.var.names, always.use)
      if (sum(always.use) > 0) {
        always.use <- which(always.use)
      }
    }
    p0 <- length(which.x.na)
    mfOriginal <- FALSE
    if (mf.q == 0) {
      stop("mf.q must be greater than zero")
    }
    if (mf.q == 1 && is.null(always.use)) {
      mfOriginal <- TRUE
    }
    ## Change D: guard when only one variable has missingness
    if (p0 == 1) {
      K <- 1
    } else {
      if (mf.q >= 1) {
        mf.q <- min(p0 - 1, mf.q) / p0
      }
      K <- max(1, round(max(1 / mf.q, 2)))
    }
    dots.rough <- dots
    dots.rough$mtry <- NULL
    dots.rough$splitrule <- "random"
    dots.rough$formula <- as.formula(zZ999Zz ~.)
    data <- do.call("generic.impute.rfsrc",
                    c(list(data = data.frame(zZ999Zz = rnorm(nrow(data)), data),
                           ntree = 10,
                           nodesize = nodesize,
                           nsplit = nsplit,
                           nimpute = 1,
                           fast = fast), dots.rough))$data
    data$zZ999Zz <- NULL
    data <- .restore.integer.data(data, integer.support,
                                  restore.integer = restore.integer)
    diff.err <- Inf
    check <- TRUE
    var.grp <- cv.folds(p0, K)
    K <- length(var.grp)
    if (verbose) {
      if (mfOriginal) {
        cat("missForest parameters:", paste0("(#max.iter, #vars)=(", max.iter, ",", p0, ")"), "\n")
      }
      else {
        cat("multivariate missForest parameters:",
            paste0("(#iter, #vars, #blks)=(", max.iter, ",", p0, ",", K, ")"), "\n")
      }
    }
    nullWhile <- lapply(1:max.iter, function(m) {
      if (!check) {
        return(NULL)
      }
      var.grp <- cv.folds(p0, K)
      if (verbose) {
        if (max.iter > 1) {
          cat("\t", paste0("iteration:", m, "\n"))
        }
      }
      data.old <- data
      nullBlocks <- lapply(blocks, function(blk) {
        nullObj <- lapply(var.grp, function(grp) {
          if (verbose) {
            cat(".")
          }
          if (!mfOriginal) {
            ynames <- unique(c(var.names[grp], all.var.names[always.use]))
            dots$formula <- as.formula(paste("Multivar(", paste(ynames, collapse = ","), paste(") ~ ."), sep = ""))
            dta <- data[blk,, drop = FALSE]
            dta[, ynames] <- lapply(ynames, function(nn) {
              xk <- data[, nn]
              xk[unlist(x.na[nn])] <- NA
              xk[blk]
            })
            mvimpute <- tryCatch({do.call("generic.impute.rfsrc",
                                      c(list(data = dta,
                                             ntree = ntree,
                                             nodesize = nodesize,
                                             nsplit = nsplit,
                                             nimpute = 1,
                                             fast = fast), dots))}, error = function(e) {NULL})
            if (!is.null(mvimpute)) {
              if (!is.null(mvimpute$missing$row)) {
                blk <- blk[-mvimpute$missing$row]
              }
              if (!is.null(mvimpute$missing$col)) {
                ynames <- ynames[-mvimpute$missing$col]
              }
              data[blk, ynames] <<- mvimpute$data[, ynames, drop = FALSE]
              rm(dta)
            }
          }
          if (mfOriginal) {
            yname <- var.names[grp]
            trn <- setdiff(blk,  unlist(x.na[yname]))
            tst <- setdiff(blk,  trn)
            if (length(trn) > 0 && length(tst) > 0) {
              dots$formula <- as.formula(paste0(yname, "~."))
              grow <- tryCatch({do.call("rfsrc",
                                        c(list(data = data[trn,, drop = FALSE],
                                               ntree = ntree,
                                               nodesize = nodesize,
                                               nsplit = nsplit,
                                               perf.type = "none",
                                               fast = fast), dots))}, error = function(e) {NULL})
              if (!is.null(grow)) {
                pred <- predict(grow, data[tst,, drop = FALSE])
                if (grow$family == "regr") {
                  data[tst, yname] <<- pred$predicted
                }
                else {
                  data[tst, yname] <<- pred$class
                }
              }
            }
          }
          NULL
        })
        NULL
      })
      diff.new.err <- mean(sapply(var.names, function(nn) {
        xo <- data.old[unlist(x.na[nn]), nn]
        xn <- data[unlist(x.na[nn]), nn]
        if (!is.numeric(xo)) {
          sum(xn != xo, na.rm = TRUE) / (.001 + length(xn))
        }
          else {
            var.xo <- var(xo, na.rm = TRUE)
            if (is.na(var.xo)) {
              var.xo <- 0
            }
            sqrt(mean((xn - xo)^2, na.rm = TRUE) / (.001 + var.xo))
          }
      }), na.rm = TRUE)
      if (verbose) {
        cat("\n")
        err <- paste("err = " , format(diff.new.err, digits = 3), sep = "")
        drp <- paste("drop = ", format(diff.err - diff.new.err, digits = 3), sep = "")
        cat("         >> ", err, ", ", drp, "\n")
      }
      check <<- ((diff.err - diff.new.err) >= eps)
      diff.err <<- diff.new.err
      rm(data.old)
      NULL
    })
  }
  ##--------------------------------------------------------------
  ##
  ## optional final sweep over originally-missing variables
  ## (applies to both methods) + progress output
  ##
  ##--------------------------------------------------------------
  if (isTRUE(full.sweep)) {
    data <- .restore.integer.data(data, integer.support,
                                  restore.integer = restore.integer)
    ## identify row indices of original missing values for each variable
    x.na.sweep <- lapply(1:ncol(which.na), function(k) {
      idx <- which(which.na[, k])
      if (length(idx) > 0) as.numeric(idx) else NULL
    })
    names(x.na.sweep) <- colnames(which.na)
    ## variables with any original missingness
    sweep.vars <- names(which(sapply(x.na.sweep, length) > 0))
    if (length(sweep.vars) > 0) {
      nvars <- length(sweep.vars)
      nullSweep <- lapply(seq_along(sweep.vars), function(j) {
        yname <- sweep.vars[[j]]
        ## progress output
        if (verbose) {
          pct <- round(100 * j / nvars)
          cat("--> full.sweep:", paste0("[", j, "/", nvars, "]"), yname,
              paste0(" (", pct, "%)"), "\n")
        }
        tst <- x.na.sweep[[yname]]
        if (length(tst) == 0) return(NULL)
        trn <- setdiff(seq_len(nrow(data)), tst)
        if (length(trn) == 0) return(NULL)
        ## fit on observed y rows using final data
        grow <- tryCatch({
          do.call("rfsrc",
                  c(list(formula   = as.formula(paste0(yname, "~.")),
                         data      = data[trn,, drop = FALSE],
                         ntree     = fs.ntree,
                         nodesize  = fs.nodesize,
                         nsplit    = fs.nsplit,
                         perf.type = "none",
                         fast      = fast), fs.dots))
        }, error = function(e) { NULL })
        if (!is.null(grow)) {
          pred <- tryCatch(predict(grow, data[tst,, drop = FALSE]),
                           error = function(e) { NULL })
          if (!is.null(pred)) {
            if (grow$family == "regr") {
              data[tst, yname] <<- pred$predicted
            }
            else {
              data[tst, yname] <<- pred$class
            }
          }
        }
        NULL
      })
      rm(nullSweep)
      if (verbose) {
        cat("full.sweep: completed (", length(sweep.vars), " variables)\n", sep = "")
      }
    }
    data <- .restore.integer.data(data, integer.support,
                                  restore.integer = restore.integer)
  }
  ##--------------------------------------------------------------
  ##
  ## return the imputed data
  ##
  ##--------------------------------------------------------------
  data <- .restore.integer.data(data, integer.support,
                                restore.integer = restore.integer)
  invisible(data)
}
impute <- impute.rfsrc

generic.impute.rfsrc <- function(data,
                                 ntree = 250,
                                 nodesize = NULL,
                                 nsplit = 1,
                                 nimpute = 1,
                                 fast = FALSE,
                                 ...)
{
  ## save the row and column names: later we will check if any rows or columns
  ## were deleted as part of the missing data preprocessing
  c.names <- colnames(data)
  r.names <- rownames(data)
  ## acquire the permissible hidden options
  dots <- list(...)
  dots$na.action <- dots$impute.only <- dots$forest <- NULL
  ## rfsrc grow call 
  if (!fast) {
    object <- do.call("rfsrc",
                    c(list(data = data,
                           ntree = ntree,
                           nodesize = nodesize,
                           nsplit = nsplit,
                           nimpute = nimpute,
                           na.action = "na.impute",
                           impute.only = TRUE), dots))
  }
  else {## user has requested the fast forest interface
    object <- do.call("rfsrc.fast",
                    c(list(data = data,
                           ntree = ntree,
                           nodesize = nodesize,
                           nsplit = nsplit,
                           nimpute = nimpute,
                           na.action = "na.impute",
                           impute.only = TRUE), dots))
  }
  ## confirm that no error has occured
  if (is.null(object)) {
    return(NULL)
  }
  ## the data is no longer needed
  rm(data)
  ## if the return object is a data frame then imputation was not
  ## performed: for example, there was no missing data either before
  ## or after processing
  if (is.data.frame(object)) {
    return(invisible(list(data = object, missing = row.col.deleted(object, r.names, c.names)))) 
  }
  ## preliminary results of imputation
  if(is.null(object$yvar.names)) {
    imputed.result <- object$xvar
  }
    else {
      imputed.result <- cbind(object$yvar, object$xvar)
    }
  colnames(imputed.result) <- c(object$yvar.names, object$xvar.names)
  ## overlay the data (only necessary when nimpute = 1)
  if (nimpute == 1) {
    imputed.result[object$imputed.indv, ] <- object$imputed.data
  }
  ## the object is no longer required
  rm(object)
  ## return the goodies
  invisible(list(data = imputed.result, missing = row.col.deleted(imputed.result, r.names, c.names)))
}

###################################################################
##
## core bootstrap function
##
###################################################################
bootsample <- function(obj, rf.prms, B, block.size, joint, xvar.names,
                       importance, performance, performance.only, verbose) {
  ##--------------------------------------------------------------
  ##
  ## extract data from the previously grown forest
  ## set the dimensions, pull the feature names
  ##
  ##--------------------------------------------------------------
  dta <- data.frame(obj$yvar, obj$xvar)
  colnames(dta)[1:length(obj$yvar.names)] <- obj$yvar.names
  n <- nrow(obj$xvar)
  ##--------------------------------------------------------------
  ##
  ## call vimp to extract importance if not available in original object
  ##
  ##--------------------------------------------------------------
  vmp <- get.mv.vimp(obj, FALSE, FALSE)
  if (!performance.only) {
    if (is.null(vmp)) {
      if (verbose) cat("no importance found: calculating it now ...\n")
      obj <- vimp(obj, perf.type = rf.prms$perf.type, block.size = block.size, importance = importance)
      vmp <- get.mv.vimp(obj, FALSE, FALSE)
      rf.prms$block.size <- block.size
      if (verbose) cat("done\n")
    }
    else {
      importance <- obj$forest$importance
    }
  }
  ##--------------------------------------------------------------
  ##
  ## call joint vimp - reference vimp value for pure noise settings
  ##
  ##--------------------------------------------------------------
  if (joint && !performance.only) {
    vmp.joint <- get.mv.vimp(get.subsample.joint.vimp(obj, rf.prms, xvar.names), FALSE, FALSE)
    vmp <- vimp.subsample.combine(vmp, vmp.joint)
  }
  ##--------------------------------------------------------------
  ##
  ## obtain performance value - used for error rate confidence intervals
  ##
  ##--------------------------------------------------------------
  if (performance) {
    err <- get.mv.error(obj, FALSE, FALSE)
    vmp <- vimp.subsample.combine(vmp, err, "err")
  }
  ##----------------------------------------------------------
  ##
  ## bootstrap loop for calculating VIMP confidence regions
  ##
  ##----------------------------------------------------------
  vmpB <- lapply(1:B, function(b) {
    ## progress bar
    if (verbose && B > 1) {
      custom.progress.bar(b, B)
    }
    ##---------------------------------------------------
    ##
    ## double bootstrap 
    ##
    ##---------------------------------------------------
    ## double bootstrap sample
    smp <- sample(1:n, size = n, replace = TRUE)
    smp.unq <- sort(unique(smp))
    dbl.smp <- make.double.boot.sample(rf.prms$ntree, n, smp,
                             size = rf.prms$sampsize, replace = rf.prms$samptype == "swr")
    ## calculate VIMP from the double boostrap forest
    rf.b <- do.call("rfsrc",
                    c(list(data = dta[smp.unq,, drop = FALSE], importance = importance,
                           bootstrap = "by.user", samp = dbl.smp), rf.prms))
    vmp.b <- get.mv.vimp(rf.b, FALSE, FALSE)
    if (joint && !performance.only) {
      vmp.b.joint <- get.mv.vimp(get.subsample.joint.vimp(rf.b, rf.prms, xvar.names), FALSE, FALSE)
      vmp.b <- vimp.subsample.combine(vmp.b, vmp.b.joint)
    }
    if (performance) {
      err.b <- get.mv.error(rf.b, FALSE, FALSE)
      vmp.b <- vimp.subsample.combine(vmp.b, err.b, "err")
    }
    ## combine
    rO.b <- lapply(1:length(vmp), function(j) {
      vmp.j <- vmp[[j]]
      vmp.b.j <- vmp.b[[j]]
      vmp.mtx <- data.frame(matrix(NA, nrow(vmp.j), ncol(vmp.j)))
      rownames(vmp.mtx) <- rownames(vmp.j)
      colnames(vmp.mtx) <- colnames(vmp.j)
      if (ncol(vmp.b.j) == ncol(vmp.j)) {
        vmp.mtx[rownames(vmp.b.j), ] <- vmp.b.j
        vmp.b.j <- vmp.mtx
      }
      vmp.b.j
    })
    names(rO.b) <- names(vmp)
    rO.b
  })
  ## return the potentially updated forest object
  ## return the vimp 
  list(rf = obj, vmp = vmp, vmpB = vmpB)
}
###################################################################
##
## extract confidence interval + other goodies from bootstrap object
##
###################################################################
extract.bootsample <- function(obj, alpha = .05, target = 0, m.target = NULL, standardize = TRUE, raw = FALSE)
{
  ## confirm this is the right kind of object
  if (!sum(c(grepl("rfsrc", class(obj))), grepl("bootsample", class(obj))) == 2) {
    stop("this must be a double-bootstrap object obtained by calling the 'subsample' function")
  }
  ## coerce the (potentially) multivariate rf object
  m.target <- get.univariate.target(obj$rf, m.target)
  obj$rf <- coerce.multivariate(obj$rf, m.target)
  ## coerce the (potentially) multivariate vimp object
  if (is.null(m.target)) {
    obj$vmpB <- lapply(obj$vmpB, data.frame)
  }
  else {
    obj$vmpB <- lapply(obj$vmpB, function(oo) {
      oo[[m.target]]
    })
  }  
  ## extract vimp
  theta.boot <- get.subsample.standardize.vimp(obj$rf, rbind(sapply(obj$vmpB, "[[", 1 + target)), standardize)
  rownames(theta.boot) <- rownames(obj$vmp[[1]][, 1 + target, drop = FALSE])
  ## bootstrap VIMP
  vmp.boot <- rowMeans(theta.boot, na.rm = TRUE)
  ## bootstrap standard error
  se.boot <- apply(theta.boot, 1, sd, na.rm = TRUE)
  ## nonparametric bootstrap confidence interval
  ci.boot <- do.call(cbind, lapply(1:length(vmp.boot), function(rowj) {
    c(quantile(theta.boot[rowj,], alpha/2, na.rm = TRUE),
      quantile(theta.boot[rowj,], .25, na.rm = TRUE),
      vmp.boot[rowj],
      quantile(theta.boot[rowj,], .75, na.rm = TRUE),
      quantile(theta.boot[rowj,], 1 - alpha/2, na.rm = TRUE))
  }))
  rownames(ci.boot) <- paste(round(100 * c(alpha/2,.25,.50,.75,1-alpha/2), 1), "%", sep = "")
  colnames(ci.boot) <- names(vmp.boot)
  ## parametric bootstrap confidence interval
  ci.boot.Z <- do.call(cbind, lapply(1:length(vmp.boot), function(rowj) {
    c(vmp.boot[rowj] - qnorm(1 - alpha/2) * se.boot[rowj],
      vmp.boot[rowj] - qnorm(.75) * se.boot[rowj],
      vmp.boot[rowj],
      vmp.boot[rowj] + qnorm(.75) * se.boot[rowj],
      vmp.boot[rowj] + qnorm(1 - alpha/2) * se.boot[rowj])
  }))
  rownames(ci.boot.Z) <- paste(round(100 * c(alpha/2,.25,.50,.75,1-alpha/2), 1), "%", sep = "")
  colnames(ci.boot.Z) <- names(vmp.boot)
  ## bootstrap variable selection object
  var.sel.boot <- data.frame(lower  =  ci.boot[1, ],
                             mean   =  ci.boot[3, ],
                             upper  =  ci.boot[5, ],
                             signif =  ci.boot[1, ] > 0)
  rownames(var.sel.boot) <- names(vmp.boot)
  var.sel.boot.Z <- data.frame(lower  =  ci.boot.Z[1, ],
                               mean   =  ci.boot.Z[3, ],
                               upper  =  ci.boot.Z[5, ],
                               pvalue = pnorm(vmp.boot, 0, se.boot, FALSE),
                               signif =  ci.boot.Z[1, ] > 0)
  rownames(var.sel.boot.Z) <- names(vmp.boot)
  if (raw) {
    list(ci = ci.boot,
         ci.Z = ci.boot.Z,
         vmp = vmp.boot,
         vmpS = theta.boot,
         se = se.boot,
         var.sel = var.sel.boot,
         var.sel.Z = var.sel.boot.Z)
  }
  else {
    list(se = se.boot,
         var.sel = var.sel.boot,
         var.sel.Z = var.sel.boot.Z)
  }
}
###################################################################
##
## print function from bootstrap ci
##
###################################################################
print.bootsample.rfsrc <- function(x, alpha = .05, standardize = TRUE, ...) {
  vmp <- x$vmp
  nullO <- lapply(1:length(vmp), function(j) {
    m.target <- names(vmp)[j]
    p.vmp <- ncol(vmp[[j]])
    vmp.col.names <- colnames(vmp[[j]])
    if (length(vmp) > 1) {
      cat("processing a multivariate family, outcome:", m.target, "\n")
    }
    lapply(1:p.vmp, function(k) {
      oo <- extract.bootsample(x, alpha = alpha, target = k - 1,
                   m.target = m.target, standardize = standardize, raw = TRUE)
      cat("===== VIMP confidence regions for", vmp.col.names[k], " =====\n")
      cat("nonparametric:\n")
      print(round(oo$ci, 3))
      cat("parametric:\n")
      print(round(oo$ci.Z, 3))
      NULL
    })
  })
}
print.bootsample <- print.bootsample.rfsrc

.onAttach <- function(libname, pkgname) {
  rfsrc.version <- read.dcf(file=system.file("DESCRIPTION", package=pkgname), 
                            fields="Version")
  packageStartupMessage(paste("\n",
                              pkgname,
                              rfsrc.version,
                              "\n",
                              "\n",
                              "Type rfsrc.news() to see new features, changes, and bug fixes.",
                              "\n",
                              "\n"))
}

partial.rfsrc <- function(
  object,
  oob = TRUE,
  partial.type = NULL,
  partial.xvar = NULL,
  partial.values = NULL,
  partial.xvar2 = NULL,
  partial.values2 = NULL,
  partial.time = NULL,
  get.tree = NULL,
  seed = NULL,
  do.trace = FALSE,
  ...)
{
  ## hidden options
  user.option <- list(...)
  ## Object consistency
  ## TBD Version checking is NOT implemented in this mode
  if (missing(object)) {
    stop("object is missing!")
  }
  ## coherence checks
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2    &
      sum(inherits(object, c("rfsrc", "forest"), TRUE) == c(1, 2)) != 2) {
    stop("this function only works for objects of class `(rfsrc, grow)' or '(rfsrc, forest)'")
  }
  ## anonymous not allowed
  if (inherits(object, "anonymous")) {
    stop("this function does work with anonymous forests")
  }
  ## jitt=FALSE if not specified
  jitt <- is.hidden.jitt(user.option, "none", "na.omit", FALSE, partial = TRUE)
  ## data-pass
  data.pass <- is.hidden.data.pass(user.option)
  ## acquire the forest --> hereafter the object is the forest
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) == 2) {
    if (is.forest.missing(object)) {
      stop("The forest is empty.  Re-run rfsrc (grow call) with forest=TRUE")
    }
    object <- object$forest
  }
  else {
    ## object is already a forest.
  }
  ## convert oob to ensemble
  if (oob) {
    ensemble <- "oob"
  }
  else {
    ensemble <- "inbag"
  }
  ## multivariate family details
  family <- object$family
  splitrule <- object$splitrule
  ## pull the x-variable and y-outcome names from the grow object
  xvar.names <- object$xvar.names
  yvar.names <- object$yvar.names
  ## verify the x-var
  if (length(which(xvar.names == partial.xvar)) != 1) {
    stop("x-variable specified incorrectly:  ", partial.xvar)
  }
  ## uniquify the x-var
  partial.values <- sort(unique(partial.values))
  ## verify the x-var2
  if (!is.null(partial.xvar2)) {   
      if (length(partial.xvar2) != length(partial.values2)) {
          stop("second order x-variable and value vectors not of same length:  ", length(partial.xvar2), "vs", length(partial.values2))
      }
      for (i in 1:length(partial.xvar2)) {
          if (length(which(xvar.names == partial.xvar2[i])) != 1) {
              stop("second order x-variable element", i, "specified incorrectly:  ", partial.xvar2[i])
          }
      }
  }
  ## Caution:  There are no checks on partial.type.  Note that "rel.freq" and "mort"
  ## are equivalent from a native code perspective.
  ## Determine the immutable yvar factor map
  yfactor <- object$yvar.factor
  ## multivariate family details
  ##m.target.idx <- get.outcome.target(family, yvar.names, NULL)
  m.target.idx <- 1:length(yvar.names)
  ## short cut to get the y-outcome type and number of levels
  yvar.types <- yfactor$types
  yvar.nlevels  <- yfactor$nlevels
  yvar.numeric.levels  <- yfactor$numeric.levels
  ## recover the individual subject identifiers, if they exist.
  subj <- object$subj
  ## Get event information for survival families.
  event.info <- object$event.info
  event.type  <- event.info$event.type
  ## CR.bits assignment.
  cr.bits <- get.cr.bits(family)
  ## Determine the immutable xvar factor map.
  xfactor <- object$xvar.factor
  ## short cut to get the x-variable type and number of levels
  xvar.types <- xfactor$types
  xvar.nlevels <- xfactor$nlevels
  xvar.numeric.levels <- xfactor$numeric.levels
  ## Initialize the number of trees in the forest.
  ntree <- object$ntree
  ## Use the training data na.action protocol.
  na.action = object$na.action
  ## Data conversion for training data.
  xvar <- as.matrix(data.matrix(object$xvar))
  yvar <- as.matrix(data.matrix(object$yvar))
  ## Set the y dimension.
  r.dim <- ncol(cbind(yvar))
  ## remove row and column names for proper processing by the native code
  ## set the dimensions.
  rownames(xvar) <- colnames(xvar) <- NULL
  n.xvar <- ncol(xvar)
  n <- nrow(xvar)
  sampsize <- round(object$sampsize(n))
  ## Recover the case weights and bootstrap sample. 
  case.wt <- object$case.wt
  samp <- object$samp
  ## There is no test data.
  outcome = "train"
  ## Process the get.tree vector that specifies which trees we want
  get.tree <- get.tree.index(get.tree, ntree)
  ## Initialize the low bits.
  ensemble.bits <- get.ensemble.bits(ensemble)
  bootstrap.bits <- get.bootstrap.bits(object$bootstrap)
  na.action.bits <- get.na.action.bits(na.action)
  ## Initalize the high bits
  samptype.bits <- get.samptype.bits(object$samptype)
  partial.bits <- get.partial.bits(length(partial.values))
  terminal.qualts.bits <- get.terminal.qualts.predict.bits(object$terminal.qualts)
  terminal.quants.bits <- get.terminal.quants.predict.bits(object$terminal.quants)
  data.pass.bits <- get.data.pass.bits(data.pass)
  ## Just in Time Tree Topology Restoration.  Note that we do not
  ## support augmented data, SGT, TDC, here, but we do no checks for
  ## these. We defer to the user option for JITT.
  jitt.bits <- get.jitt.bits(jitt)
  seed <- get.seed(seed)
  do.trace <- get.trace.bits(do.trace)
  ## The pivot in this predict related function is different from
  ## the pivot used in the grow related function.  In rfsrc we are
  ## referencing the list nativeOutput[[]].  Here we are referencing
  ## the $nativeArray[[]] object which is a massaged version of
  ## nativeOutput[[]].  Here, pivot points to the record PRIOR to
  ## parmID2.  We adjust for the presence of interactions, and
  ## synthetics.  The default chunk is parmIDx, contPTx, contPTRx,
  ## mwcpSZx, fsrecIDx.
  ## WARNING: Note that the maximum number of slots in the following
  ## foreign function call is 64.  Ensure that this limit is not
  ## exceeded.  Otherwise, the program will error on the call.
  pivot = 0
  chunk = 0
  nativeOutput <- tryCatch({.Call("rfsrcPredict",
                                  as.integer(do.trace),
                                  as.integer(seed),
                                  as.integer(ensemble.bits +
                                             bootstrap.bits +
                                             cr.bits),
                                  as.integer(samptype.bits +
                                             na.action.bits +
                                             terminal.qualts.bits +
                                             terminal.quants.bits +
                                             partial.bits + 
                                             data.pass.bits +
                                             jitt.bits),
                                  ## >>>> start of maxi forest object >>>>
                                  as.numeric(1),
                                  as.integer(ntree),
                                  as.integer(n),
                                  list(as.integer(length(yvar.types)),
                                       if (is.null(yvar.types)) NULL else as.character(yvar.types),
                                       if (is.null(yvar.types)) NULL else as.integer(yvar.nlevels),
                                       if (is.null(yvar.numeric.levels)) NULL else sapply(1:length(yvar.numeric.levels), function(nn) {as.integer(length(yvar.numeric.levels[[nn]]))}),
                                       if (is.null(subj)) NULL else as.integer(subj),
                                       if (is.null(event.type)) NULL else as.integer(length(event.type)),
                                       if (is.null(event.type)) NULL else as.integer(event.type)),
                                  if (is.null(yvar.numeric.levels)) {
                                      NULL
                                  }
                                  else {
                                      lapply(1:length(yvar.numeric.levels),
                                             function(nn) {as.integer(yvar.numeric.levels[[nn]])})
                                  },
                                  if (is.null(yvar.types)) NULL else as.double(as.vector(yvar)),
                                  list(as.integer(n.xvar),
                                       as.character(xvar.types),
                                       if (is.null(xvar.types)) NULL else as.integer(xvar.nlevels),
                                       if (is.null(xvar.numeric.levels)) NULL else sapply(1:length(xvar.numeric.levels), function(nn) {as.integer(length(xvar.numeric.levels[[nn]]))}),
                                       NULL,
                                       NULL),
                                  if (is.null(xvar.numeric.levels)) {
                                      NULL
                                  }
                                  else {
                                      lapply(1:length(xvar.numeric.levels),
                                             function(nn) {as.integer(xvar.numeric.levels[[nn]])})
                                  },
                                  as.double(as.vector(xvar)),
                                  list(as.integer(length(case.wt)),
                                       if (is.null(case.wt)) NULL else as.double(case.wt),
                                       as.integer(sampsize),
                                       if (is.null(samp)) NULL else as.integer(samp)),
                                  list(if(is.null(event.info$time.interest)) as.integer(0) else as.integer(length(event.info$time.interest)),
                                       if(is.null(event.info$time.interest)) NULL else as.double(event.info$time.interest)),
                                  as.integer(object$totalNodeCount),
                                  as.integer(object$leafCount),
                                  list(as.integer(object$seed),
                                       if (is.null(object$seedVimp)) NULL else as.integer(object$seedVimp),
                                       as.integer(object$optLoGrow)),
                                  as.integer(0),
                                  ## Deleted base learner. 
                                  NULL,
                                  as.integer((object$nativeArray)$treeID),
                                  as.integer((object$nativeArray)$nodeID),
                                  as.integer((object$nativeArray)$nodeSZ),
                                  as.integer((object$nativeArray)$brnodeID),
                                  ## This is hc_zero.  It is never NULL.
                                  list(as.integer((object$nativeArray)$parmID),
                                  as.double((object$nativeArray)$contPT),
                                  as.integer((object$nativeArray)$mwcpSZ),
                                  as.integer((object$nativeArray)$fsrecID),
                                  if (is.null((object$nativeFactorArray)$mwcpPT)) NULL else as.integer((object$nativeFactorArray)$mwcpPT)),
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  as.integer(object$nativeArrayTNDS$tnRMBR),
                                  as.integer(object$nativeArrayTNDS$tnAMBR),
                                  as.integer(object$nativeArrayTNDS$tnRCNT),
                                  as.integer(object$nativeArrayTNDS$tnACNT),
                                  as.double((object$nativeArrayTNDS$tnSURV)),
                                  as.double((object$nativeArrayTNDS$tnMORT)),
                                  as.double((object$nativeArrayTNDS$tnNLSN)),
                                  as.double((object$nativeArrayTNDS$tnCSHZ)),
                                  as.double((object$nativeArrayTNDS$tnCIFN)),
                                  as.double((object$nativeArrayTNDS$tnREGR)),
                                  as.integer((object$nativeArrayTNDS$tnCLAS)),
                                  ## <<<< end of maxi forest object <<<<
                                  list(if (is.null(m.target.idx)) as.integer(0) else as.integer(length(m.target.idx)),
                                       if (is.null(m.target.idx)) NULL else as.integer(m.target.idx)),
                                  as.integer(0),  ## Pruning disabled
                                  NULL,
                                  list(as.integer(0), NULL), ## Importance disabled.
                                  ## Partial variables enabled.  Note the as.integer is needed.
                                  list(as.integer(get.partial.type(family, partial.type)),
                                       as.integer(which(xvar.names == partial.xvar)),
                                       as.integer(length(partial.values)),
                                       as.double(partial.values),
                                       as.integer(length(partial.xvar2)),
                                       if (length(partial.xvar2) == 0) NULL else as.integer(match(partial.xvar2, xvar.names)),
                                       as.double(partial.values2)),
                                  as.integer(0),    ## New data disabled.
                                  as.integer(0),    ## New data disabled.
                                  as.double(NULL),  ## New data disabled.
                                  as.double(NULL),  ## New data disabled.
                                  as.integer(ntree), ## block.size is hard-coded.
                                  list(as.integer(0),
                                       NULL,
                                       as.double(0)), ## Quantiles disabled.
                                  as.integer(get.tree),
                                  NULL,
                                  as.integer(get.rf.cores()))}, error = function(e) {
                                    print(e)
                                    NULL})
  ## check for error return condition in the native code
  if (is.null(nativeOutput)) {
    stop("An error has occurred in prediction.  Please turn trace on for further analysis.")
  }
  rfsrcOutput <- list(call = match.call(),
                      family = family,
                      partial.time = partial.time)
  ## Subset the user time vector from the grow time interest vector.
  if (grepl("surv", family)) {
      ## Get the indices of the closest points of the grow time interest vector.
      ## Exact:
      ## partial.time.idx <- match(partial.time, event.info$time.interest)
      ## Closest:
      partial.time.idx <- sapply(partial.time, function(x) {max(which(event.info$time.interest <= x))})
      if (sum(is.na(partial.time.idx)) > 0) {
          stop("partial.time must be a subset of the time interest vector contained in the model")
      }
  }
  if (family == "surv") {
    if ((partial.type == "rel.freq") || (partial.type == "mort")) {
      mort.names <- list(NULL, NULL)
      ## Incoming from the native code:
      ##   type = mort
      ##   -> of dim [length(partial.values)] x [1] x [1] x [n]
      ## Outgoing to the R code:
      ##   -> of dim [n] x [length(partial.values)]  
      survOutput <- (if (!is.null(nativeOutput$partialSurv))
                       array(nativeOutput$partialSurv,
                             c(n, length(partial.values)),
                             dimnames=mort.names) else NULL)
    }
    else if (partial.type == "chf") {
      nlsn.names <- list(NULL, NULL, NULL)
      ## Incoming from the native code:
      ##   type = chf
      ##   -> of dim [length(partial.values)] x [1] x [length(partial.time)] x [n]
      ## Outgoing to the R code:
      ##   -> of dim [n] x [length(partial.time)] x [length(partial.values)]  
      survOutput <- (if (!is.null(nativeOutput$partialSurv))
                       array(nativeOutput$partialSurv,
                             c(n, length(event.info$time.interest), length(partial.values)),
                             dimnames=nlsn.names) else NULL)
      if (!is.null(survOutput)) {
        survOutput <- survOutput[, partial.time.idx, , drop=FALSE]
      }
    }
      else if (partial.type == "surv") {
        surv.names <- list(NULL, NULL, NULL)
        ## Incoming from the native code:
        ##   type = surv
        ##   -> of dim [length(partial.values)] x [1] x [length(partial.time)] x [n]
        ## Outgoing to the R code:
        ##   -> of dim [n] x [length(partial.time)] x [length(partial.values)]  
        survOutput <- (if (!is.null(nativeOutput$partialSurv))
                         array(nativeOutput$partialSurv,
                               c(n, length(event.info$time.interest), length(partial.values)),
                               dimnames=surv.names) else NULL)
        if (!is.null(survOutput)) {
          survOutput <- survOutput[, partial.time.idx, , drop=FALSE]
        }
      }
        else {
          stop("Invalid choice for 'partial.type' option:  ", partial.type)
        }
    rfsrcOutput <- c(rfsrcOutput, survOutput = list(survOutput))
  }
  else if (family == "surv-CR") {
    if (partial.type == "years.lost") {
      yrls.names <- list(NULL, NULL, NULL)
      ## Incoming from the native code:
      ##   type = years.lost
      ##   -> of dim [length(partial.values)] x [length(event.info$event.type)] x [1] x [n]
      ## Outgoing to the R code:
      ##   -> of dim [n] x [length(event.info$event.type)] x [length(partial.values)]  
      survOutput <- (if (!is.null(nativeOutput$partialSurv))
                       array(nativeOutput$partialSurv,
                             c(n, length(event.info$event.type), length(partial.values)),
                             dimnames=yrls.names) else NULL)
    }
      else if (partial.type == "cif") {
        cifn.names <- list(NULL, NULL, NULL, NULL)
        ## Incoming from the native code:
        ##   type = cif
        ##   -> of dim [length(partial.values)] x [length(event.info$event.type)] x [length(partial.time)] x [n]
        ## Outgoing to the R code:
        ##   -> of dim [n] x [length(partial.time)] x [length(event.info$event.type)] x [length(partial.values)]
        survOutput <- (if (!is.null(nativeOutput$partialSurv))
                         array(nativeOutput$partialSurv,
                               c(n, length(event.info$time.interest), length(event.info$event.type), length(partial.values)),
                               dimnames=cifn.names) else NULL)
        if (!is.null(survOutput)) {
          survOutput <- survOutput[, partial.time.idx, , , drop=FALSE]
        }
      }
        else if (partial.type == "chf") {
          chfn.names <- list(NULL, NULL, NULL, NULL)
          ## Incoming from the native code:
          ##   type = chfn
          ##   -> of dim [length(partial.values)] x [length(event.info$event.type)] x [length(partial.time)] x [n]
          ## Outgoing to the R code:
          ##   -> of dim [n] x [length(partial.time)] x [length(event.info$event.type)] x [length(partial.values)]
          survOutput <- (if (!is.null(nativeOutput$partialSurv))
                           array(nativeOutput$partialSurv,
                                 c(n, length(event.info$time.interest), length(event.info$event.type), length(partial.values)),
                                 dimnames=chfn.names) else NULL)
          if (!is.null(survOutput)) {
            survOutput <- survOutput[, partial.time.idx, , , drop=FALSE]
          }
        }
          else {
            stop("Invalid choice for 'partial.type' option:  ", partial.type)
          }
    rfsrcOutput <- c(rfsrcOutput, survOutput = list(survOutput))
  }
    else {
      ## We consider "R", "I", and "C" outcomes.  The outcomes are grouped
      ## by type and sequential.  That is, the first "C" encountered in the
      ## response type vector is in position [[1]] in the classification output
      ## list, the second "C" encountered is in position [[2]] in the
      ## classification output list, and so on.  The same applies to the
      ## regression outputs.  We also have a mapping from the outcome slot back
      ## to the original response vector type, given by the following:
      ## Given yvar.types = c("R", "C", "R", "C", "R" , "I")
      ## regr.index[1] -> 1
      ## regr.index[2] -> 3
      ## regr.index[3] -> 5
      ## clas.index[1] -> 2
      ## clas.index[2] -> 4
      ## clas.index[3] -> 6
      ## This will pick up all "C" and "I".
      class.index <- which(yvar.types != "R")
      class.count <- length(class.index)
      regr.index <- which(yvar.types == "R")
      regr.count <- length(regr.index)
      if (class.count > 0) {
        ## Create and name the classification outputs.
        classOutput <- vector("list", class.count)
        names(classOutput) <- yvar.names[class.index]
        ## Vector to hold the number of levels in each factor response. 
        levels.count <- array(0, class.count)
        ## List to hold the names of levels in each factor response. 
        levels.names <- vector("list", class.count)
        counter <- 0
        for (i in class.index) {
            counter <- counter + 1
            ## Note that [i] is the actual index of the y-variables and not a sequential iterator.
            ## The sequential iteratior is [counter]
            levels.count[counter] <- yvar.nlevels[i]
            if (yvar.types[i] == "C") {
              ## This an unordered factor.
              ## Here, we don't know the sequence of the unordered factor list, so we identify the factor by name.
              levels.names[[counter]] <- yfactor$levels[[which(yfactor$factor == yvar.names[i])]]
            }
              else {
                ## This in an ordered factor.
                ## Here, we don't know the sequence of the ordered factor list, so we identify the factor by name.
                levels.names[[counter]] <- yfactor$order.levels[[which(yfactor$order == yvar.names[i])]]
              }
        }
        iter.start <- 0
        iter.end   <- 0
        offset <- vector("list", class.count)
        for (p in 1:length(partial.values)) {
          for (k in 1:length(m.target.idx)) {
            target.idx <- which (class.index == m.target.idx[k])
            if (length(target.idx) > 0) {
              iter.start <- iter.end
              iter.end <- iter.start + ((1 + levels.count[target.idx]) * n)
              offset[[target.idx]] <- c(offset[[target.idx]], (iter.start+1):iter.end)
            }
          }
        }
        for (i in 1:length(m.target.idx)) {
          target.idx <- which (class.index == m.target.idx[i])
          if (length(target.idx) > 0) {
            ens.names <- list(NULL, c("all", levels.names[[target.idx]]), NULL)
            ## Incoming from the native code:
            ##   type = NULL
            ##   -> of dim [length(partial.values)] x [length(1 + yvar.nlevels[.]] x [n]
            ## Outgoing to the R code:
            ##   -> of dim [n] x [1 + yvar.nlevels[.]] x [length(partial.values)]
            ensemble <- (if (!is.null(nativeOutput$partialClas))
                             array(nativeOutput$partialClas[offset[[target.idx]]],
                                   c(n, 1 + levels.count[target.idx], length(partial.values)),
                                   dimnames=ens.names) else NULL)
            classOutput[[target.idx]] <- ensemble
            remove(ensemble)
          }
        }
        rfsrcOutput <- c(rfsrcOutput, classOutput = list(classOutput))        
      }
      if (regr.count > 0) {
        ## Create and name the classification outputs.
        regrOutput <- vector("list", regr.count)
        names(regrOutput) <- yvar.names[regr.index]
        iter.start <- 0
        iter.end   <- 0
        offset <- vector("list", regr.count)
        for (p in 1:length(partial.values)) {
          for (k in 1:length(m.target.idx)) {
            target.idx <- which (regr.index == m.target.idx[k])
            if (length(target.idx) > 0) {
              iter.start <- iter.end
              iter.end <- iter.start + n
              offset[[target.idx]] <- c(offset[[target.idx]], (iter.start+1):iter.end)
            }
          }
        }
        for (i in 1:length(m.target.idx)) {
          target.idx <- which (regr.index == m.target.idx[i])
          if (length(target.idx) > 0) {
            ens.names <- list(NULL, NULL)
            ## Incoming from the native code:
            ##   type = NULL
            ##   -> of dim [length(partial.values)] x [1] x [n]
            ## Outgoing to the R code:
            ##   -> of dim [n] x [length(partial.values)]
            ensemble <- (if (!is.null(nativeOutput$partialRegr))
                             array(nativeOutput$partialRegr[offset[[target.idx]]],
                                   c(n, length(partial.values)),
                                   dimnames=ens.names) else NULL)
            regrOutput[[target.idx]] <- ensemble
            remove(ensemble)
          }
        }
        rfsrcOutput <- c(rfsrcOutput, regrOutput = list(regrOutput))
      }
    }
  ## append other useful information here
  rfsrcOutput$partial.values <- partial.values
  rfsrcOutput$yvar.names <- object$yvar.names
  rfsrcOutput$event.info <- object$event.info 
  ## return the goodies
  class(rfsrcOutput) <- c("rfsrc", "partial",   family)
  return (rfsrcOutput)
}
partial <- partial.rfsrc
##################################################################
##
##
## work-horse for converting raw data to useable format for plots
##
##
##################################################################
get.partial.plot.data <- function(o, target, m.target = NULL, granule = FALSE) {
  ##--------------------------------------------------------------
  ##
  ## coherency checks
  ##
  ##--------------------------------------------------------------
  if (sum(inherits(o, c("rfsrc", "partial"), TRUE) == c(1, 2)) != 2) {
    stop("this function only works for objects of class `(rfsrc, partial)'")
  }
  ##--------------------------------------------------------------
  ##
  ## extract the partial data, family specific
  ##
  ##--------------------------------------------------------------
  family <- class(o)[3]
  yvar.names <- o$yvar.names
  x <-  o$partial.values
  ## multivariate families
  if (family == "regr+" | family == "class+" | family == "mix+") {
    ## determine the y-outcome to be used
    if (missing(m.target)) {
      m.target <- yvar.names[1]
    }
    else {
      m.target <- match.arg(m.target, yvar.names)
    }      
    ## pull the data
    pdta  <- unlist(list(o$classOutput, o$regrOutput), recursive = FALSE)[[m.target]]
    ## convert to either regression or classification
    ## classification
    if (length(dim(pdta)) > 2) {
      ## identify target class level
      if (missing(target)) {
        target <- 1
      }
      else {
        target <- match(match.arg(target, yvar.names), yvar.names)
      }
      pdta <- pdta[, 1 + target, ]
      pdta[is.infinite(pdta)] <- NA
      yhat <- partialhat(pdta, x, granule)
    }
    ## regression
    else {
      pdta[is.infinite(pdta)] <- NA
      yhat <- partialhat(pdta, x, granule)
    }
  }
  ## regression
  else if (family == "regr") {
    pdta <- o$regrOutput[[1]]
    pdta[is.infinite(pdta)] <- NA
    yhat <- partialhat(pdta, x, granule)    
  }
  ## classification
  else if (family == "class") {
    pdta <- o$classOutput[[1]]
    yvar.names <- colnames(pdta[,,1])[-1]
    if (missing(target)) {
      target <- 1
    }
    else {
      target <- match(match.arg(target, yvar.names), yvar.names)
    }
    pdta <- pdta[, 1 + target, ]
    pdta[is.infinite(pdta)] <- NA
    yhat <- partialhat(pdta, x, granule)    
  }
  ## survival
  else if (family == "surv") {
    pdta <- o$survOutput
    pdta[is.infinite(pdta)] <- NA
    ## surv, chf
    if (length(dim(pdta)) > 2) {
      yhat <- do.call(cbind, lapply(1:dim(pdta)[2], function(k) {
        partialhat(pdta[, k, ], x, granule)
      }))
    }
    ## mortality
    else {
      yhat <- partialhat(pdta, x, granule)
    }
  }
  ## competing risk
  else if (family == "surv-CR") {
    pdta <- o$survOutput
    pdta[is.infinite(pdta)] <- NA
    ## identify target cause specific event
    if (missing(target)) {
        target <- 1
      }
      else {
        target <- match(target, o$event.info$event.type)
      }
    ## cif, chf
    if (length(dim(pdta)) > 3) {
      yhat <- do.call(cbind, lapply(1:dim(pdta)[2], function(k) {
        partialhat(pdta[, k, target, ], x, granule)
      }))
    }
    ## years-lost
    else {
      yhat <- partialhat(pdta[, target, ], x, granule)
    }
  }
  else {
    stop("family specified is not supported")
  }
  ##--------------------------------------------------------------
  ##
  ## return the goodies
  ##
  ##--------------------------------------------------------------
  if (granule) {
    x.unq <- sort(unique(x))
    n <- dim(pdta)[1]
    n.x <- length(x.unq)
    list(x = rep(x.unq, rep(n, n.x)), yhat = yhat, partial.time = o$partial.time)
  }
  else {
    list(x = x, yhat = yhat, partial.time = o$partial.time)
  }
}
##################################################################
##
##
## utility functions
##
##
##################################################################
partialhat <- function(pdta, x, granule = FALSE) {
  if (granule) {
    c(pdta)
  }
  else {
    colMeans(pdta, na.rm = TRUE)
  }
}
get.partial.bits <- function (partial.length) {
  ## Convert partial option into native code parameter.
  if (!is.null(partial.length)) {
    if (partial.length > 0) {
      bits <- 2^14
    }
      else if (partial.length == 0) {
        bits <- 0
      }
        else {
          stop("Invalid choice for 'partial.length' option:  ", partial.length)
        }
  }
    else {
      stop("Invalid choice for 'partial.length' option:  ", partial.length)
    }
  return (bits)
}
get.partial.type <- function (family, partial.type) {
  if (family == "surv") {
    ## The native code interprets "rel.freq" as "mort".  The R-side
    ## handles the difference downstream after native code exit.
    if (partial.type == "rel.freq") {
      partial.type <- "mort"
    }
    ## Warning:  Hard coded in global.h
    type <- match(partial.type, c("mort", "chf", "surv"))
  }
    else if (family == "surv-CR") {
      ## Warning:  Hard coded in global.h
      type <- match(partial.type, c("years.lost", "cif", "chf"))
    }
      else {
        type <- 0
      }
  if (is.na(type)) {
    stop("Invalid choice for 'partial.type' option:  ", partial.type)
  }
  return (type)
}

###################################################################
##
## performance measures (with random-baseline as an attribute)
##
###################################################################
.safe_div <- function(num, den) {
  num <- as.numeric(num); den <- as.numeric(den)
  ifelse(den == 0, NA_real_, num / den)
}
get.imbalanced.performance <- function(obj,
                                       prob = NULL,
                                       threshold = NULL,
                                       confusion = FALSE,
                                       robust = FALSE) {
  ## determine if a forest object is provided or two vectors (yvar, prob)
  if (is.null(prob)) {
    if (class(obj)[1] != "rfsrc") {
      stop("obj must be a forest object")
    } else {
      yvar <- obj$yvar
      if (!is.null(obj$predicted.oob) && !all(is.na(obj$predicted.oob))) {
        prob <- obj$predicted.oob
      } else {
        prob <- obj$predicted
      }
    }
  } else {
    ## user has supplied (yvar, prob)
    yvar <- obj
  }
  ## if this is not a two-class problem, return NULL
  if (!is.factor(yvar) || length(levels(yvar)) != 2) {
    return(NULL)
  }
  ## compute
  perf.o <- get.imbalanced.performance.workhorse(
    yvar, prob,
    threshold = threshold,
    confusion = confusion,
    robust = robust
  )
  ## carry the baseline attribute through compact printing
  rand.attr <- attr(perf.o, "rand")
  if (!confusion) {
    v <- unlist(perf.o)        # default behavior preserved
    attr(v, "rand") <- rand.attr
    class(v) <- c("imbalanced.performance", class(v))
    v
  } else {
    attr(perf.o, "rand") <- rand.attr
    class(perf.o) <- c("imbalanced.performance", class(perf.o))
    perf.o
  }
}
get.imbalanced.performance.workhorse <- function (yvar, prob,
                                                  threshold = NULL,
                                                  confusion = FALSE,
                                                  robust = FALSE) {
  ## not two-class -> NULL
  if (!is.factor(yvar) || length(levels(yvar)) != 2) return(NULL)
  ## recode to {0,1}: 0=majority, 1=minority
  y.frq <- table(yvar)
  if (length(y.frq) != 2 || any(is.na(y.frq))) return(NULL)
  class.labels <- names(y.frq)
  minority <- which.min(y.frq)
  majority <- setdiff(1:2, minority)
  n.eff    <- sum(y.frq)
  pihat    <- as.numeric(y.frq[minority]) / n.eff
  iratio   <- max(y.frq, na.rm = TRUE) / min(y.frq, na.rm = TRUE)
  y <- rep(0, length(yvar))
  y[yvar == class.labels[minority]] <- 1
  y <- factor(y, levels = c(0,1))
  ## map probabilities --> (majority, minority)
##
## Important: some callers may pass a 2-column *data.frame* of probabilities.
## Downstream metrics (brier/auc/logloss) expect a numeric matrix and may
## fail on data.frames (which are lists).  We coerce to a double matrix here.
if (!is.null(ncol(prob)) && ncol(prob) == 2) {
  prob.matx <- prob[, c(majority, minority), drop = FALSE]
  ## coerce to numeric matrix (handles data.frame inputs robustly)
  prob.matx <- as.matrix(prob.matx)
  storage.mode(prob.matx) <- "double"
  ## minority probability is the 2nd column after re-ordering
  prob <- prob.matx[, 2]
  ## clamp to [0,1]
  prob <- pmin(pmax(prob, 0), 1)
  prob.matx[, 2] <- prob
  prob.matx[, 1] <- pmin(pmax(prob.matx[, 1], 0), 1)
} else {
  ## user supplied a vector / 1-col object => assume it is minority prob
  if (is.data.frame(prob) && ncol(prob) == 1) {
    prob <- prob[[1]]
  }
  prob <- as.numeric(prob)
  prob <- pmin(pmax(prob, 0), 1)
  prob.matx <- cbind(1 - prob, prob)
  storage.mode(prob.matx) <- "double"
}
colnames(prob.matx) <- levels(y)
  ## threshold
  if (is.null(threshold)) threshold <- as.numeric(pihat) else threshold <- as.numeric(threshold)
  threshold <- max(0, min(1, threshold))
  ## confusion at threshold
  yhat <- factor(1 * (prob >= threshold), levels = c(0, 1))
  confusion.matx <- table(y, yhat)
  N <- sum(confusion.matx)
  if (N > 0) {
    TN <- confusion.matx[1, 1]; FP <- confusion.matx[1, 2]
    FN <- confusion.matx[2, 1]; TP <- confusion.matx[2, 2]
    if (robust) {
      sens <- .safe_div(1 + TP, 1 + FN + TP)
      spec <- .safe_div(1 + TN, 1 + TN + FP)
      prec <- .safe_div(1 + TP, 1 + TP + FP)
      npv  <- .safe_div(1 + TN, 1 + TN + FN)
    } else {
      sens <- .safe_div(TP, FN + TP)
      spec <- .safe_div(TN, TN + FP)
      prec <- .safe_div(TP, TP + FP)
      npv  <- .safe_div(TN, TN + FN)
    }
    misclass   <- (FP + FN) / N
    F1         <- ifelse((prec + sens) > 0, 2 * (prec * sens) / (prec + sens), NA_real_)
    F1mod      <- ifelse(all(is.finite(c(sens, spec, prec, npv))) &&
                         all(c(sens, spec, prec, npv) > 0),
                         4 / (1/sens + 1/spec + 1/prec + 1/npv), NA_real_)
    gmean      <- sqrt(sens * spec)
    F1gmean    <- (F1 + gmean) / 2
    F1modgmean <- (F1mod + gmean) / 2
  } else {
    sens <- spec <- prec <- npv <- misclass <-
      F1 <- F1mod <- gmean <- F1gmean <- F1modgmean <- NA_real_
  }
  ## probability-based metrics
  brier       <- get.brier.error(y, prob.matx, normalized = FALSE)
  brier.norm  <- get.brier.error(y, prob.matx)   # normalized per your function
  auc         <- get.auc(y, prob.matx)
  logloss     <- get.logloss(y, prob.matx)
  prO         <- get.pr.auc(y, prob)             # vector or 1-col matrix is OK
  pr.auc      <- prO[1]
  pr.auc.rand <- prO[2]
  ## ------------------------------
  ## Analytical random baseline
  ## ------------------------------
  t  <- threshold
  pi <- pihat
  TP.e <- n.eff * pi * (1 - t)
  FN.e <- n.eff * pi * t
  TN.e <- n.eff * (1 - pi) * t
  FP.e <- n.eff * (1 - pi) * (1 - t)
  if (robust) {
    sens.rand <- (1 + TP.e) / (1 + FN.e + TP.e)
    spec.rand <- (1 + TN.e) / (1 + TN.e + FP.e)
    prec.rand <- (1 + TP.e) / (1 + TP.e + FP.e)
    npv.rand  <- (1 + TN.e) / (1 + TN.e + FN.e)
  } else {
    sens.rand <- 1 - t
    spec.rand <- t
    prec.rand <- pi
    npv.rand  <- 1 - pi
  }
  misclass.rand   <- (FP.e + FN.e) / n.eff                     # == pi*t + (1-pi)*(1-t)
  F1.rand         <- ifelse((prec.rand + sens.rand) > 0,
                            2 * (prec.rand * sens.rand) / (prec.rand + sens.rand), NA_real_)
  F1mod.rand      <- ifelse(all(is.finite(c(sens.rand, spec.rand, prec.rand, npv.rand))) &&
                            all(c(sens.rand, spec.rand, prec.rand, npv.rand) > 0),
                            4 / (1/sens.rand + 1/spec.rand + 1/prec.rand + 1/npv.rand), NA_real_)
  gmean.rand      <- sqrt(sens.rand * spec.rand)
  F1gmean.rand    <- (F1.rand + gmean.rand) / 2
  F1modgmean.rand <- (F1mod.rand + gmean.rand) / 2
  auc.rand        <- 0.5
  ## Brier baselines aligned to your get.brier.error():
  ##   - unnormalized (J=2): E[(p - Y)^2] = 1/3
  ##   - normalized: factor is 4x the unnormalized in binary case
  brier.rand      <- 1/3
  brier.norm.rand <- 4 * brier.rand
  logloss.rand    <- 1.0
  ## assemble primary output (unchanged fields)
  rO <- list(n.majority = as.numeric(y.frq[majority]),
             n.minority = as.numeric(y.frq[minority]),
             iratio = iratio,
             threshold = threshold,
             sens = sens,
             spec = spec,
             prec = prec,
             npv = npv,
             misclass = misclass,
             brier = brier,
             brier.norm = brier.norm,
             auc = auc,
             logloss = logloss,
             F1 = F1,
             F1mod = F1mod,
             pr.auc.rand = pr.auc.rand,
             pr.auc = pr.auc,
             F1gmean = F1gmean,
             F1modgmean = F1modgmean,
             gmean = gmean)
  if (confusion) {
    class.error <- 1 - diag(confusion.matx) / rowSums(confusion.matx, na.rm = TRUE)
    rO$confusion <- cbind(confusion.matx, class.error = round(class.error, 4))
  }
  ## attach random-baseline as an attribute so default print stays compact
  rand.metrics <- c(
    sens.rand = sens.rand,
    spec.rand = spec.rand,
    prec.rand = prec.rand,
    npv.rand  = npv.rand,
    misclass.rand = misclass.rand,
    F1.rand = F1.rand,
    F1mod.rand = F1mod.rand,
    F1gmean.rand = F1gmean.rand,
    F1modgmean.rand = F1modgmean.rand,
    gmean.rand = gmean.rand,
    auc.rand = auc.rand,
    brier.rand = brier.rand,
    brier.norm.rand = brier.norm.rand,
    logloss.rand = logloss.rand
  )
  confusion.rand.expected <- matrix(c(TN.e, FP.e, FN.e, TP.e), nrow = 2, byrow = TRUE)
  dimnames(confusion.rand.expected) <- list(Truth = levels(y), Pred = levels(y))
  attr(rO, "rand") <- list(
    metrics = rand.metrics,
    confusion.expected = confusion.rand.expected
  )
  rO
}
# pretty printer for imbalanced performance (model vs random baseline)
print.imbalanced.performance <- function(x, digits = 4, show.confusion = TRUE, ...) {
  # helpers to extract by name from list or named numeric
  .has <- function(nm) nm %in% names(x)
  .get <- function(nm) {
    if (.has(nm)) {
      if (is.list(x)) x[[nm]] else as.numeric(x[nm])
    } else NA_real_
  }
  .fmt <- function(z) formatC(z, digits = digits, format = "f")
  .fmt_pct <- function(z) ifelse(is.na(z), NA, paste0(.fmt(z), " %"))
  rand <- attr(x, "rand")
  rand.metrics <- if (!is.null(rand)) rand$metrics else NULL
  # prevalence pi (prefer prec.rand, else pr.auc.rand)
  pi.hat <- if (!is.null(rand.metrics) && "prec.rand" %in% names(rand.metrics)) {
    as.numeric(rand.metrics["prec.rand"])
  } else {
    .get("pr.auc.rand")
  }
  # header ----------------------------------------------------------
  cat("Imbalanced classification performance (model vs random baseline)\n")
  cat(sprintf(" n.majority = %s   n.minority = %s   iratio = %s   threshold = %s   prevalence  = %s\n\n",
              .get("n.majority"),
              .get("n.minority"),
              .fmt(.get("iratio")),
              .fmt(.get("threshold")),
              .fmt(pi.hat)))
  # metrics to display (grouped for readability)
  rate_metrics   <- c("sens","spec","prec","npv","F1","F1mod","gmean","F1gmean","F1modgmean")
  auc_metrics    <- c("auc","pr.auc")
  error_metrics  <- c("misclass","brier","brier.norm","logloss")
  all_metrics    <- c(rate_metrics, auc_metrics, error_metrics)
  # mapping to baseline names
  base_map <- stats::setNames(paste0(all_metrics, ".rand"), all_metrics)
  base_map["pr.auc"] <- "pr.auc.rand"  # special case lives in main list
  # build table -----------------------------------------------------
  model_vals <- sapply(all_metrics, .get)
  base_vals  <- sapply(all_metrics, function(m) {
    nm <- base_map[[m]]
    if (m == "pr.auc") {
      .get(nm)                       # "pr.auc.rand" is in the main object
    } else if (!is.null(rand.metrics) && nm %in% names(rand.metrics)) {
      as.numeric(rand.metrics[[nm]]) # most baselines live in the attribute
    } else {
      NA_real_
    }
  })
  # gains: higher-is-better vs lower-is-better
  hib <- c(rate_metrics, auc_metrics)                   # higher is better
  lib <- error_metrics                                  # lower is better
  delta <- rep(NA_real_, length(all_metrics)); names(delta) <- all_metrics
  gain  <- delta
  for (m in all_metrics) {
    mv <- model_vals[[m]]; bv <- base_vals[[m]]
    if (is.na(mv) || is.na(bv)) next
    if (m %in% lib) {
      delta[m] <- bv - mv
      gain[m]  <- if (bv != 0) (bv - mv) / bv * 100 else NA_real_
    } else {
      delta[m] <- mv - bv
      gain[m]  <- if (abs(bv) > .Machine$double.eps) (mv - bv) / abs(bv) * 100 else NA_real_
    }
  }
  df <- data.frame(
    Metric   = all_metrics,
    Model    = .fmt(model_vals[all_metrics]),
    Baseline = .fmt(base_vals[all_metrics]),
    Delta    = .fmt(delta[all_metrics]),
    Gain     = .fmt_pct(gain[all_metrics]),
    row.names = NULL,
    check.names = FALSE
  )
  # visual grouping by inserting blank lines (printed as separate blocks)
  print(df[df$Metric %in% rate_metrics,   , drop = FALSE], row.names = FALSE, right = TRUE)
  cat("\n")
  print(df[df$Metric %in% auc_metrics,    , drop = FALSE], row.names = FALSE, right = TRUE)
  cat("\n")
  print(df[df$Metric %in% error_metrics,  , drop = FALSE], row.names = FALSE, right = TRUE)
  # confusion matrices ---------------------------------------------
  if (show.confusion && is.list(x) && !is.null(x$confusion)) {
    cat("\nConfusion matrix (model):\n")
    print(x$confusion)
    ce <- if (!is.null(rand) && !is.null(rand$confusion.expected)) rand$confusion.expected else NULL
    if (!is.null(ce)) {
      cat("\nExpected confusion (random baseline):\n")
      ce_err <- c(
        ifelse(sum(ce[1,]) > 0, 1 - ce[1,1]/sum(ce[1,]), NA_real_),
        ifelse(sum(ce[2,]) > 0, 1 - ce[2,2]/sum(ce[2,]), NA_real_)
      )
      ce_out <- cbind(round(ce, 2), class.error = round(ce_err, 4))
      print(ce_out)
    }
  }
  invisible(x)
}
###################################################################
##
##
## rfq threshold
##
##
###################################################################
get.rfq.threshold <- function(y) {
  frq <- table(y)
  if (length(frq) != 2) {
    return(NULL)
  }
  as.numeric(min(frq, na.rm = TRUE) / sum(frq, na.rm = TRUE))
}
###################################################################
##
##
## optimize the gmean (or balanced measure) by the threshold
##
##
###################################################################
# internal: fast, vectorized sweep over thresholds (avoids mclapply)
.get.imbalanced.optimize.sweep <- function(yvar, prob, thresholds) {
  ## if this is not a two-class problem, return NULL
  if (!is.factor(yvar) || length(levels(yvar)) != 2) {
    return(NULL)
  }
  ## recode to {0,1}: 0=majority, 1=minority (same as get.imbalanced.performance.workhorse)
  y.frq <- table(yvar)
  if (length(y.frq) != 2 || any(is.na(y.frq))) {
    return(NULL)
  }
  class.labels <- names(y.frq)
  minority <- which.min(y.frq)
  y <- rep.int(0L, length(yvar))
  y[yvar == class.labels[minority]] <- 1L
## extract minority probabilities (robust to matrices/data.frames/vectors)
if (!is.null(ncol(prob))) {
  if (ncol(prob) == 2) {
    p <- prob[, minority]
  } else if (ncol(prob) == 1) {
    p <- prob[, 1]
  } else {
    ## unexpected column count: fall back to first column
    p <- prob[, 1]
  }
} else {
  p <- prob
}
p <- as.numeric(p)
p <- pmin(pmax(p, 0), 1)
  ## mimic table(y, factor(prob>=th)) behavior: drop NA probabilities
  ok <- !is.na(p)
  if (!any(ok)) {
    return(data.frame(
      threshold = thresholds,
      gmean = NA_real_,
      F1 = NA_real_,
      F1mod = NA_real_,
      F1modgmean = NA_real_
    ))
  }
  y <- y[ok]
  p <- p[ok]
  ## sort by probability (ascending)
  o <- order(p)
  p <- p[o]
  y <- y[o]
  ## prefix sums (length n+1 to handle k=0)
  n <- length(p)
  cum_pos <- c(0L, cumsum(y))
  cum_neg <- c(0L, cumsum(1L - y))
  total_pos <- cum_pos[n + 1L]
  total_neg <- cum_neg[n + 1L]
  ## for each threshold t: k = number of probs < t  (since prediction is prob >= t)
  k <- findInterval(thresholds, p, left.open = TRUE)
  idx <- k + 1L
  FN <- cum_pos[idx]
  TN <- cum_neg[idx]
  TP <- total_pos - FN
  FP <- total_neg - TN
  sens <- .safe_div(TP, FN + TP)
  spec <- .safe_div(TN, TN + FP)
  prec <- .safe_div(TP, TP + FP)
  npv  <- .safe_div(TN, TN + FN)
  F1 <- ifelse((prec + sens) > 0, 2 * (prec * sens) / (prec + sens), NA_real_)
  ok2 <- is.finite(sens) & is.finite(spec) & is.finite(prec) & is.finite(npv) &
    (sens > 0) & (spec > 0) & (prec > 0) & (npv > 0)
  F1mod <- ifelse(ok2, 4 / (1/sens + 1/spec + 1/prec + 1/npv), NA_real_)
  gmean <- sqrt(sens * spec)
  F1modgmean <- (F1mod + gmean) / 2
  data.frame(
    threshold = thresholds,
    gmean = gmean,
    F1 = F1,
    F1mod = F1mod,
    F1modgmean = F1modgmean
  )
}
get.imbalanced.optimize <- function(obj,
                                    prob = NULL,
                                    newdata = NULL,
                                    measure = c("gmean", "F1", "F1mod", "F1modgmean"),
                                    ngrid = 1000,
                                    plot.it = TRUE) {
  ## determine if a forest object is provided or two vectors (yvar, prob)
  if (is.null(prob)) {
    if (class(obj)[1] != "rfsrc") {
      stop("obj must be a forest object")
    }
    ## object is a forest ---> parse for (yvar, prob)
    else {
      ## if newdata is provided, we swap the grow object with the predict object
      if (!is.null(newdata)) {
        obj <- predict(obj, newdata)
      }
      yvar <- obj$yvar
      if (!is.null(obj$predicted.oob) && !all(is.na(obj$predicted.oob))) {
        prob <- obj$predicted.oob
      }
      else {
        prob <- obj$predicted
      }
    }
  }
  ## user has supplied (yvar, prob)
  else {
    yvar <- obj
  }
  ## check that measure requested is coherent
  measure <- match.arg(measure, c("gmean", "F1", "F1mod", "F1modgmean"))
  thresholds <- seq(0, 1, length = ngrid)
  ## fast, vectorized threshold sweep (avoids mclapply and is usually faster)
  sweep <- .get.imbalanced.optimize.sweep(yvar, prob, thresholds)
  if (is.null(sweep)) {
    return(NULL)
  }
  best <- which.max(sweep[[measure]])
  best.threshold <- sweep$threshold[best]
  if (plot.it) {
    par(mfrow = c(2,2))
    pt <- sweep$threshold < 3 * best.threshold
    plot(sweep$threshold[pt], sweep$gmean[pt], xlab = "threshold", ylab = "gmean")
    abline(v = best.threshold, lty = 2, col = "blue")
    plot(sweep$threshold[pt], sweep$F1[pt], xlab = "threshold", ylab = "F1")
    abline(v = best.threshold, lty = 2, col = "blue")
    plot(sweep$threshold[pt], sweep$F1mod[pt], xlab = "threshold", ylab = "F1mod")
    abline(v = best.threshold, lty = 2, col = "blue")
    plot(sweep$threshold[pt], sweep$F1modgmean[pt], xlab = "threshold", ylab = "F1modgmean")
    abline(v = best.threshold, lty = 2, col = "blue")
  }
  ## return full performance metrics at the optimal threshold (same as previous behavior)
  perf <- get.imbalanced.performance(yvar, prob, threshold = best.threshold)
  if (is.null(perf)) {
    return(NULL)
  }
  out <- data.frame(as.list(perf))
  rownames(out) <- best
  out
}
###################################################################
##
##
## precision recall utilities (these are internal functions)
## assumes y=0--> majority, y=1--> minority
##
##
###################################################################
get.pr.auc <- function(truth, yhat) {
  ## failsafe (in case directly called)
  if (is.factor(truth) && !(sum(levels(truth) == c(0, 1)) == 2)) {
    y.frq <- table(truth)
    class.labels <- names(y.frq)
    minority <- which.min(y.frq)
    majority <- setdiff(1:2, minority)
    y <- rep(NA, length(truth))
    y[truth==class.labels[minority]] <- 1
    y[truth==class.labels[majority]] <- 0
    truth <- y
  }
  ## if yhat is a matrix, extract probabilities for minority class
  if (!is.null(ncol(yhat)) && ncol(yhat) == 2) {  
    yhat <- yhat[, which.min(table(truth))]  
  }
  ## create pr (x,y) data
  x <- yhat[truth == 1]
  y <- yhat[truth == 0]
  rO <- c(NA, NA)
  if (length(x) > 0 && length(y) > 0) {
    ## precision recall with random baseline
    pr.o <- tryCatch({get.pr.workhorse(x, y, rand.compute = TRUE)}, error = function(ex) {NULL})
    if (!is.null(pr.o)) {
      rO <- c(pr.o$auc.integral, pr.o$auc.integral.rand)
    }
  }
  rO
}
get.pr.curve <- function(truth, yhat) {
  ## failsafe (in case directly called)
  if (is.factor(truth) && !(sum(levels(truth) == c(0, 1)) == 2)) {
    y.frq <- table(truth)
    class.labels <- names(y.frq)
    minority <- which.min(y.frq)
    majority <- setdiff(1:2, minority)
    y <- rep(NA, length(truth))
    y[truth==class.labels[minority]] <- 1
    y[truth==class.labels[majority]] <- 0
    truth <- y
  }
  ## if yhat is a matrix, extract probabilities for minority class
  if (!is.null(ncol(yhat)) && ncol(yhat) == 2) {  
    yhat <- yhat[, which.min(table(truth))]  
  }
  ## create pr (x,y) data
  x <- yhat[truth == 1]
  y <- yhat[truth == 0]
  rO <- NULL
  if (length(x) > 0 && length(y) > 0) {
    ## precision recall with random baseline
    pr.o <- tryCatch({get.pr.workhorse(x, y, curve = TRUE)}, error = function(ex) {NULL})
    if (!is.null(pr.o)) {
      rO <- pr.o$curve
    }
  }
  rO
}
## copied from PRROC library - we gratefully acknowledge the developer for this function
get.pr.workhorse <- function(scores.class0, scores.class1 = scores.class0, weights.class0 = NULL,
                     weights.class1 = {if (is.null(weights.class0)) {NULL} else {1 - weights.class0}},
                     sorted = FALSE,
                     minStepSize = min(1, ifelse(is.null(weights.class0), 1, sum(weights.class0)/100)),
                     rand.compute = FALSE, curve = FALSE) 
{
  ## preliminary processing
  if (!sorted) {
    o0 <- order(scores.class0)
    scores.class0 <- scores.class0[o0]
    if (!is.null(weights.class0)) {
      weights.class0 <- weights.class0[o0]
    }
    o1 <- order(scores.class1)
    scores.class1 <- scores.class1[o1]
    if (!is.null(weights.class1)) {
      weights.class1 <- weights.class1[o1]
    }
  }
  sorted.scores.class0 <- scores.class0
  sorted.scores.class1 <- scores.class1
  ## main workhorse
  weights.class0 <- c(rep(1, length(sorted.scores.class0)), 
                      rep(0, length(sorted.scores.class1)))
  sorted.scores.class0 <- c(sorted.scores.class0, sorted.scores.class1)
  o0 <- order(sorted.scores.class0)
  sorted.scores.class0 <- sorted.scores.class0[o0]
  weights.class0 <- weights.class0[o0]
  weights.class1 <- 1 - weights.class0
  sorted.scores.class1 <- sorted.scores.class0
  all.scores <- sorted.scores.class0
  all.weights.pos <- weights.class0
  all.weights.neg <- weights.class1
  o <- order(all.scores, decreasing = T)
  all.scores <- all.scores[o]
  all.weights.pos <- all.weights.pos[o]
  all.weights.neg <- all.weights.neg[o]
  cum.weights.pos <- cumsum(all.weights.pos)
  cum.weights.neg <- cumsum(all.weights.neg)
  cum.use <- c(all.scores[-length(all.scores)] != all.scores[-1], TRUE)
  all.scores <- all.scores[cum.use]
  cum.weights.pos <- cum.weights.pos[cum.use]
  cum.weights.neg <- cum.weights.neg[cum.use]
  r.fg <- sum(all.weights.pos)
  tp <- cum.weights.pos
  fp <- cum.weights.neg
  tp.prev <- c(0, cum.weights.pos[-length(cum.weights.pos)])
  fp.prev <- c(0, cum.weights.neg[-length(cum.weights.neg)])
  h <- (fp - fp.prev)/(tp - tp.prev)
  a <- 1 + h
  b <- (fp.prev - h * tp.prev)/r.fg
  h[tp == tp.prev] <- 1
    a[tp == tp.prev] <- 1
  b[tp == tp.prev] <- 0
  v <- (tp/r.fg - tp.prev/r.fg - b/a * (log(a * tp/r.fg + b) - log(a * tp.prev/r.fg + b)))/a
  v2 <- (tp/r.fg - tp.prev/r.fg)/a
  v[b == 0] <- v2[b == 0]
  vals <- v
  auc.integral <- sum(vals)
  res <- list(auc.integral = auc.integral)
  rand.auc <- pr.curve <- NULL
  ## random auc
  if (rand.compute) {
    if (is.null(weights.class0)) {
      rand.auc <- length(sorted.scores.class0)/(length(sorted.scores.class0) + 
                                                length(sorted.scores.class1))
    }
    else {
      rand.auc <- sum(weights.class0)/sum(weights.class0 + weights.class1)
    }
  }
  ## pr curve
  if (curve) {
    minStepSize.2 <- minStepSize/r.fg
    min.curve <- cbind(tp/r.fg, tp/(tp + fp), all.scores)
    idxs <- which((tp - tp.prev)/r.fg > minStepSize.2 & tp/(tp + fp) != tp.prev/(tp.prev + fp.prev))
    idxs <- idxs[idxs > 1]
    if (length(idxs) > 0) {
      m <- sapply(idxs, function(i) {
        x <- seq(0, min.curve[i, 1] - min.curve[i - 1, 1], by = minStepSize.2)
        sns <- min.curve[i - 1, 1] + x
        prcs <- (min.curve[i - 1, 1] + x)/(min.curve[i - 1, 1] + x + fp[i - 1]/r.fg + h[i] * x)
        temp <- rbind(sns, prcs, rep(all.scores[i], length(x)))
        temp
      })
      m <- matrix(unlist(m), ncol = 3, byrow = T)
      m <- rbind(min.curve, m)
    }
    else {
      m <- min.curve
    }
    m <- m[order(m[, 1], -m[, 3], decreasing = T), ]
    m <- rbind(c(1, m[1, 2:3]), m, c(0, m[nrow(m), 2:3]))
    dimnames(m) <- c(NULL, NULL)
    colnames(m) <- c("recall", "precision", "threshold")
    pr.curve <- m
  }
  c(res, list(auc.integral.rand = rand.auc, curve = pr.curve))
}
###################################################################
##
##
## one versus one  OVO
##
##
###################################################################
# Function to train one-vs-one classifiers using randomForestSRC's imbalanced
train_ovo <- function(formula, data, verbose = TRUE) {
  label_col <- all.vars(formula)[1]
  class_pairs <- combn(as.character(unique(data[[label_col]])), 2, simplify = FALSE)
  classifiers <- list()
  cat("training one-vs-one\n")
  for (pair in class_pairs) {
    # Verbose output
    if (verbose) print(pair)
    # Filter data for the current pair of classes
    binary_data <- data[data[[label_col]] %in% pair, ]
    binary_data[[label_col]] <- droplevels(binary_data[[label_col]])
    # No training if two classes are not present
    if (length(levels(binary_data[, label_col])) < 2) {
      model <- NULL
    }
    # Train random forest classifier with imbalanced function
    else {
      model <- tryCatch({suppressWarnings(imbalanced(formula, data = binary_data))},
                      error = function(ex) {NULL})
    }
    # Assign the classifier
    classifiers[[paste(pair, collapse = "_")]] <- model
  }
  classifiers <- classifiers[lengths(classifiers) != 0]
  attr(classifiers, "formula") <- formula
  attr(classifiers, "data") <- data
  return(classifiers)
}
train.ovo <- train_ovo
# Function to predict using one-vs-one classifiers
predict_ovo <- function(classifiers, data = NULL, verbose = TRUE) {
  ## true testing?
  test <- TRUE
  if (is.null(data)) {
    data <- attr(classifiers, "data")
    test <- FALSE
  }
  ## set up the vote matrix
  formula <- attr(classifiers, "formula")
  label_col <- all.vars(formula)[1]
  votes <- matrix(0, nrow = nrow(data), ncol = length(unique(data[[label_col]])))
  colnames(votes) <- levels(droplevels(data[[label_col]]))
  cat("prediction one-vs-one\n")
  for (pair in names(classifiers)) {
    # Split the pair name to get class labels
    classes <- unlist(strsplit(pair, "_"))
    # Verbose output
    if (verbose) print(classes)
    # Get the relevant model
    model <- classifiers[[pair]]
    # Get predictions for the entire data
    full_predictions <- predict(model, data[, model$xvar.names, drop = FALSE])$class
    # Use OOB predictions for the training data (this is not a true train/test scenario)
    if (!test) {
      binary_data_indices <- which(data[[label_col]] %in% classes)
      full_predictions[binary_data_indices] <- model$class.oob
    }
    # Assign votes based on predictions
    for (i in 1:nrow(data)) {
      cl <- which(colnames(votes) == full_predictions[i])
      if (length(cl) > 0) {
        votes[i, cl] <- votes[i, cl] + 1
      }
    }
  }
  # Determine final class by majority vote
  final_predictions <- apply(votes, 1, function(vote) colnames(votes)[which.max(vote)])
  return(final_predictions)
}
## predict.ovo <- predict_ovo
# Function to evaluate the model and calculate AUC and PR-AUC
one.versus.one <- function(formula, data, verbose = TRUE) {
  # Clean the data
  o <- rfsrc.cart(formula, data, nodedepth = 0, perf.type = "none")
  if (o$family != "class") {
    stop("this function only works for classification families\n")
  }
  data <- droplevels(data.frame(o$yvar, o$xvar))
  colnames(data) <- c(o$yvar.names, o$xvar.names)
  # Train classifiers
  classifiers <- train_ovo(formula, data, verbose)
  attr(classifiers, "formula") <- formula
  # Predict on the entire data
  predictions <- predict_ovo(classifiers, verbose = verbose)
  # Evaluate the predictions
  confusionMatrixCustom(predictions, o$yvar)
}
ovo <- one.versus.one
confusionMatrixCustom <- function(predictions, actuals) {
  # Ensure that predictions and actuals are factors
  predictions <- factor(predictions, levels = levels(actuals))
  # Create the confusion matrix
  confusion <- table(Predicted = predictions, Actual = actuals)
  # Calculate overall accuracy
  accuracy <- sum(diag(confusion)) / sum(confusion)
  n <- sum(confusion)
  # Calculate the 95% confidence interval for accuracy
  se <- sqrt(accuracy * (1 - accuracy) / n)
  ci <- c(accuracy - qnorm(0.975) * se, accuracy + qnorm(0.975) * se)
  # Calculate the no information rate (NIR)
  nir <- max(table(actuals) / n)
  # Calculate the p-value for accuracy greater than NIR
  p_value <- binom.test(sum(diag(confusion)), n, p = nir, alternative = "greater")$p.value
  # Initialize metrics
  sensitivity <- numeric(ncol(confusion))
  specificity <- numeric(ncol(confusion))
  pos_pred_value <- numeric(ncol(confusion))
  neg_pred_value <- numeric(ncol(confusion))
  prevalence <- numeric(ncol(confusion))
  detection_rate <- numeric(ncol(confusion))
  detection_prevalence <- numeric(ncol(confusion))
  balanced_accuracy <- numeric(ncol(confusion))
  for (i in 1:ncol(confusion)) {
    sensitivity[i] <- confusion[i, i] / sum(confusion[, i])
    specificity[i] <- sum(confusion[-i, -i]) / sum(confusion[, -i])
    pos_pred_value[i] <- confusion[i, i] / sum(confusion[i, ])
    neg_pred_value[i] <- sum(confusion[-i, -i]) / sum(confusion[-i, ])
    prevalence[i] <- sum(confusion[, i]) / n
    detection_rate[i] <- confusion[i, i] / n
    detection_prevalence[i] <- sum(confusion[i, ]) / n
    balanced_accuracy[i] <- (sensitivity[i] + specificity[i]) / 2
  }
  # Compile results into a list
  results <- list(
    confusion = confusion,
    overall = c(
      Accuracy = accuracy,
      "95% CI" = paste0("(", round(ci[1], 4), ", ", round(ci[2], 4), ")"),
      "No Information Rate" = nir,
      "P-Value [Acc > NIR]" = format_p_value(p_value)
    ),
    byClass = data.frame(rbind(
      Sensitivity = sensitivity,
      Specificity = specificity,
      Pos_Pred_Value = pos_pred_value,
      Neg_Pred_Value = neg_pred_value,
      Prevalence = prevalence,
      Detection_Rate = detection_rate,
      Detection_Prevalence = detection_prevalence,
      Balanced_Accuracy = balanced_accuracy
    ))
  )
  colnames(results$byClass) <- levels(actuals)
  # Print results
  cat("Overall Statistics\n")
  cat("\n")
  cat(sprintf("               Accuracy : %.4s\n", results$overall["Accuracy"]))
  cat(sprintf("                 95%% CI : %s\n", results$overall["95% CI"]))
  cat(sprintf("    No Information Rate : %.4s\n", results$overall["No Information Rate"]))
  cat(sprintf("    P-Value [Acc > NIR] : %s\n", results$overall["P-Value [Acc > NIR]"]))
  cat("\n")
  cat("Confusion Matrix\n")
  print(results$confusion)
  cat("\n")
  cat("Class Statistics\n")
  print(results$byClass)
  # Return
  invisible(results)
}
# Helper function to format p-values
format_p_value <- function(p_value) {
  if (p_value < 2.2e-16) {
    return("< 2.2e-16")
  } else {
    return(sprintf("%.4g", p_value))
  }
}

plot.quantreg.rfsrc <- function(x, prbL = .25, prbU = .75,
                                m.target = NULL, crps = TRUE, subset = NULL,
                                xlab = NULL, ylab = NULL, ...) {
  ##--------------------------------------------------------------
  ##
  ## prelimary checks
  ##
  ##--------------------------------------------------------------
  ## does not apply to predict objects without y
  if (sum(grepl("predict", class(x))) > 0 && is.null(x$yvar)) {
    stop("no yvar present in quantreg predict object")
  }
  ## check probs
  if (prbL < 0 || prbU > 1) {
    stop("requested probabilities must lie in (0, 1)")
  }
  if (prbL >= prbU) {
    stop("prbL must be less than prbU")
  }
  prbM <- max(prbL, .5)
  if (prbM == prbL) {
    prbM <- (prbL + prbU) / 2
  }
  ##--------------------------------------------------------------
  ##
  ## subset assignment
  ##
  ##--------------------------------------------------------------
  if (is.null(subset)) {
    subset <- 1:x$n
  }
  else {
    if (is.logical(subset)) {
      subset <- which(subset)
    }
    subset <- subset[subset >=1 & subset <= x$n]
  }
  if (length(subset) == 0) {
    stop("requested subset analysis has subset that is empty")
  }
  ##--------------------------------------------------------------
  ##
  ## assemble the quantile data for plotting
  ##
  ##--------------------------------------------------------------
  quant.dat <- get.quantile(x, c(prbL, prbM, prbU), FALSE)
  ## we have a univariate quantile regression object
  if (length(quant.dat) == 1) {
    y.names <- names(quant.dat)[1]
    quant.dat <- quant.dat[[1]][subset,, drop = FALSE]
    y <- x$yvar[subset]
    if (crps) crps.dat <- get.quantile.crps(x, subset = subset)
  }
  ## the quantile regression object is multivariate
  else {
    if (is.null(m.target) || length(intersect(m.target, names(quant.dat))) == 0) {
      y.names <- names(quant.dat)[1]
      quant.dat <- quant.dat[[1]][subset,, drop = FALSE]
      y <- x$yvar[, y.names][subset]
      if (crps) crps.dat <- get.quantile.crps(x, subset = subset)[[1]]
    }
    else {
      y.names <- m.target
      quant.dat <- quant.dat[[y.names]][subset,, drop = FALSE]
      y <- x$yvar[, y.names][subset]
      if (crps) crps.dat <- get.quantile.crps(x, subset = subset)[[y.names]]
    }
  }
  ##--------------------------------------------------------------
  ##
  ## quantile regression plot
  ##
  ##--------------------------------------------------------------
  if (is.null(xlab)) {
    xlab <- y.names
  }
  if (is.null(ylab)) {
    ylab <- "Target Quantiles"
  }
  jitter.y <- jitter(y, 10)
  rng <- range(c(y, quant.dat, jitter.y))
  plot(rng, rng, xlab = xlab, ylab = ylab, type = "n")
  points(jitter.y, quant.dat[, 2], pch = 15, col = 4, cex = 0.75)
  segments(jitter.y, quant.dat[, 2], jitter.y, quant.dat[, 1], col = "grey")
  segments(jitter.y, quant.dat[, 2], jitter.y, quant.dat[, 3], col = "grey")
  points(jitter.y, quant.dat[, 1], pch = "-", cex = 1)
  points(jitter.y, quant.dat[, 3], pch = "-", cex = 1)
  abline(0, 1, lty = 2, col = 2)
  ##--------------------------------------------------------------
  ##
  ## inset the CRPS
  ##
  ##--------------------------------------------------------------
  if (crps) {
    old.par <- par(no.readonly = TRUE)
    on.exit(par(old.par))
    u <- par("usr")
    v <- c(
      grconvertX(u[1:2], "user", "ndc"),
      grconvertY(u[3:4], "user", "ndc")
    )
    v <- c(v[1], (v[1]+v[2]) / 3,  (v[3]+v[4]) / (1.5), v[4])
    figO <- tryCatch({par(fig = v, new = TRUE, mar = c(0,0,0,0), mgp = c(0,0,0))}, error=function(ex){NULL})
    if (!is.null(figO)) {
      plot(crps.dat, yaxt = "n", col = 2, type = "l", lwd = 2, tck = .05, cex.axis = .75, ...)
      axis(4, cex.axis = .75, tck = .05)
      box()
    }
  }
}
plot.quantreg <- plot.quantreg.rfsrc

#' @keywords internal
.resolve_papply <- function(papply,
                            cores = getOption("mc.cores", 1L),
                            cl = NULL,
                            packages = NULL,
                            export = NULL,
                            envir = parent.frame()) {
  ## Accept either a function (lapply / mclapply) or a string ("lapply" / "mclapply")
  if (is.character(papply)) {
    papply <- match.arg(papply, c("lapply", "mclapply"))
    papply <- if (papply == "lapply") base::lapply else parallel::mclapply
  } else {
    papply <- match.fun(papply)
  }
  cores <- as.integer(cores)
  if (is.na(cores) || cores < 1L) cores <- 1L
  ## Always return something with lapply-like signature (X, FUN, ...)
  if (identical(papply, base::lapply) || cores < 2L) {
    return(function(X, FUN, ...) base::lapply(X, FUN, ...))
  }
  ## Legacy request for mclapply => use PSOCK instead
  if (identical(papply, parallel::mclapply)) {
    ## In package code, this helps because workers start as "empty" R sessions:
    ## load the calling package namespace (which also loads Imports).
    pkg <- utils::packageName()
    force(cl); force(packages); force(export); force(envir)
    return(function(X, FUN, ...,
                    mc.cores = cores,
                    mc.preschedule = TRUE,
                    mc.set.seed = TRUE,
                    mc.seed = NULL) {
      mc.cores <- as.integer(mc.cores)
      if (is.na(mc.cores) || mc.cores < 2L) {
        return(base::lapply(X, FUN, ...))
      }
      ## Create a cluster per call unless a cluster is supplied for reuse
      local_cluster <- is.null(cl)
      cl_use <- cl
      if (local_cluster) {
        cl_use <- parallel::makePSOCKcluster(mc.cores)
        on.exit(parallel::stopCluster(cl_use), add = TRUE)
      }
      ## Ensure your package (and its Imports) are available on workers
      if (!is.null(pkg) && nzchar(pkg)) {
        parallel::clusterCall(
          cl_use,
          function(p) { loadNamespace(p); NULL },
          pkg
        )
      }
      ## Optionally attach extra packages on workers (if you rely on search path)
      if (!is.null(packages) && length(packages)) {
        parallel::clusterCall(
          cl_use,
          function(pkgs) {
            for (p in pkgs) suppressPackageStartupMessages(
              require(p, character.only = TRUE)
            )
            NULL
          },
          packages
        )
      }
      ## Optional explicit exports (useful mainly for objects in globalenv())
      if (!is.null(export) && length(export)) {
        parallel::clusterExport(cl_use, export, envir = envir)
      }
      ## Optional reproducible parallel RNG
      if (isTRUE(mc.set.seed)) {
        if (is.null(mc.seed)) mc.seed <- sample.int(.Machine$integer.max, 1L)
        parallel::clusterSetRNGStream(cl_use, iseed = mc.seed)
      }
      ## Scheduling: map mclapply's mc.preschedule to parLapply vs parLapplyLB
      if (isTRUE(mc.preschedule)) {
        parallel::parLapply(cl_use, X, FUN, ...)
      } else {
        parallel::parLapplyLB(cl_use, X, FUN, ...)
      }
    })
  }
  ## Otherwise, user provided some other apply-like function; respect it
  if (is.function(papply)) return(papply)
  stop("papply must be 'lapply', 'mclapply', or a function like lapply().")
}

## mortality-stratified Brier curves used by plot.survival
.surv.brier.strata <- function(mort, brier.matx, brier.all) {
  ntime <- length(brier.all)
  brier.strata <- matrix(NA_real_, nrow = ntime, ncol = 4L)
  mort.ok <- is.finite(mort)
  if (any(mort.ok)) {
    mort.perc <- c(min(mort[mort.ok]) - 1e-5,
                   quantile(mort[mort.ok], (1:4) / 4, na.rm = TRUE))
    for (k in 1:4) {
      mort.pt <- mort.ok & mort > mort.perc[k] &
        mort <= mort.perc[k + 1L]
      if (any(mort.pt)) {
        brier.strata[, k] <- colMeans(
          brier.matx[mort.pt, , drop = FALSE], na.rm = TRUE)
      }
    }
  }
  brier.strata[!is.finite(brier.strata)] <- NA_real_
  out <- data.frame(brier.strata, bs.all = brier.all)
  colnames(out) <- c("bs.q25", "bs.q50", "bs.q75", "bs.q100", "bs.all")
  out
}
## cumulative integrated Brier curves used by plot.survival
.surv.crps.curves <- function(time, brier.score) {
  crps <- matrix(NA_real_, nrow = length(time), ncol = ncol(brier.score))
  for (k in seq_len(ncol(brier.score))) {
    for (j in seq_along(time)) {
      if (j > 1L) {
        denom <- diff(range(time[seq_len(j)]))
        if (is.finite(denom) && denom > 0) {
          crps[j, k] <- trapz(time[seq_len(j)],
                              brier.score[seq_len(j), k]) / denom
        }
      }
    }
  }
  crps <- data.frame(crps)
  colnames(crps) <- c("crps.q25", "crps.q50", "crps.q75",
                      "crps.q100", "crps.all")
  crps
}
## draw individual and mean survival curves
.plot.survival.curves <- function(time, surv.ensb, surv.mean.ensb,
                                  surv.aalen, subset, collapse, ylab, ...) {
  curve <- surv.ensb
  if (!collapse && length(subset) > 500L) {
    r.pt <- sample(seq_along(subset), 500L, replace = FALSE)
    curve <- surv.ensb[, r.pt, drop = FALSE]
  }
  if (!any(is.finite(curve))) {
    plot.new()
    title(xlab = "Time", ylab = ylab)
    return(invisible(NULL))
  }
  matplot(time,
          curve,
          xlab = "Time",
          ylab = ylab,
          type = "l",
          col = 1,
          lty = 3, ...)
  if (!is.null(surv.aalen)) {
    lines(time, surv.aalen, lty = 1, col = 3, lwd = 3)
  }
  lines(time, surv.mean.ensb, lty = 1, col = 2, lwd = 3)
  rug(time, ticksize = -0.03)
  invisible(NULL)
}
plot.survival.rfsrc <- function(x,
                                 show.plots = TRUE,
                                 subset, collapse = FALSE,
                                 cens.model = c("km", "rfsrc"),
                                 ...)
{
  ## incoming parameter checks
  if (is.null(x)) {
    stop("object x is empty!")
  }
  if ((!inherits(x, "rfsrc") || !inherits(x, "grow")) &&
      (!inherits(x, "rfsrc") || !inherits(x, "predict"))) {
    stop("This function only works for objects of class `(rfsrc, grow)' or '(rfsrc, predict)'")
  }
  if (x$family != "surv") {
    stop("this function only supports right-censored survival settings")
  }
  cens.model <- match.arg(cens.model)
  subset.arg <- if (missing(subset)) NULL else subset
  ## Extract predictions before requesting a performance score. This permits
  ## prediction objects without outcomes to display survival curves.
  plot.data <- .get.survival.performance.data(x,
                                               subset = subset.arg,
                                               times = NULL)
  has.outcomes <- !is.null(plot.data$eval.yvar) &&
    any(is.finite(plot.data$eval.event.info$time) &
        is.finite(plot.data$eval.event.info$cens))
  ## A prediction object without outcomes has no observable target for Brier
  ## score, CRPS, or mortality-versus-time diagnostics.
  if (!has.outcomes) {
    surv.ensb <- t(plot.data$survival)
    surv.mean.ensb <- rowMeans(surv.ensb, na.rm = TRUE)
    surv.plot <- if (collapse) surv.mean.ensb else surv.ensb
    if (show.plots) {
      old.par <- par(no.readonly = TRUE)
      on.exit(par(old.par), add = TRUE)
      par(mfrow = c(1, 1), cex = 1.0)
      .plot.survival.curves(plot.data$time,
                            surv.plot,
                            surv.mean.ensb,
                            surv.aalen = NULL,
                            subset = plot.data$subset,
                            collapse = collapse,
                            ylab = "Survival", ...)
    }
    return(invisible(list(time = plot.data$time,
                          survival = surv.ensb,
                          survival.mean = surv.mean.ensb,
                          subset = plot.data$subset)))
  }
  ## Evaluation outcomes are aligned with the predictions, while the
  ## censoring distribution is always estimated from the full grow data.
  brier.obj <- get.brier.survival(x,
                                   subset = subset.arg,
                                   cens.model = cens.model,
                                   conf.int = FALSE,
                                   keep.matrix = TRUE)
  brier.matx <- brier.obj$brier.matx
  brier.all <- brier.obj$brier.score$brier.score
  mort <- brier.obj$mort
  surv.ensb <- brier.obj$surv.ensb
  surv.aalen <- brier.obj$surv.aalen
  event.info <- brier.obj$eval.event.info
  subset <- brier.obj$subset
  time <- brier.obj$time
  ## Brier and cumulative integrated Brier processing
  brier.score <- .surv.brier.strata(mort, brier.matx, brier.all)
  crps <- .surv.crps.curves(time, brier.score)
  ## labels and titles
  if (identical(brier.obj$prediction, "oob")) {
    ylab.1 <- "OOB Survival"
    ylab.2 <- "OOB Brier"
    ylab.3 <- "OOB CRPS"
    ylab.4 <- "OOB Mortality vs Time"
  }
  else {
    ylab.1 <- "Survival"
    ylab.2 <- "Brier Score"
    ylab.3 <- "CRPS"
    ylab.4 <- "Mortality vs Time"
  }
  ## mean ensemble survival
  surv.mean.ensb <- rowMeans(surv.ensb, na.rm = TRUE)
  surv.plot <- if (collapse) surv.mean.ensb else surv.ensb
  ## ------------------------------------------------------------
  ##
  ## plots
  ##
  ## ------------------------------------------------------------
  if (show.plots) {
    old.par <- par(no.readonly = TRUE)
    on.exit(par(old.par), add = TRUE)
    par(mfrow = c(2, 2), cex = 1.0)
    ## ----survival plot----
    .plot.survival.curves(time,
                          surv.plot,
                          surv.mean.ensb,
                          surv.aalen,
                          subset,
                          collapse,
                          ylab.1, ...)
    ## ----brier plot----
    brier.values <- unlist(brier.score)
    point.x <- max(1L, round(3 * length(time) / 4))
    if (any(is.finite(brier.values))) {
      matplot(time, brier.score,
              xlab = "Time",
              ylab = ylab.2,
              type = "l",
              lwd = c(rep(1, 4), 2),
              col = c(rep(1, 4), 2),
              lty = c(1:4, 1), ...)
      rng <- range(brier.values[is.finite(brier.values)])
      abline(h = seq(rng[1], rng[2], length = 20),
             col = gray(.6), lty = 3, lwd = .85)
      text(time[point.x], brier.score[point.x, 1], "0-25", col = 4)
      text(time[point.x], brier.score[point.x, 2], "25-50", col = 4)
      text(time[point.x], brier.score[point.x, 3], "50-75", col = 4)
      text(time[point.x], brier.score[point.x, 4], "75-100", col = 4)
      rug(time, ticksize = 0.03)
    }
    else {
      plot.new()
      title(xlab = "Time", ylab = ylab.2)
    }
    ## ----CRPS plot----
    crps.values <- unlist(crps)
    if (any(is.finite(crps.values))) {
      matplot(time, crps,
              xlab = "Time",
              ylab = ylab.3,
              type = "l",
              lwd = c(rep(1, 4), 2),
              col = c(rep(1, 4), 2),
              lty = c(1:4, 1), ...)
      rng <- range(crps.values[is.finite(crps.values)])
      abline(h = seq(rng[1], rng[2], length = 20),
             col = gray(.6), lty = 3, lwd = .85)
      text(time[point.x], crps[point.x, 1], "0-25", col = 4)
      text(time[point.x], crps[point.x, 2], "25-50", col = 4)
      text(time[point.x], crps[point.x, 3], "50-75", col = 4)
      text(time[point.x], crps[point.x, 4], "75-100", col = 4)
      rug(time, ticksize = 0.03)
    }
    else {
      plot.new()
      title(xlab = "Time", ylab = ylab.3)
    }
    ## ----mortality plot----
    ytime <- event.info$time
    status <- event.info$cens
    plot.pt <- is.finite(ytime) & is.finite(mort)
    if (any(plot.pt)) {
      plot(ytime[plot.pt], mort[plot.pt],
           xlab = "Time", ylab = ylab.4, type = "n", ...)
      cex <- if (length(mort) > 500L) 0.5 else 0.75
      event.pt <- plot.pt & is.finite(status) & status != 0
      cens.pt <- plot.pt & is.finite(status) & status == 0
      points(ytime[event.pt], mort[event.pt], pch = 16,
             col = 4, cex = cex)
      points(ytime[cens.pt], mort[cens.pt], pch = 16, cex = cex)
      if (sum(event.pt) > 1L) {
        ord <- order(ytime[event.pt])
        lines(supsmu(ytime[event.pt][ord], mort[event.pt][ord]),
              lty = 3, col = 4)
      }
      if (sum(cens.pt) > 1L) {
        ord <- order(ytime[cens.pt])
        lines(supsmu(ytime[cens.pt][ord], mort[cens.pt][ord]),
              lty = 3)
      }
      if (any(event.pt)) {
        rug(ytime[event.pt], ticksize = -0.03)
      }
    }
    else {
      plot.new()
      title(xlab = "Time", ylab = ylab.4)
    }
  }
  ## Invisibly return the Brier and cumulative integrated Brier curves.
  invisible(data.frame(time = time, brier.score, crps))
}
plot.survival <- plot.survival.rfsrc
## draw a confidence polygon over contiguous finite portions of a curve
.surv.plot.band <- function(x, lower, upper, col, border = NA) {
  ok <- is.finite(x) & is.finite(lower) & is.finite(upper)
  if (!any(ok)) {
    return(invisible(NULL))
  }
  run <- rle(ok)
  run.end <- cumsum(run$lengths)
  run.start <- run.end - run$lengths + 1L
  for (k in which(run$values)) {
    idx <- run.start[k]:run.end[k]
    if (length(idx) >= 2L) {
      graphics::polygon(c(x[idx], rev(x[idx])),
                        c(lower[idx], rev(upper[idx])),
                        col = col, border = border)
    }
  }
  invisible(NULL)
}
## draw one Brier or AUCT performance curve using base R graphics
.surv.plot.performance.curve <- function(score, estimate,
                                         ylab, main, reference = NULL,
                                         dots = list()) {
  time <- score$time
  value <- score[[estimate]]
  has.band <- all(c("lower", "upper") %in% names(score))
  yy <- value
  if (has.band) {
    yy <- c(yy, score$lower, score$upper)
  }
  if (!is.null(reference)) {
    yy <- c(yy, reference)
  }
  yy <- yy[is.finite(yy)]
  ylim <- if (length(yy) > 0L) range(yy) else c(0, 1)
  if (diff(ylim) == 0) {
    ylim <- ylim + c(-0.04, 0.04) * max(1, abs(ylim[1L]))
  }
  line.args <- list(col = 2, lwd = 2, lty = 1, type = "l")
  for (nm in intersect(c("col", "lwd", "lty", "type"), names(dots))) {
    line.args[[nm]] <- dots[[nm]]
    dots[[nm]] <- NULL
  }
  band.col <- if (!is.null(dots$band.col)) {
    dots$band.col
  }
  else {
    grDevices::adjustcolor(gray(.5), alpha.f = .25)
  }
  band.border <- if (!is.null(dots$band.border)) dots$band.border else NA
  dots$band.col <- NULL
  dots$band.border <- NULL
  plot.args <- list(x = time, y = value, type = "n",
                    xlab = "Time", ylab = ylab,
                    main = main, ylim = ylim)
  dot.names <- names(dots)
  if (!is.null(dot.names)) {
    named <- nzchar(dot.names)
    for (nm in dot.names[named]) {
      plot.args[[nm]] <- dots[[nm]]
    }
    if (any(!named)) {
      plot.args <- c(plot.args, dots[!named])
    }
  }
  do.call(plot, plot.args)
  if (has.band) {
    .surv.plot.band(time, score$lower, score$upper,
                    col = band.col, border = band.border)
  }
  if (!is.null(reference)) {
    abline(h = reference, lty = 3, col = gray(.5))
  }
  do.call(lines, c(list(x = time, y = value), line.args))
  rug(time, ticksize = -0.025)
}
## base-R plot helper for Brier score and time-dependent AUC
plotBrierAUC <- function(x,
                           subset = NULL,
                           cens.model = c("km", "rfsrc"),
                           papply = lapply,
                           times = NULL,
                           conf.int = TRUE,
                           plots = c("brier", "auct"),
                           show.plots = TRUE,
                           ...) {
  plots <- as.character(plots)
  if (length(plots) == 0L) {
    stop("'plots' must include 'brier', 'auct', or both.")
  }
  plots[plots == "auc"] <- "auct"
  plots <- unique(match.arg(plots, c("brier", "auct"), several.ok = TRUE))
  ans <- .get.survival.performance(x,
                                   subset = subset,
                                   cens.model = cens.model,
                                   papply = papply,
                                   times = times,
                                   conf.int = conf.int,
                                   keep.matrix = FALSE,
                                   metrics = plots)
  if (is.null(ans)) {
    return(invisible(NULL))
  }
  if (show.plots) {
    old.par <- par(no.readonly = TRUE)
    on.exit(par(old.par), add = TRUE)
    par(mfrow = c(length(plots), 1L), cex = 1.0)
    dots <- list(...)
    for (what in plots) {
      if (what == "brier") {
        .surv.plot.performance.curve(
          ans$brier$brier.score,
          estimate = "brier.score",
          ylab = "Brier Score",
          main = "Brier Score Over Time",
          dots = dots
        )
      }
      else {
        .surv.plot.performance.curve(
          ans$auct$auct.score,
          estimate = "auct",
          ylab = "AUC",
          main = "Time-Dependent AUC",
          reference = 0.5,
          dots = dots
        )
      }
    }
  }
  invisible(ans)
}

plot.rfsrc <- function (x, m.target = NULL, plots.one.page = TRUE, sorted = TRUE, verbose = TRUE, ...)
{
  sf.flag <- FALSE
  ## is this a synthetic forest?  Printing is different in that case
  if (sum(inherits(x, c("rfsrc", "synthetic"), TRUE) == c(1, 2)) == 2) {
    if (sum(inherits(x, c("rfsrc", "synthetic", "oob"), TRUE) == c(1, 2, 3)) != 3) {
      sf.flag <- TRUE
      sf.message <- "OOB was not used for synthetic forests, error rates/VIMP will be unreliable"
    }
    x <- x$rfSyn
  }
  ## check that object is interpretable
  if (sum(inherits(x, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2 &
      sum(inherits(x, c("rfsrc", "predict"), TRUE) == c(1, 2)) != 2) {
    stop("this function only works for objects of class `(rfsrc, grow)' or '(rfsrc, predict)'")
  }
  ## pull the unnamed options
  dots <- list(...)
  cex <- par("cex")
  if (!is.null(dots$cex)) {
    cex <- dots$cex
  }
  ## print a subsampled object
  if (sum(inherits(x, c("rfsrc", "subsample"), TRUE) == c(1, 4)) == 2 ||
      sum(inherits(x, c("rfsrc", "bootsample"), TRUE) == c(1, 4)) == 2) {
    return(plot.subsample(x, m.target = m.target, cex = cex))
  }
  ## coerce the (potentially) multivariate object if necessary.
  m.target <- get.univariate.target(x, m.target)
  x <- coerce.multivariate(x, m.target)
  ## grow objects under non-standard bootstrapping are devoid of
  ## performance values
  if (is.null(x$err.rate)) {
    stop("object is devoid of performance values")
  }
  ## set importance to NA if it is NULL
  if (is.null(x$importance)) {
    x$importance <- NA
  }
  ## return when everything is NA
  if (all(is.na(x$err.rate)) & all(is.na(x$importance))) {
    stop("performance values are all NA")
  }
  ## Check that the error rate vector is assigned for all trees.  If
  ## it is not, fill in the slots for 1:ntree.  This will result in
  ## the plot flat-lining between transitions or blocks.
  if (x$block.size != 1) {
      x$err.rate <- cbind(x$err.rate)
      fill.err.row <-  x$block.size
      nullO <- lapply(1:x$ntree, function(i) {
        x$err.rate[i, ] <<- x$err.rate[fill.err.row, ]
        if (i == fill.err.row) {
          fill.err.row <<- min(fill.err.row + x$block.size, x$ntree)
        }
        NULL
      })
      rm(nullO)
  }
  ## convert drc two-classifier error rate to one column
  if (!is.null(x$forest$perf.type) && (x$forest$perf.type == "g.mean" || x$forest$perf.type == "g.mean.rfq")) {
    x$err.rate <- x$err.rate[, 1, drop = FALSE]
  }
  ## for surv/CR we do not plot the yvar.names: we set them to blank
  if (x$family == "surv-CR" | x$family == "surv") {
    plot.yvar.names <- ""
  }
  else {
    plot.yvar.names <- paste("(", x$yvar.names, ")", sep = "")
  }
  ## decide what plots to generate
  if (all(is.na(x$importance))) {
    if (x$ntree > 1 && !all(is.na(x$err.rate))) {
      err <- cbind(x$err.rate)      
      par(mfrow = c(1,1))
      plot_err(err, plot.yvar.names)
    }
  }
    else {
      ## convert err/vimp to matrix format
      err <- cbind(x$err.rate)
      imp <- cbind(x$importance)
      ## convert drc two-classifier vimp to one column
      if (!is.null(x$forest$perf.type) && (x$forest$perf.type == "g.mean" || x$forest$perf.type == "g.mean.rfq")) {
        imp <- imp[, 1, drop = FALSE]
      }
      x.var.names <- rownames(imp)
      n.pred <- nrow(imp)
      if (sorted) pred.order <- order(imp[, 1]) else pred.order <- n.pred:1
      if (ncol(imp) == 1) max.pred <- 100 else max.pred <- 80/ncol(imp)
      if (n.pred > max.pred) {
        dotchart.labels <- rep("",n.pred)
        pretty.pt <- pretty(1:n.pred, n = max.pred)
        dotchart.labels[pretty.pt] <- x.var.names[pred.order][pretty.pt]
      }
        else {
          dotchart.labels <- x.var.names[pred.order]
        }
      if (x$ntree > 1 & !all(is.na(x$err.rate)) & plots.one.page) {
        par(mfrow = c(1,2))
      }
        else {
          par(mfrow = c(1,1))
        }
      if (x$ntree > 1 & !all(is.na(x$err.rate))) {
        plot_err(err, plot.yvar.names)
      }
      ## CR/classification scenarios
      if (ncol(imp) > 1) {
        imp.out <- imp[rev(pred.order),, drop = FALSE]
        dot_chart(imp[pred.order,, drop = FALSE], plot.yvar.names, dotchart.labels, cex = cex)
      }
      ## other scenarios
      if (ncol(imp) == 1) {
        dot_chart(imp[pred.order, ], plot.yvar.names, dotchart.labels, cex = cex)
        if (!is.null(x$xvar.wt) & length(unique(x$xvar.wt)) > 1 ) {
          if (length(unique(x$xvar.wt)) == 1) x$xvar.wt <- 1
          imp.out <- as.data.frame(cbind(imp, imp/max(abs(imp), na.rm = TRUE), x$xvar.wt),
                                   row.names = x.var.names)[rev(pred.order),]
          if (nrow(imp.out) == 1) imp.out[1 , 2] <- 1
          colnames(imp.out) <- c("Importance","Relative Imp","xvar weight")
        }
          else {
            imp.out=as.data.frame(cbind(imp, imp/max(abs(imp), na.rm = TRUE)),
              row.names=x.var.names)[rev(pred.order),]
            if (nrow(imp.out) == 1) imp.out[1 , 2] <- 1
            colnames(imp.out) <- c("Importance","Relative Imp")
          }
      }
      cat("\n")
      if (verbose) {
        print(round(imp.out[1:min(n.pred, max.pred),, drop = FALSE],4), justify="right", print.gap=3)
      }    
    }
#################################################################################
  ##
  ## synthetic forest flag
  ##
################################################################################# 
  if (sf.flag) {
    message(sf.message)
  }
}
## error rate plot
plot_err <- function(err, yname = NULL) {
  matplot(1:nrow(err), err,
          xlab = "Number of Trees",
          ylab = paste("Error rate", yname),
          type = c("p", "l")[1 + 1 * (nrow(err) > 1)], pch = 16, lty = 1, lwd = 3)
  if (ncol(err) > 1) {
    legend("topright",
           legend = colnames(err), col = 1:ncol(err), lty = 1, lwd = 3)
  }
}
## pretty dotchart
dot_chart <- function(x, yname = NULL, labels = NULL, cex = cex) {
  if (!is.null(dim(x))) {
    ncol  <- ncol(x)
    x.dot <- NULL
    for (k in ncol(x):1) {x.dot <- c(x.dot, x[, k])}
    gcolor <- 1:ncol
  }
    else {
      x.dot <- x
      gcolor <- par("fg")
    }
  y.dot <- dot_chart_main(x, labels = labels, xlab = paste("Variable Importance", yname),
                          cex = cex, pch="", lwd = 2, lcolor = "white", gcolor = gcolor)
  segments(rep(max(0, min(x.dot, na.rm = TRUE)) - 1e-6, length(y.dot)),
           y.dot, x.dot, y.dot, col=c(2,4)[1 + 1 * (x.dot > 0)], lwd = 4)
  if (min(x.dot, na.rm = TRUE) < 0) abline(v=0, lwd = 2, lty = 2, col = 1)
}
## workhorse for dotchart
dot_chart_main <- function (x, labels = NULL, groups = NULL, gdata = NULL, cex = NULL,
                            pch = 21, gpch = 21, bg = par("bg"), color = par("fg"), gcolor = par("fg"), 
                            lcolor = "gray", xlim = range(x[is.finite(x)]), main = NULL, 
                            xlab = NULL, ylab = NULL, ...) 
{
  opar <- par("mai", "mar", "cex", "yaxs")
  on.exit(par(opar))
  par(yaxs = "i")
  if (!is.numeric(x)) 
    stop("'x' must be a numeric vector or matrix")
  n <- length(x)
  if (is.matrix(x)) {
    if (is.null(labels)) 
      labels <- rownames(x)
    if (is.null(labels)) 
      labels <- as.character(1:nrow(x))
    labels <- rep(labels, length.out = n)
    if (is.null(groups)) 
      groups <- col(x, as.factor = TRUE)
    glabels <- levels(groups)
  }
    else {
      if (is.null(labels)) 
        labels <- names(x)
      glabels <- if (!is.null(groups)) 
                   levels(groups)
    }
  plot.new()
  linch <- if (!is.null(labels)) 
             max(strwidth(labels, "inch"), na.rm = TRUE)
             else 0
  if (is.null(glabels)) {
    ginch <- 0
    goffset <- 0
  }
    else {
      ginch <- max(strwidth(glabels, "inch"), na.rm = TRUE)
      goffset <- 0.4
    }
  if (!(is.null(labels) && is.null(glabels))) {
    nmai <- par("mai")
    nmai[2] <- nmai[4] + max(linch + goffset, ginch) + 0.1
    par(mai = nmai)
  }
  if (is.null(groups)) {
    o <- 1:n
    y <- o
    ylim <- c(0, n + 1)
  }
    else {
      o <- sort.list(as.numeric(groups), decreasing = TRUE)
      x <- x[o]
      groups <- groups[o]
      color <- rep(color, length.out = length(groups))[o]
      lcolor <- rep(lcolor, length.out = length(groups))[o]
      offset <- cumsum(c(0, diff(as.numeric(groups)) != 0))
      y <- 1:n + 2 * offset
      ylim <- range(0, y + 2)
    }
  plot.window(xlim = xlim, ylim = ylim, log = "")
  lheight <- par("csi")
  if (!is.null(labels)) {
    linch <- max(strwidth(labels, "inch"), na.rm = TRUE)
    loffset <- (linch + 0.1)/lheight
    labs <- labels[o]
    mtext(labs, side = 2, line = loffset, at = y, adj = 0, 
          col = color, las = 2, cex = cex, ...)
  }
  abline(h = y, lty = "dotted", col = lcolor)
  points(x, y, pch = pch, col = color, bg = bg)
  if (!is.null(groups)) {
    gpos <- rev(cumsum(rev(tapply(groups, groups, length)) + 
                         2) - 1)
    ginch <- max(strwidth(glabels, "inch"), na.rm = TRUE)
    goffset <- (max(linch + 0.2, ginch, na.rm = TRUE) + 0.1)/lheight
    mtext(glabels, side = 2, line = goffset, at = gpos, adj = 0, 
          col = gcolor, las = 2, cex = cex, ...)
    if (!is.null(gdata)) {
      abline(h = gpos, lty = "dotted")
      points(gdata, gpos, pch = gpch, col = gcolor, bg = bg, 
             ...)
    }
  }
  axis(1)
  box()
  title(main = main, xlab = xlab, ylab = ylab, ...)
  invisible(y)
}

max.subtree.rfsrc <- function(object,
                              max.order = 2,
                              sub.order = FALSE,
                              conservative = FALSE,
                              ...)
{
  ## Incoming parameter checks.  All are fatal.
  if (is.null(object)) stop("Object is empty!")
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2    &
      sum(inherits(object, c("rfsrc", "forest"), TRUE) == c(1, 2)) != 2)
    stop("This function only works for objects of class `(rfsrc, grow)' or '(rfsrc, forest)'")
  ## Acquire the forest 
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) == 2) {
    if (is.null(object$forest)) 
      stop("Forest is empty!  Re-run grow call with forest set to 'TRUE'.")
    object <- object$forest
  }
  nativeArray <- object$nativeArray
  if (is.null(nativeArray)) {
    stop("RFSRC nativeArray content is NULL.  Please ensure the object is valid.")
  }
  xvar.names <- object$xvar.names
  if (is.null(xvar.names)) {
    stop("RFSRC xvar.names content is NULL.  Please ensure the object is valid.")
  }
  if (is.null(object$xvar)) {
    stop("RFSRC xvar content is NULL.  Please ensure the object is valid.")
  }
  ## verify the max.order option
  max.order <- floor(max.order)
  if (max.order < 0) {
    stop("RFSRC 'max.order' requested for distance order statistic must be an integer greater than zero (0).")
  }
  ## max.order=0 defaults to non-conservative thresholding
  if (max.order == 0) {
    conservative <- FALSE
  }
  ## Maximum depth monitored for nodes at depth counts
  ## TBD Increased from 1000 to 10000 but need a better way to set this value. TBD 
  MAX.DEPTH <- 10000
  ## Count the number of trees in the forest.
  numTree <- length(as.vector(unique(nativeArray$treeID)))
  ## Pre-compute the starting offset for each tree to avoid an O(N) scan per tree.
  treeID <- nativeArray$treeID
  if (is.factor(treeID)) {
    treeID <- as.integer(as.character(treeID))
  }
  treeOffset <- match(1:numTree, treeID)
  if (any(is.na(treeOffset))) {
    ## Fallback for unexpected/non-consecutive tree IDs.
    treeOffset <- sapply(1:numTree, function(b) min(which(treeID == b)))
  }
  ## Count the number of parameters in the data set.
  numParm <- length(xvar.names)
  ## Determine the sample size of the data set.
  numSamp <- nrow(object$xvar)
  ## ------------------------------------------------------------------
  ## The maximal distance from the root node of a tree or sub-tree is
  ## zero-based.  This means that the root node represents a distance
  ## of zero (0).
  ##
  ## $count   - of length [numParm].
  ##          - integer count of the number of maximal subtrees for each
  ##            xvar in each tree.
  ##          - this will be zero (0) if the xvar is not used in
  ##            the forest.
  ## $order   - if max.order > 0, matrix of dim [numParm] x [max.order]
  ##            representing the order statistic for the maximal
  ##            subtree distance for each xvar
  ##          - if max.order == 0, matrix of [numParm] x [numTree]
  ##            representing the first order statistic for the maximal
  ##            subtree distance for each xvar by tree
  ##          - rows represent the distribution for the order
  ##            statistic for a xvar
  ##          - thus [2][1] gives the minimum distance for xvar 2.
  ##          - elements will have the value (maximum + 1), the
  ##            penalty, if the xvar is not used in the tree, or
  ##            the statistic does not exist
  ## $meanSum - vector of length [numParm]
  ##          - the pre-mean mimimum maximal subtree distance for each
  ##            xvar in each tree.
  ##          - if the xvar is not used in the tree, this will be
  ##            the maximum depth for the tree, the penalty.
  ## $depth   - integer depth (defined as the maximum terminal node
  ##            distance from the root node) possible in the tree
  ##            topology.  This will be always greater than zero (0)
  ##            for non-trivial trees
  ## $terminalDepthSum
  ##          - integer sum of the depths of the terminal
  ##            nodes of the tree
  ##          - used in calculating the average depth of the
  ##            nodes for tree
  ## $subOrder
  ##          - matrix of [numParm] x [numParm] representing the
  ##            normalized minimum maximal w-subtree distance for
  ##            parameter v's maximal subtree.
  ##          - thus if v=2, w=1, consider a maximal 2-subtree with
  ##            the "root" node of this subtree associated with
  ##            relative depth = 0.  Then consider a 1-subtree of the
  ##            2-subtree.  Then, [2][1] gives the normalized minimum
  ##            maximal subtree distance of all 1-subtrees within the
  ##            2-subtree.  If the maximal w-subtree does not exist,
  ##            the distance is set to be one (1).  The normalized
  ##            distance is calculated as follows.  First, the
  ##            maximum depth of the terminal nodes of the v-subtree
  ##            is determined.  Then the relative depth of the
  ##            w-subtree (from the root node) of the v-subtree is
  ##            determined.  This latter quantity is divided by the
  ##            prior maximum terminal node depth quantity to give a
  ##            number between (0,1].
  ##          - the diagonal [i][i] is the normalized minimum maximal
  ##            v-subtree distance for the parameter i.  If it is not
  ##            split on in the tree, it is assigned the value one
  ##            (1).
  ##          - the resulting matrix is NA-free.
  ## $subOrderDiag
  ##          - interim vector of length [numParm] representing the 
  ##            minimum maximal v-subtree distance for the tree that
  ##            will constitute the diagonal of $subOrder.
  ##          - variables that are not split on are assigned the
  ##            penalty value of the maximum depth.
  ## $nodesAtDepth
  ##          - number of nodes at each depth, where [1] = 2 for
  ##            a non-trivial tree, [2] = 4 for a balanced tree.
  ##            In general, for a balanced tree, at depth d, the
  ##            count is 2^d.
  ##
  ## --------------------------------------------------------------
  ## Loop through all trees
  ## This is a LOCAL operation
  ## Global dependencies:  (xvar.names, forest)
  ##
  ## It is inefficient to use a do-loop, therefore we use lapply.
  ## Later we parse the list to extract the necessary objects
  ## --------------------------------------------------------------
  subtree.obj <- lapply(1:numTree, function(b) {
    ## Create the (local) subtree object for recursion.  This is NOT the
    ## output object, but is closely related to it.
    subtree <- vector("list", 8)
    names(subtree) <- c("count",
                        "order",
                        "meanSum",
                        "depth",
                        "terminalDepthSum",
                        "subOrder",
                        "subOrderDiag",
                        "nodesAtDepth")
    ## Create the recursive output object.  This would be unnecessary 
    ## if it was possible to declare global variables in a package.
    recursiveObject <- list(offset = treeOffset[b],
                            subtree = subtree,
                            diagnostic = 0,
                            diagnostic2 = 0)
    ## Reset the nodes at depth count.
    recursiveObject$subtree$nodesAtDepth <- rep(NA, MAX.DEPTH)
    ## Reset the mean maximal subtree distance.
    ## This will be NA if the xvar is not used in this tree.
    recursiveObject$subtree$meanSum <- rep(NA, numParm)
    ## Reset the order statistic matrix representing maximal distance.
    recursiveObject$subtree$order <- matrix(NA, nrow=numParm, ncol=max(max.order, 1))
    if (sub.order) {
      ## Reset the sub-order statistic matrix representing the minimum
      ## maximal w-subtree distance for v != w.  
      recursiveObject$subtree$subOrder <- matrix(1.0, nrow=numParm, ncol=numParm)
      ## Reset the subOrder diagonals to NA.
      recursiveObject$subtree$subOrderDiag <- rep(NA, numParm)
    }
    ## Reset the maximum maximal subtree distance.
    recursiveObject$subtree$depth <- 0
    ## Reset the terminal node depth sums.
    recursiveObject$subtree$terminalDepthSum <- 0
    ## Reset the count of the maximal subtrees in the tree.
    recursiveObject$subtree$count <- rep(0, numParm)
    ## Identify the root node split parameter.
    rootParmID <- nativeArray$parmID[recursiveObject$offset] 
    ## Save the previous value of the offset.  This is used in
    ## determining the number of terminal nodes in the tree.
    offsetMark <- recursiveObject$offset
    ## Set the stump count.  Gets updated when the tree is a stump.
    stumpCnt <- 0
    ## ----------------------------
    ##  PRIMARY RECURSION PRIMARY
    ## ----------------------------
    ## Recursively parse the tree in the primary protocol.
    recursiveObject <- rfsrcParseTree(
      recursiveObject,
      max(max.order, 1),
      sub.order,
      nativeArray,
      b,
      distance=0,
      subtreeFlag=rep(FALSE, numParm))
    ## Check that the current tree is not a stump.
    if (rootParmID != 0) {
      ## Update the forest sum ...
      ## Determine which xvar _were_not_ split on, in the current tree.
      index <- which(recursiveObject$subtree$count == 0)
      ## Penalize these unused xvar by making their mean distance the worst.
      recursiveObject$subtree$meanSum[index] <- recursiveObject$subtree$depth
      forestMeanSum <- recursiveObject$subtree$meanSum
      ## This is tree specific.  Here we view the matrix as a vector.
      index <- which(is.na(recursiveObject$subtree$order))
      ## Penalize these order statistics making their values the worst.      
      recursiveObject$subtree$order[index] <- recursiveObject$subtree$depth
      orderTree <- recursiveObject$subtree$order
      ## Explanation of demominator: The total number of nodes N
      ## (internal and external) in a tree is given by the number of
      ## records in a tree.  In this case, offset - offsetMark = N,
      ## due to the nature of the pointer incrementation.  The number
      ## of external (terminal) nodes N(TERM) is given by (N + 1) / 2.
      ## The maximum number of subtrees for the tree is given by
      ## N(TERM) / 2 which is the denominator.
      ## We normalize the actual number of maximal subtrees found in
      ## the tree by the maximum number maximal subtrees possible.
      subtreeCountSum <- (recursiveObject$subtree$count / ((recursiveObject$offset - offsetMark + 1) / 4))
      ## Get the average depth of the terminal nodes of this tree.
      ## The denominator is just the number of terminal nodes.
      terminalDepth <- recursiveObject$subtree$terminalDepthSum / ((recursiveObject$offset - offsetMark + 1) / 2)
      if (sub.order) {
        ## Determine which xvar _were_ split on, in the current tree.
        index <- which(recursiveObject$subtree$count > 0)
        ## Over-ride those diagonal elements with the minimum depth encountered.
        diag(recursiveObject$subtree$subOrder)[index] <- recursiveObject$subtree$subOrderDiag[index]
        ## Determine which xvar _were_not_ split on, in the
        ## current tree.  This can also be determined by examining the
        ## diagonal vector, and noting those elements with the value
        ## NA
        index <- which(recursiveObject$subtree$count == 0)
        ## Over-ride those missing diagonal elements with penalty value.
        diag(recursiveObject$subtree$subOrder)[index] <- recursiveObject$subtree$depth
        ## Normalize the depths.
        diag(recursiveObject$subtree$subOrder) <-  diag(recursiveObject$subtree$subOrder) / recursiveObject$subtree$depth
        ## Add the result to the forest sum.
        subOrderSum <- recursiveObject$subtree$subOrder
      }
      else {
        subOrderSum <- NULL
      }
      ## Update the nodes at depth count.
      nodesAtDepthMatrix <- recursiveObject$subtree$nodesAtDepth
    }
    else {
      ## The tree is a stump.  It will be excluded.
      stumpCnt <- 1
      ## All other parameters are set to NULL
      forestMeanSum <- orderTree <- subtreeCountSum <-
        terminalDepth <- subOrderSum <- nodesAtDepthMatrix <- NULL
    }
    ## return the local subtree objects for parsing
    return(list(forestMeanSum = forestMeanSum,
                orderTree = orderTree,
                subtreeCountSum = subtreeCountSum,
                terminalDepth = terminalDepth,
                subOrderSum = subOrderSum,
                stumpCnt = stumpCnt,
                nodesAtDepthMatrix = nodesAtDepthMatrix))
  })
  ## ---------------------------------------------------------------------
  ## Parse the local object
  ## Pre-mean maximal subtree distance by parameter over the forest.
  forestMeanSum <- rep(0, numParm)
  ## This is by tree over the forest.
  ## Minimum maximal subree distance by parameter by tree over the forest.
  order.tree <- matrix(NA, nrow=numParm, ncol=numTree)
  ## Order statistics by parameter of maximal subtree distance.
  if (max.order > 0) {
    orderSum <- matrix(0, nrow=numParm, ncol=max.order)
  }
  ## Normalized maximal subtree count sum over the forest.  Linked to $count.
  subtreeCountSum <- rep(0, numParm)
  ## Average depth of each tree.  Linked to $terminalDepthSum 
  terminalDepth <- rep(0, numTree)
  ## Sub-order interim sum.
  subOrderSum <- matrix(0, nrow=numParm, ncol=numParm)
  ## Final list of count of nodes at depth by tree.
  nodesAtDepthMatrix <- matrix(NA, nrow = MAX.DEPTH, ncol = numTree)
  ## Initialize the stump count
  stumpCnt <- 0
  ## iterate over trees
  ## (avoid an explicit for-loop; use vapply/lapply/Reduce instead)
  is.valid.tree <- vapply(subtree.obj, function(z) z$stumpCnt == 0, logical(1))
  stumpCnt <- sum(!is.valid.tree)
  ## Minimal depth by variable by tree (stumps remain NA)
  if (max.order > 0) {
    order.tree <- vapply(subtree.obj, function(z) {
      if (z$stumpCnt == 0) {
        z$orderTree[, 1]
      } else {
        rep(NA_real_, numParm)
      }
    }, numeric(numParm))
  } else {
    order.tree <- vapply(subtree.obj, function(z) {
      if (z$stumpCnt == 0) {
        as.vector(z$orderTree)
      } else {
        rep(NA_real_, numParm)
      }
    }, numeric(numParm))
  }
  ## Update the nodes at depth count (stumps remain NA)
  nodesAtDepthMatrix <- vapply(subtree.obj, function(z) {
    if (z$stumpCnt == 0) {
      z$nodesAtDepthMatrix
    } else {
      rep(NA_real_, MAX.DEPTH)
    }
  }, numeric(MAX.DEPTH))
  ## Tree terminal node depth (stumps coded as 0)
  terminalDepth <- vapply(subtree.obj, function(z) {
    if (z$stumpCnt == 0) {
      z$terminalDepth
    } else {
      0
    }
  }, numeric(1))
  ## Aggregate sums over non-stump trees
  valid.obj <- subtree.obj[is.valid.tree]
  if (length(valid.obj) > 0) {
    forestMeanSum <- Reduce(`+`, lapply(valid.obj, `[[`, "forestMeanSum"))
    subtreeCountSum <- Reduce(`+`, lapply(valid.obj, `[[`, "subtreeCountSum"))
    if (max.order > 0) {
      orderSum <- Reduce(`+`, lapply(valid.obj, `[[`, "orderTree"))
    }
    if (sub.order == TRUE) {
      subOrderSum <- Reduce(`+`, lapply(valid.obj, `[[`, "subOrderSum"))
    }
  }
  ## ---------------------------------------------------------------------
  ## Output preparation  
  ##
  ##
  ## Prepare the ensemble values for the output object.
  ##
  ## $mean - vector of length [numParm]
  ##       - mean minimum maximal subtree distance by parameter over the
  ##         forest.
  ## $order - if max.order > 0, matrix of dim [numParm] x [max.order]
  ##            representing the order statistic for the maximal
  ##            subtree distance for each xvar over the forest
  ##          - if max.order == 0, matrix of [numParm] x [numTree]
  ##            representing the first order statistic for the maximal
  ##            subtree distance for each xvar by tree
  ## $count - vector of length [numParm]
  ##        - normalized maximal subtree count by parameter over the
  ##          forest
  ##        - normalization is by the number of topologically possible
  ##          subtrees in each tree
  ## $sub.order - matrix of [numParm] x [numParm] representing the
  ##             normalized minimum maximal w-subtree distance for
  ##             parameter v's maximal subtree averaged over all trees.
  ##           - see the tree specific description for details.
  ##           - output if sub.order=TRUE
  ## $terminal - vector of length [numTree]
  ##           - average terminal node depth of each tree.
  ## $nodes.at.depth - matrix of [MAX.DEPTH] x [numTree]
  ##               - number of nodes at each depth, where [1] = 2 for
  ##                 a non-trivial tree, [2] = 4 for a balanced tree.
  ##                 In general, for a balanced tree, at depth d, the
  ##                 count is [d] = 2^d.
  nameVector <- c("mean",
                  "order",
                  "count",
                  "terminal",
                  "nodes.at.depth",
                  "sub.order",
                  "threshold",
                  "threshold.1se",
                  "topvars",
                  "topvars.1se",
                  "percentile",
                  "density")
  result <- vector("list", length(nameVector))
  names(result) <- nameVector
  ## Precautionary check for all stumps.
  if(numTree != stumpCnt) {
    ## Attach the average terminal node depth to the output object.
    result$terminal <- terminalDepth
    ## Determine the forest average maximal subtree distance.
    result$mean <- forestMeanSum / (numTree - stumpCnt)
    names(result$mean) <- xvar.names
    ## Minimal depth by variable by tree
    minDepthVarTree <- order.tree
    if (max.order > 0) {
      ## Summarize the order statistics containing the maximal subtree distances. 
      result$order <- orderSum / (numTree - stumpCnt)
      rownames(result$order) <- xvar.names
    }
    else {
      result$order <- minDepthVarTree
      rownames(result$order) <- xvar.names
    }
    ## Determine the forest average subtree count.
    result$count <- subtreeCountSum / (numTree - stumpCnt)
    names(result$count) <- xvar.names
    ## Copy the matrix for the number of nodes at each depth.
    result$nodes.at.depth <- nodesAtDepthMatrix
    if (sub.order == TRUE) {      
      ## Calculate the forest average of the minimum maximal w-subtree distances.
      result$sub.order   <- subOrderSum / (numTree - stumpCnt)
      rownames(result$sub.order) <- xvar.names
      colnames(result$sub.order) <- xvar.names
    }
  }
  ## print (recursiveObject$diagnostic)
  ## print (recursiveObject$diagnostic2)
  ## Minimal depth for thresholding weak variables
  if (conservative == TRUE) {
    parseDepthOrderObj <- parseDepthOrder(result, max.order)
    if (!is.null(parseDepthOrderObj)) {
      result$threshold <- parseDepthOrderObj$first.moment
      result$threshold.1se <- result$threshold + parseDepthOrderObj$sd / sqrt(numSamp)
      result$density <- parseDepthOrderObj$prob
      result$second.order.threshold <- parseDepthOrderObj$second.order.moment
    }
  }
  else {
    parseDepthOrderObj <- parseDepthOrder(result, 0)
    if (!is.null(parseDepthOrderObj)) {
      threshold.con <- parseDepthOrder(result, 1)$first.moment
      threshold.lib <- mean(parseDepthOrderObj$first.moment, na.rm = TRUE)
      result$threshold <- ifelse(threshold.lib - 0.5 > threshold.con, threshold.lib - 0.5, threshold.lib)
      result$threshold.1se <- result$threshold + mean(parseDepthOrderObj$sd, na.rm = TRUE) / sqrt(numSamp)    
      result$density <- apply(parseDepthOrderObj$prob, 1, mean, na.rm = TRUE)
      result$second.order.threshold <- mean(parseDepthOrderObj$second.order.moment, na.rm = TRUE)
    }
  }
  if (!is.null(result$density)) {
    names(result$density) <- paste(0:(length(result$density) - 1))
  }
  ## variable selection
  if (max.order > 0) {
    minDepthVar <- result$order[, 1]
  }
  else {
    if (!is.null(result$order)) {
      minDepthVar <- apply(minDepthVarTree, 1, mean, na.rm = TRUE)
    }
    else {
      minDepthVar <- NULL
    }
  }
  top.pt <- minDepthVar <= result$threshold
  top.pt[is.na(top.pt)] <- FALSE
  if (sum(top.pt) > 0) {
    result$topvars <- xvar.names[top.pt]
  }
  top.1se.pt <- minDepthVar <= result$threshold.1se
  top.1se.pt[is.na(top.1se.pt)] <- FALSE
  if (sum(top.1se.pt) > 0) {
    result$topvars.1se <- xvar.names[top.1se.pt]
  }
  ## minimal depth percentile
  if (!is.null(result$order)) {
    cdf <- cumsum(result$density)
    if (conservative == TRUE) {
      result$percentile <- cdf[pmax(round(minDepthVar), 1)]      
    }
    else {
      result$percentile <- apply(rbind(sapply(1:ncol(minDepthVarTree), function(b) {
        cdf[pmax(minDepthVarTree[, b], 1)]
      })), 1, mean, na.rm = TRUE)
    }
    names(result$percentile) <- xvar.names
  }
  ## discard values no longer needed
  result$terminal <- result$mean <- NULL
  ## exit; invisible return
  invisible(result)
}
## -----------------------------------------------------------------------
##
##  PRIMARY RECURSION
##
## -----------------------------------------------------------------------
## Recursive function to determine first order maximal distances.
rfsrcParseTree <- function(recursiveObject,
                           max.order,
                           sub.order,
                           nativeArray,
                           b,
                           distance,
                           subtreeFlag) {
  ## Diagnostic count of calls.
  recursiveObject$diagnostic <- recursiveObject$diagnostic + 1
  ## Weak consistency check to ensure that the iteration matches the treeID in the nativeArray record.
  if(b != nativeArray$treeID[recursiveObject$offset]) {
    stop("Invalid nativeArray input record (treeID) at ", recursiveObject$offset, ".  Please contact Technical Support.")
  }
  ## Check if we are at the root node.  Otherwise increment the node at depth count.
  if (distance > 0) {
    ## We are not at the root node.
    if (distance <= length(recursiveObject$subtree$nodesAtDepth)) {    
      if (is.na(recursiveObject$subtree$nodesAtDepth[distance])) {
        ## This is the first split at this depth, so initialize the nodes at depth count.
        recursiveObject$subtree$nodesAtDepth[distance] <- 1
      }
      else {
        ## Increment the count of nodes at this depth.
        recursiveObject$subtree$nodesAtDepth[distance] <- recursiveObject$subtree$nodesAtDepth[distance] + 1
      }
    }
  }
  ## Read the current nativeArray split parameter.
  splitParameter <- nativeArray$parmID[recursiveObject$offset]
  ## Determine whether this is a terminal node.  
  if (splitParameter == 0) {
    terminalFlag <- TRUE
  }
  else if (splitParameter != 0) {
    terminalFlag <- FALSE
  }
  if (!terminalFlag) {
    ## Update the maximal subtree information for the parameter if it
    ## has not already been encountered in the tree.
    if (subtreeFlag[splitParameter] == FALSE) {
      ## Increment the subtree count for the parameter.
      recursiveObject$subtree$count[splitParameter] <- recursiveObject$subtree$count[splitParameter] + 1
      ## Update the (pre) mean distance.
      if (is.na(recursiveObject$subtree$meanSum[splitParameter])) {
        recursiveObject$subtree$meanSum[splitParameter] <- distance
      }
      else {
        recursiveObject$subtree$meanSum[splitParameter] <- recursiveObject$subtree$meanSum[splitParameter] + distance
      }
      if (max.order > 0) {
        ## Update distance encountered for the order statistic.
        ## The last element, NA, in the temporary vector is a filler to handle which.max() silliness. 
        orderVector <- c(recursiveObject$subtree$order[splitParameter, ], distance, NA)
        ## Identify the first NA occurence.
        index <- which.max(is.na(orderVector))
        ## The first (relevant) NA is filled with the current distance.
        orderVector[index] <- distance
        ## The vector is then sorted.
        sortedVector <- sort(orderVector[1:index])
        ## Restore the length of the order vector, if necessary, by apppending NA's.
        if (index <= max.order) {
          orderVector <- c(sortedVector, rep(NA, max.order-index))
        }
        else {
          orderVector <- sortedVector[1:max.order]
        }
        ## Update the order statistic.
        recursiveObject$subtree$order[splitParameter, ] <- orderVector
      }
      else {
        ## Update the minimum distance encountered for the parameter.
        if (is.na(recursiveObject$subtree$order[splitParameter])) {
          recursiveObject$subtree$order[splitParameter] <- distance
        }
        else {
          recursiveObject$subtree$order[splitParameter] <- min(recursiveObject$order[splitParameter], distance, na.rm = TRUE)
        }
      }
      ## Indicate that the parameter has been split on.
      subtreeFlag[splitParameter] <- TRUE
      if (sub.order == TRUE) {
        ## Update the diagonal element with the minimum distance encountered for the parameter.
        if (is.na(recursiveObject$subtree$subOrderDiag[splitParameter])) {
          recursiveObject$subtree$subOrderDiag[splitParameter] <- distance
        }
        else {
          recursiveObject$subtree$subOrderDiag[splitParameter] <- min(recursiveObject$subtree$subOrderDiag[splitParameter], distance, na.rm = TRUE)
        }
        ## Create the recursive2 list object.  Note that the relative depth starts at zero (0) for the subtree.  Also note
        ## that the object returns a vector containing the unnormalized minimum maximal w-subtree distance.  
        recursive2Object <- list(offset = recursiveObject$offset,
                                 depth = 0,
                                 minimumVector = rep(NA, dim(recursiveObject$subtree$subOrder)[2]),
                                 diagnostic = recursiveObject$diagnostic2)
        ## Initialize the split flags for the sub-recursion.  Ignore
        ## the diagonal element by setting the split flag.
        subtree2Flag <- rep(FALSE, dim(recursiveObject$subtree$subOrder)[2])        
        subtree2Flag[splitParameter] <- TRUE
        ## Do the sub-recursion.
        recursive2Object <- rfsrcParse2Tree(recursive2Object,
                                            nativeArray,
                                            b,
                                            distance=0,
                                            subtreeFlag=subtree2Flag)
        recursiveObject$diagnostic2 <- recursiveObject$diagnostic2 + recursive2Object$diagnostic
        ## Set the element corresponding to the diagonal element temporarily to the penalized value, to avoid being manipulated.
        recursive2Object$minimumVector[splitParameter] <- recursive2Object$depth
        ## Determine which w-subtrees in this v-subtree do not exist and set them to the penalized value.
        recursive2Object$minimumVector[which(is.na(recursive2Object$minimumVector))] <- recursive2Object$depth
        ## Normalize the relative distances by the maximum terminal node depth encountered for this v-subtree.
        recursive2Object$minimumVector <- recursive2Object$minimumVector / recursive2Object$depth
        ## Update the minimum w-subtree distance encountered in the tree.
        recursiveObject$subtree$subOrder[splitParameter, ] <- pmin(recursiveObject$subtree$subOrder[splitParameter, ], recursive2Object$minimumVector)
      }  ## if (sub.order == TRUE) ...
    }  ## if (subtreeFlag[splitParameter] == FALSE) ...
  }  ## if (terminalFlag and distance > 0) then adjust the node count by - 1
  else if (distance > 0) {
    recursiveObject$subtree$nodesAtDepth[distance] <- recursiveObject$subtree$nodesAtDepth[distance] - 1
  }
  ## Update the maximum depth encountered for this tree.
  recursiveObject$subtree$depth <- max(recursiveObject$subtree$depth, distance)
  ## Increment the offset.
  recursiveObject$offset <- recursiveObject$offset + 1
  ## Parse left and then right, if this is not a terminal node.
  if (terminalFlag == FALSE) {
    ## Increment the (parsed) tree distance. 
    distance <- distance + 1
    ## Parse left:
    recursiveObject <- rfsrcParseTree(recursiveObject, max.order, sub.order, nativeArray, b, distance, subtreeFlag)
    ## Parse right:
    recursiveObject <- rfsrcParseTree(recursiveObject, max.order, sub.order, nativeArray, b, distance, subtreeFlag)
  }
  else {
    ## Update the terminal node depth sum. 
    recursiveObject$subtree$terminalDepthSum <- recursiveObject$subtree$terminalDepthSum + distance
  }
  return(recursiveObject)
}
## -----------------------------------------------------------------------
##
##  SECONDARY RECURSION
##
## -----------------------------------------------------------------------
## Determine the second order maximal subtree distance.
## Recursive Object Definition:
##
## recursiveObject =
##   list(offset,
##        depth = 0,
##        minimumVector = rep(NA, numParm),
##        diagnostic)
##
## Inputs:
##   distance
##     - distance (depth) from the root of the (primary)
##       v-subtree.  Thus, this always starts at zero (0).
##     
##
## Outputs:  
##   recursiveObject$minimumDistance
##     - vector of length [numParm]
##     - second order unnormalized minimum maximum w-subtree distances
##       relative to the subtree defined by the initial root call.
##   recursiveObject$depth
##     - maximum terminal node depth, relative to the v-subtree.
##
## Algorithm Notes:  
##
## We use the terminology (v,w) for a w-maximal subtree within
## (relative to) a v-maximal subtree.
## 
## STEP-1:
## 
## For a given v, and a given v-max subtree, we search for all w-max
## subtrees within this v-max subtree, where v != w.  We single out
## the w-subtree with minimum depth -- that is, the w-split closest to
## the v-split.  Thus, when we talk about depths of w, these are
## relative depths, relative to the v-max subtree.  So the depth at
## root of the v-max subtree is defined to be zero (0).
## 
## For a given v, and a given v-max subtree, we also determine the
## maximum terminal node depth.  Let's call this the penalty depth.
## Again, this is relative to the v-max subtree.
## 
## If a particular w-max subtree does not exist for a given v and
## v-max subtree, it is assigned the penalty depth.
## 
## Now, there are p-1 minimum w-max subtree depths, given by analyzing
## all w != v.  These distances are normalized (divided by) by the
## penalty depth for the v-max subtree, so that all p-1 values are in
## (0,1].
## 
## 
## STEP-2:
## Consider v as fixed.  We do STEP-1 for all v-max subtrees.  Say
## there are count(v) of these.  We then take the MINIMUM normalized depth across the count(v)
## vectors of length p-1 that give the minimum w-max subtree depths by
## dividing by count(v).
## 
## STEP-3:
## Currently, if the v-max subtree does not exist, the row
## corresponding to v in the $subOrder matrix, namely [v, ] will be
## one (1) except for the element corresponding to the diagonal [v,v].
## It will be zero (0).
## 
## STEP-4:
## 
## All of the above are repeated for each tree in the forest such that
## the [p]x[p] matrix for each tree is normalized (divided by) the the
## number of valid (non-stumped) trees in the forest.
## 
## There is the rare possibility that a xvar v will never be
## split upon within the forest.  The entire row for this xvar,
## given by [v, ], will be NA.
rfsrcParse2Tree <- function(recursiveObject,
                            nativeArray,
                            b,
                            distance,
                            subtreeFlag) {
  ## Diagnostic count of calls.
  recursiveObject$diagnostic = recursiveObject$diagnostic + 1
  ## Weak consistency check to ensure that the iteration matches the treeID in the nativeArray record.
  if(b != nativeArray$treeID[recursiveObject$offset]) {
    stop("Invalid nativeArray input record (treeID) at ", recursiveObject$offset, ".  Please contact Technical Support.")
  }
  ## Read the current nativeArray split parameter.
  splitParameter = nativeArray$parmID[recursiveObject$offset]
  ## Determine whether this is a terminal node.  
  if (splitParameter == 0) {
    terminalFlag = TRUE
  }
  else if (splitParameter != 0) {
    terminalFlag = FALSE
  }
  if (splitParameter != 0) {
    ## Check that the parameter has not already been encountered along this path.
    if (subtreeFlag[splitParameter] == FALSE) {
      ## Update the minimum maximal subtree distance encountered for the parameter.
      if (is.na(recursiveObject$minimumVector[splitParameter])) {
        recursiveObject$minimumVector[splitParameter] = distance
      }
      else {
        recursiveObject$minimumVector[splitParameter] = min(recursiveObject$minimumVector[splitParameter], distance, na.rm = TRUE)
      }
      ## Indicate that the parameter has been split on
      subtreeFlag[splitParameter] = TRUE
    }
  }
  ## Update the (relative) maximum depth encountered for this tree.
  recursiveObject$depth = max(recursiveObject$depth, distance)
  ## Increment the (parsed) tree distance. 
  distance = distance + 1
  ## Increment the offset.
  recursiveObject$offset = recursiveObject$offset + 1
  ## Parse left and then right, if this is not a terminal node.
  if (terminalFlag == FALSE) {
    ## Parse left:
    recursiveObject = rfsrcParse2Tree(recursiveObject, nativeArray, b, distance, subtreeFlag)
    ## Parse right:
    recursiveObject = rfsrcParse2Tree(recursiveObject, nativeArray, b, distance, subtreeFlag)
  }
  return(recursiveObject)
}
## -----------------------------------------------------------------------
##
##  Supporting Functions
##
## -----------------------------------------------------------------------
## minimal depth density
minDepthProb <- function(p, D, l) {
  if (D <= 0) {
    return(1)
  }
  log.term <- ifelse(p > 1, log(1 - 1 / p), 0)
  prob <- rep(0, D + 1)
  if (is.null(l)) {
    d <- 0:(D - 1)
    ld <- 2^d
    Ld <- ld - 1
  } else {
    l <- l[1:D]
    ld <- l
    if (D > 1) {
      ## cumulative number of nodes above depth d
      Ld <- c(0, cumsum(l[1:(D - 1)]))
    } else {
      Ld <- 0
    }
  }
  prob[1:D] <- exp(Ld * log.term) * (1 - exp(ld * log.term))
  prob[D + 1] <- 1 - sum(prob[1:D])
  if (prob[D + 1] < 0) {
    prob[D + 1] <- 0
    prob <- prob / sum(prob)
  }
  prob
}
## second order depth density
secondOrderDepthProb <- function(p, D, l) {
  if (D <= 0) {
    return(1)
  }
  log.term <- ifelse(p > 1, log(1 - 1 / p), 0)
  if (is.null(l)) {
    l <- sapply(0:(D - 1), function(d) 2^d)
  } else {
    l <- l[1:D]
  }
  ## First compute the minimal depth density
  md.prob <- rep(0, D + 1)
  ld <- l
  if (D > 1) {
    Ld <- c(0, cumsum(l[1:(D - 1)]))
  } else {
    Ld <- 0
  }
  md.prob[1:D] <- exp(Ld * log.term) * (1 - exp(ld * log.term))
  md.prob[D + 1] <- 1 - sum(md.prob[1:D])
  if (md.prob[D + 1] < 0) {
    md.prob[D + 1] <- 0
    md.prob <- md.prob / sum(md.prob)
  }
  ## Now define the second order depth density
  prob <- rep(0, D + 1)
  if (D >= 2) {
    g <- exp(-l * log.term) - 1
    cumg <- cumsum(g)
    prob[2:D] <- md.prob[2:D] * cumg[1:(D - 1)]
  }
  prob[D + 1] <- 1 - sum(prob[1:D])
  if (prob[D + 1] < 0) {
    prob[D + 1] <- 0
    prob <- prob / sum(prob)
  }
  prob
}
## calculates minimal depth density under the null
## calculates first/second moments and robust estimate of s.d
## treat stumps as equivalent to trees with tree depth = 1.
minDepthStat <- function(p, D=NULL, l=NULL) {
  if (is.null(D) & is.null(l)) stop("set D or l")
  if (!is.null(l)) {
    D <- length(l)
  }
  D.support <- c(0:D)
  prob <- minDepthProb(p, D=D, l=l)
  prob.robust <- prob
  prob.robust[length(prob)] <- 0
  prob.robust <- prob.robust / sum(prob.robust, na.rm = TRUE) 
  first.moment <- sum(D.support * prob, na.rm = TRUE)
  second.moment <- sum((D.support^2) * prob, na.rm = TRUE)
  sd.robust <- sqrt(sum((D.support^2) * prob.robust, na.rm = TRUE)
                    - (sum(D.support * prob.robust, na.rm = TRUE))^2)
  return(list(prob = prob, first.moment = first.moment, second.moment = second.moment, sd.robust = sd.robust))
}
## calculates second order depth density under the null
## returns its first moment
## treat stumps as equivalent to trees with tree depth = 1.
secondOrderDepthStat <- function(p, D=NULL, l=NULL) {
  if (is.null(D) & is.null(l)) stop("set D or l")
  if (!is.null(l)) {
    D <- length(l)
  }
  D.support <- c(0:D)
  prob <- secondOrderDepthProb(p, D=D, l=l)
  sum(D.support * prob, na.rm = TRUE)
}
## parses minimal depth and second order depth
## returns the md density
## returns the md first (*threshold*) and second moment
## returns a robust estimate of standard deviation (gets rid of atom in the right tail)
## returns (new) threshold for second order statistic
## v is the max.subtree object
parseDepthOrder <- function(v, max.order) {
  ## preliminary checks: can this calculation even be done?
  if (is.null(v$mean) & max.order > 0) {
    return(NULL)
  }
  if (is.null(v$order) & max.order == 0) {
    return(NULL)
  }
  ## root node is assigned a value of 1 nodes at depth
  ## although not technically correct, it is relatively benign
  nodes.at.depth <- rbind(1, v$nodes.at.depth)
  ## determine the tree height (height = 1 + depth)
  ## determine the average tree height
  ## determine the maximum tree height
  treeHeight <- apply(nodes.at.depth, 2, function(l) {sum(!is.na(l))})
  avgTreeHeight <- mean(treeHeight, na.rm=TRUE)
  maxTreeHeight <- max(treeHeight, na.rm=TRUE)
  ## determine the number of trees in the forest
  ntree <- length(v$terminal)
  ## mean minimal depth
  ## break this up into two cases
  if (max.order > 0) {
    ## set the dimension
    p <- length(v$mean)
    ## case I: substitute forest averaged values for parameters
    nodes.at.depth.avg <- apply(nodes.at.depth, 1, mean, na.rm = TRUE)
    l <- nodes.at.depth.avg[1:max(avgTreeHeight - 1, 1)]
    minDepthStatObj <- minDepthStat(p, l = l)
    first.moment <- minDepthStatObj$first.moment
    second.moment <- minDepthStatObj$second.moment
    sd.robust <- minDepthStatObj$sd.robust
    prob <- minDepthStatObj$prob
    second.order.moment <- secondOrderDepthStat(p, l = l)
  }
  else {
    ## case II: tree-averaged (by-tree) analysis
    ## set the dimension
    p <- nrow(v$order)
    ## extract the tree density in a matrix format
    ## extract each trees depth
    prob <- sapply(1:ntree, function(tree) {
      l <- nodes.at.depth[, tree][1:sum(nodes.at.depth[, tree] > 0, na.rm = TRUE)]
      c(minDepthStat(p, l = l)$prob, rep(0, maxTreeHeight - treeHeight[tree] * (1 + 1 * (treeHeight[tree] == 1))))
    })
    ## extract the tree minimal threshold value
    first.moment <- unlist(sapply(1:ntree, function(tree) {
      l <- nodes.at.depth[, tree][1:sum(nodes.at.depth[, tree] > 0, na.rm = TRUE)]
      minDepthStat(p, l = l)$first.moment
    }))
    ## extract the tree second moment
    second.moment <- unlist(sapply(1:ntree, function(tree) {
      l <- nodes.at.depth[, tree][1:sum(nodes.at.depth[, tree] > 0, na.rm = TRUE)]
      minDepthStat(p, l = l)$second.moment
    }))
    ## extract the tree robust standard deviation
    sd.robust  <- unlist(sapply(1:ntree, function(tree) {
      l <- nodes.at.depth[, tree][1:sum(nodes.at.depth[, tree] > 0, na.rm = TRUE)]
      minDepthStat(p, l = l)$sd.robust
    }))
    ## extract the tree second order threshold value
    second.order.moment <- unlist(sapply(1:ntree, function(tree) {
      l <- nodes.at.depth[, tree][1:sum(nodes.at.depth[, tree] > 0, na.rm = TRUE)]
      secondOrderDepthStat(p, l = l)
    }))
  }
  return(list(prob = prob,
              first.moment = first.moment,
              sd = sqrt(second.moment - first.moment^2),
              sd.robust = sd.robust,
              second.order.moment = second.order.moment))
}   
max.subtree <- max.subtree.rfsrc

generic.predict.rfsrc <-
  function(object,
           newdata,
           m.target = NULL,
           importance = FALSE,
           get.tree = NULL,
           block.size = NULL,
           importance.xvar,
           na.action = c("na.omit", "na.impute", "na.random"),
           outcome = c("train", "test"),
           perf.type = NULL,
           proximity = FALSE,
           forest.wt = FALSE,
           ptn.count = 0,
           distance = FALSE,
           var.used = c(FALSE, "all.trees", "by.tree"),
           split.depth = c(FALSE, "all.trees", "by.tree"),
           case.depth = FALSE,
           seed = NULL,
           do.trace = FALSE,
           membership = FALSE,
           marginal.xvar = NULL,
           ...)
{
  univariate.nomenclature <- TRUE
  ## get any hidden options
  user.option <- list(...)
  ## incoming parameter checks: all are fatal
  if (missing(object)) {
    stop("object is missing!")
  }
  ## incoming object must be a grow forest or a forest object
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2    &
      sum(inherits(object, c("rfsrc", "forest"), TRUE) == c(1, 2)) != 2)
    stop("this function only works for objects of class `(rfsrc, grow)' or '(rfsrc, forest)'")
  ## grow forests must have true forest information
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) == 2) {
    if (is.forest.missing(object)) {
      stop("Forest information for prediction is missing.  Re-run rfsrc (grow call) with forest=TRUE")
    }
  }
  ## case specific performance
  cse  <- is.hidden.cse(user.option)
  csv  <- is.hidden.csv(user.option)
  ## verify importance (needed for jitt)
  importance <- match.arg(as.character(importance)[1], c(FALSE, TRUE,
          "none", "anti", "permute", "random", "anti.joint", "permute.joint", "random.joint"))
  if (grepl("joint", importance)) {
    vimp.joint <- TRUE
  }
  else {
    vimp.joint <- FALSE
  }
  ## verify na.action (needed for jitt)
  na.action <- match.arg(na.action, c("na.omit", "na.impute", "na.random"))
  if (inherits(object, "anonymous") && na.action == "na.impute") {
      na.action <- "na.mean"
  }
  ## verify other key options
  outcome <- match.arg(outcome, c("train", "test"))
  proximity <- match.arg(as.character(proximity), c(FALSE, TRUE, "inbag", "oob", "all", "hybrid"))
  forest.wt <- match.arg(as.character(forest.wt), c(FALSE, TRUE, "inbag", "oob", "all"))  
  distance <- match.arg(as.character(distance), c(FALSE, TRUE, "inbag", "oob", "all", "hybrid"))
  ## set restore.mode and the ensemble option
  ## newdata missing --> restore.mode = TRUE
  ## outcome = "test" --> restore.mode = FALSE for R-processing but switched later for C-code function call
  if (missing(newdata)) {
    restore.mode <- TRUE
    outcome <- "train"
    ensemble <- "all"##remove option to select between "all", "oob", "inbag"
  }
  else {##there is test data present
    restore.mode <- FALSE
    ## special treatment for outcome=="test" (which is really restore mode)
    if (outcome == "test") {
      ensemble <- "all"##remove option to select between "all", "oob", "inbag"
    }
    ## standard prediction scenario on new test data - there is no OOB
    else {
      ensemble <- "inbag"
    }
  }
  ## check if this is an anonymous object
  ## coerce values as necessary
  ## graceful return if restore.mode = TRUE which is not allowed for anonymous
  if (inherits(object, "anonymous")) {
    anonymize.bits <- 2^26
    outcome <- "train"
    if (restore.mode) {
      stop("in order to predict with anonymous forests please provide a test data set")
    }
  }
  else {
    anonymize.bits <- 0
  }
  ## jitt
  jitt <- is.hidden.jitt(user.option, importance, na.action, inherits(object, "anonymous"), partial = FALSE)
  ## rfq
  rfq <- is.hidden.rfq(user.option)
  ## quantile regression
  gk.quantile <- is.hidden.gk.quantile(user.option)
  prob <- is.hidden.prob(user.option)
  prob.epsilon <- is.hidden.prob.epsilon(user.option)
  ## vimp
  vimp.threshold  <- is.hidden.vimp.threshold(user.option)
  ## set the family
  family <- object$family
  ## pull the x-variable and y-outcome names from the grow object
  xvar.names <- object$xvar.names
  yvar.names <- object$yvar.names
  subj.names <- object$subj.names
  importance.xvar <- get.importance.xvar(importance.xvar, importance, object)
  importance.xvar.idx <- match(importance.xvar, xvar.names)
  marginal.xvar <- get.marginal.xvar(marginal.xvar, object)
  marginal.xvar.idx <- NULL
  if (length(marginal.xvar) > 0) {
    marginal.xvar.idx <- match(marginal.xvar, xvar.names)
  }
  var.used <- match.arg(as.character(var.used), c("FALSE", "all.trees", "by.tree"))
  if (var.used == "FALSE") var.used <- FALSE
  split.depth <- match.arg(as.character(split.depth),  c("FALSE", "all.trees", "by.tree"))
  if (split.depth == "FALSE") split.depth <- FALSE
  ## Currently we do not support split.depth for test data. TBD2
  split.depth = FALSE
  ## initialize the seed
  seed <- get.seed(seed)
  ## REDUCES THE OBJECT TO THE FOREST -- REDUCTION STARTS HERE
  ## hereafter we only need the forest and reassign "object" to the forest
  ## (TBD3) memory management "big.data" not currently implemented:
  big.data <- FALSE
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) == 2) {
    if (inherits(object, "bigdata")) {
      big.data <- TRUE
    }
    object <- object$forest
  }
  else {
    ## object is already a forest
    if (inherits(object, "bigdata")) {
      big.data <- TRUE
    }
  }
  ## confirm version coherence
  if (is.null(object$version)) {
    cat("\n  This function only works with objects created with the following minimum version of the package:")
    cat("\n    Minimum version:  ")
    cat("2.3.0")
    cat("\n    Your version:     ")
    cat("unknown")
    cat("\n")
    stop()
  }
    else {
      object.version <- as.integer(unlist(strsplit(object$version, "[.]")))
      installed.version <- as.integer(unlist(strsplit("3.7.0", "[.]")))
      minimum.version <- as.integer(unlist(strsplit("2.3.0", "[.]")))
      object.version.adj <- object.version[1] + (object.version[2]/10) + (object.version[3]/100)
      installed.version.adj <- installed.version[1] + (installed.version[2]/10) + (installed.version[3]/100)
      minimum.version.adj <- minimum.version[1] + (minimum.version[2]/10) + (minimum.version[3]/100)
      ## Minimum object version must be satisfied for us to proceed.  This is the only way
      ## terminal node restoration is guaranteed, due to RNG coherency.
      if (object.version.adj >= minimum.version.adj) {
        ## We are okay
      }
        else {
          cat("\n  This function only works with objects created with the following minimum version of the package:")
          cat("\n    Minimum version:  ")
          cat("2.3.0")
          cat("\n    Your version:     ")
          cat(object$version)
          cat("\n")
          stop()
        }
  }
  ## classification specific details related to rfq and perf.type
  class.relfrq <- NULL
  if (family == "class") {
    ## rfq specific details
    if (!is.null(rfq)) {##predict has specified rfq
      if (!rfq) {##predict does not want rfq
        ## nothing 
      }
      else {##predict has requested rfq
        class.relfrq <- prop.table(table(object$yvar))
      }
    }
    if (is.null(rfq)) {##predict  ambivalent about rfq
      if (!object$rfq) {##grow did not use rfq
        ## nothing -> rfq = FALSE
      }
      else {##grow used rfq - use grow spec
        class.relfrq <- prop.table(table(object$yvar))
        rfq <- TRUE
      }
    }
    ## performance details
    if (is.null(perf.type) && !is.null(object$perf.type)) {
      perf.type <- object$perf.type
    }
  }
  ## recover the split rule
  splitrule <- object$splitrule
  ## gk processing
  if (!is.null(gk.quantile) || object$gk.quantile) { 
    if (is.null(gk.quantile)) {##predict ambivalent about gk - use grow spec
      gk.quantile <- object$gk.quantile
    }
  }
  ## !!! here's where prob and prob.epsilon are set globally !!!
  gk.quantile <- get.gk.quantile(gk.quantile)
  prob.assign <- global.prob.assign(if (is.null(prob)) object$prob else prob,
                                    if (is.null(prob.epsilon)) object$prob.epsilon else prob.epsilon,
                                    gk.quantile, object$quantile.regr, splitrule, object$n)
  ## Determine the immutable yvar factor map which is needed for
  ## classification sexp dimensioning.  But, first convert object$yvar
  ## to a data frame which is required for factor processing
  #object$yvar <- as.data.frame(object$yvar)
  #colnames(object$yvar) <- yvar.names
  yfactor <- object$yvar.factor
  ## multivariate family details
  m.target.idx <- get.outcome.target(family, yvar.names, m.target)
  ## short cut to get the y-outcome type and number of levels
  yvar.types <- yfactor$types
  yvar.nlevels  <- yfactor$nlevels
  yvar.numeric.levels  <- yfactor$numeric.levels
  ## recover the individual subject identifiers, if they exist.
  subj <- object$subj
  ## get event information for survival families
  event.info <- object$event.info
  event.type <- event.info$event.type
  ## CR.bits assignment
  cr.bits <- get.cr.bits(family)
  ## determine the immutable xvar factor map
  xfactor <- object$xvar.factor
  any.xvar.factor <-  (length(xfactor$factor) + length(xfactor$order)) > 0
  ## short cut to get the x-variable type and number of levels
  xvar.types <- xfactor$types
  xvar.nlevels <- xfactor$nlevels
  xvar.numeric.levels <- xfactor$numeric.levels
  ## set dimensions
  n.xvar <- length(xvar.names)
  n <- object$n
  r.dim <- event.info$rdim
  ## set flags appropriately for unsupervised forests
  ## there are layers of checks appearing later, so some of these are redundant
  if (family == "unsupv") {
    outcome <- "train"
    perf.type <- "none"
    importance <- "none"
  }
  ## Override ptn.count by family.  Pruning is based on variance,
  ## thus, this is not yet well-defined in [S] settings.
  if (grepl("surv", family)) {
    ptn.count <- 0
  }
  ## by default: uno.weights are NULL; use.uno set by training object 
  uno.weights.test <- NULL
  use.uno <- !is.null(object$uno.weights)
  if (!is.null(user.option$use.uno) && isFALSE(user.option$use.uno)) {
    use.uno <- FALSE
  }
  if (!is.null(user.option$use.uno) && isTRUE(user.option$use.uno)) {
    use.uno <- TRUE
  }
  ## ----------------------------------------------------------------
  ## From the native code's perspective, PRED mode can process one
  ## or two data sets.  If one data set is sent in, we assume
  ## we wish to restore the forest with original-training or
  ## pseudo-training data.  If two data sets are sent in, we
  ## assume we are sending in the original-training data set, and a
  ## test data set.  When outcome="test" we make a call to the native
  ## code with only one data set (the test data which becomes pseudo-training
  ## data).  We set the test outcome bit to allow the native
  ## code to distinguish this case from the case of the restore
  ## with the original training data.
  ## ----------------------------------------------------------------
  ##--------------------------------------------------------
  ##
  ## NON-RESTORE MODE PROCESSING (includes outcome=="test")
  ##
  ##--------------------------------------------------------
  if (!restore.mode) {
    ## Filter the test data based on the formula
    newdata <- newdata[, is.element(names(newdata), c(yvar.names, xvar.names)), drop = FALSE]
    ## impute.mean for anonymous forests
    if (inherits(object, "anonymous") && na.action == "na.mean") {
      newdata <- assign.impute.mean(newdata, object$impute.mean)
      na.action <- "na.omit"
    }
    ## Check that test/train factors are the same.  If factor has an
    ## NA in its levels, remove it.  Confirm factor labels overlap.
    newdata <- rm.na.levels(newdata, xvar.names)
    #newdata.xfactor <- extract.factor(newdata, xvar.names)
    #if (!setequal(xfactor$factor, newdata.xfactor$factor)) {
    #  stop("x-variable factors from test data do not match original training data")
    #}
    #if (!setequal(xfactor$order, newdata.xfactor$order)) {
    #  stop("(ordered) x-variable factors from test data do not match original training data")
    #}
    any.outcome.factor <- family == "class"
    if (family == "class+" | family ==  "mix+") {
      if (length(intersect("R", yfactor$generic.types[m.target.idx])) == 0) {
        any.outcome.factor <- TRUE
      }
    }
    ## If the outcomes contain factors we need to check that train/test y-outcomes are compatible
    if (any.outcome.factor) {
      if (sum(is.element(names(newdata), yvar.names)) > 0) {
        newdata <- rm.na.levels(newdata, yvar.names)
        #newdata.yfactor <- extract.factor(newdata, yvar.names)
        #if (!setequal(yfactor$factor, newdata.yfactor$factor)) {
        #  stop("class outcome from test data does not match original training data")
        #}
        #if (!setequal(yfactor$order, newdata.yfactor$order)) {
        #  stop("(ordered) class outcome from test data does not match original training data")
        #}
      }
    }
    ## one final (possibly redundant) check confirming coherence of train/test xvars
    ## previously commented out: but put we this back because error message is helpful
    if (length(xvar.names) != sum(is.element(xvar.names, names(newdata)))) {
      stop("x-variables in test data do not match original training data")
    }
    ## coherence of train/test yvars (assuming test yvars are available)
    yvar.present <- sum(is.element(yvar.names, names(newdata))) > 0
    if (yvar.present && length(yvar.names) != sum(is.element(yvar.names, names(newdata)))) {
      stop("y-variables in test data do not match original training data")
    }
    ## Force test factor levels to equal grow factor levels
    ## this is crucial to ensuring an immutable map
    if (any.xvar.factor) {
      newdata <- check.factor(newdata, xfactor)
    }
    ## If the outcomes contain factors we need to check factor coherence.
    if (any.outcome.factor) {
      if (yvar.present) {
        ### COMMENT THIS OUT IF YOU WANT TO ELIMINATE BAD Y-TEST OUTCOME LABELS
        ### newdata <- check.factor(newdata, yfactor, ignore = FALSE)
        newdata <- check.factor(newdata, yfactor, ignore = TRUE)
      }
    }
    ## Extract test yvar names (if any) and xvar names.
    if (yvar.present) {
      fnames <- c(yvar.names, xvar.names)
    }
      else {
        fnames <- xvar.names
      }
    ## Data conversion to numeric mode
    newdata <- finalizeData(fnames, newdata, na.action)
    ## Extract the test x-matrix and sort the columns as in the original training data.
    ## this accomodates incoming test x-matrix in a different order.
    xvar.newdata  <- as.matrix(newdata[, xvar.names, drop = FALSE])
    n.newdata <- nrow(newdata)
    ## Save the row names for later overlay
    newdata.row.names <- rownames(xvar.newdata)
    ## Process the y-outcomes and set their dimension
    ## note that in unsupervised mode there are no responses
    ## r.dim.newdata is set correctly by get.grow.event.info() in this case
    if (yvar.present) {
      yvar.newdata <- as.matrix(newdata[, yvar.names, drop = FALSE])
      event.info.newdata <- get.grow.event.info(yvar.newdata, family, need.deaths = FALSE)
      r.dim.newdata <- event.info.newdata$r.dim
      ## Survival specific coherency checks
      ## if there are no deaths, turn off performance values and VIMP
      if (grepl("surv", family) && all(na.omit(event.info.newdata$cens) == 0)) {
        perf.type <- "none"
        importance <- "none"
      }
      ## Ensure consistency of event types
      if (grepl("surv", family) &&
          length(setdiff(na.omit(event.info.newdata$cens), na.omit(event.info$cens))) > 1) {
        stop("survival events in test data do not match training data")
      }
      ## uno weights
      if (!is.null(object$uno.weights) && isTRUE(use.uno)) {
        uno.weights.test <- uno.prepare.test(yvar.newdata[, 1],
                                             yvar.newdata[, 2],
                                             object$uno.weights$fit)$weight
      }      
      if (is.null(object$uno.weights) && isTRUE(use.uno)) {
        uno.weights.fit <- get.uno.weights.train(object$event.info$time, object$event.info$cens)$fit
        uno.weights.test <- uno.prepare.test(yvar.newdata[, 1],
                                             yvar.newdata[, 2],
                                             uno.weights.fit)$weight
      }      
    }
    else {
        ## Disallow outcome=TEST without y-outcomes
        if (outcome == "test") {
          stop("outcome=TEST, but the test data has no y values, which is not permitted")
        }
        ## There are no outcomes.
        r.dim.newdata <- 0
        yvar.newdata <-  NULL
        perf.type <- "none"
        importance <- "none"
    }
    ## Remove xvar row and column names for proper processing by the native library
    ## does not apply when outcome = TEST because the xvar TEST data has been made NULL
    if (outcome != "test") {
      rownames(xvar.newdata) <- colnames(xvar.newdata) <- NULL
    }
    ## We don't need the test data anymore
    remove(newdata)
  }
  ##--------------------------------------------------------
  ##
  ## RESTORE MODE PROCESSING (excludes outcome=="test")
  ##
  ##--------------------------------------------------------
  else {
    ## There cannot be test data in restore mode
    ## The native code switches based on n.newdata being zero (0).  Be careful.
    n.newdata <- 0
    r.dim.newdata <- 0
    xvar.newdata <- NULL
    yvar.newdata <-  NULL
    ## Outcome is set to train for the native code
    ## Determine whether performance values are requested
    outcome <- "train"
    if (object$bootstrap == "none" || object$bootstrap == "by.node" || family == "unsupv") {
      importance <- "none"
      perf.type <- "none"
    }
    else {
      ## do nothing.
    }
    ## restore hidden parameters
    if (is.null(user.option$vimp.threshold)) {
      vimp.threshold <- object$vimp.threshold 
    }
    ## uno weights -- assign training weights
    if (!is.null(object$uno.weights) && isTRUE(use.uno)) {
        uno.weights.test <- object$uno.weights$weight
    }
    if (is.null(object$uno.weights) && isTRUE(use.uno)) {
      uno.weights.test <- get.uno.weights.train(object$event.info$time, object$event.info$cens)$weight
    }
  } ## ends restore.mode check
  ## ------------------------------------------------------------
  ## We have completed the restore/non-restore mode processing
  ## ------------------------------------------------------------
  ## Final processing of xvar and yvar test data
  ## depends on "outcome"
  ## outcome=train
  if (outcome == "train") {
    ## data conversion for y-training data
    yvar <- as.matrix(data.matrix(data.frame(object$yvar)))
    ## respect the training options related to bootstrapping:
    sampsize <- round(object$sampsize(n))
    case.wt <- object$case.wt
    samp <- object$samp
    ## data conversion for x-training data
    if (anonymize.bits == 0) {
      xvar <- as.matrix(data.matrix(object$xvar))
      rownames(xvar) <- colnames(xvar) <- NULL
    }
    else {
      xvar <- NULL
    }
  }
  if (outcome == "test") {
    ## outcome=test
    ## From the native code perspective we are in pseudo-restore mode
    ## From the R side, it is convenient to now pretend we are
    ## in restore mode, so we swap the training data out with the test data
    ## Performance is always requested for this setting
    ## swap the data
    xvar <- xvar.newdata
    yvar <- yvar.newdata
    rownames(xvar) <- colnames(xvar) <- NULL
    restore.mode <- TRUE
    ## set the dimensions
    n <- nrow(xvar)
    r.dim <- r.dim.newdata
    ## pretend there is no test data, but do *not* get rid of it
    ## we need (and use) this data *after* the native code call
    n.newdata <- 0
    r.dim.newdata <- 0
    ## set the sample size
    ## "swor" now handled because we now make sampsize a function of n
    sampsize <- round(object$sampsize(n))
    case.wt <- get.weight(NULL, n)
    samp <- NULL
  }
  ## initialize the number of trees in the forest
  ntree <- object$ntree
  ## process the get.tree vector that specifies which trees we want
  ## to extract from the forest.  This is only relevant to restore mode.
  ## The parameter is ignored in predict mode.
  get.tree <- get.tree.index(get.tree, ntree)
  ## get performance and rfq, gk bits
  perf.type <- get.perf(perf.type, FALSE, family)
  perf.bits <-  get.perf.bits(perf.type)
  rfq <- get.rfq(rfq)
  rfq.bits <- get.rfq.bits(rfq, family)
  gk.quantile.bits <- get.gk.quantile.bits(gk.quantile)
  bootstrap.bits <- get.bootstrap.bits(object$bootstrap)
  ## initialize the low bits
  ensemble.bits <- get.ensemble.bits(ensemble)
  importance.bits <- get.importance.bits(importance, perf.type)
  proximity.bits <- get.proximity.bits(restore.mode, proximity)
  distance.bits <- get.distance.bits(restore.mode, distance)
  split.depth.bits <- get.split.depth.bits(split.depth)
  var.used.bits <- get.var.used.bits(var.used)
  outcome.bits <- get.outcome.bits(outcome)
  case.depth.bits  <- get.case.depth.bits(case.depth)
  ## Initalize the high bits
  samptype.bits <- get.samptype.bits(object$samptype)
  ## forest weights
  forest.wt.bits <- get.forest.wt.bits(restore.mode, object$bootstrap, forest.wt)
  membership.bits <-  get.membership.bits(membership)
  terminal.qualts.bits <- get.terminal.qualts.predict.bits(object$terminal.qualts)
  terminal.quants.bits <- get.terminal.quants.predict.bits(object$terminal.quants)
  cse.bits = get.cse.bits(cse)
  csv.bits = get.csv.bits(csv)
  jitt.bits <- get.jitt.bits(jitt)
  ## set the data.pass flags: we do this here because the restore.mode flag is now finalized
  ## training data.pass acquires the grow data.pass flag 
  data.pass.bits <- get.data.pass.bits(object$data.pass)
  ## testing data.pass is na.action AND restore.mode dependent
  if (restore.mode == FALSE) {
    if (na.action == "na.omit") {
      data.pass.predict.bits  <- get.data.pass.predict.bits(TRUE)
    }
    else {
      data.pass.predict.bits  <- get.data.pass.predict.bits(FALSE)
    }
  }
  else {
    ## we are in restore mode -> we are safe and the test data.pass flag irrelevant
    data.pass.predict.bits  <- get.data.pass.predict.bits(FALSE)
  }
  ## We over-ride block-size in the case that get.tree is user specified
  block.size <- min(get.block.size.bits(block.size, ntree), sum(get.tree))
  ## Turn off partial option.
  partial.bits <- get.partial.bits(0)
  ## na.action bit
  ## in restore mode, we send in the protocol used for the training data.
  if (restore.mode) {
    na.action = object$na.action
  }
  na.action.bits <- get.na.action.bits(na.action)
  do.trace <- get.trace.bits(do.trace)
  ## The pivot in this predict related function is different from
  ## the pivot used in the grow related function.  In rfsrc we are
  ## referencing the list nativeOutput[[]].  Here we are referencing
  ## the $nativeArray[[]] object which is a massaged version of
  ## nativeOutput[[]].
  ## WARNING: Note that the maximum number of slots in the following
  ## foreign function call is 64.  Ensure that this limit is not
  ## exceeded.  Otherwise, the program will error on the call.
      pivot = 0
      chunk = 0
  ## set the maximum class levels
  max.class.levels <- 0
  ## Start the C external timer.
  ctime.external.start  <- proc.time()
  nativeOutput <- tryCatch({.Call("rfsrcPredict",
                                  as.integer(do.trace),
                                  as.integer(seed),
                                  as.integer(ensemble.bits +
                                             importance.bits +
                                             bootstrap.bits +
                                             proximity.bits +
                                             split.depth.bits +
                                             var.used.bits +
                                             outcome.bits +
                                             perf.bits +
                                             rfq.bits +
                                             cr.bits +
                                             gk.quantile.bits +
                                             anonymize.bits +
                                             case.depth.bits),
                                  as.integer(forest.wt.bits +
                                             distance.bits +
                                             samptype.bits +
                                             na.action.bits +
                                             membership.bits +
                                             terminal.qualts.bits +
                                             terminal.quants.bits +
                                             cse.bits +
                                             csv.bits +
                                             data.pass.bits +
                                             data.pass.predict.bits +
                                             jitt.bits),
                                  ## >>>> start of maxi forest object >>>>
                                  as.double(vimp.threshold),
                                  as.integer(ntree),
                                  as.integer(n),
                                  list(as.integer(length(yvar.types)),
                                       if (is.null(yvar.types)) NULL else as.character(yvar.types),
                                       if (is.null(yvar.types)) NULL else as.integer (sapply(1:length(yvar.nlevels), function(nn) {
                                         if(yvar.nlevels[nn] > 0) max(max.class.levels, yvar.nlevels[nn]) else 0})),
                                       if (is.null(yvar.numeric.levels)) NULL else sapply(1:length(yvar.numeric.levels), function(nn) {
                                         as.integer(length(yvar.numeric.levels[[nn]]))}),
                                       if (is.null(subj)) NULL else as.integer(subj),
                                       if (is.null(event.type)) NULL else as.integer(length(event.type)),
                                       if (is.null(event.type)) NULL else as.integer(event.type)),
                                  if (is.null(yvar.numeric.levels)) {
                                      NULL
                                  }
                                  else {
                                      lapply(1:length(yvar.numeric.levels),
                                             function(nn) {as.integer(yvar.numeric.levels[[nn]])})
                                  },
                                  if (is.null(yvar.types)) NULL else as.double(as.vector(yvar)),
                                  list(as.integer(n.xvar),
                                       as.character(xvar.types),
                                       if (is.null(xvar.types)) NULL else as.integer(xvar.nlevels),                                       
                                       if (is.null(xvar.numeric.levels)) NULL else sapply(1:length(xvar.numeric.levels), function(nn) {as.integer(length(xvar.numeric.levels[[nn]]))}),
                                       NULL,
                                       NULL),
                                  if (is.null(xvar.numeric.levels)) {
                                      NULL
                                  }
                                  else {
                                      lapply(1:length(xvar.numeric.levels),
                                             function(nn) {as.integer(xvar.numeric.levels[[nn]])})
                                  },
                                  if (is.null(xvar)) NULL else as.double(as.vector(xvar)),
                                  list(as.integer(length(case.wt)),
                                       if (is.null(case.wt)) NULL else as.double(case.wt),
                                       as.integer(sampsize),
                                       if (is.null(samp)) NULL else as.integer(samp)),
                                  list(if(is.null(event.info$time.interest)) as.integer(0) else as.integer(length(event.info$time.interest)),
                                       if(is.null(event.info$time.interest)) NULL else as.double(event.info$time.interest)),
                                  as.integer(object$totalNodeCount),
                                  as.integer(object$leafCount),
                                  list(as.integer(object$seed),
                                       if (is.null(object$seedVimp)) NULL else as.integer(object$seedVimp),
                                       as.integer(object$optLoGrow)),
                                  as.integer(0),
                                  ## Deleted base learner slot.
                                  NULL,
                                  as.integer((object$nativeArray)$treeID),
                                  as.integer((object$nativeArray)$nodeID),
                                  as.integer((object$nativeArray)$nodeSZ),
                                  as.integer((object$nativeArray)$brnodeID),
                                  ## This is hc_zero.  It is never NULL.
                                  list(as.integer((object$nativeArray)$parmID),
                                  as.double((object$nativeArray)$contPT),
                                  as.integer((object$nativeArray)$mwcpSZ),
                                  as.integer((object$nativeArray)$fsrecID),
                                  if (is.null((object$nativeFactorArray)$mwcpPT)) NULL else as.integer((object$nativeFactorArray)$mwcpPT)),
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  as.integer(object$nativeArrayTNDS$tnRMBR),
                                  as.integer(object$nativeArrayTNDS$tnAMBR),
                                  as.integer(object$nativeArrayTNDS$tnRCNT),
                                  as.integer(object$nativeArrayTNDS$tnACNT),
                                  as.double((object$nativeArrayTNDS$tnSURV)),
                                  as.double((object$nativeArrayTNDS$tnMORT)),
                                  as.double((object$nativeArrayTNDS$tnNLSN)),
                                  as.double((object$nativeArrayTNDS$tnCSHZ)),
                                  as.double((object$nativeArrayTNDS$tnCIFN)),
                                  as.double((object$nativeArrayTNDS$tnREGR)),
                                  as.integer((object$nativeArrayTNDS$tnCLAS)),
                                  ## <<<< end of maxi forest object <<<<
                                  list(if (is.null(m.target.idx)) as.integer(0) else as.integer(length(m.target.idx)),
                                       if (is.null(m.target.idx)) NULL else as.integer(m.target.idx)),
                                  as.integer(ptn.count),
                                  list(if (is.null(marginal.xvar.idx)) as.integer(0) else as.integer(length(marginal.xvar.idx)),
                                       if (is.null(marginal.xvar.idx)) NULL else as.integer(marginal.xvar.idx)),
                                  list(if (is.null(importance.xvar.idx)) as.integer(0) else as.integer(length(importance.xvar.idx)),
                                       if (is.null(importance.xvar.idx)) NULL else as.integer(importance.xvar.idx)),
                                  ## Partial variables disabled.
                                  list(as.integer(0),
                                       as.integer(0),
                                       as.integer(0),
                                       NULL,
                                       as.integer(0),
                                       NULL,
                                       NULL),
                                  as.integer(n.newdata),
                                  as.integer(r.dim.newdata),
                                  as.double(if (outcome != "test") yvar.newdata else NULL),
                                  as.double(if (outcome != "test") xvar.newdata else NULL),
                                  as.integer(block.size),
                                  list(if (is.null(prob.assign$prob)) as.integer(0) else as.integer(length(prob.assign$prob)),
                                       if (is.null(prob.assign$prob)) NULL else as.double(prob.assign$prob),
                                       if (is.null(prob.assign$prob.epsilon)) as.double(0) else as.double(prob.assign$prob.epsilon)),
                                  as.integer(get.tree),
                                  if (is.null(uno.weights.test)) NULL else as.double(uno.weights.test),
                                  as.integer(get.rf.cores()))}, error = function(e) {
                                    print(e)
                                    NULL})
  ## Stop the C external timer.
  ctime.external.stop <- proc.time()
  ## check for error return condition in the native code
  if (is.null(nativeOutput)) {
    stop("An error has occurred in prediction.  Please turn trace on for further analysis.")
  }
  ## determine missingness which is used for imputed data
  ## not done for anonymous forests or na.random imputation
  if (!inherits(object, "anonymous") && na.action != "na.random") {
    if (restore.mode) {
      n.miss <- get.nmiss(xvar, yvar)
    }
    else {
      n.miss <- get.nmiss(xvar.newdata, yvar.newdata)
     }
  }
  else {
    n.miss <- 0
  }
  ## extract the imputed data if there was missingness
  if (n.miss > 0) {
    imputed.data <- matrix(nativeOutput$imputation, nrow = n.miss)
    nativeOutput$imputation <- NULL
    imputed.indv <- imputed.data[, 1]
    imputed.data <- as.data.frame(imputed.data[, -1, drop = FALSE])
    if ((r.dim.newdata > 0) | (perf.type != "none")) {
      colnames(imputed.data) <- c(yvar.names, xvar.names)
    }
      else {
        colnames(imputed.data) <- xvar.names
      }
  }
  ## post-process the test data
  ## for restore mode there is no test data *except* when outcome=TEST
  if (!restore.mode | outcome == "test") {
    ## add row and column names to test xvar matrix
    xvar.newdata <- as.data.frame(xvar.newdata)
    rownames(xvar.newdata) <- newdata.row.names
    colnames(xvar.newdata) <- xvar.names
    ## map xvar factors back to original values
    xvar.newdata <- map.factor(xvar.newdata, xfactor)
    if (perf.type != "none") {
      ## add column names to test response matrix
      yvar.newdata <- as.data.frame(yvar.newdata)
      colnames(yvar.newdata) <- yvar.names
      ## map response factors back to original values
      yvar.newdata <- map.factor(yvar.newdata, yfactor)
    }
  }
  ## Map imputed data factors back to original values
  if (n.miss > 0) {
    imputed.data <- map.factor(imputed.data, xfactor)
    if (perf.type != "none") {
      imputed.data <- map.factor(imputed.data, yfactor)
    }
  }
  ## proximity
  if (proximity != FALSE) {
    if (restore.mode) {
      prox.n <- n
    }
    else {
        prox.n <- n.newdata
    }
    ## Are we conducting hybrid proximity?
    if (proximity == "hybrid") {
        proximity.out <- matrix(nativeOutput$proximity,
                                nrow = n.newdata,
                                ncol = n,
                                byrow = TRUE)
    }
    else {
        proximity.out <- matrix(0, prox.n, prox.n)
        count <- 0
        for (k in 1:prox.n) {
            proximity.out[k,1:k] <- nativeOutput$proximity[(count+1):(count+k)]
            proximity.out[1:k,k] <- proximity.out[k,1:k]
            count <- count + k
        }
    }
    nativeOutput$proximity <- NULL
  }
  else {
      proximity.out <- NULL
  }
  ## distance
  if (distance != FALSE) {
    if (restore.mode) {
      dist.n <- n
    }
    else {
      dist.n <- n.newdata
    }
    ## Are we conducting hybrid distance?
    if (distance == "hybrid") {
      distance.out <- matrix(nativeOutput$distance,
                             nrow = n.newdata,
                             ncol = n,
                             byrow = TRUE)
    }
    else {
      distance.out <- matrix(0, dist.n, dist.n)
      count <- 0
      for (k in 1:dist.n) {
        distance.out[k,1:k] <- nativeOutput$distance[(count+1):(count+k)]
        distance.out[1:k,k] <- distance.out[k,1:k]
        count <- count + k
      }
    }
    nativeOutput$distance <- NULL
  }
  else {
    distance.out <- NULL
  }
  ## forest weight matrix
  if (forest.wt != FALSE) {
    if (restore.mode) {
      forest.wt.n <- c(n, n)
    }
      else {
        forest.wt.n <- c(n.newdata, n)
      }
    forest.wt.out <- matrix(nativeOutput$weight, forest.wt.n, byrow = TRUE)
    nativeOutput$weight <- NULL
  }
    else {
      forest.wt.out <- NULL
    }
  ## forest case.depth
  if (case.depth != FALSE) {
    if (restore.mode) {
      case.depth.out <- matrix(nativeOutput$caseDepth, c(ntree, n), byrow = TRUE)
      nativeOutput$caseDepth <- NULL
    }
    else {
      case.depth.out <- matrix(nativeOutput$caseDepth, c(ntree, n.newdata), byrow = TRUE)
    }
    nativeOutput$caseDepth <- NULL
  }
  else {
    case.depth.out <- NULL
  }
  n.observed = if (restore.mode) n else n.newdata
  ## membership
  if (membership) {
    membership.out <- matrix(nativeOutput$nodeMembership, c(n.observed, ntree))
    nativeOutput$nodeMembership <- NULL
    if (restore.mode) {
      inbag.out <- matrix(nativeOutput$bootMembership, c(n.observed, ntree))
      nativeOutput$bootMembership <- NULL
    }
      else {
        inbag.out <- NULL
      }
    if (ptn.count > 0) {
      ptn.membership.out <- matrix(nativeOutput$pstnMembership, c(n.observed, ntree))
      nativeOutput$pstnMembership <- NULL
    }
      else {
        ptn.membership.out <- NULL
      }
  }
    else {
      membership.out <- NULL
      inbag.out <- NULL
      ptn.membership.out <- NULL
    }
  ## variables used
  if (var.used != FALSE) {
    if (var.used == "all.trees") {
      var.used.out <- nativeOutput$varUsed
      names(var.used.out) <- xvar.names
    }
      else {
        var.used.out <- matrix(nativeOutput$varUsed, nrow = ntree, byrow = TRUE)
        colnames(var.used.out) <- xvar.names
      }
    nativeOutput$varUsed <- NULL
  }
    else {
      var.used.out <-  NULL
    }
  ## split depth
  if (split.depth != FALSE) {
    if (split.depth == "all.trees") {
      split.depth.out <- array(nativeOutput$splitDepth, c(n, n.xvar))
    }
      else {
        split.depth.out <- array(nativeOutput$splitDepth, c(n, n.xvar, ntree))
      }
    nativeOutput$splitDepth <- NULL
  }
    else {
      split.depth.out <-  NULL
    }
  ## make the output object
  rfsrcOutput <- list(
    call = match.call(),
    family = family,
    n = n.observed,
    ntree = ntree,
    yvar = (if ((outcome == "train" & restore.mode) | (perf.type != "none")) {
      if (outcome == "train" & restore.mode)
        amatrix.remove.names(object$yvar) else amatrix.remove.names(yvar.newdata)} else NULL),
    yvar.names = yvar.names,
    xvar = (if(outcome != "test" & restore.mode) object$xvar else xvar.newdata),
    xvar.names = xvar.names,
    leaf.count = nativeOutput$leafCount,
    proximity = proximity.out,
    forest = object,
    forest.wt = forest.wt.out,
    case.depth = case.depth.out,
    distance = distance.out,
    ptn.membership = ptn.membership.out,
    membership = membership.out,
    splitrule = splitrule,
    inbag = inbag.out,
    var.used = var.used.out,
    uno.weights = uno.weights.test,
    imputed.indv = (if (n.miss>0) imputed.indv else NULL),
    imputed.data = (if (n.miss>0) imputed.data else NULL),
    split.depth  = split.depth.out,
    block.size = block.size,
    perf.type = perf.type,
    ctime.internal = nativeOutput$cTimeInternal,
    ctime.external = ctime.external.stop - ctime.external.start
  )
  ## memory management
  nativeOutput$leafCount <- NULL
  remove(object)
  remove(proximity.out)
  remove(forest.wt.out)
  remove(case.depth.out)
  remove(distance.out)
  remove(ptn.membership.out)
  remove(membership.out)
  remove(inbag.out)
  if (n.miss > 0) remove(imputed.indv)
  if (n.miss > 0) remove(imputed.data)
  remove(var.used.out)
  remove(split.depth.out)
  ## Safe the outputs.
  survOutput <- NULL
  classOutput <- NULL
  regrOutput <- NULL
  if(vimp.joint) {
    vimp.count <- 1
  }
    else {
      vimp.count <- length(importance.xvar)
    }
  ## family specific additions to the predict object
  if (grepl("surv", family)) {
      if ((length(event.info$event.type) > 1) &&
          (splitrule != "l2.impute") &&
          (splitrule != "logrankscore")) {
      coerced.event.count <- length(event.info$event.type)
    }
      else {
        coerced.event.count <- 1
      }
    if (family == "surv") {
      ## Right Censored names.
      ens.names <- list(NULL, NULL)
      mortality.names <- list(NULL, NULL)
      err.names <- list(NULL, NULL)
      vimp.names <- list(NULL, if (vimp.joint) "joint" else importance.xvar)
    }
      else {
        ## Competing Risk names.
        ens.names <- list(NULL, NULL, c(paste("condCHF.", 1:length(event.info$event.type), sep = "")))
        mortality.names <- list(NULL, paste("event.", 1:length(event.info$event.type), sep = ""))
        cif.names <- list(NULL, NULL, c(paste("CIF.", 1:length(event.info$event.type), sep = "")))
        err.names <- list(c(paste("event.", 1:length(event.info$event.type), sep = "")), NULL)
        vimp.names <- list(paste("event.", 1:length(event.info$event.type), sep = ""),
                           if(vimp.joint) "joint" else importance.xvar)
      }
    ## From the native code:
    ##   "allEnsbCHF"
    ##   "oobEnsbCHF"
    ## -> of dim [length(event.info$event.type)] x [RF_sortedTimeInterestSize] x [n]
    ##    where [length(event.info$event.type)] may be equal to [1].
    ## To the R code:
    ## -> of dim [n] x [RF_sortedTimeInterestSize] x [length(event.info$event.type)]
    chf <- (if (!is.null(nativeOutput$allEnsbCHF))
              adrop3d.last(array(nativeOutput$allEnsbCHF,
                                 c(n.observed, length(event.info$time.interest), length(event.info$event.type)),
                                 dimnames=ens.names), length(event.info$event.type)) else NULL)
    nativeOutput$allEnsbCHF <- NULL
    survOutput <- list(chf = chf)
    remove(chf)
    chf.oob <- (if (!is.null(nativeOutput$oobEnsbCHF))
                  adrop3d.last(array(nativeOutput$oobEnsbCHF,
                                     c(n.observed, length(event.info$time.interest), length(event.info$event.type)),
                                     dimnames=ens.names), length(event.info$event.type)) else NULL)
    nativeOutput$oobEnsbCHF <- NULL
    survOutput = c(survOutput, chf.oob = list(chf.oob))
    remove(chf.oob)
    ## From the native code:
    ##   "allEnsbMRT"
    ##   "oobEnsbMRT"
    ## -> of dim [length(event.info$event.type)] x [n]
    ## To the R code:
    ## -> of dim [n] x [length(event.info$event.type)]
    predicted <- (if (!is.null(nativeOutput$allEnsbMRT))
                    adrop2d.last(array(nativeOutput$allEnsbMRT,
                                       c(n.observed, length(event.info$event.type)), dimnames=mortality.names), coerced.event.count) else NULL)
    nativeOutput$allEnsbMRT <- NULL
    survOutput = c(survOutput, predicted = list(predicted))
    remove(predicted)
    predicted.oob <- (if (!is.null(nativeOutput$oobEnsbMRT))
                        adrop2d.last(array(nativeOutput$oobEnsbMRT,
                                           c(n.observed, length(event.info$event.type)), dimnames=mortality.names), coerced.event.count) else NULL)
    nativeOutput$oobEnsbMRT <- NULL
    survOutput <- c(survOutput, predicted.oob = list(predicted.oob))
    remove(predicted.oob)
    ## From the native code:
    ##   "allEnsbSRV"
    ##   "oobEnsbSRV"
    ## -> of dim [RF_sortedTimeInterestSize] x [n]
    ## To the R code:
    ## -> of dim [n] x [RF_sortedTimeInterestSize]
    survival <-  (if (!is.null(nativeOutput$allEnsbSRV))
                    matrix(nativeOutput$allEnsbSRV,
                           c(n.observed, length(event.info$time.interest))) else NULL)
    nativeOutput$allEnsbSRV <- NULL
    survOutput <- c(survOutput, survival = list(survival))
    remove(survival)
    survival.oob <-  (if (!is.null(nativeOutput$oobEnsbSRV))
                        matrix(nativeOutput$oobEnsbSRV,
                               c(n.observed, length(event.info$time.interest))) else NULL)
    nativeOutput$oobEnsbSRV <- NULL
    survOutput <- c(survOutput, survival.oob = list(survival.oob))
    remove(survival.oob)
    ## From the native code:
    ##   "allEnsbCIF"
    ##   "oobEnsbCIF"
    ##   -> of dim [length(event.info$event.type)] x [RF_sortedTimeInterestSize] x [n]
    ## To the native code:
    ##   -> of dim  [n] x [RF_sortedTimeInterestSize] x [length(event.info$event.type)]
    cif <- (if (!is.null(nativeOutput$allEnsbCIF))
              array(nativeOutput$allEnsbCIF,
                    c(n.observed, length(event.info$time.interest), length(event.info$event.type)),
                    dimnames=cif.names) else NULL)
    nativeOutput$allEnsbCIF <- NULL
    survOutput <- c(survOutput, cif = list(cif))
    remove(cif)
    cif.oob <- (if (!is.null(nativeOutput$oobEnsbCIF))
                  array(nativeOutput$oobEnsbCIF,
                        c(n.observed, length(event.info$time.interest), length(event.info$event.type)),
                        dimnames=cif.names) else NULL)
    nativeOutput$oobEnsbCIF <- NULL
    survOutput = c(survOutput, cif.oob = list(cif.oob))
    remove(cif.oob)
    ## From the native code:
    ##   "perfSurv"
    ##   -> of dim [ntree] x length(event.info$event.type)]
    ## To the R code:
    ##   -> of dim [ntree] x length(event.info$event.type)]
    if (!is.null(nativeOutput$perfSurv)) {
      err.rate <- adrop2d.first(array(nativeOutput$perfSurv,
                                      c(length(event.info$event.type), ntree),
                                      dimnames=err.names),
                                coerced.event.count)
      nativeOutput$perfSurv <- NULL
      if (family == "surv-CR") {
        survOutput = c(survOutput, err.rate = list(t(err.rate)))
      }
        else {
          survOutput = c(survOutput, err.rate = list(err.rate))
        }
      remove(err.rate)
    }
      ## From the native code:
      ##   "blockSurv"
      ##   -> of dim [block.cnt] x length(event.info$event.type)]
      ## To the R code:
      ##   -> of dim [block.cnt] x length(event.info$event.type)]
      if (!is.null(nativeOutput$blockSurv)) {
          err.block.rate <- adrop2d.first(array(nativeOutput$blockSurv,
                                                c(length(event.info$event.type), floor(ntree/block.size)),
                                                dimnames=err.names),
                                          coerced.event.count)
          nativeOutput$blockSurv <- NULL
          if (family == "surv-CR") {
              survOutput = c(survOutput, err.block.rate = list(t(err.block.rate)))
          }
          else {
              survOutput = c(survOutput, err.block.rate = list(err.block.rate))
          }
          remove(err.block.rate)
      }
    ## From the native code:
    ##   "vimpSurv"
    ##   -> of dim [n.xvar] x length(event.info$event.type)]
    ## To the R code:
    ##   -> of dim length(event.info$event.type)] x [n.xvar]
    if (!is.null(nativeOutput$vimpSurv)) {
      importance <- adrop2d.first(array(nativeOutput$vimpSurv,
                                        c(length(event.info$event.type), vimp.count),
                                        dimnames = vimp.names),
                                  coerced.event.count)
      nativeOutput$vimpSurv <- NULL
      if (family == "surv-CR") {
        survOutput = c(survOutput, importance = list(t(importance)))
      }
        else {
          survOutput = c(survOutput, importance = list(importance))
        }
      remove(importance)
    }
    survOutput = c(
      survOutput, list(
        time.interest = event.info$time.interest,
        ndead = (if (perf.type != "none") sum((if (restore.mode) yvar[, 2] else yvar.newdata[, 2]) !=0 , na.rm=TRUE) else NULL))
    )
    ## When TRUE we revert to univariate nomenclature for all the outputs.
    if(univariate.nomenclature) {
      rfsrcOutput <- c(rfsrcOutput, survOutput)
    }
      else {
        rfsrcOutput <- c(rfsrcOutput, survOutput = list(survOutput))
      }
  }
    else {
      ## We consider "R", "I", and "C" outcomes.  The outcomes are grouped
      ## by type and sequential.  That is, the first "C" encountered in the
      ## response type vector is in position [[1]] in the classification output
      ## list, the second "C" encountered is in position [[2]] in the
      ## classification output list, and so on.  The same applies to the
      ## regression outputs.  We also have a mapping from the outcome slot back
      ## to the original response vector type, given by the following:
      ## Given yvar.types = c("R", "C", "R", "C", "R" , "I")
      ## regr.index[1] -> 1
      ## regr.index[2] -> 3
      ## regr.index[3] -> 5
      ## clas.index[1] -> 2
      ## clas.index[2] -> 4
      ## clas.index[3] -> 6
      ## This will pick up all "C" and "I".
      class.index <- which(yvar.types != "R")
      class.count <- length(class.index)
      regr.index <- which(yvar.types == "R")
      regr.count <- length(regr.index)
      if (class.count > 0) {
        classOutput <- vector("list", class.count)
        ## Names of the classification outputs.
        names(classOutput) <- yvar.names[class.index]
        ## Vector to hold the number of levels in each factor response.
        levels.count <- array(0, class.count)
        ## List to hold the names of levels in each factor response.
        levels.names <- vector("list", class.count)
        counter <- 0
        for (i in class.index) {
          counter <- counter + 1
          ## Note that [i] is the actual index of the y-variables and not a sequential iterator.
          ## The sequential iteratior is [counter]
          levels.count[counter] <- yvar.nlevels[i]
          if (yvar.types[i] == "C") {
            ## This an unordered factor.
            ## Here, we don't know the sequence of the unordered factor list, so we identify the factor by name.
            levels.names[[counter]] <- yfactor$levels[[which(yfactor$factor == yvar.names[i])]]
          }
            else {
              ## This in an ordered factor.
              ## Here, we don't know the sequence of the ordered factor list, so we identify the factor by name.
              levels.names[[counter]] <- yfactor$order.levels[[which(yfactor$order == yvar.names[i])]]
            }
        }
        ## Incoming error rates: T=tree R=response L=level
        ## T1R1L0 T1R1L1 T1R1L2 T1R1L3 T1R2L0 T1R2L1 T1R2L2, T2R1L0 T2R1L1 T2R1L2 T2R1L3 T2R2L0 T2R2L1 T2R2L2, ...
        ## In GROW mode, all class objects are represented in the tree offset calculation.
        ## In PRED mode, the offsets are dependent on the only those targets that are requested!
        ## Yeilds tree.offset = c(1, 8, ...)
        tree.offset <- array(1, ntree)
        ## Sum of all level counts targetted classification.  For example,
        ## if R1 has 3 levels, and R2 has 2 levels, and R3 has 6 levels, and
        ## only R1 and R3 and targetted, then levels.total = 9 + 3
        levels.total <- 0
        if (ntree > 1) {
          ## Iterate over all the target outcomes and map them to the class list.
          for (i in 1:length(m.target.idx)) {
            ## This is the slot in the class list.  The class list spans all the classification
            ## outputs, some of which may be NULL.
            target.idx <- which (class.index == m.target.idx[i])
            ## Is the target a classification y-variable.
            if (length(target.idx) > 0) {
              levels.total <- levels.total + 1 + levels.count[target.idx]
            }
          }
          tree.offset[2:ntree] <- levels.total
        }
        tree.offset <-  cumsum(tree.offset)
        ## The block offset calculation mirrors the tree offset calculation, but differs in only the primary dimension.
        block.offset <- array(1, floor(ntree/block.size))
        if (floor(ntree/block.size) > 1) {
          block.offset[2:floor(ntree/block.size)] <- levels.total
        }
        block.offset <-  cumsum(block.offset)
        ## Incoming vimp rates: V=xvar R=response L=level
        ## V1R1L0 V1R1L1 V1R1L2 V1R1L3 V1R1L0 V1R2L1 V1R2L2, V2R1L0 V2R1L1 V2R1L2 V2R1L3 V2R2L0 V2R2L1 V2R2L2, ...
        ## Yeilds vimp.offset = c(1, 8, ...)
        vimp.offset <- array(1, vimp.count)
        if (vimp.count > 1) {
          vimp.offset[2:vimp.count] <- levels.total
        }
        vimp.offset <-  cumsum(vimp.offset)
        iter.ensb.start <- 0
        iter.ensb.end   <- 0
        ## From the native code:
        ##   "allEnsbCLS"
        ##   "oobEnsbCLS"
        ## -> of dim [[class.count]] x [levels.count[]] x [n]
        ##    where this is a ragged array.
        ## From the native code:
        ##   "perfClas"
        ## -> of dim [ntree] x [class.count] x [1 + levels.count[]]
        ## where the slot [.] x [.] x [1] holds the unconditional error rate.
        ## Note that this is a ragged array.
        ## To the R code:
        ## -> of dim [[class.count]] x [ntree] x [1 + levels.count[]]
        ## From the native code:
        ##   "blockClas"
        ## -> of dim [block.cnt] x [class.count] x [1 + levels.count[]]
        ## where the slot [.] x [.] x [1] holds the unconditional error rate.
        ## Note that this is a ragged array.
        ## To the R code:
        ## -> of dim [[class.count]] x [block.cnt] x [1 + levels.count[]]
        ## From the native code:
        ##   "vimpClas"
        ## -> of dim [n.xvar] x [class.count] x [1 + levels.count]
        ## where the slot [.] x [.] x [1] holds the unconditional vimp.
        ## Note that this is a ragged array.
        ## To the R code:
        ## -> of dim [[class.count]] x [1 + levels.count] x [n.xvar]
        ## From the native code:
        ##   "cseNum"
        ## -> of dim [resp.class.count] x [n]
        ## To the R code:
        ## -> of dim [[resp.class.count]] x [n]
        ## From the native code:
        ##   "cseDen"
        ## -> of dim [n]
        ## To the R code:
        ## -> of dim [n]
        ## From the native code:
        ##   "csvNum"
        ## -> of dim [n.xvar] x [resp.class.count] x [n]
        ## To the R code:
        ## -> of dim [[resp.class.count]] x [n] x [n.xvar]
        ## From the native code:
        ##   "csvDen"
        ## -> of dim [n.xvar] x [n] 
        ## To the R code:
        ## -> of dim  x [n]  x [n.xvar]
        for (i in 1:length(m.target.idx)) {
          target.idx <- which(class.index == m.target.idx[i])
          if (length(target.idx) > 0) {
            iter.ensb.start <- iter.ensb.end
            iter.ensb.end <- iter.ensb.end + (levels.count[target.idx] * n.observed)
            ens.names <- list(NULL, levels.names[[target.idx]])
            err.names <- c("all", levels.names[[target.idx]])
            vimp.names <- list(c("all", levels.names[[target.idx]]), if(vimp.joint) "joint" else importance.xvar)
            predicted <- (if (!is.null(nativeOutput$allEnsbCLS))
                            array(nativeOutput$allEnsbCLS[(iter.ensb.start + 1):iter.ensb.end],
                                  c(n.observed, levels.count[target.idx]), dimnames=ens.names) else NULL)
            classOutput[[target.idx]] <- list(predicted = predicted)
            response <- (if (!is.null(predicted)) get.bayes.rule(predicted, class.relfrq) else NULL)
            classOutput[[target.idx]] <- c(classOutput[[target.idx]], class = list(response))
            remove(predicted)
            remove(response)
            predicted.oob <- (if (!is.null(nativeOutput$oobEnsbCLS))
                                array(nativeOutput$oobEnsbCLS[(iter.ensb.start + 1):iter.ensb.end],
                                      c(n.observed, levels.count[target.idx]), dimnames=ens.names) else NULL)
            classOutput[[target.idx]] <- c(classOutput[[target.idx]], predicted.oob = list(predicted.oob))
            response.oob <- (if (!is.null(predicted.oob)) get.bayes.rule(predicted.oob, class.relfrq) else NULL)
            classOutput[[target.idx]] <- c(classOutput[[target.idx]], class.oob = list(response.oob))
            remove(predicted.oob)
            remove(response.oob)
            ## New case specific arrays, numerator:
            cse.num <- (if (!is.null(nativeOutput$cseClas))
                            array(nativeOutput$cseClas[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
            classOutput[[target.idx]] <- c(classOutput[[target.idx]], cse.num = list(cse.num))
            remove(cse.num)
            ## New case specific arrays, denominator:
            cse.den <- (if (!is.null(nativeOutput$cseDen))
                            array(nativeOutput$cseDen[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
            classOutput[[target.idx]] <- c(classOutput[[target.idx]], cse.den = list(cse.den))
            remove(cse.den)
            if (!is.null(nativeOutput$perfClas)) {
              err.rate <- array(0, c(1 + levels.count[target.idx], ntree))
              for (j in 1: (1 + levels.count[target.idx])) {
                err.rate[j, ]  <- nativeOutput$perfClas[tree.offset]
                tree.offset <- tree.offset + 1
              }
              row.names(err.rate) <- err.names
              classOutput[[target.idx]] <- c(classOutput[[target.idx]], err.rate = list(t(err.rate)))
              remove(err.rate)
            }
            if (!is.null(nativeOutput$blockClas)) {
              err.block.rate <- array(0, c(1 + levels.count[target.idx], floor(ntree/block.size)))
              for (j in 1: (1 + levels.count[target.idx])) {
                err.block.rate[j, ]  <- nativeOutput$blockClas[block.offset]
                block.offset <- block.offset + 1
              }
              row.names(err.block.rate) <- err.names
              classOutput[[target.idx]] <- c(classOutput[[target.idx]], err.block.rate = list(t(err.block.rate)))
              remove(err.block.rate)
            }
            if (!is.null(nativeOutput$vimpClas)) {
              importance <- array(0, c(1 + levels.count[target.idx], vimp.count), dimnames=vimp.names)
              for (j in 1: (1 + levels.count[target.idx])) {
                importance[j, ]  <- nativeOutput$vimpClas[vimp.offset]
                vimp.offset <- vimp.offset + 1
              }
              classOutput[[target.idx]] <- c(classOutput[[target.idx]], importance = list(t(importance)))
              remove(importance)
            }
            ## See GROW call for explanation of arrays:
            csv.idx  <- array(0, n.observed * n.xvar) 
            j2seq <- (target.idx - 1) * n.observed
            for (j in 1:n.xvar) {
              j1seq <- (j - 1) * n.observed
              csv.idx[(j1seq + 1) : (j1seq + n.observed)]  <- ( (j2seq + j1seq * class.count + 1) : (j2seq + j1seq * class.count + n.observed) ) 
            }
            ## New case specific arrays, numerator:
            csv.num <- (if (!is.null(nativeOutput$csvClas))
                          array(nativeOutput$csvClas[csv.idx], c(n.observed, n.xvar)) else NULL)
            classOutput[[target.idx]] <- c(classOutput[[target.idx]], csv.num = list(csv.num))
            remove(csv.num)
            ## New case specific arrays, denominator:
            csv.den <- (if (!is.null(nativeOutput$csvDen))
                          array(nativeOutput$csvDen, c(n.observed, n.xvar)) else NULL)
            classOutput[[target.idx]] <- c(classOutput[[target.idx]], csv.den = list(csv.den))
            remove(csv.den)
          }
        }
        nativeOutput$allEnsbCLS <- NULL
        nativeOutput$oobEnsbCLS <- NULL
        nativeOutput$perfClas <- NULL
        nativeOutput$vimpClas <- NULL
        nativeOutput$blockClas <- NULL
        ## When TRUE we revert to univariate nomenclature for all the outputs.
        if(univariate.nomenclature) {
          if ((class.count == 1) & (regr.count == 0)) {
            names(classOutput) <- NULL
            rfsrcOutput <- c(rfsrcOutput, unlist(classOutput, recursive=FALSE))
          }
            else {
              rfsrcOutput <- c(rfsrcOutput, classOutput = list(classOutput))
            }
        }
          else {
            rfsrcOutput <- c(rfsrcOutput, classOutput = list(classOutput))
          }
      }
      if (regr.count > 0) {
        regrOutput <- vector("list", regr.count)
        names(regrOutput) <- yvar.names[regr.index]
        ## Incoming: T=tree R=response
        ## T1R1 T1R2, T2R1 T2R2, T3R1 T3R2, ...
        ## Yeilds tree.offset = c(1, 3, 5)
        tree.offset <- array(1, ntree)
        if (ntree > 1) {
          tree.offset[2:ntree] <- length(regr.index)
        }
        tree.offset <-  cumsum(tree.offset)
        ## The block offset calculation mirrors the tree offset calculation, but differs in only the primary dimension.
        block.offset <- array(1, floor(ntree/block.size))
        if (floor(ntree/block.size) > 1) {
          block.offset[2:floor(ntree/block.size)] <- length(regr.index)
        }
        block.offset <-  cumsum(block.offset)
        ## Incoming vimp rates: V=xvar R=response L=level
        ## V1R1 V1R2, V2R1 V2R2, V3R1 V3R2, ...
        ## Yeilds vimp.offset = c(1, 3, 5, ...)
        vimp.offset <- array(1, vimp.count)
        if (vimp.count > 1) {
          vimp.offset[2:vimp.count] <- length(regr.index)
        }
        vimp.offset <-  cumsum(vimp.offset)
        iter.ensb.start <- 0
        iter.ensb.end   <- 0
        iter.qntl.start <- 0
        iter.qntl.end   <- 0
        ## From the native code:
        ##   "allEnsbRGR"
        ## -> of dim [regr.count] x [obsSize]
        ## From the native code:
        ##   "allEnsbQNT"
        ##   "oobEnsbQNT"
        ## -> of dim [regr.count] x [length(prob)] x [n]
        ## From the native code:
        ##   "perfRegr"
        ## -> of dim [regr.count] x [ntree]
        ## To the R code:
        ## -> of dim  [[regr.count]] x [ntree]
        ## From the native code:
        ##   "blockRegr"
        ## -> of dim [block.cnt] x [regr.count]
        ## To the R code:
        ## -> of dim [[regr.count]] x [block.cnt]
        ## From the native code:
        ##   "vimpRegr"
        ## -> of dim [n.vxar] x [regr.count]
        ## To the R code:
        ## -> of dim  [[regr.count]] x [n.xvar]
        ## From the native code:
        ##   "cseNum"
        ## -> of dim [regr.count] x [n]
        ## To the R code:
        ## -> of dim [[regr.count]] x [n]
        ## From the native code:
        ##   "cseDen"
        ## -> of dim [n]
        ## To the R code:
        ## -> of dim [n]
        ## From the native code:
        ##   "csvNum"
        ## -> of dim [n.xvar] x [regr.count] x [n]
        ## To the R code:
        ## -> of dim [[regr.count]] x [n] x [n.xvar]
        ## From the native code:
        ##   "csvDen"
        ## -> of dim [n.xvar] x [n] 
        ## To the R code:
        ## -> of dim  x [n]  x [n.xvar]
        for (i in 1:length(m.target.idx)) {
          target.idx <- which(regr.index == m.target.idx[i])
          if (length(target.idx) > 0) {
            iter.ensb.start <- iter.ensb.end
            iter.ensb.end <- iter.ensb.end + n.observed
            iter.qntl.start <- iter.qntl.end
            iter.qntl.end <- iter.qntl.end + (length(prob) * n.observed)
            vimp.names <- if(vimp.joint) "joint" else importance.xvar
            predicted <- (if (!is.null(nativeOutput$allEnsbRGR))
                            array(nativeOutput$allEnsbRGR[(iter.ensb.start + 1):iter.ensb.end], n.observed) else NULL)
            regrOutput[[target.idx]] <- list(predicted = predicted)
            remove(predicted)
            predicted.oob <- (if (!is.null(nativeOutput$oobEnsbRGR))
                                array(nativeOutput$oobEnsbRGR[(iter.ensb.start + 1):iter.ensb.end], n.observed) else NULL)
            regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], predicted.oob = list(predicted.oob))
            remove(predicted.oob)
            ## New case specific arrays, numerator:
            cse.num <- (if (!is.null(nativeOutput$cseRegr))
                            array(nativeOutput$cseRegr[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
            regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], cse.num = list(cse.num))
            remove(cse.num)
            ## New case specific arrays, denominator:
            cse.den <- (if (!is.null(nativeOutput$cseDen))
                            array(nativeOutput$cseDen[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
            regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], cse.den = list(cse.den))
            remove(cse.den)
            quantile <- (if (!is.null(nativeOutput$allEnsbQNT))
                             array(nativeOutput$allEnsbQNT[(iter.qntl.start + 1):iter.qntl.end],
                                   c(n.observed, length(prob))) else NULL)
            regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], quantile = list(quantile))
            remove(quantile)
            quantile.oob <- (if (!is.null(nativeOutput$oobEnsbQNT))
                                 array(nativeOutput$oobEnsbQNT[(iter.qntl.start + 1):iter.qntl.end],
                                       c(n.observed, length(prob))) else NULL)
            regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], quantile.oob = list(quantile.oob))
            remove(quantile.oob)
            if (!is.null(nativeOutput$perfRegr)) {
              err.rate <- nativeOutput$perfRegr[tree.offset]
              tree.offset <- tree.offset + 1
              regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], err.rate = list(err.rate))
              remove(err.rate)
            }
            if (!is.null(nativeOutput$blockRegr)) {
              err.block.rate <- nativeOutput$blockRegr[block.offset]
              block.offset <- block.offset + 1
              regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], err.block.rate = list(err.block.rate))
              remove(err.block.rate)
            }
            if (!is.null(nativeOutput$vimpRegr)) {
              importance <- nativeOutput$vimpRegr[vimp.offset]
              names(importance) <- vimp.names
              vimp.offset <- vimp.offset + 1
              regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], importance = list(importance))
              remove(importance)
            }
            ## See GROW call for explanation of arrays:
            csv.idx  <- array(0, n.observed * n.xvar)
            j2seq <- (target.idx - 1) * n.observed
            for (j in 1:n.xvar) {
              j1seq <- (j - 1) * n.observed
              csv.idx[(j1seq + 1) : (j1seq + n.observed)]  <- ( (j2seq + j1seq * regr.count + 1) : (j2seq + j1seq * regr.count + n.observed) ) 
            }
            ## New case specific arrays, numerator:
            csv.num <- (if (!is.null(nativeOutput$csvRegr))
                          array(nativeOutput$csvRegr[csv.idx], c(n.observed, n.xvar)) else NULL)
            regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], csv.num = list(csv.num))
            remove(csv.num)
            ## New case specific arrays, denominator:
            csv.den <- (if (!is.null(nativeOutput$csvDen))
                          array(nativeOutput$csvDen, c(n.observed, n.xvar)) else NULL)
            regrOutput[[target.idx]] <- c(regrOutput[[target.idx]], csv.den = list(csv.den))
            remove(csv.den)
          }
        }
        nativeOutput$allEnsbRGR <- NULL
        nativeOutput$oobEnsbRGR <- NULL
        nativeOutput$allEnsbQNT <- NULL
        nativeOutput$oobEnsbQNT <- NULL
        nativeOutput$perfRegr <- NULL
        nativeOutput$vimpRegr <- NULL
        nativeOutput$blockRegr <- NULL
        ## When TRUE we revert to univariate nomenclature for all the outputs.
        if(univariate.nomenclature) {
          if ((class.count == 0) & (regr.count == 1)) {
            names(regrOutput) <- NULL
            rfsrcOutput <- c(rfsrcOutput, unlist(regrOutput, recursive=FALSE))
          }
            else {
              rfsrcOutput <- c(rfsrcOutput, regrOutput = list(regrOutput))
            }
        }
          else {
            rfsrcOutput <- c(rfsrcOutput, regrOutput = list(regrOutput))
          }
      }
    }
  class(rfsrcOutput) <- c("rfsrc", "predict",   family)
  return(rfsrcOutput)
}

## --------------------------------------------------------------------------------------------
##
## extraction function when partial=FALSE
##
## --------------------------------------------------------------------------------------------
extract.pred <- function(obj, type, subset, time, m.target, target, oob = oob) {
  ## coerce the (potentially) multivariate object if necessary.
  obj <- coerce.multivariate(obj, m.target)
  ## extract the sample size, assign the default subset
  if (missing(subset)) {
    subset <- 1:obj$n
  }
  n <- length(subset)
  ## decide if OOB or in-bag values are requested
  if (oob == FALSE) {
    if (is.null(obj$predicted)) {
      stop("inbag requested but inbag values are not available in the object")
    }
    pred <- obj$predicted
    surv <- obj$survival
    chf <- obj$chf
    cif <- obj$cif
  }
  else {
    if (is.null(obj$predicted.oob)) {
      stop("oob requested but oob values are not available in the object")
    }
    pred <- obj$predicted.oob
    surv <- obj$survival.oob
    chf <- obj$chf.oob
    cif <- obj$cif.oob
  }
  ## Survival and competing risk families.
  if (grepl("surv", obj$family)) {
    ## Competing risks:
    if (obj$family == "surv-CR") {
      ## Default is first event type.
      if (missing(target)) target <- 1
      type <- match.arg(type, c("years.lost", "cif", "chf"))
      ## Get the index of the closest point of the grow time interest vector.
      time.idx <-  max(which(obj$time.interest <= time))
      return(switch(type,
                    "years.lost" = pred[subset, target],
                    "cif"        = cif[subset, time.idx, target],
                    "chf"        = chf[subset, time.idx, target]
                    ))
    }
    ## Right-censored:
    else {
      type <- match.arg(type, c("rel.freq", "mort", "chf", "surv"))
      ## Get the index of the closest point of the grow time interest vector.
      time.idx <-  max(which(obj$time.interest <= time))
      return(switch(type,
                    "rel.freq" = pred[subset]/max(n, na.omit(pred)),
                    "mort"     = pred[subset],
                    "chf"      = 100 * chf[subset, time.idx],
                    "surv"     = 100 * surv[subset, time.idx]
                    ))
    }
  }
  ## The object has been coerced.  It will be univariate.
  else {
    if (obj$family == "class") {
      type <- match.arg(type, c("bayes", "rfq", "prob"))
      ## The default is first class label.
      if (missing(target)) target <- 1
      ## extract the subsetted values
      prob <- pred[subset,, drop = FALSE]
      y <- obj$yvar[subset]
      ## rfq only permitted in two class problems
      if (type == "rfq" && nlevels(y) != 2) {
        type <- "bayes"
      }
      ## return the requested performance
      return(switch(type,
                    "bayes" =  get.bayes.rule(prob),
                    "rfq"   =  get.gmean.rule(y, prob),
                    "prob"  =  prob[, target]))
    }
    else {
      return(pred[subset])
    }
  }
}
## --------------------------------------------------------------------------------------------
##
## extraction function when partial=TRUE
## Note that currently only one (1) time point is requested via the
## plot.variable() R-wrapper.  Only one (1) is allowed.
## Additional time points can be specified
## using the partial.rfsrc() wrapper, and then extracted without any
## additional calls to the native code.  The extract call
## is limited to only one (1) time point on each call.
## The same protocol holds for m.target.
##
## --------------------------------------------------------------------------------------------
extract.partial.pred <- function(obj, type, subset, m.target, target) {
  ## Survival families
  if (grepl("surv", obj$family)) {
    ## extract the sample size, assign the default subset
    if (missing(subset)) {
      subset <- 1:obj$n
    }
    n <- length(subset)
    ## array dimension
    time.idx <-  1
    ## Competing risks:
    if (obj$family == "surv-CR") {
      ## Default is first event type.
      if (missing(target)) target <- 1
      type <- match.arg(type, c("years.lost", "cif", "chf"))
      return(switch(type,
                    "years.lost" = obj$survOutput[subset, target, ],
                    "cif" = obj$survOutput[subset, time.idx, target, ],
                    "chf" = obj$survOutput[subset, time.idx, target, ]
                    ))
    }
    ## Right censored:
      else {
        if (type == "rel.freq") {
          sz <- apply(obj$survOutput[subset, ], 2, function(x) {length(na.omit(x))})
          rs <- t(apply(obj$survOutput[subset, ], 1, function(x) {x / sz}))
        }
        return(switch(type,
                      "rel.freq" = rs,
                      "mort"     = obj$survOutput[subset, ],
                      "chf"      =  obj$survOutput[subset, time.idx, ],
                      "surv"     =  obj$survOutput[subset, time.idx, ]
                      ))
      }
  }
    else {
      ## Univariate and multivariate families:
      ## The incoming partial object is always presented in a mulitvariate list, regardless of whether
      ## the grow object was univariate or multivariate.  There is no coersion.  In a multivariate case,
      ## m.target is coherent.  But in a univariate case, it will be NULL.  As a result, we
      ## accomodate for the NULL case by grabbing the first (and actually only) valid output below, in
      ## order to parse the multivariate list correctly.
      ## We are either classification or regression but we don't know which yet.
      regrClassTarget <- NULL
      ## It doesn't matter whether we process regression or classification first.
      ## But one and only one family will be processed.
      if (is.null(regrClassTarget)) {
        if (!is.null(obj$classOutput)) {
          if (is.null(m.target)) {
            ## Grab the first outcome!  It will exist, as this is the univariate case.
            m.target <- names(obj$classOutput)[1]
          }
          regrClassTarget <- which(names(obj$classOutput) == m.target)
          ## Classification.
          if (length(regrClassTarget) > 0) {
            if (length(regrClassTarget) > 1) {
              ## Safety check in case m.target was incorrect coming into this routine.
              stop("Invalid number of target outcomes specified in partial plot extraction.")
            }
            type <- match.arg(type, c("prob", "bayes"))
            ## Default is first class label.
            if (missing(target)) target <- 1
            n <- dim(obj$classOutput[[regrClassTarget]])[1]
            if (missing(subset)) subset <- 1:n
            return(switch(type,
                          "prob" = obj$classOutput[[regrClassTarget]][subset, 1 + target, ],
                          "bayes" =  obj$classOutput[[regrClassTarget]][subset, 1, ]))
          }
        }
      }
      ## It doesn't matter whether we process regression or classification first.
      ## But one and only one family will be processed.
      if (is.null(regrClassTarget)) {
        if (!is.null(obj$regrOutput)) {
          if (is.null(m.target)) {
            ## Grab the first outcome!  It will exist, as this is the univariate case.
            m.target <- names(obj$regrOutput)[1]
          }
          regrClassTarget <- which (names(obj$regrOutput) == m.target)
          ## Regression.
          if (length(regrClassTarget) > 0) {
            if (length(regrClassTarget) > 1) {            
              ## Safety check in case m.target was incorrect coming into this routine.
              stop("Invalid number of target outcomes specified in partial plot extraction.")
            }
            n <- dim(obj$classOutput[[regrClassTarget]])[1]
            if (missing(subset)) subset <- 1:n
            return(obj$regrOutput[[regrClassTarget]][subset, ])
          }
        }
      }
      ## If after parsing both classification and regression we have not found the target, we error.
      if (is.null(regrClassTarget)) {
        stop("Invalid target specified in partial plot extraction:  ", m.target)
      }
    }
}

## o must be a list of forests!
fast.save.list <- function(o, path=NULL, testing=FALSE, units="Mb") {
  if (is.null(path)) {
    path <- paste0(getwd(), "/forest")
  }
  unlink(path, recursive=TRUE)
  dir.create(path, showWarnings=FALSE, recursive=TRUE)
  lapply(1:length(o), function(j) {
    fast.save(o[[j]], paste0(path, "/forest", j), testing=testing, units=units)
  })
}
## o must be a list of forests!
fast.load.list <- function(directory, path = NULL, testing = FALSE, units="Mb") {
  if (is.null(path)) {
    path <- paste0(getwd(), "/", directory)
  }
  else {
    path <- paste0(path, "/", directory)
  }
  files <- list.files(path)
  lapply(1:length(files), function(j) {
    fast.load(files[j], path, testing=testing, units=units)
  })
}
fast.save <- function(o, path=NULL, testing=TRUE, units="Mb") {
  if (is.null(path)) {
    path <- paste0(getwd(), "/forest")
  }
  unlink(path, recursive=TRUE)
  dir.create(path, showWarnings=FALSE, recursive=TRUE)
  ## coherence check
  if (sum(inherits(o, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2)
    stop("this function only works for objects of class `(rfsrc, grow)'")
  ## extract the forest - hereafter this is what we work with
  o <- o$forest
  ## nativeArrayTNDS
  check <- !sapply(o$nativeArrayTNDS, is.null)
  if (sum(check) > 0) {
    lapply(names(o$nativeArrayTNDS)[check], function(nm) {
      fst::write_fst(data.table::data.table(o$nativeArrayTNDS[[nm]]), paste0(path, "/nativeArrayTDNS_", nm, ".rda"))
    })
    o$nativeArrayTNDS[names(o$nativeArrayTNDS)[check]] <- NULL
  }
  ## nativeArray
  fst::write_fst(data.table::data.table(o$nativeArray), paste0(path, "/nativeArray.rda"))
  o$nativeArray <- NULL
  ## x and y
  if (!is.null(o$xvar)) {
    fst::write_fst(data.table::data.table(o$xvar), paste0(path, "/xvar.rda"))
    o$xvar <- NULL
  }
  if (!is.null(o$yvar)) {
    fst::write_fst(data.frame(o$yvar), paste0(path, "/yvar.rda"))
    o$yvar <- NULL
  }
  ## misc.
  fst::write_fst(data.frame(o$case.wt), paste0(path, "/case.wt.rda"))
  fst::write_fst(data.frame(o$leafCount), paste0(path, "/leafCount.rda"))
  fst::write_fst(data.frame(o$seed), paste0(path, "/seed.rda"))
  o$case.wt <- o$yvar <- o$leafCount <- o$seed <- NULL
  ## convert sample size to a number
  o$sampfrac <- o$sampsize(1)
  o$sampsize <- NULL
  ## survival
  if (o$family == "surv" || o$family == "surv-CR") {
    check <- !sapply(o$event.info, is.null)
    lapply(names(o$event.info)[check], function(nm) {
      fst::write_fst(data.table::data.table(o$event.info[[nm]]), paste0(path, "/event.info_", nm, ".rda"))
    })
    o$event.info[names(o$event.info[check])] <- NULL
  }
  ## output size (used for testing)
  if (testing) {
    print(lsos(o, units=units, n=length(o)))
  }
  ## forest
  saveRDS(o, file=paste0(path, "/forest.rda"), compress=FALSE)
  gc(FALSE)
  #cat("finished\n")
}
fast.load <- function(directory, path = NULL, testing = FALSE, units="Mb") {
  if (is.null(path)) {
    path <- paste0(getwd(), "/", directory)
  }
  else {
    path <- paste0(path, "/", directory)
  }
  files <- list.files(path)
  ## forest
  o <- readRDS(paste0(path, "/forest.rda"))
  gc()
  ## nativeArrayTNDS
  if (any(grepl("nativeArrayTDNS", files))) {
    target <- files[grepl("nativeArrayTDNS", files)]
    target <- gsub(".rda", "", target)
    target <- gsub("nativeArrayTDNS_", "", target)
    lapply(target, function(nm) {
      o$nativeArrayTNDS[[nm]] <<- fst::read_fst(paste0(path, "/nativeArrayTDNS_", nm, ".rda"))[[1]]
    })
  }
  ## nativeArrayTNDS
  o$nativeArray <- fst::read_fst(paste0(path, "/nativeArray.rda"))
  ## x and y
  if (any(grepl("xvar.rda", files))) {
    o$xvar <- fst::read_fst(paste0(path, "/xvar.rda"))
  }
  if (any(grepl("yvar.rda", files))) {
    o$yvar <- fst::read_fst(paste0(path, "/yvar.rda"))
    if (!is.data.frame(o$yvar)) {
      o$yvar <- o$yvar[[1]]
    }
  }
  ## misc.
  o$case.wt <- fst::read_fst(paste0(path, "/case.wt.rda"))[[1]]
  o$leafCount <- fst::read_fst(paste0(path, "/leafCount.rda"))[[1]]
  o$seed <- fst::read_fst(paste0(path, "/seed.rda"))[[1]]
  ## sample size --- hard coded
  o$sampsize <- function(x){x * o$sampfrac} 
  ## survival
  if (any(grepl("event.info", files))) {
    target <- files[grepl("event.info", files)]
    target <- gsub(".rda", "", target)
    target <- gsub("event.info_", "", target)
    lapply(target, function(nm) {
      o$event.info[[nm]] <<- fst::read_fst(paste0(path, "/event.info_", nm, ".rda"))[[1]]
    })
  }
  ## output size (used for testing)
  if (testing) {
    print(lsos(o, units=units, n=length(o)))
  }
  ##return the object
  o
}
## improved list of objects
.ls.objects <- function (pos = 1, pattern, order.by,
                         decreasing=FALSE, head=FALSE, n=5,
                         units="Gb") {
    napply <- function(names, fn) sapply(names, function(x)
                                         fn(get(x, pos = pos)))
    names <- ls(pos = pos, pattern = pattern)
    obj.class <- napply(names, function(x) as.character(class(x))[1])
    obj.mode <- napply(names, mode)
    obj.type <- ifelse(is.na(obj.class), obj.mode, obj.class)
    obj.prettysize <- napply(names, function(x) {
                           format(object.size(x), units = units)})
    obj.size <- napply(names, object.size)
    obj.dim <- t(napply(names, function(x)
                        as.numeric(dim(x))[1:2]))
    vec <- is.na(obj.dim)[, 1] & (obj.type != "function")
    obj.dim[vec, 1] <- napply(names, length)[vec]
    out <- data.frame(obj.type, obj.size, obj.prettysize, obj.dim)
    names(out) <- c("Type", "Size", "PrettySize", "Rows", "Columns")
    if (!missing(order.by))
        out <- out[order(out[[order.by]], decreasing=decreasing), ]
    if (head)
        out <- head(out, n)
    out
}
## shorthand for improved list of objects
lsos <- function(..., n=10) {
    .ls.objects(..., order.by="Size", decreasing=TRUE, head=TRUE, n=n)
}

`%||%` <- function(x, y) if (is.null(x)) y else x
.check.fst <- function() {
  if (!requireNamespace("fst", quietly = TRUE)) {
    stop("Package 'fst' is required for save/load support.", call. = FALSE)
  }
  invisible(TRUE)
}
.timestamp <- function() {
  format(Sys.time(), "%Y-%m-%d %H:%M:%S")
}
.msg <- function(..., verbose = TRUE) {
  if (isTRUE(verbose)) {
    cat(sprintf("[%s] ", .timestamp()), ..., "\n", sep = "")
    flush.console()
  }
}
.safe.dir.create <- function(path) {
  if (!dir.exists(path)) {
    dir.create(path, recursive = TRUE, showWarnings = FALSE)
  }
  invisible(path)
}
.safe.unlink.dir <- function(path) {
  if (dir.exists(path)) {
    unlink(path, recursive = TRUE, force = TRUE)
  }
  invisible(path)
}
.safe.name <- function(x) {
  x <- gsub("[^[:alnum:].]+", ".", x)
  x <- gsub("\\.+", ".", x)
  x <- gsub("^\\.+|\\.+$", "", x)
  if (!nzchar(x)) x <- "var"
  x
}
.is.real.valued <- function(x) {
  (is.double(x) || is.integer(x)) &&
    !inherits(x, c("Date", "POSIXt", "difftime"))
}
.coerce.supported.column <- function(x, nm) {
  if (is.factor(x) || .is.real.valued(x)) {
    return(x)
  }
  out <- tryCatch(
    factor(x),
    error = function(e) e
  )
  if (inherits(out, "error")) {
    stop("Column `", nm, "` is neither real-valued nor factor and could not be coerced to factor: ",
         conditionMessage(out),
         call. = FALSE)
  }
  out
}
.normalize.training.data <- function(data) {
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  for (nm in names(data)) {
    data[[nm]] <- .coerce.supported.column(data[[nm]], nm)
  }
  data
}
.drop.all.na.train <- function(data) {
  which.na <- is.na(data)
  keep.rows <- rowSums(which.na) < ncol(data)
  keep.cols <- colSums(which.na) < nrow(data)
  list(
    data = data[keep.rows, keep.cols, drop = FALSE],
    keep.rows = keep.rows,
    keep.cols = keep.cols
  )
}
.build.schema <- function(data) {
  out <- lapply(names(data), function(nm) {
    x <- data[[nm]]
    integer.info <- .integer.support.info(x)
    list(
      class = class(x),
      is.factor = is.factor(x),
      ordered = is.ordered(x),
      levels = if (is.factor(x)) levels(x) else NULL,
      is.integer = is.integer(x),
      is.numeric = .is.real.valued(x),
      integer.support = isTRUE(integer.info$restore),
      integer.storage = isTRUE(integer.info$storage.integer),
      integer.frac = integer.info$integer.frac
    )
  })
  names(out) <- names(data)
  out
}
.normalize.schema.integer.support <- function(schema) {
  if (is.null(schema) || length(schema) == 0L) {
    return(schema)
  }
  for (nm in names(schema)) {
    sc <- schema[[nm]]
    if (!is.list(sc)) next
    if (is.null(sc$integer.storage)) {
      sc$integer.storage <- isTRUE(sc$is.integer)
    }
    if (is.null(sc$integer.support)) {
      sc$integer.support <- isTRUE(sc$is.integer)
    }
    if (is.null(sc$integer.frac)) {
      sc$integer.frac <- if (isTRUE(sc$is.integer)) 1 else NA_real_
    }
    schema[[nm]] <- sc
  }
  schema
}
.schema.restores.integer <- function(sc) {
  is.list(sc) && (isTRUE(sc$integer.support) || isTRUE(sc$is.integer))
}
.schema.integer.info <- function(sc) {
  list(
    restore = .schema.restores.integer(sc),
    storage.integer = isTRUE(sc$is.integer) || isTRUE(sc$integer.storage),
    integer.frac = sc$integer.frac %||% (if (isTRUE(sc$is.integer)) 1 else NA_real_)
  )
}
.as.numeric.safe <- function(x) {
  if (is.factor(x)) {
    x <- as.character(x)
  }
  suppressWarnings(as.numeric(x))
}
.as.numeric.from.schema <- function(x, schema = NULL) {
  .as.numeric.safe(x)
}
.mode.value <- function(x) {
  x <- x[!is.na(x)]
  if (length(x) == 0L) return(NA)
  tb <- sort(table(x), decreasing = TRUE)
  names(tb)[1]
}
.numeric.init <- function(x) {
  x <- .as.numeric.safe(x)
  x <- x[is.finite(x)]
  if (length(x) == 0L) return(NA_real_)
  mean(x)
}
.compute.init <- function(data, schema) {
  out <- lapply(names(data), function(nm) {
    x <- data[[nm]]
    sc <- schema[[nm]]
    if (isTRUE(sc$is.factor)) {
      .mode.value(as.character(x))
    }
    else {
      .numeric.init(.as.numeric.from.schema(x, sc))
    }
  })
  names(out) <- names(data)
  out
}
.compute.scale <- function(data, schema) {
  out <- lapply(names(data), function(nm) {
    sc <- schema[[nm]]
    if (isTRUE(sc$is.factor)) {
      1
    }
    else {
      x <- .as.numeric.from.schema(data[[nm]], sc)
      s <- stats::sd(x, na.rm = TRUE)
      if (!is.finite(s) || s <= 0) s <- 1
      s
    }
  })
  names(out) <- names(data)
  out
}
.resolve.targets <- function(which.na, target.mode = c("missing.only", "all")) {
  target.mode <- match.arg(target.mode)
  miss.count <- colSums(which.na)
  if (target.mode == "missing.only") {
    names(miss.count)[miss.count > 0]
  }
  else {
    colnames(which.na)
  }
}
.resolve.predictor.map <- function(targets, all.names, deployment.xvars = NULL) {
  if (is.null(deployment.xvars)) {
    out <- lapply(targets, function(y) setdiff(all.names, y))
    names(out) <- targets
    return(out)
  }
  if (is.character(deployment.xvars)) {
    unknown <- setdiff(deployment.xvars, all.names)
    if (length(unknown) > 0L) {
      warning("Ignoring deployment.xvars predictors not found in training data: ",
              paste(unknown, collapse = ", "),
              call. = FALSE)
    }
    xvars <- intersect(all.names, deployment.xvars)
    out <- lapply(targets, function(y) setdiff(xvars, y))
    names(out) <- targets
    return(out)
  }
  if (is.list(deployment.xvars)) {
    nms <- names(deployment.xvars)
    if (is.null(nms) || length(nms) != length(deployment.xvars) ||
        any(is.na(nms) | !nzchar(nms))) {
      stop("'deployment.xvars' supplied as a list must be named by target variable.",
           call. = FALSE)
    }
    dup <- unique(nms[duplicated(nms)])
    if (length(dup) > 0L) {
      stop("'deployment.xvars' contains duplicated target names: ",
           paste(dup, collapse = ", "),
           call. = FALSE)
    }
    extra.targets <- setdiff(nms, targets)
    if (length(extra.targets) > 0L) {
      warning("Ignoring deployment.xvars entries for non-target variables: ",
              paste(extra.targets, collapse = ", "),
              call. = FALSE)
    }
    bad.entries <- names(deployment.xvars)[vapply(deployment.xvars, function(x) {
      !(is.null(x) || is.character(x))
    }, logical(1))]
    if (length(bad.entries) > 0L) {
      stop("Each 'deployment.xvars' list entry must be NULL or a character vector. ",
           "Problem entries: ", paste(bad.entries, collapse = ", "),
           call. = FALSE)
    }
    listed.predictors <- unique(unlist(deployment.xvars[intersect(nms, targets)],
                                       use.names = FALSE))
    unknown <- setdiff(listed.predictors, all.names)
    if (length(unknown) > 0L) {
      warning("Ignoring deployment.xvars predictors not found in training data: ",
              paste(unknown, collapse = ", "),
              call. = FALSE)
    }
    out <- lapply(targets, function(y) {
      xvars <- deployment.xvars[[y]]
      if (is.null(xvars)) {
        xvars <- setdiff(all.names, y)
      }
      setdiff(intersect(all.names, xvars), y)
    })
    names(out) <- targets
    return(out)
  }
  stop("'deployment.xvars' must be NULL, a character vector, or a named list.",
       call. = FALSE)
}
.normalize.impute.learn.manifest <- function(manifest) {
  if (is.null(manifest$spec.version)) {
    manifest$spec.version <- 1L
  }
  if (is.null(manifest$supervised) || !is.list(manifest$supervised)) {
    manifest$supervised <- list(enabled = FALSE)
  }
  if (is.null(manifest$supervised$enabled)) {
    manifest$supervised$enabled <- FALSE
  }
  manifest$schema <- .normalize.schema.integer.support(manifest$schema)
  if (is.null(manifest$supervised$response.schema)) {
    manifest$supervised$response.schema <- NULL
  }
  else {
    manifest$supervised$response.schema <- .normalize.schema.integer.support(
      manifest$supervised$response.schema
    )
  }
  if (is.null(manifest$predictor.map.raw)) {
    manifest$predictor.map.raw <- manifest$predictor.map
  }
  manifest
}
.is.supervised.manifest <- function(manifest) {
  is.list(manifest$supervised) && isTRUE(manifest$supervised$enabled)
}
.get.rfsrc.internal <- function(name) {
  fn <- get0(name, mode = "function")
  if (!is.null(fn)) {
    return(fn)
  }
  ns <- tryCatch(
    asNamespace("randomForestSRC"),
    error = function(e) NULL
  )
  if (is.null(ns) || !exists(name, envir = ns, inherits = FALSE)) {
    return(NULL)
  }
  get(name, envir = ns, inherits = FALSE)
}
.parse.formula.terms <- function(formula, data, arg.name = "formula") {
  tt <- tryCatch(
    stats::terms(formula, data = data, keep.order = TRUE),
    error = function(e) e
  )
  if (inherits(tt, "error")) {
    stop("Failed to parse '", arg.name, "': ", conditionMessage(tt),
         call. = FALSE)
  }
  tt
}
.make.supervised.formula.env <- function(parent = environment()) {
  if (is.null(parent)) {
    parent <- baseenv()
  }
  env <- new.env(parent = parent)
  surv.fun <- get0("Surv", envir = parent, mode = "function", inherits = TRUE)
  if (is.null(surv.fun)) {
    surv.fun <- tryCatch(
      get("Surv", envir = asNamespace("survival"), inherits = FALSE),
      error = function(e) NULL
    )
  }
  if (!is.null(surv.fun)) {
    env$Surv <- surv.fun
  }
  multivar.fun <- .get.rfsrc.internal("Multivar")
  if (!is.null(multivar.fun)) {
    env$Multivar <- multivar.fun
  }
  env$cbind <- base::cbind
  env
}
.project.supervised.formula <- function(supervised.formula, data, train.names,
                                        arg.name = "supervised.formula") {
  tt <- .parse.formula.terms(supervised.formula, data = data, arg.name = arg.name)
  xvar.names <- attr(tt, "term.labels") %||% character(0)
  if (length(xvar.names) == 0L) {
    stop("'", arg.name, "' must include at least one predictor.", call. = FALSE)
  }
  if (!all(xvar.names %in% names(data))) {
    bad <- setdiff(xvar.names, names(data))
    stop(
      "'", arg.name, "' must reference raw predictor columns in 'data'. ",
      "Derived terms, interactions, and transformations are not currently supported. ",
      "Problem terms: ", paste(bad, collapse = ", "),
      call. = FALSE
    )
  }
  xvar.names <- intersect(xvar.names, train.names)
  if (length(xvar.names) == 0L) {
    stop("No retained predictor columns remain for '", arg.name,
         "' after training preprocessing.", call. = FALSE)
  }
  lhs <- paste(deparse(supervised.formula[[2L]], width.cutoff = 500L), collapse = "")
  stats::reformulate(
    xvar.names,
    response = lhs,
    intercept = attr(tt, "intercept") %||% 1L,
    env = .make.supervised.formula.env(environment(supervised.formula))
  )
}
.parse.supervised.spec <- function(supervised.formula, data) {
  if (is.null(supervised.formula)) {
    return(list(enabled = FALSE))
  }
  if (!inherits(supervised.formula, "formula")) {
    stop("'supervised.formula' must be a formula.", call. = FALSE)
  }
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  tt <- .parse.formula.terms(supervised.formula, data = data,
                             arg.name = "supervised.formula")
  response.pos <- attr(tt, "response") %||% 0L
  if (response.pos != 1L) {
    stop("'supervised.formula' must include a response on the left-hand side.",
         call. = FALSE)
  }
  xvar.names <- attr(tt, "term.labels") %||% character(0)
  if (length(xvar.names) == 0L) {
    stop("'supervised.formula' must include at least one predictor.",
         call. = FALSE)
  }
  if (!all(xvar.names %in% names(data))) {
    bad <- setdiff(xvar.names, names(data))
    stop(
      "'supervised.formula' must reference raw predictor columns in 'data'. ",
      "Derived terms, interactions, and transformations are not currently supported. ",
      "Problem terms: ", paste(bad, collapse = ", "),
      call. = FALSE
    )
  }
  parser <- .get.rfsrc.internal("parseFormula")
  parsed <- NULL
  if (!is.null(parser)) {
    parsed <- tryCatch(
      parser(supervised.formula, data = data),
      error = function(e) e
    )
    if (inherits(parsed, "error")) {
      parsed <- NULL
    }
  }
  if (is.null(parsed)) {
    yvar.names <- all.vars(supervised.formula[[2L]])
    subj.names <- character(0)
    family.guess <- NA_character_
  }
  else {
    yvar.names <- parsed$yvar.names %||% character(0)
    subj.names <- parsed$subj.names %||% character(0)
    family.guess <- parsed$family %||% NA_character_
  }
  response.names <- unique(c(subj.names, yvar.names))
  if (length(response.names) == 0L) {
    stop("Unable to resolve supervised response variables from 'supervised.formula'.",
         call. = FALSE)
  }
  if (!all(response.names %in% names(data))) {
    bad <- setdiff(response.names, names(data))
    stop("The supervised response variables were not found in 'data': ",
         paste(bad, collapse = ", "), call. = FALSE)
  }
  response.predictor.overlap <- intersect(response.names, xvar.names)
  if (length(response.predictor.overlap) > 0L) {
    stop("'supervised.formula' must not include supervised response variables ",
         "on the right-hand side. Problem variables: ",
         paste(response.predictor.overlap, collapse = ", "),
         call. = FALSE)
  }
  y.data <- data[, response.names, drop = FALSE]
  for (nm in names(y.data)) {
    y.data[[nm]] <- .coerce.supported.column(y.data[[nm]], nm)
  }
  list(
    enabled = TRUE,
    formula = supervised.formula,
    family.guess = family.guess,
    xvar.names = xvar.names,
    yvar.names = yvar.names,
    subj.names = subj.names,
    response.names = response.names,
    x.data = data[, xvar.names, drop = FALSE],
    y.data = y.data
  )
}
.supervised.usable.rows <- function(ydata) {
  out <- tryCatch(
    stats::complete.cases(ydata),
    error = function(e) e
  )
  if (inherits(out, "error")) {
    ydf <- tryCatch(
      as.data.frame(ydata, stringsAsFactors = FALSE),
      error = function(e) e
    )
    if (inherits(ydf, "error")) {
      stop("Failed to determine which supervised responses are observed: ",
           conditionMessage(out), call. = FALSE)
    }
    out <- stats::complete.cases(ydf)
  }
  out
}
.parse.supervised.args <- function(supervised.args = list()) {
  if (is.null(supervised.args)) {
    supervised.args <- list()
  }
  if (!is.list(supervised.args)) {
    stop("'supervised.args' must be a list.", call. = FALSE)
  }
  if (length(supervised.args) > 0L) {
    arg.names <- names(supervised.args)
    if (is.null(arg.names) || any(is.na(arg.names) | !nzchar(arg.names))) {
      stop("'supervised.args' must be a named list.", call. = FALSE)
    }
    if (anyDuplicated(arg.names)) {
      warning("Duplicate supervised.args entries were supplied; keeping the last occurrence for each name.",
              call. = FALSE)
      supervised.args <- supervised.args[!duplicated(arg.names, fromLast = TRUE)]
    }
  }
  blocked <- intersect(names(supervised.args), c("formula", "data", "forest"))
  if (length(blocked) > 0L) {
    warning("Ignoring supervised.args entries controlled internally: ",
            paste(blocked, collapse = ", "),
            call. = FALSE)
    supervised.args[blocked] <- NULL
  }
  supervised.args
}
.resolve.supervised.fit.meta <- function(supervised.fit, fallback = NULL) {
  if (is.null(supervised.fit) || !is.list(supervised.fit)) {
    stop("The supervised forest fit did not return a valid randomForestSRC object.",
         call. = FALSE)
  }
  forest <- .resolve.predict.forest(supervised.fit)
  family <- supervised.fit$family %||%
    (if (!is.null(forest)) forest$family else NULL) %||%
    (if (!is.null(fallback)) fallback$family.guess else NULL) %||%
    NA_character_
  xvar.names <- supervised.fit$xvar.names %||%
    (if (!is.null(forest)) forest$xvar.names else NULL) %||%
    (if (!is.null(fallback)) fallback$xvar.names else NULL)
  yvar.names <- supervised.fit$yvar.names %||%
    (if (!is.null(forest)) forest$yvar.names else NULL) %||%
    (if (!is.null(fallback)) fallback$yvar.names else NULL)
  if (is.null(xvar.names) || length(xvar.names) == 0L) {
    stop("The supervised forest fit did not return 'xvar.names'.",
         call. = FALSE)
  }
  if (is.null(yvar.names) || length(yvar.names) == 0L) {
    stop("The supervised forest fit did not return 'yvar.names'.",
         call. = FALSE)
  }
  list(
    family = family,
    xvar.names = xvar.names,
    yvar.names = yvar.names
  )
}
.infer.supervised.pred.n <- function(x) {
  n <- x$n %||% NULL
  if (!is.null(n) && length(n) == 1L && is.finite(n)) {
    return(as.integer(n))
  }
  if (!is.null(x$xvar)) {
    xdf <- tryCatch(as.data.frame(x$xvar, check.names = FALSE,
                                  stringsAsFactors = FALSE),
                    error = function(e) NULL)
    if (!is.null(xdf)) {
      return(nrow(xdf))
    }
  }
  pred <- x$predicted %||% x$predicted.oob %||% NULL
  if (!is.null(pred)) {
    return(NROW(pred))
  }
  for (slot in c("classOutput", "regrOutput")) {
    obj <- x[[slot]]
    if (!is.null(obj) && length(obj) > 0L) {
      for (nm in names(obj)) {
        comp <- obj[[nm]]
        if (!is.null(comp$predicted)) {
          return(NROW(comp$predicted))
        }
        if (!is.null(comp$predicted.oob)) {
          return(NROW(comp$predicted.oob))
        }
      }
    }
  }
  0L
}
.patch.supervised.prediction.object <- function(x, family,
                                                response.schema = NULL,
                                                yvar.names = NULL) {
  if (!(family %in% c("regr+", "class+", "mix+"))) {
    return(x)
  }
  n <- .infer.supervised.pred.n(x)
  if (n <= 0L) {
    return(x)
  }
  y <- x$yvar %||% NULL
  if (is.null(y)) {
    if (is.null(response.schema) || length(response.schema) == 0L) {
      return(x)
    }
    y <- .na.data.from.schema(response.schema, n)
  }
  else {
    y <- tryCatch(
      as.data.frame(y, check.names = FALSE, stringsAsFactors = FALSE),
      error = function(e) NULL
    )
    if (is.null(y)) {
      if (is.null(response.schema) || length(response.schema) == 0L) {
        return(x)
      }
      y <- .na.data.from.schema(response.schema, n)
    }
    else if (!is.null(response.schema) && length(response.schema) > 0L) {
      wanted <- names(response.schema)
      extra.cols <- setdiff(names(y), wanted)
      miss.cols <- setdiff(wanted, names(y))
      if (length(extra.cols) > 0L) {
        y <- y[, setdiff(names(y), extra.cols), drop = FALSE]
      }
      if (length(miss.cols) > 0L) {
        y.miss <- .na.data.from.schema(response.schema[miss.cols], nrow(y))
        y <- data.frame(y, y.miss, check.names = FALSE, stringsAsFactors = FALSE)
      }
      y <- y[, wanted, drop = FALSE]
      y <- .restore.schema(y, response.schema, restore.integer = TRUE)
    }
  }
  if (!is.null(yvar.names) && length(yvar.names) > 0L) {
    have <- intersect(yvar.names, names(y))
    if (length(have) == length(yvar.names)) {
      y <- y[, yvar.names, drop = FALSE]
    }
  }
  x$yvar <- y
  if (!is.null(yvar.names) && length(yvar.names) > 0L) {
    x$yvar.names <- yvar.names
  }
  x
}
.supervised.materialize.aux <- function(x, family, oob = FALSE,
                                        response.schema = NULL,
                                        yvar.names = NULL) {
  x <- .patch.supervised.prediction.object(
    x,
    family = family,
    response.schema = response.schema,
    yvar.names = yvar.names
  )
  if (isTRUE(oob)) {
    if (family %in% c("regr+", "class+", "mix+")) {
      getter <- .get.rfsrc.internal("get.mv.predicted")
      if (is.null(getter)) {
        stop("get.mv.predicted() is required to materialize supervised multivariate predictions.",
             call. = FALSE)
      }
      pred <- getter(x, oob = TRUE)
    }
    else {
      pred <- x$predicted.oob
    }
  }
  else {
    if (family %in% c("regr+", "class+", "mix+")) {
      getter <- .get.rfsrc.internal("get.mv.predicted")
      if (is.null(getter)) {
        stop("get.mv.predicted() is required to materialize supervised multivariate predictions.",
             call. = FALSE)
      }
      pred <- getter(x)
    }
    else {
      pred <- x$predicted
    }
  }
  out <- tryCatch(
    data.frame(predicted = pred, check.names = FALSE),
    error = function(e) e
  )
  if (inherits(out, "error")) {
    stop("Failed to materialize supervised predicted values: ",
         conditionMessage(out), call. = FALSE)
  }
  out
}
.align.supervised.aux <- function(aux, expected.names = NULL) {
  aux <- as.data.frame(aux, check.names = FALSE, stringsAsFactors = FALSE)
  if (ncol(aux) == 0L) {
    stop("The supervised prediction block has no columns.", call. = FALSE)
  }
  for (nm in names(aux)) {
    aux[[nm]] <- .as.numeric.safe(aux[[nm]])
  }
  if (is.null(expected.names)) {
    return(aux)
  }
  if (length(expected.names) != ncol(aux)) {
    stop("The supervised prediction block produced ", ncol(aux),
         " column(s), but ", length(expected.names),
         " were learned during training.", call. = FALSE)
  }
  names(aux) <- expected.names
  aux[, expected.names, drop = FALSE]
}
.fill.supervised.aux <- function(aux, init) {
  aux <- as.data.frame(aux, check.names = FALSE, stringsAsFactors = FALSE)
  if (length(aux) == 0L) {
    return(aux)
  }
  for (nm in names(aux)) {
    x <- .as.numeric.safe(aux[[nm]])
    bad <- !is.finite(x)
    if (any(bad)) {
      x[bad] <- as.numeric(init[[nm]] %||% NA_real_)
    }
    aux[[nm]] <- x
  }
  aux
}
.augment.predictor.map <- function(targets, predictor.map.raw, aux.names = NULL) {
  predictor.map.raw <- predictor.map.raw[targets]
  if (is.null(aux.names) || length(aux.names) == 0L) {
    return(predictor.map.raw)
  }
  out <- lapply(targets, function(target) {
    unique(c(predictor.map.raw[[target]], aux.names))
  })
  names(out) <- targets
  out
}
.fast.load.named <- function(name, root, label = "saved object", strict = TRUE) {
  .check.fst()
  out <- tryCatch(
    fast.load(name, root),
    error = function(e) e
  )
  if (inherits(out, "error")) {
    msg <- paste0("Failed to load ", label, " from ",
                  file.path(root, name), ": ",
                  conditionMessage(out))
    if (isTRUE(strict)) {
      stop(msg, call. = FALSE)
    }
    return(list(model = NULL, error = msg))
  }
  if (isTRUE(strict)) {
    return(out)
  }
  list(model = out, error = NULL)
}
.resolve.predict.forest <- function(model) {
  if (is.null(model)) return(NULL)
  if (is.list(model) && !is.null(model$forest) &&
      is.list(model$forest) && !is.null(model$forest$xvar.names)) {
    return(model$forest)
  }
  if (is.list(model) && !is.null(model$xvar.names)) {
    return(model)
  }
  NULL
}
.conform.x.to.forest <- function(x, model, ignore.levels = TRUE) {
  x <- data.frame(x, check.names = FALSE, stringsAsFactors = FALSE)
  forest <- .resolve.predict.forest(model)
  if (is.null(forest) || is.null(forest$xvar.names)) {
    return(x)
  }
  wanted <- forest$xvar.names
  ## add/drop/reorder to match the learner exactly
  miss.cols <- setdiff(wanted, names(x))
  extra.cols <- setdiff(names(x), wanted)
  if (length(extra.cols) > 0) {
    x <- x[, setdiff(names(x), extra.cols), drop = FALSE]
  }
  if (length(miss.cols) > 0) {
    for (nm in miss.cols) x[[nm]] <- NA
  }
  x <- x[, wanted, drop = FALSE]
  ## restore unordered and ordered factors using forest metadata
  if (!is.null(forest$xvar.factor)) {
    x <- check.factor(x, forest$xvar.factor, ignore = ignore.levels)
    gtypes <- forest$xvar.factor$generic.types
    if (!is.null(gtypes) && length(gtypes) == ncol(x)) {
      real.idx <- which(gtypes == "R")
      if (length(real.idx) > 0) {
        x[real.idx] <- lapply(x[real.idx], .as.numeric.safe)
      }
    }
  }
  x
}
.parse.full.sweep.options <- function(full.sweep.options = NULL) {
  fs <- full.sweep.options %||% list()
  allowed <- c(
    "mtry", "splitrule", "bootstrap", "sampsize", "samptype",
    "perf.type", "rfq", "save.memory", "importance", "proximity"
  )
  structural <- c("ntree", "nodesize", "nsplit")
  unknown <- setdiff(names(fs), c(structural, allowed))
  if (length(unknown) > 0L) {
    warning("Ignoring unknown full.sweep.options entries: ",
            paste(unknown, collapse = ", "),
            call. = FALSE)
  }
  list(
    ntree = fs$ntree %||% 100L,
    nodesize = fs$nodesize,
    nsplit = fs$nsplit %||% 10L,
    dots = fs[names(fs) %in% allowed]
  )
}
.make.learner.name <- function(i, target, prefix = "impute.learner.") {
  sprintf("%s%03d.%s", prefix, i, .safe.name(target))
}
.make.response.name <- function(existing.names = character()) {
  nm <- ".impute.learn.response."
  while (nm %in% existing.names) {
    nm <- paste0(nm, ".")
  }
  nm
}
.coerce.factor.levels <- function(x, levels, ordered = FALSE) {
  x <- as.character(x)
  x[!(x %in% levels)] <- NA_character_
  factor(x, levels = levels, ordered = ordered)
}
.harmonize.newdata <- function(newdata, manifest, verbose = TRUE) {
  newdata <- as.data.frame(newdata, stringsAsFactors = FALSE)
  required.cols <- manifest$columns
  added.cols <- setdiff(required.cols, names(newdata))
  extra.cols <- setdiff(names(newdata), required.cols)
  response.cols <- character(0)
  if (.is.supervised.manifest(manifest) &&
      !is.null(manifest$supervised$response.schema)) {
    response.cols <- intersect(names(manifest$supervised$response.schema), extra.cols)
  }
  other.extra.cols <- setdiff(extra.cols, response.cols)
  if (length(added.cols) > 0L) {
    for (nm in added.cols) {
      newdata[[nm]] <- NA
    }
    .msg("Added missing columns to newdata: ", paste(added.cols, collapse = ", "),
         verbose = verbose)
  }
  if (length(response.cols) > 0L) {
    .msg("Dropping supervised response columns from newdata: ",
         paste(response.cols, collapse = ", "),
         verbose = verbose)
  }
  if (length(other.extra.cols) > 0L) {
    .msg("Dropping extra columns from newdata: ", paste(other.extra.cols, collapse = ", "),
         verbose = verbose)
  }
  newdata <- newdata[, required.cols, drop = FALSE]
  unseen.levels <- vector("list", length(required.cols))
  names(unseen.levels) <- required.cols
  unseen.mask <- matrix(FALSE, nrow = nrow(newdata), ncol = length(required.cols),
                        dimnames = list(NULL, required.cols))
  for (nm in required.cols) {
    sc <- manifest$schema[[nm]]
    x <- newdata[[nm]]
    if (isTRUE(sc$is.factor)) {
      xchr <- as.character(x)
      bad <- !is.na(xchr) & !(xchr %in% sc$levels)
      unseen <- unique(xchr[bad])
      unseen <- stats::na.omit(unseen)
      if (length(unseen) > 0L) {
        unseen.levels[[nm]] <- as.character(unseen)
      }
      if (nrow(newdata) > 0L) {
        unseen.mask[, nm] <- bad
      }
      newdata[[nm]] <- .coerce.factor.levels(xchr, sc$levels, ordered = sc$ordered)
    }
    else {
      newdata[[nm]] <- .as.numeric.from.schema(x, sc)
    }
  }
  list(
    data = newdata,
    added.cols = added.cols,
    extra.cols = extra.cols,
    response.cols = response.cols,
    unseen.levels = unseen.levels,
    unseen.mask = as.data.frame(unseen.mask, stringsAsFactors = FALSE),
    unseen.rows = if (nrow(newdata) == 0L) logical(0) else rowSums(unseen.mask) > 0L
  )
}
.apply.init <- function(data, init, schema) {
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  for (nm in names(data)) {
    miss <- is.na(data[[nm]])
    if (!any(miss)) next
    sc <- schema[[nm]]
    val <- init[[nm]]
    if (isTRUE(sc$is.factor)) {
      x <- as.character(data[[nm]])
      x[miss] <- as.character(val)
      data[[nm]] <- .coerce.factor.levels(x, sc$levels, ordered = sc$ordered)
    }
    else {
      x <- .as.numeric.from.schema(data[[nm]], sc)
      x[miss] <- as.numeric(val)
      data[[nm]] <- x
    }
  }
  data
}
.restore.schema <- function(data, schema, restore.integer = TRUE) {
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  for (nm in names(data)) {
    sc <- schema[[nm]]
    if (isTRUE(sc$is.factor)) {
      data[[nm]] <- .coerce.factor.levels(data[[nm]], sc$levels, ordered = sc$ordered)
    }
    else if (.schema.restores.integer(sc)) {
      x <- .as.numeric.from.schema(data[[nm]], sc)
      data[[nm]] <- .restore.integer.vector(
        x,
        .schema.integer.info(sc),
        restore.integer = restore.integer
      )
    }
    else {
      data[[nm]] <- .as.numeric.from.schema(data[[nm]], sc)
    }
  }
  data
}
.na.data.from.schema <- function(schema, nrow) {
  if (is.null(schema) || length(schema) == 0L) {
    return(data.frame())
  }
  out <- setNames(vector("list", length(schema)), names(schema))
  for (nm in names(schema)) {
    sc <- schema[[nm]]
    if (isTRUE(sc$is.factor)) {
      out[[nm]] <- factor(rep(NA_character_, nrow),
                          levels = sc$levels,
                          ordered = sc$ordered)
    }
    else if (isTRUE(sc$is.integer) || isTRUE(sc$integer.storage)) {
      out[[nm]] <- rep(NA_integer_, nrow)
    }
    else {
      out[[nm]] <- rep(NA_real_, nrow)
    }
  }
  data.frame(out, check.names = FALSE, stringsAsFactors = FALSE)
}
.extract.prediction <- function(pred, family, target.schema,
                                restore.integer = TRUE) {
  if (is.null(pred)) return(NULL)
  if (identical(family, "regr")) {
    out <- pred$predicted
    if (.schema.restores.integer(target.schema)) {
      out <- .restore.integer.vector(
        out,
        .schema.integer.info(target.schema),
        restore.integer = restore.integer
      )
    }
    return(out)
  }
  out <- pred$class
  if (is.null(out) && !is.null(pred$predicted)) {
    out <- pred$predicted
  }
  if (isTRUE(target.schema$is.factor)) {
    out <- .coerce.factor.levels(out, target.schema$levels,
                                 ordered = target.schema$ordered)
  }
  out
}
.extract.class.prob <- function(pred, target.schema = NULL) {
  if (is.null(pred)) return(NULL)
  prob <- pred$predicted
  if (is.data.frame(prob)) {
    prob <- as.matrix(prob)
  }
  if (!is.matrix(prob)) {
    return(NULL)
  }
  if (is.null(colnames(prob)) && !is.null(target.schema$levels) &&
      ncol(prob) == length(target.schema$levels)) {
    colnames(prob) <- target.schema$levels
  }
  prob
}
.ood.delta.regression <- function(observed, pred, target.schema = NULL) {
  if (is.null(pred) || is.null(pred$predicted)) {
    return(rep(NA_real_, length(observed)))
  }
  out <- pred$predicted
  if (.schema.restores.integer(target.schema)) {
    out <- .restore.integer.vector(
      out,
      .schema.integer.info(target.schema),
      restore.integer = TRUE
    )
  }
  obs <- .as.numeric.from.schema(observed, target.schema)
  est <- .as.numeric.from.schema(out, target.schema)
  abs(obs - est)
}
.ood.delta.classification <- function(observed, pred, target.schema,
                                      prob.floor = 1e-12) {
  y <- as.character(observed)
  out <- rep(NA_real_, length(y))
  missing.y <- is.na(y)
  prob <- .extract.class.prob(pred, target.schema)
  if (!is.null(prob) && nrow(prob) == length(y)) {
    lev <- colnames(prob)
    pos <- match(y, lev)
    ok <- !missing.y & !is.na(pos)
    if (any(ok)) {
      p.true <- prob[cbind(which(ok), pos[ok])]
      p.true[!is.finite(p.true)] <- 0
      out[ok] <- -log(pmax(p.true, prob.floor))
    }
    if (any(!missing.y & is.na(pos))) {
      out[!missing.y & is.na(pos)] <- Inf
    }
    return(out)
  }
  cls <- pred$class
  if (is.null(cls) && !is.null(pred$predicted)) {
    cls <- pred$predicted
  }
  cls <- as.character(cls)
  ok <- !missing.y & !is.na(cls)
  if (!any(ok)) {
    return(out)
  }
  if (isTRUE(target.schema$ordered)) {
    obs.code <- match(y[ok], target.schema$levels)
    cls.code <- match(cls[ok], target.schema$levels)
    denom <- max(1L, length(target.schema$levels) - 1L)
    out[ok] <- abs(obs.code - cls.code) / denom
  }
  else {
    out[ok] <- as.numeric(y[ok] != cls[ok])
  }
  out
}
.compute.ood.delta <- function(observed, pred, target.schema) {
  if (isTRUE(target.schema$is.factor)) {
    .ood.delta.classification(observed, pred, target.schema = target.schema)
  }
  else {
    .ood.delta.regression(observed, pred, target.schema = target.schema)
  }
}
.make.ood.reference <- function(x, probs = seq(0, 1, length.out = 257)) {
  x <- as.numeric(x)
  x <- x[is.finite(x)]
  probs <- unique(sort(pmin(1, pmax(0, probs))))
  if (length(probs) < 2L) {
    probs <- c(0, 1)
  }
  if (length(x) == 0L) {
    return(list(
      probs = c(0, 1),
      quantiles = c(0, 0),
      n = 0L
    ))
  }
  q <- as.numeric(stats::quantile(x, probs = probs, names = FALSE,
                                  na.rm = TRUE, type = 8))
  q <- cummax(q)
  list(
    probs = probs,
    quantiles = q,
    n = length(x)
  )
}
.eval.ood.reference <- function(x, ref) {
  out <- rep(NA_real_, length(x))
  if (length(out) == 0L || is.null(ref) || is.null(ref$quantiles)) {
    return(out)
  }
  ok <- is.finite(x)
  if (!any(ok)) {
    return(out)
  }
  q <- as.numeric(ref$quantiles)
  p <- as.numeric(ref$probs)
  if (length(q) == 0L || length(p) == 0L) {
    return(out)
  }
  if (length(q) == 1L || all(q == q[1L])) {
    out[ok] <- ifelse(x[ok] <= q[1L], p[1L], p[length(p)])
    return(out)
  }
  idx <- findInterval(x[ok], q, rightmost.closed = TRUE, all.inside = TRUE)
  idx <- pmax(1L, pmin(length(p), idx))
  out[ok] <- p[idx]
  out
}
.canonicalize.ood.aggregate <- function(aggregate = "weighted.mean",
                                        aggregate.args = NULL) {
  choices <- c(
    "bounded.product",
    "weighted.mean",
    "weighted.lp",
    "weighted.lp.log",
    "top.k"
  )
  if (is.null(aggregate) || length(aggregate) == 0L) {
    aggregate <- "weighted.mean"
  }
  aggregate <- match.arg(aggregate, choices = choices)
  if (is.null(aggregate.args)) {
    aggregate.args <- list()
  }
  if (!is.list(aggregate.args)) {
    stop("'aggregate.args' must be a list.", call. = FALSE)
  }
  used <- character(0)
  args <- list()
  if (aggregate %in% c("weighted.lp", "weighted.lp.log")) {
    p <- aggregate.args$p %||% 2
    if (!is.numeric(p) || length(p) != 1L || !is.finite(p) || p < 1) {
      stop("The row aggregate power 'p' must be a finite scalar >= 1.",
           call. = FALSE)
    }
    args$p <- as.numeric(p)
    used <- c(used, "p")
  }
  if (aggregate == "weighted.lp.log") {
    agg.eps <- aggregate.args$eps %||% 1e-12
    if (!is.numeric(agg.eps) || length(agg.eps) != 1L ||
        !is.finite(agg.eps) || agg.eps <= 0) {
      stop("The row aggregate 'eps' must be a finite positive scalar.",
           call. = FALSE)
    }
    args$eps <- as.numeric(agg.eps)
    used <- c(used, "eps")
  }
  if (aggregate == "top.k") {
    k <- aggregate.args$k %||% aggregate.args$top.k %||% 1L
    if (!is.numeric(k) || length(k) != 1L || !is.finite(k) || k < 1) {
      stop("The top-k row aggregate requires a finite scalar 'k' >= 1.",
           call. = FALSE)
    }
    args$k <- as.integer(max(1L, round(k)))
    used <- c(used, "k", "top.k")
  }
  if (aggregate == "bounded.product") {
    agg.eps <- aggregate.args$eps %||% 1e-12
    if (!is.numeric(agg.eps) || length(agg.eps) != 1L ||
        !is.finite(agg.eps) || agg.eps <= 0) {
      stop("The bounded-product row aggregate requires a finite positive 'eps'.",
           call. = FALSE)
    }
    args$eps <- as.numeric(agg.eps)
    used <- c(used, "eps")
  }
  extra <- setdiff(names(aggregate.args), unique(used))
  if (length(extra) > 0L) {
    warning("Ignoring unknown aggregate.args entries: ",
            paste(extra, collapse = ", "),
            call. = FALSE)
  }
  list(name = aggregate, args = args)
}
.same.ood.aggregate <- function(aggregate, aggregate.args = NULL,
                                ref.aggregate, ref.aggregate.args = NULL,
                                tolerance = sqrt(.Machine$double.eps)) {
  ax <- tryCatch(
    .canonicalize.ood.aggregate(aggregate, aggregate.args),
    error = function(e) NULL
  )
  ay <- tryCatch(
    .canonicalize.ood.aggregate(ref.aggregate, ref.aggregate.args),
    error = function(e) NULL
  )
  if (is.null(ax) || is.null(ay)) {
    return(FALSE)
  }
  if (!identical(ax$name, ay$name)) {
    return(FALSE)
  }
  if (!setequal(names(ax$args), names(ay$args))) {
    return(FALSE)
  }
  if (length(ax$args) == 0L) {
    return(TRUE)
  }
  isTRUE(all(vapply(names(ax$args), function(nm) {
    isTRUE(all.equal(ax$args[[nm]], ay$args[[nm]],
                     tolerance = tolerance,
                     check.attributes = FALSE))
  }, logical(1))))
}
.ood.aggregate.vector <- function(x, weight, aggregate.spec) {
  ok <- is.finite(x) & is.finite(weight) & weight > 0
  if (!any(ok)) {
    return(NA_real_)
  }
  x <- as.numeric(x[ok])
  weight <- as.numeric(weight[ok])
  denom <- sum(weight)
  if (!is.finite(denom) || denom <= 0) {
    return(NA_real_)
  }
  switch(
    aggregate.spec$name,
    weighted.mean = {
      sum(weight * x) / denom
    },
    weighted.lp = {
      p <- aggregate.spec$args$p
      (sum(weight * (x ^ p)) / denom) ^ (1 / p)
    },
    weighted.lp.log = {
      p <- aggregate.spec$args$p
      agg.eps <- aggregate.spec$args$eps
      u <- pmin(1, pmax(0, x))
      z <- -log(pmax(1 - u, agg.eps))
      (sum(weight * (z ^ p)) / denom) ^ (1 / p)
    },
    top.k = {
      k <- min(length(x), aggregate.spec$args$k)
      ord <- order(x, decreasing = TRUE, na.last = NA)
      sel <- ord[seq_len(k)]
      sum(weight[sel] * x[sel]) / sum(weight[sel])
    },
    bounded.product = {
      agg.eps <- aggregate.spec$args$eps
      u <- pmin(1, pmax(0, x))
      wnorm <- weight / denom
      surv <- pmax(1 - u, agg.eps)
      pmin(1, pmax(0, 1 - exp(sum(wnorm * log(surv)))))
    },
    stop("Unknown OOD row aggregate: ", aggregate.spec$name, call. = FALSE)
  )
}
.max.ood.aggregate.value <- function(aggregate = "weighted.mean",
                                    aggregate.args = NULL) {
  aggregate.spec <- .canonicalize.ood.aggregate(aggregate, aggregate.args)
  switch(
    aggregate.spec$name,
    weighted.lp.log = -log(aggregate.spec$args$eps),
    1
  )
}
.aggregate.ood.row <- function(mat, weight = NULL,
                               aggregate = "weighted.mean",
                               aggregate.args = NULL) {
  mat <- as.matrix(mat)
  if (ncol(mat) == 0L) {
    return(rep(NA_real_, nrow(mat)))
  }
  if (is.null(weight)) {
    weight <- rep(1, ncol(mat))
  }
  if (length(weight) != ncol(mat)) {
    stop("'weight' must have one entry per target.", call. = FALSE)
  }
  weight <- as.numeric(weight)
  aggregate.spec <- .canonicalize.ood.aggregate(aggregate, aggregate.args)
  out <- rep(NA_real_, nrow(mat))
  for (i in seq_len(nrow(mat))) {
    out[i] <- .ood.aggregate.vector(mat[i, ], weight, aggregate.spec)
  }
  out
}
.as.ood.train.score.matrix <- function(x, targets = NULL) {
  if (is.null(x)) {
    return(NULL)
  }
  mat <- tryCatch(as.matrix(x), error = function(e) NULL)
  if (is.null(mat)) {
    return(NULL)
  }
  storage.mode(mat) <- "double"
  if (is.null(colnames(mat)) && !is.null(targets) && ncol(mat) == length(targets)) {
    colnames(mat) <- targets
  }
  mat
}
.rebuild.ood.row.reference <- function(train.target.score, targets, weight,
                                       saved.targets = NULL,
                                       aggregate = "weighted.mean",
                                       aggregate.args = NULL) {
  mat <- .as.ood.train.score.matrix(train.target.score, targets = saved.targets)
  if (is.null(mat)) {
    return(list(
      reference = NULL,
      row.score = NULL,
      n.finite = 0L,
      reason = paste(
        "No saved training OOD target scores are available.",
        "Refit with the updated imputer to enable row-level percentile",
        "recalibration for arbitrary target subsets, weights, and row aggregates."
      )
    ))
  }
  if (length(targets) == 0L || ncol(mat) == 0L) {
    return(list(
      reference = NULL,
      row.score = rep(NA_real_, nrow(mat)),
      n.finite = 0L,
      reason = "No saved training OOD target scores are available for the requested targets."
    ))
  }
  if (is.null(colnames(mat))) {
    return(list(
      reference = NULL,
      row.score = NULL,
      n.finite = 0L,
      reason = paste(
        "Saved training OOD target scores lack target names.",
        "Refit with the updated imputer to enable row-level percentile",
        "recalibration for arbitrary target subsets, weights, and row aggregates."
      )
    ))
  }
  missing.targets <- setdiff(targets, colnames(mat))
  if (length(missing.targets) > 0L) {
    return(list(
      reference = NULL,
      row.score = NULL,
      n.finite = 0L,
      reason = paste0(
        "Saved training OOD target scores are missing requested targets: ",
        paste(missing.targets, collapse = ", "), "."
      )
    ))
  }
  row.score <- .aggregate.ood.row(
    mat[, targets, drop = FALSE],
    weight = weight,
    aggregate = aggregate,
    aggregate.args = aggregate.args
  )
  n.finite <- sum(is.finite(row.score))
  if (n.finite == 0L) {
    return(list(
      reference = NULL,
      row.score = row.score,
      n.finite = 0L,
      reason = paste(
        "No finite row-level training OOD scores were available for the",
        "requested targets, weights, and row aggregates."
      )
    ))
  }
  list(
    reference = .make.ood.reference(row.score),
    row.score = row.score,
    n.finite = n.finite,
    reason = NULL
  )
}
.resolve.ood.weight <- function(targets, weight = NULL, default = NULL,
                                warn.extra = TRUE) {
  targets <- as.character(targets %||% character(0))
  if (length(targets) == 0L) {
    return(setNames(numeric(0), character(0)))
  }
  validate.named.weight <- function(x, what = "'weight'") {
    if (is.null(names(x))) {
      return(invisible(NULL))
    }
    dup <- unique(names(x)[duplicated(names(x))])
    if (length(dup) > 0L) {
      stop(what, " contains duplicated target names: ",
           paste(dup, collapse = ", "),
           call. = FALSE)
    }
    invisible(NULL)
  }
  align.named.weight <- function(x, what = "'weight'",
                                 warn.extra = TRUE,
                                 require.match = TRUE,
                                 fill = 0) {
    validate.named.weight(x, what = what)
    out <- setNames(rep(fill, length(targets)), targets)
    extra.targets <- setdiff(names(x), targets)
    if (isTRUE(warn.extra) && length(extra.targets) > 0L) {
      warning("Ignoring ", what, " entries for unknown targets: ",
              paste(extra.targets, collapse = ", "),
              call. = FALSE)
    }
    matched.targets <- intersect(targets, names(x))
    if (length(matched.targets) == 0L) {
      if (isTRUE(require.match)) {
        stop(what, " did not match any requested targets.", call. = FALSE)
      }
      return(out)
    }
    out[matched.targets] <- as.numeric(x[matched.targets])
    out
  }
  if (is.null(weight) || length(weight) == 0L) {
    if (is.null(default) || length(default) == 0L) {
      weight <- setNames(rep(1, length(targets)), targets)
    }
    else if (is.null(names(default))) {
      if (length(default) != length(targets)) {
        stop("'default' must have one entry per target.", call. = FALSE)
      }
      weight <- setNames(as.numeric(default), targets)
    }
    else {
      weight <- align.named.weight(
        x = default,
        what = "'default'",
        warn.extra = FALSE,
        require.match = FALSE,
        fill = 0
      )
    }
  }
  else if (is.null(names(weight))) {
    if (length(weight) != length(targets)) {
      stop("'weight' must have one entry per target.", call. = FALSE)
    }
    weight <- setNames(as.numeric(weight), targets)
  }
  else {
    weight <- align.named.weight(
      x = weight,
      what = "'weight'",
      warn.extra = warn.extra,
      require.match = TRUE,
      fill = 0
    )
  }
  if (any(!is.finite(weight) | weight < 0)) {
    stop("'weight' must contain finite nonnegative values.", call. = FALSE)
  }
  if (!any(weight > 0)) {
    stop("At least one target weight must be positive.", call. = FALSE)
  }
  weight
}
.same.ood.weight <- function(targets, x, y,
                             tolerance = sqrt(.Machine$double.eps)) {
  if (is.null(x) || is.null(y)) {
    return(FALSE)
  }
  wx <- tryCatch(
    .resolve.ood.weight(targets, x, warn.extra = FALSE),
    error = function(e) NULL
  )
  wy <- tryCatch(
    .resolve.ood.weight(targets, y, warn.extra = FALSE),
    error = function(e) NULL
  )
  if (is.null(wx) || is.null(wy)) {
    return(FALSE)
  }
  isTRUE(all.equal(unname(wx), unname(wy), tolerance = tolerance,
                   check.attributes = FALSE))
}
.compute.pass.diff <- function(old.data, new.data, missing.mask, schema, scale, targets) {
  diffs <- sapply(targets, function(y) {
    idx <- missing.mask[[y]]
    if (!any(idx)) return(NA_real_)
    xo <- old.data[[y]][idx]
    xn <- new.data[[y]][idx]
    sc <- schema[[y]]
    if (isTRUE(sc$is.factor)) {
      return(sum(as.character(xn) != as.character(xo), na.rm = TRUE) /
               (0.001 + length(xn)))
    }
    xo <- .as.numeric.from.schema(xo, sc)
    xn <- .as.numeric.from.schema(xn, sc)
    sy <- scale[[y]] %||% 1
    if (!is.finite(sy) || sy <= 0) sy <- 1
    sqrt(mean((xn - xo)^2, na.rm = TRUE) / (0.001 + sy^2))
  })
  mean(diffs, na.rm = TRUE)
}
.prepare.impute.learn.newdata <- function(object, newdata,
                                          targets = NULL,
                                          max.predict.iter = 3L,
                                          eps = 1e-3,
                                          restore.integer = TRUE,
                                          cache.learners = c("session", "none", "all"),
                                          verbose = TRUE) {
  if (!inherits(object, "impute.learn.rfsrc")) {
    stop("'object' must inherit from class 'impute.learn.rfsrc'.", call. = FALSE)
  }
  object$manifest <- .normalize.impute.learn.manifest(object$manifest)
  if (!is.null(targets)) {
    bad.targets <- setdiff(targets, object$manifest$targets)
    if (length(bad.targets) > 0L) {
      warning("Ignoring unknown targets: ",
              paste(bad.targets, collapse = ", "),
              call. = FALSE)
    }
  }
  use.targets <- if (is.null(targets)) object$manifest$targets else {
    intersect(object$manifest$targets, targets)
  }
  if (length(use.targets) == 0L) {
    stop("No valid targets requested.", call. = FALSE)
  }
  cache.learners <- match.arg(cache.learners)
  harmonized <- .harmonize.newdata(newdata, object$manifest, verbose = verbose)
  data <- harmonized$data
  original.missing <- as.data.frame(is.na(data[, use.targets, drop = FALSE]))
  names(original.missing) <- use.targets
  data <- .apply.init(data, object$manifest$init, object$manifest$schema)
  ## make the raw working copy respect the training x-schema before the sweep
  data <- .restore.schema(data, object$manifest$schema,
                          restore.integer = restore.integer)
  working.data <- data
  supervised.info <- list(enabled = FALSE)
  if (.is.supervised.manifest(object$manifest)) {
    sup.info <- .predict.get.supervised.model(object)
    if (is.null(sup.info$model)) {
      stop(sup.info$error %||% "The supervised forest could not be loaded.",
           call. = FALSE)
    }
    aux <- tryCatch(
      {
        pred.newdata <- .conform.x.to.forest(
          data,
          sup.info$model,
          ignore.levels = TRUE
        )
        pred <- predict(sup.info$model, pred.newdata)
        aux <- .supervised.materialize.aux(
          pred,
          family = object$manifest$supervised$family,
          oob = FALSE,
          response.schema = object$manifest$supervised$response.schema,
          yvar.names = object$manifest$supervised$yvar.names
        )
        aux <- .align.supervised.aux(
          aux,
          expected.names = object$manifest$supervised$aux.names
        )
        .fill.supervised.aux(aux, object$manifest$supervised$aux.init)
      },
      error = function(e) e
    )
    if (inherits(aux, "error")) {
      stop("Failed to compute supervised auxiliary predictors for newdata: ",
           conditionMessage(aux), call. = FALSE)
    }
    working.data <- data.frame(data, aux, check.names = FALSE)
    supervised.info <- list(
      enabled = TRUE,
      family = object$manifest$supervised$family,
      aux.names = names(aux),
      n.aux = ncol(aux),
      response.names = names(object$manifest$supervised$response.schema %||% list()),
      dropped.response.columns = harmonized$response.cols,
      loaded.from.disk = isTRUE(sup.info$loaded.from.disk)
    )
  }
  pass.history <- numeric(0)
  sweep.order <- object$manifest$sweep.order
  sweep.order <- sweep.order[sweep.order %in% use.targets]
  cache.env <- if (identical(cache.learners, "none")) NULL else new.env(parent = emptyenv())
  disk.load.targets <- character(0)
  target.issues <- setNames(vector("list", length(use.targets)), use.targets)
  record.issue <- function(target, message) {
    current <- target.issues[[target]]
    if (is.null(current)) current <- character(0)
    if (!(message %in% current)) {
      target.issues[[target]] <<- c(current, message)
    }
    invisible(NULL)
  }
  if (identical(cache.learners, "all")) {
    .msg("Preloading learner bank...", verbose = verbose)
    for (target in use.targets) {
      info <- object$manifest$learners[[target]]
      if (!identical(info$status, "ok")) next
      mdl.info <- .predict.get.model(object, target, cache.env = cache.env)
      if (isTRUE(mdl.info$loaded.from.disk)) {
        disk.load.targets <- c(disk.load.targets, target)
      }
      if (is.null(mdl.info$model) && !is.null(mdl.info$error)) {
        record.issue(target, mdl.info$error)
      }
    }
  }
  any.target.missing <- length(sweep.order) > 0L &&
    any(as.matrix(original.missing[, sweep.order, drop = FALSE]))
  if (isTRUE(any.target.missing)) {
    .msg("Starting prediction-time sweep...", verbose = verbose)
    for (iter in seq_len(max.predict.iter)) {
      old.data <- data
      .msg("  prediction pass ", iter, "/", max.predict.iter, verbose = verbose)
      for (target in sweep.order) {
        miss.idx <- which(original.missing[[target]])
        if (length(miss.idx) == 0L) next
        info <- object$manifest$learners[[target]]
        if (!identical(info$status, "ok")) {
          msg <- paste0("No trained learner is available (status = ",
                        info$status %||% "unknown", ").")
          record.issue(target, msg)
          .msg("    skipping `", target, "` (", msg, ")", verbose = verbose)
          next
        }
        mdl.info <- .predict.get.model(object, target, cache.env = cache.env)
        mdl <- mdl.info$model
        if (isTRUE(mdl.info$loaded.from.disk)) {
          disk.load.targets <- c(disk.load.targets, target)
        }
        if (is.null(mdl)) {
          msg <- mdl.info$error %||% "learner could not be loaded"
          record.issue(target, msg)
          .msg("    skipping `", target, "` (", msg, ")", verbose = verbose)
          next
        }
        xvars <- object$manifest$predictor.map[[target]]
        pred.df <- working.data[miss.idx, xvars, drop = FALSE]
        pred.df <- .conform.x.to.forest(pred.df, mdl)
        pred <- tryCatch(
          predict(mdl, pred.df),
          error = function(e) e
        )
        if (inherits(pred, "error")) {
          msg <- paste0("Prediction failed: ", conditionMessage(pred))
          record.issue(target, msg)
          .msg("    prediction failed for `", target, "`: ", conditionMessage(pred),
               verbose = verbose)
          next
        }
        values <- .extract.prediction(
          pred,
          info$family,
          object$manifest$schema[[target]],
          restore.integer = restore.integer
        )
        data[miss.idx, target] <- values
        working.data[miss.idx, target] <- values
        if (identical(cache.learners, "none") && is.null(object$models[[target]])) {
          rm(mdl)
          gc()
        }
      }
      diff.err <- .compute.pass.diff(old.data, data, original.missing,
                                     object$manifest$schema,
                                     object$manifest$scale,
                                     sweep.order)
      pass.history <- c(pass.history, diff.err)
      .msg("    pass diff = ", format(diff.err, digits = 4), verbose = verbose)
      if (is.finite(diff.err) && diff.err < eps) {
        .msg("    convergence criterion met; stopping early.", verbose = verbose)
        break
      }
    }
  }
  else {
    .msg("No missing values were found among requested targets; ",
         "skipping iterative sweep.", verbose = verbose)
  }
  data <- .restore.schema(data, object$manifest$schema,
                          restore.integer = restore.integer)
  working.data[names(data)] <- data
  target.issues <- target.issues[lengths(target.issues) > 0L]
  list(
    data = data,
    working.data = working.data,
    use.targets = use.targets,
    harmonized = harmonized,
    cache.learners = cache.learners,
    cache.env = cache.env,
    info = list(
      n.passes = length(pass.history),
      pass.diff = pass.history,
      targets = use.targets,
      added.columns = harmonized$added.cols,
      dropped.extra.columns = harmonized$extra.cols,
      dropped.response.columns = harmonized$response.cols,
      unseen.levels = harmonized$unseen.levels,
      unseen.rows = harmonized$unseen.rows,
      cache.learners = cache.learners,
      n.disk.loads = length(disk.load.targets),
      disk.load.targets = unique(disk.load.targets),
      supervised = supervised.info,
      target.issues = target.issues
    )
  )
}
.fast.load.learner <- function(target, info, learner.root, strict = TRUE) {
  .check.fst()
  out <- tryCatch(
    fast.load(info$learner.name, learner.root),
    error = function(e) e
  )
  if (inherits(out, "error")) {
    msg <- paste0("Failed to load learner for `", target, "` from ",
                  file.path(learner.root, info$learner.name), ": ",
                  conditionMessage(out))
    if (isTRUE(strict)) {
      stop(msg, call. = FALSE)
    }
    return(list(model = NULL, error = msg))
  }
  if (isTRUE(strict)) {
    return(out)
  }
  list(model = out, error = NULL)
}
.predict.get.model <- function(object, target, cache.env = NULL) {
  if (!inherits(object, "impute.learn.rfsrc")) {
    stop("'object' must inherit from class 'impute.learn.rfsrc'.", call. = FALSE)
  }
  object$manifest <- .normalize.impute.learn.manifest(object$manifest)
  if (!is.null(cache.env) && exists(target, envir = cache.env, inherits = FALSE)) {
    return(list(
      model = get(target, envir = cache.env, inherits = FALSE),
      loaded.from.disk = FALSE,
      cached = TRUE,
      error = NULL
    ))
  }
  if (!is.null(object$models[[target]])) {
    mdl <- object$models[[target]]
    if (!is.null(cache.env)) assign(target, mdl, envir = cache.env)
    return(list(model = mdl, loaded.from.disk = FALSE,
                cached = !is.null(cache.env), error = NULL))
  }
  info <- object$manifest$learners[[target]]
  if (is.null(info) || !identical(info$status, "ok")) {
    return(list(model = NULL, loaded.from.disk = FALSE,
                cached = FALSE, error = NULL))
  }
  if (is.null(object$path)) {
    return(list(
      model = NULL,
      loaded.from.disk = FALSE,
      cached = FALSE,
      error = paste0("Learner for `", target, "` is not available in memory ",
                     "and no saved imputer path is attached to 'object'.")
    ))
  }
  learner.root <- file.path(object$path, object$manifest$learner.root)
  loaded <- .fast.load.learner(target, info, learner.root, strict = FALSE)
  mdl <- loaded$model
  if (!is.null(cache.env) && !is.null(mdl)) {
    assign(target, mdl, envir = cache.env)
  }
  list(model = mdl,
       loaded.from.disk = !is.null(mdl),
       cached = !is.null(cache.env) && !is.null(mdl),
       error = loaded$error)
}
.predict.get.supervised.model <- function(object) {
  if (!inherits(object, "impute.learn.rfsrc")) {
    stop("'object' must inherit from class 'impute.learn.rfsrc'.", call. = FALSE)
  }
  object$manifest <- .normalize.impute.learn.manifest(object$manifest)
  if (!.is.supervised.manifest(object$manifest)) {
    return(list(
      model = NULL,
      loaded.from.disk = FALSE,
      enabled = FALSE,
      error = NULL
    ))
  }
  if (!is.null(object$supervised.model)) {
    return(list(
      model = object$supervised.model,
      loaded.from.disk = FALSE,
      enabled = TRUE,
      error = NULL
    ))
  }
  if (is.null(object$path)) {
    return(list(
      model = NULL,
      loaded.from.disk = FALSE,
      enabled = TRUE,
      error = "The supervised forest is not available in memory and no saved imputer path is attached to 'object'."
    ))
  }
  learner.root <- file.path(object$path, object$manifest$learner.root)
  loaded <- .fast.load.named(
    name = object$manifest$supervised$model.name,
    root = learner.root,
    label = "supervised forest",
    strict = FALSE
  )
  list(
    model = loaded$model,
    loaded.from.disk = !is.null(loaded$model),
    enabled = TRUE,
    error = loaded$error
  )
}

quantreg.rfsrc <- function(formula, data, object, newdata,
                           method = "local", splitrule = NULL,
                           prob = NULL, prob.epsilon = NULL,
                           oob = TRUE, fast = FALSE, maxn = 1e3, ...)
{
  ## we intialize the grow primary operation as false
  grow <- FALSE
  ## which method will be used?
  method <- match.arg(method, c("forest", "local", "gk", "GK", "G-K", "g-k"))
  ## we now allow other regression splitting rules to be used 
  ## splitrule <- "quantile.regr"
  ## --------------------------------------------------------------------------------
  ##
  ## grow mode
  ##
  ## if no object is present, then the user is requesting a quantile regression forest
  ##
  ## --------------------------------------------------------------------------------
  if (missing(object)) {
    ## the primary operation is grow 
    grow <- TRUE
    ## list of forest parameters
    rfnames <- get.rfnames(hidden = TRUE)
    ## restrict to allowed values
    rfnames <- rfnames[rfnames != "formula"            &
                       rfnames != "data"               &
                       rfnames != "splitrule"          ]
    ## get the permissible hidden options
    dots <- list(...)
    dots <- dots[names(dots) %in% rfnames]
    ## manually set key hidden options 
    dots$prob <- prob
    dots$prob.epsilon <- prob.epsilon
    ## set the default split rule
    if (is.null(splitrule)) {
      splitrule <- "la.quantile.regr"
    }
    ## quantile splitrule cannot be used for multivariate regression
    fmly <- parseFormula(formula, data)$family
    if (fmly == "regr+" || fmly == "class+" || fmly == "mix+") {
      if (splitrule == "la.quantile.regr" || splitrule == "la.quantile.regr") {
        splitrule <- NULL
      }
    }
    ## determine the grow interface - rfsrc or rfsrc.fast?
    if (!fast) {
      rfsrc.grow <- "rfsrc"
    }
    else {
      rfsrc.grow <- "rfsrc.fast"
    }
    ## if the user wants method=forest, set GK to false and request forest weights
    if (method == "forest") {
      dots$gk.quantile <- FALSE
      dots$forest.wt <- if (oob) "oob" else TRUE
    }
    ## if user wants local method, set GK to false
    else if (method == "local") {
      dots$gk.quantile <- FALSE
    }
    ## user is requesting GK method
    else {
      dots$gk.quantile <- TRUE
    }
    ## set the hidden quantile.regr flag as TRUE
    dots$quantile.regr <- TRUE
    ## grow a quantile regression forest
    object <- do.call(rfsrc.grow, c(list(formula = formula, data = data, splitrule = splitrule), dots))
    ## save the grow yvar - this will be crucial downstream for prediction
    object$yvar.grow <- object$yvar
    ## ----------------------------------------------------------------------
    ##
    ## save the grow residual - this will be crucial downstream for prediction
    ##
    ## ----------------------------------------------------------------------
    ## pull the target outcome names
    if (object$family == "regr") {
      ynames <- object$yvar.names
    }
    else {
      ynames <- names(object$regrOutput)
    }
    ## cycle over ynames, extract residual
    if (ncol(cbind(object$yvar.grow)) > 1) {
      object$res.grow <- do.call(cbind, lapply(ynames, function(yn) {
        if (oob && !is.null(object$regrOutput[[yn]]$predicted.oob)) {
          object$yvar[, yn] - object$regrOutput[[yn]]$predicted.oob
        }
        else {
          object$yvar[, yn] - object$regrOutput[[yn]]$predicted
        }
      }))
      colnames(object$res.grow) <- ynames
    }
    else {##only one yvar - vectorize it
      if (oob && !is.null(object$predicted.oob)) {
        object$res.grow <- object$yvar - object$predicted.oob
      }
        else {
        object$res.grow <- object$yvar - object$predicted
      }
    }
  }
  ## user has passed in an object - is it a quantile object?
  else {
    if (sum(grepl("quantreg", class(object))) == 0) { 
      stop("object must be a quantreg object")
    }
  }
  ## -----------------------------------------------------------------------
  ##
  ## forest object must contain regression outcomes
  ## this includes multivariate regression and mixed multivariate regression
  ##
  ## -----------------------------------------------------------------------
  if (!(object$family == "regr" | !is.null(object$regrOutput))) {
    stop("this function requires at least one outcome that is continuous\n")
  }
  ## ---------------------------------------
  ##
  ##  save grow prob for later calculations
  ##  can be over-written in predict mode
  ##
  ## -----------------------------------------
  prob.grow <- object$forest$prob
  ## ---------------------------------------
  ##
  ##  prediction mode
  ##
  ## -----------------------------------------
  if (!grow || !missing(newdata)) {
    ## the user can over-ride grow options in predict mode
    ## if prob and prob.epsilon are non-NULL, over-ride grow values
    ## make sure to overwrite prob.grow 
    dots <- list(...)
    if (!is.null(prob)) {
      prob.grow <- dots$prob <- prob
    }
    else {
      dots$prob <- prob.grow
    }
    dots$prob.epsilon <- prob.epsilon
    ## grow mode is in effect, but new data is present --> oob cannot be TRUE
    if (!missing(newdata)) {
      oob <- FALSE
    }
    ## forest method -> set GK to false and request forest weights
    if (method == "forest") {
      dots$gk.quantile <- FALSE
      dots$forest.wt <- TRUE
      if (missing(newdata) && oob) {
        dots$forest.wt <- "oob"
      }
    }
    ## local method -> set GK to false and request forest weights
    else if (method == "local") {
      dots$gk.quantile <- FALSE
    }
    ## user is requesting GK method
    else {
      dots$gk.quantile <- TRUE
    }
    ## save grow information before restore/predict over-writes it
    yvar.grow <- object$yvar.grow
    res.grow <- object$res.grow
    ## restore call on grow data 
    if (missing(newdata)) {
      object <- do.call(predict, c(list(object = object), dots))
    }
    ## predict call with new data
    else {
      object <- do.call(predict, c(list(object = object, newdata = newdata), dots))
    }
    ## restore grow information
    object$yvar.grow <- yvar.grow
    object$res.grow <- res.grow
  }
  ## -----------------------------------
  ##
  ## grow/test done: final value of prob
  ##
  ## -----------------------------------
  prob <- prob.grow
  ## -----------------------------------
  ##
  ## pull the target outcome names
  ##
  ## -----------------------------------
  if (object$family == "regr") {
    ynames <- object$yvar.names
  }
  else {
    ynames <- names(object$regrOutput)
  }
  ## -----------------------------------
  ##
  ## finally - extract quantile information
  ##
  ## -----------------------------------
  rO <- lapply(ynames, function(yn) {
    ## ------------------------------------
    ##
    ## GK method - also clean up the object   
    ##
    ## ------------------------------------
    if (!(method == "forest" || method == "local")) {
      if (ncol(cbind(object$yvar.grow)) > 1) {
        y <- object$yvar.grow[, yn]
        if (oob && !is.null(object$regrOutput[[yn]]$quantile.oob)) {
        quant <- object$regrOutput[[yn]]$quantile.oob
        }
        else {
          quant <- object$regrOutput[[yn]]$quantile
        }
        object$regrOutput[[yn]]$quantile.oob <<- object$regrOutput[[yn]]$quantile <<- NULL
      }
      else {
        y <- object$yvar.grow
        if (oob && !is.null(object$quantile.oob)) {
          quant <- object$quantile.oob
        }
        else {
          quant <- object$quantile
        }
        object$quantile.oob <<- object$quantile <<- NULL
      }
      ## append a minimum and maximum y value corresponding to prob = 0, 1
      quant.temp <- cbind(min(y, na.rm = TRUE) - 1, quant, max(y, na.rm = TRUE))
      prob.temp <- c(0, prob, 1)
      ## unique y values
      yunq <- sort(unique(y))
      ## cdf calculation
      ## trim atoms: useful for big data otherwise O(n^2)
      if (length(yunq) > maxn) {
        yunq <- yunq[unique(round(seq(1, length(yunq), length.out = maxn)))]
      }
      cdf <- t(apply(quant.temp, 1, function(q) {prob.temp[sIndex(q, yunq)]}))
      ## density
      density <- t(apply(cbind(0, cdf), 1, diff))
    }
    ## ------------------------------------
    ##
    ## forest weight or local method
    ##
    ## ------------------------------------
    else {
      ## pull grow y, grow residual, and yhat (either grow/predicted)
      if (ncol(cbind(object$yvar.grow)) > 1) {
        y <- object$yvar.grow[, yn]
        res <- object$res.grow[, yn]
        if (oob && !is.null(object$regrOutput[[yn]]$predicted.oob)) {
          yhat <- object$regrOutput[[yn]]$predicted.oob
        }
        else {
          yhat <- object$regrOutput[[yn]]$predicted
        }
      }
      else {
        y <- object$yvar.grow
        res <- object$res.grow
         if (oob && !is.null(object$predicted.oob)) {
          yhat <- object$predicted.oob
        }
        else {
          yhat <- object$predicted
        }
      }
      ## extract unique y values
      yunq <- sort(unique(y))
      ## trim atoms for density/cdf (useful for big data otherwise O(n^2) memory)
      if (length(yunq) > maxn) {
        yunq <- yunq[unique(round(seq(1, length(yunq), length.out = maxn)))]
      }
      ## fast CDF construction
      ## ------------------------------------------------------------
      if (method == "forest") {
        ord <- order(res)
        res_sorted <- res[ord]
        n_res <- length(res_sorted)
        n_pred <- length(yhat)
        K <- length(yunq)
        cdf <- matrix(0, nrow = n_pred, ncol = K)
        cw0 <- numeric(n_res + 1L)
        for (i in seq_len(n_pred)) {
          cw0[-1L] <- cumsum(object$forest.wt[i, ord])
          pos <- findInterval(yunq - yhat[i], res_sorted)
          cdf[i, ] <- cw0[pos + 1L]
        }
      } else {
        ## locally adjusted cdf 
        res_sorted <- sort(res)
        n_res <- length(res_sorted)
        n_pred <- length(yhat)
        K <- length(yunq)
        cdf <- matrix(0, nrow = n_pred, ncol = K)
        for (i in seq_len(n_pred)) {
          pos <- findInterval(yunq - yhat[i], res_sorted)
          cdf[i, ] <- pos / n_res
        }
      }
      ## quantiles
      quant <- t(apply(cdf, 1, function(pr) {
        c(min(yunq, na.rm = TRUE), yunq)[1 + sIndex(pr, prob)]
      }))
      ## density
      density <- t(apply(cbind(0, cdf), 1, diff))
    }
    ## ------------------------------------
    ##
    ## ends loop for specific yvar
    ## return various quantities as list
    ##
    ## ------------------------------------
    list(quantiles = quant,
         prob = prob,
         cdf = cdf,
         density = density,
         yunq = yunq)
  })
  ## ------------------------------------
  ##
  ## finished -- return final goodies
  ##
  ## ------------------------------------
  if (object$family == "regr") {
    rO <- rO[[1]]
  }
  else {
    names(rO) <- ynames
  }
  object$quantreg <- rO
  class(object)[4] <- "quantreg"
  object
}
quantreg <- quantreg.rfsrc

### AUC workhorse
get.auc.workhorse <- function(roc.data) {
  x <- roc.data[, 1][roc.data[, 2] == 1]
  y <- roc.data[, 1][roc.data[, 2] == 0]
  if (length(x) > 1 & length(y) > 1) {
    AUC  <- tryCatch({wilcox.test(x, y, exact=F)$stat/(length(x)*length(y))}, error=function(ex){NA})
  }
  else {
    AUC <- NA
  }
  AUC
}
### Multiclass AUC -- Hand & Till (2001) definition
#' @name get.auc
#' @export
get.auc <- function(y, prob) {
  if (is.factor(y)) {
    y.uniq <- levels(y)
  }
  else {
    y.uniq <- sort(unique(y))
  }
  nclass <- length(y.uniq)
  AUC <- NULL
  for (i in 1:(nclass - 1)) {
    for (j in (i + 1):nclass) {
      pt.ij <- (y == y.uniq[i] | y == y.uniq[j])
      if (sum(pt.ij) > 1) {
        y.ij <- y[pt.ij]
        pij <- prob[pt.ij, j]
        pji <- prob[pt.ij, i]
        Aij <-  get.auc.workhorse(cbind(pij, 1 * (y.ij == y.uniq[j])))
        Aji <-  get.auc.workhorse(cbind(pji, 1 * (y.ij == y.uniq[i])))
        AUC <- c(AUC, (Aij + Aji)/2)
      }
    } 
  }
  if (is.null(AUC)) {
    NA
  }
  else {
    mean(AUC, na.rm = TRUE)
  }
}                          
## bayes rule 
get.bayes.rule <- function(prob, class.relfrq = NULL) {
  class.labels <- colnames(prob)
  if (is.null(class.relfrq)) {
    factor(class.labels[apply(prob, 1, function(x) {
      if (!all(is.na(x))) {
        resample(which(x == max(x, na.rm = TRUE)), 1)
      }
      else {
        NA
      }
    })], levels = class.labels)
  }
  ## added to handle the rfq classifier
  else {
    minority <- which.min(class.relfrq)
    majority <- setdiff(1:2, minority)      
    rfq.rule <- rep(majority, nrow(prob))
    rfq.rule[prob[, minority] >= min(class.relfrq, na.rm = TRUE)] <- minority
    factor(class.labels[rfq.rule], levels = class.labels)
  }
}
## normalized brier (normalized to one for strawman coin toss)
get.brier.error <- function(y, prob, normalized = TRUE, vector = FALSE) {
  if (is.null(colnames(prob))) {
    colnames(prob) <- levels(y)
  }
  cl <- colnames(prob)
  J <- length(cl)
  bs <- rep(NA, J)
  nullO <- sapply(1:J, function(j) {
    bs[j] <<- mean((1 * (y == cl[j]) - prob[, j]) ^ 2, na.rm = TRUE)
    NULL
  })
  if (normalized) {##normalized to 100
    norm.const <- (J / (J - 1))
  }
  else {##standard, normalized to 0.25
    norm.const <- (1 / J)
  }
  ## return the summary value 
  if (!vector) {
    sum(bs * norm.const, na.rm = TRUE)
  }
  ## return the vector of brier values
  else {
    bs * norm.const
  }
}
## get confusion matrix
get.confusion <- function(y, class.or.prob) {
  ## response or probability?
  if (is.factor(class.or.prob)) {
    confusion <- table(y, class.or.prob)
  }
  else {
    if (is.null(colnames(class.or.prob))) {
      colnames(class.or.prob) <- levels(y)
    }
    confusion <- table(y, get.bayes.rule(class.or.prob))
  }
  class.error <- 1 - diag(confusion) / rowSums(confusion, na.rm = TRUE)
  cbind(confusion, class.error = round(class.error, 4))
}
## cindex - extended to CR + uno/fenwick
get.cindex <- function(time, censoring, predicted, weight, fast, do.trace = FALSE) {
  size <- length(time)
  if (size != length(censoring)) {
    stop("time, censoring, and predicted must have the same length")
  }
  ## determine competing risks (CR) vs right-censoring
  isCR <- any(censoring > 1, na.rm = TRUE)
  ## weight is optional
  if (missing(weight)) {
    weight <- NULL
  } else {
    if (length(weight) != size) {
      stop("weight must have the same length as time")
    }
  }
  ## fast switch is dynamic - now set by the native library
  if (missing(fast)) {
    fast <- -1
  }
  ## -----------------------------
  ## Right-censoring
  ## -----------------------------
  if (!isCR) {
    ## accept predicted as a vector or a 1-column matrix/data.frame
    if (is.matrix(predicted) || is.data.frame(predicted)) {
      if (NROW(predicted) != size) {
        stop("predicted must have the same length as time")
      }
      if (NCOL(predicted) != 1) {
        stop("predicted must be a vector for right-censoring")
      }
      predicted <- predicted[, 1]
    }
    if (length(predicted) != size) {
      stop("time, censoring, and predicted must have the same length")
    }
    miss <- is.na(time) | is.na(censoring) | is.na(predicted)
    if (!is.null(weight)) {
      miss <- miss | is.na(weight)
    }
    if (sum(miss) == size) {
      stop("no valid pairs found, too much missing data")
    }
    ## Flag missing members so we can exclude them in the pairs.
    denom <- as.double(!miss)
    nativeOutput <- .Call("rfsrcCIndex",
                          as.integer(do.trace),
                          as.integer(fast),
                          as.integer(size),
                          as.double(time),
                          as.double(censoring),
                          as.double(predicted),
                          as.double(denom),
                          if (is.null(weight)) NULL else as.double(weight))
    if (is.null(nativeOutput)) {
      stop("An error has occurred in rfsrcCIndex.  Please turn trace on for further analysis.")
    }
    return(nativeOutput$err)
  }
  ## -----------------------------
  ## Competing Risks (CR path)
  ## -----------------------------
  ## Convert predicted into an n x J matrix (one column per event type).
  ## Allow: matrix/data.frame, or list of J vectors.
  baseMiss <- is.na(time) | is.na(censoring)
  if (!is.null(weight)) {
    baseMiss <- baseMiss | is.na(weight)
  }
  if (all(baseMiss)) {
    stop("no valid pairs found, too much missing data")
  }
  J <- max(censoring[!baseMiss], na.rm = TRUE)
  if (is.matrix(predicted) || is.data.frame(predicted)) {
    predMat <- as.matrix(predicted)
  } else if (is.list(predicted) && is.null(dim(predicted))) {
    if (length(predicted) < J) {
      stop("for competing risks, predicted must have at least J event-specific prediction vectors")
    }
    predMat <- do.call(cbind, lapply(predicted[1:J], as.double))
  } else {
    stop("for competing risks, predicted must be a matrix/data.frame with one column per event type, or a list of event-specific prediction vectors")
  }
  if (NROW(predMat) != size) {
    stop("for competing risks, predicted must have nrow == length(time)")
  }
  if (NCOL(predMat) < J) {
    stop("for competing risks, predicted must have at least one column per event type")
  }
  predMat <- predMat[, seq_len(J), drop = FALSE]
  err <- rep(NA_real_, J)
  ## (A) CR + weights: IPCW Fenwick, event-by-event
  if (!is.null(weight)) {
    for (j in seq_len(J)) {
      miss.j <- baseMiss | is.na(predMat[, j])
      denom.j <- as.double(!miss.j)
      nativeOutput <- .Call("rfsrcCIndexFenwick",
                            as.integer(do.trace),
                            as.integer(j),
                            as.integer(size),
                            as.double(time),
                            as.double(censoring),
                            as.double(predMat[, j]),
                            as.double(denom.j),
                            as.double(weight))
      if (is.null(nativeOutput)) {
        stop("An error has occurred in rfsrcCIndexFenwick.  Please turn trace on for further analysis.")
      }
      err[j] <- nativeOutput$err
    }
    return(err)
  }
  ## (B) CR + no weights: legacy conditional approach
  ##     subset to {0, j} and call right-censored concordance
  for (j in seq_len(J)) {
    keep <- !baseMiss & !is.na(predMat[, j])
    idx  <- which(keep & (censoring == 0 | censoring == j))
    if (length(idx) < 2) {
      err[j] <- NA_real_
      next
    }
    denom.sub <- rep(1, length(idx))
    nativeOutput <- .Call("rfsrcCIndex",
                          as.integer(do.trace),
                          as.integer(fast),  
                          as.integer(length(idx)),
                          as.double(time[idx]),
                          as.double(censoring[idx]),      # values are 0 or j (j>0 => event)
                          as.double(predMat[idx, j]),
                          as.double(denom.sub),
                          NULL)
    if (is.null(nativeOutput)) {
      stop("An error has occurred in rfsrcCIndex.  Please turn trace on for further analysis.")
    }
    err[j] <- nativeOutput$err
  }
  return(err)
}
## gmean rule
get.gmean.rule <- function(y, prob) {
  ## determine frequencies: exit if this is not a two-class problem
  frq <- table(y)  
  if (length(frq) > 2) {
    return(NULL)
  }
  ## call bayes rule using class relative frequencies 
  get.bayes.rule(prob, frq / sum(frq, na.rm = TRUE))
}
## gmean for imbalanced classification
get.gmean <- function(y, prob, rfq = FALSE, robust = FALSE) {
  ## determine frequencies: exit if this is not a two-class problem
  frq <- table(y)  
  if (length(frq) > 2) {
    return(NULL)
  }
  ## determine threshold
  if (rfq) {
    threshold <- min(frq, na.rm = TRUE) / sum(frq, na.rm = TRUE)
  }
  else {
    threshold <- 0.5
  }
  ## convert yhat to a class label: 0 = majority, 1 = minority
  minority <- which.min(frq)
  majority <- setdiff(1:2, minority)
  yhat <- factor(1 * (prob[, minority] >= threshold), levels = c(0,1))
  ## compute the confusion matrix and extract TN,TP etc.
  confusion.matrix <- table(y, yhat)
  if (nrow(confusion.matrix) > 1) {##check that dimension is correct
    ## calculate the various rates
    TN <- confusion.matrix[minority, 2]
    FP <- confusion.matrix[minority, 1]
    FN <- confusion.matrix[majority, 2]
    TP <- confusion.matrix[majority, 1]
    ## assemble the sensitivity/specificity values
    if (robust) {##modified with 1 added to diagonals
      sensitivity <- (1 + TP) / (1 + TP + FN)
      specificity <- (1 + TN) / (1 + TN + FP)
    }
    else {
      sensitivity <- TP / (TP + FN)
      specificity <- TN / (TN + FP)
    }
    ## return the g mean
    sqrt(sensitivity * specificity)
  }
  else {
    NA
  }
}
## misclassification error
get.misclass.error <- function(y, yhat) {
  cl <- sort(unique(y))
  err <- rep(NA, length(cl))
  for (k in 1:length(cl)) {
    cl.pt  <- (y == cl[k])
    if (sum(cl.pt) > 0) {
        err[k] <- mean(y[cl.pt] != yhat[cl.pt])
    }
  }
  err
}
## cross-entropy, log-loss
get.logloss <- function(y, prob, robust = TRUE) {
  lgl <- lapply(levels(y), function(yn) {
    phat <- prob[, yn]
    pt <- y == yn
    if (sum(pt) > 0) {
      -log(phat[pt])
    }
    else {
      0
    }
  })
  if (robust) {
    lgl <- unlist(lgl[lengths(lgl) != 0])
    mean(lgl[!is.infinite(lgl)], na.rm = TRUE)
  }
  else {
    mean(unlist(lgl[lengths(lgl) != 0]), na.rm = TRUE)
  }
}                          

get.block.size.bits <- function (block.size, ntree) {
  ## Check for user silliness.
  if (!is.null(block.size)) {
    ## for backwards compatibility allow TRUE/FALSE
    if (is.logical(block.size)) {
      if (block.size) {
        block.size <- 1
      }
      else {
        block.size <- ntree
      }
    }
    else if ((block.size < 1) || (block.size > ntree)) {
      block.size <- ntree
    }
    else {
      block.size <- round(block.size)
    }
  }
  else {
    block.size <- ntree
  }
  return (block.size)
}
get.bootstrap.bits <- function (bootstrap) {
  if (bootstrap == "by.root") {
    bootstrap <- 0
  }
  else if (bootstrap == "by.node") {
    bootstrap <- 2^19
  }
  else if (bootstrap == "none") {
    bootstrap <- 2^20
  }
  else if (bootstrap == "by.user") {
    bootstrap <- 2^19 + 2^20
  }
  else {
    stop("Invalid choice for 'bootstrap' option:  ", bootstrap)
  }
  return (bootstrap)
}
get.case.depth.bits <- function (case.depth) {
  if (!is.null(case.depth)) {
    if (case.depth == TRUE) {
      case.depth <- 2^11
    }
    else if (case.depth == FALSE) {
      case.depth <- 0
    }
    else {
      stop("Invalid choice for 'case.depth' option:  ", case.depth)
    }
  }
  else {
    stop("Invalid choice for 'case.depth' option:  ", case.depth)
  }
  return (case.depth)
}
get.cr.bits <- function (fmly) {
  if (fmly == "surv-CR") {
    return(2^21)
  } else {
    return(0)
  }
}
get.cse.bits <- function (cse) {
  if (!is.null(cse)) {
    if (cse == TRUE) {
      cse <- 2^28
    }
    else if (cse == FALSE) {
      cse <- 0
    }
    else {
      stop("Invalid choice for 'cse' option:  ", cse)
    }
  }
  else {
    stop("Invalid choice for 'cse' option:  ", cse)
  }
  return (cse)
}
get.csv.bits <- function (csv) {
  if (!is.null(csv)) {
    if (csv == TRUE) {
      csv <- 2^29
    }
    else if (csv == FALSE) {
      csv <- 0
    }
    else {
      stop("Invalid choice for 'csv' option:  ", csv)
    }
  }
  else {
    stop("Invalid choice for 'csv' option:  ", csv)
  }
  return (csv)
}
get.data.pass.bits <- function (data.pass) {
  if (!is.null(data.pass)) {
    if (data.pass == TRUE) {
      data.pass <- 2^15
    }
    else if (data.pass == FALSE) {
      data.pass <- 0
    }
    else {
      stop("Invalid choice for 'data.pass' option:  ", data.pass)
    }
  }
  else {
    stop("Invalid choice for 'data.pass' option:  ", data.pass)
  }
  return (data.pass)
}
get.data.pass.predict.bits <- function (data.pass) {
  if (!is.null(data.pass)) {
    if (data.pass == TRUE) {
      data.pass <- 2^27
    }
    else if (data.pass == FALSE) {
      data.pass <- 0
    }
    else {
      stop("Invalid choice for 'data.pass' option:  ", data.pass)
    }
  }
  else {
    stop("Invalid choice for 'data.pass' option:  ", data.pass)
  }
  return (data.pass)
}
get.distance.bits <- function (grow.equivalent, distance) {
  ## Convert distance option into native code parameter.
  if (!is.null(distance)) {
    if (distance == FALSE) {
      dist.bits <- 0
    }
    else if (grow.equivalent == TRUE) {
      if (distance == TRUE) {
        ## inbag is the default
        dist.bits <- 2^20 + 2^21
      }
      else if (distance == "inbag") {
        dist.bits <- 2^20 + 2^21
      }
      else if (distance == "oob") {
        dist.bits <- 2^20 + 2^22
      }
      else if (distance == "all") {
        dist.bits <- 2^20 + 2^21 + 2^22
      }
      else {
        stop("Invalid choice for 'distance' option:  ", distance)
      }
    }
    else if (grow.equivalent == FALSE) {
      if (distance == TRUE) {
        ## We equate oob with test data.
        dist.bits <- 2^20 + 2^22
      }
      else if (distance == "all") {
        ## We equate oob with test data.
        dist.bits <- 2^20 + 2^22
      }
      else if (distance == "hybrid") {
        ## We equate oob with test data.
        dist.bits <- 2^20 + 2^21 + 2^22
      }
      else {
        stop("Invalid choice for 'distance' option:  ", distance)
      }
    }
    else {
      stop("Invalid choice for 'grow.equivalent' in distance:  ", grow.equivalent)
    }
  }
  else {
    stop("Invalid choice for 'distance' option:  ", distance)
  }
  return (dist.bits)
}
## convert ensemble option into native code parameter.
get.ensemble.bits <- function (ensemble) {
  if (ensemble == "oob") {
    ensemble <- 2^1
  }
  else if (ensemble == "inbag") {
    ensemble <- 2^0
  }
  else if (ensemble == "all") {
    ensemble <- 2^0 + 2^1
  }    
  else {
    stop("Invalid choice for 'ensemble' option:  ", ensemble)
  }
  return (ensemble)
}
get.experimental.bits <- function (experimental) {
  ## Convert experimental option into native code parameter.
  bits <- 0
  for (i in 1:length(experimental)) {
    if (experimental[i] == TRUE) {
      bits <- bits + 2^(i-1)
    }
  }
  return (bits)
}
get.forest.bits <- function (forest) {
  ## Convert forest option into native code parameter.
  if (!is.null(forest)) {
    if (forest == TRUE) {
      forest <- 2^5
    }
    else if (forest == FALSE) {
      forest <- 0
    }
    else {
      stop("Invalid choice for 'forest' option:  ", forest)
    }
  }
  else {
    stop("Invalid choice for 'forest' option:  ", forest)
  }
  return (forest)
}
get.forest.wt.bits <- function (grow.equivalent, bootstrap, weight) {
  ## Convert weight option into native code parameter.
  if (!is.null(weight)) {
    if (weight == FALSE) {
      wght.bits <- 0
    }
    else if (grow.equivalent == TRUE) {
      if (weight == TRUE) {
        wght.bits <- 2^0 + 2^1
      }
      else if (weight == "inbag") {
        wght.bits <- 2^0 + 2^1
      }
      else if (weight == "oob") {
        wght.bits <- 2^0 + 2^2
      }
      else if (weight == "all") {
        wght.bits <- 2^0 + 2^1 + 2^2
      }
      else {
        stop("Invalid choice for 'weight' option:  ", weight)
      }
    }
    else if (grow.equivalent == FALSE) {
      if (weight == TRUE) {
        wght.bits <- 2^0 + 2^1 + 2^2
      }
      else if (weight == "all") {
        wght.bits <- 2^0 + 2^1 + 2^2
      }
      else {
        stop("Invalid choice for 'weight' option:  ", weight)
      }
    }
    else {
      stop("Invalid choice for 'grow.equivalent' in weight:  ", grow.equivalent)
    }
  }
  else {
    stop("Invalid choice for 'weight' option:  ", weight)
  }
  return (wght.bits)
}
get.gk.quantile <- function(gk.quantile) {
  if (is.null(gk.quantile)) {
    gk.quantile <- FALSE
  }
  gk.quantile
}
get.gk.quantile.bits <-  function (gk.quantile) {
  if (gk.quantile) {
    return (2^24)
  }
  else {
    return (0)
  }
}
## This has been deprecated!
get.insitu.ensemble.bits <- function (insitu.ensemble) {
  ## Convert insitu.ensemble option into native code parameter.
  if (!is.null(insitu.ensemble)) {
    if (insitu.ensemble == TRUE) {
      insitu.ensemble <- 2^11
    }
    else if (insitu.ensemble == FALSE) {
      insitu.ensemble <- 0
    }
    else {
      stop("Invalid choice for 'insitu.ensemble' option:  ", insitu.ensemble)
    }
  }
  else {
    stop("Invalid choice for 'insitu.ensemble' option:  ", insitu.ensemble)
  }
  return (insitu.ensemble)
}
get.importance <-  function(importance, perf.type = NULL) {
  ## convert importance option into native code parameter
  if (!is.null(importance)) {
    ## use permute with gmean as default
    if (!is.null(perf.type) && perf.type == "gmean" && as.character(importance) == TRUE) {
      importance <- "permute"
    }
    ## default vimp is now anti
    if (importance == TRUE) {
      importance <- "anti"
    }
    else if (importance == FALSE) {
      importance <- "none"
    }
  }
  importance
}
get.importance.bits <-  function(importance, perf.type = NULL) {
  ## convert importance option into native code parameter
  if (!is.null(importance)) {
    importance <- get.importance(importance, perf.type)
    if (importance == "none") {
      importance <- 0
    }
    else if (importance == "anti") {
      importance <- 2^25 + 0
    }
    else if (importance == "permute") {
      importance <- 2^25 + 2^8
    }
    else if (importance == "random") {
      importance <- 2^25 + 2^9
    }
    else if (importance == "anti.joint") {
      importance <- 2^25 + 2^10 + 0
    }
    else if (importance == "permute.joint") {
      importance <- 2^25 + 2^10 + 2^8
    }
    else if (importance == "random.joint") {
      importance <- 2^25 + 2^10 + 2^9
    }
    else {
      stop("Invalid choice for 'importance' option:  ", importance)
    }
  }
  else {
    stop("Invalid choice for 'importance' option:  ", importance)
  }
  return (importance)
}
get.impute.only.bits <-  function (impute.only, nMiss) {
  if (impute.only) {
    if (nMiss > 0) {
      return (2^16)
    }
    else {
      stop("Data has no missing values, using 'impute' makes no sense.")
    }
  }
  else {
    return (0)
  }
}
## Convert jitt option into native code parameter.
get.jitt.bits <- function (jitt) {
  if (!is.null(jitt)) {
    if (jitt == TRUE) {
      jitt <- 2^23
    }
    else if (jitt == FALSE) {
      jitt <- 0
    }
    else {
      stop("Invalid choice for 'jitt' option:  ", jitt)
    }
  }
  else {
    stop("Invalid choice for 'jitt' option:  ", jitt)
  }
  return (jitt)
}
get.membership.bits <- function (membership) {
  ## Convert option into native code parameter.
  bits <- 0
  if (!is.null(membership)) {
    if (membership == TRUE) {
      bits <- 2^6
    }
    else if (membership != FALSE) {
      stop("Invalid choice for 'membership' option:  ", membership)
    }
  }
  else {
    stop("Invalid choice for 'membership' option:  ", membership)
  }
  return (bits)
}
get.na.action.bits <- function (na.action) {
  if (na.action == "na.omit") {
    ## This is the high byte!
    na.action <- 0
  }
  else if (na.action == "na.impute" || na.action == "na.random") {
    ## This is the high byte!
    na.action <- 2^4
    ## To recover the original functionality in which the split
    ## statistic uses missing in-node imputed values, uncomment 
    ## the following statement:
    ## na.action <- 0
  }
  else {
    stop("Invalid choice for 'na.action' option:  ", na.action)
  }
  return (na.action)
}
get.outcome.bits <- function (outcome) {
  ## Convert outcome option into native code parameter.
  if (outcome == "train") {
    outcome <- 0
  }
  else if (outcome == "test") {
    outcome <- 2^17
  }
  else {
    stop("Invalid choice for 'outcome' option:  ", outcome)
  }
  return (outcome)
}
get.perf <-  function (perf, impute.only, family) {
  ## first deal with impute.only where there is no performance
  if (impute.only == TRUE) {
    return("none")
  }
  ## now deal with non-classification
  if (family != "class") {
    if (is.null(perf)) {
      return("default")
    }
    perf <- match.arg(perf, c("none", "default", "standard"))##only allowed values
    if (perf == "standard") {
      perf <- "default"
    }
    return(perf)
  }
  ## now deal with classification
  if (is.null(perf)) {
    return("default")
  }
  perf <- match.arg(perf, c("none", "default", "standard", "misclass", "brier", "gmean", "g.mean"))
  if (perf == "standard" || perf == "misclass") {
    perf <- "default"
  }
  if (perf == "g.mean") {
    perf <- "gmean"
  }
  perf
}
get.perf.bits <- function (perf) {
  if (perf == "default") {
    return (2^2)
  }
  else if (perf == "gmean" || perf == "g.mean") {
    return (2^2 + 2^14)
  }
  else if (perf == "brier") {
    return (2^2 + 2^3)
  }
  else {#everything else becomes "none"
    return (0)
  }
}
get.proximity.bits <- function (grow.equivalent, proximity) {
  ## Convert proximity option into native code parameter.
  if (!is.null(proximity)) {
    if (proximity == FALSE) {
      prox.bits <- 0
    }
    else if (grow.equivalent == TRUE) {
      if (proximity == TRUE) {
          ## inbag is the default
          prox.bits <- 2^28 + 2^29
      }
      else if (proximity == "inbag") {
        prox.bits <- 2^28 + 2^29
      }
      else if (proximity == "oob") {
        prox.bits <- 2^28 + 2^30
      }
      else if (proximity == "all") {
        prox.bits <- 2^28 + 2^29 + 2^30
      }
      else {
        stop("Invalid choice for 'proximity' option:  ", proximity)
      }
    }
    else if (grow.equivalent == FALSE) {
      if (proximity == TRUE) {
        ## We equate oob with test data.
          prox.bits <- 2^28 + 2^30
      }
      else if (proximity == "all") {
        ## We equate oob with test data.
        prox.bits <- 2^28 + 2^30
      }
      else if (proximity == "hybrid") {
        ## We equate oob with test data.
        prox.bits <- 2^28 + 2^29 + 2^30
      }
      else {
        stop("Invalid choice for 'proximity' option:  ", proximity)
      }
    }
    else {
      stop("Invalid choice for 'grow.equivalent' in proximity:  ", grow.equivalent)
    }
  }
  else {
    stop("Invalid choice for 'proximity' option:  ", proximity)
  }
  return (prox.bits)
}
get.rfq <- function(rfq) {
  if (is.null(rfq)) {
    rfq <- FALSE
  }
  rfq
}
get.rfq.bits <- function (rfq, family) {
  result <- 0
  if (family == "class") {
    if (rfq) {
      result <- 2^15
    }
  }
  return (result)
}
get.rf.cores <- function () {
    ## PART I:  Two ways for the user to specify cores:
    ## (1) R-option "rf.cores"
    ## (2) Shell-environment-option "RF_CORES"
    if (is.null(getOption("rf.cores", NULL))) {
        if (!is.na(as.numeric(Sys.getenv("RF_CORES")))) {
            options(rf.cores = as.integer(Sys.getenv("RF_CORES")))
        }
    }
    ## If the user has set the cores using either of the two methods, we respect it.
    if (!is.null(getOption("rf.cores", NULL))) {
        return (getOption("rf.cores"))
    }
    ## PART II:  Respect R CMD check limit
    chk <- tolower(Sys.getenv("_R_CHECK_LIMIT_CORES_", ""))
    if (nzchar(chk) && chk != "false") {
        ## under R CMD check --as-cran (CRAN sets this)
        return(2L)
    }
    ## PART III:  Use everything.
    return (-1L)
}
## convert samptype option into native code parameter.
get.samptype.bits <- function (samptype) {
  if (samptype == "swr") {
    bits <- 0
  }
  else if (samptype == "swor") {
    bits <- 2^12
  }
  else {
    stop("Invalid choice for 'samptype' option:  ", samptype)
  }
  return (bits)
}
get.seed <- function (seed) {
  if ((is.null(seed)) || (abs(seed) < 1)) {
    seed <- runif(1,1,1e6)
  }
  seed <- -round(abs(seed))
  return (seed)
}
get.split.cust.bits <- function (split.cust) {
  ## Convert split.cust option into native code parameter.
  if (!is.null(split.cust)) {
    if ((split.cust >= 1) && (split.cust <= 16)) {
      ## Bit shift eight left.
      split.cust <- 256 * (split.cust - 1)
    }
    else {
      stop("Invalid choice for 'split.cust' option:  ", split.cust)
    }
  }
  else {
    split.cust <- 0
  }
  return (split.cust)
}
get.split.depth.bits <- function (split.depth) {
  ## Convert split.depth option into native code parameter.
  if (!is.null(split.depth)) {
    if (split.depth == "all.trees") {
      split.depth <- 2^22
    }
    else if (split.depth == "by.tree") {
      split.depth <- 2^23
    }
    else if (split.depth == FALSE) {
      split.depth <- 0
    }
    else {
      stop("Invalid choice for 'split.depth' option:  ", split.depth)
    }
  }
  else {
    stop("Invalid choice for 'split.depth' option:  ", split.depth)
  }
  return (split.depth)
}
get.split.null.bits <- function (split.null) {
  ## Convert split.null option into native code parameter.
  if (!is.null(split.null)) {
    if (split.null == TRUE) {
      split.null <- 2^18
    }
    else if (split.null == FALSE) {
      split.null <- 0
    }
    else {
      stop("Invalid choice for 'split.null' option:  ", split.null)
    }
  }
  else {
    stop("Invalid choice for 'split.null' option:  ", split.null)
  }
  return (split.null)
}
get.tdc.rule.bits <- function (tdc.rule) {
  if (is.null(tdc.rule)) {
    ## Allow splits on everything.
    result <- 2^24 + 2^25 + 2^26
  }
  else if (tdc.rule == "all") {
    ## Allow splits on everything.
    result <- 2^24 + 2^25 + 2^26
  }
  else if (tdc.rule == "time") {
    ## Allow splits only on time.
    result <- 2^24
  }
  else if (tdc.rule == "tdx") {
    ## Allow splits on time dependent x-variables only, and not on time.
    result <- 2^25
  }
  else if (tdc.rule == "tsx") {
    ## Allow splits on time static x-variables only, and not on time.
    result <- 2^26
  }
  else if (tdc.rule == "time.and.tsx") {
    ## Allow splits on time and time static x-variables only.
    result <- 2^24 + 2^26
  }
  else if (tdc.rule == "time.and.tdx") {
    ## Allow splits on time and time dependent x-variables only.
    result <- 2^24 + 2^25
  }
  return (result)
}
get.terminal.qualts.bits <- function(terminal.qualts) {
  bits <- 0
  if (!is.null(terminal.qualts)) {
    if (terminal.qualts == TRUE) {
      bits <- bits + 2^16
    }
    else if (terminal.qualts != FALSE) {
      stop("Invalid choice for 'terminal.qualts' option:  ", terminal.qualts)
    }
  }
  else {
    stop("Invalid choice for 'terminal.qualts' option:  ", terminal.qualts)
  }
  return(bits)
}
get.terminal.qualts.predict.bits <- function(incoming.flag) {
  bits <- 2^16
  if (!is.null(incoming.flag)) {
    if (incoming.flag == TRUE) {
      bits <- bits + 2^17
    }
  }
  else {
    stop("Invalid choice for 'incoming.flag':  ", incoming.flag)
  }
  return(bits)
}
get.terminal.quants.bits <- function(terminal.quants) {
  bits <- 0
  if (!is.null(terminal.quants)) {
    if (terminal.quants == TRUE) {
      bits <- bits + 2^18
    }
    else if (terminal.quants != FALSE) {
      stop("Invalid choice for 'terminal.quants' option:  ", terminal.quants)
    }
  }
  else {
    stop("Invalid choice for 'terminal.quants' option:  ", terminal.quants)
  }
  return(bits)
}
get.terminal.quants.predict.bits <- function(incoming.flag) {
  bits <- 2^18
  if (!is.null(incoming.flag)) {
    if (incoming.flag == TRUE) {
      bits <- bits + 2^19
    }
  }
  else {
    stop("Invalid choice for 'incoming.flag':  ", incoming.flag)
  }
  return(bits)
}
get.trace.bits <- function (do.trace) {
  ## Convert trace into native code parameter.
  if (!is.logical(do.trace)) {
    if (do.trace >= 1) {
      do.trace <- round(do.trace)
    }
    else {
      do.trace <- 0
    }
  }
  else {
    do.trace <- 1 * do.trace
  }
  return (do.trace)
}
get.tree.index <- function(get.tree, ntree) {
  ## NULL --> default setting
  if (is.null(get.tree)) {
    rep(1, ntree)
  }
  ## the user has specified a subset of trees
  else {
    pt <- get.tree >=1 & get.tree <= ntree
    if (sum(pt) > 0) {
      get.tree <- get.tree[pt]
      get.tree.temp <- rep(0, ntree)
      get.tree.temp[get.tree] <- 1
      get.tree.temp
    }
    else {
      rep(1, ntree)
    }
  }
}
get.var.used.bits <- function (var.used) {
  ## Convert var.used option into native code parameter.
  if (!is.null(var.used)) {
    if (var.used == "all.trees") {
      var.used <- 2^12
    }
    else if (var.used == "by.tree") {
      var.used <- 2^13
    }
    else if (var.used == FALSE) {
      var.used <- 0
    }
    else {
      stop("Invalid choice for 'var.used' option:  ", var.used)
    }
  }
  else {
    stop("Invalid choice for 'var.used' option:  ", var.used)
  }
  return (var.used)
}
get.vimp.only.bits <-  function (vimp.only) {
  ## Convert vimp.only option into native code parameter.
  if (!is.null(vimp.only)) {
    if (vimp.only) {
      return (2^27)
    }
    else if (!vimp.only) {
      return (0)
    }
    else {
      stop("Invalid choice for 'vimp.only' option:  ", vimp.only)
    }
  }
  else {
    stop("Invalid choice for 'vimp.only' option:  ", vimp.only)
  }
}
## Check for presence of forest
is.forest.missing <- function(object) {
  ## for backwards compatability
  if(is.null(object$forest$forest)) {
    is.null(object$forest)
  }
  ## current stealth build moving forward
  else {
    !object$forest$forest
  }
}
is.hidden.cse <-  function(user.option) {
  if (is.null(user.option$cse)) {
    FALSE
  }
  else {
    as.logical(as.character(user.option$cse))
  }
}
is.hidden.csv <-  function(user.option) {
  if (is.null(user.option$csv)) {
    FALSE
  }
  else {
    as.logical(as.character(user.option$csv))
  }
}
is.hidden.data.pass <-  function(user.option) {
  if (is.null(user.option$data.pass)) {
    FALSE
  }
  else {
    as.logical(as.character(user.option$data.pass))
  }
}
is.hidden.do.trace <-  function(user.option) {
  if (is.null(user.option$do.trace)) {
    FALSE
  }
  else {
    user.option$do.trace
  }
}
is.hidden.experimental <-  function(user.option) {
  if (is.null(user.option$experimental)) {
    e <- FALSE
  }
  else {
    e  <- as.logical(as.character(user.option$experimental))
  }
  return (e)
}
is.hidden.gk.quantile <-  function(user.option) {
  if (is.null(user.option$gk.quantile)) {
    NULL
  }
  else {
    as.logical(as.character(user.option$gk.quantile))
  }
}
is.hidden.holdout.array <-  function(user.option) {
  if (is.null(user.option$holdout.array)) {
    NULL
  }
  else {
    user.option$holdout.array
  }
}
is.hidden.holdout.specs <-  function(user.option) {
  ## Default value is NULL
  if (is.null(user.option$holdout.specs)) {
    obj <- NULL
  }
  else {
    obj <- user.option$holdout.specs
  }
  return (obj)
}
is.hidden.impute.only <-  function(user.option) {
  if (is.null(user.option$impute.only)) {
    FALSE
  }
  else {
    as.logical(as.character(user.option$impute.only))
  }
}
## This has been deprecated!
is.hidden.insitu.ensemble <-  function(user.option) {
  if (is.null(user.option$insitu.ensemble)) {
    ## From > 2.14.0 we now assert insitu.ensemble as the default. 
    TRUE
  }
  else {
    as.logical(as.character(user.option$insitu.ensemble))
  }
}
is.hidden.jitt <-  function (user.option, importance, na.action = "na.omit", anonymous = FALSE,  partial = FALSE) {
  ## no missing data
  ## anonymous = TRUE, jitt = TRUE,
  ## anonymous = TRUE, jitt = FALSE,
  ## anonymous = FALSE, jitt = TRUE,
  ## anonymous = FALSE, jitt = FALSE,
  ## missing data
  ## anonymous = TRUE, na.action=na.mean  =>>  no restriction on jitt
  ## anonymous = TRUE, jitt = TRUE   =>>  na.random
  ## anonymous = TRUE, jitt = FALSE  =>>  failure and exit
  ## anonymous = FALSE, jitt = TRUE  =>>  na.random
  ## anonymous = FALSE, jitt = FALSE =>>  na.impute (old-school)
  if (na.action == "na.random") {
    return(TRUE)
  }
  if (na.action == "na.impute" && !anonymous) {
    return(FALSE)
  }
  ## now that the danger is passed, proceed forward ...
  if (is.null(user.option$jitt)) {
    if (partial || importance == "permute" || importance == "permute.joint") {
      ##partial is generally a full restore
      ##permutation importance is synthetic, so default JITT = FALSE
      FALSE
    }
    else {
      TRUE
    }
  }
  else {
    as.logical(as.character(user.option$jitt))
  }
}
is.hidden.mahalanobis.sigma <- function(user.option) {
  if (is.null(user.option$mahalanobis.sigma) & is.null(user.option$sigma)) {
    NULL
  }
  else {
    if (!is.null(user.option$mahalanobis.sigma)) {
      ginverse(user.option$mahalanobis.sigma)
    }
    else {
      ginverse(user.option$sigma)
    }
  }
}
is.hidden.perf.type <-  function(user.option) {
  ## Default value is NULL
  if (is.null(user.option$perf.type)) {
    NULL
  }
  else {
    as.character(user.option$perf.type)
  }
}
is.hidden.prob <-  function(user.option) {
  if (is.null(user.option$prob)) {
    NULL
  }
  else {
    prob <- user.option$prob
    sort(prob[prob>0 & prob<1])
  }
}
is.hidden.prob.epsilon <-  function(user.option) {
  if (is.null(user.option$prob.epsilon)) {
    NULL
  }
  else {
    prob.epsilon <- user.option$prob.epsilon
    if ((prob.epsilon <= 0) || (prob.epsilon >= 0.50)) {
      stop("parameter 'prob.epsilon' must be in range (0, 1/2) :  ", prob.epsilon)
    }
    prob.epsilon
  }
}
is.hidden.quantile.regr <-  function(user.option) {
  if (is.null(user.option$quantile.regr)) {
    FALSE
  }
  else {
    as.logical(as.character(user.option$quantile.regr))
  }
}
is.hidden.rfq <-  function(user.option) {
  if (is.null(user.option$rfq)) {
    NULL
  }
  else {
    as.logical(as.character(user.option$rfq))
  }
}
is.hidden.tdc.rule <-  function(user.option) {
  if (is.null(user.option$tdc.rule)) {
    NULL
  }
  else {
    user.option$tdc.rule
  }
}
is.hidden.terminal.qualts <-  function(user.option) {
  ## From > 2.14.0 we now assert terminal qualts and quants as the default. 
  if (is.null(user.option$terminal.qualts)) {
    TRUE
  }
  else {
    as.logical(as.character(user.option$terminal.qualts))
  }
}
is.hidden.terminal.quants <-  function (user.option, save.memory = FALSE) {
  ## From > 2.14.0 we now assert terminal qualts and quants as the default - save.memory override though 
  if (is.null(user.option$terminal.quants)) {
    TRUE & !save.memory
  }
  else {
    as.logical(as.character(user.option$terminal.quants))
  }
}
is.hidden.vtry <-  function(user.option) {
  ## Default value is 0
  if (is.null(user.option$vtry)) {
    0
  }
  else {
    user.option$vtry
  }
}
is.hidden.ytry <-  function(user.option) {
  if (is.null(user.option$ytry)) {
    NULL
  }
  else {
    as.integer(user.option$ytry)
  }
}
is.hidden.vimp.threshold <-  function(user.option) {
  if (is.null(user.option$vimp.threshold)) {
    ## This is value from zero (0) to one (1) that represents the
    ## probability that variables will be noised up for random and
    ## anti vimp.  Thus zero means that vimp is effectively turned off.
    ## One means that the vimp protocol will be respected all the
    ## time.  If the threshold value alpha is such that 0 < alpha
    ## < 1, we draw a random value b. If b <= alpha, we respect
    ## the vimp protocol.
    1.0
  }
  else {
    user.option$vimp.threshold
  }
}

subsample.rfsrc <- function(obj,                
                            B = 100,
                            block.size = 1,
                            importance,
                            subratio = NULL,
                            stratify = TRUE,
                            performance = FALSE,
                            performance.only = FALSE,
                            joint = FALSE,
                            xvar.names = NULL,
                            bootstrap = FALSE,
                            verbose = TRUE) 
{
  ##--------------------------------------------------------------
  ##
  ## coherence checks
  ##
  ##--------------------------------------------------------------
  ## incoming object must be a grow forest
  if (sum(inherits(obj, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2) {
    stop("This function only works for objects of class `(rfsrc, grow)'")
  }
  ## grow forests must have true forest information
  if (sum(inherits(obj, c("rfsrc", "grow"), TRUE) == c(1, 2)) == 2) {
    if (is.forest.missing(obj)) {
      stop("Forest information for prediction is missing.  Re-run rfsrc (grow call) with forest=TRUE")
    }
  }
  if (inherits(obj, "anonymous")) {
    stop("this function does work with anonymous forests")
  }
  ##--------------------------------------------------------------
  ##
  ## is this a multivariate family?
  ##
  ##--------------------------------------------------------------
  fmly <- obj$family
  mv.fmly <- fmly == "regr+" || fmly == "class+" || fmly == "mix+" 
  ##--------------------------------------------------------------
  ##
  ## set the sample size
  ## set subratio (if not supplied)
  ## confirm subratio is appropriately set
  ##
  ##--------------------------------------------------------------
  n <- obj$n
  if (is.null(subratio)) {
    subratio <- get.subsample.subratio(n)
  }
  if (!bootstrap && (subratio < 0 || subratio > 1)) {
    stop("subratio must be between 0 and 1")
  }
  ##--------------------------------------------------------------
  ##
  ## extract random forest parmaters
  ##
  ##--------------------------------------------------------------
  ## list of parameters to be obtained from the grow object
  grow.prms <- c("call",
                 "ntree",
                 "mtry",
                 "nodesize",
                 "nodedepth",
                 "splitrule",
                 "nsplit",
                 "xvar.wt",
                 "split.wt",
                 "cause.wt",
                 "block.size")
  ## list of parameters to be obtained from the forest object
  forest.prms <- c("forest",
                   "bootstrap",
                   "sampsize",
                   "samptype",
                   "case.wt",
                   "perf.type",
                   "vimp.threshold",
                   "rfq")
  ## pull the parameters (grow first, then forest object)
  rf.prms <- c(obj[grow.prms], obj$forest[forest.prms])
  ## rename parameters to their correct random forest names
  rf.prms$call <- formula(rf.prms$call)
  names(rf.prms)[names(rf.prms) == "call"] <- "formula"
  names(rf.prms)[names(rf.prms) == "cause.wt"] <- "cause"
  ## subsamping cannot be applied when case weights are non-standard
  if (!all(diff(rf.prms$case.wt) == 0)) {
    stop("subsampling is not permitted on forests grown under non-standard case weights")
  }
  ## everthing is OK - nuke the case weights
  rf.prms$case.wt <- NULL
  ## need time.interest if this is a survival object
  if (fmly == "surv" || fmly == "surv-CR") {
    rf.prms$ntime <- obj$time.interest
  }
  ## nuke xvar and split weighting if they are standard
  if (all(diff(rf.prms$xvar.wt) == 0)) {
    rf.prms$xvar.wt <- NULL
  }
  if (all(diff(rf.prms$split.wt) == 0)) {
    rf.prms$split.wt <- NULL
  }
  ## nuke nodedepth if it's standard
  if (rf.prms$nodedepth <= 0) {
    rf.prms$nodedepth <- NULL
  }
  ## subsampling is not permitted if bootstrapping is not by root
  if (!bootstrap && rf.prms$bootstrap != "by.root") {
    stop("subsampling is not permitted on grow objects where sampling is not 'by.root'")
  }
  ## for subsampling we need to determine the subsampling tree sample size
  ## this is now handled by forest$sampsize which is a function
  ## for double bootstrapping we do NOT allow custom sampling for grow object
  if (bootstrap && rf.prms$bootstrap != "by.root") {
    stop("double bootstrapping is only permitted for breiman (bootstrap) forests")
  }
  else {
    ##everything is OK - set bootstrap parameter to NULL for later custom bootstrap
    ## leave $samptype and $sampsize in place to be used for making double bootstrap
    rf.prms$bootstrap <- NULL
  }
  ## use uno weights?
  rf.prms$use.uno <- !is.null(obj$forest$uno.weights)
  ##--------------------------------------------------------------
  ##
  ## performance only?  (user is requesting generalization error only)
  ##
  ##--------------------------------------------------------------
  if (performance.only) {
    performance <- TRUE
    importance <- "none"
    vmp <- NULL
  }
  ##--------------------------------------------------------------
  ##
  ## call the bootstrap subroutine if double bootstrap was requested
  ##
  ##--------------------------------------------------------------
  if (bootstrap) {
    if (missing(importance) && !performance.only) {
      importance <- TRUE
    }
    bootO <- bootsample(obj, rf.prms, B = B, block.size = block.size,
                        joint = joint, xvar.names = xvar.names,
                        importance = importance,
                        performance = performance, performance.only = performance.only,
                        verbose = verbose)
    rO <- list(rf = bootO$rf, vmp = bootO$vmp, vmpB = bootO$vmpB,
               subratio = NULL, performance.only = performance.only)
    class(rO) <- c(class(obj), "bootsample")
    return(rO)
  }
  ##--------------------------------------------------------------
  ##
  ## extract data from the previously grown forest
  ## set the dimension, pull the feature names
  ##
  ##--------------------------------------------------------------
  dta <- data.frame(obj$yvar, obj$xvar)
  colnames(dta)[1:length(obj$yvar.names)] <- obj$yvar.names
  ##--------------------------------------------------------------
  ##
  ## call vimp to extract importance if not available in original object
  ##
  ##--------------------------------------------------------------
  if (!performance.only) {
    vmp <- get.mv.vimp(obj, FALSE, FALSE)
    if (is.null(vmp)) {
      if (verbose) cat("no importance found: calculating it now ...\n")
      if (missing(importance)) {
        importance <- TRUE
      }
      obj <- vimp(obj,
                  perf.type = rf.prms$perf.type,
                  use.uno = rf.prms$use.uno,
                  block.size = block.size,
                  importance = importance)
      vmp <- get.mv.vimp(obj, FALSE, FALSE)
      rf.prms$block.size <- block.size
      if (verbose) cat("done\n")
    }
    else {
      importance <- obj$forest$importance
    }
  }
  ##--------------------------------------------------------------
  ##
  ## call joint vimp - reference vimp value for pure noise settings
  ##
  ##--------------------------------------------------------------
  if (joint && !performance.only) {
    vmp.joint <- get.mv.vimp(get.subsample.joint.vimp(obj, rf.prms, xvar.names), FALSE, FALSE)
    vmp <- vimp.subsample.combine(vmp, vmp.joint, "joint")
  }
  ##--------------------------------------------------------------
  ##
  ## obtain performance value - used for error rate confidence intervals
  ##
  ##--------------------------------------------------------------
  if (performance) {
    err <- get.mv.error(obj, FALSE, FALSE)
    vmp <- vimp.subsample.combine(vmp, err, "err")
  }
  ##----------------------------------------------------------
  ##
  ## subsampling loop for calculating VIMP confidence regions
  ##
  ##----------------------------------------------------------
  vmpS <- lapply(1:B, function(b) {
    ## progress bar
    if (verbose && B > 1) {
      custom.progress.bar(b, B)
    }
    ## draw the subsample
    ## use stratified sampling for classification/CR if requested (default)
    ## stratified sampling not allowed for mv-families
    if (!mv.fmly && stratify) {
      if (fmly == "class") {
        pt <- make.strat.sample(obj$yvar, subratio)
      }
      else if (fmly == "surv" || fmly == "surv-CR") {
        pt <- make.strat.sample(obj$yvar[, 2], subratio)
      }
      else {
        pt <- sample(1:n, size = (n * subratio), replace = FALSE)
      }
    }
    else {
      pt <- sample(1:n, size = (n * subratio), replace = FALSE)
    }
    ## calculate VMP on the subsampled data
    rf.b <-  do.call("rfsrc",
                     c(list(data = dta[pt,, drop = FALSE], importance = importance), rf.prms))
    vmp.b <- get.mv.vimp(rf.b, FALSE, FALSE)
    if (joint && !performance.only) {
      vmp.b.joint <- get.mv.vimp(get.subsample.joint.vimp(rf.b, rf.prms, xvar.names), FALSE, FALSE)
      vmp.b <- vimp.subsample.combine(vmp.b, vmp.b.joint, "joint")
    }
    if (performance) {
      err.b <- get.mv.error(rf.b, FALSE, FALSE)
      vmp.b <- vimp.subsample.combine(vmp.b, err.b, "err")
    }
    ## combine
    rO.b <- lapply(1:length(vmp), function(j) {
      vmp.j <- vmp[[j]]
      vmp.b.j <- vmp.b[[j]]
      vmp.mtx <- data.frame(matrix(NA, nrow(vmp.j), ncol(vmp.j)))
      rownames(vmp.mtx) <- rownames(vmp.j)
      colnames(vmp.mtx) <- colnames(vmp.j)
      if (ncol(vmp.b.j) == ncol(vmp.j)) {
        vmp.mtx[rownames(vmp.b.j), ] <- vmp.b.j
        vmp.b.j <- vmp.mtx
      }
      vmp.b.j
    })
    names(rO.b) <- names(vmp)
    rO.b
  })
  ##------------------------------------------------------------------
  ##
  ## return the goodies
  ##
  ##------------------------------------------------------------------
  rO <- list(rf = obj, vmp = vmp, vmpS = vmpS, subratio = subratio, performance.only = performance.only)
  class(rO) <- c(class(obj), "subsample")
  rO
}
subsample <- subsample.rfsrc

rfsrc.fast <- function(formula, data,
                       ntree = 500,
                       nsplit = 10,
                       bootstrap = "by.root",
                       sampsize = function(x){min(x * .632, max(150, x ^ (3/4)))},
                       samptype = "swor",
                       samp = NULL,
                       ntime = 50,
                       forest = FALSE,
                       save.memory = TRUE,
                       ...)
{
  ## --------------------------------------------------------------
  ##   
  ##   preliminary processing
  ##
  ## --------------------------------------------------------------
  ## verify key options
  if (!is.function(sampsize) && !is.numeric(sampsize)) {
    stop("sampsize must be a function or number specifying size of subsampled data")
  }
  ##--------------------------------------------------------------
  ##
  ## extract additional options specified by user
  ## we lock this down to allowed types
  ##
  ##--------------------------------------------------------------
  ## list of forest parameters
  rfnames <- get.rfnames(hidden = TRUE)
  ## restrict to allowed values
  rfnames <- rfnames[rfnames != "data"               &
                     rfnames != "ntree"              &
                     rfnames != "nsplit"             &
                     rfnames != "bootstrap"          &
                     rfnames != "sampsize"           &
                     rfnames != "samptype"           &
                     rfnames != "ntime"              &
                     rfnames != "forest"             &
                     rfnames != "save.memory"        ]
  ## get the permissible hidden options
  ## add formula if present
  dots <- list(...)
  dots <- dots[names(dots) %in% rfnames]
  if (!missing(formula)) {
    dots$formula <- formula
  }
  ## set bootstrap accordingly if the user has provided their own sampling
  ## ntree and sampsize are handled in rfsrc
  if (!is.null(samp)) {
    bootstrap <- "by.user"
  }
  ## manually set key hidden options if forest is not requested
  if (!forest) {
    dots$terminal.qualts <- FALSE
    dots$terminal.quants <- FALSE
    dots$save.memory <- TRUE
  }
  ##--------------------------------------------------------------
  ##
  ## make the grow call and return the object
  ##
  ##--------------------------------------------------------------
  return(do.call("rfsrc",
                 c(list(data = data,
                 ntree = ntree,
                 nsplit = nsplit,
                 bootstrap = bootstrap,
                 sampsize = sampsize,
                 samptype = samptype,
                 samp = samp,
                 ntime = ntime,
                 forest = forest), dots)))
}

## Finalized: uses numeric 'ssize' during tuning (OOB/holdout), supports method = "grid" or "golden"
tune.nodesize.rfsrc <- function(formula, data,
                                nodesize.try = c(1:9, seq(10, 150, by = 5)),
                                ntree.try = 100,
                                sampsize = function(x) {min(x * 0.632, max(150, x^(4/5)))},
                                nsplit = 1,
                                method = c("grid", "golden"),
                                ## golden-search controls (ignored for grid)
                                final.window = 5L,
                                reps.initial = 2L,
                                reps.final = 3L,
                                max.iter = 50L,
                                trace = TRUE,
                                seed = NULL,
                                ...) {
  method <- match.arg(method)
  ## capture and sanitize dots (avoid duplicate formal arguments)
  dots <- list(...)
  perf.type <- if ("perf.type" %in% names(dots)) dots$perf.type else "default"
  drop.from.dots <- c("perf.type","forest","save.memory","ntree","nodesize",
                      "sampsize","nsplit","data","formula")
  if (length(dots)) {
    dots <- dots[setdiff(names(dots), drop.from.dots)]
  }
  ## reproducibility: save/restore RNG; choose base seed even if user did not pass one
  have.old.seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  if (have.old.seed) old.seed <- get(".Random.seed", envir = .GlobalEnv)
  on.exit({
    if (exists("old.seed", inherits = FALSE)) assign(".Random.seed", old.seed, envir = .GlobalEnv)
  }, add = TRUE)
  if (!is.null(seed)) set.seed(seed)
  base.seed <- if (is.null(seed)) sample.int(.Machine$integer.max, 1L) else seed
  ## build stump to reconstruct a clean model frame (handles NA, factors, Surv, etc.)
  stump <- rfsrc(formula, data, nodedepth = 0, perf.type = "none",
                 save.memory = TRUE, ntree = 1, splitrule = "random")
  n <- stump$n
  yvar.names <- stump$yvar.names
  data <- data.frame(stump$yvar, stump$xvar)
  colnames(data)[seq_along(yvar.names)] <- yvar.names
  rm(stump)
  ## compute effective subsample size and coerce to integer bounds
  if (is.function(sampsize)) {
    ssize <- sampsize(n)
  } else if (is.numeric(sampsize) && length(sampsize) == 1L) {
    ssize <- sampsize
  } else if (is.numeric(sampsize) && length(sampsize) > 1L) {
    ## class-specific vector: use its total count per tree during tuning
    ssize <- sum(sampsize)
  } else {
    stop("'sampsize' must be a function or numeric (scalar or vector).")
  }
  ssize <- max(2L, min(n, as.integer(round(ssize))))
  if (trace) cat("effective sampsize:", ssize, "of", n, "\n")
  ## holdout equal to ssize if feasible; otherwise rely on OOB
  use.holdout <- (2L * ssize) < n
  if (use.holdout) {
    tst <- sample.int(n, ssize)
    trn <- setdiff(seq_len(n), tst)
    newdata <- data[tst, , drop = FALSE]
  } else {
    trn <- seq_len(n)
    newdata <- NULL
  }
  ## admissible upper bound on nodesize from the tree subsample size
  U.calc <- max(10L, ssize %/% 2L)
  ## filter the provided grid to admissible values
  grid.cand <- unique(sort(as.integer(nodesize.try[nodesize.try <= U.calc])))
  if (!length(grid.cand) && method == "grid") {
    stop("No admissible 'nodesize.try' after filtering by ssize/2. Increase 'ssize' or adjust the grid.")
  }
  ## evaluation kernel; uses numeric 'ssize' during tuning
  ## seeds are deterministic for reproducibility when 'seed' is supplied
  eval.nodesize <- function(nsz, reps) {
    vals <- numeric(reps)
    for (j in seq_len(reps)) {
      set.seed(base.seed + j)  ## shared across candidates within replicate j
      if (is.null(newdata)) {
        tr.args <- c(list(formula = formula,
                          data = data[trn, , drop = FALSE],
                          ntree = ntree.try,
                          nodesize = nsz,
                          sampsize = ssize,   ## numeric ssize during tuning
                          nsplit = nsplit,
                          perf.type = perf.type,
                          forest = FALSE,
                          save.memory = TRUE),
                     dots)
        fit <- tryCatch(do.call(rfsrc.fast, tr.args), error = function(e) e)
        if (inherits(fit, "error")) return(NA_real_)
        vals[j] <- tryCatch(mean(get.mv.error(fit, TRUE), na.rm = TRUE),
                            error = function(e) NA_real_)
      } else {
        tr.args <- c(list(formula = formula,
                          data = data[trn, , drop = FALSE],
                          ntree = ntree.try,
                          nodesize = nsz,
                          sampsize = ssize,   ## numeric ssize during tuning
                          nsplit = nsplit,
                          perf.type = "none",
                          forest = TRUE,
                          save.memory = FALSE),
                     dots)
        fit <- tryCatch(do.call(rfsrc.fast, tr.args), error = function(e) e)
        if (inherits(fit, "error")) return(NA_real_)
        pr <- tryCatch(predict(fit, newdata = newdata, perf.type = perf.type),
                       error = function(e) e)
        if (inherits(pr, "error")) return(NA_real_)
        vals[j] <- tryCatch(mean(get.mv.error(pr, TRUE), na.rm = TRUE),
                            error = function(e) NA_real_)
      }
    }
    if (trace && is.finite(mean(vals))) {
      cat("nodesize =", nsz, " reps =", reps, " error =", paste0(100 * round(mean(vals), 4), "%"), "\n")
    }
    mean(vals)
  }
  ## ----------------------------
  ## method = "grid"
  ## ----------------------------
  if (method == "grid") {
    err <- vapply(grid.cand, function(nsz) eval.nodesize(nsz, reps = 1L), numeric(1))
    if (all(is.na(err))) {
      msg <- if (is.null(newdata)) "All OOB errors are NA; check settings (especially 'sampsize')."
             else "All holdout errors are NA; check settings (especially 'sampsize')."
      warning(msg)
      return(list(nsize.opt = NA_integer_,
                  err = data.frame(nodesize = grid.cand, err = err)))
    }
    nsize.opt <- grid.cand[which.min(err)]
    if (trace) cat("optimal nodesize:", nsize.opt, "\n")
    return(list(nsize.opt = nsize.opt,
                err = data.frame(nodesize = grid.cand, err = err)))
  }
  ## ----------------------------
  ## method = "golden" (guarded)
  ## ----------------------------
  ## always probe the left boundary region (1..min(9, U.calc))
  L <- 1L; U <- U.calc
  if (U <= L) stop("Admissible nodesize range is empty; increase 'ssize' or check data.")
  left.max <- min(9L, U)
  left.cand <- seq.int(1L, left.max)
  left.err <- vapply(left.cand, function(nsz) eval.nodesize(nsz, reps.initial), numeric(1))
  best.left.idx <- which.min(left.err)
  best.left.nsz <- left.cand[best.left.idx]
  best.left.err <- left.err[best.left.idx]
  ## interior points for golden shrinkage
  phi <- (1 + sqrt(5)) / 2
  invphi <- 1 / phi
  xL2 <- max(L, min(U, floor(U - invphi * (U - L))))
  xR2 <- max(L, min(U, ceiling(L + invphi * (U - L))))
  if (xL2 == xR2) xR2 <- min(U, xL2 + 1L)
  fL2 <- eval.nodesize(xL2, reps.initial)
  fR2 <- eval.nodesize(xR2, reps.initial)
  ## if left boundary looks best, jump to a local sweep
  if (is.finite(best.left.err) && (best.left.err <= min(fL2, fR2, na.rm = TRUE))) {
    Lf <- max(L, best.left.nsz - final.window)
    Uf <- min(U, best.left.nsz + final.window)
    cand <- sort(unique(c(seq.int(Lf, Uf), left.cand)))
    err <- vapply(cand, function(nsz) eval.nodesize(nsz, reps.final), numeric(1))
    nsize.opt <- cand[which.min(err)]
    if (trace) cat("optimal nodesize:", nsize.opt, "\n")
    return(list(nsize.opt = nsize.opt, err = data.frame(nodesize = cand, err = err)))
  }
  ## otherwise proceed with golden shrinkage
  iter <- 0L
  while ((U - L) > final.window && iter < max.iter) {
    if (!is.finite(fL2) || !is.finite(fR2)) break
    if (fR2 < fL2) {
      L <- xL2; xL2 <- xR2; fL2 <- fR2
      xR2 <- max(L, min(U, ceiling(L + invphi * (U - L))))
      if (xR2 == xL2) xR2 <- min(U, xR2 + 1L)
      fR2 <- eval.nodesize(xR2, reps.initial)
    } else {
      U <- xR2; xR2 <- xL2; fR2 <- fL2
      xL2 <- max(L, min(U, floor(U - invphi * (U - L))))
      if (xL2 == xR2) xL2 <- max(L, xL2 - 1L)
      fL2 <- eval.nodesize(xL2, reps.initial)
    }
    iter <- iter + 1L
  }
  ## final local sweep
  cand <- sort(unique(c(seq.int(L, U), left.cand)))
  err <- vapply(cand, function(nsz) eval.nodesize(nsz, reps.final), numeric(1))
  if (all(is.na(err))) {
    msg <- if (is.null(newdata)) "All OOB errors are NA; check 'sampsize' and other settings."
           else "All holdout errors are NA; check 'sampsize' and other settings."
    warning(msg)
    return(list(nsize.opt = NA_integer_, err = data.frame(nodesize = cand, err = err)))
  }
  nsize.opt <- cand[which.min(err)]
  if (trace) cat("optimal nodesize:", nsize.opt, "\n")
  list(nsize.opt = nsize.opt, err = data.frame(nodesize = cand, err = err))
}
tune.nodesize <- tune.nodesize.rfsrc

imbalanced.rfsrc <- function(formula, data, ntree = 3000,
                             method = c("rfq", "brf", "standard"), splitrule = "auc",
                             perf.type = NULL, block.size = NULL, fast = FALSE,
                             ratio = NULL,  ...)
{
  ## preliminary checks: all are fatal
  ## parse the formula to ensure this is a two-class problem
  formulaPrelim <- parseFormula(formula, data)
  if (formulaPrelim$family != "class") {
    stop("this function only applies to classification problems")
  }
  yvar <- data[, formulaPrelim$yvar.names]
  if (length(levels(yvar)) != 2) {
    stop("this function only applies to two-class problems")
  }
  ## check that method is set correctly
  method <- match.arg(method, c("rfq", "brf", "standard"))
  rfq.flag <- NULL
  if (method == "rfq") {
    rfq.flag <- TRUE
  }
  ## set default performance
  if (is.null(perf.type)) {
    if (method == "brf" || method == "rfq") {
      perf.type <- "gmean"
    }
    else {
      perf.type <- "default"##equivalent to misclass
    }
  }
  ## check performance type is properly set
  perf.type <- match.arg(perf.type,
        c("none", "default", "standard", "misclass", "brier", "gmean", "g.mean"))
  if (perf.type == "g.mean") {##legacy
    perf.type <- "gmean"
  }
  ##-----------------------------------------
  ##
  ## rfsrc call - depends on what's requested
  ##
  ##-----------------------------------------
  ##----------------------------------------
  ## brf method
  ##----------------------------------------
  if (method == "brf") {
    ## TBD2 currently cannot handle missing values
    data <- na.omit(data)
    yvar <- data[, formulaPrelim$yvar.names]
    ## for legacy reasons we maintain swr here
    ## TBD2 consider swor
    o <- rfsrc(formula, data, ntree = ntree,
               splitrule = splitrule,
               perf.type = perf.type,
               block.size = block.size,
               case.wt = make.wt(yvar),
               sampsize = make.size(yvar),
               samptype = "swr", ...)
  }
  ##----------------------------------------
  ## standard and rfq method
  ##----------------------------------------
  else {
    ##-------------------------------------------------
    ## determine the grow interface - rfsrc or rfsrc.fast?
    ##-------------------------------------------------
    if (!fast) {
      rfsrc.grow <- "rfsrc"
    }
    else {
      rfsrc.grow <- "rfsrc.fast"
    }
    ##-------------------------------------------------
    ## acquire the user specified additional options
    ##-------------------------------------------------
    dots <- list(...)
    ##-------------------------------------------------
    ## undersampling of the majority class if requested
    ##-------------------------------------------------
    if (!is.null(ratio)) {
      ## TBD2 currently cannot handle missing values
      data <- na.omit(data)
      yvar <- data[, formulaPrelim$yvar.names]
      samp <- make.imbalanced.sample(ntree = ntree, ratio = ratio, y = yvar)
      dots$bootstrap <- dots$samp <- NULL
      o <- do.call(rfsrc.grow, c(list(formula = formula, data = data, ntree = ntree,
                                      rfq = rfq.flag, splitrule = splitrule,
                                      perf.type = perf.type, block.size = block.size,
                                      samp = samp, bootstrap = "by.user"), dots))
    }
    ##-----------------------------------------------------------
    ## proceed to standard and rfq analysis without undersampling
    ## this is the default scenario
    ## allow fast random forests if requested
    ##-----------------------------------------------------------
    else {
      o <- do.call(rfsrc.grow, c(list(formula = formula, data = data, ntree = ntree,
                                      rfq = rfq.flag, splitrule = splitrule,
                                      perf.type = perf.type, block.size = block.size), dots))    
    }
  }
  ## return the object
  o
}
imbalanced <- imbalanced.rfsrc

adrop3d.last <- function(x, d, keepColNames = FALSE) {
  ## this function is for arrays only
  if (!is.array(x)) {
    x
  }
  else {
    if (d > 1) {
      x[,,1:d, drop = FALSE]
    }
    else {
      if (dim(x)[1] == 1) {
        rbind(x[,,1, drop = TRUE])
      }
      else {
        if (dim(x)[2] == 1) {
          if (keepColNames) {
            ##needed for J=1 classification pathology
            xnew <- cbind(x[,,1, drop = TRUE])
            colnames(xnew) <- colnames(x)
            xnew
          }
          else {
            cbind(x[,,1, drop = TRUE])
          }
        }
        else {
          x[,,1, drop = TRUE]
        }
      }
    }
  }
}
adrop2d.first <- function(x, d, keepColNames = FALSE) {
  ## this function is for arrays only
  if (!is.array(x)) {
    x
  }
  else {
    if (d > 1) {
      x[1:d,, drop = FALSE]
    }
    else {
      x[1, , drop = TRUE]
    }
  }
}
adrop2d.last <- function(x, d, keepColNames = FALSE) {
  ## this function is for arrays only
  if (!is.array(x)) {
    x
  }
  else {
    if (d > 1) {
      x[,1:d, drop = FALSE]
    }
    else {
      x[,1, drop = TRUE]
    }
  }
}
amatrix <- function(x, d, names) {
  x <- matrix(x, d, dimnames = names)
  if (ncol(x) > 1) {
    x
  }
  else {
    c(x)
  }
}
amatrix.remove.names <- function(x) {
  if (!is.null(dim(x)) && ncol(x) == 1) {
    unlist(c(x), use.names = FALSE)
  }
  else {
    x
  }
}
atmatrix <- function(x, d, names, keep.names = FALSE) {
  x <- t(matrix(x, ncol = d, dimnames = names))
  if (ncol(x) > 1) {
    x
  }
  else {
    if (keep.names == FALSE) {
      c(x)
    }
    else {
      x.names <- rownames(x)
      x <- c(x)
      names(x) <- x.names
      x
    }
  }
}
avector <- function(x, name = FALSE) {
  if (!is.null(dim(x)) && nrow(x) > 1 && ncol(x) == 1) {
    x.names <- rownames(x)
    x <- unlist(c(x))
    if (name) names(x) <- x.names else names(x) <- NULL
    x
  }
  else if (!is.null(dim(x)) && nrow(x) == 1 && ncol(x) > 1) {
    x.names <- colnames(x)
    x <- unlist(c(x))
    if (name) names(x) <- x.names else names(x) <- NULL
    x
  }
  else if (!is.null(dim(x)) && nrow(x) == 1 && ncol(x) == 1) {
    unlist(c(x))
  }
  else {
    x
  }
}
available <- function (package, lib.loc = NULL, quietly = TRUE) {
  package <- as.character(substitute(package))
  installed <- package %in% installed.packages()
  if (installed) {
    require(package, quietly = TRUE, character.only = TRUE)
  }
    else {
      return(invisible(FALSE))
    }
}
cv.folds <- function (n, folds = 10) {
  split(resample(1:n), rep(1:folds, length = n))
}
data.matrix <- function(x) {
  as.data.frame(lapply(x, function(xi) {
    if (is.integer(xi) || is.numeric(xi)) {
      xi
    }
      else if (is.logical(xi) || is.factor(xi)) {
        as.integer(xi)
      }
        else {
          as.numeric(xi)
        }
  }))
}
digits.pretty <- function(x, digits = 8) {
  paste(round(x, digits), collapse=", ", sep = "")
}
family.pretty <- function(x) {
  fmly <- x$family
  if (!is.null(x$forest$rfq) && x$forest$rfq) {
    fmly <- "rfq"
  }
  switch(fmly,
         "surv"     = "RSF",
         "surv-CR"  = "RSF",
         "surv-TDC" = "RHF",
         "regr"     = "RF-R",
         "class"    = "RF-C",
         "unsupv"   = "RF-U",
         "regr+"    = "mRF-R",
         "class+"   = "mRF-C",
         "mix+"     = "mRF-RC",
         "rfq"      = "RFQ"
         )
}
finalizeFormula <- function(formula.obj, data) {
    ## parse the formula object
    yvar.names <- formula.obj$yvar.names
    subj.names <- formula.obj$subj.names
    all.names  <- formula.obj$all.names
    fmly       <- formula.obj$family
    ytry       <- formula.obj$ytry
    index <- length(yvar.names)
    ## Adjust the index for the presence of subject names.
    if (fmly == "surv") {
        if (!is.null(subj.names)) {
            index <- index + 1
        }
    }
    ## total number of variables should exceed number of yvars
    if (length(all.names) <= index) {
        stop("formula is misspecified: total number of variables does not exceed total number of y-variables")
    }
    ## extract the xvar names
    if (all.names[index + 1] == ".") {
        if(index == 0) {
            xvar.names <- names(data)
        }
        else {
            xvar.names <- names(data)[!is.element(names(data), all.names[1:index])]
        }
    }
    else {
        if(index == 0) {
            xvar.names <- all.names
        }
        else {
            xvar.names <- all.names[-c(1:index)]
        }
        not.specified <- !is.element(xvar.names, names(data))
        if (sum(not.specified) > 0) {
            stop("formula is misspecified, object ", xvar.names[not.specified], " not found")
        }
    }
    ## return the goodies
    return (list(family=fmly, subj.names=subj.names, yvar.names=yvar.names, xvar.names=xvar.names, ytry=ytry))
}
finalizeData <- function(fnames, data, na.action, miss.flag = TRUE) {
  ## restrict data to the target variables
  data <- data[ , is.element(names(data), fnames), drop = FALSE]
  ## data conversion to numeric mode for factors
  ## no need for conversion if factors are not present
  factor.names <- unlist(lapply(data, is.factor))
  if (sum(factor.names) > 0) {
    data[, factor.names] <- data.matrix(data[, factor.names, drop = FALSE])
  }
  ## omit the na's
  if (miss.flag == TRUE && na.action == "na.omit") {
    ## no need to make the call if there is no missing data
    if (any(is.na(data))) {
      data <- na.omit(data)
    }
  }
  ## canvert nan to na's (github reported bug 06/27/2018)
  if (miss.flag == TRUE && na.action != "na.omit") {
    nan.names <- sapply(data, function(x){any(is.nan(x))})
    if (sum(nan.names) > 0) {
      data[, nan.names] <- data.frame(lapply(which(nan.names), function(j) {
        x <- data[, j]
        x[is.nan(x)] <- NA
        x
      }))
    }
  }
  ## is anything left?
  if (nrow(data) == 0) {
    stop("no records in the NA-processed data: consider using 'na.action=na.impute'")
  }
  ## convert logicals to 0/1 real (bug reported by John Ehrlinger)
  logical.names <- unlist(lapply(data, is.logical))
  if (sum(logical.names) > 0) {
    data[, logical.names] <- 1 * data[, logical.names, drop = FALSE]
  }
  ## characters not allowed (bug reported by John Ehrlinger)
  character.names <- unlist(lapply(data, is.character))
  if (sum(character.names) > 0) {
    stop("data types cannot be character: please convert all characters to factors")
  }
  return (data)
}
get.coerced.survival.fmly <- function(fmly, subj, event.type, splitrule = NULL) {
    if (grepl("surv", fmly)) {
        ## assume no coercion
        coerced.fmly <- "surv"
        if (!is.null(splitrule)) {
            if (is.null(subj)) {             
                ## either competing risks or right censoring competing risks
                ## is coerced to right censoring in some settings
                if ((length(event.type) > 1) &&
                    (splitrule != "l2.impute") &&
                    (splitrule != "logrankscore")) {
                    coerced.fmly <- "surv-CR"
                }
            }
            else {
                coerced.fmly <- "surv-TDC"
            }
        }
        else {
            if (is.null(subj)) { 
                if (length(event.type) > 1) {
                    coerced.fmly <- "surv-CR"
                }
            }
            else {
                coerced.fmly <- "surv-TDC"
            }
        }
    }
    else {
        stop("attempt to coerce a non-survival family")
    }
    coerced.fmly
}
## get grow mtry 
get.grow.mtry <- function (mtry = NULL, n.xvar, fmly, splitrule = NULL) {
  if (!is.null(mtry)) {
    mtry <- round(mtry)
    if (mtry < 1 | mtry > n.xvar) mtry <- max(1, min(mtry, n.xvar))
  }
  else {
    if (grepl("regr", fmly)) {
      mtry <- max(ceiling(n.xvar/3), 1)
    }
    else {
      mtry <- max(ceiling(sqrt(n.xvar)), 1)
    }
  }
  if (!is.null(splitrule) && splitrule == "random") {
    mtry <- 1
  }
  mtry
}
get.grow.nodesize <- function(fmly, nodesize) {
  ## Default node size for right-censored survival
  if (fmly == "surv"){
    if (is.null(nodesize)) {
      nodesize <- 15
    }
  }
  ## Default node size for competing risks
    else if (fmly == "surv-CR"){
      if (is.null(nodesize)) {
        nodesize <- 15
      }
    }
  ## Default node size for time dependent covariates
    else if (fmly == "surv-TDC"){
      if (is.null(nodesize)) {
        nodesize <- 15
      }
    }
  ## Default node size for classification
      else if (fmly == "class" | fmly == "class+") {
        if (is.null(nodesize)) {
          nodesize <- 1
        }
      }
  ## Default node size for regression
        else if (fmly == "regr" | fmly == "regr+") {
          if (is.null(nodesize)) {
            nodesize <- 5
          }
        }
  ## Default node size for mixed outcomes
          else if (fmly == "mix+") {
            if (is.null(nodesize)) {
              nodesize <- 3
            }
          }
  ## Default node size for unsupervised splitting
            else if (fmly == "unsupv") {
              if (is.null(nodesize)) {
                nodesize <- 3
              }
            }
  ## The family is misspecified
              else if (is.null(nodesize)) {
                stop("family is misspecified")
              }
  ## Nodesize should be rounded if non-integer
  nodesize <- round(nodesize)
}
get.grow.splitinfo <- function (formula.detail, splitrule, nsplit, event.info) {
  ## CAUTION:  HARD CODED ON NATIVE SIDE
  splitrule.names <- c("logrank",              ##  1
                       "logrankscore",         ##  2
                       "logrankCR",            ##  3
                       "random",               ##  4
                       "mse",                  ##  5
                       "mse.unwt",             ##  6
                       "mse.hvwt",             ##  7
                       "gini",                 ##  8
                       "gini.unwt",            ##  9
                       "gini.hvwt",            ## 10
                       "unsupv",               ## 11
                       "unsupv.unwt",          ## 12
                       "unsupv.hvwt",          ## 13
                       "mv",                   ## 14
                       "mv.unwt",              ## 15
                       "mv.hvwt",              ## 16
                       "custom",               ## 17
                       "quantile.regr",        ## 18
                       "la.quantile.regr",     ## 19
                       "bs.gradient",          ## 20
                       "auc",                  ## 21
                       "entropy",              ## 22
                       "sg.regr",              ## 23
                       "sg.class",             ## 24
                       "sg.surv",              ## 25
                       "mahalanobis",          ## 26
                       "logrankCRGeneral")     ## 27
  ## set the family
  fmly <- formula.detail$family
  ## initialization
  cust.idx <- NULL
  splitpass <- FALSE
  ## set splitpass
  if (!is.null(splitrule)) {
    if(grepl("custom", splitrule)) {
      splitrule.idx <- which(splitrule.names == "custom")
      cust.idx <- as.integer(sub("custom", "", splitrule))
      if (is.na(cust.idx)) cust.idx <- 1
      splitpass <- TRUE
    }
    else if (splitrule == "random") {
      splitrule.idx <- which(splitrule.names == "random")
      splitpass <- TRUE
    }
  }
  ## do this unless pass given for split 
  if (!splitpass) {
    ## if splitrule is present, confirm it is correctly set
    if (!is.null(splitrule)) {
      splitrule <- match.arg(splitrule, splitrule.names)
    }
    ## survival
    if (grepl("surv", fmly)) {
      if (is.null(splitrule)) {
        ## No split rule specified, use default.
        if (event.info$r.dim == 2) {
          ## Survival or Competing Risk:
          if (length(event.info$event.type) ==  1) {
            splitrule.idx <- which(splitrule.names == "logrank")
          }
          else {
            splitrule.idx <- which(splitrule.names == "logrankCR")
          }
        }
        else if (event.info$r.dim == 3) {
          ## Time Dependent Covariates:                    
          splitrule.idx <- which(splitrule.names == "tdc.gradient")
        }
        else {
          stop("Invalid r.dim encountered in split rule:  ", event.info$r.dim)
        }
        splitrule <- splitrule.names[splitrule.idx]
      }
      else {
        ## User split rule specified.
        splitrule.idx <- which(splitrule.names == splitrule)
        if (length(splitrule.idx) != 1) {
          stop("Invalid split rule specified for survival:  ", splitrule)
        }
        if (event.info$r.dim == 2) {
          if ((length(event.info$event.type) ==  1) & (splitrule.idx == which(splitrule.names == "logrankCR"))) {
            stop("Cannot specify logrankCR splitting for right-censored data")
          }
          if ((length(event.info$event.type) >   1) & (splitrule.idx == which(splitrule.names == "logrank"))) {
              ## Override the splitrule to access the generalized CR split rule (see 3.3.1 in the paper).
              ## The default splitrule is 3.3.2 or Gray's test.
            splitrule.idx <- which(splitrule.names == "logrankCRGeneral")
          }
        }
        else if (event.info$r.dim == 3) {
          if (splitrule.idx != which(splitrule.names == "tdc.gradient")) {
            stop("Must specify tdc.gradient for time dependent covariates")                        
          }
        }
        else {
          stop("Invalid r.dim encountered in split rule:  ", event.info$r.dim)
        }
      }
    }
    ## class
    if (fmly == "class") {
      if (is.null(splitrule)) {
        ## No split rule specified, use default.
        splitrule.idx <- which(splitrule.names == "gini")
        splitrule <- splitrule.names[splitrule.idx]
      }
      else {
        ## User specified split rule.
        if ((splitrule != "gini") &
            (splitrule != "gini.unwt") &
            (splitrule != "gini.hvwt") &
            (splitrule != "auc") &
            (splitrule != "entropy") &
            (splitrule != "sg.class")) {
          stop("Invalid split rule specified for classification:  ", splitrule)
        }
        splitrule.idx <- which(splitrule.names == splitrule)
      }
    }
    ## regression
    if (fmly == "regr") {
      if (is.null(splitrule)) {
        ## No split rule specified, use default.
        splitrule.idx <- which(splitrule.names == "mse")
        splitrule <- splitrule.names[splitrule.idx]
      }
      else {
        ## User specified split rule.
        if ((splitrule != "mse") &
            (splitrule != "mse.unwt") &
            (splitrule != "mse.hvwt") &
            (splitrule != "sg.regr") &
            (splitrule != "la.quantile.regr") &
            (splitrule != "quantile.regr")) {
          stop("Invalid split rule specified for regression:  ", splitrule)
        }
        splitrule.idx <- which(splitrule.names == splitrule)
      }
    }
    ## multivariate
    if (fmly == "regr+") {
      if (is.null(splitrule)) {
        ## No split rule specified, use default
        splitrule.idx <- which(splitrule.names == "mv")
        splitrule <- splitrule.names[splitrule.idx]
      }
      else {
        ## User specified split rule
        if ((splitrule != "mv") & (splitrule != "mv.unwt") & (splitrule != "mv.hvwt") & (splitrule != "mahalanobis")) {
          stop("Invalid split rule specified for multivariate regression:  ", splitrule)
        }
        splitrule.idx <- which(splitrule.names == splitrule)
      }
    }
    if (fmly == "class+") {
      if (is.null(splitrule)) {
        ## No split rule specified, use default
        splitrule.idx <- which(splitrule.names == "mv")
        splitrule <- splitrule.names[splitrule.idx]
      }
      else {
        ## User specified split rule.
        if ((splitrule != "mv") & (splitrule != "mv.unwt") & (splitrule != "mv.hvwt")) {
          stop("Invalid split rule specified for multivariate classsification:  ", splitrule)
        }
        splitrule.idx <- which(splitrule.names == splitrule)
      }
    }
    if (fmly == "mix+") {
      if (is.null(splitrule)) {
        ## No split rule specified, use default.
        splitrule.idx <- which(splitrule.names == "mv")
        splitrule <- splitrule.names[splitrule.idx]
      }
      else {
        ## User specified split rule.
        if ((splitrule != "mv") & (splitrule != "mv.unwt") & (splitrule != "mv.hvwt")) {
          stop("Invalid split rule specified for mixed multivariate regression:  ", splitrule)
        }
        splitrule.idx <- which(splitrule.names == splitrule)
      }
    }
    ## unsupervised
    if (fmly == "unsupv") {
      if (is.null(splitrule)) {
        ## No split rule specified, use default.
        splitrule.idx <- which(splitrule.names == "unsupv")
        splitrule <- splitrule.names[splitrule.idx]
      }
      else {
        ## User specified split rule.
        if ((splitrule != "unsupv") & (splitrule != "unsupv.unwt") & (splitrule != "unsupv.hvwt")) {
          stop("Invalid split rule specified:  ", splitrule)
        }
        splitrule.idx <- which(splitrule.names == splitrule)
      }
    }
  }## completes !splitpass
  ## set nsplit based on splitrule
  if(!is.null(nsplit)) {
    nsplit <- round(nsplit)    
    if (nsplit < 0) {
      stop("Invalid nsplit value: set nsplit >= 0")
    }
  }
  else {
    ## pure random
    if (splitrule == "random") {
      nsplit <- 1
    }
    ## fast families -- revert back to 10 due to factors
    else if ((splitrule == "mse")  | (splitrule == "mse.unwt")  | (splitrule == "mse.hvwt") |
             (splitrule == "gini") | (splitrule == "gini.unwt") | (splitrule == "gini.hvwt")) {  
      nsplit <- 10
    }
    ## sg families
    else if ((splitrule == "sg.regr") | (splitrule == "sg.class") | (splitrule == "sg.surv")) {
      nsplit <- 10 ## can change this as needed later
    }
    else {
      nsplit <- 10
    }
  }    
  splitinfo <- list(name = splitrule, index = splitrule.idx, cust = cust.idx, nsplit = nsplit)
  return (splitinfo)
}
get.importance.xvar <- function(importance.xvar, importance, object) {
  ## Check that importance has been requested
  if (!is.null(importance)) {
    ## Map vimp names to columns of GROW x-matrix
    ## Ensure names are coherent
    if (missing(importance.xvar) || is.null(importance.xvar)) {
      importance.xvar <- object$xvar.names
    }
      else {
        importance.xvar <- unique(importance.xvar)
        importance.xvar <- intersect(importance.xvar, object$xvar.names)
      }
    if (length(importance.xvar) == 0) {
      stop("xvar names do not match object xvar matrix")
    }
  }
    else {
      ## This was previously zero.  We set this so as to get length zero to the native code.
      importance.xvar <- NULL
    }
  return (importance.xvar)
}
get.marginal.xvar <- function(marginal.xvar, object) {
  ## Check the coherency of the incoming vector.
  if (missing(marginal.xvar) | is.null(marginal.xvar)) {
    marginal.xvar <- NULL
  }
    else {
      marginal.xvar <- unique(marginal.xvar)
      marginal.xvar <- intersect(marginal.xvar, object$xvar.names)
      if (length(marginal.xvar) == 0) {
        stop("marginal xvar names must be a subset of xvar names")
      }
    }
  return (marginal.xvar)
}
get.nmiss <- function(xvar, yvar = NULL) {
  if (!is.null(yvar)) {
    sum(apply(yvar, 1, function(x){any(is.na(x))}) | apply(xvar, 1, function(x){any(is.na(x))}))
  }
    else {
      sum(apply(xvar, 1, function(x){any(is.na(x))}))
    }
}
get.outcome.target <- function(family, yvar.names, outcome.target) {
  if (family == "regr" | family == "regr+" | family == "class" | family == "class+" | family == "mix+") {
    if (is.null(outcome.target)) {
      outcome.target <- yvar.names
    }
    ## Map target names to outcome names and ensure coherency.
    outcome.target <- unique(outcome.target)
    outcome.target <- intersect(outcome.target, yvar.names)
    if (length(outcome.target) == 0) {
      stop("yvar target names do not match object yvar names")
    }
    outcome.target <- match(outcome.target, yvar.names)
  }
    else {
      ## This is surv or surv-CR
      outcome.target <- 0
    }
}
get.rfnames <- function(hidden = TRUE, stealth = FALSE) {
  rfnames <- names(formals(rfsrc))
  if (hidden) {
    rfnames <- c(rfnames,              
               "impute.only",
               "experimental",
               "rfq",
               "perf.type",
               "gk.quantile",
               "prob",
               "prob.epsilon",
               "vtry",
               "holdout.array")
  }
  rfnames
}
get.weight <- function(weight, n) {
  ## set the default weight
  if (!is.null(weight)) {
    if (any(weight < 0)      ||
##        all(weight == 0)     ||
        length(weight) != n  ||
        any(is.na(weight))) {
      stop("Invalid weight vector specified.")
    }
  }
    else {
      weight <- rep(1, n)
    }
  return (weight)
}
get.ytry <- function(f) {
}
get.yvar.type <- function(fmly, generic.types, yvar.names) {
  if (fmly == "unsupv") {
    yvar.type <- NULL
  }
    else {
      if (grepl("surv", fmly)) {
          if (length(yvar.names) == 2) {
              yvar.type <- c("T", "S")
          }
          else {
              yvar.type <- c("t", "T", "S")
          }
      }
        else {
          yvar.type <- generic.types
        }
    }
  yvar.type
}
get.yvar.nlevels <- function(fmly, nlevels, yvar.names, yvar) {
  if (fmly == "unsupv") {
    NULL
  }
  else {
    nlevels
  }
}
get.numeric.levels <- function(fmly, nlevels, gvar) {
    gvar.numeric.levels  <- lapply(1:length(nlevels),
                                   function(nn) {if(nlevels[nn] > 0) unique(sort(gvar[, nn])) else NULL})
    ## Remove null elements in the list
    gvar.numeric.levels <- gvar.numeric.levels[!sapply(gvar.numeric.levels,is.null)]
    ## We are uncomfortable in sending a ist of length zero into the C-code, so we add an additional check.
    if (length(gvar.numeric.levels) == 0) gvar.numeric.levels = NULL
    gvar.numeric.levels
}
global.prob.assign <- function(prob, prob.epsilon, gk.quantile, quantile.regr, splitrule, n) {
  ## default values
  prob.default <- (1:99) / 100
  prob.epsilon.default <- .005
  prob.bs.default <- .9
  ## is quantile.regr in effect or is a quantile regression splitting rule in place?
  if (quantile.regr  || grepl("quantile.regr", splitrule) || gk.quantile) {
    ## is gk requested?
    if (gk.quantile) {
      ## gk quantile requires prob and prob.epsilon
      ## if both are missing, set to optimal value -- SLOW SLOW !!!!
      if (is.null(prob) && is.null(prob.epsilon)) {
        n <- max(n, 1)
        prob <- (1 : (2 * n)) / (2 * n + 1)
        prob.epsilon <- diff(prob)[1] * .99
      }
      ## if prob missing, set to default value
      else if (is.null(prob) && !is.null(prob.epsilon)) {
        prob <- prob.default
      }
      ## if prob.epsilon missing, set to default value
      else if (!is.null(prob) && is.null(prob.epsilon)) {
        prob.epsilon <- mean(diff(sort(prob)), na.rm = TRUE) * .99
      }
      ## neither are missing
      else {
        ##nothing        
      }
    }
    ## quantile splitting is in effect but no gk estimation
    else {
      prob.epsilon <- prob.epsilon.default
      ## here's where we set the default prob if its missing
      if (is.null(prob)) {
        prob <-  prob.default
      }
    }
    ## make sure prob values are coherent
    if (sum(prob > 0 & prob < 1) == 0) {
      stop("parameter 'prob' is not between (0,1)", prob)
    }
    prob <- sort(prob[prob>0 & prob<1])
  }
  ## survival has quantile splitting rules, here's where we deal with those
  if (splitrule ==  "bs.gradient") {
    if (is.null(prob)) {
      prob <- prob.bs.default
    }
    ## make sure prob values are coherent
    if (sum(prob > 0 & prob < 1) == 0) {
      stop("parameter 'prob' is not between (0,1)", prob)
    }
    prob <- prob[prob>0 & prob<1][1]
  }
  ## for all other families leave the values at default NULL
  #if (is.null(prob)) {
  #  prob <-  prob.default
  #}
  #if (is.null(prob.epsilon)) {
  #  prob.epsilon <-  prob.epsilon.default
  #}
  return(list(prob = prob, prob.epsilon = prob.epsilon))
}
make.samplesize.function <- function(fraction = 1) {
  f <- paste("x * ", paste(eval(fraction)))
  expr <- parse(text = f)
  function(x) eval(expr, list(x = x))
}
make.holdout.array <- function(vtry = 0, p, ntree, ntree.allvars = NULL) {
  ##default is triggered by vtry = 0
  if (vtry == 0) {
    return(NULL)
  }
  ## check that 1<= vtry <= p-1
  if (vtry < 0 || vtry >= p) {
    stop("vtry must be positive and less than the feature dimension")
  }
  ## everything is OK, go ahead and make the p x ntree array
  holdout <- do.call(cbind, lapply(1:ntree, function(b) {
    ho <- rep(0, p)
    ho[resample(1:p, size = vtry, replace = FALSE)] <- 1
    ho
  }))
  if (is.null(ntree.allvars)) {
    holdout
  }
  ## otherwise pad the array with 0's so that no variables are held out
  ## put them at the front for split-optimization  
  else {
    cbind(matrix(0, p, ntree.allvars), holdout)
  }
}
## imbalanced sampling - uses SWOR
make.imbalanced.sample <- function(ntree, ratio = 0.5, y, replace = FALSE) {
  ## ratio must be between 0 and 1
  if (ratio < 0 | ratio > 1) {
    stop("undersampling ratio must be between 0 and 1:", ratio)
  }
  ## determine minority/majority frequencies
  frq <- table(y)
  class.labels <- names(frq)
  minority <- which.min(frq)
  majority <- setdiff(1:2, minority)
  n.minority <- min(frq, na.rm = TRUE)
  n.majority <- max(frq, na.rm = TRUE)
  samp.size <- length(y)
  ## ratio must be larger than minority ratio
  ratio <- max((2 + n.minority) / length(y), ratio)
  ## use rbind to ensure a matrix
  rbind(sapply(1:ntree, function(bb){
    inb <- rep(0, samp.size)
    smp.min <- resample(which(y == class.labels[minority]), size = n.minority, replace = TRUE)
    smp.maj <- resample(which(y == class.labels[majority]), size = ratio * n.majority, replace = replace)
    smp <- c(smp.min, smp.maj)
    inb.frq <- tapply(smp, smp, length)
    idx <- as.numeric(names(inb.frq))
    inb[idx] <- inb.frq
    inb
  }))
}
## modified to default to "SWR"
## make inbag bootstrap samples (for use with bootstrap = "by.user")
## allows for under/over sampling
make.sample <- function(ntree, samp.size, boot.size = NULL, replace = FALSE) {
  ## samp.size cannot be negative
  if (samp.size < 0) {
    stop("samp.size cannot be negative:", samp.size)
  }
  ## default setting
  if (is.null(boot.size)) {
    if (replace == TRUE) {
      boot.size <- samp.size
    }
    else {
      boot.size <- .632 * samp.size
    }
  }
  ## use rbind at the end to ensure a matrix
  rbind(sapply(1:ntree, function(bb){
    inb <- rep(0, samp.size)
    smp <- resample(1:samp.size, size = boot.size, replace = replace)
    frq <- tapply(smp, smp, length)
    idx <- as.numeric(names(frq))
    inb[idx] <- frq
    inb
  }))
}
## make stratified inbag bootstrap samples for imbalanced two class problem
## allows for arbitrary under-sampling ratio of the majority class
## sample size for subsampling
make.size <- function(y) {
  ## extract the relative frequencies
  frq <- table(y)
  min(length(y), min(frq, na.rm = TRUE) * length(frq))
}
## imbalanced weights for subsampling
make.wt <- function(y) {
  ## extract the relative frequencies
  frq <- table(y)
  class.labels <- names(frq)
  wt <- rep(1, length(y))
  ## weights for class j equal product of all non-j class frequencies
  nullO <- sapply(1:length(frq), function(j) {
    wt[y == class.labels[j]] <<- prod(frq[-j], na.rm = TRUE)
    NULL
  })
  ## weights can become large, so divide by the maximum value
  ## this ensures 0 < wt <= 1
  wt / max(wt, na.rm = TRUE)
}
parseFormula <- function(f, data, ytry = NULL, coerce.factor = NULL) {
  ## confirm coherency of the formula
  if (!inherits(f, "formula")) {
    stop("'formula' is not a formula object.")
  }
  if (is.null(data)) {
    stop("'data' is missing.")
  }
  if (!is.data.frame(data)) {
    stop("'data' must be a data frame.")
  }
  ## pull the family and y-variable names
  fmly <- all.names(f, max.names = 1e7)[2]
  all.names <- all.vars(f, max.names = 1e7)
  yvar.names <- all.vars(formula(paste(as.character(f)[2], "~ .")), max.names = 1e7)
  yvar.names <- yvar.names[-length(yvar.names)]
  ## Default scenario, no subject information when family is not
  ## time dependent covariates.  Can be overridden later.
  subj.names <- NULL
  ## is coerce.factor at play for the y-outcomes?
  coerce.factor.org <- coerce.factor
  coerce.factor <- vector("list", 2)
  names(coerce.factor) <- c("xvar.names", "yvar.names")
  if (!is.null(coerce.factor.org)) {
    coerce.factor$yvar.names <- intersect(yvar.names, coerce.factor.org)
    if (length(coerce.factor$yvar.names) == 0) {
      coerce.factor$yvar.names <- NULL
    }
    coerce.factor$xvar.names <- intersect(setdiff(colnames(data), yvar.names), coerce.factor.org)
  }
  ## survival forests
  if (fmly == "Surv") {
    ## Survival and competing risk will have 2 slots, namely time and censoring.
    ## Time dependent covariates will have 4 slots, namely id, start, stop, and event.
    ## If TDC is in effect, we remove the id from the yvars, and tag is an the subject identifier.
    if ((sum(is.element(yvar.names, names(data))) != 2) &&
        (sum(is.element(yvar.names, names(data))) != 4)) {
      stop("Survival formula incorrectly specified.")
    }
    else {
      if (sum(is.element(yvar.names, names(data))) == 4) {
        ## Time dependent covariates is in effect.
        subj.names <- yvar.names[1]
        yvar.names <- yvar.names[-1]
      }
    }
    family <- "surv"
    ytry <- 0
  }
  ## multivariate forests
  else if ((fmly == "Multivar" || fmly == "cbind")  && length(yvar.names) > 1) {
    if (sum(is.element(yvar.names, names(data))) < length(yvar.names)) {
      stop("Multivariate formula incorrectly specified: y's listed in formula are not in data.")
    }
    ## determine the family: now handles mixed outcomes
    Y <- data[, yvar.names, drop = FALSE]
    ## convert logical to 0/1 real (bug reported by John Ehrlinger)
    logical.names <- unlist(lapply(Y, is.logical))
    if (sum(logical.names) > 0) {
      Y[, logical.names] <- 1 * Y[, logical.names, drop = FALSE]
    }
    ## are all the responses factors?
    ## caution: ordered factors are factors!
    if ((sum(unlist(lapply(Y, is.factor))) + 
         length(coerce.factor$yvar.names)) == length(yvar.names)) {
      family <- "class+"
    }
    ## are all the responses continuous?
    ## caution: ordered factors are factors!
    else if ((sum(unlist(lapply(Y, is.factor))) + 
              length(coerce.factor$yvar.names)) == 0) {
      family <- "regr+"
    }
    ## are the responses a combination of factors and continuous?
    ## caution: ordered factors are factors!
    else if (((sum(unlist(lapply(Y, is.factor))) +
               length(coerce.factor$yvar.names)) > 0) && 
             ((sum(unlist(lapply(Y, is.factor))) +
               length(coerce.factor$yvar.names)) < length(yvar.names))) {
      family <- "mix+"
    }
    ## failure
    else {
      stop("y-outcomes must be either real or factors in multivariate forests.")
    }
    if (!is.null(ytry)) {
      ## Check that incoming ytry is consistent.
      if ((ytry < 1) || (ytry > length(yvar.names))) {
        stop("invalid value for ytry:  ", ytry)
      }
    }
    else {
      ytry <- length(yvar.names)
    }
  }
  ## unsupervised forests
  else if (fmly == "Unsupervised") {
    ## unsupervised forests
    if (length(yvar.names) != 0) {
      stop("Unsupervised forests require no y-responses")
    }
    family <- "unsupv"
    yvar.names <- NULL
    ## Strip away the family from the formula, leaving ytry.
    temp <- gsub(fmly, "", as.character(f)[2])
    temp <- gsub("\\(|\\)", "", temp)
    ytry <- as.integer(temp)
    if (is.na(ytry)) {
      ytry <- 1
    }
    else {
      if (ytry <= 0) {
        stop("Unsupervised forests require positive ytry value")
      }
    }
  }
  ## univariate forests (regression or classification)
  else {
    ## must be a (univariate) regresssion or classification
    if (sum(is.element(yvar.names, names(data))) != 1) {
      stop("formula is incorrectly specified.")
    }
    Y <- data[, yvar.names]
    ## logicals are treated as 0/1 real (bug reported by John Ehrlinger)
    if (is.logical(Y)) {
      Y <- as.numeric(Y)
    }
    ## check whether we have a factor or a continuous variable
    if (!(is.factor(Y) | is.numeric(Y))) {
      stop("the y-outcome must be either real or a factor.")
    }
    if (is.factor(Y) || length(coerce.factor$yvar.names) == 1) {
      family <- "class"
    }
    else {
      family <- "regr"
    }
    ytry <- 1
  }
  ## done: return the goodies
  return (list(all.names=all.names, family=family, subj.names=subj.names, yvar.names=yvar.names, ytry=ytry,
               coerce.factor = coerce.factor))
}
is.all.na <- function(x) {all(is.na(x))}
parseMissingData <- function(formula.obj, data) {
  ## parse the formula object
  yvar.names <- formula.obj$yvar.names
  if (length(yvar.names) > 0) {
    resp <- data[, yvar.names, drop = FALSE]
    ## determine whether all the outcomes are missing
    ## works for any dimension
    col.resp.na <- unlist(lapply(data[, yvar.names, drop = FALSE], is.all.na))
    if (any(col.resp.na)) {
      stop("All records are missing for one (or more) yvar(s)")
    }
  }
  ## remove all x variables with missing values for all records
  colPt <- unlist(lapply(data, is.all.na))
  ## terminate if all columns have all missing data
  if (sum(colPt) > 0 && sum(colPt) >= (ncol(data) - length(yvar.names))) {
    stop("All x-variables have all missing data:  analysis not meaningful.")
  }
  data <- data[, !colPt, drop = FALSE]
  ## remove all records with missing values for all outcomes(s) and xvar
  rowPt <- apply(data, 1, is.all.na)
  if (sum(rowPt) == nrow(data)) {
    stop("Rows of the data have all missing data:  analysis not meaningful.")
  }
  data <- data[!rowPt,, drop = FALSE]
  ## return the NA processed data
  return(data)
}
## robust resampling
resample <- function(x, size, ...) {
  if (length(x) <= 1) {
    if (!missing(size) && size == 0) x[FALSE] else x
  }
    else {
      sample(x, size, ...)
    }
}
## determine rows and columns missing from dat
row.col.deleted <- function(dat, r.n, c.n) {
  which.r <- setdiff(r.n, rownames(dat))
  if (length(which.r) > 0) {
    which.r <- match(which.r, r.n)
  }
    else {
      which.r <- NULL
    }
  which.c <- setdiff(c.n, colnames(dat))
  if (length(which.c) > 0) {
    which.c <- match(which.c, c.n)
  }
    else {
      which.c <- NULL
    }
  return(list(row = which.r, col = which.c))
}

holdout.vimp.rfsrc <- function(formula, data,
                               ntree = function(p, vtry){1000 * p / vtry},
                               nsplit = 10,
                               ntime = 50,
                               sampsize = function(x){x * .632},
                               samptype = "swor",
                               block.size = 10,
                               vtry = 1,
                               ...)
{
  ## --------------------------------------------------------------
  ##   
  ##   preliminary processing
  ##
  ## --------------------------------------------------------------
  ## pull unnamed parameters to pass to rfsrc
  ## lock down options that cannot be used
  dots <- list(...)
  dots$formula <- dots$data <- dots$ntree <- dots$nsplit <- dots$ntime <-
    dots$samp <- dots$sampsize <- dots$samptype <- dots$block.size <-
      dots$importance <- NULL
  dots$terminal.qualts <- FALSE
  ## get dimension p - make a fast single stump call for accurate dimension
  p <- length(rfsrc(formula, data, nodedepth = 1, ntime = 1,
              ntree = 1, splitrule = "random")$xvar.names)
  ## process ntree - is it a function or number?
  ## targeted hold out vimp must be handled carefully - checks are put in place
  if (!is.function(ntree) && !is.numeric(ntree)) {
    stop("ntree must be a function or number specifying requested number of grow trees")
  }
  if (is.function(ntree)) {
    if (!is.list(vtry)) {
      ntree <- ntree(p, vtry)
    }
    ## special case when vtry is a list and user wants targeted hold out vimp
    else {
      xvar <- unlist(vtry[1])
      if (any(xvar < 1) | any(xvar > p)) {
        stop("x-variables specified in vtry should be integer values between 1 and p=number of features")
      }      
      joint <- unlist(vtry[2])
      if (!is.logical(joint)) {
        stop("joint vimp in vtry must be specified as being either TRUE or FALSE")
      }
      if (!joint) {
        ntree <- ntree(length(xvar), 1)
      }
      else {
        if (length(xvar) == p) {
          stop("number of hold out variables for joint analysis must be less than number of features")
        }
        ntree <- ntree(1, 1)
      }
    }
  }
  ## initialize the holdout specs object
  holdout.specs <- get.holdout.specs(vtry, p, ntree, block.size, mode = "baseline")
  ## --------------------------------------------------------------
  ##   
  ##   There are two native-code calls to affect holdout vimp:
  ##
  ##   The first call grows a baseline forest.  This is called
  ##   "baseline" mode, and is specificaly recognized by the native
  ##   code via the incoming holdout specs.  In this mode, an x-var IS
  ##   NOT HELD OUT during growth, but performance is calculated
  ##   over the relevant blocks as if the x-vars WAS held out.
  ##
  ##   The second call grows a holdout forest.  This is called
  ##   "holdout" mode, and is specificaly recognized by the native
  ##   code via the incoming parameters.  In this mode, an x-var IS
  ##   HELD OUT during growth, and performance is calculated over
  ##   the relevant blocks.
  ##
  ##   Note that the trees over which a particular x-var is held
  ##   are determined by the holdout array.  Theses trees are
  ##   divided in blocks of block.size.  Incomplete blocks are
  ##   ignored in the resulting performance calculation.
  ##   Blocks are non-contiguous and specific to each x-var.
  ##
  ## --------------------------------------------------------------
  ## Note that we use the same seed for both forests, in an attempt
  ## to reduce variance. This will only result in identical forests
  ## for both modes if the holdout array is the zero matrix, there
  ## is no missing data, and we do deterministic splitting.  It's a
  ## little complicated, and we should probably test the for
  ## variance without identical seeds.  In fact, in the presence of
  ## missing data, it might be the choice of identical seeds is
  ## irrelevant.
  seed <- dots$seed
  dots$seed <- NULL
  ## initialize the seed
  seed <- get.seed(seed)
  ## baseline (harness) forest call:
  base.obj <- do.call(rfsrc, c(list(formula = formula,
                                    data = data,
                                    ntree = ntree,
                                    nsplit = nsplit,
                                    ntime = ntime,
                                    sampsize = sampsize,
                                    samptype = samptype,
                                    holdout.specs = holdout.specs,
                                    importance = FALSE,
                                    forest = FALSE,
                                    terminal.qualts = FALSE,
                                    seed = seed),
                               dots))
  ## change the specifiction object to holdout mode. 
  holdout.specs <- switch.holdout.specs(holdout.specs, "holdout")
  ## holdout forest call:    
  hold.obj <- do.call(rfsrc, c(list(formula = formula,
                                    data = data,
                                    ntree = ntree,
                                    nsplit = nsplit,
                                    ntime = ntime,
                                    sampsize = sampsize,
                                    samptype = samptype,
                                    holdout.specs = holdout.specs,
                                    importance = FALSE,
                                    forest = FALSE,
                                    terminal.qualts = FALSE,
                                    seed = seed),
                               dots))
  ## We are now in a position to analyze the resulting blocked vimp.
  ## The delta between the two call will give you the holdout vimp.
  ## The objects are multivariate compliant and return the performance in
  ## the response specific lists in the multivariate case.  They follow
  ## the simplified list format in the univaratiate case.
  ## Note: obj$holdout.blk is a vector of length [p] indicating how many blocks
  ## each x-var contains.  This is variable and might be zero if the x-var is
  ## never held out.
  ## --------------------------------------------------------------    
  ## The outputs are as follows:
  ## SURV:  obj$holdout.vimp
  ## To the R code:
  ##   -> of dim [[p]] x [RF_eventTypeSize] x [holdout.blk[.]]
  ##              ^^^
  ##   list element can be null
  ##   if no blocks for this x-var  
  ## CLAS:  obj$holdout.vimp
  ## To the R code:
  ##   -> of dim [[p]] x [1 + levels.count[.]] x [holdout.blk[.]]
  ##              ^^^
  ##   list element can be null
  ##   if no blocks for this x-var  
  ## REGR:  obj$holdout.vimp
  ## To the R code:
  ##   -> of dim [[p]] x [holdout.blk[.]]
  ##              ^^^
  ##   list element can be null
  ##   if no blocks for this x-var  
  ##
  ## --------------------------------------------------------------
  ##cat("\nBlocks in each x-var:  \n")
  ##print(base.obj$holdout.blk)
  mv.flag <- FALSE
  ##--------------------------------
  ##
  ## multivariate object
  ##
  ##--------------------------------
  if ((class(base.obj)[3] == "regr+") | (class(base.obj)[3] == "class+") | (class(base.obj)[3] == "mix+")) { 
    ## set the mv flag
    mv.flag <- TRUE
    ## create the holdout vimp object
    hvimp <- list()
    ## regr outcomes
    if (length(base.obj$regrOutput) > 0) {
      hvimp$regrOutput <- lapply(1:length(base.obj$regrOutput), function(i) {
        hv <- sapply(1:length(base.obj$regrOutput[[i]]$holdout.vimp), function(j) {
          mean(hold.obj$regrOutput[[i]]$holdout.vimp[[j]] - base.obj$regrOutput[[i]]$holdout.vimp[[j]], na.rm = TRUE)
        })
        names(hv) <- names(base.obj$regrOutput[[i]]$holdout.vimp)
        hv
      })
      names(hvimp$regrOutput) <- names(base.obj$regrOutput)
    }
    ## class outcomes
    if (length(base.obj$classOutput) > 0) {
      hvimp$classOutput <- lapply(1:length(base.obj$classOutput), function(i) {
        yval.i <- colnames(base.obj$classOutput[[i]]$predicted)
        hv <- do.call(rbind, lapply(1:length(base.obj$classOutput[[i]]$holdout.vimp), function(j) {
          if (!is.null(dim(base.obj$classOutput[[i]]$holdout.vimp[[j]]))) {
            rowMeans(hold.obj$classOutput[[i]]$holdout.vimp[[j]] - base.obj$classOutput[[i]]$holdout.vimp[[j]], na.rm = TRUE)
          }
          else {
            rep(NA, length(yval.i) + 1)
          }
        }))
        colnames(hv) <- c("all", yval.i)
        rownames(hv) <- names(base.obj$classOutput[[i]]$holdout.vimp)
        hv
      })
      names(hvimp$classOutput) <- names(base.obj$classOutput)
    }
  }
  ##--------------------------------
  ##
  ## regression object
  ##
  ##--------------------------------
  else if (class(base.obj)[3] == "regr") {  
    hvimp <- sapply(1:length(base.obj$holdout.vimp), function(j) {
      if (!is.null(base.obj$holdout.vimp[[j]])) {
        mean(hold.obj$holdout.vimp[[j]] - base.obj$holdout.vimp[[j]], na.rm = TRUE)
      }
      else {
        NA
      }
    })
    names(hvimp) <- names(base.obj$holdout.vimp)
  }
  ##--------------------------------
  ##
  ## classification object
  ##
  ##--------------------------------
  else if (class(base.obj)[3] == "class") {  
    yval <- levels(base.obj$yvar)
    hvimp <- do.call(rbind, lapply(1:length(base.obj$holdout.vimp), function(j) {
      if (!is.null(dim(base.obj$holdout.vimp[[j]]))) {
        rowMeans(hold.obj$holdout.vimp[[j]] - base.obj$holdout.vimp[[j]], na.rm = TRUE)
      }
      else {
        rep(NA, length(yval) + 1)
      }
    }))
    colnames(hvimp) <- c("all", yval)
    rownames(hvimp) <- names(base.obj$holdout.vimp)
  }
  ##--------------------------------
  ##
  ## survival, CR objects
  ##
  ##--------------------------------
  else if (class(base.obj)[3] == "surv") {
    hvimp <- sapply(1:length(base.obj$holdout.vimp), function(j) {
      if (!is.null(base.obj$holdout.vimp[[j]])) {
        mean(hold.obj$holdout.vimp[[j]] - base.obj$holdout.vimp[[j]], na.rm = TRUE)
      }
      else {
        rep(NA)
      }
    })
    names(hvimp) <- names(base.obj$holdout.vimp)
  }
  else if (class(base.obj)[3] == "surv-CR") {
    hvimp <- do.call(rbind, lapply(1:length(base.obj$holdout.vimp), function(j) {
      if (!is.null(dim(base.obj$holdout.vimp[[j]]))) {
        rowMeans(hold.obj$holdout.vimp[[j]] - base.obj$holdout.vimp[[j]], na.rm = TRUE)
      }
      else {
        rep(NA, length(base.obj$holdout.vimp))
      }
    }))
    colnames(hvimp) <- paste("event.", 1:length(get.event.info(base.obj)$event.type), sep = "")
    rownames(hvimp) <- names(base.obj$holdout.vimp)
  }
  ## return the promised holdout vimp
  ## processing is different for multivariate families
  rO <- list()
  if (mv.flag) {
    rO$baseline <- rO$holdout <- rO$importance <- list()
    if (length(base.obj$regrOutput) > 0) {
      rO$baseline$regrOutput <- lapply(base.obj$regrOutput, function(o) {o$holdout.vimp})
      rO$holdout$regrOutput <- lapply(hold.obj$regrOutput, function(o) {o$holdout.vimp})
      rO$importance$regrOutput <- hvimp$regrOutput
    }
    if (length(base.obj$classOutput) > 0) {
      rO$baseline$classOutput <- lapply(base.obj$classOutput, function(o) {o$holdout.vimp})
      rO$holdout$classOutput <- lapply(hold.obj$classOutput, function(o) {o$holdout.vimp})
      rO$importance$classOutput <- hvimp$classOutput
    }
  }
  else {
    rO$baseline <- base.obj$holdout.vimp
    rO$holdout <- hold.obj$holdout.vimp
    rO$importance <- hvimp
  }
  return(invisible(rO))
}
holdout.vimp <- holdout.vimp.rfsrc
## --------------------------------------------------------------
##
## function to create holdout native code parameters
## - no checks on block.size are made
##
## --------------------------------------------------------------
get.holdout.specs <- function(vtry, p, ntree, block.size, mode) {
  ## make holdout array
  ## check that vtry is not a list --> default case and default processing
  if (!is.list(vtry)) {
    if (vtry > 0) {
      holdout <- make.holdout.array(vtry, p, ntree, NULL)
    }
    else {
      stop("vtry must be positive")
    }
  }
  ## vtry is a list --> user wants targeted hold out vimp
  else {
    ## parse vtry
    xvar <- unlist(vtry[1])
    joint <- unlist(vtry[2])
    ## set vtry to integer value
    if (!joint) {
      vtry <- 1
    }
    else {
      vtry <- length(xvar)
    }
    if (!joint) {
      holdout <- do.call(cbind, lapply(1:ntree, function(b) {
        ho <- rep(0, p)
        ho[resample(xvar, size = 1, replace = FALSE)] <- 1
        ho
      }))
    }
    else {
      holdout <- matrix(0, p, ntree)
      holdout[xvar, ] <- 1
    }
  }
  ## assemble the return object and return it
  obj <- list(as.integer(vtry), as.integer(holdout), as.integer(block.size), as.integer(0))
  names(obj) = c("vtry", "holdout", "block.size", "mode")
  return(switch.holdout.specs(obj, mode))
}
## --------------------------------------------------------------
##
## function to switch holdout mode for native code parameters
## mode changes between "baseline" <-> "holdout"
##
## --------------------------------------------------------------
switch.holdout.specs<- function(obj, mode) {
  if (mode == "baseline") {
    new.mode = 1
  }
  else if (mode == "holdout") {
    new.mode = 2
  }
  else if (mode == "natural") {
    new.mode = 3
  }
  else {
    stop("vtry mode invalid")
  }
  obj$mode = as.integer(new.mode)
  return(obj)
}

# internal-utils.R -- Dummy file for documentation linkage
# This file does not define actual functions. It exists to tie together 
# the manual documentation (man/internal-utils.Rd) with the exported names,
# preventing \"Undocumented object\" warnings during R CMD check.
NULL

predict.rfsrc <-  function(object,
   newdata,
   importance = c(FALSE, TRUE, "none", "anti", "permute", "random"),
   get.tree = NULL,
   block.size = if (any(is.element(as.character(importance), c("none", "FALSE")))) NULL else 10,
   na.action = c("na.omit", "na.impute", "na.random"),
   outcome = c("train", "test"),
   perf.type = NULL,
   proximity = FALSE,
   forest.wt = FALSE,
   ptn.count = 0,
   distance = FALSE,
   var.used = c(FALSE, "all.trees", "by.tree"),
   split.depth = c(FALSE, "all.trees", "by.tree"),
   case.depth = FALSE,
   seed = NULL,
   do.trace = FALSE,
   membership = FALSE,
   marginal.xvar = NULL,
   ...)
{
  dots <- list(...)
  m.target <- dots$m.target
  dots$m.target <- NULL
  importance.xvar <- dots$importance.xvar
  dots$importance.xvar <- NULL
  args <- c(list(
    object = object,
    m.target = m.target,
    importance = importance,
    get.tree = get.tree,
    block.size = block.size,
    importance.xvar = importance.xvar,
    na.action = na.action,
    outcome = outcome,
    perf.type = perf.type,
    proximity = proximity,
    forest.wt = forest.wt,
    ptn.count = ptn.count,
    distance = distance,
    var.used = var.used,
    split.depth = split.depth,
    case.depth = case.depth,
    seed = seed,
    do.trace = do.trace,
    membership = membership,
    marginal.xvar = marginal.xvar
  ), dots)
  ## if newdata is not missing
  if (!missing(newdata)) {
    args$newdata <- newdata
  }
  return(do.call("generic.predict.rfsrc", args))
}

tune.rfsrc <- function(formula, data,
      mtry.start = ncol(data) / 2,
      nodesize.try = c(1:9, seq(10, 100, by = 5)), ntree.try = 100,
      sampsize = function(x){min(x * .632, max(150, x ^ (3/4)))},
      nsplit = 1, step.factor = 1.25, improve = 1e-3, strikeout = 3, max.iter = 25,
      method = c("grid","golden"),
      ## golden controls (ignored for grid)
      final.window = 5L, reps.initial = 2L, reps.final = 3L,
      trace = FALSE, do.best = TRUE, seed = NULL, ...) {
  method <- match.arg(method)
  ## input checks 
  if (improve < 0) stop("improve must be non-negative.")
  if (step.factor <= 1) stop("step.factor must be great than 1.")
  if (missing(formula)) stop("a formula must be supplied (only supervised forests allowed).")
  ## capture/sanitize dots (avoid duplicate formals)
  dots <- list(...)
  perf.type <- if ("perf.type" %in% names(dots)) dots$perf.type else "default"
  drop.from.dots <- c("perf.type","forest","save.memory","ntree","mtry","nodesize",
                      "sampsize","nsplit","data","formula")
  if (length(dots)) dots <- dots[setdiff(names(dots), drop.from.dots)]
  ## reproducibility and a base seed for CRN
  have.old.seed <- exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  if (have.old.seed) old.seed <- get(".Random.seed", envir = .GlobalEnv)
  on.exit({
    if (exists("old.seed", inherits = FALSE)) assign(".Random.seed", old.seed, envir = .GlobalEnv)
  }, add = TRUE)
  if (!is.null(seed)) set.seed(seed)
  base.seed <- if (is.null(seed)) sample.int(.Machine$integer.max, 1L) else seed
  ## clean model frame via stump (handles NA & factor levels consistently)
  stump <- rfsrc(formula, data, nodedepth = 0, perf.type = "none",
                 save.memory = TRUE, ntree = 1, splitrule = "random")
  n <- stump$n
  yvar.names <- stump$yvar.names
  data <- data.frame(stump$yvar, stump$xvar)
  colnames(data)[seq_along(yvar.names)] <- yvar.names
  p <- ncol(data) - length(yvar.names)  ## number of predictors
  rm(stump)
  ## effective numeric ssize for tuning and holdout
  if (is.function(sampsize)) {
    ssize <- sampsize(n)
  } else if (is.numeric(sampsize) && length(sampsize) == 1L) {
    ssize <- sampsize
  } else if (is.numeric(sampsize) && length(sampsize) > 1L) {
    ssize <- sum(sampsize)
  } else {
    stop("'sampsize' must be a function or numeric (scalar or vector).")
  }
  ssize <- max(2L, min(n, as.integer(round(ssize))))
  if (trace) cat("effective sampsize:", ssize, "of", n, "\n")
  ## holdout equal to ssize if possible; else rely on OOB
  use.holdout <- (2L * ssize) < n
  if (use.holdout) {
    tst <- sample.int(n, ssize)
    trn <- setdiff(seq_len(n), tst)
    newdata <- data[tst, , drop = FALSE]
  } else {
    trn <- seq_len(n)
    newdata <- NULL
  }
  ## admissible nodesize bound
  U.calc <- max(10L, ssize %/% 2L)
  ## evaluation kernel for a pair (nodesize, mtry) with CRN
  eval.pair <- function(nsz, mtry, reps) {
    vals <- numeric(reps)
    for (j in seq_len(reps)) {
      s <- as.integer((base.seed + 10007L * as.integer(nsz) + 101L * as.integer(mtry) + j) %% .Machine$integer.max)
      if (s <= 0L) s <- j
      set.seed(s)
      if (is.null(newdata)) {
        tr.args <- c(list(formula = formula,
                          data = data[trn, , drop = FALSE],
                          ntree = ntree.try,
                          mtry = mtry,
                          nodesize = nsz,
                          sampsize = ssize,      # numeric subsample during tuning
                          nsplit = nsplit,
                          perf.type = perf.type,
                          forest = FALSE,
                          save.memory = TRUE),
                     dots)
        fit <- tryCatch(do.call(rfsrc.fast, tr.args), error = function(e) e)
        if (inherits(fit, "error")) return(NA_real_)
        vals[j] <- tryCatch(mean(get.mv.error(fit, TRUE), na.rm = TRUE), error = function(e) NA_real_)
      } else {
        tr.args <- c(list(formula = formula,
                          data = data[trn, , drop = FALSE],
                          ntree = ntree.try,
                          mtry = mtry,
                          nodesize = nsz,
                          sampsize = ssize,
                          nsplit = nsplit,
                          perf.type = "none",
                          forest = TRUE,
                          save.memory = FALSE),
                     dots)
        fit <- tryCatch(do.call(rfsrc.fast, tr.args), error = function(e) e)
        if (inherits(fit, "error")) return(NA_real_)
        pr <- tryCatch(predict(fit, newdata = newdata, perf.type = perf.type), error = function(e) e)
        if (inherits(pr, "error")) return(NA_real_)
        vals[j] <- tryCatch(mean(get.mv.error(pr, TRUE), na.rm = TRUE), error = function(e) NA_real_)
      }
    }
    if (trace && is.finite(mean(vals))) {
      cat("nodesize =", nsz, " mtry =", mtry, " reps =", reps,
          " error =", paste0(100 * round(mean(vals), 4), "%"), "\n")
    }
    mean(vals)
  }
  ## ---------------------------------------
  ## method == "grid": original method
  ## ---------------------------------------
  if (method == "grid") {
    grid.cand <- unique(sort(as.integer(nodesize.try[nodesize.try <= U.calc])))
    if (!length(grid.cand)) stop("No admissible 'nodesize.try' after filtering by ssize/2.")
    res <- list()
    counter1 <- 0
    for (nsz in grid.cand) {
      counter1 <- counter1 + 1
      ## acquire starting error at mtry.start
      if (is.null(newdata)) {
        o <- tryCatch(do.call(rfsrc.fast,
                              c(list(formula = formula, data = data,
                                     ntree = ntree.try, mtry = mtry.start, nodesize = nsz,
                                     sampsize = ssize, nsplit = nsplit, perf.type = perf.type),
                                dots)),
                      error = function(e) e)
        if (inherits(o, "error")) stop(conditionMessage(o))
        error.old <- mean(get.mv.error(o, TRUE), na.rm = TRUE)
      } else {
        o <- tryCatch(do.call(rfsrc.fast,
                              c(list(formula = formula, data = data[trn, , drop = FALSE],
                                     ntree = ntree.try, mtry = mtry.start, nodesize = nsz,
                                     sampsize = ssize, nsplit = nsplit, perf.type = "none",
                                     forest = TRUE, save.memory = FALSE),
                                dots)),
                      error = function(e) e)
        if (inherits(o, "error")) stop(conditionMessage(o))
        error.old <- mean(get.mv.error(predict(o, newdata, perf.type = perf.type), TRUE), na.rm = TRUE)
      }
      mtry.start <- o$mtry
      mtry.max <- p
      if (trace) {
        cat("nodesize = ", nsz, " mtry =", mtry.start,
            " error =", paste(100 * round(error.old, 4), "%", sep = ""), "\n")
      }
      if (!is.null(o$leaf.count) && mean(o$leaf.count) <= 2) break
      oob.error <- list()
      oob.error[[1]] <- error.old
      names(oob.error)[1] <- mtry.start
      for (direction in c("left", "right")) {
        if (trace) cat("Searching", direction, "...\n")
        Improve <- 1.1 * improve
        mtry.best <- mtry.start
        mtry.cur <- mtry.start
        counter2 <- 1
        strikes <- 0
        while (counter2 <= max.iter && (Improve >= improve || (Improve < 0 && strikes < strikeout))) {
          counter2 <- counter2 + 1
          if (Improve < 0) strikes <- strikes + 1
          mtry.old <- mtry.cur
          if (direction == "left") {
            mtry.cur <- max(1, min(ceiling(mtry.cur / step.factor), mtry.cur - 1))
          } else {
            mtry.cur <- min(mtry.max, max(floor(mtry.cur * step.factor), mtry.cur + 1))
          }
          if (mtry.cur == mtry.old) break
          error.cur <- eval.pair(nsz, mtry.cur, reps = 1L)
          if (trace) {
            cat("nodesize = ", nsz, " mtry =", mtry.cur,
                " error =", paste(100 * round(error.cur, 4), "%", sep = ""), "\n")
          }
          oob.error[[as.character(mtry.cur)]] <- error.cur
          Improve <- 1 - error.cur / error.old
          if (trace) cat(Improve, improve, "\n")
          if (Improve > improve) {
            error.old <- error.cur
            mtry.best <- mtry.cur
          }
        }
      }
      mtry.vec <- sort(as.numeric(names(oob.error)))
      err.vec <- unlist(oob.error[as.character(mtry.vec)])
      res[[counter1]] <- cbind(nodesize = nsz, mtry = mtry.vec, err = err.vec)
    }
    if (is.null(res)) stop("NULL results - check tuning settings.")
    res <- do.call(rbind, res)
    res <- res[order(res[, 1], res[, 2]), , drop = FALSE]
    rownames(res) <- seq_len(nrow(res))
    opt.idx <- which.min(res[, 3])
    rf <- NULL
    if (do.best) {
      rf <- do.call(rfsrc.fast,
                    c(list(formula = formula, data = data,
                           mtry = res[opt.idx, 2], nodesize = res[opt.idx, 1],
                           sampsize = sampsize, nsplit = nsplit, forest = TRUE),
                      dots))
    }
    return(list(results = res, optimal = res[opt.idx, -3], rf = rf))
  }
  ## ---------------------------------------
  ## method == "golden": guarded coordinate line search
  ## ---------------------------------------
  ## 1D golden on nodesize with left guard; final local sweep
  golden.search.nodesize <- function(mfix) {
    L <- 1L; U <- U.calc
    if (U <= L) stop("Admissible nodesize range is empty; increase 'ssize' or check data.")
    left.max <- min(9L, U)
    left.cand <- seq.int(1L, left.max)
    left.err <- vapply(left.cand, function(nsz) eval.pair(nsz, mfix, reps.initial), numeric(1))
    best.left.idx <- which.min(left.err); best.left.nsz <- left.cand[best.left.idx]
    best.left.err <- left.err[best.left.idx]
    phi <- (1 + sqrt(5)) / 2; invphi <- 1 / phi
    xL2 <- max(L, min(U, floor(U - invphi * (U - L))))
    xR2 <- max(L, min(U, ceiling(L + invphi * (U - L))))
    if (xL2 == xR2) xR2 <- min(U, xL2 + 1L)
    fL2 <- eval.pair(xL2, mfix, reps.initial)
    fR2 <- eval.pair(xR2, mfix, reps.initial)
    if (is.finite(best.left.err) && (best.left.err <= min(fL2, fR2, na.rm = TRUE))) {
      Lf <- max(L, best.left.nsz - final.window)
      Uf <- min(U, best.left.nsz + final.window)
      cand <- sort(unique(c(seq.int(Lf, Uf), left.cand)))
      err <- vapply(cand, function(nsz) eval.pair(nsz, mfix, reps.final), numeric(1))
      return(list(best = cand[which.min(err)], grid = data.frame(nodesize = cand, mtry = mfix, err = err)))
    }
    iter <- 0L
    while ((U - L) > final.window && iter < max.iter) {
      if (!is.finite(fL2) || !is.finite(fR2)) break
      if (fR2 < fL2) {
        L <- xL2; xL2 <- xR2; fL2 <- fR2
        xR2 <- max(L, min(U, ceiling(L + invphi * (U - L))))
        if (xR2 == xL2) xR2 <- min(U, xR2 + 1L)
        fR2 <- eval.pair(xR2, mfix, reps.initial)
      } else {
        U <- xR2; xR2 <- xL2; fR2 <- fL2
        xL2 <- max(L, min(U, floor(U - invphi * (U - L))))
        if (xL2 == xR2) xL2 <- max(L, xL2 - 1L)
        fL2 <- eval.pair(xL2, mfix, reps.initial)
      }
      iter <- iter + 1L
    }
    cand <- sort(unique(c(seq.int(L, U), left.cand)))
    err <- vapply(cand, function(nsz) eval.pair(nsz, mfix, reps.final), numeric(1))
    list(best = cand[which.min(err)], grid = data.frame(nodesize = cand, mtry = mfix, err = err))
  }
  ## 1D golden on mtry with left guard; final local sweep
  golden.search.mtry <- function(nfix) {
    L <- 1L; U <- p
    if (U <= L) stop("Admissible mtry range is empty; check data.")
    left.max <- min(9L, U)
    left.cand <- seq.int(1L, left.max)
    left.err <- vapply(left.cand, function(m) eval.pair(nfix, m, reps.initial), numeric(1))
    best.left.idx <- which.min(left.err); best.left.m <- left.cand[best.left.idx]
    best.left.err <- left.err[best.left.idx]
    phi <- (1 + sqrt(5)) / 2; invphi <- 1 / phi
    xL2 <- max(L, min(U, floor(U - invphi * (U - L))))
    xR2 <- max(L, min(U, ceiling(L + invphi * (U - L))))
    if (xL2 == xR2) xR2 <- min(U, xL2 + 1L)
    fL2 <- eval.pair(nfix, xL2, reps.initial)
    fR2 <- eval.pair(nfix, xR2, reps.initial)
    if (is.finite(best.left.err) && (best.left.err <= min(fL2, fR2, na.rm = TRUE))) {
      Lf <- max(L, best.left.m - final.window)
      Uf <- min(U, best.left.m + final.window)
      cand <- sort(unique(c(seq.int(Lf, Uf), left.cand)))
      err <- vapply(cand, function(m) eval.pair(nfix, m, reps.final), numeric(1))
      return(list(best = cand[which.min(err)], grid = data.frame(nodesize = nfix, mtry = cand, err = err)))
    }
    iter <- 0L
    while ((U - L) > final.window && iter < max.iter) {
      if (!is.finite(fL2) || !is.finite(fR2)) break
      if (fR2 < fL2) {
        L <- xL2; xL2 <- xR2; fL2 <- fR2
        xR2 <- max(L, min(U, ceiling(L + invphi * (U - L))))
        if (xR2 == xL2) xR2 <- min(U, xR2 + 1L)
        fR2 <- eval.pair(nfix, xR2, reps.initial)
      } else {
        U <- xR2; xR2 <- xL2; fR2 <- fL2
        xL2 <- max(L, min(U, floor(U - invphi * (U - L))))
        if (xL2 == xR2) xL2 <- max(L, xL2 - 1L)
        fL2 <- eval.pair(nfix, xL2, reps.initial)
      }
      iter <- iter + 1L
    }
    cand <- sort(unique(c(seq.int(L, U), left.cand)))
    err <- vapply(cand, function(m) eval.pair(nfix, m, reps.final), numeric(1))
    list(best = cand[which.min(err)], grid = data.frame(nodesize = nfix, mtry = cand, err = err))
  }
  ## coordinate descent: alternate golden on nodesize and mtry
  res.grid <- list()
  res.row <- 0L
  add.grid <- function(df) {
    if (is.null(df) || !nrow(df)) return()
    res.row <<- res.row + 1L
    res.grid[[res.row]] <<- df
  }
  mtry.cur <- as.integer(max(1L, min(p, round(mtry.start))))
  best.err <- Inf
  strikes <- 0L
  for (iter.cd in seq_len(max.iter)) {
    nsres <- golden.search.nodesize(mtry.cur)
    add.grid(nsres$grid)
    nsz.cur <- nsres$best
    mtres <- golden.search.mtry(nsz.cur)
    add.grid(mtres$grid)
    mtry.new <- mtres$best
    cur.err <- eval.pair(nsz.cur, mtry.new, reps.final)
    if (trace) {
      cat("CD iter", iter.cd, ": nodesize =", nsz.cur, " mtry =", mtry.new,
          " error =", paste0(100 * round(cur.err, 4), "%"), "\n")
    }
    if (is.finite(best.err)) {
      rel.imp <- 1 - cur.err / best.err
      if (rel.imp <= improve) {
        strikes <- strikes + 1L
      } else {
        strikes <- 0L
      }
      if (rel.imp <= improve && strikes >= strikeout) break
    }
    if (cur.err < best.err) best.err <- cur.err
    mtry.cur <- mtry.new
  }
  if (!length(res.grid)) stop("No evaluations; check settings.")
  res <- do.call(rbind, res.grid)
  res <- stats::aggregate(err ~ nodesize + mtry, data = res, FUN = mean)  # average if duplicates
  res <- res[order(res$nodesize, res$mtry), , drop = FALSE]
  colnames(res) <- c("nodesize", "mtry", "err")
  opt.idx <- which.min(res$err)
  opt.nsz <- res$nodesize[opt.idx]
  opt.mtry <- res$mtry[opt.idx]
  rf <- NULL
  if (do.best) {
    rf <- do.call(rfsrc.fast,
                  c(list(formula = formula, data = data,
                         mtry = opt.mtry, nodesize = opt.nsz,
                         sampsize = sampsize, nsplit = nsplit, forest = TRUE),
                    dots))
  }
  list(results = as.matrix(res), optimal = c(nodesize = opt.nsz, mtry = opt.mtry), rf = rf)
}
tune <- tune.rfsrc

vimp.rfsrc <- function(object,
                       xvar.names,
                       importance = c("anti", "permute", "random"),
                       block.size = 10,
                       joint = FALSE,
                       seed = NULL,
                       do.trace = FALSE,
                       ...)
{
  ## incoming parameter checks - all are fatal
  if (missing(object)) {
    stop("object is missing")
  }
  if (object$family == "unsupv") {
    stop("vimp does not apply to unsupervised forests: consider using max.subtree or varpro")
  }
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2    &
      sum(inherits(object, c("rfsrc", "forest"), TRUE) == c(1, 2)) != 2) {
    stop("This function only works for objects of class `(rfsrc, grow)' or '(rfsrc, forest)'")
  }
  ## process the importance specification
  if (!is.logical(joint)) {
    stop("joint must be a logical value")
  }
  importance <- importance[1]
  if (joint & importance != "none") {
    i.str <- unlist(strsplit(importance, "\\."))
    if (length(i.str) == 1) {
      importance <- paste(i.str[1], ".joint", sep = "")
    }
      else if (length(i.str) == 2) {
        importance <- paste(i.str[1], ".joint.", i.str[2], sep = "")
      }
  }
  importance <- match.arg(as.character(importance),
      c(TRUE, "anti", "permute", "random", "anti.joint", "permute.joint", "random.joint"))
  ## grow objects under non-standard bootstrapping are devoid of performance values
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) == 2) {
    if (is.null(object$forest)) {
      stop("The forest is empty.  Re-run rfsrc (grow) call with forest=TRUE")
    }
      else {
        bootstrap <- object$forest$bootstrap
      }
  }
    else {
      bootstrap <- object$bootstrap
    }
  if (bootstrap == "none" || bootstrap == "by.node") {
    stop("grow objects under non-standard bootstrapping are devoid of performance values")
  }
  ## legacy m.target
  dots <- list(...)
  m.target <- dots$m.target
  dots$m.target <- NULL
  ## generic predict call
  args <- c(list(
    object = object,
    m.target = m.target,
    importance = importance,
    block.size = block.size,
    seed = seed,
    do.trace = do.trace,
    membership = FALSE
    ), dots)
  ## if xvar.names is not missing  
  if (!missing(xvar.names)) {
    args$importance.xvar <- xvar.names
  }
  return(do.call("generic.predict.rfsrc", args))
}
vimp <- vimp.rfsrc

get.varselect.imp <- function(f.o, target.dim) {
  if (!is.null(f.o$importance)) {
    c(cbind(f.o$importance)[, target.dim])
  }
    else {
      rep(NA, length(f.o$xvar.names))
    }
}
get.varselect.imp.all <- function(f.o) {
  if (!is.null(f.o$importance)) {
    imp.all <- cbind(f.o$importance)
    if (ncol(imp.all) == 1) {
      colnames(imp.all) <- "vimp"
    }
      else {
        colnames(imp.all) <- paste("vimp.", colnames(imp.all), sep = "")
      }
    imp.all
  }
    else {
      rep(NA, length(f.o$xvar.names))
    }
}
get.varselect.err <- function(f.o) {
  if (!is.null(f.o$err.rate)) {
    if (grepl("surv", f.o$family)) {
      err <- 100 * cbind(f.o$err.rate)[f.o$ntree, ]
    }
      else {
        err <- cbind(f.o$err.rate)[f.o$ntree, ]
      }
  }
    else {
      err = NA
    }
  err
}
get.varselect.length <- function(x, y) {
  (length(x) > 0 & length(y) > 0)
}
get.varselect.mtry <- function(x, y) {
  mtry <- round((length(x) - length(y))/3)
  if (mtry == 0) {
    round(length(x)/3)
  }
    else {
      mtry
    }
}
get.varselect.sd <- function(x) {
  if (all(is.na(x))) {
    NA
  }
    else {
      sd(x, na.rm = TRUE)
    }
}
permute.rows <-function(x) {
  n <- nrow(x)
  p <- ncol(x)
  mm <- runif(length(x)) + rep(seq(n) * 10, rep(p, n))
  matrix(t(x)[order(mm)], n, p, byrow = TRUE)
}
balanced.folds <- function(y, nfolds = min(min(table(y)), 10)) {
  y[is.na(y)] <- resample(y[!is.na(y)], size = sum(is.na(y)), replace = TRUE)
  totals <- table(y)
  if (length(totals) < 2) {
    return(cv.folds(length(y), nfolds))
  }
    else {
      fmax <- max(totals)
      nfolds <- min(nfolds, fmax)     
      nfolds <- max(nfolds, 2)
      folds <- as.list(seq(nfolds))
      yids <- split(seq(y), y) 
      bigmat <- matrix(NA, ceiling(fmax/nfolds) * nfolds, length(totals))
      for(i in seq(totals)) {
        if(length(yids[[i]])>1){bigmat[seq(totals[i]), i] <- sample(yids[[i]])}
        if(length(yids[[i]])==1){bigmat[seq(totals[i]), i] <- yids[[i]]}
      }
      smallmat <- matrix(bigmat, nrow = nfolds)
      smallmat <- permute.rows(t(smallmat)) 
      res <- vector("list", nfolds)
      for(j in 1:nfolds) {
        jj <- !is.na(smallmat[, j])
        res[[j]] <- smallmat[jj, j]
      }
      return(res)
    }
}

get.tree.rfsrc <- function(object,
                           tree.id,
                           target,
                           m.target = NULL,
                           time,
                           surv.type = c("mort", "rel.freq", "surv", "years.lost", "cif", "chf"),
                           class.type = c("bayes", "rfq", "prob"),
                           ensemble = FALSE,
                           oob = TRUE,
                           show.plots = TRUE,
                           do.trace = FALSE)
{
  ##----------------------------------------------------------------
  ##
  ## tolerance value for dealing with float issues related to
  ## splitting left at an exact split value
  ##
  ##----------------------------------------------------------------
  tolerance <- sqrt(.Machine$double.eps)
  ##----------------------------------------------------------------
  ##
  ## The following two utilities were copied from the BMS CRAN
  ## package.  Thanks to Martin Feldkircher and Stefan Zeugne for
  ## these little quickies.
  ##
  ##----------------------------------------------------------------
  hex2bin <- function (hexcode) 
  {
      if (!is.character(hexcode)) 
          stop("please input a character like '0af34c'")
      hexcode <- paste("0", tolower(hexcode), sep = "")
      hexobj <- .hexcode.binvec.convert(length(hexcode) * 16L)
      return(hexobj$as.binvec(hexcode))
  }
  .hexcode.binvec.convert <- function (length.of.binvec) 
  {
      if (length(length.of.binvec) > 1) 
          length.of.binvec = length(length.of.binvec)
      addpositions = 4 - length.of.binvec%%4
      positionsby4 = (length.of.binvec + addpositions)/4
      hexvec = c(0:9, "a", "b", "c", "d", "e", "f")
      hexcodelist = list(`0` = numeric(4),
                         `1` = c(0, 0, 0, 1), 
                         `2` = c(0, 0, 1, 0),
                         `3` = c(0, 0, 1, 1),
                         `4` = c(0, 1, 0, 0),
                         `5` = c(0, 1, 0, 1),
                         `6` = c(0, 1, 1, 0), 
                         `7` = c(0, 1, 1, 1),
                         `8` = c(1, 0, 0, 0),
                         `9` = c(1, 0, 0, 1),
                         a   = c(1, 0, 1, 0),
                         b   = c(1, 0, 1, 1),
                         c   = c(1, 1, 0, 0),
                         d   = c(1, 1, 0, 1),
                         e   = c(1, 1, 1, 0),
                         f   = c(1, 1, 1, 1))
      return(list(
          as.hexcode = function(binvec) {
              incl = c(numeric(addpositions), binvec)
              dim(incl) = c(4, positionsby4)
              return(paste(hexvec[crossprod(incl, 2L^(3:0)) + 1], collapse = ""))
          },
          as.binvec = function(hexcode) {
              return(unlist(
                  hexcodelist[unlist(strsplit(hexcode, "", fixed = TRUE),
                                     recursive = FALSE, use.names = FALSE)], 
                  recursive = FALSE, use.names = FALSE)[-(1:addpositions)])
          }))
  }
  ##----------------------------------------------------------------
  ##
  ## coherence checks
  ##
  ##----------------------------------------------------------------
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2) {
    stop("This function only works for objects of class `(rfsrc, grow)'")
  }
  if (is.forest.missing(object)) {
    stop("forest is missing.  Re-run rfsrc (grow call) with forest=TRUE")
  }
  if (inherits(object, "anonymous")) {
    ## anonymous <- TRUE
    stop("get.tree does not currently work with anonymous forests\n")
  }
  else {
    anonymous <- FALSE
  }
  ## only one tree allowed
  tree.id <- tree.id[1]
  ## must be integer from 1...ntree
  if (tree.id < 1 || tree.id > object$ntree) {
    stop("tree id must be integer from 1 to ntree")
  }
  ## ensure coherency of the multivariate target
  m.target <- get.univariate.target(object, m.target)
  ##----------------------------------------------------------------
  ##
  ##
  ## take care of some preliminary processing for prediction
  ## this will be used for labels on terminal nodes
  ##
  ##
  ##----------------------------------------------------------------
  ## process the object depending on the underlying family
  family <- object$forest$family
  ## coherence for prediction
  predict.flag <- TRUE
  if (family == "unsupv" || family == "surv-TDC") {
    predict.flag <- FALSE
  }
  if (predict.flag) {
    ## survival and competing risk families.
    if (grepl("surv", family)) {
      ## extract event information
      event.info <- object$event.info
      cens <- event.info$cens
      event.type <- event.info$event.type
      ## assign time if missing
      if (missing(time)) {
        time <- median(event.info$time.interest, na.rm = TRUE)
      }
      ## check for single time point
      if (length(time) > 1) {
        stop("time must be a single value:  ", time)
      }
      ## competing risk
      if (family == "surv-CR") {
        if (missing(target)) {
          target <- 1
        }
        else {
          if (target < 1 || target > max(event.type, na.rm = TRUE)) {
            stop("'target' is specified incorrectly")
          }
        }
        ## set the surv.type
        surv.type <- setdiff(surv.type, c("mort", "rel.freq", "surv"))[1]
        pred.type <- match.arg(surv.type, c("years.lost", "cif", "chf"))
      }
      ## survival
      else {
        target <- 1
        ## set the surv.type
        surv.type <- setdiff(surv.type, c("years.lost", "cif", "chf"))[1]
        pred.type <- match.arg(surv.type, c("rel.freq", "mort", "chf", "surv"))
      }
    }
    ## univariate and multivariate families.
    else {
      ## assign a null time value
      event.info <- time <- NULL
      ## factor outcome
      if(is.factor(coerce.multivariate(object, m.target)$yvar)) {
        object.yvar.levels <- levels(coerce.multivariate(object, m.target)$yvar)
        pred.type <- match.arg(class.type, c("bayes", "rfq", "prob"))
        if (missing(target)) {
          target <- object.yvar.levels[1]
        }
        if (is.character(target)) {
          target <- match(match.arg(target, object.yvar.levels), object.yvar.levels)
        }
        else {
          if ((target > length(object.yvar.levels)) | (target < 1)) {
            stop("target is specified incorrectly:", target)
          }
        }
      }
      ## regression or unsupervised
      else {
        pred.type <- "y"
        target <- NULL
      }
    }
  }
  ##----------------------------------------------------------------
  ##
  ## define the target subset of cases
  ## if ensemble=TRUE  --> this is the entire data
  ## if ensemble=FALSE --> this is inbag or oob cases
  ##
  ##----------------------------------------------------------------
  if (ensemble) {
    subset <- 1:object$n
  }
  ## restrict the object to the target
  ## oob not allowed
  else {
    object <- predict(object, get.tree = tree.id, membership = TRUE, do.trace = do.trace)
    subset <- which(object$inbag[, tree.id] != 0)
    oob <- FALSE
  }
  ##----------------------------------------------------------------
  ##
  ## extract x data for the target subset
  ## convert the data to numeric mode, apply the na.action protocol
  ## missing data not allowed
  ##
  ##----------------------------------------------------------------
  xvar.names <- object$forest$xvar.names
  xvar.factor <- object$forest$xvar.factor
  if (!anonymous) {
    x.data <- object$forest$xvar
    if (any(is.na(x.data))) {
      stop("missing data not allowed")
    }
    x.data <- finalizeData(xvar.names, x.data, miss.flag = FALSE)
    x.data <- x.data[subset,, drop = FALSE]
  }
  ##----------------------------------------------------------------
  ##
  ## now acquire the predicted values for the tree labels
  ##
  ##----------------------------------------------------------------
  if (predict.flag) {
    if (ensemble) {
      object <- predict.rfsrc(object, m.target = m.target, do.trace = do.trace)
    }
    yhat <- extract.pred(object, pred.type, subset, time, m.target, target, oob = oob)
  }
  ##----------------------------------------------------------------
  ##
  ## get tree data
  ##
  ##----------------------------------------------------------------
  ## pull the arrays
  native.array <- object$forest$nativeArray
  native.f.array <- object$forest$nativeFactorArray[[1]]
  ## added processing needed for factors
  f.ctr <- 0
  factor.flag <- FALSE
  if (!is.null(native.f.array)) {
    pt.f <- which(native.array$mwcpSZ != 0)
    factPT <- lapply(pt.f, function(j) {
      f.ctr <<- f.ctr + 1
      step <- native.array$mwcpSZ[j] - 1
      mwcpPT <- native.f.array[f.ctr:(f.ctr+step)]
      mwcpPT <- paste0(sapply(mwcpPT, function(mwc) {
        format(as.hexmode(mwcpPT), 8)
      }))      
      mwcpSZ <- hex2bin(mwcpPT)
      paste(mwcpSZ, collapse = "")
    })
    native.array$contPT[pt.f] <- factPT
    factor.flag <- TRUE
  }
  ## define the display tree
  display.tree <- native.array[native.array$treeID == tree.id,, drop = FALSE]
  ## check to see if any factors are left
  ## store relevant informatio for later split-inequality encodings
  if (factor.flag) {
    pt.f <- display.tree$mwcpSZ !=0
    if (sum(pt.f) > 0) {
      f.names <- unique(xvar.names[display.tree$parmID[pt.f]])
    }
    else {
      factor.flag <- FALSE
    }
  }
  ##----------------------------------------------------------------
  ##
  ## prepare the tree to be converted into a network
  ##
  ##----------------------------------------------------------------
  ## conversion
  converted.tree <- display.tree
  vars.id <- data.frame(var = c("<leaf>", xvar.names), parmID = 0:length(xvar.names), stringsAsFactors = FALSE)
  converted.tree$var <- vars.id$var[match(display.tree$parmID, vars.id$parmID)]
  ## special symbol to be used for encoding the counter for variables (see note below)
  special <- "999_999"
  # note: we append a counter to the variables, because the data.tree package has trouble when
  # nodes are not unique.
  var.count <- 1:nrow(converted.tree)
  lapply(unique(converted.tree$var), function(vv) {
    pt <- converted.tree$var == vv
    var.count[which(pt)] <<- 1:sum(pt)
  })
  converted.tree$var_count <- var.count
  converted.tree$var_conc <- paste0(converted.tree$var, special, converted.tree$var_count)
  ##----------------------------------------------------------------
  ##
  ## convert the tree to a network data frame, using the fact that the
  ## nativeArray output is a pre-order traversal 
  ##
  ##----------------------------------------------------------------
  ## preliminary
  from_node <- ""
  network <- data.frame()
  num.children <- data.frame(converted.tree, children = 0)
  num.children <- num.children[num.children$var != "<leaf>",, drop = FALSE]
  num.children <- num.children[!duplicated(num.children$var_conc),, drop = FALSE]
  num_children <- as.list(rep(0, nrow(num.children)))
  names(num_children) <- num.children$var_conc
  ## loop (using lapply)
  lapply(1:nrow(converted.tree), function(i) {
    rowi <- converted.tree[i, ]
    xs <- converted.tree$contPT[converted.tree$var_conc == from_node]
    if(i == 1){
      from_node <<- rowi$var_conc
    }
    else{
      ## develop the split encoding
      if(num_children[[from_node]] == 0) {#left split
        side <- "<="
        contPT.pretty <- round(as.numeric(xs), 3)
        split_ineq_pretty <- paste0(side, contPT.pretty)
      }
      else {#right split
        side <- ">"
        split_ineq_pretty <- ""
      }
      ## both numeric and factors are encoded as <= > but factors are secretely in hex notation
      ## !!ADD MACHINE TOLERANCE WHEN SPLIT IS ON ACTUAL X-VALUE !!
      if (is.numeric(xs)) {
        xs <- xs + tolerance
      }
      split_ineq <- paste0(side, xs)
      ## update the network
      to_node <- rowi$var_conc
      new_node <- list(from = from_node, to = to_node, split = split_ineq, split.pretty = split_ineq_pretty)
      network <<- data.frame(rbind(network, new_node, stringsAsFactors = FALSE))
      num_children[[from_node]] <<- num_children[[from_node]] + 1
      if(rowi$var != "<leaf>")
        from_node <<- to_node
      else{
        if(i != nrow(converted.tree)){
          while(num_children[[from_node]] == 2){
            from_node <<- network$from[network$to == from_node]
          }
        }
      }
    }
  })
  ##----------------------------------------------------------------
  ##
  ## process network for factors - clean up the split encoding
  ## encode as complementary pair string
  ##
  ##----------------------------------------------------------------
  if (factor.flag) {
    ## identify which splits need to be cleaned up
    from.names <- network$from
    pt.f <- sort(unique(unlist(lapply(f.names, function(ptrn) {
      grep(ptrn, from.names)
    }))))
    ## identify levels of the factor
    ## otherwise we would have huge sets full of levels that aren't used
    fs <- gsub(paste0(special,".*"),"",from.names[pt.f])
    fs.levels <- sapply(fs, function(fsn) {
      #length(levels(x.org.data[, fsn]))
      xvar.factor$nlevels[which(fsn == xvar.names)]
    })
    ## clean the splits up and encode as complementary pair sets
    split.str <- lapply(1:length(pt.f), function(j) {
      str <- network$split[pt.f[j]]
      ## left split
      if (grepl("<=", str)) {
        str <- sub("<=", "", str)
        str <- strsplit(str, "")[[1]]
        cpr <- 1 + length(str) - which(str != "0")
        cpr <- cpr[cpr <= fs.levels[j]]
        paste0("{", paste(cpr, collapse = ","), "}")
      }
      ## right split
      else {
        str <- sub(">", "", str)
        str <- strsplit(str, "")[[1]]
        cpr <- 1 + length(str) - which(str == "0")
        cpr <- cpr[cpr <= fs.levels[j]]
        paste0("{", paste(cpr, collapse = ","), "}")
      }
    })
    ## pretty complementary pairs
    split.str.pretty <- lapply(1:length(pt.f), function(j) {
      str <- network$split[pt.f[j]]
      ## left split
      if (grepl("<=", str)) {
        str <- sub("<=", "", str)
        str <- strsplit(str, "")[[1]]
        cpr <- 1 + length(str) - which(str != "0")
        cpr <- cpr[cpr <= fs.levels[j]]
        paste0("{", paste(cpr, collapse = ","), "}")
      }
      else {
        ""
      }
    })
    ## replace the previous fake encoding with the now correct "set encoding"
    network$split[pt.f] <- split.str
    network$split.pretty[pt.f] <- split.str.pretty
  }
  ##----------------------------------------------------------------
  ##
  ## create a tree object, see the data.tree package
  ##
  ##----------------------------------------------------------------
  data.tree.network <- data.tree::FromDataFrameNetwork(network, "split")
  ## INTERNAL NODES
  ## label the edges with the splits, and relabel the nodes so that the appended counters are not visible
  data.tree.network$Get(function(node) {
    data.tree::SetEdgeStyle(node, color = "black", label = node$split.pretty, fontcolor = "black")
    data.tree::SetNodeStyle(node, color = "black", fontcolor = "black", penwidth = 3,
                            label = strsplit(node$name, special)[[1]][1])
  })
  ## TERMINAL NODES
  ## loop through the leaves of the tree, each time getting the path down to the leaf, and using this path
  ## to establish a filtering condition to obtain all cases that match the splitting on the way to this leaf.
  ## we filter out cases to correspond to the leaf, and then modify the leaf display to include the number
  ## of cases, followed by the user requested predicted value
  lapply(data.tree.network$leaves, function(node) {
    path_list <- node$path
    var_list <- sapply(path_list, function(x){strsplit(x, special)[[1]][1]})
    var_list[length(var_list)] <- ""
    node_iter <- data.tree.network
    ## make boolean string operatore
    call <- lapply(2:(length(path_list)), function(i) {
      node_iter <<- node_iter[[path_list[[i]]]]
      str <- node_iter$split
      ## numeric boolean operator
      if (!any(grepl("\\{", str))) {
        str <- paste0("x.data$", paste0(var_list[i-1], str))
      }
      ## complementary pair boolean operator
      else {
        str <- gsub("\\{", "", str)
        str <- gsub("\\}", "", str)
        str <- strsplit(str, ",")[[1]]
        str <- paste("==", str, sep = "")
        str <- paste0("(",paste(paste0("x.data$", var_list[i-1], str), collapse = "|"),")")
      }
      str
    })
    call <- paste(call, collapse = " & ")
    ## evaluate the boolean operator
    ## this yields the id's for the cases in the node
    if (!anonymous) {
      pt <- which(eval(parse(text=call)))
      n.cases <- length(pt)
    }
    ## set the edgelabel
    edge.label <- node$split.pretty
    ## set the node label
    ## extract the predicted value in the node
    if (predict.flag && !anonymous) {
      if (!is.factor(yhat)) {
        yhat <- round(mean(yhat[pt], na.rm = TRUE), 2)
        node.label <- paste0("n=", n.cases, "\n", yhat)
      }
      else {
        frqtable <- tapply(yhat[pt], yhat[pt], length)
        pred <- names(frqtable)[which.max(frqtable)]
        node.label <- paste0("n=", n.cases, "\n", pred)
      }
    }
    ## unsupervised family --> no predicted value
    else if (!predict.flag && !anonymous) {
      node.label <- paste0("n=", n.cases)
    }
    ## anonymous
    else {
      node.label <- NULL
    }
    ## set styles
    data.tree::SetGraphStyle(node, rankdir = "TB")
    data.tree::SetEdgeStyle(node, arrowhead = "vee", color = "grey35",
                            penwidth = 3, label = edge.label)
    data.tree::SetNodeStyle(node, style = "filled,rounded", shape = "box",
                 fillcolor = "GreenYellow", penwidth = 3, 
                 fontname = "helvetica", tooltip = data.tree::GetDefaultTooltip,
                 label = node.label)
  })
  invisible(data.tree.network)
}
get.tree <- get.tree.rfsrc

plot.subsample.rfsrc <- function(x, alpha = .01, xvar.names,
                          standardize = TRUE, normal = TRUE, jknife = FALSE,
                          target, m.target = NULL, pmax = 75, main = "",
                          sorted = TRUE, show.plots = TRUE,
                          ...)
{
  ##--------------------------------------------------------------
  ##
  ## was subsampling or double-bootstrap used?
  ##
  ##--------------------------------------------------------------
  if (sum(c(grepl("rfsrc", class(x))), grepl("subsample", class(x))) == 2) {
    subsample <- TRUE
  }
  else if (sum(c(grepl("rfsrc", class(x))), grepl("bootsample", class(x))) == 2) {
    subsample <- FALSE
  }
  else {
    stop("object must be obtained from call to 'subsample' function")
  }
  ##--------------------------------------------------------------
  ##
  ## coerce the (potentially) multivariate rf object
  ##
  ##--------------------------------------------------------------
  m.target <- get.univariate.target(x$rf, m.target)
  x$rf <- coerce.multivariate(x$rf, m.target)
  ##--------------------------------------------------------------
  ##
  ## family specific details
  ## - set the target if not supplied
  ## - assign pretty labels for the horizontal axis
  ##
  ##--------------------------------------------------------------
  fmly <- x$rf$family
  ## labels
  if (standardize) {
    if (fmly == "regr") {
      xlab <- "standardized vimp"
    }
    else {
      xlab <- "vimp"
    }
  }
  else {
    xlab <- "vimp"
  }
  ## over-ride x label when only generalization error is requested
  if (x$performance.only) {
    xlab <- "generalization error"
  }
  ## extract vimp and vimp column names - depends on family and univariate/multivariate
  ## special mod needed for generalization error
  if (is.null(m.target)) {  
    vmp <- x$vmp[[1]]
    if (!x$performance.only) {
      vmp.col.names <- colnames(x$vmp[[1]])
    }
    else {
      vmp.col.names <- names(x$vmp)
    }
  }
  else {
    vmp <- x$vmp[[m.target]]
    if (!x$performance.only) {
      vmp.col.names <- colnames(x$vmp[[m.target]])
    }
    else {
      vmp.col.names <- m.target
    }
  }
  if (fmly == "regr" || fmly == "surv") {
    target <- 0
    vmp <- c(vmp)
    xlab <- paste(xlab, " (", vmp.col.names, ")", sep = "")
  }
  else if (fmly == "class") {
    if (missing(target)) {
      target <- 0
    }
    else {
      yvar.levels <- levels(x$rf$yvar)
      if (is.character(target)) {
        target <- match(match.arg(target, yvar.levels), yvar.levels)
      }
      else {
        if (target < 0 || target > length(yvar.levels)) {
          stop("'target' is specified incorrectly")
        }
      }
    }
    vmp <- c(vmp[, 1 + target])
    xlab <- paste(xlab, " (", vmp.col.names[1 + target], ")", sep = "")
  }
  else if (fmly == "surv-CR") {    
    if (missing(target)) {
      target <- 0
    }
    else {
      n.event <- length(get.event.info(x$rf)$event.type)
      if (target < 1 || target > n.event) {
        stop("'target' is specified incorrectly")
      }
      target <- target - 1
    }
    vmp <- c(vmp[, 1 + target])
    xlab <- paste(xlab, " (", vmp.col.names[1 + target], ")", sep = "")
  }
  ##--------------------------------------------------------------
  ##
  ## over-ride x label if the user has supplied their own
  ##
  ##--------------------------------------------------------------
  if (!is.null(list(...)$xlab)) {
    xlab <- list(...)$xlab
  }
  ##--------------------------------------------------------------
  ##
  ## extract necessary objects
  ##
  ##--------------------------------------------------------------
  if (subsample) {
    oo <- extract.subsample(x, alpha = alpha, target = target, m.target = m.target,
                            standardize = standardize, raw = TRUE)
    boxplot.dta <- oo$boxplot.dta
  }
  else {
    oo <- extract.bootsample(x, alpha = alpha, target = target, m.target = m.target,
                             standardize = standardize, raw = TRUE)
    boxplot.dta <- oo[[1]]
  }
  if (normal) {
    if (subsample && jknife) {
      ci <- oo$ci.jk.Z
    }
    else {
      ci <- oo$ci.Z
    }
  }
  else {
    ci <- oo$ci
  }
  ##--------------------------------------------------------------
  ##
  ## final processing
  ##
  ##--------------------------------------------------------------
  ## (a) user has provided a list of target variables
  if (!missing(xvar.names)) {
    trim.pt <- colnames(boxplot.dta) %in% xvar.names
    if (sum(trim.pt) > 0) {
      boxplot.dta <- boxplot.dta[, trim.pt, drop = FALSE]
      ci <- ci[, trim.pt, drop = FALSE]
      vmp <- vmp[trim.pt]
    }
    else {
      stop("user supplied variable names do not match any of the variables from analysis")
    }
  }
  ## (b) sort by vimp?
  p <- ncol(boxplot.dta)
  if (sorted || p > pmax) {
    o.pt <- order(vmp, decreasing = TRUE)
  }
  else {
    o.pt <- 1:p
  }
  ## (c) trim if too many variables
  if (p > pmax) {
    o.pt <- o.pt[1:pmax]
  }
  ## redefine the data to match the processing
  boxplot.dta <- boxplot.dta[, o.pt, drop = FALSE]
  ci <- ci[, o.pt, drop = FALSE]
  ##--------------------------------------------------------------
  ##
  ## skeleton boxplot 
  ##
  ##--------------------------------------------------------------
  bp <- boxplot(boxplot.dta,
                yaxt="n",
                outline = FALSE,
                horizontal = TRUE,
                plot = FALSE)
  bp$stats <- ci
  col.pt <- bp$stats[1, ] > 0 
  col.pt[is.na(col.pt)] <- FALSE
  colr <- c("blue", "red")[1 + col.pt]
  ##--------------------------------------------------------------
  ##
  ## finesse ... unnamed options to be passed to bxp and axis
  ##
  ##--------------------------------------------------------------
  ## pull the unnamed options
  dots <- list(...)
  bxp.names <- c(names(formals(bxp)),
                 "xaxt", "yaxt", "las", "cex.axis", 
                 "col.axis", "cex.main",
                 "col.main", "sub", "cex.sub", "col.sub", 
                 "ylab", "cex.lab", "col.lab")
  axis.names <- formals(axis)
  axis.names$tick <- axis.names$las <- axis.names$labels <- NULL
  axis.names <- c(names(axis.names), "cex.axis") 
  ## override xlab
  if (!is.null(dots$xlab)) {
    xlab <- dots$xlab
  }
  ## overlay ylab when user mistakenly uses it  
  if (!is.null(dots$ylab) && length(dots$ylab) == p) {
    bp$names <- dots$ylab[o.pt]
    bxp.names <- bxp.names[bxp.names != "ylab"]
  }
  ## overlay names
  if (!is.null(dots$names)) {
    bp$names <- dots$names[o.pt]
  }
  ## special mod needed for generalization error
  if (x$performance.only) {
    bp$names <- ""
  }
  ##--------------------------------------------------------------
  ##
  ## draw the core bxp plot
  ##
  ##--------------------------------------------------------------
  if (show.plots) {
    do.call("bxp", c(list(z = bp, main = main, xlab = xlab, 
                          boxfill = colr, xaxt = "n", yaxt = "n",
                          outline = FALSE, horizontal = TRUE),
                     dots[names(dots) %in% bxp.names]))
    do.call("axis", c(list(side = 1, at = pretty(c(bp$stats)), tick = .02), dots[names(dots) %in% axis.names]))
    do.call("axis", c(list(side = 2, at = 1:length(bp$names), labels = bp$names,
                           las = 2, tick = FALSE), dots[names(dots) %in% axis.names]))
    abline(h = 1:length(bp$names), col = gray(.9), lty = 1)
    abline(v = 0, lty = 1, lwd = 1.5, col = gray(.8))
    bxp(bp, boxfill=colr,xaxt="n",yaxt="n",
        outline=FALSE,horizontal=TRUE,add=TRUE,
        whisklty=1,whisklwd=2)
  }
  ##--------------------------------------------------------------
  ##
  ## return the invisible boxplot data
  ##
  ##--------------------------------------------------------------
  invisible(bp)
}
plot.subsample <- plot.subsample.rfsrc

##  make SH data (modes 1 and 2)
make.sh <- function(dat, mode = 1) {
  ## extract sample size dimension
  nr <- dim(dat)[[1]]
  nc <- dim(dat)[[2]]
  if (nc == 0) {
    stop("can't make SH data ... not enough unique values\n")
  }
  ## coerce to data frame format
  if (!is.data.frame(dat)) {
    dat <- data.frame(dat)
  }
  ## mode 1
  if (mode == 1) {
    data.frame(classes = factor(c(rep(1, nr), rep(2, nr))),
      rbind(dat, data.frame(lapply(dat, sample, replace = TRUE))))
  }
  ## mode 2
  else {
    data.frame(classes = factor(c(rep(1, nr), rep(2, nr))),
      rbind(dat, data.frame(lapply(dat, function(x) {
        if (is.factor(x)) {
          resample(x, replace = TRUE)
        }
        else {
          runif(nr, min(x, na.rm = TRUE), max(x, na.rm = TRUE))
        }
      }))))
  }
}
##  make sid (staggered interaction data)
make.sid <- function(dat, order.by.range = TRUE, delta = NULL, papply = mclapply) {
  ## coerce to data frame format
  if (!is.data.frame(dat)) {
    dat <- data.frame(dat)
  }
  ## remove any column with less than two unique values
  void.var <- sapply(dat, function(x) {length(unique(x, na.rm = TRUE)) < 2})
  if (sum(void.var) > 0) {
    dat[, which(void.var)] <- NULL
  }
  ## there might be nothing left (small sample size issue)
  if (ncol(dat) == 0) {
    stop("can't make sid data ... not enough unique values\n")
  }
  ## order columns by range of values: noted improvement Alex 04/30/2018
  ## we do this before positivity and translation - but we could just as
  ## well have done this afterwards
  if (order.by.range) {
    order.range <- order(sapply(dat, function(x) {
    if (is.factor(x)) {
      1##changed from -Inf to 1 which is the correct value of a factor noted by Alex 06/05/2018
    }
    else {
      diff(range(x, na.rm = TRUE))
    }
  }), decreasing = TRUE)
  dat <- dat[, order.range, drop = FALSE]
  }
  ## make continuous variables positive
  posdat <- dat
  lapply(1:ncol(posdat), function(i) {
    if(!is.factor(posdat[, i]) && min(posdat[, i], na.rm = TRUE) < 0) {
      posdat[, i] <<- abs(min(posdat[, i], na.rm = TRUE)) + posdat[, i]
    }
    NULL
  })
  ## define the delta offset value (delta = 1 by default)
  ## if ordering is in effect, translate features to have the same maximum value
  ## this was observed in counter-example of theorem of revised paper 05/04/2018 
  if (is.null(delta)) {
    delta <- 1
  }
  if (order.by.range) {
    cont.feature <- sapply(posdat, function(x) {!is.factor(x)})
    ## acquire the maximum value
    if (sum(cont.feature) > 1) {
      maxV <- max(c(0, sapply(which(cont.feature), function(i) {
        max(posdat[, i], na.rm = TRUE)
      })), na.rm = TRUE)
      ## translate the features to have the same max
      lapply(which(cont.feature), function(i) {
        posdat[, i] <<- (maxV - max(posdat[, i], na.rm = TRUE)) + posdat[, i]
        NULL
      })	
    }
  }		
  ## two level factors are converted to numeric
  binary.fac <- sapply(posdat, function(x) {is.factor(x) & length(levels(x)) == 2})
  if (sum(binary.fac) > 0) {
    lapply(which(binary.fac), function(i) {
      names(posdat)[i] <<- names(posdat)[i]
      posdat[, i] <<- as.numeric(posdat[, i])
      NULL 
    })
  }
  ## the following list makes it possible to map SID back to original variable names 
  org.names <- list()
  ## make anova data for remaining factors - i.e. make everything numeric
  ## updated to make column names more appealing for binary factors (03/14/2018)
  counter <- 0
  numdat <- data.frame(lapply(1:ncol(posdat), function(j) {
    m.j <- model.matrix(~.-1, posdat[,j, drop = FALSE])
    ##SID main effect names mapped back to original names
    org.names[(counter + 1):(counter + ncol(m.j))] <<- colnames(posdat)[j]
    counter <<- counter + ncol(m.j)
    m.j
  }))
  ## stagger the positive data using delta
  ## use integer values when staggering to minimize creating double precision values
  ## the latter is accomplished using ceiling
  numvar <- dim(numdat)[2]
  staggerdat <- numdat
  staggerdat[, 1] <- staggerdat[, 1] + delta
  if (numvar > 1) {##handles pathological case of one column design matrix - maybe add this as a failure check?
    lapply(2:numvar, function(i) {
      staggerdat[, i] <<- staggerdat[, i] + ceiling(max(staggerdat[, (i-1)], na.rm = TRUE)) + delta
    })
  }
  ## make interactions
  ## -- do not create interactions WITHIN levels of the same factor -- noted by Alex M. 04/25/17
  ##    such instances are caught using unstaggered numerical data and checking x[,i]*x[,j]=constant
  ## -- interactions BETWEEN distinct factors (two or more levels) must always produce a factor
  ##    we catch this by checking the number of distinct values of each variable
  ##    all factors at this point have <= 2 unique values (whether they are dummy anova or numerical)
  intdat <- staggerdat
  if (numvar > 1) {##interactions do not apply if only one variable is supplied - maybe add this as a failure check?
    ## resolve papply
    papply <- .resolve_papply(
      papply,
      cores = getOption("mc.cores", 1L),
      envir = environment()
    )
    ##SID interaction names mapped back to original names
    ##we do this first inside of an lapply and not in next loop
    counter <- numvar
    lapply(1:(numvar-1), function(i) {
      for (j in (i + 1):numvar) {
        if (length(unique(numdat[,i] * numdat[,j])) > 1) {
          counter <<- counter + 1
          org.names[[counter]] <<- c(org.names[[i]], org.names[[j]])
        }
      }
      NULL
    })
    ## suggested by Alex M. 02/19/2019
    ## make interaction matrix after creating interactions - faster on big p
    ints <- papply(1:(numvar - 1), function(i) {
      d <- data.frame(rep(NA, dim(dat)[1]))
      d <- d[, -1]
      counter.ints <- 1
      for (j in (i + 1):numvar) {
        if (length(unique(numdat[,i] * numdat[,j])) > 1) {
          if (length(unique(numdat[, i])) <= 2 & length(unique(numdat[, j])) <= 2) {
            holder <- factor(staggerdat[, i] * staggerdat[, j])
            ## suggested by Alex M. 11/12/17
            levels(holder) <- c('FF','FT','TF','TT') #new factor levels, should always come out in same order
            d <- cbind(d, holder)
          }
          else {
            d <- cbind(d, staggerdat[, i] * staggerdat[, j])
          }
          names(d)[counter.ints] <- paste(names(intdat)[i], "_", names(intdat)[j], sep = "")
          counter.ints <- counter.ints + 1
        }
      }
      d
    })
    intdat <- data.frame(intdat, ints)
  }
  ## suggested by Alex M. 11/12/17
  ## final processing of numeric 0/1 variables to convert them to a binary factor with new levels
  binary.fac2 <- sapply(intdat, function(x){is.numeric(x) & length(unique(x)) == 2})
  if (sum(binary.fac2) > 0) {
    lapply(which(binary.fac2), function(i) {
      intdat[, i] <<- as.factor(intdat[, i])
      levels(intdat[, i]) <<- c('F','T') #new levels, similarly tested
      NULL
    })
  }
  ## pull the "x" and "y" features
  ## y=staggered main effects
  ## x=staggered interactions
  y <- intdat[, 1:numvar, drop = FALSE]
  names(org.names) <- colnames(intdat)
  y.names <- org.names[1:numvar]
  if (numvar > 1) {
    x <- intdat[, -(1:numvar), drop = FALSE]
    x.names <- org.names[-(1:numvar)]
  }
  else {
    x <- x.names <- NULL
  }
  ## return the goodies
  list(y = y, x = x, y.names = y.names, x.names = x.names, delta = delta)
}
#weighted gini/entropy performance metric
#mode is equal to either 'gini' or 'entropy'
sid.perf.metric <- function(truth,cluster,mode=c("entropy", "gini")){
  ## verify mode option
  mode <- match.arg(mode, c("entropy", "gini"))
  ## confusion matrix
  k=length(unique(truth))
  tab=table(truth,cluster)
  clustersizes=colSums(tab)
  clustersizesnorm=clustersizes/sum(clustersizes)
  tabprop=tab
  lapply(1:(dim(tab)[2]),function(i){
    tabprop[,i]<<-tabprop[,i]/clustersizes[i]
    NULL
  })
  if(mode=="entropy"){
    measure=0
    maxmeasure=0
    lapply(1:(dim(tabprop)[2]),function(i){
      clustermeasure=0
      maxclustermeasure=0
      lapply(1:(dim(tabprop)[1]),function(j){
        if(tabprop[j,i]!=0){
          clustermeasure<<-clustermeasure+-tabprop[j,i]*log2(tabprop[j,i])
        }
        maxclustermeasure<<-maxclustermeasure+-1/k*log2(1/k)
        NULL
      })
      measure<<-measure+clustersizesnorm[i]*clustermeasure
      maxmeasure<<-maxmeasure+clustersizesnorm[i]*maxclustermeasure
    })
  }
  if(mode=="gini"){
    measure=0
    maxmeasure=0
    lapply(1:(dim(tabprop)[2]),function(i){
      clustermeasure=1
      maxclustermeasure=1
      lapply(1:(dim(tabprop)[1]),function(j){
        if(tabprop[j,i]!=0){
          clustermeasure<<-clustermeasure+(-tabprop[j,i]^2)
        }
        maxclustermeasure<<-maxclustermeasure+(-1/k^2)
        NULL
      })
      measure<<-measure+clustersizesnorm[i]*clustermeasure
      maxmeasure<<-maxmeasure+clustersizesnorm[i]*maxclustermeasure
      NULL
    })
  }
  list(result=measure,measure=mode,normalized_measure=measure/maxmeasure)
}

rfsrc <- function(formula, data, ntree = 500,
                  mtry = NULL, ytry = NULL,
                  nodesize = NULL, nodedepth = NULL,
                  splitrule = NULL, nsplit = NULL,
                  importance = c(FALSE, TRUE, "none", "anti", "permute", "random"),
                  block.size = if (any(is.element(as.character(importance), c("none", "FALSE")))) NULL else 10,
                  bootstrap = c("by.root", "none", "by.user"),
                  samptype = c("swor", "swr"),  samp = NULL, membership = FALSE,
                  sampsize = if (samptype == "swor") function(x){x * .632} else function(x){x},
                  na.action = c("na.omit", "na.impute"), nimpute = 1,
                  ntime = 150, cause,
                  perf.type = NULL,
                  proximity = FALSE, distance = FALSE, forest.wt = FALSE,
                  xvar.wt = NULL, yvar.wt = NULL, split.wt = NULL, case.wt = NULL, case.depth = FALSE, 
                  forest = TRUE,
                  use.uno = TRUE, save.memory = FALSE,
                  var.used = c(FALSE, "all.trees", "by.tree"),
                  split.depth = c(FALSE, "all.trees", "by.tree"),
                  seed = NULL,
                  do.trace = FALSE,
                   ...)
{
  univariate.nomenclature = TRUE
  ## get any hidden options
  user.option <- list(...)
  impute.only <- is.hidden.impute.only(user.option)
  terminal.qualts <- is.hidden.terminal.qualts(user.option)
  terminal.quants <- is.hidden.terminal.quants(user.option, save.memory)
  ## set the big data flag
  big.data <- FALSE
  ## manually over-ride parameters if forest is not requested
  if (!forest) {
    terminal.qualts <- FALSE
    terminal.quants <- FALSE
    big.data <- TRUE
  }
  cse  <- is.hidden.cse(user.option)
  csv  <- is.hidden.csv(user.option)
  ## experimental option(s)
  experimental <- is.hidden.experimental(user.option)
  ## this is always false by default
  data.pass <- is.hidden.data.pass(user.option)
  ## rfq
  rfq <- is.hidden.rfq(user.option)
  ## quantile regression
  gk.quantile <- is.hidden.gk.quantile(user.option)
  quantile.regr <- is.hidden.quantile.regr(user.option)
  prob <- is.hidden.prob(user.option)
  prob.epsilon <- is.hidden.prob.epsilon(user.option)
  ## mahalanobis sigma matrix
  mahalanobis.sigma <- is.hidden.mahalanobis.sigma(user.option)
  ## misc.
  vtry <- is.hidden.vtry(user.option)
  holdout.array <- is.hidden.holdout.array(user.option)
  holdout.specs <- is.hidden.holdout.specs(user.option)
  tdc.rule <- is.hidden.tdc.rule(user.option)
  vimp.threshold  <- is.hidden.vimp.threshold(user.option)
  ## verify key options
  ensemble <- "all"##remove option to select between "all", "oob", "inbag"
  bootstrap <- match.arg(bootstrap, c("by.root", "none", "by.user"))
  if (bootstrap == "none") {
    ensemble <- "inbag"
  }
  importance <- match.arg(as.character(importance), c(FALSE, TRUE, "none", "anti", "permute", "random"))
  na.action <- match.arg(na.action, c("na.omit", "na.impute"))
  proximity <- match.arg(as.character(proximity), c(FALSE, TRUE, "inbag", "oob", "all"))
  distance <- match.arg(as.character(distance), c(FALSE, TRUE, "inbag", "oob", "all"))
  var.used <- match.arg(as.character(var.used), c("FALSE", "all.trees", "by.tree"))
  split.depth <- match.arg(as.character(split.depth),  c("FALSE", "all.trees", "by.tree"))
  if (var.used == "FALSE") var.used <- FALSE
  if (split.depth == "FALSE") split.depth <- FALSE
  ## data cannot be missing
  if (missing(data)) stop("data is missing")
  ## conduct preliminary formula validation
  if (missing(formula) | (!missing(formula) && is.null(formula))) {
    if (is.null(ytry)) {
      formula <- as.formula("Unsupervised() ~ .")
    }
    else {
      formula <- as.formula(paste("Unsupervised(", ytry, ")~."))
    }
  } 
  formulaPrelim <- parseFormula(formula, data, ytry)
  ## save the call/formula for the return object
  my.call <- match.call()
  my.call$formula <- eval(formula)
  ## conduct preliminary processing of missing data
  ## record whether there's no missing data: efficiency step
  if (data.pass == TRUE) {
    ## Give the data a complete pass and assume no missing data.
    ## This is a user directive and should be used with extreme caution.
    ## In general, control of this parameter should be left to the algorithm.
    miss.flag = FALSE
  }
  else {
    if (any(is.na(data))) {
      data <- parseMissingData(formulaPrelim, data)
      ## We detect missingness.
      miss.flag <- TRUE
    }
    else {
      ## We detect no missingness.
      miss.flag <- FALSE
    }
  }
  ## Set the C-side protocol for the data pass option. If there is
  ## no missingness, allow the C-side to skip a lot of missingness
  ## checks, and hand over control of many split routines to their
  ## non-missing counterparts.
  if (miss.flag == FALSE) {
    data.pass = TRUE
  }
  ## finalize the formula based on the pre-processed data
  formulaDetail <- finalizeFormula(formulaPrelim, data)
  ## coherence checks on option parameters
  ntree <- round(ntree)
  if (ntree < 1) stop("Invalid choice of 'ntree'.  Cannot be less than 1.")
  if (!is.null(nodesize) && nodesize < 1) stop("Invalid choice of 'nodesize'. Cannot be less than 1.")
  if (!is.null(nodedepth)) nodedepth = round(nodedepth) else nodedepth = -1
  nimpute <- round(nimpute)
  if (nimpute < 1) stop("Invalid choice of 'nimpute'.  Cannot be less than 1.")
  ## initialize the seed
  seed <- get.seed(seed)
  ## save the family for convenient access
  family <- formulaDetail$family
  ## save the names for convenient access
  xvar.names <- formulaDetail$xvar.names
  yvar.names <- formulaDetail$yvar.names
  subj.names <- formulaDetail$subj.names
  ## reality check on x and y
  ## are there any x-variables?  (can happen when for multivariate formula)
  if (length(xvar.names) == 0) {
    stop("something seems wrong: your formula did not define any x-variables")
  }
  ## .. are there any y-variables?  (do not test for the unsupervised case)
  if (family != "unsupv" && length(yvar.names) == 0) {
    stop("something seems wrong: your formula did not define any y-variables")
  }
  ## missing levels are allowed.
  if (family == "class") {
    if (length(setdiff(levels(data[, yvar.names]), unique(data[, yvar.names]))) > 0) {
      warning("empty classes found when implementing classification\n")
    }
  }
  ## mark missing factor levels as NA.
  data <- rm.na.levels(data, xvar.names)
  data <- rm.na.levels(data, yvar.names)
  ## Determine the immutable yvar factor map.
  ## get y-outcome type and nlevels and append to y-immutable map (saved in forest later)
  yfactor <- extract.factor(data, yvar.names)
  yfactor$types <- yvar.types <- get.yvar.type(family, yfactor$generic.types, yvar.names)
  yfactor$nlevels <- yvar.nlevels <- get.yvar.nlevels(family, yfactor$nlevels, yvar.names, data)
  ## Determine the immutable xvar factor map.
  xfactor <- extract.factor(data, xvar.names)
  xfactor$types <- xvar.types <- xfactor$generic.types
  xvar.nlevels <- xfactor$nlevels
  ## Convert the data to numeric mode, apply the na.action protocol.
  data <- finalizeData(c(subj.names, yvar.names, xvar.names), data, na.action, miss.flag)
  ## Save the row and column names for later overlay
  ## data.col.names <- colnames(data)
  data.row.names <- rownames(data)
  ## Finalize the xvar matrix.
  xvar <- as.matrix(data[, xvar.names, drop = FALSE])
  rownames(xvar) <- colnames(xvar) <- NULL
  ## get numeric levels for x-variables and append to x-immutable map (saved in forest)
  xfactor$numeric.levels <- xvar.numeric.levels <- get.numeric.levels(family, xfactor$nlevels, xvar)
  ## initialize sample size, set mtry, samptype
  n <- nrow(xvar)
  n.xvar <- length(xvar.names)
  mtry <- get.grow.mtry(mtry, n.xvar, family, splitrule)
  samptype <- match.arg(samptype, c("swor", "swr"))
  ## Get the indvidual subject identifiers if they exist - do this *after* na.protocol
  ## The default is that we do not support time dependent x-variables.
  subj.unique.count <- n
  subj <- NULL
  xvar.time <- NULL
  subj.time <- NULL
  ## bootstrap case
  if (bootstrap == "by.root") {
    ## now applies to Time Dependent Covariates (TDC)
    ## Nominal bootstrap options.
    if (!is.function(sampsize) && !is.numeric(sampsize)) {
      stop("sampsize must be a function or number specifying size of subsampled data")
    }
    if (is.function(sampsize)) {
      sampsize.function <- sampsize
    }
    else {
      ## convert user passed sample size to a function
      sampsize.function <- make.samplesize.function(sampsize / subj.unique.count)
    }
    sampsize <- round(sampsize.function(subj.unique.count))
    if (sampsize < 1) {
      stop("sampsize must be greater than zero")
    }
    ## for swor size is limited by the number of cases
    if (samptype == "swor" && (sampsize > subj.unique.count)) {
      sampsize.function <- function(x){x}
      sampsize <- subj.unique.count
    }
    samp <- NULL
    case.wt  <- get.weight(case.wt, subj.unique.count)
  }
  ## user specified bootstrap
  else if (bootstrap == "by.user") {    
    ## check for coherent sample: it will of be dim [.] x [ntree]
    ## where [.] is the number of rows in the data set or unique subjects in
    ## the case of time dependent covariates.
    if (is.null(samp)) {
      stop("samp must not be NULL when bootstrapping by user")
    }
    ## ntree should be coherent with the sample provided
    ntree <- ncol(samp)
    sampsize <- colSums(samp)
    if (sum(sampsize == sampsize[1]) != ntree) {
      stop("sampsize must be identical for each tree")
    }
    ## set the sample size and function
    sampsize <- sampsize[1]
    sampsize.function <- make.samplesize.function(sampsize[1] / subj.unique.count)
    ## set the case weights
    case.wt  <- get.weight(NULL, subj.unique.count)
  }
  ## This is "none".
  else {
    sampsize <- subj.unique.count
    sampsize.function <- function(x){x}
    case.wt  <- get.weight(case.wt, sampsize)
  }
  ## Set the weight matrix for xvar, split.
  ## Set the flag for forest weight.
  split.wt <- get.weight(split.wt, n.xvar)
  forest.wt <- match.arg(as.character(forest.wt), c(FALSE, TRUE, "inbag", "oob", "all"))
  ## Note that yvar.types is based on the formula:
  ## If the family is unsupervised, yvar.types is NULL
  ## If the family is regression, classification, or
  ## multivariate/mixed, it is of length greater than zero (0).  In the case
  ## of surival and competing risk, it is specifically of length two (2) for
  ## coherency, though it is ignored in the native-code.
  ## Note that ytry is set by the formula:
  ## If the family is unsupervised, ytry assumes the value specified
  ## via the formula.  If the family is regression, classification,
  ## or multivariate/mixed, it defaults to length(yvar.types).  In
  ## the case of survival, competing risk, and time dependent
  ## covariates, it is zero (0) and ignored in the native-code.
  ## With time dependent covariates, the weight y-vector is
  ## implicitly of length 3, one for "start", "stop", and "event".
  ## But "start" and "event" are ignored.  Send in a weight in the
  ## correct slot for "stop" (slot 2 based on the formula), and it
  ## will be appended, in the native code, to xvar.wt.  For example,
  ## sending in something like ywar.wt=c(0,2,0), and xvar.wt =
  ## rep(1, 17) will favour time over indigenous x-vars 2 to 1.
  if (family == "unspv") {
    ## Override any incoming weights.
    yvar.wt <- NULL
  }
  else {
    yvar.wt <- get.weight(yvar.wt, length(yvar.types))
  }
  xvar.wt  <- get.weight(xvar.wt, n.xvar)
  ## Get the y-outcomes
  yvar <- as.matrix(data[, yvar.names, drop = FALSE])
  if(dim(yvar)[2] == 0) {
    ## Override yvar if it is not present.
    yvar <- 
      yvar.nlevels  <-  yvar.numeric.levels <- yfactor <- NULL
  }
  else {
    yfactor$numeric.levels <- yvar.numeric.levels <- get.numeric.levels(family, yfactor$nlevels, yvar)
  }
  ## Determine the number of missing values
  if (miss.flag) {
    n.miss <- get.nmiss(xvar, yvar)
  }
  else {
    n.miss <- 0
  }
  ## In impute.only, if there is no missing data,
  ## return the data and do nothing else.
  if (impute.only && n.miss == 0) {
    return(data)
  }
  ## Don't need the data anymore.
  remove(data)
  ## TBD I think we need to separate event related data and dimensioning
  ## of the response which this function combines.
  ## Get event information and dimensioning for families
  event.info <- get.grow.event.info(yvar, family, ntime = ntime)
  ## Initialize nsplit, noting that it may be been overridden.
  splitinfo <- get.grow.splitinfo(formulaDetail, splitrule, nsplit, event.info)
  ## Set the cause weights for the split statistic calculation.
  if (family == "surv" || family == "surv-CR") {
    if (length(event.info$event.type) > 1) {
      ## This may or may not be CR, but we do have multiple events.
      ## This requires associated weights.
      if (missing(cause) || is.null(cause)) {
        cause <- NULL
        cause.wt <- rep(1, length(event.info$event.type))
      }
      else {
        if (length(cause) == 1) {
          ## Event specific selection
          if (cause >= 1 && cause <= length(event.info$event.type)) {
            cause.wt <- rep(0, length(event.info$event.type))
            cause.wt[cause] <- 1
          }
          else {
            cause.wt <- rep(1, length(event.info$event.type))
          }
        }
        else {
          ## The user has specified event specific weights
          if (length(cause) == length(event.info$event.type) && all(cause >= 0) && !all(cause == 0)) {
            cause.wt <- cause / sum(cause)
          }
          else {
            cause.wt <- rep(1, length(event.info$event.type))
          }
        }
      }
    }
    else {
      ## We are conducting non-CR, without multiple events.  Send in a dummy value.
      cause <- NULL
      cause.wt = 1
    }
    ## Set the coerced family.
    family <- get.coerced.survival.fmly(family, subj, event.info$event.type, splitinfo$name)
  }
  else {
    ## We pass in NULL for the cause weight for all other families
    cause <- cause.wt <- NULL
  }
  ## set the Uno-weights for survival
  uno.weights <- NULL
  if (use.uno && (family == "surv" || family == "surv-CR")) {
    uno.weights <- get.uno.weights.train(event.info$time, event.info$cens)
  }
  ## Nodesize determination
  nodesize <- get.grow.nodesize(family, nodesize)
  ## Turn ensemble outputs off unless bootstrapping by root or user.
  if ((bootstrap != "by.root") && (bootstrap != "by.user")) {
    importance <- "none"
    perf.type <- "none"
  }
  ## Turn ensemble outputs off for unsupervised mode only
  if (family == "unsupv") {
    importance <- "none"
    perf.type <- "none"
  }
  ## Impute only mode
  ## Some of this may be redundant
  if (impute.only) {
    forest       <- FALSE
    big.data     <- FALSE
    proximity    <- FALSE
    distance     <- FALSE
    var.used     <- FALSE
    split.depth  <- FALSE
    membership   <- FALSE
    perf.type    <- "none"
    importance   <- "none"
    case.depth   <- FALSE
    terminal.qualts <- FALSE
    terminal.quants <- FALSE
    cse  <- FALSE
    csv  <- FALSE
    ## We have deleted the following line because it was incorrectly
    ## overriding the user specified option na.action in 
    ## impute.rfsrc( generic.impute.rfsrc( rfsrc() ) ).  Now, the user specified option
    ## is respected.  Note that calls from impute.rfsrc()
    ## set the impute.only bit which results in turning of all outputs
    ## other than imputed data.
    ## na.action    <- "na.impute"
  }
  ## deal with user specified holdout array
  ## confirm dimension is right
  ## set vtry to 1 to trigger holdout algorithm
  if (!is.null(holdout.array)) {
    if (nrow(holdout.array) != n.xvar | ncol(holdout.array) != ntree) {
      stop("dimension of holdout.array does not conform to p x ntree")
    }
    vtry <- 1##this only needs to be non-zero to trigger holdout
  }
  ## !!! here's where prob and prob.epsilon are set globally !!!
  ## for the user convenience if not set make coherent
  ## global assingments for  prob and prob.epsilon values
  ## takes into account splitrule and whether gk.quantile is requested
  gk.quantile <- get.gk.quantile(gk.quantile)
  prob.assign <- global.prob.assign(prob, prob.epsilon, gk.quantile, quantile.regr, splitinfo$name, n)
  prob <- prob.assign$prob
  prob.epsilon <- prob.assign$prob.epsilon
  ## mahalanobis user specified sigma matrix
  if (splitinfo$name ==  "mahalanobis" && !is.null(mahalanobis.sigma)) {
    ## check that sigma is the correct dimension
    if (length(yvar.names) != nrow(mahalanobis.sigma)
        || length(yvar.names) != ncol(mahalanobis.sigma)) {
          stop("dimension of mahalanobis sigma matrix is incorrect: must match dimension of y response")
    }
  }
  ## Bit dependencies:
  if (terminal.qualts | terminal.quants) {
    forest <- TRUE
  }
  ## get performance and rfq, gk bits
  perf.type <- get.perf(perf.type, impute.only, family)
  perf.bits <-  get.perf.bits(perf.type)
  rfq <- get.rfq(rfq)
  rfq.bits  <- get.rfq.bits(rfq, family)
  gk.quantile.bits <- get.gk.quantile.bits(gk.quantile)
  tdc.rule.bits <- get.tdc.rule.bits(tdc.rule)
  ## Assign low bits for the native code
  ensemble.bits <- get.ensemble.bits(ensemble)
  impute.only.bits <- get.impute.only.bits(impute.only, n.miss)
  var.used.bits <- get.var.used.bits(var.used)
  split.depth.bits <- get.split.depth.bits(split.depth)
  importance.value <-  get.importance(importance, perf.type)
  importance.bits <- get.importance.bits(importance, perf.type)
  bootstrap.bits <- get.bootstrap.bits(bootstrap)
  forest.bits <- get.forest.bits(forest)
  proximity.bits <- get.proximity.bits(TRUE, proximity)
  distance.bits <- get.distance.bits(TRUE, distance)
  membership.bits <-  get.membership.bits(membership)
  split.cust.bits <- get.split.cust.bits(splitinfo$cust)
  case.depth.bits  <- get.case.depth.bits(case.depth)
  ## Assign high bits for the native code
  samptype.bits <- get.samptype.bits(samptype)
  forest.wt.bits <- get.forest.wt.bits(TRUE, bootstrap, forest.wt)
  na.action.bits <- get.na.action.bits(na.action)
  block.size <- get.block.size.bits(block.size, ntree)
  terminal.qualts.bits <- get.terminal.qualts.bits(terminal.qualts)
  terminal.quants.bits <- get.terminal.quants.bits(terminal.quants)
  cse.bits = get.cse.bits(cse)
  csv.bits = get.csv.bits(csv)
  data.pass.bits <- get.data.pass.bits(data.pass)
  experimental.bits <- get.experimental.bits(experimental)
  ## set the maximum class levels
  max.class.levels <- 0
  ## Set the trace
  do.trace <- get.trace.bits(do.trace)
  ## Start the C external timer.
  ctime.external.start  <- proc.time()
  nativeOutput <- tryCatch({.Call("rfsrcGrow",
                                  as.integer(do.trace),
                                  as.integer(seed),
                                  as.integer(ensemble.bits +
                                             impute.only.bits +
                                             var.used.bits +
                                             split.depth.bits +
                                             importance.bits +
                                             bootstrap.bits +
                                             forest.bits +
                                             proximity.bits +
                                             perf.bits +
                                             rfq.bits +
                                             gk.quantile.bits +
                                             case.depth.bits),
                                  as.integer(samptype.bits +
                                             forest.wt.bits +
                                             distance.bits + 
                                             na.action.bits +
                                             split.cust.bits +
                                             membership.bits +
                                             terminal.qualts.bits +
                                             terminal.quants.bits +
                                             tdc.rule.bits +
                                             cse.bits +
                                             csv.bits +
                                             data.pass.bits),
                                  as.integer(experimental.bits),
                                  as.integer(splitinfo$index),
                                  as.integer(splitinfo$nsplit),
                                  as.integer(mtry),
                                  NULL, ## object containing lot settings
                                  NULL, ## Object containing base learner settings.
                                  as.integer(vtry),
                                  as.integer(holdout.array),
                                  holdout.specs, ## object containing speculative holdout settings
                                  as.integer(formulaDetail$ytry),
                                  as.integer(nodesize),
                                  as.integer(nodedepth),
                                  as.integer(length(cause.wt)),
                                  as.double(cause.wt),
                                  as.double(vimp.threshold),
                                  as.integer(ntree),
                                  as.integer(n),
                                  list(as.integer(length(yvar.types)),
                                       if (is.null(yvar.types)) NULL else as.character(yvar.types),
                                       if (is.null(yvar.types)) NULL else as.integer (sapply(1:length(yvar.nlevels), function(nn) {
                                         if(yvar.nlevels[nn] > 0) max(max.class.levels, yvar.nlevels[nn]) else 0})),
                                       if (is.null(yvar.numeric.levels)) NULL else sapply(1:length(yvar.numeric.levels), function(nn) {
                                         as.integer(length(yvar.numeric.levels[[nn]]))}),
                                       if (is.null(subj)) NULL else as.integer(subj),
                                       if (is.null(event.info)) NULL else as.integer(length(event.info$event.type)),
                                       if (is.null(event.info)) NULL else as.integer(event.info$event.type)),
                                  if (is.null(yvar.numeric.levels)) {
                                    NULL
                                  }
                                  else {
                                    lapply(1:length(yvar.numeric.levels),
                                           function(nn) {as.integer(yvar.numeric.levels[[nn]])})
                                  },
                                  if (is.null(yvar)) NULL else as.double(as.vector(yvar)),
                                  list(as.integer(n.xvar),
                                       as.character(xvar.types),
                                       if (is.null(xvar.types)) NULL else as.integer(xvar.nlevels),
                                       if (is.null(xvar.numeric.levels)) NULL else sapply(1:length(xvar.numeric.levels), function(nn) {as.integer(length(xvar.numeric.levels[[nn]]))}),
                                       if (is.null(xvar.time)) NULL else as.integer(xvar.time),
                                       if (is.null(subj.time)) NULL else as.integer(subj.time)),
                                  if (is.null(xvar.numeric.levels)) {
                                    NULL
                                  }
                                  else {
                                    lapply(1:length(xvar.numeric.levels),
                                           function(nn) {as.integer(xvar.numeric.levels[[nn]])})
                                  },
                                  as.double(as.vector(xvar)),
                                  list(as.integer(length(case.wt)),
                                       if (is.null(case.wt)) NULL else as.double(case.wt),
                                       as.integer(sampsize),
                                       if (is.null(samp)) NULL else as.integer(samp)),
                                  as.double(split.wt),
                                  if (is.null(yvar.wt)) NULL else as.double(yvar.wt),
                                  as.double(xvar.wt),
                                  list(if(is.null(event.info$time.interest)) as.integer(0) else as.integer(length(event.info$time.interest)),
                                       if(is.null(event.info$time.interest)) NULL else as.double(event.info$time.interest)),
                                  as.integer(nimpute),
                                  as.integer(block.size),
                                  list(if (is.null(prob)) as.integer(0) else as.integer(length(prob)),
                                       if (is.null(prob)) NULL else as.double(prob),
                                       if (is.null(prob.epsilon)) as.double(0) else as.double(prob.epsilon)),
                                  if (is.null(mahalanobis.sigma)) NULL else as.double(as.vector(mahalanobis.sigma)),
                                  if (is.null(uno.weights)) NULL else as.double(uno.weights$weight),
                                  as.integer(get.rf.cores()))}, error = function(e) {
                                    print(e)
                                    NULL})
  ## Stop the C external timer.
  ctime.external.stop <- proc.time()
  ## check for error return condition in the native code
  if (is.null(nativeOutput)) {
    if (impute.only) {
      ## in impute mode we proceed and return NULL
      return(NULL)
    }
    else {
      ## currently all other modes will error out with the RFSRC error message
      stop("An error has occurred in the grow algorithm.  Please turn trace on for further analysis.")
    }
  }
  ## check if there was missing data, and assign imputed data if so.
  if (n.miss > 0) {
    imputed.data <- matrix(nativeOutput$imputation, nrow = n.miss, byrow = FALSE)
    imputed.indv <- imputed.data[, 1]
    imputed.data <- as.matrix(imputed.data[, -1, drop = FALSE])
    ##if (n.miss == 1) {
    ##  imputed.data <- t(imputed.data)
    ##}
    nativeOutput$imputation <- NULL
    ## fill NA's in original GROW data with multiply imputed values.
    ## This will now serve as the forest data set and will enable
    ## recovery of terminal node membership with the head of the
    ## seed chain
    if (nimpute > 1) {
      ## the forest was grown using overlaid full (all) summary values
      if (grepl("surv", family)) {
        yvar[imputed.indv, 1] <- imputed.data[, 1]
        yvar[imputed.indv, 2] <- imputed.data[, 2]
        xvar[imputed.indv, ] <- imputed.data[, -c(1:2), drop = FALSE]
      }
      else {
        if (!is.null(yvar.types)) {
          yvar[imputed.indv, ] <- imputed.data[, 1:length(yvar.types), drop = FALSE]
          xvar[imputed.indv, ] <- imputed.data[, -c(1:length(yvar.types)), drop = FALSE]
        }
        else {
          xvar[imputed.indv, ] <- imputed.data
        }
      }
      ## remove the imputed data outputs
      imputed.indv <- NULL
      imputed.data <- NULL
      ## the missingness action for the training data is formally safed.  This is saved as part of the
      ## forest object, but the setting is now irrelevant as the training data has no missingness.
      na.action = "na.omit"
    }
    else {
      ## add column names to the imputed data outputs in the absence
      ## of multiple imputation.
      colnames(imputed.data) <- c(yvar.names, xvar.names)
      imputed.data <- as.data.frame(imputed.data)
    }
    ## now map the imputed.data columns back to their original order
    ## commented out because of difficulty in maintaining order across different modes
    ## imputed.data <- imputed.data[, data.col.names]
  }
  ## add row and column names to xvar matrix
  xvar <- as.data.frame(xvar)
  rownames(xvar) <- data.row.names
  colnames(xvar) <- xvar.names
  ## map xvar factors back to original values
  xvar <- map.factor(xvar, xfactor)
  ## add column names to response matrix
  ## does not apply for unsupervised mode
  if (family != "unsupv") {
    yvar <- as.data.frame(yvar)
    colnames(yvar) <- yvar.names
  }
  else {
    yvar <- NULL
  }
  ## map response factors back to original values
  ## does not apply for unsupervised mode
  if (family != "unsupv") {
    if (family == "regr+" | family == "class+" | family == "mix+") {
      yvar <- map.factor(yvar, yfactor)
    }
    else {
      yvar <- amatrix.remove.names(map.factor(yvar, yfactor))
    }
  }
  ## class imbalanced processing
  class.relfrq <- NULL
  if (family == "class" && rfq) {
    class.relfrq <- prop.table(table(yvar))
  }
  ## map imputed data factors back to original values
  ## does NOT apply for y under unsupervised mode
  if ((n.miss > 0) & (nimpute < 2)) {
    imputed.data <- map.factor(imputed.data, xfactor)
    if (family != "unsupv") {
      imputed.data <- map.factor(imputed.data, yfactor)
    }
  }
  ## Define the forest.
  if (forest) {
    nativeArraySize = 0
    mwcpCountSummary <- rep (0, 1)
    nativeFactorArray <- vector("list", 1)
    ## Marker for start of forest topology.  This can
    ## change with the outputs requested.  For the arithmetic
    ## related to the pivot point, refer to
    ## stackOutput.c and in particular,
    ## stackForestOutputObjects().  Do not reference internal.c or
    ## global.h for proxies of these as they may not always map to
    ## stackForestOutputObjects(). The order in the list represented by
    ## nativeOutput[[]] is TOTALLY dependent on the order in
    ## which stackAndProtect() in the C-code is executed.  So,
    ## pay attention to this, when you are not calling the list
    ## elements by their name and instead using the offset nativeOutput[[...]].
    ## The pivot is "one" based.
    pivot <- which(names(nativeOutput) == "treeID")
    ## The offset is irrelevant, no calculations are conducted, so we just safe it.
    offset = 0
    nullO <- lapply(1:ntree, function(b) {
      if (nativeOutput$leafCount[b] > 0) {
        ## The tree was not rejected.  Count the number of internal
        ## and external (terminal) nodes in the forest.
        nativeArraySize <<- nativeArraySize + (2 * nativeOutput$leafCount[b]) - 1
        mwcpCountSummary[1] <<- mwcpCountSummary[1] + nativeOutput$mwcpCT[b]
      }
      else {
        ## The tree was rejected.  However, it acts as a
        ## placeholder, being a stump topologically and thus
        ## adds to the total node count.
        nativeArraySize <<- nativeArraySize + 1
      }
      NULL
    })
    rm(nullO)##memory saving device
    ## Note that we don't save the $blnodeID that is output from the C-side.
    nativeArray <- as.data.frame(cbind(nativeOutput$treeID[1:nativeArraySize],
                                       nativeOutput$nodeID[1:nativeArraySize],
                                       nativeOutput$nodeSZ[1:nativeArraySize],
                                       nativeOutput$brnodeID[1:nativeArraySize]))
    nativeArrayHeader <- c("treeID", "nodeID", "nodeSZ", "brnodeID")
    nativeArray <- as.data.frame(cbind(nativeArray,
                                       nativeOutput$parmID[1:nativeArraySize],
                                       nativeOutput$contPT[1:nativeArraySize],
                                       nativeOutput$mwcpSZ[1:nativeArraySize],
                                       nativeOutput$fsrecID[1:nativeArraySize]))
    nativeArrayHeader <- c(nativeArrayHeader, "parmID", "contPT", "mwcpSZ", "fsrecID") 
    if (mwcpCountSummary[1] > 0) {
      ## This can be NULL if there are no factor splits along this dimension.
      nativeFactorArray[[1]] <- nativeOutput$mwcpPT[1:mwcpCountSummary[1]]
    }
    nativeFactorArrayHeader <- "mwcpPT"
    names(nativeArray) <- nativeArrayHeader
    names(nativeFactorArray) <- nativeFactorArrayHeader
    if (terminal.qualts | terminal.quants) {
      totalLeafCount <- sum(nativeOutput$leafCount)
      valid.mcnt.indices <- 1:totalLeafCount
      if (terminal.quants) {
        ## family specific additions to the grow object
        if (grepl("surv", family)) {
          valid.2D.surv.indices <- 1:(totalLeafCount * length(event.info$event.type) * length(event.info$time.interest))
          valid.1D.surv.indices <- 1:(totalLeafCount * length(event.info$time.interest))
          valid.mort.indices <- 1:(totalLeafCount * length(event.info$event.type))
        }
        else {
          ## This will pick up all "C" and "I".
          class.index <- which(yvar.types != "R")
          class.count <- length(class.index)
          regr.index <- which(yvar.types == "R")
          regr.count <- length(regr.index)
          if (class.count > 0) {
            ## Vector to hold the number of levels in each factor response. 
            levels.count <- array(0, class.count)
            counter <- 0
            for (i in class.index) {
              counter <- counter + 1
              ## Note that [i] is the actual index of the y-variables and not a sequential iterator.
              ## The sequential iteratior is [counter]
              levels.count[counter] <- yvar.nlevels[i]
            }
            valid.clas.indices <- 1:(totalLeafCount * sum(levels.count))
          }
          if (regr.count > 0) {
            valid.regr.indices <- 1: (totalLeafCount * regr.count)
          }
        }
      }
      nativeArrayTNDS <- list(if(!is.null(nativeOutput$tnSURV)) nativeOutput$tnSURV[valid.1D.surv.indices] else NULL,
                              if(!is.null(nativeOutput$tnMORT)) nativeOutput$tnMORT[valid.mort.indices] else NULL,
                              if(!is.null(nativeOutput$tnNLSN)) nativeOutput$tnNLSN[valid.1D.surv.indices] else NULL,
                              if(!is.null(nativeOutput$tnCSHZ)) nativeOutput$tnCSHZ[valid.2D.surv.indices] else NULL,
                              if(!is.null(nativeOutput$tnCIFN)) nativeOutput$tnCIFN[valid.2D.surv.indices] else NULL,
                              if(!is.null(nativeOutput$tnREGR)) nativeOutput$tnREGR[valid.regr.indices] else NULL,
                              if(!is.null(nativeOutput$tnCLAS)) nativeOutput$tnCLAS[valid.clas.indices] else NULL,
                              nativeOutput$rmbrMembership,
                              nativeOutput$ambrMembership,
                              nativeOutput$ombrMembership,
                              nativeOutput$imbrMembership,
                              nativeOutput$tnRCNT[valid.mcnt.indices],
                              nativeOutput$tnACNT[valid.mcnt.indices],
                              nativeOutput$tnOCNT[valid.mcnt.indices],
                              nativeOutput$tnICNT[valid.mcnt.indices],                             
                              nativeOutput$oobSZ,
                              nativeOutput$ibgSZ);
      names(nativeArrayTNDS) <- c("tnSURV","tnMORT","tnNLSN","tnCSHZ","tnCIFN","tnREGR","tnCLAS", "tnRMBR", "tnAMBR", "tnOMBR", "tnIMBR", "tnRCNT", "tnACNT", "tnOCNT", "tnICNT", "oobSZ", "ibgSZ")
    }
    else {
      nativeArrayTNDS <- NULL
    }
    forest.out <- list(forest = TRUE,
                       nativeArray = nativeArray,
                       nativeFactorArray = nativeFactorArray,
                       leafCount = nativeOutput$leafCount,
                       totalNodeCount = dim(nativeArray)[1],
                       nodesize = nodesize,
                       nodedepth = nodedepth,
                       ntree = ntree,
                       family = family,
                       n = n,
                       splitrule = splitinfo$name,
                       xvar = xvar,
                       xvar.names = xvar.names,
                       xvar.factor = xfactor,
                       yvar = yvar,
                       yvar.names = yvar.names,
                       yvar.factor = yfactor,
                       block.size = block.size,
                       bootstrap = bootstrap,
                       case.wt = case.wt,
                       event.info = event.info,
                       uno.weights = uno.weights,
                       gk.quantile = gk.quantile,
                       na.action = na.action,
                       nativeArrayTNDS = nativeArrayTNDS,
                       optLoGrow = nativeOutput$optLoGrow,
                       experimental = experimental,
                       data.pass = data.pass,
                       perf.type = perf.type,
                       prob = prob,
                       prob.epsilon = prob.epsilon,
                       quantile.regr = quantile.regr,
                       rfq = rfq,
                       sampsize = sampsize.function,
                       samptype = samptype,
                       samp = samp,
                       subj = subj,
                       subj.names = subj.names,
                       seed = nativeOutput$seed,
                       seedVimp = if (importance == FALSE) NULL else nativeOutput$seedVimp,
                       terminal.qualts = terminal.qualts,
                       terminal.quants = terminal.quants,
                       importance = importance.value,
                       vimp.threshold = vimp.threshold,
                       version = "3.7.0")
    ## family specific additions to the forest object
    if (grepl("surv", family)) {
      forest.out$time.interest <- event.info$time.interest
    }
    ## Initialize the default class of the forest.
    class(forest.out) <- c("rfsrc", "forest", family)
    if (big.data) {
      class(forest.out) <- c(class(forest.out), "bigdata")
    }
  }
  ## the forest is NULL (the user has requested not to save the forest)
  ## add basic information needed for downstream niceties like printing
  else {
    forest.out <- list(forest = FALSE,
                       nodesize = nodesize,
                       nodedepth = nodedepth,
                       ntree = ntree,
                       family = family,
                       n = n,
                       splitrule = splitinfo$name,
                       yvar = yvar,
                       yvar.names = yvar.names,
                       #xvar = xvar,
                       xvar.names = xvar.names,
                       subj = subj,
                       subj.names = subj.names,
                       bootstrap = bootstrap,
                       sampsize = sampsize.function,
                       samptype = samptype,
                       samp = samp,
                       case.wt = case.wt,
                       version = "3.7.0",
                       na.action = na.action,
                       perf.type = perf.type,
                       rfq = rfq,
                       gk.quantile = gk.quantile,
                       quantile.regr = quantile.regr,
                       prob = prob,
                       prob.epsilon = prob.epsilon,
                       block.size = block.size)
  }
  ## process the proximity matrix
  if (proximity != FALSE) {
    proximity.out <- matrix(0, n, n)
    count <- 0
    for (k in 1:n) {
      proximity.out[k,1:k] <- nativeOutput$proximity[(count+1):(count+k)]
      proximity.out[1:k,k] <- proximity.out[k,1:k]
      count <- count + k
    }
    nativeOutput$proximity <- NULL
  }
  else {
    proximity.out <- NULL
  }
  ## process the proximity matrix
  if (distance != FALSE) {
    distance.out <- matrix(0, n, n)
    count <- 0
    for (k in 1:n) {
      distance.out[k,1:k] <- nativeOutput$distance[(count+1):(count+k)]
      distance.out[1:k,k] <- distance.out[k,1:k]
      count <- count + k
    }
    nativeOutput$distance <- NULL
  }
  else {
    distance.out <- NULL
  }
  ## process weight matrix
  if (forest.wt != FALSE) {
    forest.wt.out <- matrix(nativeOutput$weight, c(n, n), byrow = TRUE)
    nativeOutput$weight <- NULL
  }
  else {
    forest.wt.out <- NULL
  }
  ## process case.depth
  if (case.depth != FALSE) {
    case.depth.out <- matrix(nativeOutput$caseDepth, c(ntree, n), byrow = TRUE)
    nativeOutput$caseDepth <- NULL
  }
  else {
    case.depth.out <- NULL
  }
  ## membership
  if (membership) {
    membership.out <- matrix(nativeOutput$nodeMembership, c(n, ntree))
    inbag.out <- matrix(nativeOutput$bootMembership, c(n, ntree))
    nativeOutput$nodeMembership <- NULL
    nativeOutput$bootMembership <- NULL
    ## The result of this calculation is a list of dimension
    ## [[ntree]][[n]] in which each element [[b]]][[i]] is a
    ## vector containing the (multiple) terminal node identifiers
    ## for case [[i]] in tree [[b]].  These vectors are of varying
    ## lengths because each case [[i]] in each tree [[b]] can live
    ## in more than one node when splits on time occur.
    if(!is.null(subj)) {
      tdc.membership.cnt <- matrix(nativeOutput$nodeMembershipTDC[[1]], c(n, ntree))
      tdc.membership.out  <- vector("list", ntree)            
      begin.indx  <- 0
      end.indx  <- 0
      for (i in 1:ntree) {
        temp  <- vector("list", n)
        for (j in 1:n) {
          begin.indx  <- end.indx + 1
          end.indx  <- end.indx + tdc.membership.cnt[j,i]
          temp[[j]]  <- nativeOutput$nodeMembershipTDC[[2]][begin.indx:end.indx] 
        }
        tdc.membership.out[[i]]  <- temp
      }
    }
    else {
      tdc.membership.out  <- NULL
    }
  }
  else {
    membership.out <- NULL
    inbag.out <- NULL
    tdc.membership.out  <- NULL        
  }
  ## variables used
  if (var.used != FALSE) {
    if (var.used == "all.trees") {
      var.used.out <- nativeOutput$varUsed
      names(var.used.out) <- xvar.names
    }
    else {
      var.used.out <- matrix(nativeOutput$varUsed, nrow = ntree, byrow = TRUE)
      colnames(var.used.out) <- xvar.names
    }
    nativeOutput$varUsed <- NULL
  }
  else {
    var.used.out <-  NULL
  }
  ## split depth
  if (split.depth != FALSE) {
    if (split.depth == "all.trees") {
      split.depth.out <- array(nativeOutput$splitDepth, c(n, n.xvar))
    }
    else {
      split.depth.out <- array(nativeOutput$splitDepth, c(n, n.xvar, ntree))
    }
    nativeOutput$splitDepth <- NULL
  }
  else {
    split.depth.out <-  NULL
  }
  empr.risk <- NULL
  oob.empr.risk <- NULL
  if (!is.null(holdout.specs)) {
    holdout.blk <- nativeOutput$holdoutBlk
    nativeOutput$holdoutBlk <- NULL
  }
  else {
    holdout.blk = NULL
  }
  ## make the output object
  rfsrcOutput <- list(
    call = my.call,
    family = family,
    n = n,
    ntree = ntree,
    nimpute = nimpute,
    mtry = mtry,
    nodesize = nodesize,
    nodedepth = nodedepth,
    nsplit = splitinfo$nsplit,
    yvar = yvar,
    yvar.names = yvar.names,
    xvar = if (big.data) NULL else xvar,
    xvar.names = xvar.names,
    event.info = event.info,
    subj = subj,
    subj.names = subj.names,
    xvar.wt = xvar.wt,
    split.wt = split.wt,
    cause.wt = cause.wt,
    leaf.count = nativeOutput$leafCount,
    proximity = proximity.out,
    forest = forest.out,
    forest.wt = forest.wt.out,
    case.depth = case.depth.out,
    distance = distance.out,
    membership = membership.out,
    tdc.membership = tdc.membership.out,
    splitrule = splitinfo$name,
    inbag = inbag.out,
    var.used = var.used.out,
    imputed.indv = (if (n.miss > 0) imputed.indv else NULL),
    imputed.data = (if (n.miss > 0) imputed.data else NULL),
    split.depth  = split.depth.out,
    ensemble = ensemble,
    holdout.array = holdout.array,
    block.size = block.size,
    holdout.blk = holdout.blk,
    empr.risk = empr.risk,
    oob.empr.risk = oob.empr.risk,
    ctime.internal = nativeOutput$cTimeInternal,
    ctime.external = ctime.external.stop - ctime.external.start
  )
  ## memory management
  remove(yvar)
  remove(xvar)
  remove(proximity.out)
  remove(forest.out)
  remove(forest.wt.out)
  remove(case.depth.out)
  remove(distance.out)
  remove(membership.out)
  remove(inbag.out)
  remove(var.used.out)
  if (n.miss > 0) remove(imputed.indv)
  if (n.miss > 0) remove(imputed.data)
  remove(split.depth.out)
  remove(holdout.array)
  remove(empr.risk)
  remove(oob.empr.risk)
  ## save the outputs
  survOutput <- NULL
  classOutput <- NULL
  regrOutput <- NULL
  ## EFFICIENCY EFFICIENCY EFFICIENCY
  ## for efficiency - the following does not need to be executed in impute.only mode
  if (!impute.only) {
    ## family specific additions to the grow object
    if (grepl("surv", family)) {
      if ((length(event.info$event.type) > 1) &&
          (splitinfo$name != "l2.impute") &&
          (splitinfo$name != "logrankscore")) {
        coerced.event.count <- length(event.info$event.type)
      }
      else {
        coerced.event.count <- 1
      }
      if (family == "surv") {
        ## Right Censored names.
        ens.names <- list(NULL, NULL)
        mortality.names <- list(NULL, NULL)
        err.names <- list(NULL, NULL)
        vimp.names <- list(NULL, xvar.names)
      }
      else if (family == "surv-CR") {
        ## Competing Risk names.
        ens.names <- list(NULL, NULL, c(paste("condCHF.", 1:length(event.info$event.type), sep = "")))
        mortality.names <- list(NULL, paste("event.", 1:length(event.info$event.type), sep = ""))
        cif.names <- list(NULL, NULL, c(paste("CIF.", 1:length(event.info$event.type), sep = "")))
        err.names <- list(c(paste("event.", 1:length(event.info$event.type), sep = "")), NULL)
        vimp.names <- list(paste("event.", 1:length(event.info$event.type), sep = ""), xvar.names)
      }
      else {
        ## Time dependent covariates.
        ens.names <- list(NULL, NULL)
      }
      ## From the native code:
      ##   "allEnsbCHF"
      ##   "oobEnsbCHF"
      ## -> of dim [length(event.info$event.type)] x [RF_sortedTimeInterestSize] x [n]
      ##    where [length(event.info$event.type)] may be equal to [1].
      ## To the R code:
      ## -> of dim [n] x [RF_sortedTimeInterestSize] x [length(event.info$event.type)]  
      chf <- (if (!is.null(nativeOutput$allEnsbCHF))
                adrop3d.last(array(nativeOutput$allEnsbCHF,
                                   c(n, length(event.info$time.interest), length(event.info$event.type)),
                                   dimnames=ens.names), coerced.event.count) else NULL)
      nativeOutput$allEnsbCHF <- NULL
      survOutput <- list(chf = chf)
      remove(chf)
      chf.oob <- (if (!is.null(nativeOutput$oobEnsbCHF))
                    adrop3d.last(array(nativeOutput$oobEnsbCHF,
                                       c(n, length(event.info$time.interest), length(event.info$event.type)),
                                       dimnames=ens.names), coerced.event.count) else NULL)
      nativeOutput$oobEnsbCHF <- NULL
      survOutput = c(survOutput, chf.oob = list(chf.oob))
      remove(chf.oob)
      ## From the native code:
      ##   "allEnsbMRT"
      ##   "oobEnsbMRT"
      ## -> of dim [length(event.info$event.type)] x [n]
      ## To the R code:
      ## -> of dim [n] x [length(event.info$event.type)] 
      predicted <- (if (!is.null(nativeOutput$allEnsbMRT))
                      adrop2d.last(array(nativeOutput$allEnsbMRT,
                                         c(n, length(event.info$event.type)), dimnames=mortality.names), coerced.event.count) else NULL)
      nativeOutput$allEnsbMRT <- NULL
      survOutput = c(survOutput, predicted = list(predicted))
      remove(predicted)
      predicted.oob <- (if (!is.null(nativeOutput$oobEnsbMRT))
                          adrop2d.last(array(nativeOutput$oobEnsbMRT,
                                             c(n, length(event.info$event.type)), dimnames=mortality.names), coerced.event.count) else NULL)
      nativeOutput$oobEnsbMRT <- NULL
      survOutput <- c(survOutput, predicted.oob = list(predicted.oob))
      remove(predicted.oob)
      ## From the native code:
      ##   "allEnsbKHZ"
      ##   "oobEnsbKHZ"
      ## -> of dim [RF_sortedTimeInterestSize] x [subject count]
      ## To the R code:
      ## -> of dim [n] x [RF_sortedTimeInterestSize]
      hazard <-  (if (!is.null(nativeOutput$allEnsbKHZ))
                    matrix(nativeOutput$allEnsbKHZ,
                           c(subj.unique.count, length(event.info$time.interest))) else NULL)
      nativeOutput$allEnsbKHZ <- NULL
      survOutput <- c(survOutput, hazard = list(hazard))
      remove(hazard)
      hazard.oob <-  (if (!is.null(nativeOutput$oobEnsbKHZ))
                        matrix(nativeOutput$oobEnsbKHZ,
                               c(subj.unique.count, length(event.info$time.interest))) else NULL)
      nativeOutput$oobEnsbKHZ <- NULL
      survOutput <- c(survOutput, hazard.oob = list(hazard.oob))
      remove(hazard.oob)
      ## From the native code:
      ##   "allEnsbSRV"
      ##   "oobEnsbSRV"
      ## -> of dim [RF_sortedTimeInterestSize] x [n]
      ## To the R code:
      ## -> of dim [n] x [RF_sortedTimeInterestSize]
      survival <-  (if (!is.null(nativeOutput$allEnsbSRV))
                      matrix(nativeOutput$allEnsbSRV,
                             c(n, length(event.info$time.interest))) else NULL)
      nativeOutput$allEnsbSRV <- NULL
      survOutput <- c(survOutput, survival = list(survival))
      remove(survival)
      survival.oob <-  (if (!is.null(nativeOutput$oobEnsbSRV))
                          matrix(nativeOutput$oobEnsbSRV,
                                 c(n, length(event.info$time.interest))) else NULL)
      nativeOutput$oobEnsbSRV <- NULL
      survOutput <- c(survOutput, survival.oob = list(survival.oob))
      remove(survival.oob)
      ## From the native code:
      ##   "allEnsbCIF"
      ##   "oobEnsbCIF"
      ##   -> of dim [length(event.info$event.type)] x [RF_sortedTimeInterestSize] x [n]
      ## To the native code:
      ##   -> of dim  [n] x [RF_sortedTimeInterestSize] x [length(event.info$event.type)]
      cif <- (if (!is.null(nativeOutput$allEnsbCIF))
                array(nativeOutput$allEnsbCIF,
                      c(n, length(event.info$time.interest), length(event.info$event.type)),
                      dimnames=cif.names) else NULL)
      nativeOutput$allEnsbCIF <- NULL
      survOutput <- c(survOutput, cif = list(cif))
      remove(cif)
      cif.oob <- (if (!is.null(nativeOutput$oobEnsbCIF))
                    array(nativeOutput$oobEnsbCIF,
                          c(n, length(event.info$time.interest), length(event.info$event.type)),
                          dimnames=cif.names) else NULL)
      nativeOutput$oobEnsbCIF <- NULL
      survOutput = c(survOutput, cif.oob = list(cif.oob))
      remove(cif.oob)
      ## From the native code:
      ##   "perfSurv"
      ##   -> of dim [ntree] x length(event.info$event.type)]
      ## To the R code:
      ##   -> of dim [ntree] x length(event.info$event.type)]
      if (!is.null(nativeOutput$perfSurv)) {
        err.rate <- adrop2d.first(array(nativeOutput$perfSurv,
                                        c(length(event.info$event.type), ntree),
                                        dimnames=err.names),
                                  coerced.event.count)
        nativeOutput$perfSurv <- NULL
        if (family == "surv-CR") {
          survOutput = c(survOutput, err.rate = list(t(err.rate)))
        }
        else {
          survOutput = c(survOutput, err.rate = list(err.rate))
        }
        remove(err.rate)
      }
      ## From the native code:
      ##   "blockSurv"
      ##   -> of dim [block.cnt] x length(event.info$event.type)]
      ## To the R code:
      ##   -> of dim [block.cnt] x length(event.info$event.type)]
      if (!is.null(nativeOutput$blockSurv)) {
        err.block.rate <- adrop2d.first(array(nativeOutput$blockSurv,
                                              c(length(event.info$event.type), floor(ntree/block.size)),
                                              dimnames=err.names),
                                        coerced.event.count)
        nativeOutput$blockSurv <- NULL
        if (family == "surv-CR") {
          survOutput = c(survOutput, err.block.rate = list(t(err.block.rate)))
        }
        else {
          survOutput = c(survOutput, err.block.rate = list(err.block.rate))
        }
        remove(err.block.rate)
      }
      ## From the native code:
      ##   "vimpSurv"
      ##   -> of dim [n.xvar] x length(event.info$event.type)]
      ## To the R code:
      ##   -> of dim length(event.info$event.type)] x [n.xvar]
      if (!is.null(nativeOutput$vimpSurv)) {
        importance <- adrop2d.first(array(nativeOutput$vimpSurv,
                                          c(length(event.info$event.type), n.xvar),
                                          dimnames = vimp.names),
                                    coerced.event.count)
        nativeOutput$vimpSurv <- NULL
        if (family == "surv-CR") {
          survOutput = c(survOutput, importance = list(t(importance)))
        }
        else {
          survOutput = c(survOutput, importance = list(importance))
        }
        remove(importance)
      }
      survOutput = c(
        survOutput, list(
                      time.interest = event.info$time.interest,
                      ndead = sum(na.omit(event.info$cens) != 0))
      )
      ## From the native code:
      ##   "holdoutSurv"
      ##   -> of dim [p] x [holdout.blk[.]] x [RF_eventTypeSize]
      ## To the R code:
      ##   -> of dim [[p]] x [RF_eventTypeSize] x [holdout.blk[.]]
      ##              ^^^
      ##    list element can be null
      ##   if no blocks for this x-var  
      ## 2 blocks for X1                                     1 block for X2            0 blocks X3, 1 block X4
      ## --------------------------------------------------- ------------------------- -------------------------
      ## [X1]BX1[1]E1 [X1]BX1[1]E2 [X1]BX1[2]E1 [X1]BX1[2]E2 [X2]BX2[1]E1 [X2]BX2[1]E2 [X4]BX4[1]E1 [X4]BX4[1]E2  
      ## ------------ ------------ ------------ ------------ ------------ ------------ ------------ ------------
      ## 1            2            3            4            5            6            7            8
      ## holdout.blk =    c(2, 1, 0, 1, ...)
      ## offset      =    c(4, 2, 0, 2, ...)
      ## offset.sum  = c(0, 4, 6, 6, 8, ...)
      ## X1:  [[1]] is an array of dim 2 x 2
      ## X2:  [[2]] is an array of dim 2 x 1
      ## X3:  [[3]] is NULL
      ## X4:  [[4]] is an array of dim 2 x 1
      if (!is.null(nativeOutput$holdoutSurv)) {
        holdout.vimp <- vector("list", length(rfsrcOutput$holdout.blk))
        names(holdout.vimp) <- xvar.names
        ## These are lengths of of each x-var dimension.    
        holdout.offset <- rfsrcOutput$holdout.blk * length(event.info$event.type) 
        holdout.offset.sum <- c(0, cumsum(holdout.offset))
        for (i in 1:length(holdout.vimp)) {
          if (rfsrcOutput$holdout.blk[i] > 0) {
            if (length(event.info$event.type) > 1) {
              holdout.vimp[[i]] <- array(nativeOutput$holdoutSurv[(holdout.offset.sum[i] + 1) : holdout.offset.sum[i+1]], c(length(event.info$event.type), rfsrcOutput$holdout.blk[i]))
            }
            else {
              holdout.vimp[[i]] <- nativeOutput$holdoutSurv[(holdout.offset.sum[i] + 1) : holdout.offset.sum[i+1]]
            }
          }
          else {
            holdout.vimp[[i]] = NA
          }
        }
        survOutput = c(survOutput, holdout.vimp = list(holdout.vimp))
        remove(holdout.vimp)
      }
      ## When TRUE we revert to univariate nomenclature for all the outputs.
      if(univariate.nomenclature) {
        rfsrcOutput <- c(rfsrcOutput, survOutput)
      }
      else {
        rfsrcOutput <- c(rfsrcOutput, survOutput = list(survOutput))
      }
    }
    else {
      ## We consider "R", "I", and "C" outcomes.  The outcomes are grouped
      ## by type and sequential.  That is, the first "C" encountered in the
      ## response type vector is in position [[1]] in the classification output
      ## list, the second "C" encountered is in position [[2]] in the
      ## classification output list, and so on.  The same applies to the
      ## regression outputs.  We also have a mapping from the outcome slot back
      ## to the original response vector type, given by the following:
      ## Given yvar.types = c("R", "C", "R", "C", "R" , "I")
      ## regr.index[1] -> 1
      ## regr.index[2] -> 3
      ## regr.index[3] -> 5
      ## clas.index[1] -> 2
      ## clas.index[2] -> 4
      ## clas.index[3] -> 6
      ## This will pick up all "C" and "I".
      class.index <- which(yvar.types != "R")
      resp.clas.count <- length(class.index)
      regr.index <- which(yvar.types == "R")
      resp.regr.count <- length(regr.index)
      if (resp.clas.count > 0) {
        classOutput <- vector("list", resp.clas.count)
        ## Names of the classification outputs.
        names(classOutput) <- yvar.names[class.index]
        ## Vector to hold the number of levels in each factor response. 
        levels.count <- array(0, resp.clas.count)
        ## List to hold the names of levels in each factor response. 
        levels.names <- vector("list", resp.clas.count)
        counter <- 0
        for (i in class.index) {
          counter <- counter + 1
          ## Note that [i] is the actual index of the y-variables and not a sequential iterator.
          ## The sequential iteratior is [counter]
          levels.count[counter] <- yvar.nlevels[i]
          if (yvar.types[i] == "C") {
            ## This an unordered factor.
            ## Here, we don't know the sequence of the unordered factor list, so we identify the factor by name.
            levels.names[[counter]] <- yfactor$levels[[which(yfactor$factor == yvar.names[i])]]
          }
          else {
            ## This in an ordered factor.
            ## Here, we don't know the sequence of the ordered factor list, so we identify the factor by name.
            levels.names[[counter]] <- yfactor$order.levels[[which(yfactor$order == yvar.names[i])]]
          }
        }
        ## Incoming error rates: T=tree R=response L=level
        ## T1R1L0 T1R1L1 T1R1L2 T1R1L3 T1R2L0 T1R2L1 T1R2L2, T2R1L0 T2R1L1 T2R1L2 T2R1L3 T2R2L0 T2R2L1 T2R2L2, ... 
        ## In GROW mode, all class objects are represented in the tree offset calculation.
        ## In PRED mode, the offsets are dependent on the only those targets that are requested!
        ## Yields tree.offset = c(1, 8, ...) 
        tree.offset <- array(1, ntree)
        if (ntree > 1) {
          tree.offset[2:ntree] <- sum(1 + levels.count)
        }
        tree.offset <-  cumsum(tree.offset)
        ## The block offset calculation mirrors the tree offset calculation, but differs in only the primary dimension.
        block.offset <- array(1, floor(ntree/block.size))
        if (floor(ntree/block.size) > 1) {
          block.offset[2:floor(ntree/block.size)] <- sum(1 + levels.count)
        }
        block.offset <-  cumsum(block.offset)
        ## Incoming vimp rates: V=xvar R=response L=level
        ## V1R1L0 V1R1L1 V1R1L2 V1R1L3 V1R1L0 V1R2L1 V1R2L2, V2R1L0 V2R1L1 V2R1L2 V2R1L3 V2R2L0 V2R2L1 V2R2L2, ... 
        ## Yields vimp.offset = c(1, 8, ...) 
        vimp.offset <- array(1, n.xvar)
        if (n.xvar > 1) {
          vimp.offset[2:n.xvar] <- sum(1 + levels.count)
        }
        vimp.offset <-  cumsum(vimp.offset)
        iter.ensb.start <- 0
        iter.ensb.end   <- 0
        ## From the native code:
        ##   "allEnsbCLS"
        ##   "oobEnsbCLS"
        ## -> of dim [resp.clas.count] x [levels.count[]] x [n]
        ##    where this is a ragged array.
        ## From the native code:
        ##   "perfClas"
        ## -> of dim [ntree] x [resp.clas.count] x [1 + levels.count[]]
        ## where the slot [.] x [.] x [1] holds the unconditional error rate.
        ## Note that this is a ragged array.
        ## To the R code:
        ## -> of dim [[resp.clas.count]] x [ntree] x [1 + levels.count[]] 
        ## From the native code:
        ##   "blockClas"
        ## -> of dim [block.cnt] x [resp.clas.count] x [1 + levels.count[]]
        ## where the slot [.] x [.] x [1] holds the unconditional blocked error rate.
        ## Note that this is a ragged array.
        ## To the R code:
        ## -> of dim [[resp.clas.count]] x [block.cnt] x [1 + levels.count[]] 
        ## From the native code:
        ##   "vimpClas"
        ## -> of dim [n.xvar] x [resp.clas.count] x [1 + levels.count[]]
        ## where the slot [.] x [.] x [1] holds the unconditional vimp.
        ## Note that this is a ragged array.
        ## To the R code:
        ## -> of dim [[resp.clas.count]] x [1 + levels.count[]] x [n.xvar] 
        ## From the native code:
        ##   "cseNum"
        ## -> of dim [resp.regr.count] x [n]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [n]
        ## From the native code:
        ##   "cseDen"
        ## -> of dim [n]
        ## To the R code:
        ## -> of dim [n]
        ## From the native code:
        ##   "csvNum"
        ## -> of dim [n.xvar] x [resp.regr.count] x [n]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [n] x [n.xvar]
        ## From the native code:
        ##   "csvDen"
        ## -> of dim [n.xvar] x [n] 
        ## To the R code:
        ## -> of dim  x [n]  x [n.xvar]
        ## Preparing for holdout classification vimp.  See the details and examples further below for
        ## parsing the output.
        if (!is.null(nativeOutput$holdoutClas)) {
          ## These are lengths of of each x-var dimension.    
          holdout.offset.x <- rfsrcOutput$holdout.blk * (sum(1 + levels.count))
          ## This is the cumulative offset for each x-var dimension.
          holdout.offset.sum.x <- c(0, cumsum(holdout.offset.x))
          ## Relative response offset.
          holdout.offset.r <- 0
        }
        for (i in 1:resp.clas.count) {
          iter.ensb.start <- iter.ensb.end
          iter.ensb.end <- iter.ensb.end + (levels.count[i] * n)
          ens.names <- list(NULL, levels.names[[i]])
          err.names <- c("all", levels.names[[i]])
          vimp.names <- list(c("all", levels.names[[i]]), xvar.names)
          predicted <- (if (!is.null(nativeOutput$allEnsbCLS))
                          array(nativeOutput$allEnsbCLS[(iter.ensb.start + 1):iter.ensb.end],
                                c(n, levels.count[i]), dimnames=ens.names) else NULL)
          classOutput[[i]] <- list(predicted = predicted)
          response <- (if (!is.null(predicted)) get.bayes.rule(predicted, class.relfrq) else NULL)
          classOutput[[i]] <- c(classOutput[[i]], class = list(response))
          remove(predicted)
          remove(response)
          predicted.oob <- (if (!is.null(nativeOutput$oobEnsbCLS))
                              array(nativeOutput$oobEnsbCLS[(iter.ensb.start + 1):iter.ensb.end],
                                    c(n, levels.count[i]), dimnames=ens.names) else NULL)
          classOutput[[i]] <- c(classOutput[[i]], predicted.oob = list(predicted.oob))
          response.oob <- (if (!is.null(predicted.oob)) get.bayes.rule(predicted.oob, class.relfrq) else NULL)
          classOutput[[i]] <- c(classOutput[[i]], class.oob = list(response.oob))
          remove(predicted.oob)
          remove(response.oob)
          ## New case specific arrays, numerator:
          cse.num <- (if (!is.null(nativeOutput$cseClas))
                        array(nativeOutput$cseClas[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
          classOutput[[i]] <- c(classOutput[[i]], cse.num = list(cse.num))
          remove(cse.num)
          ## New case specific arrays, denominator:
          cse.den <- (if (!is.null(nativeOutput$cseDen))
                        array(nativeOutput$cseDen[1:n], n) else NULL)
          classOutput[[i]] <- c(classOutput[[i]], cse.den = list(cse.den))
          if (!is.null(nativeOutput$perfClas)) {
            err.rate <- array(0, c(1 + levels.count[i], ntree))
            for (j in 1: (1 + levels.count[i])) {
              err.rate[j, ]  <- nativeOutput$perfClas[tree.offset]
              tree.offset <- tree.offset + 1
            }
            row.names(err.rate) <- err.names
            classOutput[[i]] <- c(classOutput[[i]], err.rate = list(t(err.rate)))
            remove(err.rate)
          }
          if (!is.null(nativeOutput$blockClas)) {
            err.block.rate <- array(0, c(1 + levels.count[i], floor(ntree/block.size)))
            for (j in 1: (1 + levels.count[i])) {
              err.block.rate[j, ]  <- nativeOutput$blockClas[block.offset]
              block.offset <- block.offset + 1
            }
            row.names(err.block.rate) <- err.names
            classOutput[[i]] <- c(classOutput[[i]], err.block.rate = list(t(err.block.rate)))
            remove(err.block.rate)
          }
          ## Incoming csv rates:  V=xvar R=response n=case
          ## In elemental form:
          ## V1R1n1 V1R1n2 V1R1n3, ..., V1R2n1 V1R2n2 V1R2n3, ..., V2R1n1, V2R1n2, V2R1n3, ..., V2R2n1, V2R2n2, V2R2n3, ...
          ## Equivalent in blocks on (n):
          ##       V1R1(n)  V1R2(n)  V1R3(n)  V2R1(n)  V2R2(n)  V2R3(n)  V3R1(n)  V3R2(n)  V3R3(n)
          ## To parse we use:
          ## R1:    iter11                    iter12                     iter13
          ## R2:             iter21                    iter22                     iter23
          ## R3:                      iter31                    iter32                     iter33
          ## All zeros initially and of length for this [[i]] slot in the response list.
          ## Then we make this into an [n] x [n.xvar] array for position [[i]] in the target response.
          csv.idx  <- array(0, n * n.xvar) 
          ## Jesus Christ. For an example, try n = n.xvar = 7, resp.clas.cnt = 2
          ## For response 1, i = 1:   1  2  3  4  5  6  7  8  9 10   21 22 23 24 25 26 27 28 29 30   41 42 43 44 45 46 47 48 49 50
          ## For response 2, i = 2:  11 12 13 14 15 16 17 18 19 20   31 32 33 34 35 36 37 38 39 40   51 52 53 54 55 56 57 58 59 60
          ## Then we make this into an [n] x [n.xvar] array for position [[i]] in the target response.
          for (j in 1:n.xvar) {
            csv.idx[(((j-1) * n) + 1) : (((j-1) * n) + n)]  <- ( ((i-1) * n) + ((j-1) * n * resp.clas.count) + 1 ) : ( ((i-1) * n) + ((j-1) * n * resp.clas.count) + n )
          }
          ## New case specific arrays, numerator:
          csv.num <- (if (!is.null(nativeOutput$csvClas))
                        array(nativeOutput$csvClas[csv.idx], c(n, n.xvar)) else NULL)
          classOutput[[i]] <- c(classOutput[[i]], csv.num = list(csv.num))
          remove(csv.num)
          ## New case specific arrays, denominator:
          csv.den <- (if (!is.null(nativeOutput$csvDen))
                        array(nativeOutput$csvDen, c(n, n.xvar)) else NULL)
          classOutput[[i]] <- c(classOutput[[i]], csv.den = list(csv.den))
          remove(csv.den)
          if (!is.null(nativeOutput$vimpClas)) {
            importance <- array(0, c(1 + levels.count[i], n.xvar), dimnames=vimp.names)
            for (j in 1: (1 + levels.count[i])) {
              importance[j, ]  <- nativeOutput$vimpClas[vimp.offset]
              vimp.offset <- vimp.offset + 1
            }
            classOutput[[i]] <- c(classOutput[[i]], importance = list(t(importance)))
            remove(importance)
          }
          ## From the native code:
          ##   "holdoutClas"
          ##   -> of dim [p] x [holdout.blk[.]] x [resp.clas.count] x [1 + levels.count[]]
          ## To the R code:
          ##   -> of dim [[p]] x [1 + levels.count[]] x [holdout.blk[.]]
          ##              ^^^
          ##    list element can be null
          ##   if no blocks for this x-var  
          ##
          ##   Also note that ONLY the current response [resp.clas.count] is extracted, after its offset is calculated.
          ## Assume two responses:
          ##
          ## X1: 1 of 2 blocks, 1st response (2 class), 2nd response (3 class)
          ##
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## [X1]BX1[1]R1L0 [X1]BX1[1]R1L1 [X1]BX1[1]R1L2 [X1]BX1[1]R2L0 [X1]BX1[1]R2L1 [X1]BX1[1]R2L2 [X1]BX1[1]R2L3 
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## 1              2              3              4              5              6              7            
          ## X1: 2 of 2 blocks, 1st response (2 class), 2nd response (3 class)
          ##
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## [X1]BX1[2]R1L0 [X1]BX1[2]R1L1 [X1]BX1[2]R1L2 [X1]BX1[2]R2L0 [X1]BX1[2]R2L1 [X1]BX1[2]R2L2 [X1]BX1[2]R2L3 
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## 8              9              10             11             12             13             14            
          ##
          ## X2: 1 of 1 blocks, 1st response (2 class), 2nd response (3 class)
          ##
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## [X2]BX2[1]R1L0 [X2]BX2[1]R1L1 [X2]BX2[1]R1L2 [X2]BX2[1]R2L0 [X2]BX2[1]R2L1 [X2]BX2[1]R2L2 [X2]BX2[1]R2L3 
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## 15             16             17             18             19             20             21            
          ##
          ## X3: zero blocks
          ##
          ## X4: 1 of 1 blocks, 1st response (2 class), 2nd response (3 class)
          ##
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## [X4]BX4[1]R1L0 [X4]BX4[1]R1L1 [X4]BX4[1]R1L2 [X4]BX4[1]R2L0 [X4]BX4[1]R2L1 [X4]BX4[1]R2L2 [X4]BX4[1]R2L3 
          ## -------------- -------------- -------------- -------------- -------------- -------------- --------------
          ## 22             23             24             25             26             27             28            
          ##
          ## holdout.blk =    c( 2,  1,  0,  1, ...)
          ## offset      =    c(14,  7,  0,  7, ...)
          ## offset.sum  = c(0, 14, 21, 21, 28, ...)
          ## arrays for R1:
          ## X1:  [[1]] is an array of dim [1 + levels.count[1]] x 2 = 3 x 2
          ## X2:  [[2]] is an array of dim [1 + levels.count[1]] x 1 = 3 x 1
          ## X3:  [[3]] is NULL
          ## X4:  [[4]] is an array of dim [1 + levels.count[1]] x 1 = 3 x 1
          ## arrays for R2:        
          ## X1:  [[1]] is an array of dim [1 + levels.count[2]] x 2 = 4 x 2
          ## X2:  [[2]] is an array of dim [1 + levels.count[2]] x 1 = 4 x 1
          ## X3:  [[3]] is NULL
          ## X4:  [[4]] is an array of dim [1 + levels.count[2]] x 1 = 4 x 1
          if (!is.null(nativeOutput$holdoutClas)) {    
            holdout.vimp <- vector("list", length(rfsrcOutput$holdout.blk))
            names(holdout.vimp) <- xvar.names
            ## Update the relative response offset.
            if (i > 1) {
              holdout.offset.r <- holdout.offset.r + (1 + levels.count[i-1])
            }
            ## R1:  0
            ## R2:  3
            ## Loop over each x-var, k.
            for (k in 1:length(holdout.vimp)) {
              ## Is the block count for this x-var non-zero?
              if (rfsrcOutput$holdout.blk[k] > 0) {
                ## Get the location for this x-var.
                offset.x <- holdout.offset.sum.x[k]
                ## x = 1 -> offset.x =  0
                ## x = 2 -> offset.x = 14
                ## x = 3 -> offset.x = NA
                ## x = 4 -> offset.x = 21
                ## Relative block offset.
                offset.b <- 0
                index.m <- NULL
                ## Loop over each block for this x-var
                for (m in 1:rfsrcOutput$holdout.blk[k]) {
                  if (m > 1) {
                    ## Update the relative block offset.                            
                    offset.b <- offset.b +  (sum(1 + levels.count))
                  }
                  index.m <- c(index.m, seq(from = offset.x + offset.b + holdout.offset.r + 1, by = 1, length.out = levels.count[i] + 1))
                  ## blk = 1 -> offset.b =  0
                  ## blk = 2 -> offset.b =  7
                  ## indices for R1, X1, all blocks:  01:03, 08:10
                  ## indices for R1, X2, all blocks:  15:17
                  ## indices for R1, X3, all blocks:  NULL
                  ## indices for R1, X4, all blocks:  22:24
                  ## indices for R2, X1, all blocks:  04:07, 11:14
                  ## indices for R2, X2, all blocks:  18:21
                  ## indices for R2, X3, all blocks:  NULL
                  ## indices for R2, X4, all blocks:  25:28                
                }
                ## holdout.vimp[[k]] = index.m
                ## holdout.vimp[[k]] = nativeOutput$holdoutClas[index.m]
                holdout.vimp[[k]] = array(nativeOutput$holdoutClas[index.m], c(levels.count[i] + 1, rfsrcOutput$holdout.blk[k]))
              }
              else {
                holdout.vimp[[k]] = NA
              }
            }
            classOutput[[i]] = c(classOutput[[i]], holdout.vimp = list(holdout.vimp))
            remove(holdout.vimp)
          }
        }           
        nativeOutput$allEnsbCLS  <- NULL
        nativeOutput$oobEnsbCLS  <- NULL          
        nativeOutput$perfClas    <- NULL
        nativeOutput$blockClas   <- NULL
        nativeOutput$vimpClas    <- NULL
        nativeOutput$holdoutClas <- NULL
        ## When TRUE we revert to univariate nomenclature for all the outputs.
        if(univariate.nomenclature) {
          if ((resp.clas.count == 1) & (resp.regr.count == 0)) {
            names(classOutput) <- NULL
            rfsrcOutput <- c(rfsrcOutput, unlist(classOutput, recursive=FALSE))
          }
          else {
            rfsrcOutput <- c(rfsrcOutput, classOutput = list(classOutput))
          }
        }
        else {
          rfsrcOutput <- c(rfsrcOutput, classOutput = list(classOutput))
        }
      }
      if (resp.regr.count > 0) {
        regrOutput <- vector("list", resp.regr.count)
        names(regrOutput) <- yvar.names[regr.index]
        ## Incoming: T=tree R=response
        ## T1R1 T1R2, T2R1 T2R2, T3R1 T3R2, ... 
        ## Yields tree.offset = c(1, 3, 5, ...) 
        tree.offset <- array(1, ntree)
        if (ntree > 1) {
          tree.offset[2:ntree] <- length(regr.index)
        }
        tree.offset <-  cumsum(tree.offset)
        ## The block offset calculation mirrors the tree offset calculation, but differs in only the primary dimension.
        block.offset <- array(1, floor(ntree/block.size))
        if (floor(ntree/block.size) > 1) {
          block.offset[2:floor(ntree/block.size)] <- length(regr.index)
        }
        block.offset <-  cumsum(block.offset)
        ## Incoming vimp rates: V=xvar R=response
        ## V1R1 V1R2, V2R1 V2R2, V3R1 V3R2, ... 
        vimp.offset <- array(1, n.xvar)
        if (n.xvar > 1) {
          vimp.offset[2:n.xvar] <- length(regr.index)
        }
        ## Yields vimp.offset = c(1, 3, 5, ...) 
        vimp.offset <-  cumsum(vimp.offset)
        iter.ensb.start <- 0
        iter.ensb.end   <- 0
        iter.qntl.start <- 0
        iter.qntl.end   <- 0
        ## From the native code:
        ##   "allEnsbRGR"
        ##   "oobEnsbRGR"
        ## -> of dim [resp.regr.count] x [n]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [n] 
        ## From the native code:
        ##   "allEnsbQNT"
        ##   "oobEnsbQNT"
        ## -> of dim [resp.regr.count] x [length(prob)] x [n]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [n] x [length(prob)]
        ## From the native code:
        ##   "perfRegr"
        ## -> of dim [ntree] x [resp.regr.count]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [ntree] 
        ## From the native code:
        ##   "blockRegr"
        ## -> of dim [block.cnt] x [resp.regr.count]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [block.cnt]
        ## From the native code:
        ##   "vimpRegr"
        ## -> of dim [n.xvar] x [resp.regr.count]
        ## To the R code:
        ## -> of dim  [[resp.regr.count]] x [n.xvar]
        ## From the native code:
        ##   "cseNum"
        ## -> of dim [resp.regr.count] x [n]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [n]
        ## From the native code:
        ##   "cseDen"
        ## -> of dim [n]
        ## To the R code:
        ## -> of dim [n]
        ## From the native code:
        ##   "csvNum"
        ## -> of dim [n.xvar] x [resp.regr.count] x [n]
        ## To the R code:
        ## -> of dim [[resp.regr.count]] x [n] x [n.xvar]
        ## From the native code:
        ##   "csvDen"
        ## -> of dim [n.xvar] x [n] 
        ## To the R code:
        ## -> of dim  x [n]  x [n.xvar]
        ## Preparing for holdout regression vimp.  See the details and examples further below for
        ## parsing the output.
        if (!is.null(nativeOutput$holdoutRegr)) {
          ## These are lengths of of each x-var dimension.    
          holdout.offset.x <- rfsrcOutput$holdout.blk * resp.regr.count
          ## This is the cumulative offset for each x-var dimension.
          holdout.offset.sum.x <- c(0, cumsum(holdout.offset.x))
          ## Relative response offset.
          holdout.offset.r <- 0
        }
        for (i in 1:resp.regr.count) {
          iter.ensb.start <- iter.ensb.end
          iter.ensb.end <- iter.ensb.end + n
          iter.qntl.start <- iter.qntl.end
          iter.qntl.end <- iter.qntl.end + (length(prob) * n)
          vimp.names <- xvar.names
          predicted <- (if (!is.null(nativeOutput$allEnsbRGR))
                          array(nativeOutput$allEnsbRGR[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
          regrOutput[[i]] <- list(predicted = predicted)
          remove(predicted)
          predicted.oob <- (if (!is.null(nativeOutput$oobEnsbRGR))
                              array(nativeOutput$oobEnsbRGR[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
          regrOutput[[i]] <- c(regrOutput[[i]], predicted.oob = list(predicted.oob))
          remove(predicted.oob)
          ## New case specific arrays, numerator:
          cse.num <- (if (!is.null(nativeOutput$cseRegr))
                        array(nativeOutput$cseRegr[(iter.ensb.start + 1):iter.ensb.end], n) else NULL)
          regrOutput[[i]] <- c(regrOutput[[i]], cse.num = list(cse.num))
          remove(cse.num)
          ## New case specific arrays, denominator:
          cse.den <- (if (!is.null(nativeOutput$cseDen))
                        array(nativeOutput$cseDen[1:n], n) else NULL)
          regrOutput[[i]] <- c(regrOutput[[i]], cse.den = list(cse.den))
          quantile <- (if (!is.null(nativeOutput$allEnsbQNT))
                         array(nativeOutput$allEnsbQNT[(iter.qntl.start + 1):iter.qntl.end],
                               c(n, length(prob))) else NULL)
          regrOutput[[i]] <- c(regrOutput[[i]], quantile = list(quantile))
          remove(quantile)
          quantile.oob <- (if (!is.null(nativeOutput$oobEnsbQNT))
                             array(nativeOutput$oobEnsbQNT[(iter.qntl.start + 1):iter.qntl.end],
                                   c(n, length(prob))) else NULL)
          regrOutput[[i]] <- c(regrOutput[[i]], quantile.oob = list(quantile.oob))
          remove(quantile.oob)
          if (!is.null(nativeOutput$perfRegr)) {
            err.rate <- nativeOutput$perfRegr[tree.offset]
            tree.offset <- tree.offset + 1
            regrOutput[[i]] <- c(regrOutput[[i]], err.rate = list(err.rate))
            remove(err.rate)
          }
          if (!is.null(nativeOutput$blockRegr)) {
            err.block.rate <- nativeOutput$blockRegr[block.offset]
            block.offset <- block.offset + 1
            regrOutput[[i]] <- c(regrOutput[[i]], err.block.rate = list(err.block.rate))
            remove(err.block.rate)
          }
          if (!is.null(nativeOutput$vimpRegr)) {
            importance <- nativeOutput$vimpRegr[vimp.offset]
            names(importance) <- xvar.names
            vimp.offset <- vimp.offset + 1
            regrOutput[[i]] <- c(regrOutput[[i]], importance = list(importance))
            remove(importance)
          }
          ## Incoming csv rates:  V=xvar R=response n=case
          ## In elemental form:
          ## V1R1n1 V1R1n2 V1R1n3, ..., V1R2n1 V1R2n2 V1R2n3, ..., V2R1n1, V2R1n2, V2R1n3, ..., V2R2n1, V2R2n2, V2R2n3, ...
          ## Equivalent in blocks on (n):
          ##       V1R1(n)  V1R2(n)  V1R3(n)  V2R1(n)  V2R2(n)  V2R3(n)  V3R1(n)  V3R2(n)  V3R3(n)
          ## To parse we use:
          ## R1:    iter11                    iter12                     iter13
          ## R2:             iter21                    iter22                     iter23
          ## R3:                      iter31                    iter32                     iter33
          ## All zeros initially and of length for this [[i]] slot in the response list.
          ## Then we make this into an [n] x [n.xvar] array for position [[i]] in the target response.
          csv.idx  <- array(0, n * n.xvar) 
          ## Jesus Christ. For an example, try n = n.xvar = 7, resp.regr.cnt = 2
          ## For response 1, i = 1:   1  2  3  4  5  6  7  8  9 10   21 22 23 24 25 26 27 28 29 30   41 42 43 44 45 46 47 48 49 50
          ## For response 2, i = 2:  11 12 13 14 15 16 17 18 19 20   31 32 33 34 35 36 37 38 39 40   51 52 53 54 55 56 57 58 59 60
          ## Then we make this into an [n] x [n.xvar] array for position [[i]] in the target response.
          for (j in 1:n.xvar) {
            csv.idx[(((j-1) * n) + 1) : (((j-1) * n) + n)]  <- ( ((i-1) * n) + ((j-1) * n * resp.regr.count) + 1 ) : ( ((i-1) * n) + ((j-1) * n * resp.regr.count) + n )
          }
          ## New case specific arrays, numerator:
          csv.num <- (if (!is.null(nativeOutput$csvRegr))
                        array(nativeOutput$csvRegr[csv.idx], c(n, n.xvar)) else NULL)
          regrOutput[[i]] <- c(regrOutput[[i]], csv.num = list(csv.num))
          remove(csv.num)
          ## New case specific arrays, denominator:
          csv.den <- (if (!is.null(nativeOutput$csvDen))
                        array(nativeOutput$csvDen, c(n, n.xvar)) else NULL)
          regrOutput[[i]] <- c(regrOutput[[i]], csv.den = list(csv.den))
          remove(csv.den)
          ## From the native code:
          ##   "holdoutRegr"
          ##   -> of dim [p] x [holdout.blk[.]] x [resp.regr.count]
          ## To the R code:
          ##   -> of dim [[p]] x [holdout.blk[.]]
          ##              ^^^
          ##    list element can be null
          ##   if no blocks for this x-var  
          ##
          ##   Also note that ONLY the current response [resp.regr.count] is extracted, after its offset is calculated.
          ## Assume three responses:
          ##
          ## X1: 1 of 2 blocks, 3 responses
          ##
          ## -------------- -------------- -------------- 
          ## [X1]BX1[1]R1   [X1]BX1[1]R2   [X1]BX1[1]R3   
          ## -------------- -------------- -------------- 
          ## 1              2              3              
          ## X1: 2 of 2 blocks, 1st response (2 class), 2nd response (3 class)
          ##
          ## -------------- -------------- -------------- 
          ## [X1]BX1[2]R1   [X1]BX1[2]R2   [X1]BX1[2]R3   
          ## -------------- -------------- -------------- 
          ## 4              5              6             
          ##
          ## X2: 1 of 1 blocks, 3 responses
          ##
          ## -------------- -------------- -------------- 
          ## [X2]BX2[1]R1   [X2]BX2[1]R2   [X2]BX2[1]R3   
          ## -------------- -------------- -------------- 
          ## 7              8              9             
          ##
          ## X3: zero blocks
          ##
          ## X4: 1 of 1 blocks, 3 responses
          ##
          ## -------------- -------------- --------------
          ## [X4]BX4[1]R1   [X4]BX4[1]R2   [X4]BX4[1]R3  
          ## -------------- -------------- --------------
          ## 10             11             12            
          ##
          ## holdout.blk =    c( 2,  1,  0,  1, ...)
          ## offset      =    c(6, 3,  0,  3, ...)
          ## offset.sum  = c(0, 6, 9,  9, 12, ...)
          ## arrays for R1:
          ## X1:  [[1]] is an array of len 2
          ## X2:  [[2]] is an array of len 1
          ## X3:  [[3]] is NULL
          ## X4:  [[4]] is an array of len 1
          ## arrays for R2:        
          ## X1:  [[1]] is an array of len 2
          ## X2:  [[2]] is an array of len 1
          ## X3:  [[3]] is NULL
          ## X4:  [[4]] is an array of len 1
          if (!is.null(nativeOutput$holdoutRegr)) {
            holdout.vimp <- vector("list", length(rfsrcOutput$holdout.blk))
            names(holdout.vimp) <- xvar.names
            ## Update the relative response offset.
            if (i > 1) {
              holdout.offset.r <- holdout.offset.r + 1
            }
            ## R1:  0
            ## R2:  1
            ## R3:  2
            ## Loop over each x-var, k.
            for (k in 1:length(holdout.vimp)) {
              ## Is the block count for this x-var non-zero?
              if (rfsrcOutput$holdout.blk[k] > 0) {
                ## Get the location for this x-var.
                offset.x <- holdout.offset.sum.x[k]
                ## x = 1 -> offset.x =  0
                ## x = 2 -> offset.x =  6
                ## x = 3 -> offset.x = NA
                ## x = 4 -> offset.x =  9
                ## Relative block offset.
                offset.b <- 0
                index.m <- NULL
                ## Loop over each block for this x-var
                for (m in 1:rfsrcOutput$holdout.blk[k]) {
                  if (m > 1) {
                    ## Update the relative block offset.                            
                    offset.b <- offset.b + resp.regr.count
                  }
                  index.m <- c(index.m, offset.x + offset.b + holdout.offset.r + 1)
                  ## blk = 1 -> offset.b =  0
                  ## blk = 2 -> offset.b =  3
                  ## indices for R1, X1, all blocks:  01,03
                  ## indices for R1, X2, all blocks:  07
                  ## indices for R1, X3, all blocks:  NULL
                  ## indices for R1, X4, all blocks:  10
                  ## indices for R2, X1, all blocks:  02,05
                  ## indices for R2, X2, all blocks:  08
                  ## indices for R2, X3, all blocks:  NULL
                  ## indices for R2, X4, all blocks:  11                
                }
                ## holdout.vimp[[k]] = index.m
                ## holdout.vimp[[k]] = nativeOutput$holdoutRegr[index.m]
                holdout.vimp[[k]] <- nativeOutput$holdoutRegr[index.m]
              }
              else {
                holdout.vimp[[k]] <- NA
              }
            }
            regrOutput[[i]] <- c(regrOutput[[i]], holdout.vimp = list(holdout.vimp))        
          }
        }
        nativeOutput$allEnsbRGR  <- NULL
        nativeOutput$oobEnsbRGR  <- NULL
        nativeOutput$allEnsbQNT  <- NULL          
        nativeOutput$oobEnsbQNT  <- NULL          
        nativeOutput$perfRegr    <- NULL
        nativeOutput$blockRegr   <- NULL
        nativeOutput$vimpRegr    <- NULL
        nativeOutput$holdoutRegr <- NULL
        ## When TRUE we revert to univariate nomenclature for all the outputs.
        if(univariate.nomenclature) {
          if ((resp.clas.count == 0) & (resp.regr.count == 1)) {
            names(regrOutput) <- NULL
            rfsrcOutput <- c(rfsrcOutput, unlist(regrOutput, recursive=FALSE))
          }
          else {
            rfsrcOutput <- c(rfsrcOutput, regrOutput = list(regrOutput))
          }
        }
        else {
          rfsrcOutput <- c(rfsrcOutput, regrOutput = list(regrOutput))
        }
      }
    }
  }## BLOCK OF CODE IS NOT EXECUTED IN IMPUTE.ONLY
  class(rfsrcOutput) <- c("rfsrc", "grow", family)
  #if (big.data) {
  #  class(rfsrcOutput) <- c(class(rfsrcOutput), "bigdata")
  #}
  return(rfsrcOutput)
}

## pulls the quantile object and converts to a list
extract.quantile <- function(o) {
  ## confirm this is a quantile regression object
  if (sum(grepl("quantreg", class(o))) == 0) { 
    stop("object must be a quantreg object")
  }
  ## used to determine univariate vs multivariate objects
  mv.y.names <- intersect(names(o$quantreg), o$yvar.names)
  ## we have a univariate quantile regression object
  if (length(mv.y.names) == 0) {
    q <- list(o$quantreg)
    names(q) <- o$yvar.names
  }
  ## the quantile regression object is multivariate
  else {
    q <- lapply(mv.y.names, function(m.target) {
      o$quantreg[[m.target]]
    })
    names(q) <- mv.y.names
  }
  ## return the processed list
  q
}
## extract target quantiles
get.quantile <- function(o, target.prob = NULL, pretty = TRUE) {
  ## extract the quantile object
  qo <- extract.quantile(o)
  ## process the target probabilities
  if (!is.null(target.prob)) {
    target.prob <- sort(unique(target.prob))
  }
  else {## default is to use existing values
    target.prob <- qo[[1]]$prob
  }
  ## pull the target quantiles
  rO <- lapply(qo, function(q) {    
    q.dat <- do.call(cbind, lapply(target.prob, function(pr) {
      q$quant[, which.min(abs(pr - q$prob))]
    }))
    colnames(q.dat) <-  paste("q.", 100 * target.prob, sep = "")
    q.dat
  })
  if (pretty && length(rO) == 1) {
    rO <- rO[[1]]
  }
  rO
}
## extract crps
get.quantile.crps <- function(o, pretty = TRUE, subset = NULL, standardize = TRUE) {
  ## extract the quantile object
  qO <- extract.quantile(o)
  ## does not apply to predict objects without y
  if (sum(grepl("predict", class(o))) > 0 && is.null(o$yvar)) {
    stop("no yvar present in quantreg predict object")
  }
  ## subset assignment
  if (is.null(subset)) {
    subset <- 1:o$n
  }
  else {
    if (is.logical(subset)) {
      subset <- which(subset)
    }
    subset <- subset[subset >=1 & subset <= o$n]
  }
  if (length(subset) == 0) {
    stop("requested subset is empty")
  }
  ## pull the target stats
  rO <- lapply(1:length(qO), function(j) {
    q <- qO[[j]]
    if (is.vector(o$yvar)) {
      y <- cbind(o$yvar)
    }
    else {
      y <- o$yvar[, names(qO)[j]]
    }
    n <- length(y)
    ## brier score
    brS <- colMeans(do.call(rbind, lapply(subset, function(i) {
      (1 * (y[i] <= q$yunq) - q$cdf[i, ]) ^ 2 
    })), na.rm = TRUE)
    ## crps
    crps <- unlist(lapply(1:length(q$yunq), function(j) {
      if (standardize) {
        trapz(q$yunq[1:j], brS[1:j]) / diff(range(q$yunq[1:j]))
      }
      else {
        trapz(q$yunq[1:j], brS[1:j])
      }
    }))
    data.frame(y = q$yunq, crps = crps)
  })
  names(rO) <- names(qO)
  if (pretty && length(rO) == 1) {
    rO <- rO[[1]]
  }
  rO
}
## extract target stats
get.quantile.stat <- function(o, pretty = TRUE) {
  ## extract the quantile object
  qO <- extract.quantile(o)
  ## conditional median
  mdn <- get.quantile(o, .5, FALSE)
  ## pull the target stats
  rO <- lapply(1:length(qO), function(j) {
    q <- qO[[j]]
    ## conditional mean
    mn <- q$density %*% q$yunq
    ## conditional standard deviation
    std <- sqrt(q$density %*% q$yunq^2 - mn ^ 2)
    data.frame(mean = mn, median = c(mdn[[j]]), std = std)
  })
  names(rO) <- names(qO)
  if (pretty && length(rO) == 1) {
    rO <- rO[[1]]
  }
  rO
}

plot.variable.rfsrc <- function(
  x,
  xvar.names,
  target,
  m.target = NULL,
  time,
  surv.type = c("mort", "rel.freq", "surv", "years.lost", "cif", "chf"),
  class.type = c("prob", "bayes"),
  partial = FALSE,
  oob = TRUE,
  show.plots = TRUE,
  plots.per.page = 4,
  granule = 5,
  sorted = TRUE,
  nvar,
  npts = 25,
  smooth.lines = FALSE,
  subset,
  ...)
{
  ## --------------------------------------------------------------
  ##   
  ##   coherence checks
  ##
  ## --------------------------------------------------------------
  ## is this a synthetic forest?
  if (sum(inherits(x, c("rfsrc", "synthetic"), TRUE) == c(1, 2)) == 2) {
    x <- x$rfSyn
  }
  ## check that object is interpretable
  ## first rename x to object to avoid confusion with x matrix
  object <- x
  remove(x)
  if (sum(inherits(object, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2 &
      sum(inherits(object, c("rfsrc", "plot.variable"), TRUE) == c(1,2)) != 2) {
    stop("this function only works for objects of class `(rfsrc, grow)', '(rfsrc, synthetic)' or '(rfsrc, plot.variable)'")
  }
  if (object$family == "unsupv") {
    stop("this function does not apply to unsupervised forests")
  }
  if (inherits(object, "anonymous")) {
      stop("this function does work with anonymous forests")
  }    
  ## terminate if a partial plot is requested but no forest is found
  if (partial && (is.null(object$forest) || !object$forest$forest)) {
    stop("forest is empty:  re-run rfsrc (grow call) with forest=TRUE")
  }
  ## --------------------------------------------------------------
  ##   
  ##   preliminary processing
  ##
  ## --------------------------------------------------------------
  ## if missing data was imputed then overlay the missing data for x
  ## bug reported by John Ehrlinger
  xvar <- object$xvar
  if (!is.null(object$imputed.indv)) {
    xvar[object$imputed.indv, ] <- object$imputed.data[, object$xvar.names]
  }
  n <- nrow(xvar)
  ## obtain hidden options - set matched options to NULL
  dots <- list(...)
  plot.names <- names(formals(plot.default))
  bxp.names <- unique(c(names(formals(bxp)),
                 "xaxt", "yaxt", "las", "cex.axis", 
                 "col.axis", "main", "cex.main",
                 "col.main", "sub", "cex.sub", "col.sub", 
                 "xlab", "ylab", "cex.lab", "col.lab", "boxfill","varwidth"))
  axis.names <- unique(c(names(formals(axis)), "las", "cex.axis"))
  ## by default xlab and ylab are extracted from the data
  ## over-ridden only by hidden options
  xlab.flag <- ylab.flag <- TRUE
  if (!is.null(dots$xlab)) {
    xlab.flag <- FALSE
  }
  if (!is.null(dots$ylab)) {
    ylab.flag <- FALSE
  }
  ## default notch, boxfill
  if (is.null(dots$notch)) {
    dots$notch <- TRUE
  }
  if (is.null(dots$boxfill)) {
    dots$boxfill <- "bisque"
  }
  ## --------------------------------------------------------------------------------------------
  ## do the following if the class is NOT "plot.variable" 
  ## (i.e. the object has not already been processed by the wraper)
  if (!inherits(object, "plot.variable")) {
    ## process the subsetted index 
    ## assumes the entire data set is to be used if not specified
    if (missing(subset)) {
      subset <- 1:n
    }
      else {
        ## convert the user specified subset into a usable form
        if (is.logical(subset)) subset <- which(subset)
        subset <- unique(subset[subset >= 1 & subset <= n])
        if (length(subset) == 0) {
          stop("'subset' not set properly")
        }
      }
    ## subset the x-data
    xvar <- xvar[subset,, drop = FALSE]
    n <- nrow(xvar)
    ## process the object depending on the underlying family
    family <- object$family
    ## Ensure the coherency of multivariate target.
    m.target <- get.univariate.target(object, m.target)
    ## Survival and competing risk families.
    if (grepl("surv", family)) {
      ##extract event information
      event.info <- get.event.info(object, subset)
      cens <- event.info$cens
      event.type <- event.info$event.type
      ## assign time if missing
      if (missing(time)) {
        time <- median(event.info$time.interest, na.rm = TRUE)
      }
      ## Check for single time point.
      if (length(time) > 1) {
        stop("time must be a single value:  ", time)
      }
      ## CR analysis.
      if (family == "surv-CR") {
        if (missing(target)) {
          target <- 1
        }
          else {
            if (target < 1 || target > max(event.type, na.rm = TRUE)) {
              stop("'target' is specified incorrectly")
            }
          }
        VIMP <- object$importance[, target]
        ## set the surv.type
        surv.type <- setdiff(surv.type, c("mort", "rel.freq", "surv"))[1]
        pred.type <- match.arg(surv.type, c("years.lost", "cif", "chf"))
        ylabel <- switch(pred.type,
                         "years.lost" = paste("Years lost for event ", target),
                         "cif"        = paste("CIF for event ", target, " (time=", time, ")", sep = ""),
                         "chf"        = paste("CHF for event ", target, " (time=", time, ")", sep = ""))
      }
      ## Right-censoring analysis.
        else {
          target <- 1
          VIMP <- object$importance
          ## set the surv.type
          surv.type <- setdiff(surv.type, c("years.lost", "cif", "chf"))[1]
          pred.type <- match.arg(surv.type, c("rel.freq", "mort", "chf", "surv"))
          ylabel <- switch(pred.type,
                           "rel.freq"   = "standardized mortality",
                           "mort"       = "mortality",
                           "chf"        = paste("CHF (time=", time, ")", sep = ""),
                           "surv"       = paste("predicted survival (time=", time, ")", sep = ""))
        }
    }
    ## Univariate and multivariate families.
      else {
        ## assign a null time value
        event.info <- time <- NULL
        ## Factor outcome
        if(is.factor(coerce.multivariate(object, m.target)$yvar)) {
          object.yvar.levels <- levels(coerce.multivariate(object, m.target)$yvar)
          pred.type <- match.arg(class.type, c("prob", "bayes"))
          if (missing(target)) {
            target <- object.yvar.levels[1]
          }
          if (is.character(target)) {
            target <- match(match.arg(target, object.yvar.levels), object.yvar.levels)
          }
            else {
              if ((target > length(object.yvar.levels)) | (target < 1)) {
                stop("target is specified incorrectly:", target)
              }
            }
          if (pred.type == "prob") {
            VIMP <- coerce.multivariate(object, m.target)$importance[, 1 + target]
            ylabel <- paste("probability", object.yvar.levels[target])
            remove(object.yvar.levels)
          }
            else {
              VIMP <- coerce.multivariate(object, m.target)$importance[, 1]
              ylabel <- paste("bayes")
              remove(object.yvar.levels)
            }
        }
        ## Regression outcome.
          else {
            pred.type <- "y"
            target <- NULL
            VIMP <- coerce.multivariate(object, m.target)$importance
            ylabel <- expression(hat(y))
          }
      }
    ## extract x-var names to be plotted
    ## should x-var be sorted by importance?
    if (missing(xvar.names)) {
      xvar.names <- object$xvar.names
    }
      else {
        xvar.names <- intersect(xvar.names, object$xvar.names)
        if (length(xvar.names) ==  0){
          stop("none of the x-variable supplied match available ones:\n", object$xvar.names)
        }
      }
    if (sorted & !is.null(VIMP)) {
      xvar.names <- xvar.names[rev(order(VIMP[xvar.names]))]
    }
    if (!missing(nvar)) {
      nvar <- max(round(nvar), 1)
      xvar.names <- xvar.names[1:min(length(xvar.names), nvar)]
    }
    nvar <- length(xvar.names)
    ## Marginal plot setup
    if (!partial) {
      yhat <- extract.pred(predict(object, m.target = m.target),
                           pred.type,
                           subset,
                           time,
                           m.target,
                           target,
                           oob = oob)
    }
    ## Partial plot setup
    else {
      if (npts < 1) npts <- 1 else npts <- round(npts)
      ## Loop over all x-variables.
      prtl <- lapply(1:nvar, function(k) {
        ## We now allow subsetting of the x-values in the partial mode call.
        ## x <- na.omit(object$xvar[, object$xvar.names == xvar.names[k]])
        ## Bug reported by Amol Pande 16/11/2017
        x <- na.omit(xvar[, object$xvar.names == xvar.names[k]])
        if (is.factor(x)) x <- factor(x, exclude = NULL)          
        n.x <- length(unique(x))
        if (!is.factor(x) & n.x > npts) {
          x.uniq <- sort(unique(x))[unique(as.integer(seq(1, n.x, length = min(npts, n.x))))]
        }
        else {
          if (is.factor(x)) {
            x.uniq <- 1:length(unique(x))##partial.rfsrc requires factors as integers
          }
          else {
            x.uniq <- sort(unique(x))
          }
        }
        n.x <- length(x.uniq)
        yhat <- yhat.se <- NULL
        factor.x <- is.factor(x) | (n.x <= granule)
        pred.temp <- extract.partial.pred(partial.rfsrc(object$forest,
                                                        m.target = m.target,
                                                        partial.type = pred.type,
                                                        partial.xvar = xvar.names[k],
                                                        partial.values = x.uniq,
                                                        partial.time = time,
                                                        oob = oob),
                                          pred.type,
                                          subset,##we now allow subsetting
                                          m.target,
                                          target)
        ## Results in the mean along an x-value over n.
        if (!is.null(dim(pred.temp))) {
          mean.temp <- colMeans(pred.temp, na.rm = TRUE)
        }
        else {
          mean.temp <- mean(pred.temp, na.rm = TRUE)
        }
        if (!factor.x) {
          yhat <- mean.temp
          if (coerce.multivariate(object, m.target)$family == "class") {
            yhat.se <- mean.temp * (1 - mean.temp) / sqrt(n)
          }
          else {
            sd.temp <- apply(pred.temp, 2, sd, na.rm = TRUE)
            yhat.se <- sd.temp / sqrt(n)                
          }
        }
        else {
          if (!is.null(dim(pred.temp))) {
            pred.temp <- t(apply(pred.temp, 1, function(yhat) {
              se <- (yhat - mean.temp) / sqrt(n)
              mean.temp + sign(se) * abs(se)
            }))
          }
          else {
            pred.temp <- mean.temp + (pred.temp - mean.temp) / sqrt(n)
          }
          yhat <- c(yhat, pred.temp)
          x.uniq <- sort(unique(x))##map factor back to original labels
        }
        list(xvar.names = xvar.names[k], yhat = yhat, yhat.se = yhat.se, n.x = n.x, x.uniq = x.uniq, x = x)
      })
      names(prtl) <- xvar.names
    }
    plots.per.page <- max(round(min(plots.per.page,nvar)), 1)
    granule <- max(round(granule), 1)
    ## save the plot.variable object
    plot.variable.obj <- list(family = family,
                              partial = partial,
                              event.info = event.info,
                              target = target,
                              ylabel = ylabel,
                              n = n,
                              xvar.names = xvar.names,
                              nvar = nvar, 
                              plots.per.page = plots.per.page,
                              granule = granule,
                              smooth.lines = smooth.lines)
    if (partial) {
      plot.variable.obj$pData <- prtl
      #06/15/24: provide this object for convenience of custom user plots
      plot.variable.obj$plotthis <- lapply(prtl, function(pp) {
        if (length(pp$x.uniq) == length(pp$yhat)) {
          data.frame(x=pp$x.uniq, yhat=pp$yhat)
        }
        else {
          data.frame(x=rep(pp$x.uniq, rep(n, pp$n.x)), yhat=pp$yhat)
        }
      })
    }
    else {
      plot.variable.obj$yhat <- yhat
      plot.variable.obj$xvar <- xvar
    }
    ## assign the class
    class(plot.variable.obj) <- c("rfsrc", "plot.variable", family)
  }
  ## --------------------------------------------------------------------------------------------
  ## do the following if the class IS "plot.variable" 
  else {
    plot.variable.obj <- object
    remove(object)
    family <- plot.variable.obj$family
    partial <- plot.variable.obj$partial
    event.info <- plot.variable.obj$event.info
    target <- plot.variable.obj$target
    ylabel <- plot.variable.obj$ylabel
    n <- plot.variable.obj$n
    if (missing(xvar.names)) {
      xvar.names <- plot.variable.obj$xvar.names
    }
    else {
      xvar.names <- intersect(xvar.names, plot.variable.obj$xvar.names)
    }
    nvar <- length(xvar.names)
    plots.per.page <- min(plot.variable.obj$plots.per.page, nvar)
    granule <- plot.variable.obj$granule
    smooth.lines <- plot.variable.obj$smooth.lines
    if (partial) {
      prtl <- plot.variable.obj$pData[xvar.names]
    }
    else {
      yhat <- plot.variable.obj$yhat
      xvar <- plot.variable.obj$xvar[, xvar.names, drop = FALSE]
    }
    if (!is.null(event.info)){
      cens <- event.info$cens
      event.type <- event.info$event.type
    }
  }
  ## save par settings
  if (show.plots) {
    old.par <- par(no.readonly = TRUE)
  }
  ## --------------------------------------------------------------------------------
  ##
  ## marginal plots
  ##
  ## --------------------------------------------------------------------------------
  if (!partial && show.plots) {
    ## graphical set up
    par(mfrow = c(min(plots.per.page, ceiling(nvar / plots.per.page)), plots.per.page))
    if (n > 500) cex.pt <- 0.5 else cex.pt <- 0.75
    ## loop over variables
    for (k in 1:nvar) {
      ##x-stuff
      x <- xvar[, which(colnames(xvar) == xvar.names[k])]
      x.uniq <- unique(x)
      n.x <- length(x.uniq)
      ## hidden options
      if (xlab.flag) {
        dots$xlab <- xvar.names[k]
      }
      if (ylab.flag) {
        dots$ylab <- ylabel
      }
      ## x-continuous-plot
      if (!is.factor(x) & n.x > granule) {
        do.call("plot", c(list(x = x, y = yhat, type = "n"), dots[names(dots) %in% plot.names]))
        if (grepl("surv", family)) {
          points(x[cens == target], yhat[cens == target], pch = 16, col = 4, cex = cex.pt)
          points(x[cens == 0], yhat[cens == 0], pch = 16, cex = cex.pt)
        }
        lines(lowess(x[!is.na(x)], yhat[!is.na(x)]), col = 2, lwd=3)
      }
      ## x-factor-plots
      else {
        if (is.factor(x)) x <- factor(x, exclude = NULL)
        bp <- boxplot(yhat ~ x, na.action = "na.omit", names = rep("", n.x), plot = FALSE)
        do.call("bxp", c(list(z = bp, outline = FALSE, xaxt = "n"), dots[names(dots) %in% bxp.names]))
        at.pretty <- unique(round(pretty(1:n.x, min(30, n.x))))
        at.pretty <- at.pretty[at.pretty >= 1 & at.pretty <= n.x]
        do.call("axis", c(list(side = 1, at = at.pretty,
             labels = format(sort(x.uniq)[at.pretty], trim = TRUE, digits = 4),
             tick = TRUE), dots[names(dots) %in% axis.names]))
        }
    }
  }
  ## --------------------------------------------------------------------------------
  ##
  ## partial plots
  ##
  ## --------------------------------------------------------------------------------
  if (partial && show.plots) {
    ## graphical setup
    plots.per.page <- max(round(min(plots.per.page,nvar)), 1)
    granule <- max(round(granule),1)
    par(mfrow = c(min(plots.per.page, ceiling(nvar/plots.per.page)), plots.per.page))
    ## loop over variables
    for (k in 1:nvar) {
      ## x-stuff
      x <- prtl[[k]]$x
      if (is.factor(x)) x <- factor(x, exclude = NULL)          
      x.uniq <- prtl[[k]]$x.uniq
      n.x <- prtl[[k]]$n.x
      if (n.x > 25) cex.pt <- 0.5 else cex.pt <- 0.75
      ## y-stuff
      yhat <- prtl[[k]]$yhat
      yhat.se <- prtl[[k]]$yhat.se
      factor.x <- !(!is.factor(x) & (n.x > granule))
      ## hidden options
      if (xlab.flag) {
        dots$xlab <- prtl[[k]]$xvar.names
      }
      if (ylab.flag) {
        dots$ylab <- ylabel
      }
      ## x-continuous-plot
      if (!factor.x) {
        do.call("plot", c(list(x = c(min(x), x.uniq, max(x), x.uniq, x.uniq),
                               y = c(NA, yhat, NA, yhat + 2 * yhat.se, yhat - 2 * yhat.se),
                               type = "n"), dots[names(dots) %in% plot.names]))
        points(x.uniq, yhat, pch = 16, cex = cex.pt, col = 2)
        if (any(yhat.se > 0)) {
            if (smooth.lines && !any(is.na(yhat.se))) {
                lines(lowess(x.uniq, yhat + 2 * yhat.se), lty = 3, col = 2)
                lines(lowess(x.uniq, yhat - 2 * yhat.se), lty = 3, col = 2)
            }
            else {
                lines(x.uniq, yhat + 2 * yhat.se, lty = 3, col = 2)
                lines(x.uniq, yhat - 2 * yhat.se, lty = 3, col = 2)
            }
        }
        if (smooth.lines) {
          lines(lowess(x.uniq, yhat), lty = 2, lwd=2)
        }
          else {
            lines(x.uniq, yhat, lty = 2, lwd=2)
          }
        rug(x, ticksize=0.03)
      }
      ## x-factor-plots
        else {
          y.se <- 0.005
          bp <- boxplot(yhat ~ rep(x.uniq, rep(n, n.x)), names = rep("", n.x), plot = FALSE)
          do.call("bxp", c(list(z = bp, outline = FALSE, range = 2,
                                ylim = c(min(bp$stats[1,], na.rm = TRUE) * ( 1 - y.se ),
                                         max(bp$stats[5,], na.rm = TRUE) * ( 1 + y.se )),
                                xaxt = "n"), dots[names(dots) %in% bxp.names]))
          at.pretty <- unique(round(pretty(1:n.x, min(30,n.x))))
          at.pretty <- at.pretty[at.pretty >= 1 & at.pretty <= n.x]
          do.call("axis", c(list(side = 1, at = at.pretty,
             labels = format(sort(x.uniq)[at.pretty], trim = TRUE, digits = 4),
             tick = TRUE), dots[names(dots) %in% axis.names]))
      }
    }
  }
  ## Restore par settings
  if (show.plots) {
    par(old.par)
  }
  ## Return the plot.variable object for reuse
  invisible(plot.variable.obj)
}
plot.variable <- plot.variable.rfsrc

print.rfsrc <- function(x, outcome.target = NULL, ...) {
  has.all.classes <- function(obj, cls) all(cls %in% class(obj))
  ## simple pass-throughs for certain composite objects
  if (has.all.classes(x, c("rfsrc","forest")) ||
      has.all.classes(x, c("rfsrc","plot.variable")) ||
      has.all.classes(x, c("rfsrc","partial")) ||
      has.all.classes(x, c("rfsrc","sidClustering"))) {
    print.default(x)
    return()
  }
  ## mean pinball losses from a quantreg object
  .get.pinball.losses <- function(qobj, taus = c(0.1, 0.5, 0.9), subset = NULL) {
    qL <- extract.quantile(qobj)
    j <- 1L
    q <- qL[[j]]
    y <- if (is.vector(qobj$yvar)) c(qobj$yvar) else qobj$yvar[, names(qL)[j]]
    if (is.null(subset)) subset <- seq_along(y)
    y    <- y[subset]
    cdf  <- q$cdf[subset, , drop = FALSE]
    yunq <- q$yunq
    pinball <- function(y, q, tau) (tau - (y < q)) * (y - q)
    out <- sapply(taus, function(tt) {
      ## invert CDF to get q_tau by monotone interpolation
      qhat <- vapply(seq_len(nrow(cdf)), function(i) {
        suppressWarnings(stats::approx(x = cdf[i, ], y = yunq, xout = tt,
                                ties = "ordered", rule = 2)$y)
      }, numeric(1))
      mean(pinball(y, qhat, tt), na.rm = TRUE)
    })
    names(out) <- paste0("tau=", taus)
    out
  }
  ## subsample/bootsample
  if (has.all.classes(x, c("rfsrc","subsample"))) {
    print.subsample(x, ...)
    return()
  }
  if (has.all.classes(x, c("rfsrc","bootsample"))) {
    print.bootsample(x, ...)
    return()
  }
  ## accepted classes: (rfsrc,grow) or (rfsrc,predict)
  if (!has.all.classes(x, c("rfsrc","grow")) && !has.all.classes(x, c("rfsrc","predict"))) {
    stop("This function only works for objects of class '(rfsrc, grow)' or '(rfsrc, predict)'.")
  }
  grow.mode <- has.all.classes(x, c("rfsrc","grow"))
  ## sampling type label
  if (!is.null(x$forest) && !is.null(x$forest$bootstrap) && x$forest$bootstrap == "by.root") {
    samp.used <- x$forest$samptype
  } else {
    samp.used <- if (!is.null(x$forest)) x$forest$bootstrap else NA
  }
  ## save before coercion
  family.pretty <- family.pretty(x)
  family.org <- x$family
  yvar.dim <- if (!is.null(x$yvar)) ncol(as.data.frame(x$yvar)) else NA_integer_
  ## coerce multivariate -> univariate if requested/needed
  ## save standardized mv error rates if available
  x$univariate <- coerce.multivariate(x, 1)$univariate
  if (is.null(outcome.target) && !x$univariate) {
    mv.err.rate <- as.numeric(get.mv.error(x, FALSE))
    mv.err.rate <- c(mean(mv.err.rate, na.rm=TRUE), mv.err.rate)
  }
  if (!is.null(outcome.target) && !x$univariate) {
    outcome.target <- get.univariate.target(x, outcome.target)
    x <- coerce.multivariate(x, outcome.target)
  }
  ## ---- Metrics containers
  conf.mat <- miss.err <- iratio <- brier.err <- brier.norm.err <- auc.err <- pr.auc.err <- gmean.err <- NULL
  logloss.err <- NULL
  crps.err <- crps.std.err <- NULL
  event.freq.txt <- NULL
  k.class <- NULL
  brier.rand <- brier.norm.rand <- logloss.rand <- NULL
  q.crps <- q.crps.std <- NULL
  q.pinball <- NULL
  ## ---- Classification block
  if (x$family == "class") {
    if (!is.null(x$yvar)) {
      tab <- table(x$yvar)
      event.freq.txt <- paste(paste(names(tab), as.integer(tab), sep = "="), collapse = ", ")
      k.class <- length(tab)
    }
    ## choose predicted probs / classes (prefer OOB)
    prob.mat <- if (!is.null(x$predicted.oob) && !all(is.na(x$predicted.oob))) x$predicted.oob else x$predicted
    class.pred <- if (!is.null(x$class.oob) && !all(is.na(x$class.oob))) x$class.oob else x$class
    if (!is.null(x$err.rate) && !is.null(x$yvar) && !is.null(prob.mat)) {
      conf.mat <- get.confusion(x$yvar, class.pred)
      names(dimnames(conf.mat)) <- c("  observed", "predicted")
      ## robust misclassification: use first K columns
      if (!is.null(k.class) && k.class >= 2) {
        core <- conf.mat[, seq_len(min(k.class, ncol(conf.mat))), drop = FALSE]
        miss.err <- 1 - sum(diag(core)) / sum(core)
      }
      if (k.class > 2) {
        brier.err <- get.brier.error(x$yvar, prob.mat, normalized = FALSE)
        brier.norm.err <- get.brier.error(x$yvar, prob.mat, normalized = TRUE)
        auc.err <- get.auc(x$yvar, prob.mat)
        logloss.err <- get.logloss(x$yvar, prob.mat)
        iratio <- pr.auc.err <- gmean.err <- NULL
      } else if (!is.null(k.class) && k.class == 2) {
        thr <- if (!is.null(x$forest$rfq) && isTRUE(x$forest$rfq)) get.rfq.threshold(x$forest$yvar) else 0.5
        perO <- get.imbalanced.performance(x$yvar, prob.mat, threshold = thr, confusion = TRUE)
        iratio <- perO$iratio
        brier.err <- perO$brier
        brier.norm.err <- perO$brier.norm
        auc.err <- perO$auc
        logloss.err <- perO$logloss
        pr.auc.err <- perO$pr.auc
        gmean.err <- perO$gmean
        ## if you want RFQ confusion, you could set conf.mat <- perO$confusion here
      }
      ## --- Random-classifier strawman (uniform probabilities) ---
      if (!is.null(k.class) && k.class >= 2) {
        lev <- colnames(prob.mat)
        if (is.null(lev) && !is.null(levels(x$yvar))) lev <- levels(x$yvar)
        P0 <- matrix(1 / k.class, nrow = length(x$yvar), ncol = k.class,
                     dimnames = list(NULL, lev))
        brier.rand <- get.brier.error(x$yvar, P0, normalized = FALSE)
        brier.norm.rand <- get.brier.error(x$yvar, P0, normalized = TRUE)
        logloss.rand <- get.logloss(x$yvar, P0)
      }
    }
  }
  ## ---- Survival / CR block
  if (grepl("surv", x$family)) {
    if (!is.null(x$err.rate) && x$family == "surv") {
      bs <- get.brier.survival(x)
      crps.err <- bs$crps
      crps.std.err <- bs$crps.std
    }
    event <- x$event.info$event
    if (!is.null(event)) {
      tab <- table(event)
      event.freq.txt <- if (length(tab)) paste(paste(names(tab), as.integer(tab), sep = "="), collapse = ", ") else "0"
    }
  }
  ## ---- Quantile regression (pinball/CRPS) block
  quantile.flag <- has.all.classes(x, c("rfsrc","quantreg")) && identical(x$family, "regr")
  if (quantile.flag) {
    if (!is.null(x$yvar)) {
      ## CRPS (standardized and raw) from existing internal helper
      qcrps.std.df <- tryCatch(get.quantile.crps(x, pretty = TRUE, standardize = TRUE),
                               error = function(e) NULL)
      qcrps.raw.df <- tryCatch(get.quantile.crps(x, pretty = TRUE, standardize = FALSE),
                               error = function(e) NULL)
      if (is.data.frame(qcrps.std.df) && nrow(qcrps.std.df))
        q.crps.std <- tail(qcrps.std.df$crps, 1L)
      if (is.data.frame(qcrps.raw.df) && nrow(qcrps.raw.df))
        q.crps      <- tail(qcrps.raw.df$crps, 1L)
      ## Pinball loss at user-selectable taus (defaults 0.1, 0.5, 0.9)
      taus <- getOption("rfsrc.pinball.taus", c(0.1, 0.5, 0.9))
      q.pinball <- tryCatch(.get.pinball.losses(x, taus = taus), error = function(e) NULL)
    }
  }
  ## ---- Requested performance error (last row)
  err.rate <- NULL
  r.sq <- NULL
  if (!is.null(x$err.rate)) {
    er <- cbind(x$err.rate)
    if (grepl("surv", x$family)) {
      err.rate <- digits.pretty(er[nrow(er), , drop = TRUE], 8)
    } else if (x$family == "class") {
      if ((grow.mode && x$forest$perf.type == "gmean") || (!grow.mode && x$perf.type == "gmean")) {
        err.rate <- digits.pretty(er[nrow(er), 1], 8)
      } else {
        err.rate <- digits.pretty(er[nrow(er), , drop = TRUE], 8)
      }
    } else if (x$family == "regr") {
      if (!is.null(x$yvar)) {
        r.sq <- 1 - er[nrow(er), ] / var(x$yvar, na.rm = TRUE)
      }
      err.rate <- digits.pretty(er[nrow(er), , drop = TRUE], 8)
    }
  }
  ## adjustment for multivariate families - swap the standardized mv terror
  if (is.null(outcome.target) && !x$univariate) {
    err.rate <- digits.pretty(mv.err.rate, 3)
  }
  ## ensure nsplit exists
  if (is.null(x$nsplit)) x$nsplit <- 0
  ## helper for sampsize printing
  sampsize.pretty <- function(sf, nval) {
    out <- tryCatch({
      if (is.function(sf)) round(sf(nval)) else as.numeric(sf)
    }, error = function(e) NA_real_)
    out
  }
  ## ---------------- GROW MODE ----------------
  if (grow.mode) {
    cat("                         Sample size: ", x$n, "\n", sep = "")
    if (grepl("surv", x$family)) {
      if (!is.null(event.freq.txt) && nzchar(event.freq.txt) && grepl(",", event.freq.txt)) {
        cat("                    Number of events: ", event.freq.txt, "\n", sep = "")
      } else {
        cat("                    Number of deaths: ", x$ndead, "\n", sep = "")
      }
    }
    if (x$family == "class" && !is.null(event.freq.txt))
      cat("           Frequency of class labels: ", event.freq.txt, "\n", sep = "")
    if (!is.null(x$imputed.indv))
      cat("                    Was data imputed: ", "yes", "\n", sep = "")
    cat("                     Number of trees: ", x$ntree, "\n", sep = "")
    cat("           Forest terminal node size: ", x$nodesize, "\n", sep = "")
    cat("       Average no. of terminal nodes: ", digits.pretty(mean(x$leaf.count), 4), "\n", sep = "")
    cat("No. of variables tried at each split: ", x$mtry, "\n", sep = "")
    cat("              Total no. of variables: ", length(x$xvar.names), "\n", sep = "")
    if (!is.null(outcome.target) && !x$univariate) {
      cat("              Total no. of responses: ", yvar.dim, "\n", sep = "")
      cat("         User has requested response: ", outcome.target, "\n", sep = "")
    }
    cat("       Resampling used to grow trees: ", samp.used, "\n", sep = "")
    cat("    Resample size used to grow trees: ", sampsize.pretty(x$forest$sampsize, x$n), "\n", sep = "")
    cat("                            Analysis: ", family.pretty, "\n", sep = "")
    cat("                              Family: ", family.org, "\n", sep = "")
    if (x$nsplit > 0 && x$splitrule != "random") {
      cat("                      Splitting rule: ", paste(x$splitrule, "*random*"), "\n", sep = "")
      cat("       Number of random split points: ", x$nsplit, "\n", sep = "")
    } else {
      cat("                      Splitting rule: ", x$splitrule, "\n", sep = "")
    }
    if (!is.null(err.rate)) {
      if (x$family == "regr" && !is.null(r.sq))
        cat("                     (OOB) R squared: ", digits.pretty(r.sq, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(iratio))
        cat("                    Imbalanced ratio: ", digits.pretty(iratio, 4), "\n", sep = "")
      if (x$family == "class" && !is.null(brier.err))
        cat("                   (OOB) Brier score: ", digits.pretty(brier.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(brier.norm.err))
        cat("        (OOB) Normalized Brier score: ", digits.pretty(brier.norm.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(auc.err))
        cat("                           (OOB) AUC: ", digits.pretty(auc.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(logloss.err))
        cat("                      (OOB) Log-loss: ", digits.pretty(logloss.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(pr.auc.err))
        cat("                        (OOB) PR-AUC: ", digits.pretty(pr.auc.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(gmean.err))
        cat("                        (OOB) G-mean: ", digits.pretty(gmean.err, 8), "\n", sep = "")
      if (x$family == "surv" && !is.null(crps.err))
        cat("                          (OOB) CRPS: ", digits.pretty(crps.err, 8), "\n", sep = "")
      if (x$family == "surv" && !is.null(crps.std.err))
        cat("             (OOB) standardized CRPS: ", digits.pretty(crps.std.err, 8), "\n", sep = "")
      if (quantile.flag && !is.null(q.crps))
        cat("                 (OOB) Quantile-CRPS: ", digits.pretty(q.crps, 8), "\n", sep = "")
      if (quantile.flag && !is.null(q.crps.std))
        cat("           (OOB) Quantile-CRPS (std): ", digits.pretty(q.crps.std, 8), "\n", sep = "")
      if (quantile.flag && length(q.pinball))
        cat("           (OOB) Pinball loss by tau: ", digits.pretty(q.pinball, 8),
            " (tau = ", digits.pretty(taus), ")", "\n", sep = "")
      cat("   (OOB) Requested performance error: ", err.rate, "\n\n", sep = "")
    }
    if (x$family == "class" && !is.null(conf.mat)) {
      if (!is.null(x$predicted.oob) && any(is.na(x$predicted.oob))) {
        cat("Confusion matrix (cases with missing OOB predicted values have been removed):\n\n")
      } else {
        cat("Confusion matrix:\n\n")
      }
      print(conf.mat)
      if (!is.null(miss.err)) cat("\n      (OOB) Misclassification rate: ", miss.err, "\n", sep = "")
      ## Strawman baselines
      if (!is.null(brier.rand) && !is.null(brier.norm.rand) && !is.null(logloss.rand)) {
        cat("\nRandom-classifier baselines (uniform):\n",
            "   Brier: ", digits.pretty(brier.rand, 8),
            "   Normalized Brier: ", digits.pretty(brier.norm.rand, 8),
            "   Log-loss: ", digits.pretty(logloss.rand, 8), "\n", sep = "")
      }
    }
  }
  ## ---------------- PREDICT MODE ----------------
  else {
    cat("  Sample size of test (predict) data: ", x$n, "\n", sep = "")
    if (grepl("surv", x$family) && !is.null(event.freq.txt)) {
      if (grepl(",", event.freq.txt)) {
        cat("       Number of events in test data: ", event.freq.txt, "\n", sep = "")
      } else {
        cat("       Number of deaths in test data: ", event.freq.txt, "\n", sep = "")
      }
    }
    if (!is.null(x$imputed.data))
      cat("               Was test data imputed: ", "yes", "\n", sep = "")
    cat("                Number of grow trees: ", x$ntree, "\n", sep = "")
    cat("  Average no. of grow terminal nodes: ", digits.pretty(mean(x$leaf.count), 4), "\n", sep = "")
    cat("         Total no. of grow variables: ", length(x$xvar.names), "\n", sep = "")
    if (!x$univariate) {
      cat("         Total no. of grow responses: ", yvar.dim, "\n", sep = "")
      cat("         User has requested response: ", outcome.target, "\n", sep = "")
    }
    cat("       Resampling used to grow trees: ", samp.used, "\n", sep = "")
    if (!is.null(outcome.target) && !is.null(x$forest$n))
      cat("    Resample size used to grow trees: ", sampsize.pretty(x$forest$sampsize, x$forest$n), "\n", sep = "")
    cat("                            Analysis: ", family.pretty, "\n", sep = "")
    cat("                              Family: ", family.org, "\n", sep = "")
    if (!is.null(err.rate)) {
      if (x$family == "regr" && !is.null(r.sq))
        cat("                           R squared: ", digits.pretty(r.sq, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(iratio))
        cat("                    Imbalanced ratio: ", digits.pretty(iratio, 4), "\n", sep = "")
      if (x$family == "class" && !is.null(brier.err))
        cat("                         Brier score: ", digits.pretty(brier.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(brier.norm.err))
        cat("              Normalized Brier score: ", digits.pretty(brier.norm.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(auc.err))
        cat("                                 AUC: ", digits.pretty(auc.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(logloss.err))
        cat("                            Log-loss: ", digits.pretty(logloss.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(pr.auc.err))
        cat("                              PR-AUC: ", digits.pretty(pr.auc.err, 8), "\n", sep = "")
      if (x$family == "class" && !is.null(gmean.err))
        cat("                              G-mean: ", digits.pretty(gmean.err, 8), "\n", sep = "")
      if (x$family == "surv" && !is.null(crps.err))
        cat("                                CRPS: ", digits.pretty(crps.err, 8), "\n", sep = "")
      if (x$family == "surv" && !is.null(crps.std.err))
        cat("                     standardized CRPS: ", digits.pretty(crps.std.err, 8), "\n", sep = "")
      if (quantile.flag && !is.null(q.crps))
        cat("                       Quantile-CRPS: ", digits.pretty(q.crps, 8), "\n", sep = "")
      if (quantile.flag && !is.null(q.crps.std))
        cat("                 Quantile-CRPS (std): ", digits.pretty(q.crps.std, 8), "\n", sep = "")
            if (quantile.flag && length(q.pinball))
        cat("                 Pinball loss by tau: ", digits.pretty(q.pinball, 8),
            " (tau = ", digits.pretty(taus), ")", "\n", sep = "")
      cat("         Requested performance error: ", err.rate, "\n\n", sep = "")
    }
    if (x$family == "class" && !is.null(conf.mat)) {
      if (!is.null(x$predicted.oob) && any(is.na(x$predicted.oob))) {
        cat("Confusion matrix (cases with missing OOB predicted values have been removed):\n\n")
      } else {
        cat("Confusion matrix:\n\n")
      }
      print(conf.mat)
      if (!is.null(miss.err)) cat("\n           Misclassification error: ", miss.err, "\n", sep = "")
      ## Strawman baselines
      if (!is.null(brier.rand) && !is.null(brier.norm.rand) && !is.null(logloss.rand)) {
        cat("\nRandom-classifier baselines (uniform):\n",
            "   Brier: ", digits.pretty(brier.rand, 8),
            "   Normalized Brier: ", digits.pretty(brier.norm.rand, 8),
            "   Log-loss: ", digits.pretty(logloss.rand, 8), "\n", sep = "")
      }
    }
  }
}

plot.competing.risk.rfsrc <- function (x, plots.one.page = FALSE, ...) {
  ## Incoming parameter checks.  All are fatal.
  if (is.null(x)) {
    stop("object x is empty!")
  }
  if (sum(inherits(x, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2 &
      sum(inherits(x, c("rfsrc", "predict"), TRUE) == c(1, 2)) != 2) {
    stop("This function only works for objects of class `(rfsrc, grow)' or '(rfsrc, predict)'.")
  }
  if (x$family != "surv-CR") {
    stop("this function only supports competing risk settings")
  }
  ## work-horse plotting function
  comprplot <- function(matx, ylab = "", legend = "", pos = 2) {
    m <- dim(cbind(matx))[2]
    if (m > 1) legend <- paste(legend, 1:m, "  ")
    matplot(x$time.interest, matx, xlab = "Time", ylab = ylab, type = "l",
            col = (1:m), lty = 1, lwd = 3)
    legend(c("bottomleft", "topleft", "topright", "bottomright")[pos],
           legend = legend,
           col = (1:m),
           lty = 1,
           lwd = 3)
  }
  ## save par settings
  opar <- par(no.readonly = TRUE)
  on.exit(par(opar))
  if (plots.one.page) par(mfrow = c(1,1)) else par(mfrow = c(2,2))
  ## acquire the estimators - use OOB whenever possible
  if (!is.null(x$chf.oob)) {
    cschf <- apply(x$chf.oob, c(2, 3), mean, na.rm = TRUE)
    cif <- apply(x$cif.oob, c(2, 3), mean, na.rm = TRUE)
  }
  else {
    cschf <- apply(x$chf, c(2, 3), mean, na.rm = TRUE)
    cif <- apply(x$cif, c(2, 3), mean, na.rm = TRUE)
  }
  cpc <- do.call(cbind, lapply(1:ncol(cif), function(j) {
    cif[, j] / (1 - rowSums(cif[, -j, drop = FALSE]))
  }))
  ## plot the results 
  comprplot(cschf, "Cause-Specific CHF", "CSCHF", 2)
  comprplot(100 * cif, "Probability (%)", "CIF", 2)
  comprplot(100 * cpc, "Probability (%)", "CPC", 2)
}
plot.competing.risk <- plot.competing.risk.rfsrc

impute.learn.rfsrc <- function(formula, data,
                               ntree = 100, nodesize = 1, nsplit = 10,
                               nimpute = 2, fast = FALSE, blocks,
                               mf.q, max.iter = 10, eps = 0.01,
                               ytry = NULL, always.use = NULL, verbose = TRUE,
                               ...,
                               supervised.formula = NULL,
                               supervised.args = list(),
                               full.sweep.options = list(ntree = 100, nsplit = 10),
                               target.mode = c("missing.only", "all"),
                               deployment.xvars = NULL,
                               anonymous = TRUE,
                               learner.prefix = "impute.learner.",
                               learner.root = "learners",
                               out.dir = NULL,
                               wipe = TRUE,
                               keep.models = is.null(out.dir),
                               keep.ximp = FALSE,
                               save.on.fit = !is.null(out.dir),
                               save.ood = TRUE,
                               weight = NULL) {
  target.mode.missing <- missing(target.mode)
  supervised.enabled <- !missing(supervised.formula) && !is.null(supervised.formula)
  formula.supplied <- !missing(formula)
  formula.used <- isTRUE(formula.supplied) && !isTRUE(supervised.enabled)
  if (isTRUE(formula.supplied) && isTRUE(supervised.enabled)) {
    warning("'formula' is ignored when 'supervised.formula' is supplied. ",
            "The raw predictor block is defined by the right-hand side of ",
            "'supervised.formula'.", call. = FALSE)
  }
  if (isTRUE(supervised.enabled) && isTRUE(target.mode.missing)) {
    target.mode <- "all"
  }
  target.mode <- match.arg(target.mode)
  persist.on.fit <- !is.null(out.dir) && isTRUE(save.on.fit)
  if (isTRUE(save.on.fit) && is.null(out.dir)) {
    stop("'save.on.fit = TRUE' requires 'out.dir' to be supplied.",
         call. = FALSE)
  }
  if (!isTRUE(keep.models) && !persist.on.fit) {
    stop("At least one trained-learner storage mode must be enabled. ",
         "Use keep.models = TRUE or supply out.dir with save.on.fit = TRUE.",
         call. = FALSE)
  }
  if (persist.on.fit) {
    .check.fst()
  }
  ## make sure data is a data frame
  if (missing(data)) {
    stop("'data' is missing.", call. = FALSE)
  }
  if (is.character(data) && length(data) == 1L && is.null(dim(data))) {
    stop("'data' must be a data frame-like object, not a character string. ",
         "Did you mean data = ", data, " rather than data = \"", data, "\"?",
         call. = FALSE)
  }
  if (is.atomic(data) && is.null(dim(data)) && !is.list(data)) {
    stop("'data' must be a data frame, matrix, or other tabular object. ",
         "A bare vector is not allowed.",
         call. = FALSE)
  }
  data <- as.data.frame(data, stringsAsFactors = FALSE)
  if (isTRUE(supervised.enabled)) {
    supervised.spec <- .parse.supervised.spec(supervised.formula, data)
    train.input <- .normalize.training.data(supervised.spec$x.data)
    response.data <- supervised.spec$y.data
    supervised.args <- .parse.supervised.args(supervised.args)
  }
  else {
    supervised.spec <- list(enabled = FALSE)
    train.input <- .normalize.training.data(data)
    response.data <- NULL
    supervised.args <- list()
  }
  dropped <- .drop.all.na.train(train.input)
  train.data <- dropped$data
  if (nrow(train.data) == 0L || ncol(train.data) == 0L) {
    stop("No usable rows or columns remain after removing all-NA rows or columns.",
         call. = FALSE)
  }
  if (isTRUE(supervised.enabled)) {
    response.data <- response.data[dropped$keep.rows, , drop = FALSE]
    response.schema <- .build.schema(response.data)
  }
  else {
    response.schema <- NULL
  }
  which.na <- is.na(train.data)
  has.missing <- any(which.na)
  schema <- .build.schema(train.data)
  fs <- .parse.full.sweep.options(full.sweep.options)
  if (!has.missing) {
    if (identical(target.mode, "missing.only")) {
      stop("Training data have no missing values after preprocessing. ",
           "Use target.mode = \"all\" to fit predictive imputation learners ",
           "from complete training data.",
           call. = FALSE)
    }
    train.imputation <- "none; training data are complete"
    fit.seconds <- 0
    ximp <- as.data.frame(train.data, stringsAsFactors = FALSE)
    .msg("Training data are complete. Skipping training-time imputation and fitting ",
         "full-sweep learners directly.", verbose = verbose)
  }
  else {
    .msg("Starting training-time imputation...", verbose = verbose)
    fit.start <- proc.time()[[3]]
    impute.args <- c(
      list(
        data = train.data,
        ntree = ntree,
        nodesize = nodesize,
        nsplit = nsplit,
        nimpute = nimpute,
        fast = fast,
        max.iter = max.iter,
        eps = eps,
        ytry = ytry,
        always.use = always.use,
        verbose = verbose,
        full.sweep = FALSE
      ),
      list(...)
    )
    if (isTRUE(formula.used)) {
      impute.args$formula <- formula
    }
    if (!missing(blocks)) impute.args$blocks <- blocks
    mf.q.missing <- missing(mf.q)
    if (!mf.q.missing) impute.args$mf.q <- mf.q
    train.imputation <- if (mf.q.missing) {
      "selected by impute.rfsrc defaults"
    } else if (length(mf.q) == 1L && !is.na(mf.q) && identical(as.integer(mf.q), 1L)) {
      "missForest"
    } else {
      "multivariate missForest"
    }
    ximp <- tryCatch(
      do.call(impute.rfsrc, impute.args),
      error = function(e) e
    )
    if (inherits(ximp, "error")) {
      stop("Training-time imputation failed: ", conditionMessage(ximp),
           call. = FALSE)
    }
    fit.seconds <- proc.time()[[3]] - fit.start
    .msg("Training-time imputation finished in ", format(fit.seconds, digits = 4),
         " seconds.", verbose = verbose)
    ximp <- as.data.frame(ximp, stringsAsFactors = FALSE)
  }
  if (anyNA(ximp)) {
    stop("Training-time imputation did not return a fully completed raw predictor table.",
         call. = FALSE)
  }
  init <- .compute.init(ximp, schema)
  scale <- .compute.scale(ximp, schema)
  engine <- if (isTRUE(anonymous)) rfsrc.anonymous else rfsrc
  supervised.model <- NULL
  supervised.info <- list(enabled = FALSE)
  augmented.ximp <- ximp
  if (isTRUE(supervised.enabled)) {
    response.rows <- .supervised.usable.rows(response.data)
    if (length(response.rows) != nrow(train.data)) {
      stop("Internal supervised-response bookkeeping failed after training preprocessing.",
           call. = FALSE)
    }
    if (!any(response.rows)) {
      stop("No retained training rows have an observed supervised response.",
           call. = FALSE)
    }
    fit.supervised.formula <- .project.supervised.formula(
      supervised.formula,
      data = data,
      train.names = names(train.data),
      arg.name = "supervised.formula"
    )
    fit.sup.data <- data.frame(
      response.data[response.rows, , drop = FALSE],
      ximp[response.rows, , drop = FALSE],
      check.names = FALSE
    )
    supervised.fit.args <- list(
      formula = fit.supervised.formula,
      data = fit.sup.data,
      forest = TRUE,
      perf.type = "none",
      fast = fast
    )
    if (length(supervised.args) > 0L) {
      supervised.fit.args[names(supervised.args)] <- supervised.args
    }
    .msg("Fitting supervised forest...", verbose = verbose)
    supervised.fit <- tryCatch(
      do.call(rfsrc, supervised.fit.args),
      error = function(e) e
    )
    if (inherits(supervised.fit, "error")) {
      stop("Supervised forest fit failed: ", conditionMessage(supervised.fit),
           call. = FALSE)
    }
    supervised.meta <- .resolve.supervised.fit.meta(supervised.fit, fallback = supervised.spec)
    if (!setequal(supervised.meta$xvar.names, names(train.data))) {
      stop("The supervised forest returned predictor names that do not match the retained training predictors.",
           call. = FALSE)
    }
    aux.used <- .align.supervised.aux(
      .supervised.materialize.aux(
        supervised.fit,
        family = supervised.meta$family,
        oob = TRUE,
        response.schema = response.schema,
        yvar.names = supervised.meta$yvar.names
      )
    )
    if (nrow(aux.used) != sum(response.rows)) {
      stop("The supervised OOB prediction block produced ", nrow(aux.used),
           " row(s), but ", sum(response.rows),
           " training row(s) with observed supervised response were expected.",
           call. = FALSE)
    }
    aux.train <- as.data.frame(
      matrix(NA_real_, nrow = nrow(train.data), ncol = ncol(aux.used)),
      check.names = FALSE
    )
    names(aux.train) <- names(aux.used)
    aux.train[response.rows, ] <- aux.used
    response.omit <- which(!response.rows)
    if (length(response.omit) > 0L) {
      aux.omit.newdata <- .conform.x.to.forest(
        ximp[response.omit, , drop = FALSE],
        supervised.fit,
        ignore.levels = TRUE
      )
      aux.omit.pred <- tryCatch(
        predict(supervised.fit, aux.omit.newdata),
        error = function(e) e
      )
      if (inherits(aux.omit.pred, "error")) {
        stop("Failed to predict supervised auxiliary values for rows with missing response: ",
             conditionMessage(aux.omit.pred), call. = FALSE)
      }
      aux.omit <- .align.supervised.aux(
        .supervised.materialize.aux(
          aux.omit.pred,
          family = supervised.meta$family,
          oob = FALSE,
          response.schema = response.schema,
          yvar.names = supervised.meta$yvar.names
        ),
        expected.names = names(aux.train)
      )
      if (nrow(aux.omit) != length(response.omit)) {
        stop("The supervised prediction block for rows with missing response produced ",
             nrow(aux.omit), " row(s), but ", length(response.omit),
             " were expected.", call. = FALSE)
      }
      aux.train[response.omit, ] <- aux.omit
    }
    aux.invalid <- if (ncol(aux.train) > 0L) {
      rowSums(!is.finite(as.matrix(aux.train))) > 0L
    } else {
      rep(FALSE, nrow(aux.train))
    }
    if (any(aux.invalid)) {
      aux.fill.newdata <- .conform.x.to.forest(
        ximp[aux.invalid, , drop = FALSE],
        supervised.fit,
        ignore.levels = TRUE
      )
      aux.fill.pred <- tryCatch(
        predict(supervised.fit, aux.fill.newdata),
        error = function(e) e
      )
      if (!inherits(aux.fill.pred, "error")) {
        aux.fill <- .align.supervised.aux(
          .supervised.materialize.aux(
            aux.fill.pred,
            family = supervised.meta$family,
            oob = FALSE,
            response.schema = response.schema,
            yvar.names = supervised.meta$yvar.names
          ),
          expected.names = names(aux.train)
        )
        aux.train[aux.invalid, ] <- aux.fill
      }
    }
    if (length(intersect(names(aux.train), names(train.data))) > 0L) {
      stop("Supervised auxiliary names collide with raw predictor names: ",
           paste(intersect(names(aux.train), names(train.data)), collapse = ", "),
           call. = FALSE)
    }
    aux.schema <- .build.schema(aux.train)
    aux.init <- .compute.init(aux.train, aux.schema)
    aux.scale <- .compute.scale(aux.train, aux.schema)
    aux.train <- .fill.supervised.aux(aux.train, aux.init)
    augmented.ximp <- data.frame(ximp, aux.train, check.names = FALSE)
    supervised.model <- supervised.fit
    supervised.info <- list(
      enabled = TRUE,
      formula = paste(deparse(supervised.formula), collapse = ""),
      fit.formula = paste(deparse(fit.supervised.formula), collapse = ""),
      formula.scope = "supervised auxiliary forest",
      family = supervised.meta$family,
      xvar.names = supervised.meta$xvar.names,
      yvar.names = supervised.meta$yvar.names,
      aux.names = names(aux.train),
      aux.role = "predictor.only",
      aux.init = aux.init,
      aux.scale = aux.scale,
      fit.args = supervised.args,
      response.policy = "fit on rows with observed supervised response",
      response.n.used = sum(response.rows),
      response.n.omit = sum(!response.rows),
      subj.names = supervised.spec$subj.names %||% character(0),
      response.schema = response.schema,
      train.source = "predicted.oob then prediction fill",
      predict.source = "predicted",
      recompute.during.sweep = FALSE,
      model.name = .make.learner.name(0L, "supervised", prefix = learner.prefix)
    )
  }
  targets <- .resolve.targets(which.na, target.mode = target.mode)
  predictor.map.raw <- .resolve.predictor.map(targets, names(train.data), deployment.xvars)
  predictor.map <- if (isTRUE(supervised.info$enabled)) {
    .augment.predictor.map(targets, predictor.map.raw, supervised.info$aux.names)
  } else {
    predictor.map.raw
  }
  bad.targets <- targets[lengths(predictor.map[targets]) == 0L]
  if (length(bad.targets) > 0L) {
    stop("Some targets have no available predictors after applying deployment restrictions: ",
         paste(bad.targets, collapse = ", "), call. = FALSE)
  }
  miss.frac <- colMeans(which.na)
  sweep.order <- targets[order(miss.frac[targets], decreasing = FALSE)]
  if (persist.on.fit) {
    if (isTRUE(wipe)) .safe.unlink.dir(out.dir)
    .safe.dir.create(out.dir)
    .safe.dir.create(file.path(out.dir, learner.root))
  }
  models <- setNames(vector("list", length(targets)), targets)
  learners <- setNames(vector("list", length(targets)), targets)
  ood.delta <- if (isTRUE(save.ood)) setNames(vector("list", length(targets)), targets) else NULL
  ood.issues <- if (isTRUE(save.ood)) setNames(vector("list", length(targets)), targets) else NULL
  record.ood.issue <- function(target, message) {
    if (!isTRUE(save.ood)) return(invisible(NULL))
    current <- ood.issues[[target]]
    if (is.null(current)) current <- character(0)
    if (!(message %in% current)) {
      ood.issues[[target]] <<- c(current, message)
    }
    invisible(NULL)
  }
  if (isTRUE(supervised.info$enabled) && isTRUE(persist.on.fit)) {
    learner.path <- file.path(out.dir, learner.root, supervised.info$model.name)
    .msg("Saving supervised forest to ", learner.path, verbose = verbose)
    fast.save(supervised.model, learner.path, testing = FALSE)
  }
  .msg("Training final-sweep learner bank...", verbose = verbose)
  sweep.start <- proc.time()[[3]]
  for (i in seq_along(sweep.order)) {
    yname <- sweep.order[[i]]
    trn <- which(!which.na[, yname])
    tst <- which(which.na[, yname])
    xvars <- predictor.map[[yname]]
    learner.name <- .make.learner.name(i, yname, prefix = learner.prefix)
    learners[[yname]] <- list(
      learner.name = learner.name,
      predictors = xvars,
      predictors.raw = predictor.map.raw[[yname]],
      predictors.supervised = if (isTRUE(supervised.info$enabled)) supervised.info$aux.names else character(0),
      n.obs = length(trn),
      n.missing.train = length(tst),
      status = "pending",
      error = NULL,
      family = NA_character_
    )
    .msg("  [", i, "/", length(sweep.order), "] target = `", yname,
         "`  predictors = ", length(xvars), "  observed rows = ", length(trn),
         verbose = verbose)
    if (length(trn) == 0L) {
      learners[[yname]]$status <- "skipped.all.missing"
      record.ood.issue(yname, "No observed training rows were available for OOD reference.")
      next
    }
    yy <- ximp[trn, yname]
    xtrain <- augmented.ximp[trn, xvars, drop = FALSE]
    response.name <- .make.response.name(names(xtrain))
    fit.df <- data.frame(xtrain, check.names = FALSE)
    fit.df[[response.name]] <- yy
    fit.df <- fit.df[, c(response.name, xvars), drop = FALSE]
    fit.args <- c(
      list(
        formula = stats::as.formula(paste0("`", response.name, "` ~ .")),
        data = fit.df,
        ntree = fs$ntree,
        nodesize = fs$nodesize,
        nsplit = fs$nsplit,
        perf.type = fs$dots$perf.type %||% "none",
        fast = fast
      ),
      fs$dots[names(fs$dots) != "perf.type"]
    )
    grow <- tryCatch(
      do.call(engine, fit.args),
      error = function(e) e
    )
    if (inherits(grow, "error")) {
      learners[[yname]]$status <- "error"
      learners[[yname]]$error <- conditionMessage(grow)
      record.ood.issue(yname, paste0("Learner fit failed: ", conditionMessage(grow)))
      .msg("      fit failed for `", yname, "`: ", conditionMessage(grow),
           verbose = verbose)
      next
    }
    learners[[yname]]$status <- "ok"
    learners[[yname]]$family <- grow$family
    if (isTRUE(save.ood)) {
      pred.oob <- list(
        predicted = grow$predicted.oob,
        class = grow$class.oob
      )
      delta.oob <- tryCatch(
        .compute.ood.delta(yy, pred.oob, schema[[yname]]),
        error = function(e) e
      )
      if (inherits(delta.oob, "error")) {
        record.ood.issue(yname, paste0("Failed to compute OOD reference: ",
                                       conditionMessage(delta.oob)))
      }
      else if (length(delta.oob) != length(trn)) {
        record.ood.issue(yname, "OOB reference length did not match observed rows.")
      }
      else {
        tmp <- rep(NA_real_, nrow(train.data))
        tmp[trn] <- as.numeric(delta.oob)
        ood.delta[[yname]] <- tmp
      }
    }
    if (isTRUE(keep.models)) {
      models[[yname]] <- grow
    }
    if (persist.on.fit) {
      learner.path <- file.path(out.dir, learner.root, learner.name)
      .msg("      saving learner to ", learner.path, verbose = verbose)
      fast.save(grow, learner.path, testing = FALSE)
    }
    if (!isTRUE(keep.models)) {
      rm(grow)
      gc()
    }
  }
  sweep.seconds <- proc.time()[[3]] - sweep.start
  .msg("Final-sweep learner bank finished in ", format(sweep.seconds, digits = 4),
       " seconds.", verbose = verbose)
  ood <- NULL
  if (isTRUE(save.ood)) {
    target.reference <- list()
    valid.ood.targets <- character(0)
    for (yname in targets) {
      delta.y <- ood.delta[[yname]]
      if (is.null(delta.y) || !any(is.finite(delta.y))) {
        if (is.null(ood.issues[[yname]]) || length(ood.issues[[yname]]) == 0L) {
          record.ood.issue(yname, "No finite OOD reference values were available.")
        }
        next
      }
      target.reference[[yname]] <- .make.ood.reference(delta.y)
      valid.ood.targets <- c(valid.ood.targets, yname)
    }
    if (length(valid.ood.targets) > 0L) {
      saved.weight <- .resolve.ood.weight(
        valid.ood.targets,
        weight = weight,
        warn.extra = FALSE
      )
      target.score.train <- do.call(cbind, lapply(valid.ood.targets, function(yname) {
        .eval.ood.reference(ood.delta[[yname]], target.reference[[yname]])
      }))
      colnames(target.score.train) <- valid.ood.targets
      row.score.train <- .aggregate.ood.row(
        target.score.train,
        weight = saved.weight,
        aggregate = "weighted.mean",
        aggregate.args = list()
      )
      row.reference <- .make.ood.reference(row.score.train)
      row.reference.targets <- valid.ood.targets
      row.reference.weight <- saved.weight
      row.reference.aggregate <- "weighted.mean"
      row.reference.aggregate.args <- list()
    }
    else {
      saved.weight <- setNames(numeric(0), character(0))
      target.score.train <- matrix(numeric(0), nrow = nrow(train.data), ncol = 0L,
                                   dimnames = list(NULL, character(0)))
      row.reference <- NULL
      row.reference.targets <- character(0)
      row.reference.weight <- setNames(numeric(0), character(0))
      row.reference.aggregate <- "weighted.mean"
      row.reference.aggregate.args <- list()
    }
    ood <- list(
      reference = "oob",
      aggregate = "weighted.mean",
      aggregate.args = list(),
      target.metric = c(
        numeric = "absolute.error",
        factor = "negative.log.probability",
        fallback = "misclass.or.rank.distance"
      ),
      targets = valid.ood.targets,
      weight = saved.weight,
      target.reference = target.reference,
      train.target.score = target.score.train,
      row.reference = row.reference,
      row.reference.targets = row.reference.targets,
      row.reference.weight = row.reference.weight,
      row.reference.aggregate = row.reference.aggregate,
      row.reference.aggregate.args = row.reference.aggregate.args,
      issues = ood.issues
    )
  }
  manifest <- list(
    spec.version = 2L,
    created.at = format(Sys.time(), tz = "UTC", usetz = TRUE),
    formula = if (isTRUE(formula.used)) paste(deparse(formula), collapse = "") else NULL,
    formula.scope = if (isTRUE(formula.used)) {
      "initial imputation stage only"
    } else if (isTRUE(formula.supplied) && isTRUE(supervised.enabled)) {
      "ignored because supervised.formula defines the raw predictor block"
    } else {
      NULL
    },
    supervised = supervised.info,
    train.imputation = train.imputation,
    columns = names(train.data),
    schema = schema,
    init = init,
    scale = scale,
    targets = targets,
    sweep.order = sweep.order,
    predictor.map.raw = predictor.map.raw,
    predictor.map = predictor.map,
    deployment.xvars = deployment.xvars,
    learners = learners,
    learner.root = learner.root,
    learner.prefix = learner.prefix,
    target.mode = target.mode,
    anonymous = anonymous,
    fast = fast,
    full.sweep.options = full.sweep.options,
    save.ood = save.ood,
    ood = ood,
    train.missing.count = colSums(which.na),
    train.missing.frac = colMeans(which.na),
    n.train = nrow(train.data),
    p.train = ncol(train.data),
    dropped.all.na.rows = sum(!dropped$keep.rows),
    dropped.all.na.cols = names(train.input)[!dropped$keep.cols],
    fit.seconds = fit.seconds,
    sweep.seconds = sweep.seconds,
    call = match.call()
  )
  if (!isTRUE(keep.models) && isTRUE(supervised.info$enabled)) {
    rm(supervised.model)
    gc()
    supervised.model <- NULL
  }
  object <- list(
    manifest = manifest,
    models = if (isTRUE(keep.models)) models else setNames(vector("list", length(targets)), targets),
    supervised.model = if (isTRUE(keep.models) && isTRUE(supervised.info$enabled)) supervised.model else NULL,
    ximp = if (isTRUE(keep.ximp)) ximp else NULL,
    path = if (persist.on.fit) normalizePath(out.dir, mustWork = FALSE) else NULL
  )
  class(object) <- c("impute.learn.rfsrc", "impute.learn")
  if (persist.on.fit) {
    saveRDS(manifest, file.path(out.dir, "manifest.rds"))
    .msg("Wrote manifest: ", file.path(out.dir, "manifest.rds"), verbose = verbose)
  }
  object
}
impute.learn <- impute.learn.rfsrc
save.impute.learn.rfsrc <- function(object, path, wipe = TRUE, verbose = TRUE) {
  if (!inherits(object, "impute.learn.rfsrc")) {
    stop("'object' must inherit from class 'impute.learn.rfsrc'.", call. = FALSE)
  }
  .check.fst()
  object$manifest <- .normalize.impute.learn.manifest(object$manifest)
  learner.root <- object$manifest$learner.root %||% "learners"
  source.path <- if (is.null(object$path)) NULL else normalizePath(object$path, mustWork = FALSE)
  dest.path <- normalizePath(path, mustWork = FALSE)
  if (!is.null(source.path) && identical(source.path, dest.path) && isTRUE(wipe)) {
    .msg("Save path matches the existing imputer path; leaving directory in place.",
         verbose = verbose)
    wipe <- FALSE
  }
  if (isTRUE(wipe)) .safe.unlink.dir(path)
  .safe.dir.create(path)
  .safe.dir.create(file.path(path, learner.root))
  saveRDS(object$manifest, file.path(path, "manifest.rds"))
  .msg("Saved manifest to ", file.path(path, "manifest.rds"), verbose = verbose)
  source.root <- if (is.null(source.path)) NULL else file.path(source.path, learner.root)
  if (.is.supervised.manifest(object$manifest)) {
    sup.mdl <- object$supervised.model
    if (is.null(sup.mdl)) {
      if (is.null(source.root)) {
        stop("The supervised forest is not available in memory and no saved ",
             "imputer path is attached to 'object'.",
             call. = FALSE)
      }
      .msg("Loading supervised forest from attached path before saving.",
           verbose = verbose)
      sup.mdl <- .fast.load.named(
        object$manifest$supervised$model.name,
        source.root,
        label = "supervised forest",
        strict = TRUE
      )
    }
    learner.path <- file.path(path, learner.root, object$manifest$supervised$model.name)
    .msg("Saving supervised forest to ", learner.path, verbose = verbose)
    fast.save(sup.mdl, learner.path, testing = FALSE)
    if (is.null(object$supervised.model)) {
      rm(sup.mdl)
      gc()
    }
  }
  for (target in object$manifest$targets) {
    info <- object$manifest$learners[[target]]
    if (is.null(info) || !identical(info$status, "ok")) next
    mdl <- object$models[[target]]
    if (is.null(mdl)) {
      if (is.null(source.root)) {
        stop("Learner for `", target, "` is not available in memory and no saved ",
             "learner path is attached to 'object'.",
             call. = FALSE)
      }
      .msg("Loading learner for `", target, "` from attached path before saving.",
           verbose = verbose)
      mdl <- .fast.load.learner(target, info, source.root, strict = TRUE)
    }
    learner.path <- file.path(path, learner.root, info$learner.name)
    .msg("Saving learner for `", target, "` to ", learner.path, verbose = verbose)
    fast.save(mdl, learner.path, testing = FALSE)
    if (is.null(object$models[[target]])) {
      rm(mdl)
      gc()
    }
  }
  invisible(load.impute.learn.rfsrc(path, lazy = TRUE, verbose = FALSE))
}
save.impute.learn <- save.impute.learn.rfsrc
load.impute.learn.rfsrc <- function(path, targets = NULL, lazy = TRUE, verbose = TRUE) {
  .check.fst()
  manifest.path <- file.path(path, "manifest.rds")
  if (!file.exists(manifest.path)) {
    stop("Manifest not found: ", manifest.path, call. = FALSE)
  }
  manifest <- .normalize.impute.learn.manifest(readRDS(manifest.path))
  all.targets <- manifest$targets
  if (!is.null(targets)) {
    bad.targets <- setdiff(targets, all.targets)
    if (length(bad.targets) > 0L) {
      warning("Ignoring unknown targets: ", paste(bad.targets, collapse = ", "),
              call. = FALSE)
    }
  }
  use.targets <- if (is.null(targets)) all.targets else intersect(all.targets, targets)
  if (length(use.targets) == 0L) {
    stop("No requested targets were found in manifest.", call. = FALSE)
  }
  manifest$targets <- use.targets
  manifest$sweep.order <- manifest$sweep.order[manifest$sweep.order %in% use.targets]
  manifest$predictor.map <- manifest$predictor.map[use.targets]
  if (!is.null(manifest$predictor.map.raw)) {
    manifest$predictor.map.raw <- manifest$predictor.map.raw[use.targets]
  }
  manifest$learners <- manifest$learners[use.targets]
  models <- setNames(vector("list", length(use.targets)), use.targets)
  supervised.model <- NULL
  object <- list(
    manifest = manifest,
    models = models,
    supervised.model = supervised.model,
    ximp = NULL,
    path = normalizePath(path, mustWork = TRUE)
  )
  class(object) <- c("impute.learn.rfsrc", "impute.learn")
  if (!isTRUE(lazy)) {
    learner.root <- file.path(path, manifest$learner.root)
    if (.is.supervised.manifest(manifest)) {
      .msg("Loading supervised forest into memory...", verbose = verbose)
      object$supervised.model <- .fast.load.named(
        manifest$supervised$model.name,
        learner.root,
        label = "supervised forest",
        strict = TRUE
      )
    }
    .msg("Loading learner bank into memory...", verbose = verbose)
    for (target in use.targets) {
      info <- manifest$learners[[target]]
      if (!identical(info$status, "ok")) next
      .msg("  loading `", target, "`", verbose = verbose)
      object$models[[target]] <- .fast.load.learner(target, info, learner.root,
                                                    strict = TRUE)
    }
  }
  object
}
load.impute.learn <- load.impute.learn.rfsrc
predict.impute.learn.rfsrc <- function(object, newdata,
                                       max.predict.iter = 3L,
                                       eps = 1e-3,
                                       targets = NULL,
                                       restore.integer = TRUE,
                                       cache.learners = c("session", "none", "all"),
                                       verbose = TRUE,
                                       ...) {
  cache.learners <- match.arg(cache.learners)
  object$manifest <- .normalize.impute.learn.manifest(object$manifest)
  prep <- .prepare.impute.learn.newdata(
    object = object,
    newdata = newdata,
    targets = targets,
    max.predict.iter = max.predict.iter,
    eps = eps,
    restore.integer = restore.integer,
    cache.learners = cache.learners,
    verbose = verbose
  )
  attr(prep$data, "impute.learn.info") <- prep$info
  prep$data
}
predict.impute.learn <- predict.impute.learn.rfsrc
impute.ood.rfsrc <- function(object, newdata,
                             targets = NULL,
                             max.predict.iter = 3L,
                             eps = 1e-3,
                             cache.learners = c("all", "session", "none"),
                             weight = NULL,
                             aggregate = c("bounded.product",
                                           "weighted.mean",
                                           "weighted.lp",
                                           "weighted.lp.log",
                                           "top.k"),
                             aggregate.args = list(),
                             return.details = FALSE,
                             return.reconstruction = FALSE,
                             verbose = TRUE,
                             ...) {
  cache.learners <- match.arg(cache.learners)
  if (!inherits(object, "impute.learn.rfsrc")) {
    stop("'object' must inherit from class 'impute.learn.rfsrc'.", call. = FALSE)
  }
  object$manifest <- .normalize.impute.learn.manifest(object$manifest)
  ood.ref <- object$manifest$ood
  if (is.null(ood.ref) || length(ood.ref$targets %||% character(0)) == 0L) {
    stop("No saved OOD reference was found in 'object$manifest$ood'. ",
         "Refit with save.ood = TRUE to enable OOD scoring.",
         call. = FALSE)
  }
  if (!is.null(targets)) {
    bad.targets <- setdiff(targets, object$manifest$targets)
    if (length(bad.targets) > 0L) {
      warning("Ignoring unknown OOD targets: ",
              paste(bad.targets, collapse = ", "),
              call. = FALSE)
    }
  }
  score.targets <- if (is.null(targets)) object$manifest$targets else {
    intersect(object$manifest$targets, targets)
  }
  if (length(score.targets) == 0L) {
    stop("No valid targets requested for OOD scoring.", call. = FALSE)
  }
  prep <- .prepare.impute.learn.newdata(
    object = object,
    newdata = newdata,
    targets = NULL,
    max.predict.iter = max.predict.iter,
    eps = eps,
    restore.integer = TRUE,
    cache.learners = cache.learners,
    verbose = verbose
  )
  use.targets <- intersect(score.targets, ood.ref$targets %||% character(0))
  missing.ref.targets <- setdiff(score.targets, use.targets)
  if (length(missing.ref.targets) > 0L) {
    warning("Skipping targets without a saved OOD reference: ",
            paste(missing.ref.targets, collapse = ", "),
            call. = FALSE)
  }
  if (length(use.targets) == 0L) {
    stop("No requested targets have a saved OOD reference.", call. = FALSE)
  }
  weight <- .resolve.ood.weight(use.targets, weight, default = ood.ref$weight)
  aggregate.spec <- .canonicalize.ood.aggregate(aggregate, aggregate.args)
  completed.data <- prep$working.data
  n <- nrow(completed.data)
  target.delta <- matrix(NA_real_, nrow = n, ncol = length(use.targets),
                         dimnames = list(NULL, use.targets))
  target.score <- matrix(NA_real_, nrow = n, ncol = length(use.targets),
                         dimnames = list(NULL, use.targets))
  ## Optional: store the model-based reconstruction used by OOD scoring.
  ## For each target, this is the learner prediction "target ~ other variables"
  ## evaluated on the completed deployment data.  This is distinct from
  ## prep$data / prep$working.data, which are completed/imputed copies of
  ## newdata rather than the per-target OOD reconstructions.
  target.reconstruction <- if (isTRUE(return.reconstruction)) {
    .na.data.from.schema(object$manifest$schema[use.targets], n)
  } else {
    NULL
  }
  target.issues <- prep$info$target.issues
  if (is.null(target.issues)) {
    target.issues <- setNames(vector("list", length(use.targets)), use.targets)
  }
  for (nm in setdiff(use.targets, names(target.issues))) {
    target.issues[[nm]] <- character(0)
  }
  record.issue <- function(target, message) {
    current <- target.issues[[target]]
    if (is.null(current)) current <- character(0)
    if (!(message %in% current)) {
      target.issues[[target]] <<- c(current, message)
    }
    invisible(NULL)
  }
  disk.load.targets <- prep$info$disk.load.targets %||% character(0)
  unseen.mask <- prep$harmonized$unseen.mask
  unseen.rows <- prep$harmonized$unseen.rows
  .msg("Scoring OOD targets...", verbose = verbose)
  for (target in use.targets) {
    info <- object$manifest$learners[[target]]
    if (!identical(info$status, "ok")) {
      msg <- paste0("No trained learner is available (status = ",
                    info$status %||% "unknown", ").")
      record.issue(target, msg)
      next
    }
    mdl.info <- .predict.get.model(object, target, cache.env = prep$cache.env)
    mdl <- mdl.info$model
    if (isTRUE(mdl.info$loaded.from.disk)) {
      disk.load.targets <- c(disk.load.targets, target)
    }
    if (is.null(mdl)) {
      record.issue(target, mdl.info$error %||% "learner could not be loaded")
      next
    }
    xvars <- object$manifest$predictor.map[[target]]
    pred.df <- completed.data[, xvars, drop = FALSE]
    pred.df <- .conform.x.to.forest(pred.df, mdl)
    pred <- tryCatch(
      predict(mdl, pred.df),
      error = function(e) e
    )
    if (inherits(pred, "error")) {
      record.issue(target, paste0("Prediction failed: ", conditionMessage(pred)))
      if (identical(prep$cache.learners, "none") && is.null(object$models[[target]])) {
        rm(mdl)
        gc()
      }
      next
    }
    if (isTRUE(return.reconstruction)) {
      reconstructed <- .extract.prediction(
        pred,
        info$family,
        object$manifest$schema[[target]]
      )
      if (!is.null(reconstructed) && length(reconstructed) == n) {
        target.reconstruction[[target]] <- reconstructed
      } else {
        record.issue(target, "Failed to extract target reconstruction from prediction object.")
      }
    }
    observed <- prep$harmonized$data[[target]]
    delta <- .compute.ood.delta(observed, pred,
                                object$manifest$schema[[target]])
    target.unseen <- if (target %in% names(unseen.mask)) {
      unseen.mask[[target]]
    } else {
      rep(FALSE, n)
    }
    if (length(target.unseen) > 0L && any(target.unseen)) {
      delta[target.unseen] <- Inf
    }
    score.j <- .eval.ood.reference(delta, ood.ref$target.reference[[target]])
    if (length(target.unseen) > 0L && any(target.unseen)) {
      score.j[target.unseen] <- 1
    }
    target.delta[, target] <- delta
    target.score[, target] <- score.j
    if (identical(prep$cache.learners, "none") && is.null(object$models[[target]])) {
      rm(mdl)
      gc()
    }
  }
  score <- .aggregate.ood.row(
    target.score[, use.targets, drop = FALSE],
    weight = weight,
    aggregate = aggregate.spec$name,
    aggregate.args = aggregate.spec$args
  )
  weight.mask <- matrix(rep(weight > 0, each = n), nrow = n)
  targets.used <- rowSums(is.finite(target.score[, use.targets, drop = FALSE]) & weight.mask)
  score.percentile <- rep(NA_real_, n)
  row.reference.used <- FALSE
  row.reference.mode <- NULL
  row.reference.reason <- NULL
  row.reference.n.train <- 0L
  rebuilt.row.reference <- .rebuild.ood.row.reference(
    train.target.score = ood.ref$train.target.score,
    targets = use.targets,
    weight = weight,
    saved.targets = ood.ref$targets %||% character(0),
    aggregate = aggregate.spec$name,
    aggregate.args = aggregate.spec$args
  )
  if (!is.null(rebuilt.row.reference$reference)) {
    score.percentile <- .eval.ood.reference(score, rebuilt.row.reference$reference)
    row.reference.used <- TRUE
    row.reference.mode <- "recomputed.from.train.target.score"
    row.reference.n.train <- rebuilt.row.reference$n.finite %||% 0L
  }
  else {
    row.reference.reason <- rebuilt.row.reference$reason
    row.reference.targets <- ood.ref$row.reference.targets
    row.reference.weight <- ood.ref$row.reference.weight
    row.reference.aggregate <- ood.ref$row.reference.aggregate %||% ood.ref$aggregate %||% "weighted.mean"
    row.reference.aggregate.args <- ood.ref$row.reference.aggregate.args %||% ood.ref$aggregate.args %||% list()
    has.row.reference.meta <- !is.null(row.reference.targets) && !is.null(row.reference.weight)
    same.targets <- isTRUE(has.row.reference.meta) &&
      setequal(use.targets, row.reference.targets %||% character(0)) &&
      length(use.targets) == length(row.reference.targets %||% character(0))
    same.weight <- isTRUE(has.row.reference.meta) &&
      .same.ood.weight(use.targets, weight, row.reference.weight)
    same.aggregate <- .same.ood.aggregate(
      aggregate = aggregate.spec$name,
      aggregate.args = aggregate.spec$args,
      ref.aggregate = row.reference.aggregate,
      ref.aggregate.args = row.reference.aggregate.args
    )
    if (!is.null(ood.ref$row.reference) && isTRUE(same.targets) &&
        isTRUE(same.weight) && isTRUE(same.aggregate)) {
      score.percentile <- .eval.ood.reference(score, ood.ref$row.reference)
      row.reference.used <- TRUE
      row.reference.mode <- "saved.row.reference"
      row.reference.n.train <- ood.ref$row.reference$n %||% 0L
      row.reference.reason <- NULL
    }
    else if (is.null(row.reference.reason)) {
      if (is.null(ood.ref$row.reference)) {
        row.reference.reason <- "No saved row-level OOD reference is available."
      }
      else if (!isTRUE(has.row.reference.meta)) {
        row.reference.reason <- paste(
          "Saved row-level OOD calibration is legacy and lacks the",
          "training target-score matrix needed for arbitrary test-time",
          "weight and row-aggregate recalibration. Refit with the updated",
          "imputer to enable score.percentile for all target subsets,",
          "weights, and row aggregates."
        )
      }
      else if (!isTRUE(same.targets)) {
        row.reference.reason <- "Saved row-level OOD calibration requires scoring the original target set."
      }
      else if (!isTRUE(same.weight)) {
        row.reference.reason <- "Saved row-level OOD calibration requires the saved row-reference target weights."
      }
      else if (!isTRUE(same.aggregate)) {
        row.reference.reason <- "Saved row-level OOD calibration requires the saved row-reference aggregate."
      }
    }
  }
  if (length(unseen.rows) > 0L && any(unseen.rows)) {
    score[unseen.rows] <- .max.ood.aggregate.value(
      aggregate = aggregate.spec$name,
      aggregate.args = aggregate.spec$args
    )
    if (isTRUE(row.reference.used)) {
      score.percentile[unseen.rows] <- 1
    }
  }
  target.issues <- target.issues[lengths(target.issues) > 0L]
  reconstructed.data <- NULL
  if (isTRUE(return.reconstruction)) {
    ## User-facing reconstructed row: start from the harmonized raw deployment
    ## data and replace scored targets by their OOD reconstruction.  This keeps
    ## the original column layout and excludes supervised auxiliary columns.
    reconstructed.data <- prep$harmonized$data[, object$manifest$columns, drop = FALSE]
    for (target in use.targets) {
      reconstructed.data[[target]] <- target.reconstruction[[target]]
    }
    reconstructed.data <- .restore.schema(
      reconstructed.data,
      object$manifest$schema,
      restore.integer = TRUE
    )
  }
  out <- list(
    score = score,
    score.percentile = score.percentile,
    targets.used = targets.used,
    target.score = if (isTRUE(return.details)) target.score[, use.targets, drop = FALSE] else NULL,
    target.delta = if (isTRUE(return.details)) target.delta[, use.targets, drop = FALSE] else NULL,
    target.reconstruction = if (isTRUE(return.reconstruction)) target.reconstruction[, use.targets, drop = FALSE] else NULL,
    reconstructed.data = reconstructed.data,
    completed.data = if (isTRUE(return.details) && isTRUE(return.reconstruction)) {
      prep$data[, object$manifest$columns, drop = FALSE]
    } else {
      NULL
    },
    info = list(
      targets = use.targets,
      weight = weight,
      aggregate = aggregate.spec$name,
      aggregate.args = aggregate.spec$args,
      added.columns = prep$info$added.columns,
      dropped.extra.columns = prep$info$dropped.extra.columns,
      dropped.response.columns = prep$info$dropped.response.columns,
      unseen.levels = prep$info$unseen.levels,
      unseen.rows = unseen.rows,
      maxed.rows = unseen.rows,
      cache.learners = prep$cache.learners,
      n.disk.loads = length(unique(disk.load.targets)),
      disk.load.targets = unique(disk.load.targets),
      row.reference.used = row.reference.used,
      row.reference.mode = row.reference.mode,
      row.reference.n.train = row.reference.n.train,
      row.reference.reason = row.reference.reason,
      supervised = prep$info$supervised,
      target.issues = target.issues
    )
  )
  if (isTRUE(return.details)) {
    out$info$unseen.mask <- unseen.mask
  }
  class(out) <- c("impute.ood.rfsrc", "impute.ood")
  out
}
impute.ood <- impute.ood.rfsrc
print.impute.learn.rfsrc <- function(x, ...) {
  x$manifest <- .normalize.impute.learn.manifest(x$manifest)
  cat("Predictive imputer (randomForestSRC)\n")
  cat("  imputation:    ", x$manifest$train.imputation %||% "<unknown>", "\n", sep = "")
  cat("  training rows: ", x$manifest$n.train, "\n", sep = "")
  cat("  training cols: ", x$manifest$p.train, "\n", sep = "")
  cat("  targets:       ", length(x$manifest$targets), "\n", sep = "")
  cat("  ood targets:   ", length(x$manifest$ood$targets %||% character(0)), "\n", sep = "")
  cat("  supervised:    ", if (.is.supervised.manifest(x$manifest)) "yes" else "no", "\n", sep = "")
  if (.is.supervised.manifest(x$manifest)) {
    cat("  aux columns:   ", length(x$manifest$supervised$aux.names %||% character(0)), "\n", sep = "")
    cat("  sup family:    ", x$manifest$supervised$family %||% "<unknown>", "\n", sep = "")
  }
  cat("  learner root:  ", x$manifest$learner.root, "\n", sep = "")
  cat("  path:          ", x$path %||% "<memory>", "\n", sep = "")
  invisible(x)
}
print.impute.learn <- print.impute.learn.rfsrc

####################################################################
##
## survival functions
##
####################################################################
get.event.info <- function(obj, subset = NULL) {
  ## survival case
  if (grepl("surv", obj$family)) {
    if (!is.null(obj$yvar)) {
      if (is.null(subset)) {
        subset <- (1:nrow(cbind(obj$yvar)))
      }
        if (is.null(obj$subj)) { 
            r.dim <- 2
            time <- obj$yvar[subset, 1]
            cens <- obj$yvar[subset, 2]
        }
        else {
            r.dim <- 3
            start.time <- obj$yvar[subset, 1]
            time <- obj$yvar[subset, 2]
            cens <- obj$yvar[subset, 3]
        }
      ## censoring must be coded coherently
      if (!all(floor(cens) == abs(cens), na.rm = TRUE)) {
        stop("for survival families censoring variable must be coded as a non-negative integer")
      }
      ## Extract the unique event types.
      event <- na.omit(cens)[na.omit(cens) > 0]
      event.type <- sort(unique(event))
    }
    ##everything else
    else {
      r.dim <- 0
      event <- event.type <- cens <- cens <- time <- NULL
    }
    ## Set grid of time points.
    time.interest <- obj$time.interest
  }
  else {
    ## NULL for other families
    if ((obj$family == "regr+") | (obj$family == "class+")) {
      r.dim <- dim(obj$yvar)[2]
    }
    else {
      r.dim <- 1
    }
    event <- event.type <- cens <- time.interest <- cens <- time <- NULL
  }
  return(list(event = event, event.type = event.type, cens = cens,
              time.interest = time.interest, time = time, r.dim = r.dim))
}
get.grow.event.info <- function(yvar, fmly, need.deaths = TRUE, ntime = NULL) {
  if (grepl("surv", fmly)) {
    ##-----------------------------------------------------------
    ## survival, competing risks, or time dependent covariates
    ##-----------------------------------------------------------
    if (dim(yvar)[2] == 2) {
      ##---------------------------------
      ## survival or competing risks:
      ##---------------------------------
      r.dim <- 2
      time <- yvar[, 1]
      cens <- yvar[, 2]
      start.time <- NULL
      ## censoring must be coded coherently
      if (!all(floor(cens) == abs(cens), na.rm = TRUE)) {
        stop("for survival families censoring variable must be coded as a non-negative integer (perhaps the formula is set incorrectly?)")
      }
      ## check if deaths are available (if user specified)
      if (need.deaths && (all(na.omit(cens) == 0))) {
        stop("no deaths in data!")
      }
      ## Check for event time consistency.
      ## we over-ride this now to allow for negative time (see Stute)
      ##if (!all(na.omit(time) >= 0)) {
      ##  stop("time must be  positive")
      ##}
      ## Extract the unique event types.
      event.type <- unique(na.omit(cens))
      ## Ensure they are all greater than or equal to zero.
      if (sum(event.type >= 0) != length(event.type)) {
        stop("censoring variable must be coded as NA, 0, or greater than 0.")
      }
      ## Discard the censored state, if it exists.
      event <- na.omit(cens)[na.omit(cens) > 0]
      event.type <- unique(event)
      ## Set grid of time points.
      nonMissingOutcome <- which(!is.na(cens) & !is.na(time))
      nonMissingDeathFlag <- (cens[nonMissingOutcome] != 0)
      time.interest <- sort(unique(time[nonMissingOutcome[nonMissingDeathFlag]]))
      ## trim the time points if the user has requested it
      ## we also allow the user to pass requested time points
      if (!is.null(ntime) && !((length(ntime) == 1) && ntime == 0)) {
        if (length(ntime) == 1 && length(time.interest) > ntime) {
          time.interest <- time.interest[
            unique(round(seq.int(1, length(time.interest), length.out = ntime)))]
        }
        if (length(ntime) > 1) {
          time.interest <- unique(sapply(ntime, function(tt) {
            time.interest[max(1, sum(tt >= time.interest, na.rm = TRUE))]
          }))
        }
      }
    }
    ##-------------------------------
    ## time dependent covariates:
    ##-------------------------------
    else {
      r.dim <- 3
      start.time <- yvar[, 1]
      time <- yvar[, 2]
      cens <- yvar[, 3]
      ## censoring must be coded coherently
      if (!all(floor(cens) == abs(cens), na.rm = TRUE)) {
        stop("for survival families censoring variable must be coded as a non-negative integer (perhaps the formula is set incorrectly?)")
      }
      ## check if deaths are available (if user specified)
      if (need.deaths && (all(na.omit(cens) == 0))) {
        stop("no deaths in data!")
      }
      ## Check for event time consistency.
      if (!all(na.omit(time) >= 0)) {
        stop("time must be  positive")
      }
      ## Extract the unique event types.
      event.type <- unique(na.omit(cens))
      ## Ensure they are all greater than or equal to zero.
      if (sum(event.type >= 0) != length(event.type)) {
        stop("censoring variable must be coded as NA, 0, or greater than 0.")
      }
      ## Discard the censored state, if it exists.
      event <- na.omit(cens)[na.omit(cens) > 0]
      event.type <- unique(event)
      ## Set grid of time points.
      nonMissingOutcome <- which(!is.na(cens) & !is.na(time))
      nonMissingDeathFlag <- (cens[nonMissingOutcome] != 0)
      time.interest <- sort(unique(time[nonMissingOutcome[nonMissingDeathFlag]]))
      ## trim the time points if the user has requested it
      ## we also allow the user to pass requested time points
      if (!is.null(ntime) && !((length(ntime) == 1) && ntime == 0)) {
        if (length(ntime) == 1 && length(time.interest) > ntime) {
          ## select evenly spaced values over [0,1] and not event times 
          time.interest <- seq(0,  min(1, max(time[nonMissingOutcome])), length = ntime)
          time.interest <- time.interest[time.interest > 0]
        }
        if (length(ntime) > 1) {
          ## over-ride the default setting and allow the user to specify anything they want between [0,1]
          time.pt <- ntime <= min(1, max(time[nonMissingOutcome])) & ntime > 0
          if (sum(time.pt) == 0) {
            stop("the ntime vector supplied must be between [0,1]:", ntime)
          }
          time.interest <- sort(unique(ntime[time.pt]))
        }
      }
    }
  }
  ##---------------------
  ## other families
  ##---------------------
  else {
    if ((fmly == "regr+") | (fmly == "class+") | (fmly == "mix+")) {
      r.dim <- dim(yvar)[2]
    }
    else {
      if (fmly == "unsupv") {
        r.dim <- 0
      }
      else {
        r.dim <- 1
      }
    }
    event <- event.type <- cens <- time.interest <- cens <- time <- start.time <- NULL
  }
  return(list(event = event, event.type = event.type, cens = cens,
              time.interest = time.interest,
              time = time, start.time = start.time, r.dim = r.dim))
}
## ---------------------------------------------------------------------
##
## rmst 
##
## ---------------------------------------------------------------------
get.rmst <- function(o, tau.horizon = NULL, q = .95) {
  ## incoming parameter checks
  if (is.null(o)) {
    return(NULL)
  }
  if (o$family != "surv") {
    stop("this function only supports right-censored survival settings")
  }
  if (sum(inherits(o, c("rfsrc", "grow"), TRUE) == c(1, 2)) != 2 &
      sum(inherits(o, c("rfsrc", "predict"), TRUE) == c(1, 2)) != 2) {
    stop("This function only works for objects of class `(rfsrc, grow)' or '(rfsrc, predict)'")
  }
  ## extract time, survival (use OOB values if available)
  time <- o$time.interest
  if (is.null(o$survival.oob)) {
    surv <- o$survival.oob
  }
  else {
    surv <- o$survival
  }
  ## set the time horizon
  if (is.null(tau.horizon)) {
    ## can replace this with maximum
    ## tau.horizon <- max(time, na.rm = TRUE)
    tau.horizon <- quantile(time, probs = q, na.rm = TRUE)
  }
  ## adjustment for when time doesn't include tau.horizon
  etime <- sort(unique(c(time, tau.horizon)))
  surv <- cbind(1, surv)[, 1 + sIndex(time, etime)]
  time <- etime
  ## restrict time to tau horizon
  time.pt <- time <= tau.horizon
  ## calculate rmst for the restricted time
  c(surv[, time.pt, drop = FALSE] %*% diff(c(0, time[time.pt])))
}
## ---------------------------------------------------------------------
##
## Brier score and time-dependent AUC
##
## ---------------------------------------------------------------------
## trapezoidal rule
trapz <- function(x, y) {
  x <- as.numeric(x)
  y <- as.numeric(y)
  if (length(x) != length(y)) {
    stop("'x' and 'y' must have the same length.")
  }
  if (length(x) < 2L) {
    return(0)
  }
  idx <- 2:length(x)
  sum((x[idx] - x[idx - 1L]) * (y[idx] + y[idx - 1L]) / 2)
}
## returns an index of positions for evaluating a step function at selected times
sIndex <- function(x, y) {
  findInterval(y, x)
}
## set nodesize
set.nodesize <- function(n, p, nodesize = NULL) {
  if (is.null(nodesize)) {
    if (n <= 300 & p > n) {
      nodesize <- 2
    }
    else if (n <= 300 & p <= n) {
      nodesize <- 5
    }
    else if (n > 300 & n <= 2000) {
      nodesize <- 10
    }
    else {
      nodesize <- n / 200
    }
  }
  nodesize
}
## exact element extraction for stripped RF-SRC objects
##
## The dollar operator performs partial matching on lists.  This matters for
## rfsrc.fast(..., forest = FALSE), where xvar is absent but xvar.names is
## retained: o$forest$xvar can then partially match xvar.names.  Use exact
## extraction throughout the performance-data path.
.surv.get.element <- function(x, name) {
  if (is.null(x) || !is.list(x)) {
    return(NULL)
  }
  x[[name]]
}
## identify a supported RF-SRC object
.surv.object.type <- function(o, allow.forest = TRUE) {
  if (!inherits(o, "rfsrc")) {
    return(NULL)
  }
  if (inherits(o, "grow")) {
    return("grow")
  }
  if (inherits(o, "predict")) {
    return("predict")
  }
  if (allow.forest && inherits(o, "forest")) {
    return("forest")
  }
  NULL
}
## convert a forest object to a grow-like object with OOB predictions
.surv.prepare.forest <- function(o) {
  pred.o <- predict(o, perf.type = "none")
  o[["predicted"]] <- .surv.get.element(pred.o, "predicted")
  o[["predicted.oob"]] <- .surv.get.element(pred.o, "predicted.oob")
  o[["survival"]] <- .surv.get.element(pred.o, "survival")
  o[["survival.oob"]] <- .surv.get.element(pred.o, "survival.oob")
  if (is.null(.surv.get.element(o, "time.interest"))) {
    o[["time.interest"]] <- .surv.get.element(pred.o, "time.interest")
  }
  o[["forest"]] <- list(yvar = .surv.get.element(o, "yvar"),
                         xvar = .surv.get.element(o, "xvar"))
  o
}
## use imputed values stored on the current object, when available
.surv.apply.imputation <- function(o, yvar, xvar) {
  if (is.null(o$imputed.indv) || is.null(o$imputed.data) ||
      (is.null(yvar) && is.null(xvar))) {
    return(list(yvar = yvar, xvar = xvar))
  }
  n <- if (!is.null(yvar)) {
    nrow(yvar)
  }
  else {
    nrow(xvar)
  }
  if (is.null(n) || length(n) != 1L || n == 0L) {
    return(list(yvar = yvar, xvar = xvar))
  }
  idx <- o$imputed.indv
  if (is.logical(idx)) {
    if (length(idx) != n) {
      return(list(yvar = yvar, xvar = xvar))
    }
    idx <- which(idx)
  }
  else {
    idx <- suppressWarnings(as.integer(idx))
    idx <- idx[is.finite(idx) & idx >= 1L & idx <= n]
  }
  if (length(idx) == 0L) {
    return(list(yvar = yvar, xvar = xvar))
  }
  imp <- data.frame(o$imputed.data, check.names = FALSE)
  if (nrow(imp) == length(idx)) {
    imp.row <- seq_along(idx)
  }
  else if (nrow(imp) == n) {
    imp.row <- idx
  }
  else {
    return(list(yvar = yvar, xvar = xvar))
  }
  ## Survival imputation stores time and status in the first two columns.
  if (!is.null(yvar) && ncol(yvar) >= 2L && ncol(imp) >= 2L) {
    yvar[idx, 1:2] <- imp[imp.row, 1:2, drop = FALSE]
  }
  if (!is.null(xvar)) {
    nx <- ncol(xvar)
    if (nx > 0L && !is.null(yvar) && ncol(imp) >= 2L + nx) {
      x.col <- 2L + seq_len(nx)
    }
    else if (nx > 0L && ncol(imp) >= nx) {
      ## Accommodate prediction objects whose imputation matrix contains
      ## covariates only.
      x.col <- ncol(imp) - nx + seq_len(nx)
    }
    else {
      x.col <- integer(0L)
    }
    if (length(x.col) == nx && nx > 0L) {
      xvar[idx, ] <- imp[imp.row, x.col, drop = FALSE]
    }
  }
  list(yvar = yvar, xvar = xvar)
}
## coerce a survival prediction object to a case-by-time matrix
.surv.prediction.matrix <- function(x, ncase, ntime, name) {
  if (is.null(x)) {
    stop("survival predictions are unavailable in the supplied object.")
  }
  if (is.null(dim(x))) {
    if (ncase == 1L && length(x) == ntime) {
      x <- matrix(x, nrow = 1L)
    }
    else if (ntime == 1L && length(x) == ncase) {
      x <- matrix(x, ncol = 1L)
    }
    else {
      stop(name, " does not have dimensions compatible with the prediction object.")
    }
  }
  else {
    x <- as.matrix(x)
  }
  if (nrow(x) == ncase && ncol(x) == ntime) {
    return(x)
  }
  if (nrow(x) == ntime && ncol(x) == ncase) {
    return(t(x))
  }
  stop(name, " does not have dimensions compatible with the prediction object.")
}
## evaluate case-by-time step functions at requested times
.surv.step.matrix <- function(x, time.grid, times, initial = 1) {
  time.grid <- as.numeric(time.grid)
  times <- as.numeric(times)
  if (length(time.grid) == 0L || is.unsorted(time.grid, strictly = FALSE)) {
    stop("the survival time grid must be nondecreasing.")
  }
  idx <- findInterval(times, time.grid)
  out <- matrix(initial, nrow = nrow(x), ncol = length(times))
  use <- idx > 0L
  if (any(use)) {
    out[, use] <- x[, idx[use], drop = FALSE]
  }
  out
}
## evaluate each row of a case-by-time step function at its own time
.surv.step.row <- function(x, time.grid, times,
                           type = c("right", "left"), initial = 1) {
  type <- match.arg(type)
  times <- as.numeric(times)
  if (length(times) != nrow(x)) {
    stop("row-specific times do not match the number of predicted cases.")
  }
  out <- rep(NA_real_, length(times))
  ok <- is.finite(times)
  if (!any(ok)) {
    return(out)
  }
  idx <- if (type == "left") {
    findInterval(times[ok], time.grid, left.open = TRUE)
  }
  else {
    findInterval(times[ok], time.grid)
  }
  out[ok] <- initial
  use <- idx > 0L
  if (any(use)) {
    row <- which(ok)[use]
    out[row] <- x[cbind(row, idx[use])]
  }
  out
}
## create the event-information list used by the existing plotting code
.surv.event.info <- function(yvar, times) {
  if (is.null(yvar)) {
    return(list(event = NULL, event.type = NULL, cens = NULL,
                time.interest = times, time = NULL, r.dim = 0L))
  }
  if (ncol(yvar) != 2L) {
    stop("survival outcomes must contain exactly two columns: time and censoring.")
  }
  time <- as.numeric(yvar[, 1L])
  cens <- as.numeric(yvar[, 2L])
  valid.cens <- is.na(cens) |
    (is.finite(cens) & cens >= 0 & floor(cens) == cens)
  if (!all(valid.cens)) {
    stop("for survival families censoring variable must be coded as a non-negative integer")
  }
  event <- na.omit(cens)[na.omit(cens) > 0]
  event.type <- sort(unique(event))
  list(event = event, event.type = event.type, cens = cens,
       time.interest = times, time = time, r.dim = 2L)
}
## validate and normalize requested confidence level
.surv.conf.level <- function(conf.int) {
  if (length(conf.int) != 1L || is.na(conf.int)) {
    stop("'conf.int' must be FALSE, TRUE, or a single confidence level.")
  }
  if (is.logical(conf.int)) {
    return(if (conf.int) 0.95 else NULL)
  }
  conf.level <- suppressWarnings(as.numeric(conf.int))
  if (!is.finite(conf.level) || conf.level <= 0 || conf.level >= 1) {
    stop("a numeric 'conf.int' must be strictly between zero and one.")
  }
  conf.level
}
## extract grow and evaluation data without mixing their outcomes
.get.survival.performance.data <- function(o, subset = NULL, times = NULL) {
  if (is.null(o)) {
    return(NULL)
  }
  if (o$family != "surv") {
    stop("this function only supports right-censored survival settings")
  }
  object.type <- .surv.object.type(o, allow.forest = TRUE)
  if (is.null(object.type)) {
    stop("This function only works for objects of class `(rfsrc, grow)', `(rfsrc, forest)' or `(rfsrc, predict)'")
  }
  if (object.type == "forest") {
    o <- .surv.prepare.forest(o)
  }
  time.grid <- as.numeric(.surv.get.element(o, "time.interest"))
  if (length(time.grid) == 0L || any(!is.finite(time.grid)) ||
      is.unsorted(time.grid, strictly = FALSE)) {
    stop("the supplied object does not contain a valid survival time grid.")
  }
  if (is.null(times)) {
    times <- time.grid
  }
  else {
    times <- as.numeric(times)
    times <- sort(unique(times[is.finite(times)]))
    if (length(times) == 0L) {
      stop("'times' does not contain any finite values.")
    }
  }
  ## The full grow data always define the censoring distribution. For a
  ## prediction object, never substitute test outcomes when grow outcomes are
  ## unavailable: scoring must stop rather than estimate G from the test set.
  ## Exact extraction is essential for stripped rfsrc.fast objects, which keep
  ## xvar.names but intentionally omit xvar when forest = FALSE.
  forest <- .surv.get.element(o, "forest")
  forest.yvar <- .surv.get.element(forest, "yvar")
  forest.xvar <- .surv.get.element(forest, "xvar")
  object.yvar <- .surv.get.element(o, "yvar")
  object.xvar <- .surv.get.element(o, "xvar")
  if (object.type == "predict") {
    grow.yvar <- forest.yvar
    grow.xvar <- forest.xvar
  }
  else {
    grow.yvar <- if (!is.null(forest.yvar)) forest.yvar else object.yvar
    grow.xvar <- if (!is.null(forest.xvar)) forest.xvar else object.xvar
  }
  if (!is.null(grow.yvar) &&
      (is.null(dim(grow.yvar)) || ncol(grow.yvar) != 2L)) {
    stop("grow outcomes must contain exactly two columns: time and censoring.")
  }
  if (!is.null(grow.xvar) && is.null(dim(grow.xvar))) {
    stop("grow covariates must be stored in a matrix or data frame.")
  }
  ## Evaluation data and predictions depend on the object type. Exact
  ## extraction also prevents predicted from partially matching predicted.oob,
  ## or survival from partially matching survival.oob, in reduced objects.
  predicted <- .surv.get.element(o, "predicted")
  predicted.oob <- .surv.get.element(o, "predicted.oob")
  survival.full <- .surv.get.element(o, "survival")
  survival.oob <- .surv.get.element(o, "survival.oob")
  if (object.type == "predict") {
    eval.yvar <- object.yvar
    eval.xvar <- object.xvar
    mort <- as.numeric(predicted)
    survival <- survival.full
    prediction <- "test"
  }
  else {
    eval.yvar <- object.yvar
    eval.xvar <- object.xvar
    if (!is.null(predicted.oob) && !is.null(survival.oob)) {
      mort <- as.numeric(predicted.oob)
      survival <- survival.oob
      prediction <- "oob"
    }
    else {
      mort <- as.numeric(predicted)
      survival <- survival.full
      prediction <- "ensemble"
    }
  }
  if (length(mort) == 0L) {
    stop("mortality predictions are unavailable in the supplied object.")
  }
  ncase <- length(mort)
  survival <- .surv.prediction.matrix(survival, ncase, length(time.grid),
                                      "survival predictions")
  if (!is.null(eval.yvar)) {
    if (is.null(dim(eval.yvar)) || ncol(eval.yvar) != 2L) {
      stop("evaluation outcomes must contain exactly two columns: time and censoring.")
    }
    if (nrow(eval.yvar) != ncase) {
      stop("evaluation outcomes do not align with the predicted cases.")
    }
  }
  if (!is.null(eval.xvar)) {
    if (is.null(dim(eval.xvar))) {
      stop("evaluation covariates must be stored in a matrix or data frame.")
    }
    if (nrow(eval.xvar) != ncase) {
      stop("evaluation covariates do not align with the predicted cases.")
    }
  }
  ## Imputation information on a predict object applies to prediction data;
  ## on a grow/forest object it applies to both grow and evaluation data.
  eval.imp <- .surv.apply.imputation(o, eval.yvar, eval.xvar)
  eval.yvar <- eval.imp$yvar
  eval.xvar <- eval.imp$xvar
  if (object.type != "predict") {
    grow.imp <- .surv.apply.imputation(o, grow.yvar, grow.xvar)
    grow.yvar <- grow.imp$yvar
    grow.xvar <- grow.imp$xvar
  }
  if (is.null(subset)) {
    subset <- seq_len(ncase)
  }
  else {
    if (is.logical(subset)) {
      if (length(subset) != ncase) {
        stop("a logical 'subset' must have length equal to the number of predicted cases.")
      }
      subset <- which(subset)
    }
    else if (!is.numeric(subset)) {
      stop("'subset' must be a numeric or logical vector.")
    }
    subset <- as.numeric(subset)
    subset <- unique(subset[is.finite(subset) & subset == floor(subset) &
                              subset >= 1L & subset <= ncase])
    subset <- as.integer(subset)
    if (length(subset) == 0L) {
      stop("'subset' not set properly.")
    }
  }
  survival <- survival[subset, , drop = FALSE]
  survival <- .surv.step.matrix(survival, time.grid, times, initial = 1)
  mort <- mort[subset]
  if (!is.null(eval.yvar)) {
    eval.yvar <- eval.yvar[subset, , drop = FALSE]
  }
  if (!is.null(eval.xvar)) {
    eval.xvar <- eval.xvar[subset, , drop = FALSE]
  }
  list(object = o,
       object.type = object.type,
       prediction = prediction,
       subset = subset,
       time = times,
       prediction.time = time.grid,
       survival = survival,
       mort = mort,
       grow.yvar = grow.yvar,
       grow.xvar = grow.xvar,
       eval.yvar = eval.yvar,
       eval.xvar = eval.xvar,
       grow.event.info = .surv.event.info(grow.yvar, times),
       eval.event.info = .surv.event.info(eval.yvar, times))
}
## Nelson-Aalen survival estimator for the evaluation outcomes
.surv.aalen <- function(time, status, times) {
  ok <- is.finite(time) & is.finite(status)
  time <- as.numeric(time[ok])
  status <- as.numeric(status[ok])
  if (length(time) == 0L) {
    return(NULL)
  }
  event.time <- sort(unique(time[status != 0]))
  if (length(event.time) == 0L) {
    return(rep(1, length(times)))
  }
  increment <- vapply(event.time, function(tt) {
    y <- sum(time >= tt)
    d <- sum(time == tt & status != 0)
    if (y > 0) d / y else 0
  }, numeric(1L))
  surv <- exp(-cumsum(increment))
  idx <- findInterval(times, event.time)
  out <- rep(1, length(times))
  use <- idx > 0L
  out[use] <- surv[idx[use]]
  out
}
## fit the grow-data censoring distribution and evaluate it on eval cases
## G.time is a length-m vector for KM and an n-by-m matrix for an RF model.
.surv.censoring <- function(dat, cens.model) {
  cens.model <- match.arg(cens.model, c("km", "rfsrc"))
  grow.time <- as.numeric(dat$grow.yvar[, 1L])
  grow.status <- as.numeric(dat$grow.yvar[, 2L])
  eval.time <- as.numeric(dat$eval.yvar[, 1L])
  n <- nrow(dat$survival)
  m <- length(dat$time)
  grow.complete <- is.finite(grow.time) & is.finite(grow.status)
  if (!any(grow.complete)) {
    stop("no non-missing grow outcomes are available for estimating censoring.")
  }
  has.censoring <- any(grow.complete & grow.status == 0)
  if (!has.censoring) {
    g.minus <- rep(1, n)
    if (cens.model == "km") {
      g.time <- rep(1, m)
      cens.dist <- g.time
    }
    else {
      g.time <- matrix(1, nrow = n, ncol = m)
      cens.dist <- t(g.time)
    }
    return(list(G.time = g.time,
                G.minus = g.minus,
                cens.dist = cens.dist,
                fit = NULL))
  }
  if (cens.model == "km") {
    fit <- km_censor_fit(grow.time[grow.complete],
                         as.integer(grow.status[grow.complete] != 0))
    g.time <- .km.censor.predict(fit$time, fit$G, dat$time,
                                 type = "right")
    g.minus <- .km.censor.predict(fit$time, fit$G, eval.time,
                                  type = "left")
    return(list(G.time = g.time,
                G.minus = g.minus,
                cens.dist = g.time,
                fit = fit))
  }
  if (is.null(dat$grow.xvar) || is.null(dat$eval.xvar)) {
    stop("'cens.model = \"rfsrc\"' requires stored grow and evaluation ",
         "covariates. They are unavailable in the supplied object. For an ",
         "rfsrc.fast fit, use cens.model = \"km\" or refit with ",
         "forest = TRUE.")
  }
  if (nrow(dat$grow.xvar) != nrow(dat$grow.yvar)) {
    stop("grow covariates do not align with grow outcomes.")
  }
  ## Missing grow outcomes cannot train the censoring forest. Predictor
  ## missingness, if present, is handled by RF-SRC in the usual way.
  grow.use <- which(grow.complete)
  cens.dta <- data.frame(time = grow.time[grow.use],
                         cens = 1 * (grow.status[grow.use] == 0),
                         dat$grow.xvar[grow.use, , drop = FALSE])
  cens.o <- rfsrc(Surv(time, cens) ~ ., cens.dta,
                  ntree = 50,
                  nsplit = 1,
                  splitrule = "random",
                  nodesize = set.nodesize(nrow(cens.dta),
                                           ncol(dat$grow.xvar)),
                  perf.type = "none")
  cens.pred <- predict(cens.o, newdata = dat$eval.xvar)$survival
  cens.pred <- .surv.prediction.matrix(cens.pred, n,
                                       length(cens.o$time.interest),
                                       "censoring survival predictions")
  g.time <- .surv.step.matrix(cens.pred, cens.o$time.interest,
                              dat$time, initial = 1)
  g.minus <- .surv.step.row(cens.pred, cens.o$time.interest,
                            eval.time, type = "left", initial = 1)
  list(G.time = g.time,
       G.minus = g.minus,
       cens.dist = t(g.time),
       fit = cens.o)
}
## Brier score and empirical conditional asymptotic variance
.surv.brier <- function(dat, censoring, conf.level, papply, keep.matrix) {
  time <- as.numeric(dat$eval.yvar[, 1L])
  status <- as.numeric(dat$eval.yvar[, 2L])
  times <- dat$time
  survival <- dat$survival
  n <- nrow(survival)
  m <- ncol(survival)
  ans <- papply(seq_len(n), function(i) {
    loss <- rep(NA_real_, m)
    bad.g <- rep(FALSE, m)
    if (!is.finite(time[i]) || !is.finite(status[i])) {
      return(list(loss = loss, bad.g = bad.g))
    }
    case <- time[i] <= times & status[i] != 0
    control <- time[i] > times
    censored <- time[i] <= times & status[i] == 0
    ## Censored observations contribute zero after their censoring time.
    loss[censored] <- 0
    if (any(case)) {
      jj <- which(case)
      good.s <- is.finite(survival[i, jj])
      good.g <- is.finite(censoring$G.minus[i]) &&
        censoring$G.minus[i] > 0
      if (!good.g) {
        ## A zero censoring probability invalidates only horizons at which
        ## this case otherwise has an evaluable prediction.
        bad.g[jj[good.s]] <- TRUE
      }
      else {
        loss[jj[good.s]] <- survival[i, jj[good.s]]^2 /
          censoring$G.minus[i]
      }
    }
    if (any(control)) {
      jj <- which(control)
      g <- if (is.matrix(censoring$G.time)) {
        censoring$G.time[i, jj]
      }
      else {
        censoring$G.time[jj]
      }
      good.g <- is.finite(g) & g > 0
      good.s <- is.finite(survival[i, jj])
      bad.g[jj[good.s & !good.g]] <- TRUE
      use <- good.g & good.s
      loss[jj[use]] <- (1 - survival[i, jj[use]])^2 / g[use]
    }
    list(loss = loss, bad.g = bad.g)
  })
  brier.matx <- do.call(rbind, lapply(ans, function(z) z$loss))
  bad.g <- do.call(rbind, lapply(ans, function(z) z$bad.g))
  ## A horizon is unsupported when a required censoring probability is zero.
  unsupported <- colSums(bad.g, na.rm = TRUE) > 0L
  if (any(unsupported)) {
    brier.matx[, unsupported] <- NA_real_
  }
  point <- vapply(seq_len(m), function(j) {
    z <- brier.matx[, j]
    z <- z[is.finite(z)]
    if (length(z) > 0L) mean(z) else NA_real_
  }, numeric(1L))
  n.eval <- vapply(seq_len(m), function(j) {
    sum(is.finite(brier.matx[, j]))
  }, integer(1L))
  brier.score <- data.frame(time = times, brier.score = point)
  if (!is.null(conf.level)) {
    std.err <- vapply(seq_len(m), function(j) {
      z <- brier.matx[, j]
      z <- z[is.finite(z)]
      nz <- length(z)
      if (nz > 1L) {
        influence <- z - mean(z)
        sqrt(sum(influence^2) / (nz * (nz - 1L)))
      }
      else {
        NA_real_
      }
    }, numeric(1L))
    crit <- qnorm((1 + conf.level) / 2)
    brier.score$std.err <- std.err
    brier.score$lower <- point - crit * std.err
    brier.score$upper <- point + crit * std.err
    brier.score$n.eval <- n.eval
  }
  crps <- trapz(times, point)
  time.max <- max(times, na.rm = TRUE)
  crps.std <- if (is.finite(time.max) && time.max > 0) {
    crps / time.max
  }
  else {
    NA_real_
  }
  list(brier.matx = if (keep.matrix) brier.matx else NULL,
       brier.score = brier.score,
       crps = crps,
       crps.std = crps.std,
       n.eval = n.eval)
}
## weighted case/control placements for cumulative/dynamic AUC
.surv.auct.placements <- function(case.risk, control.risk,
                                   case.weight, control.weight) {
  if (length(case.risk) == 0L || length(control.risk) == 0L ||
      length(case.risk) != length(case.weight) ||
      length(control.risk) != length(control.weight)) {
    stop("case and control risks must have matching nonempty weight vectors.")
  }
  if (any(!is.finite(case.risk)) || any(!is.finite(control.risk)) ||
      any(!is.finite(case.weight) | case.weight <= 0) ||
      any(!is.finite(control.weight) | control.weight <= 0)) {
    stop("AUCT risks must be finite and AUCT weights must be finite and positive.")
  }
  ## Scale before normalization to avoid overflow when IPCW weights are
  ## highly concentrated. The common scale does not change the AUC.
  p <- case.weight / max(case.weight)
  p <- pmax(p, .Machine$double.xmin)
  p <- p / sum(p)
  q <- control.weight / max(control.weight)
  q <- pmax(q, .Machine$double.xmin)
  q <- q / sum(q)
  ## Control-weight distribution used for case placements.
  ord <- order(control.risk)
  risk <- control.risk[ord]
  weight <- q[ord]
  run <- rle(risk)
  run.end <- cumsum(run$lengths)
  run.start <- run.end - run$lengths + 1L
  control.unique <- run$values
  control.weight.group <- vapply(seq_along(run.start), function(k) {
    sum(weight[run.start[k]:run.end[k]])
  }, numeric(1L))
  control.cum <- cumsum(control.weight.group)
  idx <- findInterval(case.risk, control.unique)
  equal <- idx > 0L
  equal[equal] <- control.unique[idx[equal]] == case.risk[equal]
  lower.idx <- idx - as.integer(equal)
  lower <- rep(0, length(case.risk))
  use <- lower.idx > 0L
  lower[use] <- control.cum[lower.idx[use]]
  same <- rep(0, length(case.risk))
  same[equal] <- control.weight.group[idx[equal]]
  case.place <- lower + 0.5 * same
  ## Case-weight distribution used for control placements.
  ord <- order(case.risk)
  risk <- case.risk[ord]
  weight <- p[ord]
  run <- rle(risk)
  run.end <- cumsum(run$lengths)
  run.start <- run.end - run$lengths + 1L
  case.unique <- run$values
  case.weight.group <- vapply(seq_along(run.start), function(k) {
    sum(weight[run.start[k]:run.end[k]])
  }, numeric(1L))
  case.cum <- cumsum(case.weight.group)
  idx <- findInterval(control.risk, case.unique)
  equal <- idx > 0L
  equal[equal] <- case.unique[idx[equal]] == control.risk[equal]
  leq <- rep(0, length(control.risk))
  use <- idx > 0L
  leq[use] <- case.cum[idx[use]]
  same <- rep(0, length(control.risk))
  same[equal] <- case.weight.group[idx[equal]]
  control.place <- 1 - leq + 0.5 * same
  auct <- sum(p * case.place)
  list(auct = auct,
       case.place = case.place,
       control.place = control.place,
       case.weight = p,
       control.weight = q)
}
## stratified delete-one jackknife for weighted cumulative/dynamic AUC
.surv.auct.jackknife <- function(place) {
  p <- place$case.weight
  q <- place$control.weight
  n.case <- length(p)
  n.control <- length(q)
  n.case.eff <- if (n.case > 0L) 1 / sum(p^2) else NA_real_
  n.control.eff <- if (n.control > 0L) 1 / sum(q^2) else NA_real_
  max.case.weight <- if (n.case > 0L) max(p) else NA_real_
  max.control.weight <- if (n.control > 0L) max(q) else NA_real_
  ## Delete one member of a stratum and renormalize the remaining IPCW
  ## weights. Prefix and suffix sums avoid subtracting a dominant weighted
  ## placement from the full AUC, which can lose precision under heavy
  ## censoring. Each set of replicates is centered at its own replicate mean.
  jackknife.component <- function(weight, placement) {
    n <- length(weight)
    if (n <= 1L) {
      return(NA_real_)
    }
    weighted.placement <- weight * placement
    keep <- seq_len(n - 1L)
    weight.before <- c(0, cumsum(weight)[keep])
    weight.after <- rev(c(0, cumsum(rev(weight))[keep]))
    value.before <- c(0, cumsum(weighted.placement)[keep])
    value.after <- rev(c(0, cumsum(rev(weighted.placement))[keep]))
    remain.weight <- weight.before + weight.after
    remain.value <- value.before + value.after
    if (any(!is.finite(remain.weight)) ||
        any(remain.weight <= 0) ||
        any(!is.finite(remain.value))) {
      return(NA_real_)
    }
    delete.value <- remain.value / remain.weight
    if (any(!is.finite(delete.value))) {
      return(NA_real_)
    }
    delete.mean <- mean(delete.value)
    (n - 1) / n * sum((delete.value - delete.mean)^2)
  }
  var.case <- jackknife.component(p, place$case.place)
  var.control <- jackknife.component(q, place$control.place)
  std.err <- if (is.finite(var.case) && is.finite(var.control)) {
    sqrt(max(0, var.case + var.control))
  }
  else {
    NA_real_
  }
  list(std.err = std.err,
       var.case = var.case,
       var.control = var.control,
       n.case.eff = n.case.eff,
       n.control.eff = n.control.eff,
       max.case.weight = max.case.weight,
       max.control.weight = max.control.weight)
}
## cumulative/dynamic AUC and stratified delete-one jackknife variance
.surv.auct <- function(dat, censoring, conf.level, papply) {
  time <- as.numeric(dat$eval.yvar[, 1L])
  status <- as.numeric(dat$eval.yvar[, 2L])
  times <- dat$time
  risk <- 1 - dat$survival
  empty.result <- function(n.case, n.control) {
    c(auct = NA_real_, std.err = NA_real_,
      n.case = n.case, n.control = n.control,
      n.case.eff = NA_real_, n.control.eff = NA_real_,
      max.case.weight = NA_real_, max.control.weight = NA_real_)
  }
  ans <- papply(seq_along(times), function(j) {
    complete <- is.finite(time) & is.finite(status) & is.finite(risk[, j])
    case <- complete & time <= times[j] & status != 0
    control <- complete & time > times[j]
    n.case <- sum(case)
    n.control <- sum(control)
    if (n.case == 0L || n.control == 0L) {
      return(empty.result(n.case, n.control))
    }
    g.case <- censoring$G.minus[case]
    g.control <- if (is.matrix(censoring$G.time)) {
      censoring$G.time[control, j]
    }
    else {
      rep(censoring$G.time[j], n.control)
    }
    if (any(!is.finite(g.case) | g.case <= 0) ||
        any(!is.finite(g.control) | g.control <= 0)) {
      return(empty.result(n.case, n.control))
    }
    ## Only relative IPCW weights enter AUC. Scaling by the smallest
    ## censoring probability keeps the largest weight at one and avoids
    ## overflow in heavily censored settings.
    case.weight <- pmax(min(g.case) / g.case, .Machine$double.xmin)
    control.weight <- pmax(min(g.control) / g.control,
                           .Machine$double.xmin)
    if (any(!is.finite(case.weight) | case.weight <= 0) ||
        any(!is.finite(control.weight) | control.weight <= 0)) {
      return(empty.result(n.case, n.control))
    }
    place <- .surv.auct.placements(
      case.risk = risk[case, j],
      control.risk = risk[control, j],
      case.weight = case.weight,
      control.weight = control.weight
    )
    ## Weight diagnostics are useful even when confidence intervals are not
    ## requested. The jackknife calculation itself is only needed for an SE.
    p <- place$case.weight
    q <- place$control.weight
    n.case.eff <- 1 / sum(p^2)
    n.control.eff <- 1 / sum(q^2)
    max.case.weight <- max(p)
    max.control.weight <- max(q)
    std.err <- NA_real_
    if (!is.null(conf.level)) {
      jackknife <- .surv.auct.jackknife(place)
      std.err <- jackknife$std.err
    }
    c(auct = place$auct, std.err = std.err,
      n.case = n.case, n.control = n.control,
      n.case.eff = n.case.eff,
      n.control.eff = n.control.eff,
      max.case.weight = max.case.weight,
      max.control.weight = max.control.weight)
  })
  ans <- as.data.frame(do.call(rbind, ans))
  auct.score <- data.frame(time = times, auct = ans$auct)
  if (!is.null(conf.level)) {
    crit <- qnorm((1 + conf.level) / 2)
    auct.score$std.err <- ans$std.err
    auct.score$lower <- ans$auct - crit * ans$std.err
    auct.score$upper <- ans$auct + crit * ans$std.err
  }
  auct.score$n.case <- as.integer(ans$n.case)
  auct.score$n.control <- as.integer(ans$n.control)
  auct.score$n.case.eff <- ans$n.case.eff
  auct.score$n.control.eff <- ans$n.control.eff
  auct.score$max.case.weight <- ans$max.case.weight
  auct.score$max.control.weight <- ans$max.control.weight
  list(auct.score = auct.score)
}
## shared engine for Brier score and time-dependent AUC
.get.survival.performance <- function(o,
                                      subset = NULL,
                                      cens.model = c("km", "rfsrc"),
                                      papply = lapply,
                                      times = NULL,
                                      conf.int = FALSE,
                                      keep.matrix = TRUE,
                                      metrics = c("brier", "auct")) {
  cens.model <- match.arg(cens.model)
  metrics <- unique(match.arg(metrics, c("brier", "auct"),
                              several.ok = TRUE))
  if (!is.function(papply)) {
    stop("'papply' must be a function such as lapply or mclapply.")
  }
  if (length(keep.matrix) != 1L || is.na(keep.matrix) ||
      !is.logical(keep.matrix)) {
    stop("'keep.matrix' must be TRUE or FALSE.")
  }
  conf.level <- .surv.conf.level(conf.int)
  dat <- .get.survival.performance.data(o, subset = subset, times = times)
  if (is.null(dat)) {
    return(NULL)
  }
  if (is.null(dat$eval.yvar)) {
    stop("evaluation outcomes are required to calculate Brier score or time-dependent AUC.")
  }
  if (is.null(dat$grow.yvar)) {
    stop("grow outcomes are required to estimate the censoring distribution.")
  }
  eval.complete <- is.finite(as.numeric(dat$eval.yvar[, 1L])) &
    is.finite(as.numeric(dat$eval.yvar[, 2L]))
  if (!any(eval.complete)) {
    stop("no non-missing evaluation outcomes are available.")
  }
  censoring <- .surv.censoring(dat, cens.model)
  surv.aalen <- .surv.aalen(dat$eval.event.info$time,
                            dat$eval.event.info$cens,
                            dat$time)
  common <- list(cens.dist = censoring$cens.dist,
                 time = dat$time,
                 event.info = dat$grow.event.info,
                 grow.event.info = dat$grow.event.info,
                 eval.event.info = dat$eval.event.info,
                 test.event.info = dat$eval.event.info,
                 subset = dat$subset,
                 mort = dat$mort,
                 surv.aalen = surv.aalen,
                 surv.ensb = t(dat$survival),
                 prediction = dat$prediction,
                 cens.model = cens.model,
                 conf.level = conf.level)
  out <- list()
  if ("brier" %in% metrics) {
    brier <- .surv.brier(dat, censoring, conf.level, papply, keep.matrix)
    out$brier <- c(brier, common)
  }
  if ("auct" %in% metrics) {
    auct <- .surv.auct(dat, censoring, conf.level, papply)
    out$auct <- c(auct, common)
  }
  out
}
## Brier score helper
get.brier.survival <- function(o,
                               subset = NULL,
                               cens.model = c("km", "rfsrc"),
                               papply = lapply,
                               times = NULL,
                               conf.int = FALSE,
                               keep.matrix = TRUE) {
  ans <- .get.survival.performance(o,
                                   subset = subset,
                                   cens.model = cens.model,
                                   papply = papply,
                                   times = times,
                                   conf.int = conf.int,
                                   keep.matrix = keep.matrix,
                                   metrics = "brier")
  if (is.null(ans)) NULL else ans$brier
}
## cumulative/dynamic time-dependent AUC helper
get.auct.survival <- function(o,
                              subset = NULL,
                              cens.model = c("km", "rfsrc"),
                              papply = lapply,
                              times = NULL,
                              conf.int = FALSE) {
  ans <- .get.survival.performance(o,
                                   subset = subset,
                                   cens.model = cens.model,
                                   papply = papply,
                                   times = times,
                                   conf.int = conf.int,
                                   keep.matrix = FALSE,
                                   metrics = "auct")
  if (is.null(ans)) NULL else ans$auct
}
## ------------------------------------------------------------
## Uno weights
## - training mode: KM or OOB KM
## - test mode: works generically
## - censors happen after deaths at tied times
## ------------------------------------------------------------
## fit censoring KM on training outcomes only
km_censor_fit <- function(time, status) {
  stopifnot(length(time) == length(status))
  ok <- !is.na(time) & !is.na(status)
  time   <- as.numeric(time[ok])
  status <- as.integer(status[ok])
  n <- length(time)
  if (n == 0L) stop("No non-missing training outcomes.")
  ord <- order(time)
  t <- time[ord]
  s <- status[ord]
  times <- numeric(n)
  G     <- numeric(n)
  surv   <- 1.0
  n_risk <- n
  k <- 0L
  i <- 1L
  while (i <= n) {
    ti <- t[i]
    j <- i
    d_death <- 0L
    d_cens  <- 0L
    while (j <= n && t[j] == ti) {
      if (s[j] == 1L) d_death <- d_death + 1L else d_cens <- d_cens + 1L
      j <- j + 1L
    }
    ## Update censoring survival AFTER removing deaths at this time
    n_after_death <- n_risk - d_death
    if (d_cens > 0L) {
      if (n_after_death <= 0L) {
        surv <- 0.0
      } else {
        surv <- surv * (1 - d_cens / n_after_death)
      }
    }
    k <- k + 1L
    times[k] <- ti
    G[k]     <- surv
    n_risk <- n_risk - (d_death + d_cens)
    i <- j
  }
  list(time = times[1L:k], G = G[1L:k])
}
## Generic step evaluation for the censoring distribution.
## The stored values G are post-step values at time_knots.
.km.censor.predict <- function(time_knots, G, t_new,
                               type = c("left", "right")) {
  type <- match.arg(type)
  t_new <- as.numeric(t_new)
  out <- rep(NA_real_, length(t_new))
  ok <- !is.na(t_new)
  if (!any(ok)) return(out)
  if (length(time_knots) == 0L) {
    out[ok] <- 1.0
    return(out)
  }
  idx <- if (type == "left") {
    findInterval(t_new[ok], time_knots, left.open = TRUE)
  }
  else {
    findInterval(t_new[ok], time_knots)
  }
  out[ok] <- c(1.0, G)[1L + idx]
  out
}
## Generic left-limit step evaluation retained for the Uno helpers.
uno_Ghat_minus_predict <- function(time_knots, G, t_new) {
  .km.censor.predict(time_knots, G, t_new, type = "left")
}
## Effective sample size of positive weights
uno_ess <- function(w) {
  w <- w[is.finite(w) & !is.na(w) & (w > 0)]
  if (length(w) == 0L) return(NA_real_)
  (sum(w)^2) / sum(w^2)
}
## Choose gmin automatically from training event-time Ghat(t-).
##
## Input:  G_event = vector of Ghat(t-) evaluated at event times only.
## Output: list(gmin, ess_target, ess_kept, n_events, n_dropped, wmax_kept)
##
## Rule: drop the largest weights (smallest G) until ESS >= ess_target,
## where ess_target = max(ess_min, ceil(ess_frac * n_events)).
uno_choose_gmin_auto <- function(G_event,
                                 eps = 1e-12,
                                 ess_frac = 0.20,
                                 ess_min  = 20L) {
  g <- as.numeric(G_event)
  g <- g[is.finite(g) & !is.na(g)]
  d <- length(g)
  if (d <= 1L) {
    return(list(gmin = 0.0, ess_target = NA_real_, ess_kept = NA_real_,
                n_events = d, n_dropped = 0L, wmax_kept = NA_real_))
  }
  ## If essentially no censoring (G ~ 1), no trimming needed
  if (min(g) >= 1 - 1e-12) {
    return(list(gmin = 0.0, ess_target = d, ess_kept = d,
                n_events = d, n_dropped = 0L, wmax_kept = 1.0))
  }
  ## weights are monotone in g: smaller g => larger weight
  g_sorted <- sort(g)                       # ascending g
  w_desc   <- 1.0 / pmax(g_sorted, eps)^2   # descending weights
  ## prefix sums (with leading 0)
  p1 <- c(0.0, cumsum(w_desc))
  p2 <- c(0.0, cumsum(w_desc * w_desc))
  ess_target <- max(as.integer(ess_min), as.integer(ceiling(ess_frac * d)))
  ess_target <- min(ess_target, d)
  ## drop k largest weights; must keep at least ess_target events
  best_k <- 0L
  best_ess <- NA_real_
  for (k in 0L:(d - ess_target)) {
    sum_w  <- p1[d + 1L] - p1[k + 1L]
    sum_w2 <- p2[d + 1L] - p2[k + 1L]
    ess_k  <- if (sum_w2 > 0) (sum_w * sum_w) / sum_w2 else NA_real_
    if (is.finite(ess_k) && (ess_k >= ess_target)) {
      best_k <- k
      best_ess <- ess_k
      break
    }
  }
  ## If never hit the target (rare), keep only ess_target events
  if (!is.finite(best_ess)) {
    best_k <- d - ess_target
    sum_w  <- p1[d + 1L] - p1[best_k + 1L]
    sum_w2 <- p2[d + 1L] - p2[best_k + 1L]
    best_ess <- if (sum_w2 > 0) (sum_w * sum_w) / sum_w2 else NA_real_
  }
  gmin <- g_sorted[best_k + 1L]
  wmax <- 1.0 / pmax(gmin, eps)^2
  list(gmin = gmin,
       ess_target = ess_target,
       ess_kept = best_ess,
       n_events = d,
       n_dropped = best_k,
       wmax_kept = wmax)
}
## Train-mode Uno weights
get.uno.weights.train <- function(time, status,
                                  gmin = "auto",
                                  ess_frac = 0.20,
                                  ess_min  = 20L,
                                  eps = 1e-12,
                                  eps_keep = .Machine$double.eps,
                                  drop_if_G0 = FALSE,
                                  return_fit = TRUE) {
  stopifnot(length(time) == length(status))
  ## Fit KM censoring curve on training outcomes
  fit <- km_censor_fit(time, status)
  ## Global Ghat(t-) used for gating and (also) weight magnitude
  G_gate <- uno_Ghat_minus_predict(fit$time, fit$G, time)
  ## Decide gmin (train-once)
  if (is.character(gmin)) {
    gmin <- match.arg(gmin, c("auto", "none"))
    if (gmin == "none") {
      gmin_used <- 0.0
      ginfo <- list(gmin = 0.0)
    } else {
      ## events = non-censored cases (otherwise breaks for CR)
      ev <- !is.na(status) & (as.integer(status) != 0L) & !is.na(G_gate)
      ginfo <- uno_choose_gmin_auto(G_gate[ev], eps = eps,
                                    ess_frac = ess_frac, ess_min = ess_min)
      gmin_used <- ginfo$gmin
    }
  } else {
    gmin_used <- as.numeric(gmin)[1L]
    if (!is.finite(gmin_used) || gmin_used < 0) gmin_used <- 0.0
    ginfo <- list(gmin = gmin_used)
  }
  ## Missing => exclude (weight 0)
  miss <- is.na(time) | is.na(status)
  G_gate[miss] <- NA_real_
  w <- rep(0.0, length(time))
  ok <- !is.na(G_gate)
  if (!drop_if_G0) {
    ## keep-as-comparator always; event contribution trimmed by gmin
    keep <- ok & (G_gate >= gmin_used)
    drop <- ok & !keep
    if (any(keep)) {
      Gsafe <- pmax(G_gate[keep], eps)
      w[keep] <- 1.0 / (Gsafe * Gsafe)
    }
    if (any(drop)) {
      w[drop] <- eps_keep
    }
  } else {
    ## trim additionally when G is essentially 0
    keep <- ok & (G_gate >= gmin_used) & (G_gate > eps)
    drop <- ok & !keep
    if (any(keep)) {
      Gsafe <- pmax(G_gate[keep], eps)
      w[keep] <- 1.0 / (Gsafe * Gsafe)
    }
    if (any(drop)) {
      w[drop] <- eps_keep
    }
  }
  ## Store once; test will reuse automatically
  fit$gmin      <- gmin_used
  fit$eps_keep  <- eps_keep
  fit$gmin_info <- ginfo
  if (return_fit) {
    return(list(weight = w,
                Ghat_minus = G_gate,
                fit = fit))
  }
  w
}
## Test-mode Uno weights
get.uno.weights.test <- function(time_test, fit,
                                 eps = 1e-12,
                                 drop_if_G0 = FALSE) {
  if (is.null(fit$time) || is.null(fit$G))
    stop("fit must be a list with elements $time and $G")
  gmin_used <- if (!is.null(fit$gmin) && is.finite(fit$gmin)) fit$gmin else 0.0
  eps_keep  <- if (!is.null(fit$eps_keep) && is.finite(fit$eps_keep)) fit$eps_keep else .Machine$double.eps
  G_gate <- uno_Ghat_minus_predict(fit$time, fit$G, time_test)
  w <- rep(0.0, length(time_test))
  ok <- !is.na(G_gate)
  if (!drop_if_G0) {
    keep <- ok & (G_gate >= gmin_used)
  } else {
    keep <- ok & (G_gate >= gmin_used) & (G_gate > eps)
  }
  drop <- ok & !keep
  if (any(keep)) {
    Gsafe <- pmax(G_gate[keep], eps)
    w[keep] <- 1.0 / (Gsafe * Gsafe)
  }
  if (any(drop)) {
    w[drop] <- eps_keep
  }
  w
}
## ------------------------------------------------------------
## Convenience helper for test evaluation
## ------------------------------------------------------------
uno.prepare.test <- function(time_test, status_test, fit,
                             eps = 1e-12, drop_if_G0 = FALSE) {
  w <- get.uno.weights.test(time_test, fit, eps = eps, drop_if_G0 = drop_if_G0)
  list(time = time_test, status = status_test, weight = w)
}
## ------------------------------------------------------------
## Convenience one-liner: training weights only
## ------------------------------------------------------------
get.uno.weights <- function(time, status) {
  get.uno.weights.train(time, status, return_fit = FALSE)
}

distance <- function (x, 
                      method = "euclidean",
                      rowI = NULL,
                      rowJ = NULL,
                      do.trace = FALSE)
{
    n = dim(x)[1]
    p = dim(x)[2]
    if (n < 2) {
        stop("matrix must have more than one (1) row")
    }
    if (length(rowI) > 0 ) {
        if (length(rowI) != length(rowJ)) {
            stop("rowI and rowJ identifiers must have the same length")
        }
    }
    ## Legal methods.  Not all are implemented. They are placeholders.
    method.names <- c("euclidean", ## 1
                      "canberra",  ## 2
                      "maximum")   ## 3
    ## Default method is euclidean.
    if(is.null(method)) {
        method.idx <- which(method.names == "euclidean")
    }
    else {
        method.idx <- which(method.names == method)
    }
    ## Check for coherent distance method.
    if (length(method.idx) != 1) {
        stop("distance metric invalid")
    }
    ## Jump to native code.  Note that if rowI and rowJ are NULL, we return the distance matrix
    ## in it's entirity.  Only the lower-diagonal entries are returned.  If rowI and rowJ are non-NULL
    ## we return the distance for only the cell pairs (rowI[], rowJ[]).  This allows for master/slave hybrid
    ## OpenMP/MPI cluster processing.
    nativeOutput <- .Call("rfsrcDistance",
                          as.integer(method.idx),
                          as.integer(n),
                          as.integer(p),
                          as.double(x),
                          as.integer(length(rowI)),
                          as.integer(rowI),
                          as.integer(rowJ),
                          as.integer(get.rf.cores()),
                          as.integer(do.trace))
    ## check for error return condition in the native code
    if (is.null(nativeOutput)) {
        stop("An error has occurred in rfsrcDistance.  Please turn trace on for further analysis.")
    }
    if (length(rowI) > 0) {
        ## Return only the cell pairs (rowI[], rowJ[]) for processing in the master/slave scripts. 
        result <- list(rowI = rowI, rowJ = rowJ, distance = nativeOutput$distance)
    }
    else {
        ## Return the matrix in its entirety.
        result <- matrix(0, n, n)
        count <- 0
        for (k in 2:n) {
            result[k,1:(k-1)] <- nativeOutput$distance[(count + 1):(count + k - 1)]
            result[1:(k-1),k] <- result[k,1:(k-1)]
            count <- count + (k-1)
        }
    }
    return (result)
}

###################################################################
##
## extract confidence interval + other goodies from subsampled object
##
###################################################################
extract.subsample <- function(obj, alpha = .05, target = 0, m.target = NULL, standardize = TRUE, raw = FALSE)
{
  ## in case a user applies this to a bootsample object
  if (sum(c(grepl("rfsrc", class(obj))), grepl("subsample", class(obj))) == 2) {
    subsample <- TRUE
  }
  else if (sum(c(grepl("rfsrc", class(obj))), grepl("bootsample", class(obj))) == 2) {
    subsample <- FALSE
  }
  else {
    stop("this must be an object obtained by calling the 'subsample' function")
  }
  if (!subsample) {
    return(extract.bootsample(obj, alpha = alpha,
              target = target, m.target = m.target, standardize = standardize, raw = raw))    
  }
  ## coerce the (potentially) multivariate rf object
  m.target <- get.univariate.target(obj$rf, m.target)
  obj$rf <- coerce.multivariate(obj$rf, m.target)
  ## coerce the (potentially) multivariate original grow vimp
  if (is.null(m.target)) {
    vmp <- obj$vmp[[1]][, 1 + target, drop = FALSE]
  }
  else {
    vmp <- obj$vmp[[m.target]][, 1 + target, drop = FALSE]
  }
  ## coerce the (potentially) multivariate subsampled vimp 
  if (is.null(m.target)) {
    obj$vmpS <- lapply(obj$vmpS, data.frame)
  }
  else {
    obj$vmpS <- lapply(obj$vmpS, function(oo) {
      oo[[m.target]]
    })
  }  
  ## extract necessary objects
  theta.hat <- get.subsample.standardize.vimp(obj$rf, vmp, standardize)[, 1]
  theta.star <- get.subsample.standardize.vimp(obj$rf, rbind(sapply(obj$vmpS, "[[", 1 + target)), standardize)
  rownames(theta.star) <- xvar.names <- rownames(vmp)
  ## sample sizes
  n <- obj$rf$n
  m <- obj$subratio * n
  ## boxplot subsampling data
  scl <- sqrt(obj$subratio)
  theta.star.scl <- theta.hat - scl * rowMeans(theta.star, na.rm = TRUE)
  boxplot.dta <- data.frame(t(scl * theta.star + theta.star.scl))
  ## subsampled z-statistic 
  z.star <- sqrt(m) * (theta.star - theta.hat) 
  ## asymptotic standard error
  se.Z <- sqrt(m / n) * apply(theta.star, 1, sd, na.rm = TRUE)
  ## asymptotic JK standard error
  se.jk.Z <- sqrt((m / n) * rowMeans((theta.star - theta.hat)^2, na.rm = TRUE))
  ## nonparametric (default) subsampling confidence interval  
  ci <- do.call(cbind, lapply(1:nrow(z.star), function(rowj) {
    c(theta.hat[rowj] - quantile(z.star[rowj, ], 1 - alpha/2, na.rm = TRUE) / sqrt(n),
      theta.hat[rowj] - quantile(z.star[rowj, ], .75, na.rm = TRUE) / sqrt(n),
      theta.hat[rowj] - quantile(z.star[rowj, ], .50, na.rm = TRUE) / sqrt(n),
      theta.hat[rowj] - quantile(z.star[rowj, ], .25, na.rm = TRUE) / sqrt(n),
      theta.hat[rowj] - quantile(z.star[rowj, ], alpha/2, na.rm = TRUE) / sqrt(n))
  }))
  rownames(ci) <- paste(round(100 * c(alpha/2,.25,.50,.75,1-alpha/2), 1), "%", sep = "")
  colnames(ci) <- xvar.names
  ## normal-Z subsampling confidence interval
  ci.Z <- do.call(cbind, lapply(1:nrow(z.star), function(rowj) {
  c(theta.hat[rowj] - qnorm(1 - alpha/2) * se.Z[rowj],
    theta.hat[rowj] - qnorm(.75) * se.Z[rowj],
    theta.hat[rowj],
    theta.hat[rowj] + qnorm(.75) * se.Z[rowj],
    theta.hat[rowj] + qnorm(1 - alpha/2) * se.Z[rowj])
  }))
  rownames(ci.Z) <- paste(round(100 * c(alpha/2,.25,.50,.75,1-alpha/2), 1), "%", sep = "")
  colnames(ci.Z) <- xvar.names
  ## normal-Z subsampling confidence interval with JK
  ci.jk.Z <- do.call(cbind, lapply(1:nrow(z.star), function(rowj) {
  c(theta.hat[rowj] - qnorm(1 - alpha/2) * se.jk.Z[rowj],
    theta.hat[rowj] - qnorm(.75) * se.jk.Z[rowj],
    theta.hat[rowj],
    theta.hat[rowj] + qnorm(.75) * se.jk.Z[rowj],
    theta.hat[rowj] + qnorm(1 - alpha/2) * se.jk.Z[rowj])
  }))
  rownames(ci.jk.Z) <- paste(round(100 * c(alpha/2,.25,.50,.75,1-alpha/2), 1), "%", sep = "")
  colnames(ci.jk.Z) <- xvar.names
  ## nonparametric variable selection object
  var.sel <- data.frame(lower  =  ci[1, ],
                        median =  ci[3, ],
                        upper  =  ci[5, ],
                        pvalue = rowMeans((theta.star - sqrt(n / m) * theta.hat) > 0, na.rm = TRUE),
                        signif =  ci[1, ] > 0)
  rownames(var.sel) <- xvar.names
  ## normal-Z variable selection object
  var.sel.Z <- data.frame(lower  =  ci.Z[1, ],
                          mean   =  ci.Z[3, ],
                          upper  =  ci.Z[5, ],
                          pvalue = pnorm(theta.hat, 0, se.Z, FALSE),
                          signif =  ci.Z[1, ] > 0)
  rownames(var.sel.Z) <- xvar.names
  ## normal-Z variable selection object with JK
  var.jk.sel.Z <- data.frame(lower  =  ci.jk.Z[1, ],
                          mean   =  ci.jk.Z[3, ],
                          upper  =  ci.jk.Z[5, ],
                          pvalue = pnorm(theta.hat, 0, se.jk.Z, FALSE),
                          signif =  ci.jk.Z[1, ] > 0)
  rownames(var.jk.sel.Z) <- xvar.names
  if (raw) {
    list(boxplot.dta = boxplot.dta,
         ci = ci,
         ci.Z = ci.Z,
         ci.jk.Z = ci.jk.Z,
         vmp = theta.hat,
         vmpS = theta.star,
         se.Z = se.Z,
         se.jk.Z = se.jk.Z,
         var.sel = var.sel,
         var.sel.Z = var.sel.Z,
         var.jk.sel.Z = var.jk.sel.Z)
  }
  else {
    list(se.Z = se.Z,
         se.jk.Z = se.jk.Z,
         var.sel = var.sel,
         var.sel.Z = var.sel.Z,
         var.jk.sel.Z = var.jk.sel.Z)
  }
}
###################################################################
##
## set the subratio value
##
###################################################################
get.subsample.subratio <- function(n, base.n = 1000, base.r = exp(-1)) {
  if (n < base.n) {
    subratio <- base.r
  }
  else {
    subratio <- base.r * sqrt(base.n) / sqrt(n)
  }
}
###################################################################
##
## get joint vimp from a forest object
##
###################################################################
get.subsample.joint.vimp <- function(o, rf.prms = NULL, xvar.names) {
  class(o)[2] <- "grow"  ##fake the class to trick vimp()
  if (is.null(rf.prms)) {
    return(vimp(o, joint = TRUE, xvar.names = xvar.names))
  }
  else {
    return(vimp(o, joint = TRUE, xvar.names = xvar.names,
                perf.type = rf.prms$perf.type, block.size = rf.prms$block.size))
  }
}
###################################################################
##
## get standarized vimp from a matrix (p x B) of subsampled vimp
##
###################################################################
get.subsample.standardize.vimp <- function(obj, vmp, standardize = FALSE, sc = 1) {
  if (standardize && !is.null(obj)) {
    vmp <- do.call(cbind, lapply(1:ncol(vmp), function(j) {
      v <- vmp[, j]
      if (obj$family != "regr") {
        sc * v 
      }
      else {
        sc * v / var(obj$yvar, na.rm = TRUE)
      }
    }))
  }
  vmp
}
###################################################################
##
## double bootstrap sampling
##
###################################################################
make.double.boot.sample <- function(ntree, n, smp,
                                    size = function(x) { x },
                                    replace = TRUE)
{
  ntree <- as.integer(ntree)[1]
  n     <- as.integer(n)[1]
  m     <- as.integer(size(n))[1]
  if (is.na(ntree) || ntree < 1L) stop("ntree must be >= 1")
  if (is.na(n)     || n     < 1L) stop("n must be >= 1")
  if (is.na(m)     || m     < 0L) stop("size(n) must be >= 0")
  # Keep only rows that correspond to cases present in the 1st-stage bootstrap sample.
  idx <- sort(unique(smp))  # this is smp.unq, but we need it locally
  k   <- length(idx)
  # Map the multiset 'smp' (values in 1..n) onto 1..k so we can work on the reduced support.
  # This reduces memory/time when k << n.
  map <- match(smp, idx)  # length n, values in 1..k
  if (replace) {
    # Probabilities proportional to first-stage frequencies (sum(freq) = n)
    freq <- tabulate(map, nbins = k)
    prob <- freq / sum(freq)
    # k x ntree matrix of counts (rows correspond to idx in ascending order)
    dbl.smp <- rmultinom(ntree, size = m, prob = prob)
  } else {
    if (m > length(map)) stop("size(n) cannot exceed length(smp) when replace=FALSE")
    # Without replacement from the multiset: simulate by sampling positions, then tabulate.
    dbl.smp <- vapply(seq_len(ntree), function(bb) {
      tabulate(map[sample.int(length(map), size = m, replace = FALSE)], nbins = k)
    }, integer(k))
  }
  dbl.smp
}
###################################################################
##
## stratified sampling without replacement 
##
###################################################################
make.strat.sample <- function(y, subratio) {
  frq <- table(y)
  class.labels <- names(frq)
  as.numeric(unlist(sapply(class.labels, function(cl) {
    pt <- which(y == cl)
    sample(pt, size = length(pt) * subratio, replace = FALSE)    
  })))
}
###################################################################
##
## print function for subsampled ci
##
###################################################################
print.subsample.rfsrc <- function(x, alpha = .05, standardize = TRUE, ...) {
  vmp <- x$vmp
  nullO <- lapply(1:length(vmp), function(j) {
    m.target <- names(vmp)[j]
    p.vmp <- ncol(vmp[[j]])
    vmp.col.names <- colnames(vmp[[j]])
    if (length(vmp) > 1) {
      cat("processing a multivariate family, outcome:", m.target, "\n")
    }
    lapply(1:p.vmp, function(k) {
      oo <- extract.subsample(x, alpha = alpha, target = k - 1,
                  m.target = m.target, standardize = standardize, raw = TRUE)
      cat("===== VIMP/error confidence regions for", vmp.col.names[k], " =====\n")
      cat("nonparametric:\n")
      print(round(oo$ci, 3))
      cat("parametric:\n")
      print(round(oo$ci.Z, 3))
      cat("parametric (jackknife):\n")
      print(round(oo$ci.jk.Z, 3))
      NULL
    })
  })
}
print.subsample <- print.subsample.rfsrc
###################################################################
##
## combine two lists of vimp: o1 is master, o2 is appended
##
###################################################################
vimp.subsample.combine <- function(o1, o2, o2.name = NULL) {
  if (is.null(o1)) {
    rO <- lapply(1:length(o2), function(j) {
      rO.j <- rbind(NULL, o2[[j]])
      if (!is.null(o2.name)) {
        rownames(rO.j)[nrow(rO.j)] <- o2.name
      }
      rO.j
    })
    names(rO) <- names(o2)
  }
  else {  
    rO <- lapply(1:length(o1), function(j) {
      rO.j <- rbind(o1[[j]], o2[[j]])
      if (!is.null(o2.name)) {
        rownames(rO.j)[nrow(rO.j)] <- o2.name
      }
      rO.j
    })
    names(rO) <- names(o1)
  }
  rO
}
###################################################################
##
## custom progress bar
##
###################################################################
custom_progress_bar <- function(iteration, total, width = 50) {
  # Calculate the percentage of completion
  percent <- iteration / total
  # Calculate the number of characters to display
  progress <- round(width * percent)
  # ANSI color codes for green
  green_start <- "\033[32m"
  color_end <- "\033[0m"
  # Unicode block character (escaped)
  block_char <- "\u2588"
  # Create the progress bar string with color and blocks
  bar <- paste0("[", green_start, paste(rep(block_char, progress), collapse = ""), 
                color_end, paste(rep(" ", width - progress), collapse = ""), "]")
  # Create the percentage string
  percent_string <- sprintf(" %3d%%", round(100 * percent))
  # Print the progress bar to the console
  cat("\r", bar, percent_string, sep = "")
  # Flush the console output to ensure it is displayed immediately
  flush.console()
  # Clear the progress bar when complete
  if (iteration == total) {
    cat("\r", paste(rep(" ", width + 10), collapse = ""), "\r", sep = "")
  }
}
custom.progress.bar  <- custom_progress_bar

.integer.support.info <- function(x,
                                  tol = sqrt(.Machine$double.eps),
                                  frac.cut = 1) {
  if (inherits(x, c("Date", "POSIXt", "difftime")) || is.factor(x)) {
    return(list(
      restore = FALSE,
      storage.integer = FALSE,
      integer.frac = NA_real_
    ))
  }
  if (is.integer(x)) {
    return(list(
      restore = TRUE,
      storage.integer = TRUE,
      integer.frac = 1
    ))
  }
  if (!is.numeric(x)) {
    return(list(
      restore = FALSE,
      storage.integer = FALSE,
      integer.frac = NA_real_
    ))
  }
  z <- as.numeric(x)
  z <- z[!is.na(z) & is.finite(z)]
  if (length(z) == 0L) {
    return(list(
      restore = FALSE,
      storage.integer = FALSE,
      integer.frac = NA_real_
    ))
  }
  scale <- pmax(1, abs(z))
  frac <- mean(abs(z - round(z)) <= tol * scale)
  list(
    restore = isTRUE(frac >= frac.cut),
    storage.integer = FALSE,
    integer.frac = frac
  )
}
.integer.support.map <- function(data,
                                 tol = sqrt(.Machine$double.eps),
                                 frac.cut = 1) {
  out <- lapply(data, .integer.support.info, tol = tol, frac.cut = frac.cut)
  names(out) <- names(data)
  out
}
.restore.integer.vector <- function(x, info, restore.integer = TRUE) {
  if (!isTRUE(restore.integer) || !isTRUE(info$restore)) {
    return(x)
  }
  z <- as.numeric(x)
  ok <- !is.na(z) & is.finite(z)
  z[ok] <- round(z[ok])
  if (isTRUE(info$storage.integer)) {
    out <- rep(NA_integer_, length(z))
    out[ok] <- as.integer(z[ok])
    return(out)
  }
  z
}
.restore.integer.data <- function(data, integer.support, restore.integer = TRUE) {
  if (!isTRUE(restore.integer) || is.null(integer.support) || length(integer.support) == 0L) {
    return(data)
  }
  for (nm in intersect(names(integer.support), names(data))) {
    info <- integer.support[[nm]]
    if (isTRUE(info$restore)) {
      data[[nm]] <- .restore.integer.vector(data[[nm]], info, restore.integer = TRUE)
    }
  }
  data
}
assign.impute.mean <- function(data, impute.mean) {
  cn <- colnames(data)
  p  <- length(cn)
  ## Build columns in a preallocated list (faster than lapply over names)
  out <- vector("list", p)
  names(out) <- cn
  for (j in seq_len(p)) {
    nm <- cn[j]
    x  <- data[[nm]]
    na_idx <- is.na(x)
    if (any(na_idx)) {
      x[na_idx] <- impute.mean[[nm]]
    }
    out[[j]] <- x
  }
  ## Preserve original behavior: coerce character columns to factor
  ## (original code used stringsAsFactors = TRUE)
  data.frame(out, stringsAsFactors = TRUE, check.names = FALSE)
}
get.impute.mean <- function(data) {
  cn <- colnames(data)
  p  <- length(data)
  ## Preallocate list output for speed
  imean <- vector("list", p)
  names(imean) <- cn
  if (p == 0L) return(imean)
  is_fac <- vapply(data, is.factor, logical(1))
  ## Fast path for numeric/integer/logical columns: use a matrix and colMeans
  is_numlike <- vapply(data, function(x) {
    (is.numeric(x) || is.integer(x) || is.logical(x)) && !is.factor(x)
  }, logical(1))
  num_idx <- which(is_numlike)
  fac_idx <- which(is_fac)
  if (length(num_idx)) {
    xm <- data.matrix(data[, num_idx, drop = FALSE])
    ##storage.mode(xm) <- "double" : 
    cnt <- colSums(!is.na(xm))
    mu  <- colMeans(xm, na.rm = TRUE)
    mu[cnt == 0] <- NA_real_
    for (k in seq_along(num_idx)) {
      imean[[num_idx[k]]] <- mu[k]
    }
  }
  ## Factor columns: compute the modal level (fast via tabulate)
  if (length(fac_idx)) {
    for (k in fac_idx) {
      x <- data[[k]]
      if (all(is.na(x))) {
        imean[[k]] <- NA
      } else {
        tab <- tabulate(x, nbins = length(levels(x)))
        imean[[k]] <- levels(x)[which.max(tab)]
      }
    }
  }
  ## Any remaining columns (e.g., character): match original logic
  ## (mean() -> NA with warning). We suppress warnings to avoid slowdown/noise.
  other_idx <- setdiff(seq_len(p), c(num_idx, fac_idx))
  if (length(other_idx)) {
    for (k in other_idx) {
      x <- data[[k]]
      if (all(is.na(x))) {
        imean[[k]] <- NA
      } else {
        imean[[k]] <- suppressWarnings(mean(x, na.rm = TRUE))
      }
    }
  }
  imean
}
get.na.roughfix <- function(data) {
  if (!is.data.frame(data)) data <- data.frame(data)
  assign.impute.mean(data, get.impute.mean(data))
}

rfsrc.cart <- function(formula, data, ntree = 1, mtry = ncol(data), bootstrap = "none", nsplit = 0, ...)
{
  rfsrc(formula, data, ntree = ntree, mtry = mtry, bootstrap = bootstrap, nsplit = 0, ...)
}

\name{holdout.vimp.rfsrc}
\alias{holdout.vimp.rfsrc}
\alias{holdout.vimp}
\title{Hold out variable importance (VIMP)}

\description{
  Hold out VIMP is calculated from the error rate of mini ensembles of
  trees (blocks of trees) grown with and without a variable.  Applies to
  all families.
}

\usage{\method{holdout.vimp}{rfsrc}(formula, data,
  ntree = function(p, vtry){1000 * p / vtry},
  nsplit = 10,
  ntime = 50,
  sampsize = function(x){x * .632},
  samptype = "swor",
  block.size = 10,
  vtry = 1,
  ...)
}

\arguments{
  \item{formula}{A symbolic description of the model to be fit.}

  \item{data}{Data frame containing the y-outcome and x-variables.}

  \item{ntree}{Specifies the number of trees used to grow the forest. Can be a function of 
  data dimension and number of holdout variables, or a fixed numeric value.}

  \item{nsplit}{Non-negative integer specifying the number of random split points used to split a node. 
  A value of zero corresponds to deterministic splitting, which is significantly slower.}

  \item{ntime}{Integer value used for survival settings to constrain ensemble 
  calculations to a grid of \code{ntime} time points.}

  \item{sampsize}{Specifies the size of the subsampled data. Can be either a function or a numeric value.}

  \item{samptype}{Type of bootstrap used when subsampling.}

  \item{vtry}{Number of variables randomly selected to be held out when
  growing a tree.  Can also be a list for targeted holdout variable
  importance analysis. See \code{details} for more information.}

  \item{block.size}{Specifies the number of trees in a block when calculating holdout variable importance.}

  \item{...}{Further arguments passed to \code{\link{rfsrc}}.}
}

\details{

  Holdout variable importance (holdout VIMP) measures the importance of a variable 
  by comparing prediction error between two forests (blocks of trees): 
  one in which selected variables are held out during tree growing (the \emph{holdout forest}) 
  and one in which no variables are held out (the \emph{baseline forest}).

  For each variable-block combination, the bootstrap samples used to grow the trees 
  are the same in both forests. The difference in out-of-bag (OOB) prediction error 
  between the holdout and baseline forests gives the holdout VIMP for that variable-block pair. 
  The final holdout VIMP for a variable is the average of these differences over all blocks 
  in which the variable was held out.

  The option \code{vtry} controls how many variables are held out per tree. 
  The default is one, meaning a single variable is held out per tree. 
  Larger values of \code{vtry} increase the number of times each variable is held out, 
  reducing the required total number of trees. However, interpretation of holdout VIMP 
  changes when \code{vtry} exceeds one, and this option should be used cautiously.

  High accuracy requires a sufficiently large number of trees. 
  As a general guideline, we recommend using \code{ntree = 1000 * p / vtry}, 
  where \code{p} is the number of features. Accuracy also depends on \code{block.size}, 
  which determines how many trees comprise a block. Smaller values yield better accuracy 
  but are computationally more demanding. The most accurate setting is \code{block.size = 1}. 
  Ensure that \code{block.size} does not exceed \code{ntree / p}, 
  otherwise insufficient trees may be available for certain variables.

  Targeted holdout VIMP analysis can be requested by specifying \code{vtry} as a list 
  with two components: a vector of variable indices (\code{xvar}) and a logical flag 
  \code{joint} indicating whether to compute joint VIMP. For example, to compute holdout VIMP 
  only for variables 1, 4, and 5 individually:

  \code{vtry = list(xvar = c(1, 4, 5), joint = FALSE)}

  To compute the joint effect of removing these three variables together:

  \code{vtry = list(xvar = c(1, 4, 5), joint = TRUE)}

  Targeted analysis is useful when the user has prior knowledge of variables of interest 
  and can significantly reduce computation. Joint VIMP quantifies the combined importance 
  of specific groups of variables. See the Iris example below for illustration.

}

\value{
  Invisibly a list with the following components (which themselves can be lists):
  \item{importance}{Holdout VIMP.}
  \item{baseline}{Prediction error for the baseline forest.}
  \item{holdout}{Prediction error for the holdout forest.}
}

\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{

  Lu M. and Ishwaran H. (2018). Expert Opinion: A prediction-based
  alternative to p-values in regression models. \emph{J. Thoracic and
    Cardiovascular Surgery}, 155(3), 1130--1136.

  
}

\seealso{
  \command{\link{vimp.rfsrc}}
}
\examples{
\donttest{


## ------------------------------------------------------------
## regression analysis
## ------------------------------------------------------------

## new York air quality measurements
airq.obj <- holdout.vimp(Ozone ~ ., data = airquality, na.action = "na.impute")
print(airq.obj$importance)

## ------------------------------------------------------------
## classification analysis
## ------------------------------------------------------------

## iris data
iris.obj <- holdout.vimp(Species ~., data = iris)
print(iris.obj$importance)

## iris data using brier prediction error
iris.obj <- holdout.vimp(Species ~., data = iris, perf.type = "brier")
print(iris.obj$importance)

## ------------------------------------------------------------
## illustration of targeted holdout vimp analysis
## ------------------------------------------------------------

## iris data - only interested in variables 3 and 4
vtry <- list(xvar = c(3, 4), joint = FALSE)
print(holdout.vimp(Species ~., data = iris, vtry = vtry)$impor)

## iris data - joint importance of variables 3 and 4
vtry <- list(xvar = c(3, 4), joint = TRUE)
print(holdout.vimp(Species ~., data = iris, vtry = vtry)$impor)

## iris data - joint importance of variables 1 and 2
vtry <- list(xvar = c(1, 2), joint = TRUE)
print(holdout.vimp(Species ~., data = iris, vtry = vtry)$impor)


## ------------------------------------------------------------
## imbalanced classification (using RFQ)
## ------------------------------------------------------------

if (library("caret", logical.return = TRUE)) {

  ## experimental settings
  n <- 400
  q <- 20
  ir <- 6
  f <- as.formula(Class ~ .)
 
  ## simulate the data, create minority class data
  d <- twoClassSim(n, linearVars = 15, noiseVars = q)
  d$Class <- factor(as.numeric(d$Class) - 1)
  idx.0 <- which(d$Class == 0)
  idx.1 <- sample(which(d$Class == 1), sum(d$Class == 1) / ir , replace = FALSE)
  d <- d[c(idx.0,idx.1),, drop = FALSE]

  ## VIMP for RFQ with and without blocking
  vmp1 <- imbalanced(f, d, importance = TRUE, block.size = 1)$importance[, 1]
  vmp10 <- imbalanced(f, d, importance = TRUE, block.size = 10)$importance[, 1]

  ## holdout VIMP for RFQ with and without blocking
  hvmp1 <- holdout.vimp(f, d, rfq =  TRUE,
               perf.type = "g.mean", block.size = 1)$importance[, 1]
  hvmp10 <- holdout.vimp(f, d, rfq =  TRUE,
               perf.type = "g.mean", block.size = 10)$importance[, 1]
  
  ## compare VIMP values
  imp <- 100 * cbind(vmp1, vmp10, hvmp1, hvmp10)
  legn <- c("vimp-1", "vimp-10","hvimp-1", "hvimp-10")
  colr <- rep(4,20+q)
  colr[1:20] <- 2
  ylim <- range(c(imp))
  nms <- 1:(20+q)
  par(mfrow=c(2,2))
  barplot(imp[,1],col=colr,las=2,main=legn[1],ylim=ylim,names.arg=nms)
  barplot(imp[,2],col=colr,las=2,main=legn[2],ylim=ylim,names.arg=nms)
  barplot(imp[,3],col=colr,las=2,main=legn[3],ylim=ylim,names.arg=nms)
  barplot(imp[,4],col=colr,las=2,main=legn[4],ylim=ylim,names.arg=nms)

}

## ------------------------------------------------------------
## multivariate regression analysis
## ------------------------------------------------------------
mtcars.mreg <- holdout.vimp(Multivar(mpg, cyl) ~., data = mtcars,
                                    vtry = 3,
                                    block.size = 1,
                                    samptype = "swr",
                                    sampsize = dim(mtcars)[1])
print(mtcars.mreg$importance)

## ------------------------------------------------------------
## mixed outcomes analysis
## ------------------------------------------------------------

mtcars.new <- mtcars
mtcars.new$cyl <- factor(mtcars.new$cyl)
mtcars.new$carb <- factor(mtcars.new$carb, ordered = TRUE)
mtcars.mix <- holdout.vimp(cbind(carb, mpg, cyl) ~., data = mtcars.new,
                                   ntree = 100,
                                   block.size = 2,
                                   vtry = 1)
print(mtcars.mix$importance)

##------------------------------------------------------------
## survival analysis
##------------------------------------------------------------

## Primary biliary cirrhosis (PBC) of the liver
data(pbc, package = "randomForestSRC")
pbc.obj <- holdout.vimp(Surv(days, status) ~ ., pbc,
                                nsplit = 10,
                                ntree = 1000,
                                na.action = "na.impute")
print(pbc.obj$importance)

##------------------------------------------------------------
## competing risks
##------------------------------------------------------------

## WIHS analysis
## cumulative incidence function (CIF) for HAART and AIDS stratified by IDU

data(wihs, package = "randomForestSRC")
wihs.obj <- holdout.vimp(Surv(time, status) ~ ., wihs,
                                 nsplit = 3,
                                 ntree = 100)
print(wihs.obj$importance)

}}

\keyword{vimp}

\name{print.rfsrc}
\alias{print.rfsrc}
\title{Print Summary Output of a RF-SRC Analysis}
\description{
  Print summary output from a RF-SRC analysis. This is the default
  print method for the package.
}
\usage{\method{print}{rfsrc}(x, outcome.target = NULL, ...)}
\arguments{
  \item{x}{An object of class \code{(rfsrc, grow)}, 
    or \code{(rfsrc, predict)}.}
  \item{outcome.target}{Character value for multivariate families
    specifying the target outcome to be used. The default is to use the
    first coordinate from the continuous outcomes (otherwise if none,
    the first coordinate from the categorical outcomes).}
  \item{...}{Further arguments passed to or from other methods.}
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Ishwaran H. and Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7/2:25-31.
}
\examples{
  iris.obj <- rfsrc(Species ~., data = iris, ntree=10)
  print(iris.obj)
}
\keyword{print}

\name{wihs}
\docType{data}
\alias{wihs}
\title{Women's Interagency HIV Study (WIHS)}
\description{
  Competing risk data set involving AIDS in women.
}
\format{A data frame containing:
  \tabular{ll}{
    time \tab time to event\cr
    status \tab censoring status: 0=censoring, 1=HAART initiation, 2=AIDS/Death before HAART\cr
    ageatfda \tab age in years at time of FDA approval of first protease inhibitor\cr
    idu \tab history of IDU: 0=no history, 1=history\cr
    black \tab race:  0=not African-American; 1=African-American\cr
    cd4nadir \tab CD4 count (per 100 cells/ul)
  }
}
\source{
  Study included 1164 women enrolled in WIHS, who were alive, infected
  with HIV, and free of clinical AIDS on December, 1995, when the first
  protease inhibitor (saquinavir mesylate) was approved by the Federal
  Drug Administration. Women were followed until the first of the
  following occurred: treatment initiation, AIDS diagnosis, death, or
  administrative censoring (September, 2006). Variables included history
  of injection drug use at WIHS enrollment, whether an individual was
  African American, age, and CD4 nadir prior to baseline.}
\references{
  Bacon M.C, von Wyl V., Alden C., et al. (2005). The Women's Interagency
  HIV Study: an observational cohort brings clinical sciences to the
  bench, \emph{Clin Diagn Lab Immunol}, 12(9):1013-1019.
}
\examples{
\donttest{
data(wihs, package = "randomForestSRC")
wihs.obj <- rfsrc(Surv(time, status) ~ ., wihs, nsplit = 3, ntree = 100)
}
}
\keyword{datasets}

\name{plot.rfsrc}
\alias{plot.rfsrc}
\title{Plot Error Rate and Variable Importance from a RF-SRC analysis}
\description{
  Plot out-of-bag (OOB) error rates and variable importance (VIMP)
  from a RF-SRC analysis. This is the default plot method for the package.
}
\usage{\method{plot}{rfsrc}(x, m.target = NULL,
  plots.one.page = TRUE, sorted = TRUE, verbose = TRUE,  ...)}
\arguments{
  \item{x}{An object of class \code{(rfsrc, grow)}, 
    or \code{(rfsrc, predict)}.}
  \item{m.target}{Character value for multivariate families
    specifying the target outcome to be used.  If left unspecified, the
    algorithm will choose a default target.}
  \item{plots.one.page}{Should plots be placed on one page?}
  \item{sorted}{Should variables be sorted by importance values?}
  \item{verbose}{Should VIMP be printed?}
  \item{...}{Further arguments passed to or from other methods.}
}
\details{  
  Plot cumulative OOB error rates as a function of number of trees and
  variable importance (VIMP) if available.  Note that the default
  settings are now such that the error rate is no longer calculated on
  every tree and VIMP is only calculated if requested.  To get OOB error
  rates for ever tree, use the option \code{block.size = 1} when
  growing or restoring the forest.  Likewise, to view VIMP, use the option
  \code{importance} when growing or restoring the forest.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Breiman L. (2001). Random forests, \emph{Machine Learning}, 45:5-32.
  
  Ishwaran H. and Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
}
\examples{
\donttest{
## ------------------------------------------------------------
## classification example
## ------------------------------------------------------------

iris.obj <- rfsrc(Species ~ ., data = iris,
     block.size = 1, importance = TRUE)
plot(iris.obj)

## ------------------------------------------------------------
## competing risk example
## ------------------------------------------------------------

## use the pbc data from the survival package
## events are transplant (1) and death (2)
if (library("survival", logical.return = TRUE)) {
  data(pbc, package = "survival")
  pbc$id <- NULL
  plot(rfsrc(Surv(time, status) ~ ., pbc, block.size = 1))
}

## ------------------------------------------------------------
## multivariate mixed forests
## ------------------------------------------------------------

mtcars.new <- mtcars
mtcars.new$cyl <- factor(mtcars.new$cyl)
mtcars.new$carb <- factor(mtcars.new$carb, ordered = TRUE)
mv.obj <- rfsrc(cbind(carb, mpg, cyl) ~., data = mtcars.new, block.size = 1)
plot(mv.obj, m.target = "carb")
plot(mv.obj, m.target = "mpg")
plot(mv.obj, m.target = "cyl")

}
}
\keyword{plot}

\name{partial.rfsrc}
\alias{partial.rfsrc}
\alias{partial}
\alias{get.partial.plot.data}
\title{Acquire Partial Effect of a Variable}
\description{
  Direct, fast inferface for partial effect of a variable.  Works for all families.
}
\usage{
partial.rfsrc(object, oob = TRUE, 
  partial.type = NULL, partial.xvar = NULL, partial.values = NULL,
  partial.xvar2 = NULL, partial.values2 = NULL,
  partial.time = NULL, get.tree = NULL, seed = NULL, do.trace = FALSE, ...)
}
\arguments{
  \item{object}{An object of class \code{(rfsrc, grow)}.}
  \item{oob}{By default out-of-bag values are returned, but inbag
    values can be requested by setting this option to \code{FALSE}.}
  \item{partial.type}{Character vector specifying type of predicted
    value requested.  See details below.}
  \item{partial.xvar}{Character value specifying the single primary
    partial x-variable to be used.}
  \item{partial.values}{Vector of values that the primary partialy
    x-variable will assume.}
  \item{partial.xvar2}{Vector of character values specifying the second order x-variables to be
    used.}
  \item{partial.values2}{Vector of values that the second order
    x-variables will assume.  Each second order x-variable can only
    assume a single value.  This the length of \code{partial.xvar2} and
    \code{partial.values2} will be the same.  In addition, the user must
    do the appropriate conversion for factors, and represent a value as
    a numeric element.}
  \item{partial.time}{For survival families, the time at which the predicted
    survival value is evaluated at (depends on \code{partial.type}).}
  \item{get.tree}{Vector of integer(s) identifying trees over which the
    partial values are calculated over.  By default, uses all trees in the
    forest.}
  \item{seed}{Negative integer specifying seed for the random number
    generator.}
  \item{do.trace}{Number of seconds between updates to the user on
    approximate time to completion.}
  \item{...}{Further arguments passed to or from other methods.}
}
\details{

  Used for direct, efficient call to obtain partial plot effects.
  This function is intended primarily for experts. 
  
  Out-of-bag (OOB) values are returned by default.

  For factors, the partial value should be encoded as a positive integer
  reflecting the level number of the factor.  The actual label of the
  factor should not be used.

  The utility function \code{get.partial.plot.data} is supplied for
  processing returned raw partial effects in a format more convenient
  for plotting.  Options are specified as in \code{plot.variable}.  See
  examples for illustration.
  
  Raw partial plot effects data is returned either as an array or a list
  of length equal to the number of outcomes (length is one for
  univariate families) with entries depending on the underlying family:
  
  \enumerate{
    \item For regression, partial plot data is returned as a list
    in \code{regrOutput} with dim \code{[n] x  [length(partial.values)]}.
    
    \item For classification, partial plot data is returned as a list
    in \code{classOutput} of dim \code{[n] x [1 + yvar.nlevels[.]] x
      [length(partial.values)]}.
    
    \item For mixed multivariate regression, values are returned in 
    list format both in \code{regrOutput} and \code{classOutput}
    
    \item For survival, values are returned as either a matrix or array
    in \code{survOutput}. Depending on partial type specified this can be:
    \itemize{
      \item For partial type \code{surv} returns the survival function
         of dim \code{[n] x [length(partial.time)] x [length(partial.values)]}.
      \item For partial type \code{mort} returns mortality
        of dim \code{[n] x [length(partial.values)]}.
      \item For partial type \code{chf} returns the cumulative hazard function
        of dim \code{[n] x [length(partial.time)] x [length(partial.values)]}.
    }
    
    \item For competing risks, values are returned as either a matrix or
    array in \code{survOutput}.  Depending on the options specified this
    can be:
    \itemize{
      \item For partial type \code{years.lost} returns the expected number of life years lost 
      of dim \code{[n] x [length(event.info$event.type)] x
        [length(partial.values)]}.
      \item For partial type \code{cif} returns the cumulative incidence function of dim
	    \code{[n] x [length(partial.time)] x
	      [length(event.info$event.type)] x
	      [length(partial.values)]}.
      \item For partial type \code{chf} returns the cumulative hazard function of dim
            \code{[n] x [length(partial.time)] x [length(event.info$event.type)]
            x [length(partial.values)]}.
    }

  }
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Ishwaran H., Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
  
  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S.
  (2008).  Random survival forests, \emph{Ann. App.
    Statist.}, 2:841-860.
}
\seealso{
  \command{\link{plot.variable.rfsrc}}
}
\examples{
\donttest{

## ------------------------------------------------------------
##
## regression
##
## ------------------------------------------------------------

airq.obj <- rfsrc(Ozone ~ ., data = airquality)

## partial effect for wind
partial.obj <- partial(airq.obj,
                  partial.xvar = "Wind",
                  partial.values = airq.obj$xvar$Wind)
pdta <- get.partial.plot.data(partial.obj)

## plot partial values
plot(pdta$x, pdta$yhat, type = "b", pch = 16,
      xlab = "wind", ylab = "partial effect of wind")


## example where we display all the partial effects
## instead of averaging - use the granule=TRUE option
pdta <- get.partial.plot.data(partial.obj, granule = TRUE)
boxplot(pdta$yhat ~ pdta$x, xlab = "Wind", ylab = "partial effect")

## ------------------------------------------------------------
##
## regression: partial effects for two variables simultaneously
##
## ------------------------------------------------------------

airq.obj <- rfsrc(Ozone ~ ., data = airquality)

## specify wind and temperature values of interest
wind <- sort(unique(airq.obj$xvar$Wind))
temp <- sort(unique(airq.obj$xvar$Temp))

## partial effect for wind, for a given temp
pdta <- do.call(rbind, lapply(temp, function(x2) {
  o <- partial(airq.obj,
         partial.xvar = "Wind", partial.xvar2 = "Temp",
         partial.values = wind, partial.values2 = x2)
  cbind(wind, x2, get.partial.plot.data(o)$yhat)
}))
pdta <- data.frame(pdta)
colnames(pdta) <- c("wind", "temp", "effectSize")

## coplot of partial effect of wind and temp 
coplot(effectSize ~ wind|temp, pdta, pch = 16, overlap = 0)


## ------------------------------------------------------------
##
## regression: partial effects for three variables simultaneously
## (can be slow, so modify accordingly)
##
## ------------------------------------------------------------

n <- 1000
x <- matrix(rnorm(n * 3), ncol = 3)
y <- x[, 1] + x[, 1] * x[, 2] + x[, 1] * x[, 2] * x[, 3]
o <- rfsrc(y ~ ., data = data.frame(y = y, x))

## define target x values
x1 <- seq(-3, 3, length = 40)
x2 <- x3 <- seq(-3, 3, length = 10)

## extract second order partial effects
pdta <- do.call(rbind,
          lapply(x3, function(x3v) {
            cat("outer loop x3 = ", x3v, "\n")
            do.call(rbind,lapply(x2, function(x2v) {
              o <- partial(o,
                      partial.xvar = "X1",
                      partial.values = x1,
                      partial.xvar2 = c("X2", "X3"),
                      partial.values2 = c(x2v, x3v))
              cbind(x1, x2v, x3v, get.partial.plot.data(o)$yhat)
            }))
          }))
pdta <- data.frame(pdta)
colnames(pdta) <- c("x1", "x2", "x3", "effectSize")

## coplot of partial effects
coplot(effectSize ~ x1|x2*x3, pdta, pch = 16, overlap = 0)


## ------------------------------------------------------------
##
## classification
##
## ------------------------------------------------------------

iris.obj <- rfsrc(Species ~., data = iris)

## partial effect for sepal length
partial.obj <- partial(iris.obj,
                  partial.xvar = "Sepal.Length",
                  partial.values = iris.obj$xvar$Sepal.Length)

## extract partial effects for each species outcome
pdta1 <- get.partial.plot.data(partial.obj, target = "setosa")
pdta2 <- get.partial.plot.data(partial.obj, target = "versicolor")
pdta3 <- get.partial.plot.data(partial.obj, target = "virginica")

## plot the results
par(mfrow=c(1,1))
plot(pdta1$x, pdta1$yhat, type="b", pch = 16,
     xlab = "sepal length", ylab = "adjusted probability", 
     ylim = range(pdta1$yhat,pdta2$yhat,pdta3$yhat))
points(pdta2$x, pdta2$yhat, col = 2, type = "b", pch = 16)
points(pdta3$x, pdta3$yhat, col = 4, type = "b", pch = 16)
legend("topleft", legend=levels(iris.obj$yvar), fill = c(1, 2, 4))

## ------------------------------------------------------------
##
## survival
##
## ------------------------------------------------------------

data(veteran, package = "randomForestSRC")
v.obj <- rfsrc(Surv(time,status)~., veteran, nsplit = 10, ntree = 100)

## partial effect of age on mortality
partial.obj <- partial(v.obj,
  partial.type = "mort",
  partial.xvar = "age",
  partial.values = v.obj$xvar$age,
  partial.time = v.obj$time.interest)
pdta <- get.partial.plot.data(partial.obj)

plot(lowess(pdta$x, pdta$yhat, f = 1/3),
   type = "l", xlab = "age", ylab = "adjusted mortality")

## example where x is discrete - partial effect of age on mortality
## we use the granule=TRUE option
partial.obj <- partial(v.obj,
       partial.type = "mort",
       partial.xvar = "trt",
       partial.values = v.obj$xvar$trt,
       partial.time = v.obj$time.interest)
pdta <- get.partial.plot.data(partial.obj, granule = TRUE)
boxplot(pdta$yhat ~ pdta$x, xlab = "treatment", ylab = "partial effect")


## partial effects of karnofsky score on survival
karno <- quantile(v.obj$xvar$karno)
partial.obj <- partial(v.obj,
  partial.type = "surv",
  partial.xvar = "karno",
  partial.values = karno,
  partial.time = v.obj$time.interest)
pdta <- get.partial.plot.data(partial.obj)

matplot(pdta$partial.time, t(pdta$yhat), type = "l", lty = 1,
     xlab = "time", ylab = "karnofsky adjusted survival")
legend("topright", legend = paste0("karnofsky = ", karno), fill = 1:5)


## ------------------------------------------------------------
##
## competing risk
##
## ------------------------------------------------------------

data(follic, package = "randomForestSRC")
follic.obj <- rfsrc(Surv(time, status) ~ ., follic, nsplit = 3, ntree = 100)

## partial effect of age on years lost
partial.obj <- partial(follic.obj,
  partial.type = "years.lost",
  partial.xvar = "age",
  partial.values = follic.obj$xvar$age,
  partial.time = follic.obj$time.interest)
pdta1 <- get.partial.plot.data(partial.obj, target = 1)
pdta2 <- get.partial.plot.data(partial.obj, target = 2)

par(mfrow=c(2,2))
plot(lowess(pdta1$x, pdta1$yhat),
   type = "l", xlab = "age", ylab = "adjusted years lost relapse")
plot(lowess(pdta2$x, pdta2$yhat),
   type = "l", xlab = "age", ylab = "adjusted years lost death")

## partial effect of age on cif
partial.obj <- partial(follic.obj,
  partial.type = "cif",
  partial.xvar = "age",
  partial.values = quantile(follic.obj$xvar$age),
  partial.time = follic.obj$time.interest)
pdta1 <- get.partial.plot.data(partial.obj, target = 1)
pdta2 <- get.partial.plot.data(partial.obj, target = 2)

matplot(pdta1$partial.time, t(pdta1$yhat), type = "l", lty = 1,
     xlab = "time", ylab = "age adjusted cif for relapse")
matplot(pdta2$partial.time, t(pdta2$yhat), type = "l", lty = 1,
     xlab = "time", ylab = "age adjusted cif for death")


## ------------------------------------------------------------
##
## multivariate mixed outcomes
##
## ------------------------------------------------------------

mtcars2 <- mtcars
mtcars2$carb <- factor(mtcars2$carb)
mtcars2$cyl <- factor(mtcars2$cyl)
mtcars.mix <- rfsrc(Multivar(carb, mpg, cyl) ~ ., data = mtcars2)

## partial effect of displacement for each the three-outcomes
partial.obj <- partial(mtcars.mix,
                  partial.xvar = "disp",
                  partial.values = mtcars.mix$xvar$disp)
pdta1 <- get.partial.plot.data(partial.obj, m.target = "carb")
pdta2 <- get.partial.plot.data(partial.obj, m.target = "mpg")
pdta3 <- get.partial.plot.data(partial.obj, m.target = "cyl")

par(mfrow=c(2,2))
plot(lowess(pdta1$x, pdta1$yhat), type = "l", xlab="displacement", ylab="carb")
plot(lowess(pdta2$x, pdta2$yhat), type = "l", xlab="displacement", ylab="mpg")
plot(lowess(pdta3$x, pdta3$yhat), type = "l", xlab="displacement", ylab="cyl")


}
}
\keyword{partial}

\name{peakVO2}
\docType{data}
\alias{peakVO2}
\title{Systolic Heart Failure Data}

\description{

  The data involve 2231 patients with systolic heart failure who
  underwent cardiopulmonary stress testing at the Cleveland Clinic.  The
  primary end point was all-cause death.  In total, 39 variables were
  measured for each patient, including baseline clinical values and
  exercise stress test results.  A key variable of interest is
  peak VO2 (mL/kg per min), the peak respiratory exchange ratio.
  More details regarding the data can be found in Hsich et al. (2011).
  
}
\references{
  
  Hsich E., Gorodeski E.Z.,Blackstone E.H., Ishwaran H. and Lauer
  M.S. (2011). Identifying important risk factors for survival in
  systolic heart failure patients using random survival
  forests. Circulation: Cardio. Qual. Outcomes, 4(1), 39-45.
  
}
\examples{
\donttest{
## load the data
data(peakVO2, package = "randomForestSRC")

## random survival forest analysis
o <- rfsrc(Surv(ttodead, died)~., peakVO2)
print(o)

## partial effect of peak V02 on mortality
partial.o <- partial(o,
       partial.type = "mort",
       partial.xvar = "peak.vo2",
       partial.values = o$xvar$peak.vo2,
       partial.time = o$time.interest)
pdta.m <- get.partial.plot.data(partial.o)


## partial effect of peak V02 on survival
pvo2 <- quantile(o$xvar$peak.vo2)
partial.o <- partial(o,
       partial.type = "surv",
       partial.xvar = "peak.vo2",
       partial.values = pvo2,
       partial.time = o$time.interest)
pdta.s <- get.partial.plot.data(partial.o)
     

## compare the two plots
par(mfrow=c(1,2))    

plot(lowess(pdta.m$x, pdta.m$yhat, f = 2/3),
     type = "l", xlab = "peak VO2", ylab = "adjusted mortality")
rug(o$xvar$peak.vo2)

matplot(pdta.s$partial.time, t(pdta.s$yhat), type = "l", lty = 1,
          xlab = "years", ylab = "peak VO2 adjusted survival")
legend("bottomleft", legend = paste0("peak VO2 = ", pvo2),
       bty = "n", cex = .75, fill = 1:5)


}}
\keyword{datasets}

\name{quantreg.rfsrc}
\alias{quantreg.rfsrc}
\alias{quantreg}
\title{Quantile Regression Forests}
\description{
  Grows a univariate or multivariate quantile regression forest and returns
  its conditional quantile and density values.  Can be used for
  both training and testing purposes.
}
\usage{\method{quantreg}{rfsrc}(formula, data, object, newdata,
  method = "local", splitrule = NULL, prob = NULL, prob.epsilon = NULL,
  oob = TRUE, fast = FALSE, maxn = 1e3, ...)
}
\arguments{

  \item{formula}{A symbolic description of the model to be fit.
    Must be specified unless \code{object} is given.}

  \item{data}{Data frame containing the y-outcome and x-variables in
    the model. Must be specified unless \code{object} is given.}

  \item{object}{(Optional) A previously grown quantile regression
    forest.}

  \item{newdata}{(Optional) Test data frame used for prediction.  Note
    that prediction on test data must always be done with the
    \code{quantreg} function and not the \code{predict} function.  See
    example below.}

  \item{method}{Method used to calculate quantiles.  Three methods are
    provided: (1) A variation of the method used in Meinshausen (2006)
    based on forest weight (\code{method = "forest"}); (2) The
    Greenwald-Khanna algorithm, suited for big data, and specified by
    any one of the following: "gk", "GK", "G-K", "g-k"; (3) The default
    method, \code{method = "local"}, which uses the local adjusted cdf
    approach of Zhang et al. (2019).  This does not rely on forest
    weights and is reasonably fast. See below for further discussion.}

  \item{splitrule}{The default action is local adaptive quantile regression splitting,
    but this can be over-ridden by the user.  Not applicable to
    multivariate forests.  See details below.}
  
  \item{prob}{Target quantile probabilities when training.  If left unspecified,
    uses percentiles (1 through 99) for \code{method = "forest"}, and 
    for Greenwald-Khanna selects equally spaced percentiles optimized
    for accuracy (see below).}
  
  \item{prob.epsilon}{Greenwald-Khanna allowable error for quantile
    probabilities when training.}
  
  \item{oob}{Return OOB (out-of-bag) quantiles?  If false, in-bag values
    are returned.}

  \item{fast}{Use fast random forests, \code{rfsrc.fast}, in place of
    \code{rfsrc}?  Improves speed but may be less accurate.}

  \item{maxn}{Maximum number of unique y training values used when
    calculating the conditional density.}

  \item{...}{Further arguments to be passed to the \code{rfsrc}
    function used for fitting the quantile regression forest.}  
}
\details{

  \emph{Methods}

  \enumerate{

  \item
  The most common method for calculating RF quantiles uses the method
  described in Meinshausen (2006) using forest weights.  The forest
  weights method employed here (specified using \code{method}="forest"),
  however differs in that quantiles are estimated using a
  weighted local cumulative distribution function estimator.  For this
  reason, results may differ from Meinshausen (2006).  Moreover, results
  may also differ as the default splitting rule uses local adaptive
  quantile regression splitting instead of CART regression mean squared
  splitting which was used by Meinshausen (2006).  Note that local
  adaptive quantile regression splitting is not available for
  multivariate forests which reverts to the default multivariate
  composite splitting rule. In multivariate regression, users however do
  have the option to over-ride this using
  Mahalanobis splitting by setting \code{splitrule="mahalanobis"}

  \item
  A second method for estimating quantiles uses the Greenwald-Khanna
  (2001) algorithm (invoked by \code{method}="gk", "GK", "G-K" or
  "g-k").  While this will not be as accurate as forest weights, the
  high memory efficiency of Greenwald-Khanna makes it feasible to
  implement in big data settings unlike forest weights.

  The Greenwald-Khanna algorithm is implemented roughly as follows. 
  \itemize{
    \item
    To form a distribution of values for each case, from which we sample to
    determine quantiles, we create a chain of values for the case as we
    grow the forest.  Every time a case lands in a terminal node, we
    insert all of its co-inhabitants to its chain of values.
    \item
    The best case scenario is when tree node size is 1 because each case
    gets only one insert into its chain for that tree.  The worst case
    scenario is when node size is so large that trees stump. This is
    because each case receives insertions for the entire in-bag
    population.  
    \item
    What the user needs to know is that Greenwald-Khanna can become slow
    in counter-intutive settings such as when node size is large.  The
    easy fix is to change the epsilon quantile approximation that is
    requested.  You will see a significant speed-up just by doubling
    \code{prob.epsilon}.  This is because the chains stay a lot smaller as
    epsilon increases, which is exactly what you want when node sizes are
    large. Both time and space requirements for the algorithm are affected
    by epsilon.
    \item
    The best results for Greenwald-Khanna come from setting the number of
    quantiles equal to 2 times the sample size and epsilon to 1 over 2
    times the sample size which is the default values used if left
    unspecified.  This will be slow, especially for big data, and less
    stringent choices should be used if computational speed is of concern.
  }
  
  \item
  Finally, the default method, \code{method}="local", implements the
  locally adjusted cdf estimator of Zhang et al. (2019).  This does not
  use forest weights and is reasonably fast and can be used for large
  data.  However, this relies on the assumption of homogeneity of the
  error distribution, i.e. that errors are iid and therefore have equal
  variance.  While this is reasonably robust to departures of homogeneity,
  there are instances where this may perform poorly; see Zhang et
  al. (2019) for details.  If hetereogeneity is suspected we recommend
  \code{method}="forest".

  }

  \emph{Quantile-error metrics}

  \enumerate{

  \item When trees are grown with the quantile splitting rule
  (\code{splitrule = "quantile"}), \code{print.rfsrc} reports error
  metrics aligned with the quantile objective: (i) the standardized
  continuous ranked probability score (CRPS) computed from the estimated
  conditional CDF, and (ii) the pinball loss at selected quantile levels
  \eqn{\tau}.
  
  \item These metrics supplement the usual regression error rate printed
  by \code{print.rfsrc}. The line labeled "Requested performance error"
  remains the OOB mean squared error (MSE) by default (and is the
  quantity used to print \code{R squared}), regardless of the split
  rule. These quantile-metrics are printed in addition to MSE and do not
  change model fitting.

  \item Pinball levels can be controlled via
  \code{options(rfsrc.pinball.taus = c(0.1, 0.5, 0.9))}.  The CRPS
  curves used by \code{plot.quantreg} can be obtained with
  \code{get.quantile.crps()}.

  }

  
}
\value{
  
  Returns the object \code{quantreg} containing quantiles for each of
  the requested probabilities (which can be conveniently extracted using
  \code{get.quantile}).  Also contains the conditional density (and
  conditional cdf) for each case in the training data (or test data if
  provided) evaluated at each of the unique grow y-values.  The
  conditional density can be used to calculate conditional moments, such
  as the mean and standard deviation.  Use \code{get.quantile.stat} as a
  way to conveniently obtain these quantities.

  For multivariate forests, returned values will be a list of length
  equal to the number of target outcomes.

}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Greenwald M. and Khanna S. (2001).  Space-efficient online computation of
  quantile summaries. \emph{Proceedings of ACM SIGMOD}, 30(2):58-66.
  
  Meinshausen N. (2006) Quantile regression forests, \emph{Journal of
  Machine Learning Research}, 7:983-999.

  Zhang H., Zimmerman J., Nettleton D. and Nordman D.J. (2019).  Random
  forest prediction intervals. \emph{The American Statistician}. 4:1-5.
}
\seealso{
  \command{\link{rfsrc}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## regression example
## ------------------------------------------------------------

## standard call
o <- quantreg(mpg ~ ., mtcars)

## extract conditional quantiles
print(get.quantile(o))
print(get.quantile(o, c(.25, .50, .75)))

## extract conditional mean and standard deviation
print(get.quantile.stat(o))

## standardized continuous rank probabiliy score (crps) performance
plot(get.quantile.crps(o), type = "l")

## ------------------------------------------------------------
## quantile-metrics provided in print.rfsrc output 
## ------------------------------------------------------------
op <- options(rfsrc.pinball.taus = c(0.1, 0.5, 0.9))

o <- quantreg(mpg ~ ., data = mtcars)

print(o)

options(op)  ## restore options

## ------------------------------------------------------------
## train/test regression example
## ------------------------------------------------------------

## train (grow) call followed by test call
o <- quantreg(mpg ~ ., mtcars[1:20,])
o.tst <- quantreg(object = o, newdata = mtcars[-(1:20),])

## extract test set quantiles and conditional statistics
print(get.quantile(o.tst))
print(get.quantile.stat(o.tst))

## ------------------------------------------------------------
## quantile regression for Boston Housing canonical call
## ------------------------------------------------------------

if (library("mlbench", logical.return = TRUE)) {
  data(BostonHousing)
  o <- quantreg(medv ~ ., BostonHousing, nodesize = 1)
  print(o)
  plot.quantreg(o)
}


## ------------------------------------------------------------
## quantile regression for Boston Housing using forest method
## also demonstrates quantile statistics 
## ------------------------------------------------------------

if (library("mlbench", logical.return = TRUE)) {

  ## quantile regression with mse splitting
  data(BostonHousing)
  o <- quantreg(medv ~ ., BostonHousing, method = "forest", nodesize = 1)

  ## standardized continuous rank probabiliy score (crps) 
  plot(get.quantile.crps(o), type = "l")

  ## quantile regression plot
  plot.quantreg(o, .05, .95)
  plot.quantreg(o, .25, .75)

  ## (A) extract 25,50,75 quantiles
  quant.dat <- get.quantile(o, c(.25, .50, .75))

  ## (B) values expected under normality
  quant.stat <- get.quantile.stat(o)
  c.mean <- quant.stat$mean
  c.std <- quant.stat$std
  q.25.est <- c.mean + qnorm(.25) * c.std
  q.75.est <- c.mean + qnorm(.75) * c.std

  ## compare (A) and (B)
  print(head(data.frame(quant.dat[, -2],  q.25.est, q.75.est)))


}

## ------------------------------------------------------------
## multivariate mixed outcomes example
## quantiles are only returned for the continous outcomes
## ------------------------------------------------------------

dta <- mtcars
dta$cyl <- factor(dta$cyl)
dta$carb <- factor(dta$carb, ordered = TRUE)
o <- quantreg(cbind(carb, mpg, cyl, disp) ~., data = dta)

plot.quantreg(o, m.target = "mpg")
plot.quantreg(o, m.target = "disp")

## ------------------------------------------------------------
## multivariate regression example using Mahalanobis splitting
## ------------------------------------------------------------

dta <- mtcars
o <- quantreg(cbind(mpg, disp) ~., data = dta, splitrule = "mahal")

plot.quantreg(o, m.target = "mpg")
plot.quantreg(o, m.target = "disp")

## ------------------------------------------------------------
## example of quantile regression for ordinal data
## ------------------------------------------------------------

## use the wine data for illustration
data(wine, package = "randomForestSRC")

## run quantile regression
o <- quantreg(quality ~ ., wine, ntree = 100)

## extract "probabilities" = density values
qo.dens <- o$quantreg$density
yunq <- o$quantreg$yunq
colnames(qo.dens) <- yunq

## convert y to a factor 
yvar <- factor(cut(o$yvar, c(-1, yunq), labels = yunq)) 
 
## confusion matrix
qo.confusion <- get.confusion(yvar, qo.dens)
print(qo.confusion)

## normalized Brier score
cat("Brier:", 100 * get.brier.error(yvar, qo.dens), "\n")


## ------------------------------------------------------------
## example of large data using Greenwald-Khanna algorithm 
## ------------------------------------------------------------

## load the data and do quick and dirty imputation
data(housing, package = "randomForestSRC")
housing <- impute(SalePrice ~ ., housing,
         ntree = 50, nimpute = 1, splitrule = "random")

## Greenwald-Khanna algorithm 
## request a small number of quantiles 
o <- quantreg(SalePrice ~ ., housing, method = "gk",
        prob = (1:20) / 20, prob.epsilon = 1 / 20, ntree = 250)
plot.quantreg(o)
  
## ------------------------------------------------------------
## using mse splitting with local cdf method for large data
## ------------------------------------------------------------

## load the data and do quick and dirty imputation
data(housing, package = "randomForestSRC")
housing <- impute(SalePrice ~ ., housing,
         ntree = 50, nimpute = 1, splitrule = "random")

## use mse splitting and reduce number of trees
o <- quantreg(SalePrice ~ ., housing, splitrule = "mse", ntree = 250)
plot.quantreg(o)
  


}}
\keyword{quantile regression forests}

\name{rfsrc.news}
\alias{rfsrc.news}
\title{Show the NEWS file}
\description{
  Show the NEWS file of the \pkg{randomForestSRC} package.
}
\usage{
rfsrc.news(...)
}
\arguments{
  \item{...}{Further arguments passed to or from other methods.}
}
\value{
  None.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\keyword{documentation}

\name{follic}
\docType{data}
\alias{follic}
\title{Follicular Cell Lymphoma}
\description{
  Competing risk data set involving follicular cell lymphoma. 
}
\format{A data frame containing:
  \tabular{ll}{
    age      \tab age\cr
	  hgb      \tab hemoglobin (g/l)\cr
    clinstg	 \tab clinical stage: 1=stage I, 2=stage II\cr
	  ch       \tab chemotherapy\cr
	  rt       \tab radiotherapy\cr
    time     \tab first failure time\cr
    status   \tab censoring status: 0=censored, 1=relapse, 2=death
  }
}
\source{Table 1.4b, \emph{Competing Risks: A Practical Perspective}.}
\references{
  Pintilie M., (2006) \emph{Competing Risks: A
    Practical Perspective.} West Sussex: John Wiley and Sons.
}
\examples{
\donttest{
data(follic, package = "randomForestSRC")
follic.obj <- rfsrc(Surv(time, status) ~ ., follic, nsplit = 3, ntree = 100)
}
}
\keyword{datasets}

\name{vimp.rfsrc}
\alias{vimp.rfsrc}
\alias{vimp}
\title{VIMP for Single or Grouped Variables}
\description{
  Calculate variable importance (VIMP) for a single variable or group of
  variables for training or test data.
}
\usage{\method{vimp}{rfsrc}(object, xvar.names, 
  importance = c("anti", "permute", "random"), block.size = 10,
  joint = FALSE, seed = NULL, do.trace = FALSE, ...)
}

\arguments{

  \item{object}{An object of class \code{(rfsrc, grow)} or \code{(rfsrc, forest)}. The original \code{rfsrc} call must have been made with \code{forest = TRUE}.}

  \item{xvar.names}{Character vector of x-variable names to be evaluated. If not specified, all variables are used.}

  \item{importance}{Type of variable importance (VIMP) to compute.}

  \item{block.size}{Integer specifying the number of trees per block used for VIMP calculation. Balances between ensemble-level and tree-level estimates.}

  \item{joint}{Logical indicating whether to compute joint VIMP for the specified variables.}

  \item{seed}{Negative integer used to set the random number generator seed.}

  \item{do.trace}{Number of seconds between printed progress updates.}

  \item{...}{Additional arguments passed to or from other methods.}

}

\details{

Using a previously trained forest, this function calculates variable importance (VIMP) for the specified variables in \code{xvar.names}. By default, VIMP is computed using the original training data, but the user may supply a new test set via the \code{newdata} argument. See \code{\link{rfsrc}} for further details on how VIMP is computed.

If \code{joint = TRUE}, joint VIMP is returned. This is defined as the importance of a group of variables when the entire group is perturbed simultaneously.

Setting \code{csv = TRUE} returns case-specific VIMP, which provides VIMP estimates at the individual observation level. This applies to all families except survival. See examples below.

}


\value{
  An object of class \code{(rfsrc, predict)} containing importance
  values.
}

\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Ishwaran H. (2007).  Variable importance in binary regression
  trees and forests,  \emph{Electronic J. Statist.}, 1:519-537.
}
\seealso{
  \command{\link{holdout.vimp.rfsrc}},
  \command{\link{rfsrc}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## classification example
## showcase different vimp
## ------------------------------------------------------------

iris.obj <- rfsrc(Species ~ ., data = iris)

## anti vimp (default)
print(vimp(iris.obj)$importance)

## anti vimp using brier prediction error
print(vimp(iris.obj, perf.type = "brier")$importance)

## permutation vimp
print(vimp(iris.obj, importance = "permute")$importance)

## random daughter vimp
print(vimp(iris.obj, importance = "random")$importance)

## joint anti vimp 
print(vimp(iris.obj, joint = TRUE)$importance)

## paired anti vimp
print(vimp(iris.obj, c("Petal.Length", "Petal.Width"), joint = TRUE)$importance)
print(vimp(iris.obj, c("Sepal.Length", "Petal.Width"), joint = TRUE)$importance)

## ------------------------------------------------------------
## survival example
## anti versus permute VIMP with different block sizes
## ------------------------------------------------------------

data(pbc, package = "randomForestSRC")
pbc.obj <- rfsrc(Surv(days, status) ~ ., pbc)

print(vimp(pbc.obj)$importance)
print(vimp(pbc.obj, block.size=1)$importance)
print(vimp(pbc.obj, importance="permute")$importance)
print(vimp(pbc.obj, importance="permute", block.size=1)$importance)

## ------------------------------------------------------------
## imbalanced classification example
## see the imbalanced function for more details
## ------------------------------------------------------------

data(breast, package = "randomForestSRC")
breast <- na.omit(breast)
f <- as.formula(status ~ .)
o <- rfsrc(f, breast, ntree = 2000)

## permutation vimp
print(100 * vimp(o, importance = "permute")$importance)

## anti vimp using gmean performance
print(100 * vimp(o, perf.type = "gmean")$importance[, 1])

## ------------------------------------------------------------
## regression example
## ------------------------------------------------------------

airq.obj <- rfsrc(Ozone ~ ., airquality)
print(vimp(airq.obj))

## ------------------------------------------------------------
## regression example where vimp is calculated on test data
## ------------------------------------------------------------

set.seed(100080)
train <- sample(1:nrow(airquality), size = 80)
airq.obj <- rfsrc(Ozone~., airquality[train, ])

## training data vimp
print(airq.obj$importance)
print(vimp(airq.obj)$importance)

## test data vimp
print(vimp(airq.obj, newdata = airquality[-train, ])$importance)

## ------------------------------------------------------------
## case-specific vimp
## returns VIMP for each case
## ------------------------------------------------------------

o <- rfsrc(mpg~., mtcars)
v <- vimp(o, csv = TRUE)
csvimp <- get.mv.csvimp(v, standardize=TRUE)
print(csvimp)

## ------------------------------------------------------------
## case-specific joint vimp
## returns joint VIMP for each case
## ------------------------------------------------------------

o <- rfsrc(mpg~., mtcars)
v <- vimp(o, joint = TRUE, csv = TRUE)
csvimp <- get.mv.csvimp(v, standardize=TRUE)
print(csvimp)

## ------------------------------------------------------------
## case-specific joint vimp for multivariate regression
## returns joint VIMP for each case, for each outcome
## ------------------------------------------------------------

o <- rfsrc(Multivar(mpg, cyl) ~., data = mtcars)
v <- vimp(o, joint = TRUE, csv = TRUE)
csvimp <- get.mv.csvimp(v, standardize=TRUE)
print(csvimp)

}
}
\keyword{variable selection}
\keyword{predict}

\name{sidClustering.rfsrc}
\alias{sidClustering.rfsrc}
\alias{sidClustering}
\title{sidClustering using SID (Staggered Interaction Data) for Unsupervised Clustering}

\description{
  Clustering of unsupervised data using SID (Mantero and Ishwaran,
  2021).  Also implements the artificial two-class approach of
  Breiman (2003).
}

\usage{\method{sidClustering}{rfsrc}(data,
  method = "sid",
  k = NULL,
  reduce = TRUE,
  ntree = 500,
  ntree.reduce = function(p, vtry){100 * p / vtry},
  fast = FALSE,
  x.no.sid = NULL,
  use.sid.for.x = TRUE,
  x.only = NULL, y.only = NULL,
  dist.sharpen = TRUE, ...)
}

\arguments{

  \item{data}{A data frame containing the unsupervised data.}

  \item{method}{Clustering method. Default is \code{"sid"}, which implements SID clustering using Staggered Interaction Data (Mantero and Ishwaran, 2021). An alternative approach reformulates the problem as a two-class supervised learning task using artificial data, per Breiman (2003) and Shi-Horvath (2006). Mode 1 is specified via \code{"sh"}, \code{"SH"}, \code{"sh1"}, or \code{"SH1"}; Mode 2 via \code{"sh2"} or \code{"SH2"}. A third method, \code{"unsupv"}, uses a plain unsupervised forest where the data act as both features and responses, split using the multivariate rule. This is faster than SID but may be less accurate.}

  \item{k}{Requested number of clusters. Can be a single integer or a vector. If a scalar, returns a vector assigning each observation to a cluster. If a vector, returns a matrix with one column per requested value of \code{k}, each containing a clustering assignment.}

  \item{reduce}{Logical. If \code{TRUE}, applies a variable reduction step via holdout VIMP. This is conservative and computationally intensive but has strong false discovery control. Applies only when \code{method = "sid"}.}

  \item{ntree}{Number of trees used in the main SID clustering analysis.}

  \item{ntree.reduce}{Number of trees used in the holdout VIMP step during variable reduction. See \code{\link{holdout.vimp}} for details.}

  \item{fast}{Logical. If \code{TRUE}, uses the fast implementation \code{\link{rfsrc.fast}} instead of \code{\link{rfsrc}}. Improves speed at the cost of accuracy.}

  \item{x.no.sid}{Variables to exclude from SID transformation. Can be either a separate data frame (not overlapping with \code{data}) or a character vector of variable names from \code{data}. These variables will be included in the final design matrix without SID processing. Applies only when \code{method = "sid"}.}

  \item{use.sid.for.x}{Logical. If \code{FALSE}, reverses the roles of features and responses in the SID process. Staggered interactions are applied to the outcome rather than to features. This option is slower and generally less effective. Included for legacy compatibility. Applies only when \code{method = "sid"}.}

  \item{x.only}{Character vector specifying which variables to use as features. Applies only when \code{method = "unsupv"}.}

  \item{y.only}{Character vector specifying which variables to use as multivariate responses. Applies only when \code{method = "unsupv"}.}

  \item{dist.sharpen}{Logical. If \code{TRUE} (default), applies Euclidean distance to the forest distance matrix to improve clustering ("distance sharpening"). The resulting distance matrix will not be bounded between 0 and 1. Turning this off speeds up computation but may reduce clustering quality. Applies only when \code{method = "sid"} or \code{"unsupv"}.}

  \item{...}{Additional arguments passed to \code{\link{rfsrc}} to control forest construction.}

}

\details{

Given an unsupervised dataset, random forests is used to compute a distance matrix measuring dissimilarity between all pairs of observations. By default, hierarchical clustering is applied to this distance matrix, although users may apply any other clustering algorithm. See the examples below for alternative workflows.

The default method, \code{method = "sid"}, implements SID clustering (sidClustering). The algorithm begins by enhancing the original feature space using Staggered Interaction Data (SID). This transformation creates:

\itemize{
  \item SID main features: shifted and staggered versions of the original features that are made strictly positive and mutually non-overlapping in range;
  \item SID interaction features: pairwise multiplicative interactions formed between all SID main features.
}

A multivariate random forest is trained to predict SID main features
using the SID interaction features as predictors. The rationale is that
if a feature is informative for distinguishing clusters, it will exhibit
systematic variation across the data space. Because each interaction
feature is uniquely defined by the features it is formed from, node
splits on interaction terms are able to capture and separate such
variation, thus effectively identifying the clusters. See Mantero and
Ishwaran (2021) for further details.

Since SID includes all pairwise interactions, the dimensionality of the feature space grows quadratically with the number of original variables (or worse when factor variables are present). As such, the reduction step using holdout variable importance (VIMP) is strongly recommended (enabled by default). This step can be disabled using \code{reduce = FALSE}, but only when the original feature space is of manageable size.

A second approach, proposed by Breiman (2003) and refined by Shi and Horvath (2006), transforms the unsupervised task into a two-class supervised classification problem. The first class consists of the original data, while the second class is generated artificially. The goal is to separate real data from synthetic data. A proximity matrix is constructed from this supervised model, and the proximity values for the original class are extracted and converted into a distance matrix (distance = 1 - proximity) for clustering. 

Artificial data can be generated using two modes:
\itemize{
  \item \code{mode 1} (default): draws random values from the empirical distribution of each feature;
  \item \code{mode 2}: draws uniformly between the observed minimum and maximum of each feature.
}

This method is invoked by setting \code{method = "sh"}, \code{"sh1"}, or \code{"sh2"}. Mantero and Ishwaran (2021) found that while this approach works in certain settings, it can fail when clusters exist in lower-dimensional subspaces (e.g., when defined by interactions or involving both factors and continuous variables). Among the two modes, mode 1 is generally more robust.

The third method, \code{method = "unsupv"}, trains a multivariate forest using the data both as predictors and as responses. The multivariate splitting rule is applied at each node. This method is fast and simple but may be less accurate compared to SID clustering.

The package includes a helper function \code{sid.perf.metric} for evaluating clustering performance using a normalized score; smaller values indicate better performance. See Mantero and Ishwaran (2021) for theoretical background and empirical benchmarking.
}

\value{

A list with the following components:

\item{clustering}{A vector or matrix assigning each observation to a cluster. If multiple values of \code{k} were specified, this is a matrix with one column per clustering solution.}

\item{rf}{The trained random forest object used in the clustering procedure. This is typically a multivariate forest (for \code{method = "sid"} or \code{"unsupv"}) or a classification forest (for Breiman-style methods).}

\item{dist}{The distance matrix computed from the forest. Used for clustering. For \code{method = "sid"}, this is based on the forest dissimilarity; for Breiman/SH methods, this is one minus the proximity matrix.}

\item{sid}{The SID-transformed data used in the clustering (applies only to \code{method = "sid"}). Provided as a list with separate components for the staggered features and their interactions, corresponding to outcomes and predictors in the multivariate forest.}
}

  
\author{Hemant Ishwaran and Udaya B. Kogalur}

\references{

  Breiman, L. (2003). \emph{Manual on setting up, using and
  understanding random forest, V4.0}.  University of California
  Berkeley, Statistics Department, Berkeley.
  
  Mantero A. and Ishwaran H. (2021).  Unsupervised random forests.
  \emph{Statistical Analysis and Data Mining}, 14(2):144-167.
  
  Shi, T. and Horvath, S. (2006). Unsupervised learning with random forest
  predictors. \emph{Journal of Computational and Graphical Statistics},
  15(1):118-138.  
}

\seealso{
  \command{\link{rfsrc}},
  \command{\link{rfsrc.fast}}
}

\examples{
\donttest{
## ------------------------------------------------------------
## mtcars example
## ------------------------------------------------------------

## default SID method 
o1 <- sidClustering(mtcars)
print(split(mtcars, o1$cl[, 10]))

## using artifical class approach
o1.sh <- sidClustering(mtcars, method = "sh")
print(split(mtcars, o1.sh$cl[, 10]))


## ------------------------------------------------------------
## glass data set
## ------------------------------------------------------------

if (library("mlbench", logical.return = TRUE)) {

  ## this is a supervised problem, so we first strip the class label
  data(Glass)
  glass <- Glass
  y <- Glass$Type
  glass$Type <- NULL

  ## default SID call 
  o2 <- sidClustering(glass, k = 6)
  print(table(y, o2$cl))
  print(sid.perf.metric(y, o2$cl))

  ## compare with Shi-Horvath mode 1 
  o2.sh <- sidClustering(glass, method = "sh1", k = 6)
  print(table(y, o2.sh$cl))
  print(sid.perf.metric(y, o2.sh$cl))

  ## plain-vanilla unsupervised analysis
  o2.un <- sidClustering(glass, method = "unsupv", k = 6)
  print(table(y, o2.un$cl))
  print(sid.perf.metric(y, o2.un$cl))

}

## ------------------------------------------------------------
## vowel data set
## ------------------------------------------------------------

if (library("mlbench", logical.return = TRUE) &&
    library("cluster", logical.return = TRUE)) {

  ## strip the class label
  data(Vowel)
  vowel <- Vowel
  y <- Vowel$Class
  vowel$Class <- NULL

  ## SID 
  o3 <- sidClustering(vowel, k = 11)
  print(table(y, o3$cl))
  print(sid.perf.metric(y, o3$cl))

  ## compare to Shi-Horvath which performs poorly in
  ## mixed variable settings
  o3.sh <- sidClustering(vowel, method = "sh1", k = 11)
  print(table(y, o3.sh$cl))
  print(sid.perf.metric(y, o3.sh$cl))

  ## Shi-Horvath improves with PAM clustering
  ## but still not as good as SID
  o3.sh.pam <- pam(o3.sh$dist, k = 11)$clustering
  print(table(y, o3.sh.pam))
  print(sid.perf.metric(y, o3.sh.pam))

  ## plain-vanilla unsupervised analysis
  o3.un <- sidClustering(vowel, method = "unsupv", k = 11)
  print(table(y, o3.un$cl))
  print(sid.perf.metric(y, o3.un$cl))

}

## ------------------------------------------------------------
##  two-d V-shaped cluster (y=x, y=-x) sitting in 12-dimensions 
##  illustrates superiority of SID to Breiman/Shi-Horvath
## ------------------------------------------------------------

p <- 10
m <- 250
n <- 2 * m
std <- .2

x <- runif(n, 0, 1)
noise <- matrix(runif(n * p, 0, 1), n)
y <- rep(NA, n)
y[1:m] <- x[1:m] + rnorm(m, sd = std)
y[(m+1):n] <- -x[(m+1):n] + rnorm(m, sd = std)
vclus <- data.frame(clus = c(rep(1, m), rep(2,m)), x = x, y = y, noise)

## SID
o4 <- sidClustering(vclus[, -1], k = 2)
print(table(vclus[, 1], o4$cl))
print(sid.perf.metric(vclus[, 1], o4$cl))

## Shi-Horvath
o4.sh <- sidClustering(vclus[, -1], method = "sh1", k = 2)
print(table(vclus[, 1], o4.sh$cl))
print(sid.perf.metric(vclus[, 1], o4.sh$cl))

## plain-vanilla unsupervised analysis
o4.un <- sidClustering(vclus[, -1], method = "unsupv", k = 2)
print(table(vclus[, 1], o4.un$cl))
print(sid.perf.metric(vclus[, 1], o4.un$cl))


## ------------------------------------------------------------
##  two-d V-shaped cluster using fast random forests
## ------------------------------------------------------------

o5 <- sidClustering(vclus[, -1], k = 2, fast = TRUE)
print(table(vclus[, 1], o5$cl))
print(sid.perf.metric(vclus[, 1], o5$cl))


}}
\keyword{unsupervised}
\keyword{clustering}

\name{hd}
\docType{data}
\alias{hd}
\title{Hodgkin's Disease}
\description{
  Competing risk data set involving Hodgkin's disease.
}
\format{A data frame containing:
  \tabular{ll}{
    age      \tab age\cr
    sex      \tab gender\cr
	  trtgiven \tab treatment: RT=radition, CMT=Chemotherapy and radiation\cr
	  medwidsi \tab mediastinum involvement: N=no, S=small, L=Large\cr
	  extranod \tab extranodal disease: Y=extranodal disease, N=nodal disease\cr
    clinstg	 \tab clinical stage: 1=stage I, 2=stage II\cr
    time     \tab first failure time\cr
    status   \tab censoring status: 0=censored, 1=relapse, 2=death
  }
}
\source{
  Table 1.6b, \emph{Competing Risks: A Practical Perspective}.
}
\references{
  Pintilie M., (2006) \emph{Competing Risks: A
    Practical Perspective.} West Sussex: John Wiley and Sons.
}
\examples{data(hd, package = "randomForestSRC")}
\keyword{datasets}

\name{predict.rfsrc}
\alias{predict.rfsrc}
\title{Prediction for Random Forests for Survival, Regression, and Classification}
\description{
  Obtain predicted values using a forest.  Also returns performance
  values if the test data contains y-outcomes.
}
\usage{\method{predict}{rfsrc}(object,
  newdata,
  importance = c(FALSE, TRUE, "none", "anti", "permute", "random"),
  get.tree = NULL,
  block.size = if (any(is.element(as.character(importance),
                     c("none", "FALSE")))) NULL else 10,
  na.action = c("na.omit", "na.impute", "na.random"),
  outcome = c("train", "test"),
  perf.type = NULL,
  proximity = FALSE,
  forest.wt = FALSE,
  ptn.count = 0,
  distance = FALSE,
  var.used = c(FALSE, "all.trees", "by.tree"),
  split.depth = c(FALSE, "all.trees", "by.tree"),
  case.depth = FALSE,
  seed = NULL,
  do.trace = FALSE, membership = FALSE,
  marginal.xvar = NULL,  ...)
}

\arguments{
  \item{object}{An object of class \code{(rfsrc, grow)} or \code{(rfsrc, forest)}.}
  
  \item{newdata}{Test data. If omitted, the original training data is used.}
  
  \item{importance}{Method for computing variable importance (VIMP). See \code{\link{vimp}} for additional options including joint importance. See \code{\link{holdout.vimp}} for an alternative importance measure.}
  
  \item{get.tree}{Vector of integers specifying which trees to use for ensemble calculations. Defaults to all trees. Useful for extracting ensembles, VIMP, or proximity from specific trees. If specified, \code{block.size} is overridden to match the number of trees. See examples for per-tree VIMP extraction.}
  
  \item{block.size}{Controls the granularity of error rate and VIMP calculation. If \code{NULL}, error is reported only for the final tree. Set to an integer \code{k} to compute error every \code{k} trees. For VIMP, calculations are done in blocks of size \code{block.size}, balancing between tree-level and forest-level assessments.}
  
  \item{na.action}{Action to take when missing values are present. Options are \code{"na.omit"} (default), \code{"na.random"} for fast random imputation, or \code{"na.impute"} to use the imputation method in \code{rfsrc}.}
  
  \item{outcome}{Specifies whether predicted values should be based on the outcomes from the training data (\code{"train"}, default) or test data. Ignored if \code{newdata} is missing or if test outcomes are unavailable.}
  
  \item{perf.type}{Optional metric for prediction, VIMP, and error. Currently used for classification and multivariate classification. Choices: \code{"misclass"} (default), \code{"brier"}, and \code{"gmean"}.}
  
  \item{proximity}{Whether to compute a proximity matrix. Options include
    \code{"inbag"}, \code{"oob"}, \code{"all"}, \code{"hybrid"},
    \code{TRUE}, or \code{FALSE}. With \code{newdata}, \code{"hybrid"}
    compares test observations (rows) with original training observations
    (columns). Hybrid mode is valid only when \code{newdata} is supplied
    with the default \code{outcome = "train"}.}
  
  \item{distance}{Whether to compute the distance matrix. Options are the
    same as for \code{proximity}. Hybrid mode uses the same rectangular
    test-by-training layout.}
  
  \item{forest.wt}{Whether to compute the forest weight matrix. Options are the same as for \code{proximity}.}
  
  \item{ptn.count}{If nonzero, each tree is pruned to have this many terminal nodes. Only the terminal node membership is returned; no prediction is made. Default is \code{ptn.count = 0}.}
  
  \item{var.used}{If \code{TRUE}, records how many times each variable was used for splitting.}
  
  \item{split.depth}{If \code{TRUE}, returns minimal depth of each variable per case.}
  
  \item{case.depth}{If \code{TRUE}, returns a matrix of the depth at which each case first splits in each tree.}
  
  \item{seed}{Negative integer used to set the random seed.}
  
  \item{do.trace}{Number of seconds between progress updates during execution.}
  
  \item{membership}{If \code{TRUE}, returns terminal node membership and in-bag information.}
  
  \item{marginal.xvar}{Vector of variable names to marginalize over when calculating weights or proximity. If a variable is marginalized, its split does not partition the data; all cases are passed to both daughters. When all splits involve marginalized variables, terminal nodes contain the full dataset. When no marginalized variables are used, membership is unchanged. Default is \code{NULL} (no marginalization).}
  
  \item{...}{Additional arguments passed to or from other methods.}
}


\details{

Predicted values are obtained by "dropping" the test data down the trained forest-i.e., the forest grown using the training data. If the test data includes y-outcome values, performance metrics are also returned. Variable importance (VIMP), including joint VIMP, is returned if requested.

If no test data is supplied, the function uses the original training data and enters "restore" mode. This allows users to extract outputs from the trained forest that were not requested during the original grow call.

When \code{newdata} is supplied, non-hybrid proximity and distance matrices
compare test observations with one another. Setting either option to
\code{"hybrid"} instead compares each retained test observation with each
retained original training observation, producing an \code{ntest} by
\code{ntrain} matrix without constructing a square matrix over the combined
data. Hybrid mode should not be combined with \code{marginal.xvar}.

If \code{outcome = "test"}, predictions are computed using y-outcomes from the test data (which must include outcome values). Terminal node statistics are recalculated using these outcomes, while the tree topology remains fixed from training. Error rates and VIMP are then computed by bootstrapping the test set and applying out-of-bagging to maintain unbiased estimates.

Set \code{csv = TRUE} to return case-specific VIMP, and \code{cse = TRUE} to return case-specific error rates. These apply to all families except survival. Both options can also be used at training time.
}

\value{
An object of class \code{(rfsrc, predict)}, which is a list with the following components:

\describe{
  \item{call}{The original grow call to \code{rfsrc}.}
  \item{family}{The family used in the analysis.}
  \item{n}{Sample size of the test data (after handling missing values).}
  \item{ntree}{Number of trees in the trained forest.}
  \item{yvar}{Y-outcome values from the test data or original grow data (if \code{newdata} is missing).}
  \item{yvar.names}{Character vector of response variable names.}
  \item{xvar}{Data frame of test set predictor variables.}
  \item{xvar.names}{Character vector of predictor variable names.}
  \item{leaf.count}{Vector of length \code{ntree} giving the number of terminal nodes per tree.}
  \item{proximity}{Proximity matrix computed on the test data. In hybrid
    mode, rows correspond to retained test observations and columns correspond
    to retained original training observations.}
  \item{distance}{Distance matrix. Hybrid mode uses the same test-by-training
    layout as hybrid proximity.}
  \item{forest}{The trained forest object.}
  \item{forest.wt}{Forest weight matrix for test cases.}
  \item{ptn.membership}{Matrix of pruned terminal node membership. Only returned if \code{ptn.count > 0}.}
  \item{membership}{Matrix of terminal node membership for test cases. Each column corresponds to one tree.}
  \item{inbag}{Matrix indicating how many times each case appears in the bootstrap sample for each tree.}
  \item{var.used}{Number of times each variable was used in splitting.}
  \item{imputed.indv}{Indices of test observations with missing values.}
  \item{imputed.data}{Imputed version of the test data. Columns are ordered with responses first, followed by predictors.}
  \item{split.depth}{Matrix or array recording minimal depth of each variable for each case, optionally by tree.}
  \item{err.rate}{Cumulative out-of-bag (OOB) error rate, if y-outcomes are present.}
  \item{importance}{Variable importance (VIMP) for the test data. May be \code{NULL}.}
  \item{predicted}{Predicted values for the test data.}
  \item{predicted.oob}{OOB predicted values. May be \code{NULL} depending on context.}
  \item{quantile}{Estimated quantile values at the requested probabilities (quantile regression only).}
  \item{quantile.oob}{OOB quantile values. May be \code{NULL}.}

  \item{class}{(Classification only) Predicted class labels.}
  \item{class.oob}{(Classification only) OOB predicted class labels.}

  \item{regrOutput}{(Multivariate only) List of performance measures for multivariate regression outcomes.}
  \item{clasOutput}{(Multivariate only) List of performance measures for multivariate categorical outcomes.}

  \item{chf}{(Survival or competing risks) Cumulative hazard function (CHF).}
  \item{chf.oob}{(Survival or competing risks) OOB CHF. May be \code{NULL}.}
  \item{survival}{(Survival only) Survival function estimates.}
  \item{survival.oob}{(Survival only) OOB survival function. May be \code{NULL}.}
  \item{time.interest}{(Survival or competing risks) Sorted unique event times.}
  \item{ndead}{(Survival or competing risks) Number of deaths observed.}

  \item{cif}{(Competing risks only) Cumulative incidence function (CIF) for each event type.}
  \item{cif.oob}{(Competing risks only) OOB CIF. May be \code{NULL}.}
  \item{chf}{(Competing risks only) Cause-specific cumulative hazard function (CSCHF).}
  \item{chf.oob}{(Competing risks only) OOB CSCHF. May be \code{NULL}.}
}
}


\note{

The dimensions and contents of returned objects depend on the forest family and whether y-outcomes are available in the test data. In particular, performance-related components (e.g., error rate, VIMP) will be \code{NULL} if y-outcomes are missing.

For multivariate families, predicted values, VIMP, error rates, and performance metrics are stored in the lists \code{regrOutput} and \code{clasOutput}. These can be accessed using the helper functions \code{\link{get.mv.predicted}}, \code{\link{get.mv.vimp}}, and \code{\link{get.mv.error}}.

}


\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Breiman L. (2001). Random forests, \emph{Machine Learning}, 45:5-32.
  
  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S.
  (2008).  Random survival forests, \emph{Ann. App.
    Statist.}, 2:841-860.
  
  Ishwaran H. and Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
}	
\seealso{
  \command{\link{holdout.vimp.rfsrc}},
  \command{\link{plot.competing.risk.rfsrc}},
  \command{\link{plot.rfsrc}},
  \command{\link{plot.survival.rfsrc}},
  \command{\link{plot.variable.rfsrc}},
  \command{\link{rfsrc}},
  \command{\link{rfsrc.fast}},
  \command{\link{vimp.rfsrc}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## typical train/testing scenario
## ------------------------------------------------------------

data(veteran, package = "randomForestSRC")
train <- sample(1:nrow(veteran), round(nrow(veteran) * 0.80))
veteran.grow <- rfsrc(Surv(time, status) ~ ., veteran[train, ]) 
veteran.pred <- predict(veteran.grow, veteran[-train, ])
print(veteran.grow)
print(veteran.pred)

## ------------------------------------------------------------
## hybrid proximity and distance
## - rows are test cases; columns are training cases
## ------------------------------------------------------------

data(mtcars)
hybrid.train <- mtcars[1:24, ]
hybrid.test <- mtcars[25:32, ]
hybrid.obj <- rfsrc(mpg ~ ., data = hybrid.train, ntree = 50)

hybrid.proximity <- predict(hybrid.obj, hybrid.test,
                            proximity = "hybrid")$proximity
hybrid.distance <- predict(hybrid.obj, hybrid.test,
                           distance = "hybrid")$distance

## Both matrices have 8 test rows and 24 training columns.
print(dim(hybrid.proximity))
print(dim(hybrid.distance))


## ------------------------------------------------------------
## restore mode
## - if predict is called without specifying the test data
##   the original training data is used and the forest is restored
## ------------------------------------------------------------

## first train the forest
airq.obj <- rfsrc(Ozone ~ ., data = airquality)

## now we restore it and compare it to the original call
## they are identical
predict(airq.obj)
print(airq.obj)

## we can retrieve various outputs that were not asked for in
## in the original call

## here we extract the proximity matrix
prox <- predict(airq.obj, proximity = TRUE)$proximity
print(prox[1:10,1:10])

## here we extract the number of times a variable was used to grow
## the grow forest
var.used <- predict(airq.obj, var.used = "by.tree")$var.used
print(head(var.used))

## ------------------------------------------------------------
## prediction when test data has missing values
## ------------------------------------------------------------

data(pbc, package = "randomForestSRC")
trn <- pbc[1:312,]
tst <- pbc[-(1:312),]
o <- rfsrc(Surv(days, status) ~ ., trn)

## default imputation method used by rfsrc
print(predict(o, tst, na.action = "na.impute"))

## random imputation
print(predict(o, tst, na.action = "na.random"))

## ------------------------------------------------------------
## requesting different performance for classification
## ------------------------------------------------------------

## default performance is misclassification
o <- rfsrc(Species~., iris)
print(o)

## get (normalized) brier performance
print(predict(o, perf.type = "brier"))

## ------------------------------------------------------------
## vimp for each tree: illustrates get.tree 
## ------------------------------------------------------------

## regression analysis but no VIMP
o <- rfsrc(mpg~., mtcars)

## now extract VIMP for each tree using get.tree
vimp.tree <- do.call(rbind, lapply(1:o$ntree, function(b) {
     predict(o, get.tree = b, importance = TRUE)$importance
}))

## boxplot of tree VIMP
boxplot(vimp.tree, outline = FALSE, col = "cyan")
abline(h = 0, lty = 2, col = "red")

## summary information of tree VIMP
print(summary(vimp.tree))

## extract tree-averaged VIMP using importance=TRUE
## remember to set block.size to 1
print(predict(o, importance = TRUE, block.size = 1)$importance)

## use direct call to vimp() for tree-averaged VIMP
print(vimp(o, block.size = 1)$importance)

## ------------------------------------------------------------
## vimp for just a few trees
## illustrates how to get vimp if you have a large data set
## ------------------------------------------------------------

## survival analysis but no VIMP
data(pbc, package = "randomForestSRC")
o <- rfsrc(Surv(days, status) ~ ., pbc, ntree = 2000)

## get vimp for a small number of trees
print(predict(o, get.tree=1:250, importance = TRUE)$importance)


## ------------------------------------------------------------
## case-specific vimp
## returns VIMP for each case
## ------------------------------------------------------------

o <- rfsrc(mpg~., mtcars)
op <- predict(o, importance = TRUE, csv = TRUE)
csvimp <- get.mv.csvimp(op, standardize=TRUE)
print(csvimp)

## ------------------------------------------------------------
## case-specific error rate
## returns tree-averaged error rate for each case
## ------------------------------------------------------------

o <- rfsrc(mpg~., mtcars)
op <- predict(o, importance = TRUE, cse = TRUE)
cserror <- get.mv.cserror(op, standardize=TRUE)
print(cserror)


## ------------------------------------------------------------
## predicted probability and predicted class labels are returned
## in the predict object for classification analyses
## ------------------------------------------------------------

data(breast, package = "randomForestSRC")
breast.obj <- rfsrc(status ~ ., data = breast[(1:100), ])
breast.pred <- predict(breast.obj, breast[-(1:100), ])
print(head(breast.pred$predicted))
print(breast.pred$class)


## ------------------------------------------------------------
## unique feature of randomForestSRC
## cross-validation can be used when factor labels differ over
## training and test data
## ------------------------------------------------------------

## first we convert all x-variables to factors
data(veteran, package = "randomForestSRC")
veteran2 <- data.frame(lapply(veteran, factor))
veteran2$time <- veteran$time
veteran2$status <- veteran$status

## split the data into unbalanced train/test data (25/75)
## the train/test data have the same levels, but different labels
train <- sample(1:nrow(veteran2), round(nrow(veteran2) * .25))
summary(veteran2[train,])
summary(veteran2[-train,])

## train the forest and use this to predict on test data
o.grow <- rfsrc(Surv(time, status) ~ ., veteran2[train, ]) 
o.pred <- predict(o.grow, veteran2[-train , ])
print(o.grow)
print(o.pred)

## even harder ... factor level not previously encountered in training
veteran3 <- veteran2[1:3, ]
veteran3$celltype <- factor(c("newlevel", "1", "3"))
o2.pred <- predict(o.grow, veteran3)
print(o2.pred)
## the unusual level is treated like a missing value but is not removed
print(o2.pred$xvar)

## ------------------------------------------------------------
## example illustrating the flexibility of outcome = "test"
## illustrates restoration of forest via outcome = "test"
## ------------------------------------------------------------

## first we train the forest
data(pbc, package = "randomForestSRC")
pbc.grow <- rfsrc(Surv(days, status) ~ ., pbc)

## use predict with outcome = TEST
pbc.pred <- predict(pbc.grow, pbc, outcome = "test")

## notice that error rates are the same!!
print(pbc.grow)
print(pbc.pred)

## note this is equivalent to restoring the forest
pbc.pred2 <- predict(pbc.grow)
print(pbc.grow)
print(pbc.pred)
print(pbc.pred2)

## similar example, but with na.action = "na.impute"
airq.obj <- rfsrc(Ozone ~ ., data = airquality, na.action = "na.impute")
print(airq.obj)
print(predict(airq.obj))
## ... also equivalent to outcome="test" but na.action = "na.impute" required
print(predict(airq.obj, airquality, outcome = "test", na.action = "na.impute"))

## classification example
iris.obj <- rfsrc(Species ~., data = iris)
print(iris.obj)
print(predict.rfsrc(iris.obj, iris, outcome = "test"))

## ------------------------------------------------------------
## another example illustrating outcome = "test"
## unique way to check reproducibility of the forest
## ------------------------------------------------------------

## training step
set.seed(542899)
data(pbc, package = "randomForestSRC")
train <- sample(1:nrow(pbc), round(nrow(pbc) * 0.50))
pbc.out <- rfsrc(Surv(days, status) ~ .,  data=pbc[train, ])

## standard prediction call
pbc.train <- predict(pbc.out, pbc[-train, ], outcome = "train")
##non-standard predict call: overlays the test data on the grow forest
pbc.test <- predict(pbc.out, pbc[-train, ], outcome = "test")

## check forest reproducibilility by comparing "test" predicted survival
## curves to "train" predicted survival curves for the first 3 individuals
Time <- pbc.out$time.interest
matplot(Time, t(pbc.train$survival[1:3,]), ylab = "Survival", col = 1, type = "l")
matlines(Time, t(pbc.test$survival[1:3,]), col = 2)

## ------------------------------------------------------------
## multivariate forest example
## ------------------------------------------------------------

## train the forest
trn <- 1:20
o <- rfsrc(cbind(mpg, disp)~.,mtcars[trn,])

## print training results for each outcome
print(o, outcome.target="mpg")
print(o, outcome.target="disp")

## print test results for each outcome
p <- predict(o, mtcars[-trn,])
print(p, outcome.target="mpg")
print(p, outcome.target="disp")


}
}
\keyword{predict}
\keyword{forest}

\name{plot.survival.rfsrc}
\alias{plot.survival.rfsrc}
\alias{plot.survival}
\alias{get.brier.survival}
\alias{get.auct.survival}
\alias{plotBrierAUC}
\title{Survival Prediction Performance and Diagnostic Plots}
\description{
  Plot survival estimates and calculate inverse-probability-of-censoring
  weighted Brier scores and cumulative/dynamic time-dependent AUC.
}
\usage{
\method{plot.survival}{rfsrc}(x, show.plots = TRUE, subset,
  collapse = FALSE, cens.model = c("km", "rfsrc"), ...)

get.brier.survival(o, subset = NULL,
  cens.model = c("km", "rfsrc"), papply = lapply,
  times = NULL, conf.int = FALSE, keep.matrix = TRUE)

get.auct.survival(o, subset = NULL,
  cens.model = c("km", "rfsrc"), papply = lapply,
  times = NULL, conf.int = FALSE)

plotBrierAUC(x, subset = NULL,
  cens.model = c("km", "rfsrc"), papply = lapply,
  times = NULL, conf.int = TRUE,
  plots = c("brier", "auct"), show.plots = TRUE, ...)
}
\arguments{
  \item{x, o}{An object of class \code{(rfsrc, grow)} or
    \code{(rfsrc, predict)}. The two numerical helpers and
    \code{plotBrierAUC} also accept an object of class
    \code{(rfsrc, forest)}.}

  \item{show.plots}{Should plots be displayed?}

  \item{subset}{Vector indicating which predicted cases are to be used.
    All cases are used if not specified.}

  \item{collapse}{Collapse the individual survival functions into their
    ensemble mean?}

  \item{cens.model}{Method used to estimate the censoring distribution
    for inverse probability of censoring weighting. In all cases the
    censoring model is estimated from the full grow data, not from the
    evaluation outcomes:

    \describe{
      \item{\code{km}:}{Kaplan--Meier estimator.}
      \item{\code{rfsrc}:}{Random survival forest estimator of the
        conditional censoring distribution. When censoring is present,
        this option requires the grow and evaluation covariates to be stored
        in the object. Because
        \code{rfsrc.fast} uses \code{forest=FALSE} by default, use
        \code{cens.model="km"} for its default reduced object or refit with
        \code{forest=TRUE}.}
    }
  }

  \item{papply}{Function used in place of \code{lapply} for repeated
    calculations.}

  \item{times}{Optional numeric vector of prediction horizons. The
    forest survival curves are evaluated as right-continuous step
    functions. The forest time grid is used by default.}

  \item{conf.int}{Controls pointwise confidence intervals. Use
    \code{FALSE} for no intervals, \code{TRUE} for 95 percent intervals,
    or a numeric value strictly between zero and one for a different
    confidence level.}

  \item{keep.matrix}{Should the subject-by-time matrix of IPCW Brier
    contributions be retained? This matrix is used by
    \code{plot.survival} for mortality-stratified curves.}

  \item{plots}{One or both of \code{"brier"} and \code{"auct"}.
    \code{"auc"} is also accepted as an alias for \code{"auct"}.}

  \item{...}{Further graphical arguments. For \code{plotBrierAUC},
    \code{col}, \code{lwd}, \code{lty}, and \code{type} control
    the estimated curve; \code{band.col} and \code{band.border} control
    the confidence band.}
}
\details{
  \code{plot.survival} produces the following plots, going from top to
  bottom and left to right:

  \enumerate{
    \item Forest estimated survival function for each individual. The
    thick red line is the ensemble mean survival and the thick green
    line is the marginal Nelson--Aalen survival estimate.

    \item Brier score stratified by ensemble mortality. Stratification
    is into four groups corresponding to the 0--25, 25--50, 50--75 and
    75--100 percentile ranges of mortality. The red line is the overall
    Brier score.

    \item Cumulative integrated Brier score divided by elapsed time,
    labeled CRPS in the plot.

    \item Mortality (Ishwaran et al., 2008) versus observed time. Blue
    points are events and black points are censored observations.
    Mortality is estimated risk calibrated to the scale of the number of
    events. For example, a mortality value of 100 means that if all
    individuals had the same covariate values, an average of 100 events
    would be expected.

  }

  Whenever possible, out-of-bag predictions are used. For a prediction
  object, the predicted test survival curves and test outcomes are used
  for scoring, while the censoring distribution is still estimated from
  the grow data. If a prediction object contains no outcomes,
  \code{plot.survival} displays only its predicted survival curves.
  Brier score, AUC, CRPS, and mortality-versus-time cannot be calculated
  without evaluation outcomes. Direct calls to \code{get.brier.survival},
  \code{get.auct.survival}, or \code{plotBrierAUC} therefore require
  outcomes for the cases being evaluated.

  For subject \eqn{i} at horizon \eqn{t}, the IPCW Brier contribution is
  \deqn{
    L_i(t) =
    \frac{I(T_i \le t, \Delta_i > 0)}{\widehat G(T_i-)}
       \widehat S_i(t)^2
    +
    \frac{I(T_i > t)}{\widehat G(t)}
       \{1-\widehat S_i(t)\}^2,
  }
  with subject-specific values of \eqn{\widehat G} when
  \code{cens.model="rfsrc"}. The Brier score is the sample mean of these
  contributions.

  The time-dependent AUC uses a cumulative/dynamic definition. Cases
  at \eqn{t} are observed events satisfying \eqn{T_i \le t}; controls
  satisfy \eqn{T_i > t}; observations censored at or before \eqn{t} are
  not included in the case-control comparison. The time-specific risk
  score is \eqn{1-\widehat S_i(t)}, and comparisons are weighted by the
  same grow-data censoring distribution.

  When confidence intervals are requested, the Brier standard error is
  obtained by deleting one subject-level IPCW loss at a time while holding
  the survival predictions and censoring weights fixed. 

  The AUC standard error uses a stratified delete-one jackknife. One observed
  case or control is deleted at a time, the remaining IPCW weights within
  that stratum are renormalized, and the resulting case-delete and
  control-delete variance components are added. With equal weights this
  calculation reduces exactly to the ordinary DeLong variance. The
  effective case and control sample sizes and the largest normalized IPCW
  weights are returned to diagnose horizons at which a few observations
  dominate the weighted comparison.

  Both are pointwise conditional standard errors, and the reported
  intervals are pointwise normal approximations rather than simultaneous
  bands. The survival predictions and the estimated censoring distribution
  are treated as fixed. They do not account for uncertainty from training
  the survival forest or estimating the censoring model. For an independent
  prediction sample, the intervals have a direct conditional test-performance
  interpretation. For a grow object, shared out-of-bag fits and the shared
  censoring estimate induce dependence among subject-level quantities, so
  the intervals should be viewed as fixed-fit working intervals rather than
  repeated-training confidence intervals. A score and its interval are
  returned as \code{NA} at a horizon where a required censoring survival
  probability is zero. The AUC standard error is also \code{NA} if fewer
  than two cases or two controls are available, or if deleting one subject
  leaves no positive IPCW weight in its stratum.

  \code{plotBrierAUC} is a base-R plotting helper. It calls a shared
  numerical engine so that Brier score and AUC use the same fitted
  censoring model, then adds the requested pointwise confidence bands.

  Only right-censored survival families are supported. Competing-risk
  analyses should use \command{plot.competing.risk}.


}
\value{
  \code{get.brier.survival} returns a list containing
  \code{brier.score}, the optional \code{brier.matx}, integrated scores
  \code{crps} and \code{crps.std}, the estimated censoring distribution,
  aligned grow and evaluation event information, survival predictions,
  mortality, and the selected subset. When confidence intervals are
  requested, \code{brier.score} also contains \code{std.err},
  \code{lower}, \code{upper}, and \code{n.eval}.

  \code{get.auct.survival} returns a corresponding list whose
  \code{auct.score} contains \code{time}, \code{auct}, \code{n.case},
  \code{n.control}, effective sample sizes \code{n.case.eff} and
  \code{n.control.eff}, and largest normalized weights
  \code{max.case.weight} and \code{max.control.weight}. The columns
  \code{std.err}, \code{lower}, and \code{upper} are included when
  confidence intervals are requested.

  \code{plotBrierAUC} invisibly returns a list containing the requested
  Brier and/or AUC result objects.

  With evaluation outcomes, \code{plot.survival} invisibly returns the
  mortality-stratified and overall Brier and cumulative integrated Brier
  curves. Without evaluation outcomes, it invisibly returns the predicted
  survival curves and their ensemble mean.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Gerds T.A. and Schumacher M. (2006). Consistent estimation of the
  expected Brier score in general survival models with right-censored
  event times, \emph{Biometrical Journal}, 48:1029--1040.

  Graf E., Schmoor C., Sauerbrei W. and Schumacher M. (1999).
  Assessment and comparison of prognostic classification schemes for
  survival data, \emph{Statistics in Medicine}, 18:2529--2545.

  DeLong E.R., DeLong D.M. and Clarke-Pearson D.L. (1988). Comparing the
  areas under two or more correlated receiver operating characteristic
  curves: a nonparametric approach, \emph{Biometrics}, 44:837--845.

  Efron B. and Tibshirani R.J. (1993). \emph{An Introduction to the
  Bootstrap}. Chapman and Hall, New York.

  Heagerty P.J. and Zheng Y. (2005). Survival model predictive accuracy
  and ROC curves, \emph{Biometrics}, 61:92--105.

  Ishwaran H. and Kogalur U.B. (2007). Random survival forests for R,
  \emph{R News}, 7(2):25--31.

  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S. (2008).
  Random survival forests, \emph{Annals of Applied Statistics},
  2:841--860.
}
\seealso{
  \command{\link{plot.competing.risk.rfsrc}},
  \command{\link{predict.rfsrc}},
  \command{\link{rfsrc}}
}
\examples{
\donttest{
## veteran data
data(veteran, package = "randomForestSRC")
plot.survival(rfsrc(Surv(time, status) ~ ., veteran),
              cens.model = "rfsrc")

## pbc data
data(pbc, package = "randomForestSRC")
pbc.obj <- rfsrc(Surv(days, status) ~ ., pbc)

## standard survival diagnostics
plot.survival(pbc.obj)
plot.survival(pbc.obj, subset = c(3, 10), collapse = TRUE)

## Brier and AUCT helpers with pointwise intervals
brier.obj <- get.brier.survival(pbc.obj, conf.int = TRUE)
head(brier.obj$brier.score)
auct.obj <- get.auct.survival(pbc.obj, conf.int = TRUE)
head(auct.obj$auct.score)

## compare two grow-data censoring models
brier.km <- get.brier.survival(pbc.obj, cens.model = "km")
brier.rf <- get.brier.survival(pbc.obj, cens.model = "rfsrc")
plot(brier.km$brier.score$time,
     brier.km$brier.score$brier.score, type = "s", col = 2,
     xlab = "Time", ylab = "Brier Score")
lines(brier.rf$brier.score$time,
      brier.rf$brier.score$brier.score, type = "s", col = 4)
legend("bottomright",
       legend = c("cens.model = km", "cens.model = rfsrc"),
       col = c(2, 4), lty = 1)

## Brier and AUCT curves with confidence bands
plotBrierAUC(pbc.obj)
plotBrierAUC(pbc.obj, plots = "auct", conf.int = 0.90)
plotBrierAUC(pbc.obj, plots = "brier", conf.int = 0.90)



}
}
\keyword{plot}
\keyword{survival}

\name{randomForestSRC-package}
\alias{randomForestSRC-package}
\docType{package}

\title{
  Fast Unified Random Forests for Survival, Regression, and Classification (RF-SRC)
}

\description{

Fast OpenMP-parallel implementation of Breiman's random forests
(Breiman, 2001) for regression, classification, survival analysis
(Ishwaran, 2008), competing risks (Ishwaran, 2012), multivariate
outcomes (Segal and Xiao, 2011), unsupervised learning (Mantero and
Ishwaran, 2020), quantile regression (Meinshausen, 2006; Zhang et al.,
2019; Greenwald and Khanna, 2001), and imbalanced q-classification
(O'Brien and Ishwaran, 2019).


Supports deterministic and randomized splitting rules (Geurts et al.,
2006; Ishwaran, 2015) across all families. Variable importance (VIMP),
holdout VIMP, and confidence regions (Ishwaran and Lu, 2019) can be
computed for single and grouped variables. Includes minimal depth
variable selection (Ishwaran et al., 2010, 2011) and a fast interface
for missing data imputation using multiple forest-based methods (Tang
and Ishwaran, 2017).


Tree structures can be visualized in Safari or Chrome for any family;
see \code{\link{get.tree}}.

}

\section{Package Overview}{

This package contains many useful functions. Users are encouraged to read the help files in full for detailed guidance. Below is a brief overview of key functions to help navigate the package.

\enumerate{

  \item \code{\link{rfsrc}}

  The main entry point to the package. Builds a random forest using user-supplied training data. The returned object is of class \code{(rfsrc, grow)}.

  \item \code{\link{rfsrc.fast}}

  A computationally efficient version of \code{rfsrc} using subsampling.

  \item \code{\link{quantreg.rfsrc}}, \code{\link{quantreg}}

  Univariate and multivariate quantile regression forests for training and testing. Includes methods such as the Greenwald-Khanna (2001) algorithm, ideal for large data due to its memory efficiency.

  \item \code{\link{predict.rfsrc}}, \code{predict}

  Predicts outcomes by dropping test data down the trained forest. Returns an object of class \code{(rfsrc, predict)}.

  \item \code{\link{sidClustering.rfsrc}}, \code{sidClustering}

  Unsupervised clustering using SID (Staggered Interaction Data). Also includes Breiman's artificial two-class method (Breiman, 2003).

  \item \code{\link{vimp}}, \code{\link{subsample}}, \code{\link{holdout.vimp}}

  Functions for variable selection and importance assessment:
  \enumerate{
    \item \code{vimp}: Computes variable importance (VIMP) by perturbing each variable (e.g., via permutation). Can also be computed directly in \code{rfsrc} and \code{predict.rfsrc}.
    
    \item \code{subsample}: Computes confidence intervals for VIMP using subsampling.
    
    \item \code{holdout.vimp}: Measures the effect of removing a variable from the model.
    
    \item \strong{VarPro} (\pkg{VarPro} package): For advanced model-independent variable selection using rule-based variable priority. Supports regression, classification, survival, and unsupervised data. See \url{https://www.varprotools.org}.
  }

  \item \code{\link{imbalanced.rfsrc}}, \code{\link{imbalanced}}

  Implements q-classification and G-mean-based VIMP for class-imbalanced data.

  \item \code{\link{impute.rfsrc}}, \code{\link{impute}}

  A fast interface for missing data imputation. While \code{rfsrc} and \code{predict.rfsrc} can handle missing data internally, this provides a dedicated, efficient solution for imputation tasks.

  \item \code{\link{partial.rfsrc}}, \code{\link{partial}}

  Computes partial dependence functions to assess the marginal effect of one or more variables on the forest ensemble.

}
}

\section{Home page, Vignettes, Discussions, Bug Reporting, Source Code, Beta Builds}{

\enumerate{

  \item The package home page, with vignettes, manuals, GitHub links, and additional documentation, is available at:
  \url{https://www.randomforestsrc.org/index.html}

  \item Questions, comments, and general usage discussions (non-bug-related) can be posted at:
  \url{https://github.com/kogalur/randomForestSRC/discussions/}

  \item Bug reports should be submitted at:
  \url{https://github.com/kogalur/randomForestSRC/issues/}

  Please use this only for bugs, and include the following with your report:
  \itemize{
    \item Output from \code{sessionInfo()}.
    \item A minimal reproducible example including:
      \itemize{
        \item A minimal dataset required to reproduce the error.
        \item The smallest runnable code needed to reproduce the issue.
        \item Version details of R and all relevant packages.
        \item A random seed (via \code{set.seed()}) if randomness is involved.
      }
  }

  \item The latest stable release of the package is available on CRAN:
  \url{https://cran.r-project.org/package=randomForestSRC/}

  \item Development builds (unstable) with bug fixes and new features are hosted on GitHub:
  \url{https://github.com/kogalur/randomForestSRC/}
}
}

\section{OpenMP Parallel Processing -- Installation}{

This package supports OpenMP shared-memory parallel programming on systems where the architecture and operating system permit it. OpenMP is enabled by default.

Detailed instructions for configuring OpenMP parallel processing can be found at:
\url{https://www.randomforestsrc.org/articles/installation.html}

Note that running the package with OpenMP (or Open MPI) may increase memory (RAM) usage. Users are advised to understand their system's hardware limits and to monitor resource consumption to avoid overtaxing CPU and memory capacity.
}

\section{Reproducibility}{

Model reproducibility is determined by three components: the random seed, the forest topology (i.e., the structure of trees), and terminal node membership for the training data. These elements together allow the model and its terminal node statistics to be faithfully restored.

Other outputs, such as variable importance (VIMP) and performance metrics, rely on additional internal randomization and are not considered part of the model definition. As a result, such statistics are subject to Monte Carlo variability and may differ across runs, even with the same seed.
}


\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Breiman L. (2001). Random forests, \emph{Machine Learning}, 45:5-32.

  Geurts, P., Ernst, D. and Wehenkel, L., (2006). Extremely randomized
  trees. \emph{Machine learning}, 63(1):3-42.

  Greenwald M. and Khanna S. (2001).  Space-efficient online computation of
  quantile summaries. \emph{Proceedings of ACM SIGMOD}, 30(2):58-66.

  Ishwaran H. and Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
  
  Ishwaran H. (2007).  Variable importance in binary regression
  trees and forests,  \emph{Electronic J. Statist.}, 1:519-537.

  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S.
  (2008).  Random survival forests, \emph{Ann. App.
    Statist.}, 2:841-860.
  
  Ishwaran H., Kogalur U.B., Gorodeski E.Z, Minn A.J. and
  Lauer M.S. (2010).  High-dimensional variable selection for survival
  data.  \emph{J. Amer. Statist. Assoc.}, 105:205-217.

  Ishwaran H., Kogalur U.B., Chen X. and Minn A.J. (2011). Random survival
  forests for high-dimensional data. \emph{Stat. Anal. Data Mining}, 4:115-132

  Ishwaran H., Gerds T.A., Kogalur U.B., Moore R.D., Gange S.J. and Lau
  B.M. (2014). Random survival forests for competing risks.
  \emph{Biostatistics}, 15(4):757-773.

  Ishwaran H. and Malley J.D. (2014). Synthetic learning
  machines. \emph{BioData Mining}, 7:28.
  
  Ishwaran H. (2015).  The effect of splitting on random forests.
  \emph{Machine Learning}, 99:75-118.
  
  Ishwaran H. and Lu M.  (2019).  Standard errors and confidence
  intervals for variable importance in random forest regression,
  classification, and survival. \emph{Statistics in Medicine}, 38,
  558-582.

  Lu M., Sadiq S., Feaster D.J. and Ishwaran H. (2018). Estimating
  individual treatment effect in observational data using random forest
  methods. \emph{J. Comp. Graph. Statist}, 27(1), 209-219

  Mantero A. and Ishwaran H. (2021).  Unsupervised random forests.
  \emph{Statistical Analysis and Data Mining}, 14(2):144-167.

  Meinshausen N. (2006) Quantile regression forests, \emph{Journal of
  Machine Learning Research}, 7:983-999.

  O'Brien R. and Ishwaran H. (2019).  A random forests quantile
  classifier for class imbalanced data. \emph{Pattern Recognition},
  90, 232-249

  Segal M.R. and Xiao Y. Multivariate random
  forests. (2011). \emph{Wiley Interdisciplinary Reviews: Data Mining
  and Knowledge Discovery}. 1(1):80-87.

  Tang F. and Ishwaran H. (2017).  Random forest missing data
  algorithms.  \emph{Statistical Analysis and Data Mining}, 10:363-377.

  Zhang H., Zimmerman J., Nettleton D. and Nordman D.J. (2019).  Random
  forest prediction intervals. \emph{The American Statistician}. 4:1-5.

  
}
\seealso{

  \command{\link{get.tree.rfsrc}},

  \command{\link{holdout.vimp.rfsrc}},

  \command{\link{imbalanced.rfsrc}},
  \command{\link{impute.rfsrc}},

  \command{\link{max.subtree.rfsrc}},

  \command{\link{partial.rfsrc}},
  \command{\link{plot.competing.risk.rfsrc}},
  \command{\link{plot.rfsrc}},
  \command{\link{plot.survival.rfsrc}},
  \command{\link{plot.variable.rfsrc}},
  \command{\link{predict.rfsrc}},
  \command{\link{print.rfsrc}},

  \command{\link{quantreg.rfsrc}},

  \command{\link{rfsrc}},
  \command{\link{rfsrc.cart}},
  \command{\link{rfsrc.fast}},

  \command{\link{sidClustering.rfsrc}},

  \command{\link{subsample.rfsrc}},
  
  \command{\link{tune.rfsrc}},

  \command{\link{vimp.rfsrc}}
}
\keyword{package}

\name{impute.learn.rfsrc}
\alias{impute.learn.rfsrc}
\alias{impute.learn}
\alias{save.impute.learn.rfsrc}
\alias{save.impute.learn}
\alias{load.impute.learn.rfsrc}
\alias{load.impute.learn}
\alias{predict.impute.learn.rfsrc}
\alias{predict.impute.learn}
\alias{impute.ood.rfsrc}
\alias{impute.ood}
\alias{print.impute.learn.rfsrc}
\alias{print.impute.learn}
\title{Learn a predictive imputer for test-time imputation and OOD scoring}
\description{
  Learns a predictive imputer from training data for later use on new data.

  If the training data contain missing values, the function first
  imputes them using \code{impute}. It then fits one saved full-sweep
  learner per selected target on the completed training data and reuses
  those learners later to update missing values in new data without
  refitting on the test set.

  The same saved learner bank can also be used to score new data for
  out-of-distribution (OOD) behavior. Note that OOD scores are available
  even when new data have missing values. Each selected target is
  reconstructed from its saved conditional learner and compared with the
  observed value. Target-wise discrepancies are calibrated against a
  training reference calculated from out-of-bag predictions computed
  during training.

  If the training data are complete and \code{target.mode = "all"},
  the initial training-data imputation step is skipped and the
  full-sweep learners are fit directly from the complete training data.
  
  If \code{supervised.formula} is supplied, the function also fits an
  internal supervised forest from the training data.  The supervised
  forest plays no role in the training imputation but is used downstream
  to create auxiliary variables which are appended as predictors which
  can improve test-time imputation and OOD scoring. Users who want only
  the basic unsupervised learned imputer and do not want test time
  contamination from outcomes should leave \code{supervised.formula}
  unspecified.
  
}
\usage{
impute.learn.rfsrc(formula, data,
  ntree = 100, nodesize = 1, nsplit = 10,
  nimpute = 2, fast = FALSE, blocks,
  mf.q, max.iter = 10, eps = 0.01,
  ytry = NULL, always.use = NULL, verbose = TRUE,
  ...,
  supervised.formula = NULL,
  supervised.args = list(),
  full.sweep.options = list(ntree = 100, nsplit = 10),
  target.mode = c("missing.only", "all"),
  deployment.xvars = NULL,
  anonymous = TRUE,
  learner.prefix = "impute.learner.",
  learner.root = "learners",
  out.dir = NULL,
  wipe = TRUE,
  keep.models = is.null(out.dir),
  keep.ximp = FALSE,
  save.on.fit = !is.null(out.dir),
  save.ood = TRUE,
  weight = NULL)

save.impute.learn.rfsrc(object, path, wipe = TRUE, verbose = TRUE)

load.impute.learn.rfsrc(path, targets = NULL, lazy = TRUE, verbose = TRUE)

\method{predict}{impute.learn.rfsrc}(object, newdata,
  max.predict.iter = 3L,
  eps = 1e-3,
  targets = NULL,
  restore.integer = TRUE,
  cache.learners = c("session", "none", "all"),
  verbose = TRUE,
  ...)

impute.ood.rfsrc(object, newdata,
  targets = NULL,
  max.predict.iter = 3L,
  eps = 1e-3,
  cache.learners = c("all", "session", "none"),
  weight = NULL,
  aggregate = c("bounded.product",  "weighted.mean",
    "weighted.lp", "weighted.lp.log", "top.k"),
  aggregate.args = list(),
  return.details = FALSE,
  return.reconstruction = FALSE,
  verbose = TRUE,
  ...)
}
\arguments{

  \item{formula}{An optional symbolic model description passed to
    \code{impute} for the initial training-data imputation when
    \code{supervised.formula} is not supplied. It follows the same
    guidelines as in \code{impute}. This formula does not specify the
    full-sweep learner bank, the OOD targets, or the supervised
    auxiliary forest. When \code{supervised.formula} is supplied, the
    raw predictor block is taken from the right-hand side of
    \code{supervised.formula}, and \code{formula} is not used.}

  \item{data}{Training data. Variables that are not real-valued are
    coerced to factors before fitting when possible; otherwise fitting
    stops with an error. Rows and columns that are entirely missing are
    dropped before training begins. If \code{supervised.formula} is
    supplied, \code{data} should contain both the supervised response
    columns and the raw predictor columns. The learned imputer is then
    built on the raw predictor block defined by the right-hand side of
    \code{supervised.formula}.}

  \item{ntree, nodesize, nsplit, nimpute, fast, blocks,
    max.iter, ytry, always.use, verbose}{Arguments passed to
    \code{impute} for the initial training-data imputation. The
    argument \code{full.sweep} is controlled internally and should not
    be supplied. The same \code{verbose}, \code{max.predict.iter},
    \code{eps}, and \code{cache.learners} controls are also used by
    \code{predict.impute.learn} and \code{impute.ood}.}

  \item{mf.q}{Controls the imputation engine used by \code{impute}. If
    \code{mf.q = 1}, training uses standard \code{missForest}. If
    \code{mf.q > 1}, training uses the multivariate \code{missForest}
    generalization. If \code{mf.q} is omitted, the training imputation
    follows the default behavior of \code{impute}.}

  \item{eps}{Convergence threshold. In \code{impute.learn} this
    controls the initial training-data imputation. In
    \code{predict.impute.learn} and \code{impute.ood} it controls early
    stopping for the prediction-time sweep.}

  \item{\dots}{For \code{impute.learn}, additional arguments passed to
    \code{impute}. For \code{predict.impute.learn} and
    \code{impute.ood}, additional arguments are currently ignored.}

  \item{supervised.formula}{Optional supervised learning formula used to
    augment the learned imputer for improved OOD detection in supervised
    settings.  The left-hand side defines the supervised response and
    the right-hand side defines the raw predictor block to be learned by
    \code{impute.learn}. The supervised forest is fit internally after
    the raw predictor block has been completed. Its training out-of-bag
    \code{predicted} values and test-time \code{predicted} values are
    appended internally as auxiliary predictors.  These auxiliary
    predictors are internal and are not expected in \code{newdata}.}
  
  \item{supervised.args}{Optional named list of arguments passed to the
    internal supervised \code{rfsrc} fit. Entries for \code{formula},
    \code{data}, and \code{forest} are controlled internally and are
    ignored if supplied. All other entries override the defaults used
    for this supervised fit, including defaults inherited from
    top-level arguments such as \code{fast} and internally supplied
    defaults such as \code{perf.type = "none"}.}

  \item{full.sweep.options}{A \code{list} of options used when fitting
    the full sweep after the training data have been imputed.
    Recognized entries include \code{ntree}, \code{nodesize},
    \code{nsplit}, \code{mtry}, \code{splitrule}, \code{bootstrap},
    \code{sampsize}, \code{samptype}, \code{perf.type}, \code{rfq},
    \code{save.memory}, \code{importance}, and \code{proximity}.
    Unknown entries are ignored with a warning.}

  \item{target.mode}{Determines which raw variables receive a saved
    full-sweep learner. The default \code{"missing.only"} saves
    learners only for variables that were missing in the training data.
    The option \code{"all"} saves a learner for every retained raw
    variable. If the training data are complete, \code{target.mode =
    "all"} must be used. When \code{supervised.formula} is supplied,
    the default is promoted internally to \code{"all"} so that every
    retained raw predictor can later be updated and scored. For the
    broadest OOD coverage, \code{target.mode = "all"} is recommended so
    every deployment-time variable can be reconstructed.}

  \item{deployment.xvars}{Controls which raw predictors are assumed to
    be available later when the saved imputer is used on new data. If
    \code{NULL}, all raw non-target columns are used. If a character
    vector, the same raw predictor set is used for all targets. If a
    named list, names must be target variables and each entry gives the
    raw predictor set for that target; targets omitted from the list
    fall back to all raw non-target columns. Unnamed or duplicated list
    entries are not allowed. Entries named for non-target variables, and
    predictor names not found in the training data, are ignored with a
    warning. When \code{supervised.formula} is supplied, the internal
    auxiliary \code{predicted.*} columns are added automatically on top
    of these raw predictor sets. In other words,
    \code{deployment.xvars} restricts the raw predictors only.}

  \item{anonymous}{If \code{TRUE}, uses \code{rfsrc.anonymous} when
    fitting the saved target-wise full-sweep learner bank. The internal
    supervised forest, when requested through \code{supervised.formula},
    is fit with \code{rfsrc} because it must be retained for later
    prediction-time auxiliary variables. Thus \code{anonymous} usually
    reduces the size of the target-wise learner bank, but not
    necessarily the size of the supervised forest.}

  \item{learner.prefix, learner.root}{Names used when writing saved
    full-sweep learners to disk. If \code{supervised.formula} is
    supplied, the internal supervised forest is saved under the same
    learner root.}

  \item{out.dir}{Optional output directory. If supplied and
    \code{save.on.fit = TRUE}, the manifest and the saved full-sweep
    learners are written to this directory during fitting. This requires
    the \pkg{fst} package because learners are serialized with
    \code{fast.save}. If \code{supervised.formula} is supplied, the
    internal supervised forest is also saved there.}

  \item{wipe}{If \code{TRUE}, removes an existing output directory
    before writing a new one.}

  \item{keep.models}{If \code{TRUE}, keeps the fitted full-sweep
    learners in memory in the returned object. At least one storage mode
    must be enabled: either \code{keep.models = TRUE} or
    \code{out.dir} with \code{save.on.fit = TRUE}. If
    \code{supervised.formula} is supplied, the internal supervised
    forest is kept in memory under the same rule.}

  \item{keep.ximp}{If \code{TRUE}, keeps the completed raw training
    predictor table in the returned object. This is not required for
    later prediction.}

  \item{save.on.fit}{If \code{TRUE} and \code{out.dir} is supplied,
    writes the imputer to disk during fitting.}

  \item{save.ood}{If \code{TRUE}, computes and stores an OOD reference.
    The reference is built during training from target-wise out-of-bag
    reconstruction discrepancies, their target-wise calibrated training
    scores, and a default row-level weighted mean using the saved OOD
    target weights. If no \code{weight} is supplied at fit time, equal
    target weights are used. The saved target-wise training-score matrix
    allows \code{impute.ood} to rebuild a calibrated row-level
    percentile for arbitrary target subsets, test-time weight overrides,
    and alternate row aggregates besides the weighted mean.}

  \item{object}{An object returned by \code{impute.learn} or
    \code{load.impute.learn}.}

  \item{path}{Directory containing a saved imputer. Save and load
    operations require the \pkg{fst} package because learners are read
    and written with \code{fast.save} and \code{fast.load}.}

  \item{targets}{Optional subset of target variables to load, update,
    or score. Unknown names are ignored with a warning. For
    \code{impute.ood}, row-level percentile calibration is rebuilt for
    the requested target subset from the saved target-wise training OOD
    scores whenever those scores are available in the manifest.}

  \item{lazy}{If \code{TRUE}, saved learners are loaded only when they
    are needed. If \code{FALSE}, all saved learners are loaded at once.
    If \code{supervised.formula} was used at fit time, the internal
    supervised forest follows the same lazy versus eager loading rule.}

  \item{newdata}{New data to be imputed or scored. Missing columns are
    added and extra columns are dropped to match the training schema.
    Unseen factor levels are converted to \code{NA} for harmonization,
    but they are also tracked row-wise. In supervised mode,
    \code{newdata} should contain only the raw predictor columns learned
    by the imputer. The internal auxiliary columns are created
    automatically and should not be supplied by the user. If supervised
    response columns are supplied in \code{newdata} and are not part of
    the learned raw predictor block, they are treated as extra columns
    and dropped; they are not used to compute auxiliary predictors,
    imputations, or OOD scores. In \code{impute.ood},
    observed raw target values are used to compute target-wise
    discrepancies. Missing raw target values are completed for
    predictor-side reconstruction but do not themselves contribute a
    finite target-wise OOD score. Any row containing an unseen factor
    level is flagged and its row-level OOD score is set to the maximum
    value.}

  \item{max.predict.iter}{Maximum number of full-sweep passes applied to
    \code{newdata} before the saved learner bank is used for OOD
    reconstruction or returned prediction-time imputations.}

  \item{restore.integer}{If \code{TRUE}, integer-supported variables
    in the returned data are restored to the integer grid. This includes
    columns stored as R integers and numeric columns whose observed
    finite training values were all integer-valued. Numeric columns
    remain numeric; returned values are rounded to integer support.
    Columns stored as R integers are returned as integer vectors.
    Factor columns are always conformed back to the training schema.
    The package operates on real-valued and factor variables;
    inputs that are not real-valued are coerced to factors during
    preprocessing when possible, otherwise an error is raised.}

  \item{cache.learners}{How saved learners are reused during
    prediction or OOD scoring. For \code{predict.impute.learn}, the
    default \code{"session"} loads each needed learner once per call.
    The option \code{"none"} reloads a learner every time it is needed.
    The option \code{"all"} loads all requested learners before work
    starts. For \code{impute.ood}, \code{"all"} is the default because
    the saved learner bank is typically reused once for predictor-side
    completion and again for target reconstruction.}

  \item{weight}{Optional nonnegative target weights used for row-level OOD
    aggregation. In \code{impute.learn}, these weights define the
    default row-level OOD weighting scheme stored in the manifest. In
    \code{impute.ood}, they define the active row-level weighting scheme
    for the current scoring call. If supplied as a named vector, entries
    are matched to targets by name; omitted targets are set to zero, and
    extra names are ignored. If omitted in \code{impute.ood}, the saved
    training-time OOD weights are used automatically. Because the fit
    stores target-wise training OOD scores, \code{score.percentile} can
    be recalibrated automatically for test-time weight overrides rather
    than being limited to the original training-time weights.}

  \item{aggregate}{Row-level aggregation metric used by
    \code{impute.ood} to combine calibrated target-wise OOD scores. The
    default \code{"bounded.product"} applies a weighted product of
    the form \eqn{1-\prod_j (1-u_j+\varepsilon)^{\tilde w_j}} where
    \eqn{u_j} are row level calibrated scores for target feature
    \eqn{j}. \code{"weighted.mean"} is the weighted
    average. \code{"weighted.lp"} applies a weighted Minkowski \eqn{L_p}
    aggregation to the calibrated target scores.
    \code{"weighted.lp.log"} first applies the tail-stretching transform
    \eqn{-\log(1-u_j+\varepsilon)} to each calibrated target score
    \eqn{u_j} and then applies the weighted \eqn{L_p}
    aggregation. \code{"top.k"} averages only the \eqn{k} largest
    scoreable target scores among the positively weighted targets. When
    the fit stores target-wise training OOD scores,
    \code{score.percentile} is rebuilt for the requested aggregate as
    well as the requested targets and weights.}

  \item{aggregate.args}{Optional list of tuning arguments for
    \code{aggregate}. Recognized entries are \code{p} for
    \code{"weighted.lp"} and \code{"weighted.lp.log"}, \code{k}
    (or \code{top.k}) for \code{"top.k"}, and \code{eps} for
    \code{"weighted.lp.log"} and \code{"bounded.product"}. The
    default values are \code{p = 2}, \code{k = 1}, and
    \code{eps = 1e-12}. Unknown entries are ignored with a warning.}

  \item{return.details}{If \code{TRUE}, \code{impute.ood} returns the
    per-target discrepancy and calibrated target-score matrices and a
    richer \code{info} list.}
  
  \item{return.reconstruction}{If \code{TRUE}, \code{impute.ood} returns
    the target-wise reconstructed values used during OOD scoring.  The
    reconstructed values are returned on the raw target scale, with one
    column per scored target.  Typically these values are computed
    internally and discarded.}


}
\details{
  A predictive imputer is calculated in two stages.

  The training data are first processed. Variables that are not
  real-valued are coerced to factors when possible; otherwise fitting
  stops with an error. Rows and columns that are entirely missing are
  removed before the training schema is stored.

  If the resulting training data contain missing values, the first stage
  uses \code{impute} as the imputation engine to complete the training
  data.  Options are specified exactly as in \code{impute}.  In
  particular, \code{mf.q = 1} gives standard \code{missForest},
  \code{mf.q > 1} gives the multivariate \code{missForest}
  generalization, and if \code{mf.q} is omitted, on the fly imputation
  is used if \code{formula} is specified, otherwise default unsupervised
  imputation is used. If the training data are already complete and
  \code{target.mode = "all"}, this initial imputation step is skipped.

  In the second stage, a full forest sweep is fit on the completed
  training data. For each target selected by \code{target.mode}, rows
  where that target was observed are used to fit a forest with that
  target on the left-hand side and the selected deployment predictors on
  the right-hand side are specified by \code{deployment.xvars}.

  By default, \code{deployment.xvars = NULL} allows every non-target
  column to be used as a predictor. This is convenient, but it can also
  introduce leakage if the training data include outcomes, future-only
  variables, identifiers, or any fields that will not be available when
  the learned imputer is applied to new data. Restrict
  \code{deployment.xvars} when that is a concern.

  If \code{supervised.formula} is supplied, the function also fits an
  internal supervised forest from the training data. Out-of-bag
  \code{predicted} values for the training sample, and \code{predicted}
  values for new data, are appended internally as auxiliary
  predictors. These auxiliary variables are not used as imputation
  targets, however, they are available as predictors for the saved
  forests and can therefore influence both test-time imputation and OOD
  scoring. Users who want only the basic unsupervised learned imputer
  should leave \code{supervised.formula} unspecified.
  
  When \code{supervised.formula} is used, \code{deployment.xvars}
  continues to restrict only the raw predictors. The internally created
  auxiliary variables are added automatically to the predictor sets for
  the saved raw targets. This lets the learned imputer benefit from
  supervised signal without asking the user to prepare and pass those
  auxiliary columns manually.

  When the imputer is saved to disk, each full-sweep learner is written
  separately using \code{fast.save}. Loading uses \code{fast.load}. In
  practice this gives a small manifest plus a directory of saved
  learners. The \pkg{fst} package is therefore required for save and
  load operations. The explicit save method can write learners either
  from memory or by reloading them from an attached saved path. If
  \code{supervised.formula} is used, the internal supervised forest is
  saved alongside the target-wise learner bank.

  Prediction starts by matching \code{newdata} to the training schema,
  filling missing values with training means or modes, and then applying
  one or more full-sweep passes. Only the targets selected by
  \code{target.mode} are updated by saved learners. In supervised mode,
  this same test-time preparation first initializes the raw predictor
  table, then computes the internal auxiliary variables from the saved
  supervised forest, and then runs the forest test-time sweep.
  Supervised response columns, if present in \code{newdata} but not in
  the learned raw predictor block, are dropped before this step and play
  no role at prediction or scoring time.

  If \code{target.mode = "missing.only"}, a variable that was complete
  in training but missing in new data is initialized from the training
  fit but does not receive a model-based update. Use
  \code{target.mode = "all"} if missing values may appear later in any
  raw variable. Complete training data also require
  \code{target.mode = "all"}, because otherwise there are no missing
  variables from which to determine the saved targets.

  If \code{save.ood = TRUE}, the fit also stores an OOD reference in the
  manifest. For each saved target, the out-of-bag prediction from the
  fitted learner is compared with the observed training value to form a
  target-wise reconstruction discrepancy. Continuous and integer targets
  use absolute reconstruction error. Factor targets use negative log
  predictive probabilities.
  
  The row-level OOD calibration stored at fit time is built by
  aggregating the target-wise training scores with a weighted mean using
  \code{weight}. If \code{weight} is omitted at fit time, all saved OOD
  targets receive weight 1. If a named vector is supplied, entries are
  matched by target name, omitted saved OOD targets receive weight 0,
  and the resulting weighting scheme is carried forward for later
  deployment-time scoring.

  \code{impute.ood} first completes the predictor side of \code{newdata}
  using the same harmonization, initialization, and iterative sweep
  logic used by \code{predict.impute.learn}. It then reconstructs each
  requested raw target directly from its saved learner and compares the
  reconstruction with the observed raw target value supplied in
  \code{newdata}. If a scored raw target is missing in a row, that
  target does not contribute to that row's OOD score. Raw target
  discrepancies are converted to target-wise OOD scores using the saved
  target-specific training references. In supervised mode, auxiliary
  variables participate as predictors in the reconstruction, but the
  supervised response itself is not used at deployment time.

  The row-level OOD score combines those calibrated target-wise scores
  over the targets that are both observed and scoreable for that row. By
  default, \code{impute.ood} uses a bounded product rule, but the row
  aggregate can be changed to weighted mean, a weighted \eqn{L_p} rule,
  a log-tail weighted \eqn{L_p} rule, or a top-\eqn{k} rule. This makes
  it possible to explore row scores that are more sensitive to sparse
  but severe coordinate shifts. By default, \code{impute.ood} reuses
  the same OOD weights saved during \code{impute.learn}, so a pipeline
  can fix its weighting scheme once upstream and carry it forward
  automatically.

  A second component, \code{score.percentile}, is obtained by rebuilding
  the row-level training reference from the saved target-wise training
  OOD scores using the requested target subset, the active weight
  vector, and the active row aggregate. This means percentile
  calibration remains available when the user leaves the saved weights
  in place, overrides them at test time, scores only a subset of the
  saved OOD targets, or experiments with alternate row aggregates.

  Unseen factor levels are tracked row-wise during harmonization.
  Because such values are immediate anomalies relative to the training
  schema, \code{impute.ood} flags those rows and assigns them the
  maximum row-level score. If the unseen level occurs in a scored target
  itself, the corresponding target-level discrepancy is also treated as
  maximal.
}
\value{
  \code{impute.learn} returns an object of class
  \code{c("impute.learn.rfsrc", "impute.learn")}. The object
  contains a manifest, optionally the fitted full-sweep learners,
  optionally the internal supervised forest when
  \code{supervised.formula} is used, optionally the completed raw
  training predictor table, and optionally a path to the saved imputer
  on disk. If \code{save.ood = TRUE}, the manifest also contains an
  \code{ood} component storing compact target-wise OOD references, the
  saved row-by-target training OOD score matrix used for later
  percentile recalibration, and the default OOD aggregation weights.
  When supervised mode is active, the manifest also records the
  supervised family, response names, and the internally created
  auxiliary \code{predicted.*} variable names.

  \code{load.impute.learn} returns an object of the same class.

  \code{predict.impute.learn} returns a data frame with imputed values
  overlaid on the raw predictor table. An attribute named
  \code{"impute.learn.info"} contains prediction-time diagnostics such
  as the number of sweep passes, pass-difference history, caching mode,
  disk-load counts, schema harmonization details, dropped supervised
  response columns when present, row-wise unseen-factor flags,
  supervised-auxiliary diagnostics when present, and any targets
  skipped because a learner was unavailable or a prediction failed.

  \code{impute.ood} returns an object of class
  \code{c("impute.ood.rfsrc", "impute.ood")}. It is a list with the
  following components:
  \itemize{
    \item \code{score}: the row-level aggregate of calibrated
      target-wise OOD scores under the requested \code{aggregate} and
      \code{weight}. Larger values indicate greater
      out-of-distribution behavior. For \code{aggregate = "weighted.lp.log"},
      this raw score is on a positive unbounded scale.
    \item \code{score.percentile}: the percentile of \code{score}
      relative to a row-level training reference rebuilt from the saved
      target-wise training OOD scores for the requested targets,
      weights, and row aggregate. For legacy fitted objects that do not
      contain those saved training scores, the original saved row-level
      reference is used when possible; otherwise \code{NA}.
    \item \code{targets.used}: the number of weighted targets that
      contributed to each row-level score.
    \item \code{target.score}: optional matrix of target-wise calibrated
      OOD scores, returned when \code{return.details = TRUE}.
    \item \code{target.delta}: optional matrix of raw target-wise
      reconstruction discrepancies, returned when
      \code{return.details = TRUE}.
    \item \code{info}: a list of diagnostics including harmonization
      details, dropped supervised response columns when present,
      row-wise unseen-factor flags, learner-loading information,
      supervised-auxiliary diagnostics when present, the active row
      aggregate and its arguments, whether the saved row-level
      calibration was used, and any target-specific issues.
  }
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Stekhoven D.J. and Buhlmann P. (2012). MissForest--non-parametric
  missing value imputation for mixed-type data.
  \emph{Bioinformatics}, 28(1):112--118.

  Tang F. and Ishwaran H. (2017). Random forest missing data
  algorithms. \emph{Statistical Analysis and Data Mining}, 10:363--377.
}
\seealso{
  \code{\link{impute.rfsrc}},
  \code{\link{rfsrc}},
  \code{\link{predict.rfsrc}}.
}
\examples{
## ------------------------------------------------------------
## small data example: uses missForest for impute engine
## ------------------------------------------------------------

set.seed(101)
aq <- airquality[, c("Ozone", "Solar.R", "Wind", "Temp", "Month")]
aq$Month <- factor(aq$Month)

id <- sample(1:nrow(aq), 100)
train <- aq[id, ]
test <- aq[-id, ]

## training the imputer
fit <- impute.learn(
  data = train,
  ntree = 25,
  mf.q = 1,
  max.iter = 5,
  full.sweep.options = list(ntree = 25, nsplit = 5)
)

## test time imputation
test.imp <- predict(fit, test, max.predict.iter = 2, verbose = FALSE)
print(head(test.imp))

\donttest{

## OOD scoring is most informative when every deployment-time
## variable can be reconstructed, so target.mode = "all" is recommended.
## Optional named OOD weights can also be supplied here. Any omitted
## targets receive weight 0, and the saved weights are reused
## automatically later by impute.ood().

ood.fit <- impute.learn(
  data = train,
  ntree = 25,
  mf.q = 1,
  max.iter = 5,
  target.mode = "all",
  save.ood = TRUE,
  full.sweep.options = list(ntree = 25, nsplit = 5),
  verbose = FALSE
)

ood <- impute.ood(ood.fit, test, return.details = TRUE, verbose = FALSE)
print(head(ood$score))
print(head(ood$score.percentile))

## try a more spike-sensitive row aggregate
ood.lp <- impute.ood(ood.fit, test,
                     aggregate = "weighted.lp",
                     aggregate.args = list(p = 4),
                     verbose = FALSE)
print(head(ood.lp$score.percentile))

## ------------------------------------------------------------
## supervised OOD example: regression benchmark
## the user supplies only the raw x-columns at test time
## ------------------------------------------------------------

friedman1_sim <- function(n = 150, p = 10, sigma = 1) {
  X <- matrix(runif(n * p), nrow = n)
  y <- 10 * sin(pi * X[, 1] * X[, 2]) +
       20 * (X[, 3] - 0.5)^2 +
       10 * X[, 4] + 5 * X[, 5] +
       rnorm(n, sd = sigma)
  list(X = X, y = y)
}


trn <- data.frame(friedman1_sim())
tst <- data.frame(friedman1_sim())
xvars <- setdiff(names(trn), "y")


## impute data using missForest, construct a supervised forest
## - supervised forests are used to create auxiliary variables
## - improves test time OOD in supervised problems

sup.fit <- impute.learn(
  data = trn,
  mf.q = 1,
  supervised.formula = y ~ .,
  supervised.args = list(ntree = 50, nsplit = 5),
  full.sweep.options = list(ntree = 25, nsplit = 5),
  save.ood = TRUE,
  verbose = FALSE
)

## add some missing values to the test data
xnew <- tst[, xvars, drop = FALSE]
xnew[sample(seq_len(nrow(xnew)), 5), xvars[1]] <- NA
xnew[sample(seq_len(nrow(xnew)), 5), xvars[2]] <- NA

## imputation
xnew.imp <- predict(sup.fit, xnew, max.predict.iter = 2, verbose = FALSE)
print(head(xnew.imp))

## OOD score
ood.sup <- impute.ood(sup.fit, xnew, verbose = FALSE)
print(head(ood.sup$score.percentile))

## ------------------------------------------------------------
## Save the learned imputer to disk and load it later.
## This explicit save example writes learners kept in memory.
## Uses missForest for the impute engine.
## ------------------------------------------------------------

bundle.dir <- file.path(tempdir(), "aq.imputer")

fit <- impute.learn(
  data = train,
  ntree = 25,
  mf.q = 1,
  max.iter = 5,
  full.sweep.options = list(ntree = 25, nsplit = 5),
  keep.models = TRUE,
  verbose = FALSE
)

save.impute.learn(fit, bundle.dir, verbose = FALSE)
imp <- load.impute.learn(bundle.dir, lazy = TRUE, verbose = FALSE)
test.imp <- predict(imp, test, max.predict.iter = 2, verbose = FALSE)

unlink(bundle.dir, recursive = TRUE)



## ------------------------------------------------------------
## Challenging example with factors, uses save/reload
## ------------------------------------------------------------

## load pbc, convert everything to factors
data(pbc, package = "randomForestSRC")
dta <- data.frame(lapply(pbc, factor))
dta$days <- pbc$days
dta$status <- dta$status

## split the data into unbalanced train/test data (25/75)
## the train/test data have the same levels, but different labels
idx <- sample(1:nrow(dta), round(nrow(dta) * .25))
train <- dta[idx,]
test <- dta[-idx,]

## even harder ... factor level not previously encountered in training
levels(test$stage) <- c(levels(test$stage), "fake")
test$stage[sample(seq_len(nrow(test)), 10)] <- "fake"

## train forest
fit <- suppressWarnings(
  impute.learn(Surv(days, status) ~ ., train,
               target.mode = "all",
               save.ood = TRUE,
               keep.models = TRUE)
)

## save/reload
bundle.dir <- file.path(tempdir(), "pbc.imputer")
save.impute.learn(fit, bundle.dir, verbose = FALSE)
imp <- load.impute.learn(bundle.dir, lazy = TRUE, verbose = FALSE)
test.imp <- predict(imp, test, max.predict.iter = 2, verbose = FALSE)
ood <- impute.ood(imp, test, return.details = TRUE, verbose = FALSE)
print(which(ood$info$unseen.rows))
print(summary(test.imp))
unlink(bundle.dir, recursive = TRUE)
}
}
\keyword{missing data}

\name{veteran}
\docType{data}
\alias{veteran}
\title{Veteran's Administration Lung Cancer Trial}
\description{
  Randomized trial of two treatment regimens for lung cancer.
  This is a standard survival analysis data set.
}
\source{Kalbfleisch and Prentice, \emph{The Statistical
    Analysis of Failure Time Data.}}
\references{
  Kalbfleisch J. and Prentice R, (1980) \emph{The Statistical
    Analysis of Failure Time Data.}  New York: Wiley.
}
\examples{data(veteran, package = "randomForestSRC")}
\keyword{datasets}

\name{impute.rfsrc}
\alias{impute.rfsrc}
\alias{impute}
\title{Impute Only Mode}
\description{
  Fast imputation mode. A random forest is grown and used to impute
  missing data. No ensemble estimates or error rates are calculated.
  Optionally, a final sweep can be performed to re-fit each variable
  that had original missingness on the final covariates and overwrite
  only its originally-missing entries.
}
\usage{\method{impute}{rfsrc}(formula, data,
  ntree = 100, nodesize = 1, nsplit = 10,
  nimpute = 2, fast = FALSE, blocks,
  mf.q, max.iter = 10, eps = 0.01,
  ytry = NULL, always.use = NULL, verbose = TRUE,
  full.sweep = FALSE, restore.integer = TRUE,
  ...)
}
\arguments{
  
  \item{formula}{A symbolic model description used to specify on the fly
    imputation.  Can be omitted for unsupervised imputation in settings
    where the outcomes are unspecified or if distinction between
    outcomes and predictors is unnecessary. Ignored for missForest.}

  \item{data}{A data frame containing variables to be imputed.}

  \item{ntree}{Number of trees grown for each imputation.}

  \item{nodesize}{Minimum terminal node size in each tree.}

  \item{nsplit}{Non-negative integer for specifying random splitting.}

  \item{nimpute}{Number of iterations for the missing data
    algorithm. Ignored for multivariate missForest, which iterates to
    convergence unless capped by \code{max.iter}.}

  \item{fast}{If \code{TRUE}, uses \code{rfsrc.fast} instead of
    \code{rfsrc}. Increases speed but may reduce accuracy.}

  \item{blocks}{Number of row-wise blocks to divide the data into. May
    improve speed for large data, but can reduce imputation accuracy. No
    action if unspecified.}

  \item{mf.q}{Enables missForest. Either a fraction (between 0 and 1) of
    variables treated as responses, or an integer indicating number of
    response variables. \code{mf.q = 1} corresponds to standard
    missForest.}

  \item{max.iter}{Maximum number of iterations for multivariate missForest.}

  \item{eps}{Convergence threshold for multivariate missForest (change in imputed values).}

  \item{ytry}{Number of variables used as pseudo-responses in
    unsupervised forests. See Details.}

  \item{always.use}{Character vector of variables always included as
    responses in multivariate missForest. Ignored by other methods.}

  \item{verbose}{If \code{TRUE}, prints progress during multivariate
    missForest imputation.}

  \item{full.sweep}{If \code{TRUE}, performs a final sweep after the
    main imputation (both standard and missForest). For each variable
    that had any \emph{original} missingness, a forest is fit on rows
    where the variable was observed using the \emph{final} imputed
    covariates; predictions are then written back only to the originally
    missing cells. This can improve self-consistency across variables at
    the cost of extra computation.}

  \item{restore.integer}{If \code{TRUE}, variables that were integer
    supported in the original data are restored to the integer grid after
    imputation. This includes columns stored as R integers and numeric
    columns whose observed finite training values are all integer-valued.
    Numeric columns remain numeric; returned values are rounded
    to integer support. Columns stored as R integers are returned as
    integer vectors. Set to \code{FALSE} to retain the raw numeric forest
    predictions.}

  \item{...}{Additional arguments passed to or from methods.
    Recognized advanced options include \code{full.sweep.options} (a
    \code{list}) controlling the final sweep hyperparameters:
    \code{ntree} (default \code{500}), \code{nodesize} (default
    \code{NULL}), \code{nsplit} (default \code{10}); and the standard
    \code{rfsrc} controls \code{mtry}, \code{splitrule}, \code{bootstrap},
    \code{sampsize}, \code{samptype} that apply to the sweep.}
}

\details{
  \enumerate{

  \item Before imputation, observations and variables with all values
  missing are removed.

  \item A forest is grown and used solely for imputation. No ensemble
  statistics (e.g., error rates) are computed. Use this function when
  imputation is the only goal.

  \item For standard imputation (not missForest), splits are based only
  on non-missing data. If a split variable has missing values, they are
  temporarily imputed by randomly drawing from in-bag, non-missing
  values to allow node assignment.

  \item If \code{mf.q} is specified, multivariate missForest imputation
  is applied (Stekhoven and B\"uhlmann, 2012). A fraction (or integer
  count) of variables are selected as multivariate responses, predicted
  using the remaining variables with multivariate composite
  splitting. Each round imputes a disjoint set of variables, and the
  full cycle is repeated until convergence, controlled by
  \code{max.iter} and \code{eps}. Setting \code{mf.q = 1} reverts to
  standard missForest. This method is typically the most accurate, but
  also the most computationally intensive.

  \item If no formula is provided, unsupervised splitting is used. The
  default \code{ytry} is \code{sqrt(p)}, where \code{p} is the number of
  variables. For each of \code{mtry} candidate variables, a random
  subset of \code{ytry} variables is selected as pseudo-responses. A
  multivariate composite splitting rule is applied, and the split is
  made on the variable yielding the best result (Tang and Ishwaran,
  2017).

  \item If no missing values remain after preprocessing, the function
  returns the processed data without further action.

  \item All standard \code{rfsrc} options apply; see examples below for
  illustration.

  \item \emph{Optional final sweep}: if \code{full.sweep = TRUE}, a
  post-imputation sweep is performed for every variable with original
  missingness. Each such variable is re-fit on its observed rows using
  the final imputed covariates, and predictions overwrite only the
  originally missing entries. Defaults for the sweep are
  \code{ntree = 500}, \code{nodesize = NULL}, \code{nsplit = 10}, and
  can be customized via \code{full.sweep.options} passed through
  \code{...}. This applies to both standard and missForest modes.

  \item If \code{restore.integer = TRUE}, integer support is detected
  from the original retained data before imputation begins. Restoration
  is applied after the missForest rough-fix initialization, before and
  after the optional final sweep, and once more before returning the
  completed data. The missForest iterative updates themselves are not
  rounded between variable or block updates.

  }
}

\value{
  Invisibly, the data frame containing the original data with imputed
  data overlaid.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Stekhoven D.J. and Buhlmann P. (2012). MissForest--non-parametric
  missing value imputation for mixed-type data.
  \emph{Bioinformatics}, 28(1):112--118.

  Tang F. and Ishwaran H. (2017). Random forest missing data
  algorithms. \emph{Statistical Analysis and Data Mining}, 10:363--377.
}
\seealso{
  \command{\link{impute.learn.rfsrc}},
  \command{\link{rfsrc}},
  \command{\link{rfsrc.fast}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## example of survival imputation
## ------------------------------------------------------------

## default everything - unsupervised splitting
data(pbc, package = "randomForestSRC")
pbc1.d <- impute(data = pbc)

## imputation using outcome splitting
f <- as.formula(Surv(days, status) ~ .)
pbc2.d <- impute(f, data = pbc, nsplit = 3)

## random splitting can be reasonably good
pbc3.d <- impute(f, data = pbc, splitrule = "random", nimpute = 5)

## optional final sweep (standard imputation)
pbc3.fs <- impute(f, data = pbc, splitrule = "random", nimpute = 5,
                  full.sweep = TRUE)

## ------------------------------------------------------------
## example of regression imputation
## ------------------------------------------------------------

air1.d <- impute(data = airquality, nimpute = 5)
air2.d <- impute(Ozone ~ ., data = airquality, nimpute = 5)
air3.d <- impute(Ozone ~ ., data = airquality, fast = TRUE)

## final sweep with custom options (e.g., larger forest)
air3.fs <- impute(Ozone ~ ., data = airquality, nimpute = 5,
                  full.sweep = TRUE,
                  full.sweep.options = list(ntree = 1000, nodesize = 5, nsplit = 0,
                                            mtry = 3, splitrule = "random"))

## ------------------------------------------------------------
## multivariate missForest imputation
## ------------------------------------------------------------

data(pbc, package = "randomForestSRC")

## missForest algorithm - uses 1 variable at a time for the response
pbc.d <- impute(data = pbc, mf.q = 1)

## multivariate missForest - use 10 percent of variables as responses
pbc.mv <- impute(data = pbc, mf.q = .10)

## missForest but faster by using random splitting
pbc.fast <- impute(data = pbc, mf.q = 1, splitrule = "random")

## missForest + final sweep
pbc.fast.fs <- impute(data = pbc, mf.q = 1, splitrule = "random",
                      full.sweep = TRUE)

}
}
\keyword{missing data}

\name{internal.utils}
\alias{extract.subsample}
\alias{extract.bootsample}
\alias{print.bootsample}
\alias{print.bootsample.rfsrc}
\alias{print.imbalanced.performance}
\alias{print.subsample}
\alias{print.subsample.rfsrc}
\alias{fast.save}
\alias{fast.load}
\alias{fast.save.list}
\alias{fast.load.list}
\alias{lsos}
\alias{get.imbalanced.performance}
\alias{get.imbalanced.optimize}
\alias{get.pr.auc}
\alias{get.pr.curve}
\alias{get.rfq.threshold}
\alias{get.auc}
\alias{get.bayes.rule}
\alias{get.brier.error}
\alias{get.cindex}
\alias{get.confusion}
\alias{get.logloss}
\alias{get.misclass.error}
\alias{get.mv.cserror}
\alias{get.mv.csvimp}
\alias{get.mv.error}
\alias{get.mv.error.block}
\alias{get.mv.formula}
\alias{get.mv.predicted}
\alias{get.mv.vimp}
\alias{extract.quantile}
\alias{get.quantile}
\alias{get.quantile.crps}
\alias{get.quantile.stat}
\alias{sid.perf.metric}
\title{Internal Utility Functions}

\description{
  These are internal utility functions exported for advanced usage. This
  Rd file is used solely to register aliases.
}

\keyword{internal}

\name{plot.subsample.rfsrc}
\alias{plot.subsample.rfsrc}
\alias{plot.subsample}
\title{Plot Subsampled VIMP Confidence Intervals}
\description{
  Plots VIMP (variable importance) confidence regions obtained from
  subsampling a forest.
}
\usage{\method{plot.subsample}{rfsrc}(x, alpha = .01, xvar.names,
 standardize = TRUE, normal = TRUE, jknife = FALSE, target, m.target = NULL,
 pmax = 75, main = "", sorted = TRUE, show.plots = TRUE, ...)}
\arguments{

  \item{x}{An object obtained from calling \command{subample}.}

  \item{alpha}{Desired level of significance.}

  \item{xvar.names}{Names of the x-variables to be used.  If not
  specified all variables used.}
  
  \item{standardize}{Standardize VIMP?  For regression families, VIMP is
    standardized by dividing by the variance.  For all other families,
    VIMP is unaltered.}

  \item{normal}{Use parametric normal confidence regions or
    nonparametric regions?  Generally, parametric regions perform better.}

  \item{jknife}{Use the delete-d jackknife variance estimator?}

  \item{target}{For classification families, an integer or character
    value specifying the class VIMP will be conditioned on (default is
    to use unconditional VIMP).  For competing risk families, an integer
    value between 1 and \code{J} indicating the event VIMP is requested,
    where \code{J} is the number of event types.  The default is to use
    the first event.}

  \item{m.target}{Character value for multivariate families
    specifying the target outcome to be used.  If left unspecified, the
    algorithm will choose a default target.}

  \item{pmax}{Trims the data to this number of variables (sorted by VIMP).}

  \item{main}{Title used for plot.}
  
  \item{sorted}{Should variables be sorted by importance values?}

  \item{show.plots}{Should plots be displayed?  Allows users to produce
    their own custom plots.}

  \item{...}{Further arguments that can be passed to \command{bxp}.}
  
}
\details{
  Most of the options used by the R function bxp will work here and can
  be used for customization of plots.  Currently the following
  parameters will work:

  "xaxt", "yaxt", "las", "cex.axis", 
  "col.axis", "cex.main",
  "col.main", "sub", "cex.sub", "col.sub", 
  "ylab", "cex.lab", "col.lab"
		 
}
\value{
  Invisibly, returns the boxplot data that is plotted.
}

\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Ishwaran H. and Lu M.  (2017).  Standard errors and confidence
  intervals for variable importance in random forest regression,
  classification, and survival.

  Politis, D.N. and Romano, J.P. (1994). Large sample confidence
  regions based on subsamples under minimal assumptions. \emph{The
  Annals of Statistics}, 22(4):2031-2050.

  Shao, J. and Wu, C.J. (1989). A general theory for jackknife variance
  estimation. \emph{The Annals of Statistics}, 17(3):1176-1197.
}
\seealso{
  \command{\link{subsample.rfsrc}}
}
\examples{
\donttest{
o <- rfsrc(Ozone ~ ., airquality)
oo <- subsample(o)
plot.subsample(oo)
plot.subsample(oo, xvar.names = o$xvar.names[1:3])
plot.subsample(oo, jknife = FALSE)
plot.subsample(oo, alpha = .01)
plot(oo,cex.axis=.5)
}
}
\keyword{plot}

\name{plot.competing.risk.rfsrc}
\alias{plot.competing.risk.rfsrc}
\alias{plot.competing.risk}
\title{Plots for Competing Risks}
\description{
  Plot useful summary curves from a random survival forest competing risk
  analysis.
}
\usage{\method{plot.competing.risk}{rfsrc}(x, plots.one.page = FALSE, ...)}
\arguments{
  \item{x}{An object of class \code{(rfsrc, grow)} or
    \code{(rfsrc, predict)}.}
  \item{plots.one.page}{Should plots be placed on one page?}
  \item{...}{Further arguments passed to or from other methods.}
}
\details{

  Given a random survival forest object from a competing risk analysis
  (Ishwaran et al. 2014), plots from top to bottom, left to right: (1)
  cause-specific cumulative hazard function (CSCHF) for each event, (2)
  cumulative incidence function (CIF) for each event, and (3) continuous
  probability curves (CPC) for each event (Pepe and Mori, 1993).
  
  Does not apply to right-censored data.  Whenever possible, out-of-bag
  (OOB) values are displayed.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  
  Ishwaran H., Gerds T.A., Kogalur U.B., Moore R.D., Gange S.J. and Lau
  B.M. (2014). Random survival forests for competing risks.
  \emph{Biostatistics}, 15(4):757-773.

  Pepe, M.S. and Mori, M., (1993). Kaplan-Meier, marginal or conditional
  probability curves in summarizing competing risks failure time
  data? \emph{Statistics in Medicine}, 12(8):737-751.
 
  
}
\seealso{
  \command{\link{follic}},
  \command{\link{hd}},
  \command{\link{rfsrc}},
  \command{\link{wihs}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## follicular cell lymphoma
## ------------------------------------------------------------

  data(follic, package = "randomForestSRC")
  follic.obj <- rfsrc(Surv(time, status) ~ ., follic, nsplit = 3, ntree = 100)
  print(follic.obj)
  plot.competing.risk(follic.obj)

## ------------------------------------------------------------
## Hodgkin's Disease
## ------------------------------------------------------------

  data(hd, package = "randomForestSRC")
  hd.obj <- rfsrc(Surv(time, status) ~ ., hd, nsplit = 3, ntree = 100)
  print(hd.obj)
  plot.competing.risk(hd.obj)

## ------------------------------------------------------------
## competing risk analysis of pbc data from the survival package
## events are transplant (1) and death (2)
## ------------------------------------------------------------

if (library("survival", logical.return = TRUE)) {
   data(pbc, package = "survival")
   pbc$id <- NULL
   plot.competing.risk(rfsrc(Surv(time, status) ~ ., pbc))
}
}
}
\keyword{plot}

\name{tune.rfsrc}
\alias{tune.rfsrc}
\alias{tune}
\alias{tune.nodesize.rfsrc}
\alias{tune.nodesize}
\title{Tune Random Forest for optimal \code{mtry} and \code{nodesize}}

\description{

  Finds the optimal \code{mtry} and \code{nodesize} for a random forest
using out-of-bag (OOB) error.  Two search strategies are supported: a
grid-based search and a golden-section search with noise control.  Works
for all response families supported by \code{rfsrc.fast}.

}

\usage{

\method{tune}{rfsrc}(formula, data,
  mtry.start = ncol(data) / 2,
  nodesize.try = c(1:9, seq(10, 100, by = 5)), ntree.try = 100,
  sampsize = function(x) { min(x * .632, max(150, x^(3/4))) },
  nsplit = 1, step.factor = 1.25, improve = 1e-3, strikeout = 3, max.iter = 25,
  method = c("grid", "golden"),
  final.window = 5, reps.initial = 2, reps.final = 3,
  trace = FALSE, do.best = TRUE, seed = NULL, ...)

\method{tune.nodesize}{rfsrc}(formula, data,
  nodesize.try = c(1:9, seq(10, 150, by = 5)), ntree.try = 100,
  sampsize = function(x) { min(x * .632, max(150, x^(4/5))) },
  nsplit = 1, method = c("grid", "golden"),
  final.window = 5, reps.initial = 2, reps.final = 3, max.iter = 50,
  trace = TRUE, seed = NULL, ...)
}

\arguments{
  \item{formula}{A model formula.}
  \item{data}{A data frame with response and predictors.}
  \item{mtry.start}{Initial \code{mtry} for \code{tune}.}
  \item{nodesize.try}{Candidate \code{nodesize} values. Only values \eqn{\le} \code{floor(sampsize(n)/2)} are used.}
  \item{ntree.try}{Number of trees grown at each tuning evaluation.}
  \item{sampsize}{Function or numeric giving the per-tree subsample size. During tuning a single numeric size \code{ssize} is computed and passed to \code{rfsrc.fast}. If a vector is supplied (e.g., class specific), its total is used for \code{ssize}.}
  \item{nsplit}{Number of random split points to consider at each node.}
  \item{step.factor}{Multiplicative step-out factor over \code{mtry} for grid search in \code{tune}.}
  \item{improve}{Minimum relative improvement required to continue a search step in \code{tune}.}
  \item{strikeout}{Maximum number of consecutive non-improving steps allowed in \code{tune}.}
  \item{max.iter}{Maximum number of iterations for the step-out search in \code{tune} or the coordinate loop when \code{method = "golden"}.}
  \item{method}{Search strategy: \code{"grid"} (default) or \code{"golden"}.}
  \item{final.window}{For golden search, the terminal bracket width for the one-dimensional line search.}
  \item{reps.initial}{Replicates averaged at interior evaluations during golden iterations.}
  \item{reps.final}{Replicates averaged for each candidate during the final local sweep in golden search.}
  \item{trace}{If \code{TRUE}, prints progress.}
  \item{do.best}{If \code{TRUE}, \code{tune} fits and returns a forest at the optimal pair.}
  \item{seed}{Optional integer for reproducible tuning. The holdout split (when used) and all tuning fits become deterministic for a given seed.}
  \item{...}{Additional arguments passed to \code{\link{rfsrc.fast}}. Arguments that control tuning itself (\code{perf.type}, \code{forest}, \code{save.memory}, \code{ntree}, \code{mtry}, \code{nodesize}, \code{sampsize}, \code{nsplit}) are managed internally.}
}

\details{

\emph{Error estimate.} If \code{2 * ssize < n}, a disjoint holdout of
size \code{ssize} is used for evaluation; otherwise OOB error is
used.


\emph{Subsample used during tuning.} Both functions derive a single
integer \code{ssize} from \code{sampsize} and pass it to
\code{\link{rfsrc.fast}} for all tuning fits. This improves stability
and comparability across candidates. When \code{do.best = TRUE} in
\code{tune}, the final forest is fit with the user-supplied
\code{sampsize} exactly as provided.


\emph{Grid search.} \code{tune} performs a step-out search over
\code{mtry} for each \code{nodesize} in \code{nodesize.try}, using
\code{step.factor}, \code{improve}, \code{strikeout}, and
\code{max.iter}. \code{tune.nodesize} evaluates the supplied
\code{nodesize.try} grid directly.


\emph{Golden search.} Uses a guarded golden-section line search with
noise control. For each one-dimensional search (over \code{nodesize} or
\code{mtry}), the routine probes a small left-anchor grid \code{1:9},
iterates golden shrinkage until the bracket width is at most
\code{final.window}, then runs a short local sweep with
\code{reps.final} replicates. In \code{tune} the searches over
\code{nodesize} and \code{mtry} alternate in a simple coordinate loop,
with \code{improve} and \code{strikeout} as stopping controls.


}

\value{
For \code{tune}:
\itemize{
  \item \code{results}: matrix with columns \code{nodesize}, \code{mtry}, \code{err}.
  \item \code{optimal}: named numeric vector \code{c(nodesize = ..., mtry = ...)}.
  \item \code{rf}: fitted forest at the optimum if \code{do.best = TRUE}.
}

For \code{tune.nodesize}:
\itemize{
  \item \code{nsize.opt}: optimal \code{nodesize}.
  \item \code{err}: data frame with columns \code{nodesize} and \code{err}.
}
}

\author{
Hemant Ishwaran and Udaya B. Kogalur
}

\seealso{
\code{\link{rfsrc.fast}}
}

\examples{
\donttest{
## ------------------------------------------------------------
## White wine classification example
## ------------------------------------------------------------
data(wine, package = "randomForestSRC")
wine$quality <- factor(wine$quality)

## Fixed seed makes tuning reproducible
set.seed(1)

## Full tuner over nodesize and mtry (grid)
o1 <- tune(quality ~ ., wine, sampsize = 100, method = "grid")
print(o1$optimal)

## Golden search alternative
o2 <- tune(quality ~ ., wine, sampsize = 100, method = "golden",
           reps.initial = 2, reps.final = 3, seed = 1)
print(o2$optimal)

## visualize the nodesize/mtry surface
if (library("interp", logical.return = TRUE)) {

  plot.tune <- function(o, linear = TRUE) {
    x <- o$results[, 1]
    y <- o$results[, 2]
    z <- o$results[, 3]
    so <- interp(x = x, y = y, z = z, linear = linear)
    idx <- which.min(z)
    x0 <- x[idx]; y0 <- y[idx]
    filled.contour(x = so$x, y = so$y, z = so$z,
                   xlim = range(so$x, finite = TRUE) + c(-2, 2),
                   ylim = range(so$y, finite = TRUE) + c(-2, 2),
                   color.palette = colorRampPalette(c("yellow", "red")),
                   xlab = "nodesize", ylab = "mtry",
                   main = "error rate for nodesize and mtry",
                   key.title = title(main = "OOB error", cex.main = 1),
                   plot.axes = {
                     axis(1); axis(2)
                     points(x0, y0, pch = "x", cex = 1, font = 2)
                     points(x, y, pch = 16, cex = .25)
                   })
  }

  plot.tune(o1)
  plot.tune(o2)
}

## ------------------------------------------------------------
## nodesize only: grid vs golden
## ------------------------------------------------------------
o3 <- tune.nodesize(quality ~ ., wine, sampsize = 100, method = "grid",
                    trace = TRUE, seed = 1)
o4 <- tune.nodesize(quality ~ ., wine, sampsize = 100, method = "golden",
                    reps.initial = 2, reps.final = 3, trace = TRUE, seed = 1)
plot(o3$err, type = "s", xlab = "nodesize", ylab = "error")

## ------------------------------------------------------------
## Tuning for class imbalance (rfq with geometric mean performance)
## ------------------------------------------------------------
data(breast, package = "randomForestSRC")
breast <- na.omit(breast)
o5 <- tune(status ~ ., data = breast, rfq = TRUE, perf.type = "gmean",
           method = "golden", seed = 1)
print(o5$optimal)

## ------------------------------------------------------------
## Competing risks example (nodesize only)
## ------------------------------------------------------------
data(wihs, package = "randomForestSRC")
plot(tune.nodesize(Surv(time, status) ~ ., wihs, trace = TRUE)$err, type = "s")
}}

\keyword{forest}
\keyword{tune}

\name{vdv}
\docType{data}
\alias{vdv}
\title{van de Vijver Microarray Breast Cancer}
\description{
  Gene expression profiling for predicting clinical outcome of breast
  cancer (van't Veer et al., 2002).  Microarray breast cancer data set
  of 4707 expression values on 78 patients with survival information.
}
\references{
  van't Veer L.J. et al. (2002).  Gene expression profiling predicts
  clinical outcome of breast cancer.  \emph{Nature}, \bold{12},
  530--536.
}
\examples{data(vdv, package = "randomForestSRC")}
\keyword{datasets}

\name{wine}
\docType{data}
\alias{wine}
\title{White Wine Quality Data}
\description{
  The inputs include objective tests (e.g. PH values) and the output is
  based on sensory data (median of at least 3 evaluations made by wine
  experts) of white wine. Each expert graded the wine quality between 0
  (very bad) and 10 (very excellent).
}
\references{
  Cortez, P., Cerdeira, A., Almeida, F., Matos T. and Reis, J. (2009).
  Modeling wine preferences by data mining from physicochemical properties.
  In \emph{Decision Support Systems}, Elsevier, 47(4):547-553.
}
\examples{
## load wine and convert to a multiclass problem
data(wine, package = "randomForestSRC")
wine$quality <- factor(wine$quality)
}
\keyword{datasets}

\name{pbc}
\docType{data}
\alias{pbc}
\title{Primary Biliary Cirrhosis (PBC) Data}
\description{
  Data from the Mayo Clinic trial in primary biliary cirrhosis (PBC) of
  the liver conducted between 1974 and 1984.  A total of 424 PBC
  patients, referred to Mayo Clinic during that ten-year interval, met
  eligibility criteria for the randomized placebo controlled trial of
  the drug D-penicillamine.  The first 312 cases in the data set
  participated in the randomized trial and contain largely complete
  data.  
}
\source{Flemming and Harrington, 1991, Appendix D.1.}
\references{
  Flemming T.R and Harrington D.P., (1991) \emph{Counting Processes
    and Survival Analysis.} New York: Wiley.
}
\examples{
\donttest{
data(pbc, package = "randomForestSRC")
pbc.obj <- rfsrc(Surv(days, status) ~ ., pbc, nsplit = 3)
}
}
\keyword{datasets}

\name{nutrigenomic}
\docType{data}
\alias{nutrigenomic}
\title{Nutrigenomic Study}
\description{
  Investigates the effects of five dietary treatments on 21 liver lipids
  and 120 hepatic gene expressions in wild-type and PPAR-alpha deficient
  mice. A multivariate mixed random forest analysis is performed by
  regressing gene expression, diet, and genotype (x-variables) on lipid
  expression profiles (multivariate y-responses).
}
\references{
  Martin P.G. et al. (2007). Novel aspects of PPAR-alpha-mediated
  regulation of lipid and xenobiotic metabolism revealed through a
  nutrigenomic study. \emph{Hepatology}, 45(3), 767--777.
}
\examples{
\donttest{
## ------------------------------------------------------------
## multivariate regression forests using Mahalanobis splitting
## lipids (all real values) used as the multivariate y
## ------------------------------------------------------------

## load the data
data(nutrigenomic, package = "randomForestSRC")

## parse into y and x data
ydta <- nutrigenomic$lipids
xdta <- data.frame(nutrigenomic$genes,
                   diet = nutrigenomic$diet,
                   genotype = nutrigenomic$genotype)

## multivariate mixed forest call
obj <- rfsrc(get.mv.formula(colnames(ydta)),
             data.frame(ydta, xdta),
             importance=TRUE, nsplit = 10,
             splitrule = "mahalanobis")
print(obj)

## ------------------------------------------------------------
## plot the standarized performance and VIMP values
## ------------------------------------------------------------

## acquire the error rate for each of the 21-coordinates 
## standardize to allow for comparison across coordinates
serr <- get.mv.error(obj, standardize = TRUE)

## acquire standardized VIMP 
svimp <- get.mv.vimp(obj, standardize = TRUE)

par(mfrow = c(1,2))
plot(serr, xlab = "Lipids", ylab = "Standardized Performance")
matplot(svimp, xlab = "Genes/Diet/Genotype", ylab = "Standardized VIMP")


## ------------------------------------------------------------
## plot some trees
## ------------------------------------------------------------

plot(get.tree(obj, 1))
plot(get.tree(obj, 2))
plot(get.tree(obj, 3))


## ------------------------------------------------------------
##
## Compare above to (1) user specified covariance matrix
##                  (2) default composite (independent) splitting
##
## ------------------------------------------------------------

## user specified sigma matrix
obj2 <- rfsrc(get.mv.formula(colnames(ydta)),
              data.frame(ydta, xdta),
              importance = TRUE, nsplit = 10,
              splitrule = "mahalanobis",
              sigma = cov(ydta))
print(obj2)

## default independence split rule
obj3 <- rfsrc(get.mv.formula(colnames(ydta)),
              data.frame(ydta, xdta),
              importance=TRUE, nsplit = 10)
print(obj3)

## compare vimp
imp <- data.frame(mahalanobis  = rowMeans(get.mv.vimp(obj,  standardize = TRUE)),
                  mahalanobis2 = rowMeans(get.mv.vimp(obj2, standardize = TRUE)),
                  default      = rowMeans(get.mv.vimp(obj3, standardize = TRUE)))

print(head(100 * imp[order(imp$mahalanobis, decreasing = TRUE), ], 15))

}}
\keyword{datasets}

\name{rfsrc.anonymous}
\alias{rfsrc.anonymous}
\title{Anonymous Random Forests}

\description{

  Anonymous random forests is carefully modified to ensure that the
  original training data is not retained. This enables users to share
  the trained forest with others without disclosing the underlying data.

}

\usage{rfsrc.anonymous(formula, data, forest = TRUE, ...)}

\arguments{
  \item{formula}{A symbolic description of the model to be fit. If missing, unsupervised splitting is performed.}

  \item{data}{A data frame containing the y-outcome and x-variables.}

  \item{forest}{Logical. Should the forest object be returned? Required for prediction on new data and by many other package functions.}

  \item{...}{Additional arguments passed to \code{\link{rfsrc}}. See the \code{rfsrc} help file for full details.}
}


\details{

This function calls \code{\link{rfsrc}} and returns a forest object with the original training data removed. This enables users to share their forest while preserving the privacy of their data.

To enable prediction on new (test) data, certain minimal information from the training data must still be retained. This includes:
\itemize{
  \item Names of the original variables.
  \item For factor variables, the levels of each factor.
  \item Summary statistics used for imputation: the mean for continuous variables and the most frequent class for factors.
  \item Tree topology, including split points used to grow the trees.
}

For maximal privacy, users are strongly encouraged to replace variable names with non-identifiable labels and convert all variables to continuous format when possible. If factor variables are used, their levels should also be anonymized. However, the user is solely responsible for de-identifying the data and verifying that privacy is maintained. \strong{We provide NO GUARANTEES regarding data confidentiality.}

\emph{Missing data handling:} Anonymous forests do not support imputation of training data. The option \code{na.action = "na.impute"} is automatically downgraded to \code{"na.omit"}. If training data contain missing values, we recommend pre-imputing them using \code{\link{impute}}.

Test data, however, \emph{can} be imputed at prediction time:
\itemize{
  \item \code{na.action = "na.impute"} performs a fast imputation by replacing missing values with the training mean (for numeric variables) or most frequent class (for factors).
  \item \code{na.action = "na.random"} uses a fast random draw from training distributions for imputation.
}

Although anonymous forests are compatible with many package functions, they are only guaranteed to work with functions that do not explicitly require access to the original training data.

}

\value{
  An object of class \code{(rfsrc, grow, anonymous)}.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\seealso{
  \command{\link{rfsrc}}
}
\examples{
\donttest{

## ------------------------------------------------------------
## regression
## ------------------------------------------------------------
print(rfsrc.anonymous(mpg ~ ., mtcars))

## ------------------------------------------------------------
## plot anonymous regression tree (using get.tree)
## TBD CURRENTLY NOT IMPLEMENTED 
## ------------------------------------------------------------
## plot(get.tree(rfsrc.anonymous(mpg ~ ., mtcars), 10))

## ------------------------------------------------------------
## classification
## ------------------------------------------------------------
print(rfsrc.anonymous(Species ~ ., iris))

## ------------------------------------------------------------
## survival
## ------------------------------------------------------------
data(veteran, package = "randomForestSRC")
print(rfsrc.anonymous(Surv(time, status) ~ ., data = veteran))

## ------------------------------------------------------------
## competing risks
## ------------------------------------------------------------
data(wihs, package = "randomForestSRC")
print(rfsrc.anonymous(Surv(time, status) ~ ., wihs, ntree = 100))

## ------------------------------------------------------------
## unsupervised forests
## ------------------------------------------------------------
print(rfsrc.anonymous(data = iris))

## ------------------------------------------------------------
## multivariate regression
## ------------------------------------------------------------
print(rfsrc.anonymous(Multivar(mpg, cyl) ~., data = mtcars))

## ------------------------------------------------------------
## prediction on test data with missing values using pbc data
## cases 1 to 312 have no missing values
## cases 313 to 418 having missing values
## ------------------------------------------------------------
data(pbc, package = "randomForestSRC")
pbc.obj <- rfsrc.anonymous(Surv(days, status) ~ ., pbc)
print(pbc.obj)

## mean value imputation
print(predict(pbc.obj, pbc[-(1:312),], na.action = "na.impute"))

## random imputation
print(predict(pbc.obj, pbc[-(1:312),], na.action = "na.random"))

## ------------------------------------------------------------
## train/test setting but tricky because factor labels differ over
## training and test data
## ------------------------------------------------------------

# first we convert all x-variables to factors
data(veteran, package = "randomForestSRC")
veteran.factor <- data.frame(lapply(veteran, factor))
veteran.factor$time <- veteran$time
veteran.factor$status <- veteran$status

# split the data into train/test data (25/75)
# the train/test data have the same levels, but different labels
train <- sample(1:nrow(veteran), round(nrow(veteran) * .5))
summary(veteran.factor[train, ])
summary(veteran.factor[-train, ])

# grow the forest on the training data and predict on the test data
v.grow <- rfsrc.anonymous(Surv(time, status) ~ ., veteran.factor[train, ]) 
v.pred <- predict(v.grow, veteran.factor[-train, ])
print(v.grow)
print(v.pred)



}}
\keyword{forest}
\keyword{anonymous}

\name{rfsrc.fast}
\alias{rfsrc.fast}
\title{Fast Random Forests}
\description{
  Fast approximate random forests using subsampling with forest options
  set to encourage computational speed.  Applies to all families.
}
\usage{rfsrc.fast(formula, data,
  ntree = 500,
  nsplit = 10,
  bootstrap = "by.root",
  sampsize = function(x){min(x * .632, max(150, x ^ (3/4)))},
  samptype = "swor",
  samp = NULL,
  ntime = 50,
  forest = FALSE,
  save.memory = TRUE,
  ...)
}
\arguments{

  \item{formula}{Model to be fit. If missing, unsupervised splitting is
    implemented.}

  \item{data}{Data frame containing the y-outcome and x-variables.} 

  \item{ntree}{Number of trees.}

  \item{nsplit}{Non-negative integer value specifying number of 
    random split points used to split a node (deterministic splitting
    corresponds to the value zero and can be slower).}

  \item{bootstrap}{Bootstrap protocol used in growing a tree.}

  \item{sampsize}{Function specifying size of subsampled data. Can also be a number.}

  \item{samptype}{Type of bootstrap used.}

  \item{samp}{Bootstrap specification when \code{"by.user"} is used.}

  \item{ntime}{Integer value used for survival to
    constrain ensemble calculations to a grid of \code{ntime} time points.}

  \item{forest}{Save key forest values? Turn this on if you want prediction on test data.}

  \item{save.memory}{Save memory?  Setting this to \code{FALSE} stores
    terminal node quantities used for prediction on test data.  This
    yields rapid prediction but can be memory intensive for big data,
    especially competing risks and survival models.}
  
  \item{...}{Further arguments to be passed to \code{\link{rfsrc}}.}
  
}
\details{
  Calls \code{\link{rfsrc}} by choosing options (like subsampling) to
  encourage computational speeds.  This will provide a good
  approximation but will not be as good as default settings of
  \code{\link{rfsrc}}.
}
\value{
  An object of class \code{(rfsrc, grow)}.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\seealso{
  \command{\link{rfsrc}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## regression 
## ------------------------------------------------------------

## load the Iowa housing data
data(housing, package = "randomForestSRC")

## do quick and *dirty* imputation
housing <- impute(SalePrice ~ ., housing,
         ntree = 50, nimpute = 1, splitrule = "random")

## grow a fast forest
o1 <- rfsrc.fast(SalePrice ~ ., housing)
o2 <- rfsrc.fast(SalePrice ~ ., housing, nodesize = 1)
print(o1)
print(o2)

## grow a fast bivariate forest
o3 <- rfsrc.fast(cbind(SalePrice,Overall.Qual) ~ ., housing)
print(o3)

## ------------------------------------------------------------
## classification 
## ------------------------------------------------------------

data(wine, package = "randomForestSRC")
wine$quality <- factor(wine$quality)
o <- rfsrc.fast(quality ~ ., wine)
print(o)

## ------------------------------------------------------------
## grow fast random survival forests without C-calculation
## use brier score to assess model performance
## compare pure random splitting to logrank splitting
## ------------------------------------------------------------

data(peakVO2, package = "randomForestSRC")
f <- as.formula(Surv(ttodead, died)~.)
o1 <- rfsrc.fast(f, peakVO2, perf.type = "none")
o2 <- rfsrc.fast(f, peakVO2, perf.type = "none", splitrule = "random")
bs1 <- get.brier.survival(o1, cens.model = "km")
bs2 <- get.brier.survival(o2, cens.model = "km")
plot(bs2$brier.score, type = "s", col = 2)
lines(bs1$brier.score, type = "s", col = 4)
legend("bottomright", legend = c("random", "logrank"), fill = c(2,4))

## ------------------------------------------------------------
## competing risks
## ------------------------------------------------------------

data(wihs, package = "randomForestSRC")
o <- rfsrc.fast(Surv(time, status) ~ ., wihs)
print(o)

## ------------------------------------------------------------
## class imbalanced data using gmean performance
## ------------------------------------------------------------

data(breast, package = "randomForestSRC")
breast <- na.omit(breast)
f <- as.formula(status ~ .)
o <- rfsrc.fast(f, breast, perf.type = "gmean")
print(o)

## ------------------------------------------------------------
## class imbalanced data using random forests quantile-classifer (RFQ)
## fast=TRUE => rfsrc.fast
## see imbalanced function for further details
## ------------------------------------------------------------

data(breast, package = "randomForestSRC")
breast <- na.omit(breast)
f <- as.formula(status ~ .)
o <- imbalanced(f, breast, fast = TRUE)
print(o)

}}
\keyword{forest}
\keyword{fast}

\name{plot.quantreg.rfsrc}
\alias{plot.quantreg.rfsrc}
\alias{plot.quantreg}
\title{Plot Quantiles from Quantile Regression Forests}
\description{
  Plots quantiles obtained from a quantile regression forest.
  Additionally insets the continuous rank probability score (crps), a
  useful diagnostic of accuracy.
}
\usage{\method{plot.quantreg}{rfsrc}(x, prbL = .25, prbU = .75,
   m.target = NULL, crps = TRUE, subset = NULL, xlab = NULL, ylab = NULL, ...)
}
\arguments{

  \item{x}{A quantile regression object returned by a call to \command{quantreg}.}

  \item{prbL}{Lower quantile level, typically less than \code{0.5}.}

  \item{prbU}{Upper quantile level, typically greater than \code{0.5}.}

  \item{m.target}{Character string specifying the target outcome for multivariate families. 
  If not provided, a default target is selected automatically.}

  \item{crps}{Logical. If \code{TRUE}, calculates the continuous ranked probability score (CRPS) and adds it to the plot.}

  \item{subset}{Optional vector specifying a subset of the data to be plotted. 
  Defaults to plotting all data points.}

  \item{xlab}{Label for the x-axis.}

  \item{ylab}{Label for the y-axis.}

  \item{...}{Additional arguments passed to or from other methods.}
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\seealso{
  \command{\link{quantreg.rfsrc}}
}
\keyword{plot}

\name{max.subtree.rfsrc}
\alias{max.subtree.rfsrc}
\alias{max.subtree}
\title{Acquire Maximal Subtree Information}

\description{
    Extract maximal subtree information from a RF-SRC object.  Used for
    variable selection and identifying interactions between variables.
}

\usage{\method{max.subtree}{rfsrc}(object,
  max.order = 2, sub.order = FALSE, conservative = FALSE, ...)
}

\arguments{

  \item{object}{An object of class \code{(rfsrc, grow)} or \code{(rfsrc, forest)}.}

  \item{max.order}{Non-negative integer specifying the maximum interaction order for which minimal depth is calculated. Defaults to 2. Set \code{max.order=0} to return first-order depths only. When \code{max.order=0}, \code{conservative} is automatically set to \code{FALSE}.}

  \item{sub.order}{Logical. If \code{TRUE}, returns the minimal depth of each variable conditional on every other variable. Useful for investigating variable interdependence. See \code{Details}.}

  \item{conservative}{Logical. If \code{TRUE}, uses a conservative threshold for selecting variables based on the marginal minimal depth distribution (Ishwaran et al., 2010). If \code{FALSE}, uses the tree-averaged distribution, which is less conservative and typically identifies more variables in high-dimensional settings.}

  \item{...}{Additional arguments passed to or from other methods.}
}

\details{
  The maximal subtree for a variable \emph{x} is the largest subtree in which the root node splits on \emph{x}. The largest possible maximal subtree is the full tree (root node), though multiple maximal subtrees may exist for a variable. A variable may also have no maximal subtree if it is never used for splitting. See Ishwaran et al. (2010, 2011) for further discussion.

  The minimal depth of a maximal subtree-called the \emph{first-order depth}-quantifies the predictive strength of a variable. It is defined as the distance from the root node to the parent of the closest maximal subtree for \emph{x}. Smaller values indicate stronger predictive impact. A variable is flagged as strong if its minimal depth is below the mean of the minimal depth distribution.

  The \emph{second-order depth} is the distance from the root to the second-closest maximal subtree of \emph{x}. To request depths beyond first order, use the \code{max.order} option (e.g., \code{max.order = 2} returns both first and second-order depths). Set \code{max.order = 0} to retrieve first-order depths for each variable in each tree.

  Set \code{sub.order = TRUE} to obtain the relative minimal depth of
  each variable \emph{j} within the maximal subtree of another variable
  \emph{i}. This returns a \code{p x p} matrix (with \code{p} the number
  of variables) whose entry (i,j) is the normalized relative depth of
  \emph{j} in \emph{i}'s subtree. Entry (i,i) gives the depth of
  \emph{i} relative to the root. Read the matrix across rows to assess
  inter-variable relationships: small (i,j) entries suggest interactions
  between variables \emph{i} and \emph{j}. 

  For competing risks, all analyses are unconditional (non-event specific).
}


\value{

  Invisibly returns a list with the following components:

  \item{order}{Matrix of order depths for each variable up to
  \code{max.order}, averaged over trees. The matrix has \code{p} rows
  and \code{max.order} columns, where \code{p} is the number of
  variables. If \code{max.order = 0}, returns a matrix of dimension
  \code{p x ntree} containing first-order depths for each variable by
  tree.}

  \item{count}{Average number of maximal subtrees per variable, normalized by tree size.}

  \item{nodes.at.depth}{List of vectors recording the number of non-terminal nodes at each depth level for each tree.}

  \item{sub.order}{Matrix of average minimal depths of each variable relative to others (i.e., conditional minimal depth matrix). \code{NULL} if \code{sub.order = FALSE}.}

  \item{threshold}{Threshold value for selecting strong variables based on the mean of the minimal depth distribution.}

  \item{threshold.1se}{Conservative threshold equal to the mean minimal depth plus one standard error.}

  \item{topvars}{Character vector of selected variable names using the \code{threshold} criterion.}

  \item{topvars.1se}{Character vector of selected variable names using the \code{threshold.1se} criterion.}

  \item{percentile}{Percentile value of minimal depth for each variable.}

  \item{density}{Estimated density of the minimal depth distribution.}

  \item{second.order.threshold}{Threshold used for selecting strong second-order depth variables.}

}

\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Ishwaran H., Kogalur U.B., Gorodeski E.Z, Minn A.J. and 
  Lauer M.S. (2010).  High-dimensional variable selection for survival
  data.  \emph{J. Amer. Statist. Assoc.}, 105:205-217.
  
  Ishwaran H., Kogalur U.B., Chen X. and Minn A.J. (2011).  Random
  survival forests for high-dimensional data. \emph{Statist. Anal. Data
    Mining}, 4:115-132.
}
\seealso{
  \command{\link{holdout.vimp.rfsrc}},
  \command{\link{vimp.rfsrc}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## survival analysis
## first and second order depths for all variables
## ------------------------------------------------------------

data(veteran, package = "randomForestSRC")
v.obj <- rfsrc(Surv(time, status) ~ . , data = veteran)
v.max <- max.subtree(v.obj)

# first and second order depths
print(round(v.max$order, 3))

# the minimal depth is the first order depth
print(round(v.max$order[, 1], 3))

# strong variables have minimal depth less than or equal
# to the following threshold
print(v.max$threshold)

# this corresponds to the set of variables
print(v.max$topvars)

## ------------------------------------------------------------
## regression analysis
## try different levels of conservativeness
## ------------------------------------------------------------

mtcars.obj <- rfsrc(mpg ~ ., data = mtcars)
max.subtree(mtcars.obj)$topvars
max.subtree(mtcars.obj, conservative = TRUE)$topvars
}
}
\keyword{variable selection}

\name{rfsrc}
\alias{rfsrc}
\alias{rfsrc.cart}
\alias{randomForestSRC}
\title{Fast Unified Random Forests for Survival, Regression, and Classification (RF-SRC)}
\description{

Fast OpenMP-parallel implementation of random forests (Breiman, 2001)
for regression, classification, survival analysis (Ishwaran et al.,
2008), competing risks (Ishwaran et al., 2012), multivariate outcomes
(Segal and Xiao, 2011), unsupervised learning (Mantero and Ishwaran,
2020), quantile regression (Meinshausen, 2006; Zhang et al., 2019;
Greenwald and Khanna, 2001), and imbalanced q-classification (O'Brien
and Ishwaran, 2019).
  

The package supports both deterministic and randomized splitting rules
(Geurts et al., 2006; Ishwaran, 2015) across all families. Multiple
types of variable importance (VIMP) are available, including holdout
VIMP and confidence regions (Ishwaran and Lu, 2019), for both individual
and grouped variables. Variable selection can be performed using minimal
depth (Ishwaran et al., 2010, 2011). Fast interfaces for missing data
imputation are provided using several forest-based algorithms (Tang and
Ishwaran, 2017).


\strong{Highlighted updates:}

\enumerate{
  
  \item For survival and competing risk analysis, concordance-based
  performance is now computed using Uno inverse-probability-of-censoring
  weighting (Uno et al. 2001). This affects all survival performance
  values derived from the concordance index, including out-of-bag and
  test error rates and variable importance (VIMP). To revert to
  unweighted Harrell concordance, set \code{use.uno = FALSE} when
  fitting a survival forest.  Concordance calculations now use an
  efficient \eqn{O(n \log n)} algorithm based on a binary indexed tree
  (Fenwick 1994, Therneau 2024), replacing the naive \eqn{O(n^2)}
  pairwise computation for large \eqn{n}.

  
  \item For variable selection, we recommend using \pkg{VarPro}, an R
  package for model-independent variable selection using rule-based
  variable priority. It supports regression, classification, survival
  analysis, and includes a new mode for unsupervised learning. See
  \url{https://www.varprotools.org} for more information.
  

  \item For computational speed, the default VIMP method has changed
  from "permute" (Breiman-Cutler permutation) to "anti"
  (\code{importance = "anti"} or \code{importance = TRUE}). While
  faster, this may be less accurate in settings such as highly
  imbalanced classification. To revert to permutation VIMP, use
  \code{importance = "permute"}.
}


This is the main entry point to the \pkg{randomForestSRC}
package. For more information on OpenMP support and the package as a
whole, see \code{package?randomForestSRC}.

}
\usage{
rfsrc(formula, data, ntree = 500,
  mtry = NULL, ytry = NULL,
  nodesize = NULL, nodedepth = NULL,
  splitrule = NULL, nsplit = NULL,
  importance = c(FALSE, TRUE, "none", "anti", "permute", "random"),
  block.size = if (any(is.element(as.character(importance),
                     c("none", "FALSE")))) NULL else 10,
  bootstrap = c("by.root", "none", "by.user"),
  samptype = c("swor", "swr"), samp = NULL, membership = FALSE,
  sampsize = if (samptype == "swor") function(x){x * .632} else function(x){x},
  na.action = c("na.omit", "na.impute"), nimpute = 1,
  ntime = 150, cause,
  perf.type = NULL,
  proximity = FALSE, distance = FALSE, forest.wt = FALSE,
  xvar.wt = NULL, yvar.wt = NULL, split.wt = NULL, case.wt  = NULL,
  case.depth = FALSE,
  forest = TRUE,
  use.uno = TRUE, save.memory = FALSE,
  var.used = c(FALSE, "all.trees", "by.tree"),
  split.depth = c(FALSE, "all.trees", "by.tree"),
  seed = NULL,
  do.trace = FALSE,
  ...)

## convenient interface for growing a CART tree
rfsrc.cart(formula, data, ntree = 1, mtry = ncol(data),
   bootstrap = "none", nsplit = 0, ...)
}

\arguments{

  \item{formula}{A formula describing the model to fit. Interaction terms are not supported. If missing, unsupervised splitting is used.}

  \item{data}{Data frame containing the response and predictor variables.}

  \item{ntree}{Number of trees to grow.}

  \item{mtry}{Number of candidate variables randomly selected at each split. Defaults: regression uses \code{p/3}, others use \code{sqrt(p)}; rounded up.}

  \item{ytry}{Number of pseudo-response variables randomly selected for unsupervised splitting. Default is 1.}

  \item{nodesize}{Minimum terminal node size. Defaults: survival/competing risks (15), regression (5), classification (1), mixed/unsupervised (3).}

  \item{nodedepth}{Maximum tree depth. Ignored by default.}

  \item{splitrule}{Splitting rule. See Details.}

  \item{nsplit}{Number of random split points per variable. \code{0} uses all values (deterministic). Default is 10.}

  \item{importance}{Variable importance (VIMP) method. Choices: \code{FALSE}, \code{TRUE}, \code{"none"}, \code{"anti"}, \code{"permute"}, \code{"random"}. Default is \code{"none"}. VIMP can be computed later using \code{vimp} or \code{predict}.}

  \item{block.size}{Controls frequency of cumulative error/VIMP updates. Default is \code{10} if importance is requested; otherwise \code{NULL}. See Details.}

  \item{bootstrap}{Bootstrap method. Options: \code{"by.root"} (default), \code{"by.user"}, or \code{"none"} (no OOB error possible).}

  \item{samptype}{Sampling type for \code{by.root} bootstrap. Options: \code{"swor"} (without replacement, default), \code{"swr"} (with replacement).}

  \item{samp}{Bootstrap weights (only for \code{bootstrap="by.user"}). A matrix of size \code{n x ntree} giving in-bag counts per tree.}

  \item{membership}{Return inbag and terminal node membership?}

  \item{sampsize}{Bootstrap sample size (used when \code{bootstrap="by.root"}). Defaults: 0.632 x n for \code{swor}; n for \code{swr}. Can also be numeric.}

  \item{na.action}{Missing data handling. \code{"na.omit"} (default) removes rows with any \code{NA}; \code{"na.impute"} performs fast internal imputation. See also \code{impute}.}

  \item{nimpute}{Number of iterations for internal imputation. If >1, OOB error rates may be optimistic.}

  \item{ntime}{For survival models: number or grid of time points used in ensemble estimation. If \code{NULL} or \code{0}, uses all event times.}

  \item{cause}{For competing risks: event of interest (1 to \code{J}), or a vector of weights over all \code{J} events. Defaults to an average over all events.}

  \item{perf.type}{Optional performance metric for prediction, VIMP, and error. Defaults to the family-specific metric. \code{"none"} disables performance. See Details.}

  \item{proximity}{Compute proximity matrix? Options: \code{"inbag"}, \code{"oob"}, \code{"all"}, \code{TRUE} (inbag), or \code{FALSE}.}

  \item{distance}{Compute pairwise distances between cases? Similar options as \code{proximity}. See Details.}

  \item{forest.wt}{Return forest weight matrix? Same options as \code{proximity}. Default is \code{TRUE} (inbag).}

  \item{xvar.wt}{Optional weights on x-variables for sampling at splits. Does not need to sum to 1. Defaults to uniform.}

  \item{yvar.wt}{Weights on response variables (for multivariate regression). Used when \code{y} is high-dimensional.}

  \item{split.wt}{Weights applied to each variable's split statistic. Higher weight increases likelihood of splitting.}

  \item{case.wt}{Sampling weights for cases in the bootstrap. Higher values increase selection probability. See class imbalance example.}

  \item{case.depth}{Return matrix recording depth of first split for each case? Default is \code{FALSE}.}

  \item{forest}{Save forest object for future prediction? Set \code{FALSE} if prediction is not needed.}

  \item{use.uno}{Logical.  If \code{TRUE} (default for survival), concordance-based performance is computed using inverse-probability-of-censoring weights as
  proposed by Uno et al. (2011). This affects all survival performance values derived from the C-index, including error rates (C-error) and variable
  importance (VIMP). Set \code{use.uno = FALSE} to use the unweighted Harrell concordance. Applies only to survival families.}

  \item{save.memory}{Reduce memory usage by avoiding storage of prediction quantities. Recommended for large survival or competing risk forests.}

  \item{var.used}{Return variable usage statistics? Options: \code{FALSE}, \code{"all.trees"}, \code{"by.tree"}.}

  \item{split.depth}{Return minimal depth of splits for each variable? Options: \code{FALSE}, \code{"all.trees"}, \code{"by.tree"}. See Details.}

  \item{seed}{Integer seed for reproducibility (negative values only).}

  \item{do.trace}{Print progress updates every \code{do.trace} seconds.}

  \item{...}{Additional arguments passed to or from other methods.}
}

\details{
  \enumerate{

  \item \emph{Types of forests}

  The type of forest is automatically inferred from the outcome and formula. Supported forest types include:

  \itemize{
    \item Regression forests for continuous outcomes.
    \item Classification forests for factor outcomes.
    \item Multivariate forests for continuous, categorical, or mixed outcomes.
    \item Unsupervised forests when no outcome is specified.
    \item Survival forests for right-censored time-to-event data.
    \item Competing risk forests for multi-event survival settings.
  }

  \item \emph{Splitting}

  \enumerate{
    \item Splitting rules are set using the \code{splitrule} option.
    \item Random splitting is invoked via \code{splitrule = "random"}.
    \item Use \code{nsplit} to enable randomized splitting and improve speed; see \emph{Improving computational speed}.
  }

  \item \emph{Available splitting rules}

  \itemize{

    \item \strong{Regression}
    \enumerate{
      \item \code{"mse"} (default): weighted mean squared error (Breiman et al., 1984).
      \item \code{"quantile.regr"}: quantile regression via check-loss; see \code{quantreg.rfsrc}.
      \item \code{"la.quantile.regr"}: local adaptive quantile regression.
    }

    \item \strong{Classification}
    \enumerate{
      \item \code{"gini"} (default): Gini index.
      \item \code{"auc"}: AUC-based splitting; appropriate for imbalanced data.
      \item \code{"entropy"}: entropy-based splitting.
    }

    \item \strong{Survival}
    \enumerate{
      \item \code{"logrank"} (default): log-rank splitting.
      \item \code{"bs.gradient"}: Brier score gradient splitting. Uses 90th percentile of observed times by default, or set \code{prob}.
      \item \code{"logrankscore"}: log-rank score splitting.
    }

    \item \strong{Competing risks} (see Ishwaran et al., 2014)
    \enumerate{
      \item \code{"logrankCR"} (default): Gray's test-based weighted log-rank splitting.
      \item \code{"logrank"}: cause-specific weighted log-rank; use \code{cause} to target specific events.
    }

    \item \strong{Multivariate}
    \enumerate{
      \item Default: normalized composite splitting (Tang and Ishwaran, 2017).
      \item \code{"mahalanobis"}: Mahalanobis splitting with optional covariance matrix; for multivariate regression.
    }

    \item \strong{Unsupervised}
    Splitting uses pseudo-outcomes and the composite rule. See \code{sidClustering} for advanced unsupervised analysis.

    \item \strong{Custom splitting}
    Custom rules can be defined using \code{splitCustom.c}. Up to 16 rules per family are allowed. Use \code{"custom"}, \code{"custom1"}, etc. Compilation required.
  }

  \item \emph{Improving computational speed}

  See \code{\link{rfsrc.fast}}. Strategies include:

  \itemize{
    \item Increase \code{nodesize}.
    \item Set \code{save.memory = TRUE} for large survival or competing risk models.
    \item Set \code{block.size = NULL} to avoid repeated cumulative error computation.
    \item Use \code{perf.type = "none"} to disable VIMP and C-index calculations.
    \item Set \code{nsplit} to a small integer (e.g., 1-10).
    \item Reduce bootstrap size with \code{sampsize}, \code{samptype}.
    \item Set \code{ntime} to a coarse grid (e.g., 50) for survival models.
    \item Pre-filter variables; use \code{max.subtree} for fast variable selection.
  }

  \item \emph{Prediction Error}

  Error is computed using OOB data:
  \itemize{
    \item Regression: mean squared error.
    \item Classification: misclassification rate, Brier score, AUC.
    \item Survival: C-error = 1 - Harrell's concordance index.
  }

  If \code{bootstrap = "none"}, OOB error is unavailable. Use \code{predict.rfsrc} for cross-validation error instead.

  \item \emph{Variable Importance (VIMP)}

  VIMP methods:
  \itemize{
    \item \code{"permute"}: permutation VIMP (Breiman-Cutler).
    \item \code{"random"}: randomized left/right assignment.
    \item \code{"anti"} (default): anti-split assignment.
  }

  The \code{block.size} option controls granularity. For confidence intervals, see \code{subsampling}. Also see \code{holdout.vimp} for a more conservative variant.

  \item \emph{Multivariate Forests}

  Use:
  \preformatted{rfsrc(Multivar(y1, ..., yd) ~ ., data)}
  or
  \preformatted{rfsrc(cbind(y1, ..., yd) ~ ., data)}

  Use \code{get.mv.formula}, \code{get.mv.predicted}, \code{get.mv.error} for multivariate extraction.

  \item \emph{Unsupervised Forests}

  Use:
  \preformatted{rfsrc(data = X)}
  or
  \preformatted{rfsrc(Unsupervised(ytry) ~ ., data = X)}

  Random subsets of \code{ytry} pseudo-responses are used for each \code{mtry} variable. No performance metrics are computed.

  \item \emph{Survival, Competing Risks}

  \itemize{
    \item Survival: use \code{Surv(time, status) ~ .}. Status must be 0 (censored) or 1 (event).
    \item Competing risks: status = 0 (censored), 1-J (event types). Use \code{cause} to target specific events.
    \item Larger \code{nodesize} is typically needed for competing risks.
  }

  \item \emph{Missing data imputation}

  Use \code{na.action = "na.impute"}. Iteration with \code{nimpute > 1} replaces missing values using OOB predictions. Observations or variables with all missing values are removed.

  \item \emph{Allowable data types and factors}

  Variables must be numeric, integer, factor, or logical. Non-factors are coerced to numeric. For unordered factors, all complementary subsets are considered for splits.

  Factor levels are mapped to ensure consistency across training/test data. Consider converting factors to numeric for high-dimensional settings.
}}

\value{
An object of class \code{(rfsrc, grow)} with the following components:

\describe{
  \item{call}{The original call to \code{rfsrc}.}
  \item{family}{The family used in the analysis.}
  \item{n}{Sample size after applying \code{na.action}.}
  \item{ntree}{Number of trees grown.}
  \item{mtry}{Number of variables randomly selected at each node.}
  \item{nodesize}{Minimum terminal node size.}
  \item{nodedepth}{Maximum depth allowed for each tree.}
  \item{splitrule}{Splitting rule used.}
  \item{nsplit}{Number of random split points.}
  \item{yvar}{Response values.}
  \item{yvar.names}{Character vector of response variable names.}
  \item{xvar}{Data frame of predictor variables.}
  \item{xvar.names}{Character vector of predictor variable names.}
  \item{xvar.wt}{Non-negative weights specifying the selection probability of each variable.}
  \item{split.wt}{Non-negative weights adjusting each variable's split statistic.}
  \item{cause.wt}{Weights for composite competing risk splitting.}
  \item{leaf.count}{Number of terminal nodes per tree. A value of 0 indicates a rejected tree (may occur with missing data); a value of 1 indicates a stump.}
  \item{proximity}{Proximity matrix indicating how often case pairs fall in the same terminal node.}
  \item{forest}{Forest object, returned if \code{forest=TRUE}. Required for prediction and most wrappers.}
  \item{forest.wt}{Forest weight matrix.}
  \item{membership}{Terminal node membership matrix (rows: trees; columns: cases).}
  \item{inbag}{Inbag count matrix (rows: trees; columns: cases).}
  \item{var.used}{Number of times each variable is used to split a node.}
  \item{imputed.indv}{Indices of individuals with missing values.}
  \item{imputed.data}{Imputed dataset with responses followed by predictors.}
  \item{split.depth}{Matrix or array recording minimal split depth of variables by case and tree.}
  \item{err.rate}{Cumulative OOB error rate.}
  \item{err.block.rate}{Cumulative error per ensemble block (size defined by \code{block.size}). If \code{block.size = 1}, error per tree.}
  \item{importance}{Variable importance (VIMP) for each predictor.}
  \item{predicted}{In-bag predicted values.}
  \item{predicted.oob}{Out-of-bag (OOB) predicted values.}

  \item{class}{(Classification) In-bag predicted class labels.}
  \item{class.oob}{(Classification) OOB predicted class labels.}

  \item{regrOutput}{(Multivariate) List of performance results for continuous outcomes.}
  \item{clasOutput}{(Multivariate) List of performance results for categorical outcomes.}

  \item{survival}{(Survival) In-bag survival functions.}
  \item{survival.oob}{(Survival) OOB survival functions.}
  \item{chf}{(Survival or competing risks) In-bag cumulative hazard function.}
  \item{chf.oob}{(Survival or competing risks) OOB cumulative hazard function.}
  \item{time.interest}{(Survival or competing risks) Unique sorted event times.}
  \item{ndead}{(Survival or competing risks) Total number of observed events.}

  \item{cif}{(Competing risks) In-bag cumulative incidence function by cause.}
  \item{cif.oob}{(Competing risks) OOB cumulative incidence function by cause.}
}
}

\note{
Values returned by the forest depend on the family:

\itemize{
  \item \strong{Regression}: \code{predicted} and \code{predicted.oob} are vectors of predicted values.
  
  \item \strong{Classification}: \code{predicted} and \code{predicted.oob} are matrices of class probabilities. VIMP and performance metrics are returned as a matrix with \code{J+1} columns (J = number of classes). The first column ("all") gives unconditional results; remaining columns give class-conditional results.
  
  \item \strong{Survival}: \code{predicted} contains mortality estimates (Ishwaran et al., 2008). These are calibrated to the number of expected events under identical covariate profiles. Also returned are matrices of the survival function and CHF for each individual over \code{time.interest}.
  
  \item \strong{Competing risks}: \code{predicted} contains expected life years lost by cause (Ishwaran et al., 2013). Also returned are three-dimensional arrays for CIF and CSCHF indexed by case, time, and event type.
  
  \item \strong{Multivariate}: Predictions, VIMP, and error rates are returned in \code{regrOutput} and \code{clasOutput}. Use \code{get.mv.predicted}, \code{get.mv.vimp}, and \code{get.mv.error} to extract results.
}
}

  
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}

\references{
  Breiman L., Friedman J.H., Olshen R.A. and Stone C.J. (1984).
  \emph{Classification and Regression Trees}, Belmont, California.
  
  Breiman L. (2001). Random forests, \emph{Machine Learning}, 45:5-32.
  
  Cutler A. and Zhao G. (2001). PERT-Perfect random tree ensembles.
  \emph{Comp. Sci. Statist.}, 33: 490-497.

  Dietterich, T. G. (2000). An experimental comparison of three methods for
  constructing ensembles of decision trees: bagging, boosting, and randomization.
  \emph{Machine Learning}, 40, 139-157.

  Fenwick, P.M. (1994).  A new data structure for cumulative frequency
  tables.  \emph{Software: Practice and Experience}, 24(3):327--336.

  Gray R.J. (1988).  A class of k-sample tests for comparing the
  cumulative incidence of a competing risk, \emph{Ann. Statist.},
  16: 1141-1154.
  
  Geurts, P., Ernst, D. and Wehenkel, L., (2006). Extremely randomized
  trees. \emph{Machine learning}, 63(1):3-42.

  Greenwald M. and Khanna S. (2001).  Space-efficient online computation of
  quantile summaries. \emph{Proceedings of ACM SIGMOD}, 30(2):58-66.

  Harrell et al. F.E. (1982).  Evaluating the yield of medical tests,
  \emph{J. Amer. Med. Assoc.}, 247:2543-2546.
  
  Hothorn T. and Lausen B. (2003). On the exact distribution of maximally selected
  rank statistics, \emph{Comp. Statist. Data Anal.}, 43:121-137.
  
  Ishwaran H. (2007).  Variable importance in binary regression
  trees and forests,  \emph{Electronic J. Statist.}, 1:519-537.
  
  Ishwaran H. and Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
  
  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S.
  (2008).  Random survival forests, \emph{Ann. App.
    Statist.}, 2:841-860.
  
  Ishwaran H., Kogalur U.B., Gorodeski E.Z, Minn A.J. and
  Lauer M.S. (2010).  High-dimensional variable selection for survival
  data.  \emph{J. Amer. Statist. Assoc.}, 105:205-217.
  
  Ishwaran H., Kogalur U.B., Chen X. and Minn A.J. (2011). Random survival
  forests for high-dimensional data. \emph{Stat. Anal. Data Mining}, 4:115-132
  
  Ishwaran H., Gerds T.A., Kogalur U.B., Moore R.D., Gange S.J. and Lau
  B.M. (2014). Random survival forests for competing risks.
  \emph{Biostatistics}, 15(4):757-773.
  
  Ishwaran H. and Malley J.D. (2014). Synthetic learning
  machines. \emph{BioData Mining}, 7:28.

  Ishwaran H. (2015).  The effect of splitting on random forests.
  \emph{Machine Learning}, 99:75-118.

  Lin, Y. and Jeon, Y. (2006). Random forests and adaptive nearest
  neighbors. \emph{J. Amer. Statist. Assoc.}, 101(474), 578-590.
  
  Lu M., Sadiq S., Feaster D.J. and Ishwaran H. (2018). Estimating
  individual treatment effect in observational data using random forest
  methods. \emph{J. Comp. Graph. Statist}, 27(1), 209-219

  Ishwaran H. and Lu M.  (2019).  Standard errors and confidence
  intervals for variable importance in random forest regression,
  classification, and survival. \emph{Statistics in Medicine}, 38,
  558-582.
  
  LeBlanc M. and Crowley J. (1993).  Survival trees by goodness of split,
  \emph{J. Amer. Statist. Assoc.}, 88:457-467.
  
  Loh W.-Y and Shih Y.-S (1997).  Split selection methods for
  classification trees, \emph{Statist. Sinica}, 7:815-840.

  Mantero A. and Ishwaran H. (2021).  Unsupervised random forests.
  \emph{Statistical Analysis and Data Mining}, 14(2):144-167.

  Meinshausen N. (2006) Quantile regression forests, \emph{Journal of
  Machine Learning Research}, 7:983-999.

  Mogensen, U.B, Ishwaran H. and Gerds T.A. (2012). Evaluating random
  forests for survival analysis using prediction error curves,
  \emph{J.  Statist. Software}, 50(11): 1-23.
  
  O'Brien R. and Ishwaran H. (2019).  A random forests quantile
  classifier for class imbalanced data. \emph{Pattern Recognition},
  90, 232-249

  Segal M.R. (1988).  Regression trees for censored data,
  \emph{Biometrics},  44:35-47.

  Segal M.R. and Xiao Y. Multivariate random
  forests. (2011). \emph{Wiley Interdisciplinary Reviews: Data Mining
  and Knowledge Discovery}. 1(1):80-87.
  
  Tang F. and Ishwaran H. (2017).  Random forest missing data
  algorithms.  \emph{Statistical Analysis and Data Mining}, 10:363-377.

  Therneau, T.M. (2024).  Concordance.  Vignette in the \pkg{survival} R
  package.
  
  Uno, H., Cai, T., Pencina, M.J., D'Agostino, Ralph B. and Wei, L-J. (2011).
  On the C-statistics for evaluating overall adequacy of risk prediction procedures with censored survival data.
  \emph{Statistics in Medicine}, 30:1105-1117.

  Zhang H., Zimmerman J., Nettleton D. and Nordman D.J. (2019).  Random
  forest prediction intervals. \emph{The American Statistician}. 4:1-5.
  

}
\seealso{

  \command{\link{get.tree.rfsrc}},

  \command{\link{holdout.vimp.rfsrc}},

  \command{\link{imbalanced.rfsrc}},
  \command{\link{impute.rfsrc}},

  \command{\link{max.subtree.rfsrc}},

  \command{\link{partial.rfsrc}},
  \command{\link{plot.competing.risk.rfsrc}},
  \command{\link{plot.rfsrc}},
  \command{\link{plot.survival.rfsrc}},
  \command{\link{plot.variable.rfsrc}},
  \command{\link{predict.rfsrc}},
  \command{\link{print.rfsrc}},

  \command{\link{quantreg.rfsrc}},

  \command{\link{rfsrc}},
  \command{\link{rfsrc.anonymous}},
  \command{\link{rfsrc.cart}},
  \command{\link{rfsrc.fast}},

  
  \command{\link{sidClustering.rfsrc}},

  \command{\link{subsample.rfsrc}},
  
  \command{\link{tune.rfsrc}},

  \command{\link{vimp.rfsrc}}
}
\examples{
\donttest{
##------------------------------------------------------------
## survival analysis
##------------------------------------------------------------

## veteran data
## randomized trial of two treatment regimens for lung cancer
data(veteran, package = "randomForestSRC")
v.obj <- rfsrc(Surv(time, status) ~ ., data = veteran, block.size = 1)

## plot tree number 3
plot(get.tree(v.obj, 3))

## print results of trained forest
print(v.obj)

## plot results of trained forest
plot(v.obj)

## plot survival curves for first 10 individuals -- direct way
matplot(v.obj$time.interest, 100 * t(v.obj$survival.oob[1:10, ]),
    xlab = "Time", ylab = "Survival", type = "l", lty = 1)

## plot survival curves for first 10 individuals
## using function "plot.survival" 
plot.survival(v.obj, subset = 1:10)

## Brier and AUCT helpers with pointwise intervals
## see plot.survival for more details
brier.obj <- get.brier.survival(v.obj, conf.int = TRUE)
head(brier.obj$brier.score)
auct.obj <- get.auct.survival(v.obj, conf.int = TRUE)
head(auct.obj$auct.score)

## Brier and AUCT curves with confidence bands
## see plot.survival for more details
plotBrierAUC(v.obj, xlim = c(0, 600))
plotBrierAUC(v.obj, plots = "brier", xlim = c(0, 600))
plotBrierAUC(v.obj, plots = "auct", xlim = c(0, 600))

## plot CRPS (continuous rank probability score) as function of time
## here's how to calculate the CRPS for every time point
trapz <- randomForestSRC:::trapz
time <- v.obj$time.interest
bscore <- brier.obj$brier.score
crps <- sapply(1:length(time), function(j) {
  trapz(time[1:j], bscore[1:j, 2] / diff(range(time[1:j])))
})
plot(time, crps, ylab = "CRPS", type = "s", lwd = 2)


## fast nodesize optimization for veteran data
## optimal nodesize in survival is larger than other families
## see the function "tune" for more examples
tune.nodesize(Surv(time,status) ~ ., veteran)


## Primary biliary cirrhosis (PBC) of the liver
data(pbc, package = "randomForestSRC")
pbc.obj <- rfsrc(Surv(days, status) ~ ., pbc)
print(pbc.obj)


## save.memory example for survival
## growing many deep trees creates memory issue without this option!
data(pbc, package = "randomForestSRC")
print(rfsrc(Surv(days, status) ~ ., pbc, splitrule = "random",
            ntree = 25000, nodesize = 1, save.memory = TRUE))



##------------------------------------------------------------
## trees can be plotted for any family 
## see get.tree for details and more examples
##------------------------------------------------------------

## survival where factors have many levels
data(veteran, package = "randomForestSRC")
vd <- veteran
vd$celltype=factor(vd$celltype)
vd$diagtime=factor(vd$diagtime)
vd.obj <- rfsrc(Surv(time,status)~., vd, ntree = 100, nodesize = 5)
plot(get.tree(vd.obj, 3))

## classification
iris.obj <- rfsrc(Species ~., data = iris)
plot(get.tree(iris.obj, 25, class.type = "bayes"))
plot(get.tree(iris.obj, 25, target = "setosa"))
plot(get.tree(iris.obj, 25, target = "versicolor"))
plot(get.tree(iris.obj, 25, target = "virginica"))

## ------------------------------------------------------------
## simple example of VIMP using iris classification
## ------------------------------------------------------------

## directly from trained forest
print(rfsrc(Species~.,iris,importance=TRUE)$importance)

## VIMP (and performance) use misclassification error by default
## but brier prediction error can be requested
print(rfsrc(Species~.,iris,importance=TRUE,perf.type="brier")$importance)

## example using vimp function (see vimp help file for details)
iris.obj <- rfsrc(Species ~., data = iris)
print(vimp(iris.obj)$importance)
print(vimp(iris.obj,perf.type="brier")$importance)

## example using hold out vimp (see holdout.vimp help file for details)
print(holdout.vimp(Species~.,iris)$importance)
print(holdout.vimp(Species~.,iris,perf.type="brier")$importance)

## ------------------------------------------------------------
## confidence interval for vimp using subsampling
## compare with holdout vimp
## ------------------------------------------------------------

## new York air quality measurements
o <- rfsrc(Ozone ~ ., data = airquality)
so <- subsample(o)
plot(so)

## compare with holdout vimp
print(holdout.vimp(Ozone ~ ., data = airquality)$importance)


##------------------------------------------------------------
## example of imputation in survival analysis
##------------------------------------------------------------

data(pbc, package = "randomForestSRC")
pbc.obj2 <- rfsrc(Surv(days, status) ~ ., pbc, na.action = "na.impute")

## same as above but iterate the missing data algorithm
pbc.obj3 <- rfsrc(Surv(days, status) ~ ., pbc,
         na.action = "na.impute", nimpute = 3)

## fast way to impute data (no inference is done)
## see impute for more details
pbc.imp <- impute(Surv(days, status) ~ ., pbc, splitrule = "random")

##------------------------------------------------------------
## compare RF-SRC to Cox regression
## Illustrates C-error and Brier score measures of performance
## assumes "pec" and "survival" libraries are loaded
##------------------------------------------------------------

if (library("survival", logical.return = TRUE)
    & library("pec", logical.return = TRUE)
    & library("prodlim", logical.return = TRUE))

{
  ##prediction function required for pec
  predictSurvProb.rfsrc <- function(object, newdata, times, ...){
    ptemp <- predict(object,newdata=newdata,...)$survival
    pos <- sindex(jump.times = object$time.interest, eval.times = times)
    p <- cbind(1,ptemp)[, pos + 1]
    if (NROW(p) != NROW(newdata) || NCOL(p) != length(times))
      stop("Prediction failed")
    p
  }

  ## data, formula specifications
  data(pbc, package = "randomForestSRC")
  pbc.na <- na.omit(pbc)  ##remove NA's
  surv.f <- as.formula(Surv(days, status) ~ .)
  pec.f <- as.formula(Hist(days,status) ~ 1)

  ## run cox/rfsrc models
  ## for illustration we use a small number of trees
  cox.obj <- coxph(surv.f, data = pbc.na, x = TRUE)
  rfsrc.obj <- rfsrc(surv.f, pbc.na, ntree = 150)

  ## compute bootstrap cross-validation estimate of expected Brier score
  ## see Mogensen, Ishwaran and Gerds (2012) Journal of Statistical Software
  set.seed(17743)
  prederror.pbc <- pec(list(cox.obj,rfsrc.obj), data = pbc.na, formula = pec.f,
                        splitMethod = "bootcv", B = 50)
  print(prederror.pbc)
  plot(prederror.pbc)

  ## compute out-of-bag C-error for cox regression and compare to rfsrc
  ## use uno weights throughout
  rfsrc.obj <- rfsrc(surv.f, pbc.na)
  uno.wt <- rfsrc.obj$forest$uno.weights$weight
  cat("out-of-bag Cox Analysis ...", "\n")
  cox.err <- sapply(1:100, function(b) {
    if (b\%\%10 == 0) cat("cox bootstrap:", b, "\n")
    train <- sample(1:nrow(pbc.na), nrow(pbc.na), replace = TRUE)
    cox.obj <- tryCatch({coxph(surv.f, pbc.na[train, ])}, error=function(ex){NULL})
    if (!is.null(cox.obj)) {
      get.cindex(pbc.na$days[-train],
                 pbc.na$status[-train],
                 predict(cox.obj, pbc.na[-train, ]),
                 weight=uno.wt[-train])
    } else NA
  })
  cat("\n\tOOB error rates (with Uno-IPCW):\n")
  cat("\tRSF            : ", get.mv.error(rfsrc.obj), "\n")
  cat("\tCox regression : ", mean(cox.err, na.rm = TRUE), "\n")
}

##------------------------------------------------------------
## competing risks
##------------------------------------------------------------

## WIHS analysis
## cumulative incidence function (CIF) for HAART and AIDS stratified by IDU

data(wihs, package = "randomForestSRC")
wihs.obj <- rfsrc(Surv(time, status) ~ ., wihs, nsplit = 3, ntree = 100)
plot.competing.risk(wihs.obj)
cif <- wihs.obj$cif.oob
Time <- wihs.obj$time.interest
idu <- wihs$idu
cif.haart <- cbind(apply(cif[,,1][idu == 0,], 2, mean),
                   apply(cif[,,1][idu == 1,], 2, mean))
cif.aids  <- cbind(apply(cif[,,2][idu == 0,], 2, mean),
                   apply(cif[,,2][idu == 1,], 2, mean))
matplot(Time, cbind(cif.haart, cif.aids), type = "l",
        lty = c(1,2,1,2), col = c(4, 4, 2, 2), lwd = 3,
        ylab = "Cumulative Incidence")
legend("topleft",
       legend = c("HAART (Non-IDU)", "HAART (IDU)", "AIDS (Non-IDU)", "AIDS (IDU)"),
       lty = c(1,2,1,2), col = c(4, 4, 2, 2), lwd = 3, cex = 1.5)


## illustrates the various splitting rules
## illustrates event specific and non-event specific variable selection
if (library("survival", logical.return = TRUE)) {

  ## use the pbc data from the survival package
  ## events are transplant (1) and death (2)
  data(pbc, package = "survival")
  pbc$id <- NULL

  ## modified Gray's weighted log-rank splitting
  ## (equivalent to cause=c(1,1) and splitrule="logrankCR")
  pbc.cr <- rfsrc(Surv(time, status) ~ ., pbc)
 
  ## log-rank cause-1 specific splitting and targeted VIMP for cause 1
  pbc.log1 <- rfsrc(Surv(time, status) ~ ., pbc, 
              splitrule = "logrankCR", cause = c(1,0), importance = TRUE)

  ## log-rank cause-2 specific splitting and targeted VIMP for cause 2
  pbc.log2 <- rfsrc(Surv(time, status) ~ ., pbc, 
              splitrule = "logrankCR", cause = c(0,1), importance = TRUE)

  ## extract VIMP from the log-rank forests: event-specific
  ## extract minimal depth from the Gray log-rank forest: non-event specific
  var.perf <- data.frame(md = max.subtree(pbc.cr)$order[, 1],
                         vimp1 = 100 * pbc.log1$importance[ ,1],
                         vimp2 = 100 * pbc.log2$importance[ ,2])
  print(var.perf[order(var.perf$md), ], digits = 2)

}

## ------------------------------------------------------------
## regression analysis
## ------------------------------------------------------------

## new York air quality measurements
airq.obj <- rfsrc(Ozone ~ ., data = airquality, na.action = "na.impute")

# partial plot of variables (see plot.variable for more details)
plot.variable(airq.obj, partial = TRUE, smooth.lines = TRUE)

## motor trend cars
mtcars.obj <- rfsrc(mpg ~ ., data = mtcars)

## ------------------------------------------------------------
## regression with custom bootstrap
## ------------------------------------------------------------

ntree <- 25
n <- nrow(mtcars)
s.size <- n / 2
swr <- TRUE
samp <- randomForestSRC:::make.sample(ntree, n, s.size, swr)
o <- rfsrc(mpg ~ ., mtcars, bootstrap = "by.user", samp = samp)

## ------------------------------------------------------------
## classification analysis
## ------------------------------------------------------------

## iris data
iris.obj <- rfsrc(Species ~., data = iris)

## wisconsin prognostic breast cancer data
data(breast, package = "randomForestSRC")
breast.obj <- rfsrc(status ~ ., data = breast, block.size=1)
plot(breast.obj)

## ------------------------------------------------------------
## big data set, reduce number of variables using simple method
## ------------------------------------------------------------

## use Iowa housing data set
data(housing, package = "randomForestSRC")

## original data contains lots of missing data, use fast imputation
## however see impute for other methods
housing2 <- impute(data = housing, fast = TRUE)

## run shallow trees to find variables that split any tree
xvar.used <- rfsrc(SalePrice ~., housing2, ntree = 250, nodedepth = 4,
                   var.used="all.trees", mtry = Inf, nsplit = 100)$var.used

## now fit forest using filtered variables
xvar.keep  <- names(xvar.used)[xvar.used >= 1]
o <- rfsrc(SalePrice~., housing2[, c("SalePrice", xvar.keep)])
print(o)

## ------------------------------------------------------------
## imbalanced classification data
## see the "imbalanced" function for further details
##
## (a) use balanced random forests with undersampling of the majority class
## Specifically let n0, n1 be sample sizes for majority, minority
## cases.  We sample 2 x n1 cases with majority, minority cases chosen
## with probabilities n1/n, n0/n where n=n0+n1
##
## (b) balanced random forests using "imbalanced"
##
## (c) q-classifier (RFQ) using "imbalanced" 
##
## ------------------------------------------------------------

## Wisconsin breast cancer example
data(breast, package = "randomForestSRC")
breast <- na.omit(breast)

## balanced random forests - brute force
y <- breast$status
obdirect <- rfsrc(status ~ ., data = breast, nsplit = 10,
            case.wt = randomForestSRC:::make.wt(y),
            sampsize = randomForestSRC:::make.size(y))
print(obdirect)
print(get.imbalanced.performance(obdirect))

## balanced random forests - using "imbalanced" 
ob <- imbalanced(status ~ ., data = breast, method = "brf")
print(ob)
print(get.imbalanced.performance(ob))

## q-classifier (RFQ) - using "imbalanced" 
oq <- imbalanced(status ~ ., data = breast)
print(oq)
print(get.imbalanced.performance(oq))

## q-classifier (RFQ) - with auc splitting
oqauc <- imbalanced(status ~ ., data = breast, splitrule = "auc")
print(oqauc)
print(get.imbalanced.performance(oqauc))

## ------------------------------------------------------------
## unsupervised analysis
## ------------------------------------------------------------

## two equivalent ways to implement unsupervised forests
mtcars.unspv <- rfsrc(Unsupervised() ~., data = mtcars)
mtcars2.unspv <- rfsrc(data = mtcars)


## illustration of sidClustering for the mtcars data
## see sidClustering for more details
mtcars.sid <- sidClustering(mtcars, k = 1:10)
print(split(mtcars, mtcars.sid$cl[, 3]))
print(split(mtcars, mtcars.sid$cl[, 10]))


## ------------------------------------------------------------
## bivariate regression using Mahalanobis splitting
## also illustrates user specified covariance matrix
## ------------------------------------------------------------

if (library("mlbench", logical.return = TRUE)) {

  ## load boston housing data, specify the bivariate regression
  data(BostonHousing)
  f <- formula("Multivar(lstat, nox) ~.")
  
  ## Mahalanobis splitting  
  bh.mreg <- rfsrc(f, BostonHousing, importance = TRUE, splitrule = "mahal")
  
  ## performance error and vimp
  vmp <- get.mv.vimp(bh.mreg)
  pred <- get.mv.predicted(bh.mreg)

  ## standardized error and vimp
  err.std <- get.mv.error(bh.mreg, standardize = TRUE)
  vmp.std <- get.mv.vimp(bh.mreg, standardize = TRUE)

  ## same analysis, but with user specified covariance matrix
  sigma <- cov(BostonHousing[, c("lstat","nox")])
  bh.mreg2 <- rfsrc(f, BostonHousing, splitrule = "mahal", sigma = sigma)
  
}

## ------------------------------------------------------------
## multivariate mixed forests (nutrigenomic study)
## study effects of diet, lipids and gene expression for mice
## diet, genotype and lipids used as the multivariate y
## genes used for the x features
## ------------------------------------------------------------

## load the data (data is a list)
data(nutrigenomic, package = "randomForestSRC")

## assemble the multivariate y data
ydta <- data.frame(diet = nutrigenomic$diet,
                   genotype = nutrigenomic$genotype,
                   nutrigenomic$lipids)

## multivariate mixed forest call
## uses "get.mv.formula" for conveniently setting formula
mv.obj <- rfsrc(get.mv.formula(colnames(ydta)),
                data.frame(ydta, nutrigenomic$genes),
		importance=TRUE, nsplit = 10)

## print results for diet and genotype y values	    
print(mv.obj, outcome.target = "diet")
print(mv.obj, outcome.target = "genotype")

## extract standardized VIMP
svimp <- get.mv.vimp(mv.obj, standardize = TRUE)

## plot standardized VIMP for diet, genotype and lipid for each gene
boxplot(t(svimp), col = "bisque", cex.axis = .7, las = 2,
        outline = FALSE,
        ylab = "standardized VIMP",
        main = "diet/genotype/lipid VIMP for each gene")
	

## ------------------------------------------------------------
## illustrates yvar.wt which sets the probability of selecting 
## the response variables in multivariate regression
## ------------------------------------------------------------

## use mtcars: add fake responses
mult.mtcars  <- cbind(mtcars, mtcars$mpg, mtcars$mpg)
names(mult.mtcars) = c(names(mtcars), "mpg2", "mpg3")

## noise up the fake responses
mult.mtcars$mpg2 <- sample(mtcars$mpg)
mult.mtcars$mpg3 <- sample(mtcars$mpg)

formula = as.formula(Multivar(mpg, mpg2, mpg3)  ~ .)

## select 2 of the 3 responses randomly at each split with an associated weight vector.
## choose the noisy y responses which should degrade performance
yvar.wt = c(0.000001, 0.5, 0.5)
ytry = 2

mult.grow <- rfsrc(formula = formula, data = mult.mtcars, ytry = ytry, yvar.wt = yvar.wt)

print(mult.grow)
print(get.mv.error(mult.grow))

## Also, compare the following two results, as they should be similar:
yvar.wt = c(1.0, 00000.1, 00000.1)
ytry = 1

result1 = rfsrc(formula = formula, data = mult.mtcars, ytry = ytry, yvar.wt = yvar.wt)
result2 = rfsrc(mpg ~ ., mtcars)

print(get.mv.error(result1))
print(get.mv.error(result2))

## ------------------------------------------------------------
## custom splitting using the pre-coded examples
## ------------------------------------------------------------

## motor trend cars
mtcars.obj <- rfsrc(mpg ~ ., data = mtcars, splitrule = "custom")

## iris analysis
iris.obj <- rfsrc(Species ~., data = iris, splitrule = "custom1")

## WIHS analysis
wihs.obj <- rfsrc(Surv(time, status) ~ ., wihs, nsplit = 3,
                  ntree = 100, splitrule = "custom1")

}
}
\keyword{forest}

\name{breast}
\docType{data}
\alias{breast}
\title{Wisconsin Prognostic Breast Cancer Data}
\description{
  Recurrence of breast cancer from 198 breast cancer patients, all of
  which exhibited no evidence of distant metastases at the time of
  diagnosis.  The first 30 features of the data describe characteristics
  of the cell nuclei present in the digitized image of a fine needle
  aspirate (FNA) of the breast mass.
}
\source{
  The data were obtained from the UCI machine learning repository, see
  \url{http://archive.ics.uci.edu/ml/datasets/Breast+Cancer+Wisconsin+(Prognostic)}.
}
\examples{
\donttest{
## ------------------------------------------------------------
## Standard analysis
## ------------------------------------------------------------

data(breast, package = "randomForestSRC")
breast <- na.omit(breast)
o <- rfsrc(status ~ ., data = breast, nsplit = 10)
print(o)

}}
\keyword{datasets}

\name{housing}
\docType{data}
\alias{housing}
\title{Ames Iowa Housing Data}

\description{
  Data from the Ames Assessor's Office used in assessing values of
  individual residential properties sold in Ames, Iowa from 2006 to
  2010.  This is a regression problem and the goal is to predict 
  "SalePrice" which records the price of a home in thousands of dollars.
}
\references{
  De Cock, D., (2011). Ames, Iowa: Alternative to the Boston housing
  data as an end of semester regression project. \emph{Journal of Statistics
  Education}, 19(3), 1--14. 
}
\examples{
\donttest{
## load the data
data(housing, package = "randomForestSRC")

## the original data contains lots of missing data, so impute it
## use missForest, can be slow so grow trees with small training sizes
housing2 <- impute(data = housing, mf.q = 1, sampsize = function(x){x * .1})

## same idea ... but directly use rfsrc.fast and multivariate missForest 
housing3 <- impute(data = housing, mf.q = .5, fast = TRUE)

## even faster, but potentially less acurate
housing4 <- impute(SalePrice~., housing, splitrule = "random", nimpute = 1)


}}
\keyword{datasets}

\name{imbalanced.rfsrc}
\alias{imbalanced.rfsrc}
\alias{imbalanced}
\title{Imbalanced Two Class Problems}

\description{
  Implements various solutions to the two-class imbalanced problem,
  including the newly proposed quantile-classifier approach of
  O'Brien and Ishwaran (2017).  Also includes Breiman's balanced random
  forests undersampling of the majority class.  Performance is assesssed
  using the G-mean, but misclassification error can be requested.
}

\usage{\method{imbalanced}{rfsrc}(formula, data, ntree = 3000, 
  method = c("rfq", "brf", "standard"), splitrule = "auc",
  perf.type = NULL, block.size = NULL, fast = FALSE,
  ratio = NULL, ...)
}

\arguments{

  \item{formula}{A symbolic description of the model to be fit.}

  \item{data}{A data frame containing the two-class y-outcome and x-variables.}

  \item{ntree}{Number of trees to grow.}

  \item{method}{Method used to fit the classifier. The default is
  \code{"rfq"}, which implements the random forest quantile classifier
  (RFQ) of O'Brien and Ishwaran (2017). The option \code{"brf"} applies
  the balanced random forest (BRF) approach of Chen et al. (2004), which
  undersamples the majority class to match the minority class size. The
  option \code{"standard"} performs a standard random forest analysis.}


  \item{splitrule}{Splitting rule used to grow trees. The default is
  \code{"auc"}, which optimizes G-mean performance. Other supported
  options are \code{"gini"} and \code{"entropy"}.}


  \item{perf.type}{Performance metric used for evaluating the classifier
  and computing downstream quantities such as VIMP. Defaults depend on
  the method: \code{"gmean"} for RFQ and BRF; \code{"misclass"}
  (misclassification error) for standard random forests. Users may
  override this by specifying \code{"gmean"}, \code{"misclass"}, or
  \code{"brier"} (normalized Brier score). See examples for usage.}


  \item{block.size}{Controls how the cumulative error rate is
  computed. If \code{NULL}, it is calculated only once for the final
  tree. If set to an integer, cumulative error and VIMP are computed in
  blocks of that size. If unspecified, uses the default in
  \code{\link{rfsrc}}.}


  \item{fast}{Logical. If \code{TRUE}, uses the fast random forest
  implementation via \code{\link{rfsrc.fast}} instead of
  \code{rfsrc}. Improves speed at the cost of accuracy. Applies only to
  RFQ.}


  \item{ratio}{Optional and experimental. Specifies the proportion
  (between 0 and 1) of majority class cases to sample during RFQ
  training. Sampling is without replacement. Ignored for BRF.}

  
  \item{...}{Additional arguments passed to \code{\link{rfsrc}} to
    control random forest behavior.}
  

}

\details{

Imbalanced data, also known as the minority class problem, refers to
two-class classification settings where the majority class significantly
outnumbers the minority class. This function supports two approaches to
address class imbalance:


\itemize{
  \item The random forests quantile classifier (RFQ) proposed by O'Brien and Ishwaran (2017).
  \item The balanced random forest (BRF) undersampling method of Chen et al. (2004).
}

By default, the performance metric used is the G-mean (Kubat et al.,
1997), which balances sensitivity and specificity.


\strong{Handling of missing values:} Missing data are not supported for
BRF or when the \code{ratio} option is specified. In these cases,
records with missing values are removed prior to analysis.


\strong{Variable importance:} Permutation-based VIMP is used by default
in this setting, in contrast to anti-VIMP which is the default for other
families. Empirical results suggest that permutation VIMP is more
reliable in highly imbalanced settings.


\strong{Tree count recommendation:} We recommend using a relatively
large value for \code{ntree} in imbalanced problems to ensure stable
performance estimation, especially for G-mean. As a general guideline,
use at least five times the usual number of trees.

\strong{Performance metrics:} A helper function,
\code{\link{get.imbalanced.performance}}, is provided for extracting
classification performance summaries. The metric names are
self-explanatory in most cases. Some key metrics include:
\itemize{
  \item \code{F1}: The harmonic mean of precision and recall.
  \item \code{F1mod}: The harmonic mean of sensitivity, specificity,
  precision, and negative predictive value.
  \item \code{F1gmean}: The average of F1 and G-mean.
  \item \code{F1modgmean}: The average of F1mod and G-mean.
}
}

\value{
  A two-class random forest fit under the requested method and
  performance value.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{

  Chen, C., Liaw, A. and Breiman, L. (2004). Using random forest to learn
  imbalanced data. University of California, Berkeley, Technical Report
  110.
  
  Kubat, M., Holte, R. and Matwin, S. (1997). Learning when negative
  examples abound. \emph{Machine Learning}, ECML-97: 146-153.

  O'Brien R. and Ishwaran H. (2019).  A random forests quantile
  classifier for class imbalanced data. \emph{Pattern Recognition},
  90, 232-249
  
}
\seealso{
  \command{\link{rfsrc}},
  \command{\link{rfsrc.fast}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## use the breast data for illustration
## ------------------------------------------------------------

data(breast, package = "randomForestSRC")
breast <- na.omit(breast)
f <- as.formula(status ~ .)

##----------------------------------------------------------------
## default RFQ call
##----------------------------------------------------------------

o.rfq <- imbalanced(f, breast)
print(o.rfq)

## equivalent to:
## rfsrc(f, breast, rfq =  TRUE, ntree = 3000,
##       perf.type = "gmean", splitrule = "auc") 

##----------------------------------------------------------------
## detailed output using customized performance function
##----------------------------------------------------------------

print(get.imbalanced.performance(o.rfq))

##-----------------------------------------------------------------
## RF using misclassification error with gini splitting
## ------------------------------------------------------------

o.std <- imbalanced(f, breast, method = "stand", splitrule = "gini")

##-----------------------------------------------------------------
## RF using G-mean performance with AUC splitting
## ------------------------------------------------------------

o.std <- imbalanced(f, breast, method = "stand", perf.type = "gmean")

## equivalent to:
## rfsrc(f, breast, ntree = 3000, perf.type = "gmean", splitrule = "auc")

##----------------------------------------------------------------
## default BRF call 
##----------------------------------------------------------------

o.brf <- imbalanced(f, breast, method = "brf")

## equivalent to:
## imbalanced(f, breast, method = "brf", perf.type = "gmean")

##----------------------------------------------------------------
## BRF call with misclassification performance 
##----------------------------------------------------------------

o.brf <- imbalanced(f, breast, method = "brf", perf.type = "misclass")

##----------------------------------------------------------------
## train/test example
##----------------------------------------------------------------

trn <- sample(1:nrow(breast), size = nrow(breast) / 2)
o.trn <- imbalanced(f, breast[trn,], importance = TRUE)
o.tst <- predict(o.trn, breast[-trn,], importance = TRUE)
print(o.trn)
print(o.tst)
print(100 * cbind(o.trn$impo[, 1], o.tst$impo[, 1]))


##----------------------------------------------------------------
##
##  illustrates how to optimize threshold on training data
##  improves Gmean for RFQ in many situations
##
##----------------------------------------------------------------

if (library("caret", logical.return = TRUE)) {

  ## experimental settings
  n <- 2 * 5000 
  q <- 20
  ir <- 6
  f <- as.formula(Class ~ .)
 
  ## simulate the data, create minority class data
  d <- twoClassSim(n, linearVars = 15, noiseVars = q)
  d$Class <- factor(as.numeric(d$Class) - 1)
  idx.0 <- which(d$Class == 0)
  idx.1 <- sample(which(d$Class == 1), sum(d$Class == 1) / ir , replace = FALSE)
  d <- d[c(idx.0,idx.1),, drop = FALSE]

  ## split data into train and test
  trn.pt <- sample(1:nrow(d), size = nrow(d) / 2)
  trn <- d[trn.pt, ]
  tst <- d[setdiff(1:nrow(d), trn.pt), ]

  ## run rfq on training data
  o <- imbalanced(f, trn)

  ## (1) default threshold (2) directly optimized gmean threshold
  th.1 <- get.imbalanced.performance(o)["threshold"]
  th.2 <- get.imbalanced.optimize(o)["threshold"]

  ## training performance
  cat("-------- train performance ---------\n")
  print(get.imbalanced.performance(o, thresh=th.1))
  print(get.imbalanced.performance(o, thresh=th.2))

  ## test performance
  cat("-------- test performance ---------\n")
  pred.o <- predict(o, tst)
  print(get.imbalanced.performance(pred.o, thresh=th.1))
  print(get.imbalanced.performance(pred.o, thresh=th.2))
 
} 


##----------------------------------------------------------------
##
## illustrates effectiveness of blocked VIMP 
##
##----------------------------------------------------------------

if (library("caret", logical.return = TRUE)) {

  ## experimental settings
  n <- 1000
  q <- 20
  ir <- 6
  f <- as.formula(Class ~ .)
 
  ## simulate the data, create minority class data
  d <- twoClassSim(n, linearVars = 15, noiseVars = q)
  d$Class <- factor(as.numeric(d$Class) - 1)
  idx.0 <- which(d$Class == 0)
  idx.1 <- sample(which(d$Class == 1), sum(d$Class == 1) / ir , replace = FALSE)
  d <- d[c(idx.0,idx.1),, drop = FALSE]

  ## permutation VIMP for BRF with and without blocking
  ## blocked VIMP is a hybrid of Breiman-Cutler/Ishwaran-Kogalur VIMP
  brf <- imbalanced(f, d, method = "brf", importance = "permute", block.size = 1)
  brfB <- imbalanced(f, d, method = "brf", importance = "permute", block.size = 10)

  ## permutation VIMP for RFQ with and without blocking
  rfq <- imbalanced(f, d, importance = "permute", block.size = 1)
  rfqB <- imbalanced(f, d, importance = "permute", block.size = 10)

  ## compare VIMP values
  imp <- 100 * cbind(brf$importance[, 1], brfB$importance[, 1],
                     rfq$importance[, 1], rfqB$importance[, 1])
  legn <- c("BRF", "BRF-block", "RFQ", "RFQ-block")
  colr <- rep(4,20+q)
  colr[1:20] <- 2
  ylim <- range(c(imp))
  nms <- 1:(20+q)
  par(mfrow=c(2,2))
  barplot(imp[,1],col=colr,las=2,main=legn[1],ylim=ylim,names.arg=nms)
  barplot(imp[,2],col=colr,las=2,main=legn[2],ylim=ylim,names.arg=nms)
  barplot(imp[,3],col=colr,las=2,main=legn[3],ylim=ylim,names.arg=nms)
  barplot(imp[,4],col=colr,las=2,main=legn[4],ylim=ylim,names.arg=nms)

}

##----------------------------------------------------------------
##
## confidence intervals for G-mean permutation VIMP using subsampling
##
##----------------------------------------------------------------

if (library("caret", logical.return = TRUE)) {

  ## experimental settings
  n <- 1000
  q <- 20
  ir <- 6
  f <- as.formula(Class ~ .)
 
  ## simulate the data, create minority class data
  d <- twoClassSim(n, linearVars = 15, noiseVars = q)
  d$Class <- factor(as.numeric(d$Class) - 1)
  idx.0 <- which(d$Class == 0)
  idx.1 <- sample(which(d$Class == 1), sum(d$Class == 1) / ir , replace = FALSE)
  d <- d[c(idx.0,idx.1),, drop = FALSE]

  ## RFQ
  o <- imbalanced(Class ~ ., d, importance = "permute", block.size = 10)

  ## subsample RFQ
  smp.o <- subsample(o, B = 100)
  plot(smp.o, cex.axis = .7)

}


}}
\keyword{imbalanced two-class data}

\name{subsample.rfsrc}
\alias{subsample}
\alias{subsample.rfsrc}
\title{Subsample Forests for VIMP Confidence Intervals}

\description{
  Use subsampling to calculate confidence intervals and standard
  errors for VIMP (variable importance). Applies to all families.
}

\usage{\method{subsample}{rfsrc}(obj,
  B = 100,
  block.size = 1,
  importance,
  subratio = NULL,
  stratify = TRUE,
  performance = FALSE,
  performance.only = FALSE,
  joint = FALSE,
  xvar.names = NULL,
  bootstrap = FALSE,
  verbose = TRUE)
}

\arguments{

  \item{obj}{A forest grow object of class \code{(rfsrc, grow)}.}

  \item{B}{Number of subsamples (or bootstrap iterations, if \code{bootstrap = TRUE}).}

  \item{block.size}{Number of trees in each block used when calculating VIMP. If VIMP is already included in the original grow object, that setting is used instead.}

  \item{importance}{Type of variable importance (VIMP) to compute. Choices are \code{"anti"}, \code{"permute"}, or \code{"random"}. If not specified, the default importance setting from the original grow call is used (if available).}

  \item{subratio}{Subsample size as a proportion of the original sample size. The default is approximately the inverse square root of the sample size.}

  \item{stratify}{Logical. If \code{TRUE}, uses stratified subsampling to preserve class balance. See \code{Details} for more information.}

  \item{performance}{Logical. If \code{TRUE}, calculates generalization error along with standard error and confidence intervals.}

  \item{performance.only}{Logical. If \code{TRUE}, only generalization error and its uncertainty are returned; VIMP is not computed.}

  \item{joint}{Logical. If \code{TRUE}, joint VIMP is computed for all variables. To calculate joint VIMP for a subset of variables, use \code{xvar.names}.}

  \item{xvar.names}{Character vector specifying variables to be used for joint VIMP. If omitted, all variables are included.}

  \item{bootstrap}{Logical. If \code{TRUE}, uses the double bootstrap instead of subsampling. This is typically slower but may provide more accurate uncertainty estimates.}

  \item{verbose}{Logical. If \code{TRUE}, prints progress updates during computation.}

}

\details{

This function applies subsampling (or optional double bootstrapping) to a previously trained forest to estimate standard errors and construct confidence intervals for variable importance (VIMP), as described in Ishwaran and Lu (2019). It also supports inference for the out-of-bag (OOB) prediction error via the \code{performance = TRUE} option. Joint VIMP for selected or all variables can be obtained using \code{joint} and \code{xvar.names}.

If the original forest does not include VIMP, it will be computed prior to subsampling. For repeated calls to \code{subsample}, it is recommended that VIMP be requested in the original \code{rfsrc} call. This not only avoids redundant computation, but also ensures consistency of the importance type (e.g., anti, permute, or random) and related parameters, which may otherwise be unclear. Note that permutation importance is \emph{not} the default for most families.

Subsampled forests are constructed using the same tuning parameters as the original forest. While most settings are automatically recovered, certain advanced configurations (e.g., custom sampling schemes) may not be fully supported.

Both subsampled variance estimates (Politis and Romano, 1994) and delete-\(d\) jackknife variance estimates (Shao and Wu, 1989) are returned. The jackknife estimator tends to produce larger standard errors, offering a conservative bias correction, particularly for signal variables.

By default, stratified subsampling is used for classification, survival, and competing risk families:
\itemize{
  \item For classification, strata correspond to class labels.
  \item For survival and competing risks, strata include event type and censoring.
}
Stratification helps ensure representation of key outcome types and is especially important for small sample sizes. Overriding this behavior is discouraged. Note that stratification is \emph{not} available for multivariate families, and caution should be exercised when subsampling in that context.

The function \code{\link{extract.subsample}} can be used to retrieve detailed information from the subsample object. By default, returned VIMP values are standardized: for regression families, VIMP is divided by the variance of the response; for other families, no transformation is applied. To obtain raw (unstandardized) values, set \code{standardize = FALSE}. For expert users, the option \code{raw = TRUE} returns detailed internal output, including VIMP from each individual subsampled forest (constructed on a smaller sample size), which is used internally by \code{\link{plot.subsample.rfsrc}} to generate confidence intervals.

Printed and plotted outputs also standardize VIMP by default. This behavior can be disabled via \code{standardize}. The \code{alpha} option controls the confidence level and is preset in wrapper functions but can be adjusted by the user.
}

\value{

  A list with the following key components:
  \item{rf}{Original forest grow object.}
  \item{vmp}{Variable importance values for grow forest.}
  \item{vmpS}{Variable importance subsampled values.}
  \item{subratio}{Subratio used.}  

}
\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{

  Ishwaran H. and Lu M.  (2019).  Standard errors and confidence
  intervals for variable importance in random forest regression,
  classification, and survival. \emph{Statistics in Medicine}, 38,
  558-582.

  Politis, D.N. and Romano, J.P. (1994). Large sample confidence
  regions based on subsamples under minimal assumptions. \emph{The
  Annals of Statistics}, 22(4):2031-2050.

  Shao, J. and Wu, C.J. (1989). A general theory for jackknife variance
  estimation. \emph{The Annals of Statistics}, 17(3):1176-1197.

}
\seealso{
  \command{\link{holdout.vimp.rfsrc}}
  \command{\link{plot.subsample.rfsrc}},
  \command{\link{rfsrc}},
  \command{\link{vimp.rfsrc}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## regression
## ------------------------------------------------------------

## training the forest
reg.o <- rfsrc(Ozone ~ ., airquality)

## default subsample call
reg.smp.o <- subsample(reg.o)

## plot confidence regions
plot.subsample(reg.smp.o)

## summary of results
print(reg.smp.o)

## joint vimp and confidence region for generalization error
reg.smp.o2 <- subsample(reg.o, performance = TRUE,
           joint = TRUE, xvar.names = c("Day", "Month"))
plot.subsample(reg.smp.o2)

## now try the double bootstrap (slower)
reg.dbs.o <- subsample(reg.o, B = 25, bootstrap = TRUE)
print(reg.dbs.o)
plot.subsample(reg.dbs.o)

## standard error and confidence region for generalization error only
gerror <- subsample(reg.o, performance.only = TRUE)
plot.subsample(gerror)

## ------------------------------------------------------------
## classification
## ------------------------------------------------------------

## 3 non-linear, 15 linear, and 5 noise variables 
if (library("caret", logical.return = TRUE)) {
  d <- twoClassSim(1000, linearVars = 15, noiseVars = 5)

  ## VIMP based on (default) misclassification error
  cls.o <- rfsrc(Class ~ ., d)
  cls.smp.o <- subsample(cls.o, B = 100)
  plot.subsample(cls.smp.o, cex.axis = .7)

  ## same as above, but with VIMP defined using normalized Brier score
  cls.o2 <- rfsrc(Class ~ ., d, perf.type = "brier")
  cls.smp.o2 <- subsample(cls.o2, B = 100)
  plot.subsample(cls.smp.o2, cex.axis = .7)
}

## ------------------------------------------------------------
## class-imbalanced data using RFQ classifier with G-mean VIMP
## ------------------------------------------------------------

if (library("caret", logical.return = TRUE)) {

  ## experimental settings
  n <- 1000
  q <- 20
  ir <- 6
  f <- as.formula(Class ~ .)
 
  ## simulate the data, create minority class data
  d <- twoClassSim(n, linearVars = 15, noiseVars = q)
  d$Class <- factor(as.numeric(d$Class) - 1)
  idx.0 <- which(d$Class == 0)
  idx.1 <- sample(which(d$Class == 1), sum(d$Class == 1) / ir , replace = FALSE)
  d <- d[c(idx.0,idx.1),, drop = FALSE]

  ## RFQ classifier
  oq <- imbalanced(Class ~ ., d, importance = TRUE, block.size = 10)

  ## subsample the RFQ-classifier
  smp.oq <- subsample(oq, B = 100)
  plot.subsample(smp.oq, cex.axis = .7)

}

## ------------------------------------------------------------
## survival 
## ------------------------------------------------------------

data(pbc, package = "randomForestSRC")
srv.o <- rfsrc(Surv(days, status) ~ ., pbc)
srv.smp.o <- subsample(srv.o, B = 100)
plot(srv.smp.o)

## ------------------------------------------------------------
## competing risks
## target event is death (event = 2)
## ------------------------------------------------------------

if (library("survival", logical.return = TRUE)) {
  data(pbc, package = "survival")
  pbc$id <- NULL
  cr.o <- rfsrc(Surv(time, status) ~ ., pbc, splitrule = "logrankCR", cause = 2)
  cr.smp.o <- subsample(cr.o, B = 100)
  plot.subsample(cr.smp.o, target = 2)
}

## ------------------------------------------------------------
## multivariate 
## ------------------------------------------------------------

if (library("mlbench", logical.return = TRUE)) {
  ## simulate the data 
  data(BostonHousing)
  bh <- BostonHousing
  bh$rm <- factor(round(bh$rm))
  o <- rfsrc(cbind(medv, rm) ~ ., bh)
  so <- subsample(o)
  plot.subsample(so)
  plot.subsample(so, m.target = "rm")
  ##generalization error
  gerror <- subsample(o, performance.only = TRUE)
  plot.subsample(gerror, m.target = "medv")
  plot.subsample(gerror, m.target = "rm")
}

## ------------------------------------------------------------
## largish data example - use rfsrc.fast for fast forests
## ------------------------------------------------------------

if (library("caret", logical.return = TRUE)) {
  ## largish data set
  d <- twoClassSim(1000, linearVars = 15, noiseVars = 5)

  ## use a subsampled forest with Brier score performance
  ## remember to set forest=TRUE for rfsrc.fast
  o <- rfsrc.fast(Class ~ ., d, ntree = 100,
           forest = TRUE, perf.type = "brier")
  so <- subsample(o, B = 100)
  plot.subsample(so, cex.axis = .7)
}


}
}
\keyword{subsampling}
\keyword{vimp}
\keyword{confidence interval}

\name{plot.variable.rfsrc}
\alias{plot.variable.rfsrc}
\alias{plot.variable}
\title{Plot Marginal Effect of Variables}

\description{
  Plot the marginal effect of an x-variable on the class probability
  (classification), response (regression), mortality (survival), or the
  expected years lost (competing risk).  Users can select between
  marginal (unadjusted, but fast) and partial plots (adjusted, but
  slower).
}

\usage{\method{plot.variable}{rfsrc}(x, xvar.names, target,
  m.target = NULL, time, surv.type = c("mort", "rel.freq",
  "surv", "years.lost", "cif", "chf"), class.type =
  c("prob", "bayes"), partial = FALSE, oob = TRUE,
  show.plots = TRUE, plots.per.page = 4, granule = 5, sorted = TRUE,
  nvar, npts = 25, smooth.lines = FALSE, subset, ...)
}

\arguments{

  \item{x}{An object of class \code{(rfsrc, grow)}, \code{(rfsrc, synthetic)}, or \code{(rfsrc, plot.variable)}.}

  \item{xvar.names}{Character vector of x-variable names to include. If not specified, all variables are used.}

  \item{target}{For classification, an integer or character specifying the class of interest (default is the first class). For competing risks, an integer between 1 and \code{J} indicating the event of interest, where \code{J} is the number of event types. Default is the first event type.}

  \item{m.target}{Character value for multivariate families specifying the target outcome. If unspecified, a default is automatically chosen.}

  \item{time}{(Survival only) Time point at which the predicted survival value is evaluated, depending on \code{surv.type}.}

  \item{surv.type}{(Survival only) Type of predicted survival value to compute. See \code{\link{plot.variable}} details.}

  \item{class.type}{(Classification only) Type of predicted classification value to use. See \code{\link{plot.variable}} details.}

  \item{partial}{Logical. If \code{TRUE}, partial dependence plots are generated.}

  \item{oob}{Logical. If \code{TRUE}, out-of-bag predictions are used; otherwise, in-bag predictions are used.}

  \item{show.plots}{Logical. If \code{TRUE}, plots are displayed on the screen.}

  \item{plots.per.page}{Integer controlling the number of plots displayed per page.}

  \item{granule}{Integer controlling the coercion of continuous variables to factors (used to generate boxplots). Larger values increase coercion.}

  \item{sorted}{Logical. If \code{TRUE}, variables are sorted by variable importance.}

  \item{nvar}{Number of variables to plot. Defaults to all available variables.}

  \item{npts}{Maximum number of points used when generating partial plots for continuous variables.}

  \item{smooth.lines}{Logical. If \code{TRUE}, applies \code{lowess} smoothing to partial plots.}

  \item{subset}{Vector indicating which rows of \code{x$xvar} to use. Defaults to all rows. Important: do not define \code{subset} based on the original dataset (which may have been altered due to missing data or other processing); define it relative to \code{x$xvar}.}

  \item{...}{Additional arguments passed to or from other methods.}

}

\details{

The vertical axis displays the ensemble-predicted value, while x-variables are plotted along the horizontal axis.

\enumerate{

  \item For regression, the predicted response is plotted.

  \item For classification, the plotted value is the predicted class
  probability for the class specified by \code{target}, or the most
  probable class (Bayes rule) depending on whether \code{class.type} is
  set to \code{"prob"} or \code{"bayes"}.

  \item For multivariate families, the prediction corresponds to the outcome specified by \code{m.target}. If this is a classification outcome, \code{target} may also be used to indicate the class of interest.

  \item For survival, the vertical axis shows the predicted value determined by \code{surv.type}, with the following options:
  
  \itemize{
    \item \code{mort}: Mortality (Ishwaran et al., 2008), interpreted as the expected number of events for an individual with the same covariates.
    
    \item \code{rel.freq}: Relative frequency of mortality.

    \item \code{surv}: Predicted survival probability at a specified time point (default is the median follow-up time), controlled via \code{time}.
  }

  \item For competing risks, the vertical axis shows one of the following quantities, depending on \code{surv.type}:
  
  \itemize{
    \item \code{years.lost}: Expected number of life-years lost.

    \item \code{cif}: Cumulative incidence function for the specified event.

    \item \code{chf}: Cause-specific cumulative hazard function.
  }

  In all competing risks settings, the event of interest is specified using \code{target}, and \code{cif} and \code{chf} are evaluated at the time point given by \code{time}.
}

To generate partial dependence plots, set \code{partial = TRUE}. These differ from marginal plots in that they isolate the effect of a single variable \eqn{X} on the predicted value by averaging over all other covariates:

\deqn{
  \tilde{f}(x) = \frac{1}{n} \sum_{i=1}^n \hat{f}(x, x_{i,o}),
}

where \eqn{x_{i,o}} denotes the observed values of all covariates other than \eqn{X} for individual \eqn{i}, and \eqn{\hat{f}} is the prediction function. Generating partial plots can be computationally expensive; use a smaller value for \code{npts} to reduce the number of grid points evaluated for \eqn{x}.

\emph{Plot display conventions}:

\itemize{
  \item For continuous variables: red points indicate partial values; dashed red lines represent an error band of two standard errors. Black dashed lines show the raw partial values. Use \code{smooth.lines = TRUE} to overlay a \code{lowess} smoothed line.

  \item For discrete (factor) variables: boxplots are used, with whiskers extending approximately two standard errors from the mean.

  \item Standard errors are provided only as rough indicators and should be interpreted cautiously.
}

Partial plots can be slow to compute. Setting \code{npts} to a small value can improve performance.

For additional flexibility and speed, consider using \code{\link{partial.rfsrc}}, which directly computes partial plot data and allows for greater customization.
}

\author{
  Hemant Ishwaran and Udaya B. Kogalur
}
\references{
  Friedman J.H. (2001). Greedy function approximation: a gradient
  boosting machine, \emph{Ann. of Statist.}, 5:1189-1232.
  
  Ishwaran H., Kogalur U.B. (2007).  Random survival forests for R,
  \emph{Rnews}, 7(2):25-31.
  
  Ishwaran H., Kogalur U.B., Blackstone E.H. and Lauer M.S.
  (2008).  Random survival forests, \emph{Ann. App.
    Statist.}, 2:841-860.
  
  Ishwaran H., Gerds T.A., Kogalur U.B., Moore R.D., Gange S.J. and Lau
  B.M. (2014). Random survival forests for competing risks.
  \emph{Biostatistics}, 15(4):757-773.
  
}
\seealso{
  \command{\link{rfsrc}},
  \command{\link{partial.rfsrc}},
  \command{\link{predict.rfsrc}}
}
\examples{
\donttest{
## ------------------------------------------------------------
## survival/competing risk
## ------------------------------------------------------------

## survival
data(veteran, package = "randomForestSRC")
v.obj <- rfsrc(Surv(time,status)~., veteran, ntree = 100)
plot.variable(v.obj, plots.per.page = 3)
plot.variable(v.obj, plots.per.page = 2, xvar.names = c("trt", "karno", "age"))
plot.variable(v.obj, surv.type = "surv", nvar = 1, time = 200)
plot.variable(v.obj, surv.type = "surv", partial = TRUE, smooth.lines = TRUE)
plot.variable(v.obj, surv.type = "rel.freq", partial = TRUE, nvar = 2)

## example of plot.variable calling a pre-processed plot.variable object
p.v <- plot.variable(v.obj, surv.type = "surv", partial = TRUE, smooth.lines = TRUE)
plot.variable(p.v)
p.v$plots.per.page <- 1
p.v$smooth.lines <- FALSE
plot.variable(p.v)

## example using a pre-processed plot.variable to define custom plots
p.v <- plot.variable(v.obj, surv.type = "surv", partial = TRUE, show.plots = FALSE)
plotthis <- p.v$plotthis
plot(plotthis[["age"]], xlab = "age", ylab = "partial effect", type = "b")
boxplot(yhat ~ x, plotthis[["trt"]], xlab = "treatment", ylab = "partial effect")


## competing risks
data(follic, package = "randomForestSRC")
follic.obj <- rfsrc(Surv(time, status) ~ ., follic, nsplit = 3, ntree = 100)
plot.variable(follic.obj, target = 2)

## ------------------------------------------------------------
## regression
## ------------------------------------------------------------

## airquality 
airq.obj <- rfsrc(Ozone ~ ., data = airquality)
plot.variable(airq.obj, partial = TRUE, smooth.lines = TRUE)
plot.variable(airq.obj, partial = TRUE, subset = airq.obj$xvar$Solar.R < 200)

## motor trend cars
mtcars.obj <- rfsrc(mpg ~ ., data = mtcars)
plot.variable(mtcars.obj, partial = TRUE, smooth.lines = TRUE)

## ------------------------------------------------------------
## classification
## ------------------------------------------------------------

## iris
iris.obj <- rfsrc(Species ~., data = iris)
plot.variable(iris.obj, partial = TRUE)

## motor trend cars: predict number of carburetors
mtcars2 <- mtcars
mtcars2$carb <- factor(mtcars2$carb,
   labels = paste("carb", sort(unique(mtcars$carb))))
mtcars2.obj <- rfsrc(carb ~ ., data = mtcars2)
plot.variable(mtcars2.obj, partial = TRUE)

## ------------------------------------------------------------
## multivariate regression
## ------------------------------------------------------------
mtcars.mreg <- rfsrc(Multivar(mpg, cyl) ~., data = mtcars)
plot.variable(mtcars.mreg, m.target = "mpg", partial = TRUE, nvar = 1)
plot.variable(mtcars.mreg, m.target = "cyl", partial = TRUE, nvar = 1)


## ------------------------------------------------------------
## multivariate mixed outcomes
## ------------------------------------------------------------
mtcars2 <- mtcars
mtcars2$carb <- factor(mtcars2$carb)
mtcars2$cyl <- factor(mtcars2$cyl)
mtcars.mix <- rfsrc(Multivar(carb, mpg, cyl) ~ ., data = mtcars2)
plot.variable(mtcars.mix, m.target = "cyl", target = "4", partial = TRUE, nvar = 1)
plot.variable(mtcars.mix, m.target = "cyl", target = 2, partial = TRUE, nvar = 1)


}
}
\keyword{plot}

\name{get.tree.rfsrc}
\alias{get.tree.rfsrc}
\alias{get.tree}
\title{Extract a Single Tree from a Forest and plot it on your browser}
\description{
  Extracts a single tree from a forest which can then be plotted on the users
  browser.  Works for all families.  Missing data not permitted.}
\usage{\method{get.tree}{rfsrc}(object, tree.id, target, m.target = NULL,
   time, surv.type = c("mort", "rel.freq", "surv", "years.lost", "cif", "chf"),
   class.type = c("bayes", "rfq", "prob"),
   ensemble = FALSE, oob = TRUE, show.plots = TRUE, do.trace = FALSE)
}

\arguments{
  \item{object}{An object of class \code{(rfsrc, grow)}.}

  \item{tree.id}{Integer specifying the tree to extract.}

  \item{target}{For classification: integer or character indicating the class of interest (defaults to the first class). For competing risks: integer between 1 and \code{J} (number of event types) specifying the event of interest (default is the first event type).}

  \item{m.target}{Character string specifying the target outcome for multivariate families. If unspecified, a default is selected.}

  \item{time}{For survival: time point at which the predicted value is evaluated (depends on \code{surv.type}).}

  \item{surv.type}{For survival: specifies the type of predicted value returned. See \code{Details}.}

  \item{class.type}{For classification: specifies the type of predicted value. See \code{Details}.}

  \item{ensemble}{Logical. If \code{TRUE}, prediction is based on the ensemble of all trees. If \code{FALSE} (default), prediction is based on the specified tree.}

  \item{oob}{Logical. Use OOB predicted values (\code{TRUE}) or in-bag values (\code{FALSE}). Only applies when \code{ensemble=TRUE}.}

  \item{show.plots}{Logical. Should plots be displayed?}

  \item{do.trace}{Number of seconds between progress updates.}
}

\details{

  Extracts a specified tree from a forest and converts it into a
  hierarchical structure compatible with the \pkg{data.tree}
  package. Plotting the resulting object renders an interactive tree
  visualization in the user's web browser.
  

  Left-hand splits are shown. For continuous variables, the left split
  is displayed as an inequality (e.g., \code{x < value}); the right
  split is the reverse. For factor variables, the left daughter node is
  defined by a set of levels assigned to it; the right daughter is its
  complement.
  

  Terminal nodes are highlighted with color and display both sample size
  and predicted value. By default, the predicted value corresponds to
  the prediction from the selected tree, and the sample size refers to
  the in-bag cases reaching the terminal node. If \code{ensemble =
  TRUE}, the predicted value equals the forest ensemble prediction,
  allowing visualization of the full forest predictor over the selected
  tree's partition. In this case, sample sizes refer to all observations
  (not just in-bag cases).

  Predicted values displayed in terminal nodes are defined as follows:

  \enumerate{
    \item For regression: the mean of the response.

    \item For classification: depends on the \code{class.type} argument and target class:
      \itemize{
        \item If \code{class.type = "bayes"}, the predicted class with the most votes, or the RFQ classifier threshold in two-class problems.
        \item If \code{class.type = "prob"}, the class probability for the target class.
      }

    \item For multivariate families: the predicted value for the outcome specified by \code{m.target}, using the logic above depending on whether the outcome is continuous or categorical.

    \item For survival:
      \itemize{
        \item \code{mort}: estimated mortality (Ishwaran et al., 2008).
        \item \code{rel.freq}: relative frequency of mortality.
        \item \code{surv}: predicted survival probability at the specified time (\code{time}).
      }

    \item For competing risks:
      \itemize{
        \item \code{years.lost}: expected number of life years lost.
        \item \code{cif}: cumulative incidence function.
        \item \code{chf}: cause-specific cumulative hazard function.
      }
      For \code{cif} and \code{chf}, predictions are evaluated at the time point given by \code{time}, and all metrics are specific to the event type indicated by \code{target}.
  }
}


\value{
  Invisibly, returns an object with hierarchical structure formatted for use with
  the data.tree package.
}
\author{
  Hemant Ishwaran and Udaya B. Kogalur

  Many thanks to @dbarg1 on GitHub for the initial prototype of this
  function
  
  
}
\examples{
\donttest{
## ------------------------------------------------------------
## survival/competing risk
## ------------------------------------------------------------

## survival - veteran data set but with factors
## note that diagtime has many levels
data(veteran, package = "randomForestSRC")
vd <- veteran
vd$celltype=factor(vd$celltype)
vd$diagtime=factor(vd$diagtime)
vd.obj <- rfsrc(Surv(time,status)~., vd, ntree = 100, nodesize = 5)
plot(get.tree(vd.obj, 3))

## competing risks
data(follic, package = "randomForestSRC")
follic.obj <- rfsrc(Surv(time, status) ~ ., follic, nsplit = 3, ntree = 100)
plot(get.tree(follic.obj, 2))

## ------------------------------------------------------------
## regression
## ------------------------------------------------------------

airq.obj <- rfsrc(Ozone ~ ., data = airquality)
plot(get.tree(airq.obj, 10))

## ------------------------------------------------------------
## two-class imbalanced data (see imbalanced function)
## ------------------------------------------------------------

data(breast, package = "randomForestSRC")
breast <- na.omit(breast)
f <- as.formula(status ~ .)
breast.obj <- imbalanced(f, breast)

## compare RFQ to Bayes Rule
plot(get.tree(breast.obj, 1, class.type = "rfq", ensemble = TRUE))
plot(get.tree(breast.obj, 1, class.type = "bayes", ensemble = TRUE))

## ------------------------------------------------------------
## classification
## ------------------------------------------------------------

iris.obj <- rfsrc(Species ~., data = iris, nodesize = 10)

## equivalent
plot(get.tree(iris.obj, 25))
plot(get.tree(iris.obj, 25, class.type = "bayes"))

## predicted probability displayed for terminal nodes
plot(get.tree(iris.obj, 25, class.type = "prob", target = "setosa"))
plot(get.tree(iris.obj, 25, class.type = "prob", target = "versicolor"))
plot(get.tree(iris.obj, 25, class.type = "prob", target = "virginica"))


## ------------------------------------------------------------
## multivariate regression
## ------------------------------------------------------------

mtcars.mreg <- rfsrc(Multivar(mpg, cyl) ~., data = mtcars)
plot(get.tree(mtcars.mreg, 10, m.target = "mpg"))
plot(get.tree(mtcars.mreg, 10, m.target = "cyl"))


## ------------------------------------------------------------
## multivariate mixed outcomes
## ------------------------------------------------------------

mtcars2 <- mtcars
mtcars2$carb <- factor(mtcars2$carb)
mtcars2$cyl <- factor(mtcars2$cyl)
mtcars.mix <- rfsrc(Multivar(carb, mpg, cyl) ~ ., data = mtcars2)
plot(get.tree(mtcars.mix, 5, m.target = "cyl"))
plot(get.tree(mtcars.mix, 5, m.target = "carb"))

## ------------------------------------------------------------
## unsupervised analysis
## ------------------------------------------------------------

mtcars.unspv <- rfsrc(data = mtcars)
plot(get.tree(mtcars.unspv, 5))



}
}
\keyword{plot}
